/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */



package lapis;

import java.util.*;
import java.io.*;
import lapis.tc.*;
//import lapisx.io.*;
//import lapisx.net.*;
//import lapisx.util.Str;

public class Arguments implements Enumeration {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    /**
     * Argument list.
     */
    protected Vector argv = new Vector ();

    /**
     * Index of current argument (returned by getElement()).
     */
    protected int currArg = 0;

    protected Object[] originalArgs = null;

    protected Interpreter interpreter;
    protected Document input;
    protected DocumentWriter output;

    /**
     * Make an empty Arguments object.
     */
    public Arguments (Interpreter interpreter) {
        this.interpreter = interpreter;
    }

    /**
     * Add a name/value pair to the argument sequence.
     * @param name Parameter name (may be null)
     * @param value Parameter value (may be null)
     * @return this Arguments object (useful for stacking up add()
     * calls, e.g. new Arguments().add ("X", "5").add("Y", "6");
     */
    public Arguments add (String name, Object value) {
        if (name != null)
            argv.addElement (new Name (name));
        if (value != null)
            argv.addElement (value);
        return this;
    }

    /**
     * Add a parameter name to the argument sequence.
     * @param arg Parameter name
     * @return this Arguments object (useful for stacking up add() calls).
     */
    public Arguments addName (String name) {
        argv.addElement (new Name (name));
        return this;
    }

    /**
     * Add a parameter value to the argument sequence.
     * @param arg Parameter value
     * @return this Arguments object (useful for stacking up add() calls).
     */
    public Arguments addValue (Object arg) {
        argv.addElement (arg);
        return this;
    }

    /**
     * Concatenates another argument sequence to the argument sequence,
     * consuming the other sequence entirely.
     * @param args Arguments to add
     * @return this Arguments object (useful for stacking up add() calls).
     */
    public Arguments addArguments (Arguments args) {
        while (args.hasMoreElements ())
            argv.addElement (args.getElement ());
        return this;
    }

    /**
     * Parameter name.
     */
    protected static class Name {
        String name;
        public Name (String name) {
            this.name = name;
        }
        public String toString () {
            return name;
        }
    }

    //
    // Processing argument list
    // 

    /**
     * Test whether argument enumeration is exhausted.
     * @return true if more arguments to read (so getElement() or nextElement()
     * would succeed), false if not
     */
    public boolean hasMoreElements () {
        return currArg < argv.size();
    }

    /**
     * Get current argument (either name or value).
     * @return argument
     * @exception UsageException if no more arguments
     */
    public Object getElement () {
        if (currArg >= argv.size()) {
            currArg = argv.size();
            usage (); // never returns
            return null;
        }
        Object obj = argv.elementAt (currArg);
        //System.err.println ("getElement(" + currArg + "): " + obj);
        return obj;
    }

    /**
     * Test whether current argument is a name or a value.
     * @return true if name, false if not
     * @exception UsageException if no more arguments
     */
    public boolean isName () {
        Object obj = getElement ();
        return obj instanceof Name;
    }

    /**
     * Get argument (either name or value) and advance to next argument.
     * @return argument
     * @exception UsageException if no more arguments
     */
    public Object nextElement () {
        Object obj = getElement ();
        ++currArg;
        return obj;
    }

    /**
     * Try to get argument as a parameter name and advance to next argument if
     * successful.
     * @return parameter name, or null if current argument is not a name.
     */
    public String nextName () {
        return (isName ()) ? nextElement ().toString () : null;
    }

    /**
     * Try to get argument as a parameter value and advance to next argument if
     * successful.
     * @return parameter value, or null if current argument is not a value.
     */
    public Object nextValue () {
        return (!isName ()) ? nextElement () : null;
    }

    /**
     * Get next argument as a String.
     * @return next argument
     * @exception UsageException if next argument is a parameter name, not 
     * a value
     */
    public String nextString () throws UsageException {
        Object val = nextValue ();
        if (val == null)
            usage ();
        return val.toString ();
    }

    /**
     * Get next argument as an integer.
     * @return integer value of next argument
     * @exception UsageException if next argument is a parameter name, not 
     * a value
     * @exception NumberFormatException if argument isn't an integer
     */
    public int nextInt () throws UsageException, NumberFormatException {
        return Integer.parseInt (nextString ());
    }

    /**
     * Get next argument as a TC pattern.  Argument may be a String, a
     * TC, or a NamedRegionSet.
     * @return TC created from next argument
     * @exception UsageException if next argument is a parameter name, not 
     * a value, or if it's the wrong type
     * @exception TCParseException if argument isn't understood
     * be loaded
     */
    public TC nextPattern () throws UsageException, TCParseException {
        Object obj = nextValue ();

        if (obj instanceof NamedRegionSet)
            return new TC ((NamedRegionSet) obj);

        if (obj instanceof TC)
            return (TC)obj;

        if (obj instanceof String) {
            String expr = (String)obj;
            try {
                return new TC (expr);
            } catch (TCParseException e) {
                e.setStringContext (expr);
                throw e;
            }
        }

        usage ();
        return null;
    }

    /**
     * Handle a parameter common to all LAPIS tools.
     * @param name Parameter name returned by nextName().  No
     * other nextX() call should have been made since the call to nextName(). 
     * @exception UsageException if not a common parameter
     * @exception Exception if some other exception occurs in processing
     * the parameter.
     */
    public void consume (String name) throws Exception {
        if (name == null && input == null)
            setInput (nextValue ());
        else if ("help".equals (name))
            throw new UsageException (usageMessage);
        else
            usage ();
    }

    public Object[] getOriginalArguments () {
        return originalArgs;
    }

    //
    // Getting results of argument parsing  
    //

    /**
     * Get input document specified by arguments.  If no input document
     * was specified in the arguments, then input comes from the current
     * document.
     * <P>Call after
     * processing all arguments (and calling consume() on arguments that
     * the tool doesn't want to handle).
     * @return input document
     */
    public Document getInput () {
        if (input == null)
            input = (Document) interpreter.getVar (Interpreter.DOCUMENT_VAR);
        return input;
    }

    /**
     * Set input document.
     * @param doc Input document
     * @return this Arguments object (useful for stacking up set() and add() calls)
     */
    public Arguments setInput (Document doc) {
        input = doc;
        return this;
    }

    /**
     * Set input document.
     * @param obj Object to convert to an input document
     * @return this Arguments object (useful for stacking up set() and add() calls)
     */
    public Arguments setInput (Object obj) throws IOException {
        return setInput ((obj instanceof Document)
                         ? (Document) obj
                         : getFactory ().make (obj));
    }

    /**
     * Get output stream specified by arguments.  Call only after
     * processing all arguments and calling consume() on arguments that
     * the tool doesn't want to handle.
     * @return DocumentWriter to which tool should send its output
     */
    public DocumentWriter getOutput () {
        if (output == null)
            output = getFactory ().makeWriter ();
        return output;
    }

    /**
     * Set output document.
     * @param doc Output document
     * @return this Arguments object (useful for stacking up set() and add() calls)
     */
    public Arguments setOutput (DocumentWriter out) {
        output = out;
        return this;
    }

    public Interpreter getInterpreter () {
        return interpreter;
    }

    public DocumentFactory getFactory () {
        return (DocumentFactory) interpreter.getVar (Interpreter.FACTORY_VAR);
    }

    //
    // Throwing usage errors
    //

    protected String usageMessage;

    /**
     * Get usage message printed when an erroneous argument is found.
     * Usage message is null by default; must be set by tool before argument
     * processing.
     * @return tool usage message
     */
    public String getUsage () {
        return usageMessage;
    }

    /**
     * Set usage message printed when an erroneous argument is found.
     * Call before argument processing.
     * @return tool usage message
     */
    public void setUsage (String message) {
        usageMessage = message;
    }

    /**
     * Fire a UsageException containing the usage message set by setUsage().
     * Never returns.
     * @exception UsageException thrown always
     */ 
    public void usage () {
        usage (usageMessage);
    }

    /**
     * Fire a UsageException containing the given message.
     * Never returns.
     * @param message Usage message to include in thrown UsageException
     * @exception UsageException thrown always
     */ 
    public void usage (String message) {
        String context = getContextMessage ();
        throw new UsageException ((context != null ? context : "")
                                  + (message != null ? message : ""));
    }

    /**
     * Exception thrown when argument processing encounters an error.
     * Main program should catch this exception and deliver its message
     * (getMessage()) to the user.
     */ 
    public static class UsageException extends RuntimeException {
        public UsageException (String message) {
            super (message);
        }
    }

    /**
     * Get a representation of the command line with the current argument
     * position marked in it, to be printed with the usage message. 
     * @return command line (a string ending with a newline)
     */
    protected String getContextMessage () {
        if (currArg == 0)
            return "\n";

        String result = "";

        int n = Math.min (argv.size(), currArg-1);
        for (int i = 0; i < n; ++i)
            result += argv.elementAt (i) + " ";

        result += ">>>";
        if (currArg-1 < argv.size())
            result += argv.elementAt (currArg-1);
        result += "<<<\n";

        return result;
    }
}
