/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */



package lapis;
import java.io.*;
import java.net.*;
import java.util.*;

public class BasicDocument implements Document, DocumentView {
    String content;     // if non-null, already-downloaded content 
    Reader in;          // if non-null, stream that will produce 
                        // this document's content
    Hashtable metadata;
    PatternCache patternCache;

    DocumentFactory factory;

    //store version
    DocumentVersion version = null;

	/*
	DocumentView view = new DocumentView(this) {
		public String getText () {
			if (in != null)
				load ();
			return content;
		}

		public String getText (Region r) {
			if (in != null)
				load ();
			int l = content.length ();
			int s = Math.min (l, Math.max (0, r.getStart ()));
			int e = Math.min (l, Math.max (0, r.getEnd ()));
			return (s < e) ? content.substring (s, e) : "";
		}

		public Region toRegion () {
			if (in != null)
				load ();
			return new Region (0, 
							   content != null ? content.length () : 0,
							   this.version);
		}
	};
	*/

    /**
      * Factory for creating SwingDocuments.
      */

    public static class Factory extends DocumentFactory {
        public Document make (Reader in, String contentType) throws IOException {
            return new BasicDocument (in, contentType);
        }

        public Document make (String content, String contentType) {
	    return new BasicDocument (content, contentType);
        }

        public Document make (Document src, String contentType) {
            return new BasicDocument (src, contentType);
        }
        
    }

    /**
     * Factory for creating SwingDocuments.
     */
    public static final DocumentFactory defaultFactory = new Factory ();

    public BasicDocument (Reader in, String contentType) {
	this.factory = defaultFactory;
	this.in = in;
	if (contentType != null)
	    putProperty (MIMEProperty, contentType);
    }
    public BasicDocument (String content, String contentType) {
	this.factory = defaultFactory;
	this.content = content;
	if (contentType != null)
	    putProperty (MIMEProperty, contentType);	
    }
    public BasicDocument (Document src, String contentType) {
	this.factory = defaultFactory;
	this.content = src.getText ();
	this.copyMissingProperties (this, src);
	if (contentType != null)
	    putProperty (MIMEProperty, contentType);
    }
    /*
    public BasicDocument (String content, Document lastDoc, 
			  CoordinateMap map, String contentType) {
	this.in = null;
	this.factory = defaultFactory;
	this.content = content;
	this.version = new DocumentVersion(map, lastDoc.getVersion(), this);
	if (contentType != null)
	    putProperty (MIMEProperty, contentType);
    }
	*/
    public String getText () {
        if (in != null)
            load ();
        return content;
    }

    public String getText (Region r) {
        if (in != null)
            load ();
        int l = content.length ();
        int s = Math.min (l, Math.max (0, r.getStart ()));
        int e = Math.min (l, Math.max (0, r.getEnd ()));
        return (s < e) ? content.substring (s, e) : "";
    }

    public Region toRegion () {
        if (in != null)
            load ();
        return new Region (0, 
        				   content != null ? content.length () : 0,
        				   getVersion());
    }

    public Reader getReader () {    	
        return new StringReader (getText ());
    }

    public Dictionary getProperties () {
        if (metadata == null)
            metadata = new Hashtable ();
        return metadata;
    }

    public Object getProperty (Object key) {
        return getProperties ().get (key);
    }

    public void putProperty (Object key, Object value) {
        if (value == null)
            getProperties ().remove (key);
        else
            getProperties ().put (key, value);
    }

    public PatternCache getPatternCache () {
        if (patternCache == null)
            patternCache = new PatternCache ();
        return patternCache;
    }

    public DocumentFactory getFactory () {
        return factory;
    }

    public DocumentVersion getVersion () {
		return (version != null) 
				? version 
				: DocumentVersion.getOriginal((DocumentView)this); // not editable
    }

    static final int BUFFER_SIZE = 10240;
    protected void load () {
        if (in == null)
            return;

        Reader in = this.in;
        this.in = null;

        int n;
        StringBuffer contentBuf = new StringBuffer ();
        
        Main.getProgress ().fireStartProgress ("Loading document");

        try {
            char[] buf = new char[BUFFER_SIZE];
            while ((n = in.read (buf)) != -1) {
                appendSansCRs (contentBuf, buf, 0, n);
                Main.getProgress ().fireProgress ();
            }
            in.close ();
        } catch (IOException e) {
            throw new RuntimeException (e.toString ());
        } finally {
            Main.getProgress ().fireDoneProgress ();
        }

        content = contentBuf.toString ();
    }

    private void appendSansCRs (StringBuffer dst, char[] src, int offset, int len) {
        int start=offset, end=offset+len;
        for (int i = start; i < end; ++i) {
            if (src[i] == '\r') {
                if (start < i)
                    dst.append (src, start, i-start);
                start = i+1;
            }
        }
        if (start < end)
            dst.append (src, start, end-start);
    }

    /**
      * Copy content-type and base properties from on document to another.
      * @param target Target document
      * @param source Source document
      */
    public static void copyVitalProperties (Document target, Document source) {
        URL base = (URL) source.getProperty (BaseProperty);
        if (base == null)
            base = (URL) source.getProperty (URLProperty);
        target.putProperty (BaseProperty, base);

        target.putProperty (MIMEProperty, source.getProperty (MIMEProperty));
    }

    /**
      * Copy missing metadata from one document to another.
      * For every property of source that target lacks,
      * this method copies the property to target.
      * @param target Target document
      * @param source Source document
      */
    public static void copyMissingProperties (Document target, Document source) {
        Dictionary sourceMetadata = source.getProperties ();
        Dictionary targetProperties = target.getProperties ();
        for (Enumeration e = sourceMetadata.keys (); e.hasMoreElements (); ) {
            Object key = e.nextElement ();
            //if (targetProperties.get (key) == null
            //    && !CookedProperty.equals (key))
            //    targetProperties.put (key, sourceMetadata.get (key));
            if (targetProperties.get (key) == null)
            	targetProperties.put (key, sourceMetadata.get (key));
        }
    }

    public String toString () {    	
        String cmd = (String) getProperty (CommandProperty);
        if (cmd != null && cmd.length () > 0)
            return cmd;

        URL url = (URL) getProperty (URLProperty);
        if (url != null)
            return url.toString ();

        return lapisx.util.Str.abbreviate (getText (), 70);
    }
    
    public DocumentView getDefaultView () {
    	return this;
    }
    
    public DocumentView getCookedView () {
    	return this;
    }
    
    public Document getDocument () {
    	return this;
    }
}
