/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */



package lapis;
import java.io.*;
import java.net.*;
import lapisx.net.*;

public abstract class DocumentFactory {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    public abstract Document make (Reader in, String contentType) throws IOException;
    public abstract Document make (String content, String contentType);
    public abstract Document make (Document src, String contentType);

    public Document make (Object obj) throws IOException {
        if (obj instanceof String)
            return make ((String)obj);
        else if (obj instanceof File)
            return make ((File)obj);
        else if (obj instanceof URL)
            return make ((URL)obj);
        else if (obj instanceof FormURL)
            return make ((FormURL)obj);
        else if (obj instanceof URLConnection)
            return make ((URLConnection)obj);
        else if (obj instanceof InputStream)
            return make ((InputStream)obj);
        else if (obj instanceof Reader)
            return make ((Reader)obj);
        else if (obj instanceof Document)
            return make ((Document)obj);
        else
            throw new RuntimeException ("Don't know how to convert " 
                                        + obj.getClass ().getName ()
                                        + " object into Document");
    }
    
    public Document make (Document doc) {
        return make (doc, (String) doc.getProperty (Document.MIMEProperty));
    }

    public Document make (File file) throws IOException {
        return make (URLUtil.FileToURL (file));
    }

    public Document make (URL url) throws IOException {
        return make (url.openConnection ());
    }

    public Document make (FormURL url) throws IOException  {
        return make (url.openConnection ());
    }

    public Document make (URLConnection conn) throws IOException {
        String contentType = conn.getContentType ();

        Reader in = new InputStreamReader (conn.getInputStream ());
        debug.println ("reported content type is " + contentType);
        if (contentType == null
            || "content/unknown".equals (contentType)) {
            ContentTypeGuessingReader ctgr = new ContentTypeGuessingReader (in);
            contentType = ctgr.getContentType ();
            debug.println ("guessed content type is " + contentType);
            in = ctgr;
        }

        Document doc = make (in, contentType);
        doc.putProperty (Document.URLProperty, conn.getURL ());
        return doc;
    }

    public Document make (InputStream in, String contentType) throws IOException {
        return make (new InputStreamReader (in), contentType);
    }

    public Document make (InputStream in) throws IOException {
        return make (new InputStreamReader (in));
    }

    public Document make (Reader in) throws IOException {
        ContentTypeGuessingReader ctgr = new ContentTypeGuessingReader (in);
        return make (ctgr, ctgr.getContentType ());
        //return make (in, "text/plain");
    }

    public Document make (String content) {
        try {
            ContentTypeGuessingReader ctgr = new ContentTypeGuessingReader (new StringReader (content));
            String contentType = ctgr.getContentType ();
            ctgr.close ();
            return make (content, contentType);
        } catch (IOException e) {
            throw new RuntimeException (e.toString ()); // shouldn't happen
        }
    }

    public DocumentWriter makeWriter () {
        return new DocumentWriter (this);
    }
}

