/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis;

import lapis.maps.*;
import lapisx.util.Debug;
import java.util.*;

public class DocumentVersion {

	public static Debug debug = Debug.QUIET;

	CoordinateMap change;
	DocumentVersion lastVersion;
	DocumentView view;
	int num;

	public DocumentVersion(
		CoordinateMap change,
		DocumentVersion lastVersion,
		DocumentView view) {
		this.change = change;
		this.lastVersion = lastVersion;
		this.view = view;
		num = (lastVersion != null) ? lastVersion.num + 1 : 0;
	}

	public CoordinateMap diff(DocumentVersion ver) {
		debug.assertion(this.getDocument() == ver.getDocument());

		// debug
		CoordinateMap map = diff3(ver);
		return map;
	}
	private CoordinateMap diff1(DocumentVersion ver) {
		if (num == ver.num)
			return IdentityMap.IDENTITY;
		else if (lastVersion == ver)
			return change;
		else if (num < ver.num)
			return ver.diff(this).invert();
		else
			return new CompositeCoordinateMap(change, lastVersion.diff(ver));
	}

	private CoordinateMap diff3(DocumentVersion ver) {

		debug.println("this ->" + this);
		debug.println("ver  ->" + ver);

		debug.assertion(this.getDocument() == ver.getDocument());
		if ((this.num == ver.num) && (this.view == ver.view))
			return IdentityMap.IDENTITY;

		DocumentVersion lcaVersion = null;
		if (this.num < ver.num) {
			// find the least-common-ancestor in the DocumentVersion tree			
			lcaVersion = findLeastCommonAncestor(ver, this);

			debug.println("LCA  ->" + lcaVersion);

			if (lcaVersion == null)
				return IdentityMap.IDENTITY;

			if (lcaVersion.equals(this)) {
				debug.println("LCA is this");
				return ver.diffToAncestor(this).invert();
			}
		} else {
			// find the least-common-ancestor in the DocumentVersion tree
			lcaVersion = findLeastCommonAncestor(this, ver);

			debug.println("LCA ->" + lcaVersion);

			if (lcaVersion == null)
				return IdentityMap.IDENTITY;

			if (lcaVersion.equals(ver)) {
				debug.println("LCA is ver");
				return this.diffToAncestor(ver);
			}
		}
		return new CompositeCoordinateMap(
			this.diffToAncestor(lcaVersion),
			ver.diffToAncestor(lcaVersion).invert());
	}

	private DocumentVersion findLeastCommonAncestor(
		DocumentVersion vl,
		DocumentVersion vs) {
		HashSet path = new HashSet();
		path.add(vl);
		while (true) {
			vl = vl.lastVersion;
			if (vl == null)
				break;
			if (vl.equals(vs))
				return vs;
			path.add(vl);
		}
		if (path.contains(vs))
			return vs;
		while (true) {
			vs = vs.lastVersion;
			if (vs == null)
				break;
			if (path.contains(vs))
				return vs;
		}

		return null;
	}

	private CoordinateMap diffToAncestor(DocumentVersion ancestor) {
		if (this.equals(ancestor))
			return IdentityMap.IDENTITY;
		else if (this.lastVersion.equals(ancestor))
			return this.change;
		else
			return new CompositeCoordinateMap(
				this.change,
				this.lastVersion.diffToAncestor(ancestor));
	}

	// In current implementation, lessThan has only one client call (from SimulEdit.undoTo), 
	// which compares versions from the same view.
	// Therefore, num < ver.num should work.
	public boolean lessThan(DocumentVersion ver) {
		debug.assertion(this.getDocument() == ver.getDocument());
		return num < ver.num;
		//return this.toRaw().num < ver.toRaw().num;
	}

	// In current implementation, greaterThan has no client call at all.
	// By the way, greaterThan can be implemented by lessThan.
	// Therefore, no independent implementation is needed.
	/*
	public boolean greaterThan (DocumentVersion ver) {
	debug.assertion(this.getDocumentView() == ver.getDocument());
	if (ver == null)
	    return true;
	return num > ver.num;
	//return this.toRaw().num > ver.toRaw().num;
	}
	*/

	public String toString() {
		return "v" + num + "-view" + view;
	}

	public RegionSet convert(SingleDocRegionSet s) {
		if (((RegionSet) s).isEmpty())
			return (RegionSet) s;
		DocumentVersion ver = s.getVersion();
		if (this.equals(ver))
			return (RegionSet) s;
		else
			return s.convert(this.diff(ver), this);
	}

	public static DocumentVersion getOriginal(Document doc) {
		return new DocumentVersion(
			IdentityMap.IDENTITY,
			null,
			doc.getDefaultView());
	}

	public static DocumentVersion getOriginal(DocumentView view) {
		return new DocumentVersion(IdentityMap.IDENTITY, null, view);
	}

	public Document getDocument() {
		return this.view.getDocument();
	}

	public DocumentView getView() {
		return this.view;
	}

	// override equals and hashCode
	public boolean equals(Object obj) {
		if (!(obj instanceof DocumentVersion))
			return false;
		else {
			DocumentVersion ver = (DocumentVersion) obj;
			return (
				(this.change == ver.change)
					&& (this.lastVersion == ver.lastVersion)
					&& (this.num == ver.num)
					&& (this.view == ver.view));
		}
	}

	public int hashCode() {
		return ((this.change != null) ? this.change.hashCode() : 0)
			+ ((this.lastVersion != null) ? this.lastVersion.hashCode() : 0)
			+ ((this.view != null) ? this.view.hashCode() : 0)
			+ this.num;
	}
}
