/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */



package lapis;

import java.io.*;
import java.util.*;
import lapis.tools.*;
import lapisx.util.*;
import lapisx.progress.*;
import lapisx.net.URLUtil;
import lapis.tc.TCParseException;

public class Main {
    public static Debug debug = Debug.QUIET;

    static Config config;
    static File lapisDir;
    static Interpreter interp;
    static Hashtable threadProgress = new Hashtable (); // maps Thread to BasicProgressGenerator
    static BasicProgressGenerator defaultProgress = new BasicProgressGenerator ("");

    static final String USAGE = 
         "INTERACTIVE MODES\n"
        +"  lapis [-gui] [file/URL]\n"
        +"     GUI mode. Starts LAPIS web browser showing <file/URL> if given,\n"
        +"     otherwise LAPIS homepage. -gui is the default mode when no -e or -f\n"
        +"     is given.\n"
        +"\n"
        +"  lapis -gui [-e <command> | -f <script>] [file/URL]\n"
        +"     GUI mode. Starts LAPIS web browser showing optional <file/URL>,\n"
        +"     then runs script command <command> or script file <script> to\n"
        +"     initialize it. -gui is required here.\n"
        +"\n"
        +"  lapis -tty [-e <command> | -f <script>] [file/URL]\n"
        +"     TTY mode. Starts text shell, loads optional file/URL, then runs\n"
        +"     optional <command> or <script> to initialize it.\n"
        +"\n"
        +"NONINTERACTIVE MODES\n"
        +"  lapis [-pipe] -e <command> <file/URL>*\n"
        +"  lapis [-pipe] -f <script> <file/URL>*\n"
        +"     Pipe mode.  Runs <command> or <script> on every file/URL and prints.\n"
        +"     result.  -pipe is the default mode when -e or -f is specified.\n"
        +"\n"
        +"  lapis -inplace -e <command> <file>*\n"
        +"  lapis -inplace -f <script> <file>*\n"
        +"     In-place editing mode.  Runs <command> or <script> on every file, and\n"
        +"     saves result back to file.  Inplace editing of URLs not supported.\n"
        +"\n"
        +"  lapis -batch -e <command> <file/URL>*\n"
        +"  lapis -batch -f <script> <file/URL>*\n"
        +"     Batch mode.  Runs <command> or <script> on every file/URL, quietly.\n"
        +"     Script must do its own output, e.g. with puts, show, or save.\n"
        +"\n"
        +"  lapis -argv -e <command> <arg>*\n"
        +"  lapis -argv -f <script> <arg>*\n"
        +"     Argv mode.  Runs <command> or <script>, initializing argv to <arg>*.\n"
        +"     Script must parse its own arguments and generate its own output.\n"
        +"\n"
        +"Modes can be abbreviated: -g[ui] -t[ty] -p[ipe] -i[nplace] -b[atch] -a[rgv]\n"
        +"\n"
        +"Options for all modes:\n"
        +"    --                 Treat remaining arguments as files or script args\n"
        +"    -debug <classes>   Verbose debugging output for <classes>, a list of\n"
        +"                       Java class names separated by spaces or commas:\n"
        +"                           e.g. -debug lapis.swing.Browser;lapis.Main\n"
        +"    -help              Display this message\n"
        ;

    public static void main (String[] argv) throws Exception {
        try {
            String mode = null;
            String script = null;
            boolean isExpr = true;
            Vector scriptArgs = new Vector ();

            // handle arguments related to global configuration;
            // put other arguments in v
            for (int i = 0; i < argv.length; ++i) {
                String str = argv[i];
		if (str.equals ("-"))
		    scriptArgs.addElement (str);
		else if (str.startsWith ("-")) {
                    if ("-gui".startsWith (str))
                        mode = "gui";
                    else if ("-tty".startsWith (str))
                        mode = "tty";
                    else if ("-pipe".startsWith (str))
                        mode = "pipe";
                    else if ("-inplace".startsWith (str))
                        mode = "inplace";
                    else if ("-batch".startsWith (str))
                        mode = "batch";
                    else if ("-argv".startsWith (str))
                        mode = "argv";
                    else if (str.equals ("-e")) {
                        if (script != null
                            || ++i == argv.length)
                            throw new Arguments.UsageException (USAGE);
                        script = argv[i];
                        isExpr = true;
                    } else if (str.equals ("-f")) {
                        if (script != null
                            || ++i == argv.length)
                            throw new Arguments.UsageException (USAGE);
                        script = argv[i];
                        isExpr = false;
                    } else if (str.equals ("--")) {
                        for (++i; i < argv.length; ++i)
                            scriptArgs.addElement (argv[i]);
                    } else if (str.equals ("-config")) {
                        if (++i == argv.length)
                            throw new Arguments.UsageException (USAGE);
                        config = new Config (new File (argv[i]), getConfig ());
                    } else if (str.equals ("-debug")) {
                        if (++i == argv.length)
                            throw new Arguments.UsageException (USAGE);
                        enableDebugging (argv[i]);
                    } else
                        throw new Arguments.UsageException (USAGE);
                } else
		    scriptArgs.addElement (argv[i]);
            }

            // Initialize.  We delay doing this until after processing
            // the arguments so that -config has a chance to change
            // the location of the configuration file before we load it.
            init ();

            // handle the selected mode
            if (mode == null)
                mode = (script != null) ? "pipe" : "gui";

            if (mode.equals ("gui") || mode.equals ("tty")) {
                if (scriptArgs.size () > 1)
                    throw new Arguments.UsageException ("Too many files specified:\n\t"
                                                        + scriptArgs
                                                        + "\nAt most one file can be loaded from the command line.");
		printSplashText();
		
                // start interactive shell
 		interp.setVar (Interpreter.INTERACTIVE_VAR, "1");

                // create arguments
                Arguments args = new Arguments (interp);
                debug.println ("script: " + script);
                debug.println ("args: " + scriptArgs);
                if (script != null)
                    args.add (isExpr ? "e" : "f", script);
                for (Enumeration e = scriptArgs.elements (); e.hasMoreElements (); )
                    args.add ("href", e.nextElement ());
                
                // create the shell
                String shellClass = getConfig ().getProperty ("shell." + mode);
                debug.println ("creating " + shellClass);
                Tool shell = (Tool) Class.forName (shellClass).newInstance ();
            
                // run the shell
                debug.println ("starting " + shellClass);
                shell.invoke (args);

            } else {
                // must have a script
                if (script == null)
                    throw new Arguments.UsageException ("-" + mode + 
                                                        " option must be accompanied by -e <command> or -f <file>");
                    
                // run script in batch mode
 		interp.setVar (Interpreter.INTERACTIVE_VAR, "0");
                
                if (mode.equals ("argv")) {
                    // scriptArgs are arguments handled by script
                    String[] scriptArgv = new String[scriptArgs.size ()];
                    scriptArgs.copyInto (scriptArgv);
                    interp.setVar (Interpreter.ARGUMENTS_VAR, scriptArgv);

                    if (isExpr)
                        interp.eval (script);
                    else
                        interp.evalFile (script);

                } else {
                    // scriptArgs are filenames

                    // if no scriptArgs, then assume stdin
                    if (scriptArgs.isEmpty ())
                        scriptArgs.add ("-");

                    Tool show = mode.equals ("pipe")
                        ? new lapis.tools.Show ()
                        : null;
                    Tool save = mode.equals ("inplace")
                        ? new lapis.tools.Save ()
                        : null;

                    for (Enumeration e = scriptArgs.elements (); e.hasMoreElements (); ) {
                        String href = (String) e.nextElement ();
                        Document input = "-".equals (href)
                            ? BasicDocument.defaultFactory.make (System.in)
                            : BasicDocument.defaultFactory.make (URLUtil.HrefToURL (href));
                        input.putProperty (Document.CommandProperty, href);
                        interp.setVar (Interpreter.DOCUMENT_VAR, input);
                        
                        if (isExpr)
                            interp.eval (script);
                        else
                            interp.evalFile (script);

                        if (save != null) {
                            if ("-".equals (href))
                                new lapis.tools.Show ()
                                    .invoke (new Arguments (interp));
                            else
                                save.invoke (new Arguments (interp)
                                             .add ("filename", href));
                        }

                        if (show != null)
                            show.invoke (new Arguments (interp));
                    }
                }
            }

        } catch (Arguments.UsageException e) {
            System.err.println (e.getMessage ());
        } catch (TCParseException e) {
            e.printErrors (new PrintWriter (System.err));
        } catch (tcl.lang.TclException e) {
            System.err.println (e.getMessage ());
        }
    }

    
    //
    // Global state
    //

    /**
     * Initializes LAPIS.  If you started LAPIS through a method other
     * than {@link lapis.Main#main}, then you need to call this method
     * so that LAPIS will do its standard startup activities, such as
     * loading lib/properties and ~/.lapis, loading init.tcl, and
     * populating the pattern library.
     */
    public static void init () {
		// don't do anything in here;
		// the initialization is actually done
		// by the static block below
    }
    
    static {      
		// turn on debugging for classes listed in debug property
		enableDebugging (getConfig ().getProperty ("debug", ""));

		// set HTTP proxy
		setProxy (getConfig ().getProperty ("http.proxy", ""));

		// set HTTP user agent
		String newUserAgent = Main.getConfig ().getProperty ("http.user-agent");
		if (newUserAgent != null)
			setUserAgent (newUserAgent);

		// touch the pattern library so that PatternLibrary gets loaded
		// by default class loader instead of by a class loaded 
		// dynamically.  Otherwise we may end up with PatternLibrary
		// loaded more than once, and hence multiple copies of its
		// static variables.
		PatternLibrary.getGlobalLibrary ();

		// create the first interpreter, which loads init.tcl
		getInterpreter ();
    }

    public static Config getConfig () {
        if (config == null) {
            Config defaults;
            try {
            	File file = URLUtil.HrefToFile ("lib/properties", getLapisDir ());
                defaults = new Config (file);
                if (defaults.getLastException () != null)
                	throw defaults.getLastException ();
            } catch (IOException e) {
                debug.report (e);
                defaults = null;
            }
	    File testFile = new File(Config.getHomeDirectory (), ".lapis");
	    if(testFile.exists()) {
		File replaceFile = new File(Config.getHomeDirectory (), ".lapis/config");
		if(!testFile.isDirectory()) {
		    config = new Config (testFile, defaults);
		    config.rename(replaceFile);
		    config.save();
		    return config;
		}
		else if(replaceFile.exists()&&replaceFile.isDirectory())
		    replaceFile.delete();
	    }
	    config = new Config (".lapis/config", defaults);
        }
        return config;
    }

    public static File getLapisDir () {
        if (lapisDir == null)
            try {
            	String dir = System.getProperty ("lapis.dir");
                if (dir != null)
                	setLapisDir (dir);
            } catch (IOException e) {
                debug.report (e);
            }
            
		if (lapisDir == null)
            try {
            	setLapisDir (".");
            } catch (IOException e2) {
            	debug.report (e2);
            }

        return lapisDir;
    }

    public static void setLapisDir (String dir) throws IOException {
        lapisDir =new File (dir).getCanonicalFile ();
    }

    /**
     * Get global interpreter.  This is the interpreter of
     * the Browser or Shell created by {@link lapis.Main#main}.  Subsequent
     * creations of Browser windows use fresh interpreters.
     * @return Interpreter object created by main()
     */
    public static Interpreter getInterpreter () {
        if (interp == null)
            try {
                interp = makeInterpreter ();
            } catch (InstantiationException e) {
                throw new RuntimeException (e.toString ());
            } catch (IllegalAccessException e) {
                throw new RuntimeException (e.toString ());
            } catch (ClassNotFoundException e) {
                throw new RuntimeException (e.toString ());
            }
        return interp;
    }

    /**
     * Make an interpreter.  Creates an instance of the class named
     * by the "interpreter" variable in lib/properties (which may be
     * overridden in ~/.lapis).
     * @return new Interpreter object
     */
    public static Interpreter makeInterpreter () throws InstantiationException, IllegalAccessException, ClassNotFoundException {
        String interpreterClass = getConfig ().getProperty ("interpreter");
        Interpreter interp =
            (Interpreter) Class.forName (interpreterClass).newInstance ();
        debug.println ("created " + interpreterClass);
        return interp;
    }


    public static BasicProgressGenerator getProgress () {
        BasicProgressGenerator pg = (BasicProgressGenerator) threadProgress.get (Thread.currentThread ());
        return (pg != null) ? pg : defaultProgress;
    }

    public static void addProgress (BasicProgressGenerator progress) {
        threadProgress.put (Thread.currentThread (), progress);
    }

    public static void removeProgress (BasicProgressGenerator progress) {
        threadProgress.remove (Thread.currentThread ());
    }

    public static void enableDebugging (String classes) {
        StringTokenizer in = new StringTokenizer (classes, " ;,");
        while (in.hasMoreTokens ()) {
            String className = in.nextToken ();
            try {
                Debug.setDebugLevel (Class.forName (className),
                                     Debug.VERBOSE);
            } catch (ClassNotFoundException e) {
                debug.report (e);
                System.err.println ("can't enable debugging on " 
                                    + className
                                    + ": class not found");
            } catch (NoSuchFieldException e) {
                debug.report (e);
                System.err.println ("can't enable debugging on " 
                                    + className
                                    + ": no debug field found");
            }
        }
    }

    public static void setProxy (String proxy) {
        Properties props = System.getProperties ();
        if (proxy == null || proxy.length () == 0) {
            props.remove ("http.proxyHost");
            props.remove ("http.proxyPort");
        } else if (proxy.indexOf (':') == -1)
            debug.println ("misformatted http.proxy property in .lapis: should take the form <host>:<port>");
        else {
            props.put ("http.proxyHost", Str.before (proxy, ":"));
            props.put ("http.proxyPort", Str.after (proxy, ":"));
        }
        System.setProperties (props);
    }

    public static void setUserAgent (String userAgent) {
        System.setProperty ("http.agent", userAgent);
    }

    private static void printSplashText() {
      String splashText = Main.getConfig().getProperty("lapis.splash-text");
      splashText = splashText.replace('|', '\n');
      System.err.println(splashText);
    }


    //
    // Managing user interfaces
    //

    private static Vector windows = new Vector ();
    // interpreter windows (includes Browsers and Shells)

    public static int getWindowCount () {
        return windows.size ();
    }

    public static Enumeration windows () {
        return windows.elements ();
    }

    public static void windowCreated (Object window) {
        if (!windows.contains (window))
            windows.addElement (window);
    }

    public static void windowDisposed (Object window) {
        windows.removeElement (window);
        if (windows.isEmpty ()) {
	    LibraryLoader.writeToFile();
            System.exit (0);
	}
    }


}
