/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis;

import lapis.trees.*;

public class MutableRegionSet extends RegionSet {

	RegionSet set;

	// keep the original constructors
	// DocumentVersion updated or checked by insert and delete
	public MutableRegionSet() {
		set = EMPTY;
	}

	public MutableRegionSet(int size) {
		set = new RegionArray(size);
	}

	public MutableRegionSet(RegionSet s) {
        if (s instanceof Mutable)
            set = s;
        else
		  set = s.freeze();
	}

	public static MutableRegionSet makeFlattener(boolean melt, Document doc) {
		return new MutableRegionSet(new FlatteningRegionArray(melt, doc));
	}
	public static MutableRegionSet makeFlattener(
		boolean melt,
		DocumentVersion ver) {
		return new MutableRegionSet(new FlatteningRegionArray(melt, ver));
	}
	public static MutableRegionSet makeFlattener(
		int size,
		boolean melt,
		Document doc) {
		return new MutableRegionSet(new FlatteningRegionArray(size, melt, doc));
	}
	public static MutableRegionSet makeFlattener(
		int size,
		boolean melt,
		DocumentVersion ver) {
		return new MutableRegionSet(new FlatteningRegionArray(size, melt, ver));
	}

	public String toString() {
		do {
			try {
				return "*" + set.toString();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	//
	// size of region set
	//

	public int getSize() {
		do {
			try {
				return set.getSize();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	//
	// inserting and deleting regions
	//

	/**
	 * Insert a new region s.
	 * @param s Region to add
	 */
	public void insert(Region s) {
		if (s.isEmpty());
		else if (set == EMPTY) {
			set = s.freeze();
		} else {
			DocumentVersion ver = s.getVersion();
			makeMutable();

			if (needToMakeMultiDocSupport(ver))
				makeMultiDocSupport();

			try {
				((Mutable) set).insert(s);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		}
	}

	public void insert(int start, int end, Document doc) {
		this.insert(start, end, doc.getVersion());
	}

	public void insert(int start, int end, DocumentVersion ver) {
		makeMutable();

		if (needToMakeMultiDocSupport(ver))
			makeMultiDocSupport();

		try {
			((Mutable) set).insert(start, end, ver);
		} catch (RepresentationChangeException e) {
			if (set instanceof MultiDocRegionSet) {
				RegionSet mr = e.getRepresentation();
				((MultiDocRegionSet) set).replace(mr);
			} else {
				set = e.getRepresentation();
			}
		}
	}

	public void insert(RegionSet s) {
		if (s.isEmpty());
		else if (set == EMPTY) {
			set = s.freeze();
		} else {
			makeMutable();

			Versions vers = s.getVersions();
			if (vers.getSize() > 1
				|| needToMakeMultiDocSupport(
					(DocumentVersion) (vers.versions().next())))
				makeMultiDocSupport();

			try {
				((Mutable) set).insert(s);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		}
	}

	/**
	 * Delete all regions represented by s.
	 * @param s Region enclosing (in the sense of mightEnclose()) all regions to remove
	 */
	public void delete(Region s) {
		if (s.isEmpty() || isEmpty())
			return;
		else {
			makeMutable();
			try {
				((Mutable) set).delete(s);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		}
	}

	public void delete(RegionSet s) {
		if (s.isEmpty() || isEmpty())
			return;
		else {
			makeMutable();
			try {
				((Mutable) set).delete(s);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		}
	}

	void makeMutable() {
		if (set instanceof Mutable)
			return;

		RegionArray ra = new RegionArray(getSize());
		try {
			ra.insert(set);
			set = ra;
		} catch (RepresentationChangeException e) {
			if (set instanceof MultiDocRegionSet) {
				RegionSet mr = e.getRepresentation();
				((MultiDocRegionSet) set).replace(mr);
			} else {
				set = e.getRepresentation();
			}
		}
	}

	//
	// switching between mutable and immutable reps
	//

	public RegionSet freeze() {
		do {
			try {
				return set.freeze();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public RegionSet crispen() {
		do {
			try {
				return set.crispen();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public RegionSet flatten() {
		do {
			try {
				return set.flatten();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public RegionSet melt() {
		do {
			try {
				return set.melt();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	//
	// subset enumeration
	//

	public SubsetEnumeration subsets() {
		do {
			try {
				return set.subsets();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	//
	// region enumeration
	//

	// EmptyRegionSet's DocumentVersion and Document are null
	// if using RegionSet.regions(), 
	// "return regions((Region)getUniversal(getDocument()))" will throw an NUllPointerException error
	public RegionEnumeration regions() {
		if (set instanceof EmptyRegionSet)
			return set.regions();
		else
			return super.regions();
	}

	public RegionEnumeration regions(Region bbox) {
		do {
			try {
				return set.regions(bbox);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public RegionEnumeration regions(RegionSet bboxs) {
		do {
			try {
				return set.regions(bboxs);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	//
	// fast lazy implementations of a few relational operators
	//

	public RegionSet justBefore(RegionSet bgd) {
		do {
			try {
				return set.justBefore(bgd);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public RegionSet justAfter(RegionSet bgd) {
		do {
			try {
				return set.justAfter(bgd);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public RegionSet starting(RegionSet bgd) {
		do {
			try {
				return set.starting(bgd);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public RegionSet ending(RegionSet bgd) {
		do {
			try {
				return set.ending(bgd);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public RegionSet in() {
		do {
			try {
				return set.in();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public RegionSet contains() {
		do {
			try {
				return set.contains();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public SubsetEnumeration singleDocize() {
		do {
			try {
				return set.singleDocize();
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}

	public SingleDocRegionSet getSingleDocRegionSet(Document doc) {
		do {
			try {
				return set.getSingleDocRegionSet(doc);
			} catch (RepresentationChangeException e) {
				if (set instanceof MultiDocRegionSet) {
					RegionSet mr = e.getRepresentation();
					((MultiDocRegionSet) set).replace(mr);
				} else {
					set = e.getRepresentation();
				}
			}
		} while (true);
	}
    
    public Versions getVersions() {
        return this.set.getVersions();
    }

	private boolean needToMakeMultiDocSupport(DocumentVersion ver) {
		if (ver == null)
			return false;
		if (set.isEmpty())
			return false;
		if (!(set instanceof SingleDocRegionSet))
			return false;
		if (((SingleDocRegionSet) set).getVersion().equals(ver))
			return false;

		return true;
	}

	private void makeMultiDocSupport() {
		if (!(set instanceof Mutable))
			makeMutable();
		MultiDocRegionSet ms = new MultiDocRegionSet((SingleDocRegionSet) set);
		set = ms;
	}
}
