/*
* LAPIS lightweight structured text processing system
*
* Copyright (C) 1998-2002 Carnegie Mellon University,
* Copyright (C) 2003 Massachusetts Institute of Technology.
* All rights reserved.
*
* This library is free software; you can redistribute it
* and/or modify it under the terms of the GNU General
* Public License as published by the Free Software
* Foundation, version 2.
*
* LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
*/

package lapis;

import java.util.*;
import lapisx.util.*;
import lapisx.enum.*;
import lapisx.iterator.*;

/**
 * A PatternLibrary represents a collection of {@link Parser Parsers} and named {@link Pattern Patterns}.
 * <p>
 * A PatternLibrary may have a parent library, from which it inherits parsers and patterns.  The collection represented by 
 * a particular library is the union of the parsers and patterns in all of its ancestors, up to the root (global) pattern library.
 * <p> 
 * At present, LAPIS uses only two levels in this hierarchy.  At the root is the global pattern library, a singleton, which
 * can be obtained with {@link #getGlobalLibrary() getGlobalLibrary}.  Its children are the document-specific pattern 
 * libraries, one per Document.  Each document-specific library has the global pattern library as its parent.
 * <p>
 * The primary purpose of the PatternLibrary is to map between pattern names and Pattern objects.
 * A client that wants to look up a named pattern (such as Word) normally follows the following 
 * procedure:
 * <ul>
 * <li> Fetch the document-specific library using {@link #getDocumentLibrary(Document) getDocumentLibrary}.
 * <li> Use {@link #lookup(String) lookup} to translate the pattern name (e.g., "Word") into its fully-qualified canonical name (".English.Word")      
 * <li> Use {@link #get(String) get} to map the canonical name into a Pattern object
 * </ul>
 * Patterns can be stored in the pattern library with {@link #put(String,Pattern) put()}.  Patterns that could apply to any document
 * should be stored in the global pattern library.  Patterns that are meaningful only for one document should be put in the 
 * document-specific library.  
 * <p>
 * The other responsibility of the PatternLibrary is to manage a set of Parsers.  Whenever a parser is added 
 * with {@link #addParser(Parser) addParser()}, the pattern library calls Parser.bind(PatternLibrary) to direct the parser to add its named Patterns to this
 * library.  Whenever a document-specific library is fetched, Parser.bind(PatternLibrary, Document) will be
 * called on any parsers that haven't seen the document yet, so that parsers with document-specific patterns can add
 * them to the document-specific library.
 */
public class PatternLibrary implements Cloneable {

    public static Debug debug = Debug.QUIET;

    /**
     * Global pattern library.
     */
    private static PatternLibrary rootLibrary = null;

    /** 
     * String used as key in properties of Document; value is the
     *  doc-specific PatternLibrary
     */
    private static String PatternLibraryProperty = "lapis.PatternLibrary";


    /**
     * Dummy pattern returned by get() for an unbound name.  
     */
    private static final Pattern EMPTY_PATTERN = new Pattern() {
        public RegionSet match(Document doc) {
            return RegionSet.EMPTY;
        }
    };

    /**
     *  This library's parent, or null if this is the global library.
     */
    private PatternLibrary parentLibrary = null;

    /** 
     * Maps final name component (CaselessString) -> canonical name (String) or Vector of canonical names.
     * Hidden identifiers are not included in this mapping.
     */
    private Hashtable names = new Hashtable();

    /**
     * Maps canonical name (CaselessString starting with ".") -> Pattern 
     */
    private Hashtable patterns = new Hashtable();

    /**
     * Collection of parsers.
     */
    private Set parsers = new HashSet ();

    /**
     * Version number.  Incremented whenever a new parser is added,
     * so that getDocumentLibrary() can tell that it needs to ask the parsers
     * to rebind document-specific patterns.
     */
    private long version = 0L;
    
    /** 
     * Collection of LibraryListeners.
     */
    private Set listeners = new SafeIteratorSet ();


    /**
     * Suppress default no-arg constructor.  PatternLibrary instances
     * can only be obtained by getGlobalLibrary() and getDocumentLibrary().
     */
    private PatternLibrary() {
    }

    /**
     * Get the global pattern library, which is a singleton shared by all documents.
     * Patterns that should be available to all documents should be stored in this
     * library.  Document-specific patterns should be stored in a document's library
     * (see getDocumentLibrary).    For looking up patterns, it's best
     * to use getDocumentLibrary(), since a document library automatically falls back
     * to the global library if a pattern isn't found. 
     */
    public static PatternLibrary getGlobalLibrary() {
        if (rootLibrary == null)
            rootLibrary = new PatternLibrary();
        return rootLibrary;
    }

    /**
     * Get a document's pattern library.
     * @param doc Document to be searched
     * @return document-specific pattern library, which includes not only
     * document-specific patterns but also patterns from the global library.
     */
    public static PatternLibrary getDocumentLibrary(Document doc) {
        PatternLibrary globalLib = getGlobalLibrary(); 

        if (doc == null)
            return globalLib; 

        PatternLibrary docLib =
            (PatternLibrary) doc.getProperty(PatternLibraryProperty);

        if (docLib == null) {
            docLib = new PatternLibrary();
            docLib.parentLibrary = getGlobalLibrary();
            doc.putProperty(PatternLibraryProperty, docLib);
        }
        
        if (docLib.version != globalLib.version) {
            // tell all parsers in the global library to bind themselves into the new library
            docLib.version = globalLib.version; 
            for (Iterator g = globalLib.parsers (); g.hasNext (); ) {
                Parser parser = (Parser) g.next (); 
                parser.bind (docLib, doc);
            }
        }
        
        return docLib;
    }

    /**
     * Get this library's parent library. 
     * @return this library's parent, or null if this is the global library.
     */
    public PatternLibrary getParent () {
        return parentLibrary;
    }
    
    /**
     * Add a parser to this library.
     * @param parser Parser to add
     * @effects Adds parser to this library's parser collection, and calls the parser's {@link Parser#bind(PatternLibrary) bind} method.
     */
    public synchronized void addParser (Parser parser) {
        if (parsers.add (parser)) {
            ++version;
            parser.bind (this);
            fireAddedParserEvent (parser);
        }
    }
    
    /**
     * Remove a parser from this library.
     * @param parser Parser to remove
     * @effects Removes parser from this library's parser collection.
     */
    public synchronized void removeParser (Parser parser) {
        parsers.remove (parser);
        // TODO: tell parser to unbind its patterns
    }
    
    /**
     * Get the parsers in this library and all of its ancestors.
     * @return iterator that yields the parsers in this library and all of its ancestors. 
     */
    public Iterator parsers () {
        Iterator g = parsers.iterator (); 
        if (parentLibrary != null)
            g = new ConcatIterator (g, parentLibrary.parsers ());
        return g;
    }
    
    /**
     * Clears all patterns and parsers from this pattern library.  Leaves parent
     * unchanged.
     */
    public synchronized void clear() {
        fireClearEvent();
        names.clear();
        patterns.clear();
        parsers.clear ();
    }

 
    /**
     * Get the pattern corresponding to a name.
     * @param name Canonical name of a pattern, e.g. ".English.Word"
     * @return Pattern bound to the name in this library.  If the name is not bound in this library, recursively 
     * tries to get the named pattern from this library's parent library.  If the name is not bound in any of this library's
     * ancestors, then returns a dummy Pattern object that matches the empty region set on every document.
     */
    public synchronized Pattern get(String name) {
        Object docObj = patterns.get(makePatternsKey(name));
        if (docObj instanceof Pattern)
            return (Pattern) docObj;
        else if (parentLibrary != null)
            return parentLibrary.get(name);
        else {
            if (parentLibrary == null) {
                return EMPTY_PATTERN;
            } else {
                return parentLibrary.get(name);
            }
        }
    }

    /**
     * Assign a pattern to a name in this library.
     * @param name Canonical name for pattern, e.g. ".English.Word"
     * @param pattern Pattern to bind to name
     * @effects Assigns name to pattern in this library.  Doesn't affect this library's parent library.  
     */
    public synchronized void put(String name, Pattern pattern) {
        debug.println ("binding " + name + " to " + pattern.getClass ());
        CaselessString rskey = makePatternsKey(name);

        if (patterns.put(rskey, pattern) != null)
            return; // name already existed

        // Add base name -> canonical name mapping to names hash
        if (!isHidden(name)) {
            CaselessString nkey = makeNamesKey(name);
            if (name.startsWith("."))
                name = name.substring(1);
            Object obj = names.get(nkey);
            if (obj == null) 
                names.put(nkey, name);
	    else {
                Vector v;
                if (obj instanceof String) {
                    v = new Vector();
                    v.addElement(obj);
                    names.put(nkey, v);
                } else
                    v = (Vector) obj;

                v.addElement(name);
            }
        }

        //this preserves invariant of NO duplicates
        //	if (parentLibrary != null)
        //	    this.parentLibrary.recursiveRemove(name);

        fireAddedEvent(name);
    }

    /**
     * Rename a pattern.
     * @param oldName old canonical name for pattern
     * @param newName new canonical name
     * @effects Locates pattern named oldName by searching for it, first in this library, and failing that, in this library's
     * ancestors.  In <b>every</b> library where oldName is found, the assignment to oldName is removed, and the pattern is
     * rebound to newName instead.
     * @return true if oldName was found in this library or one of its ancestors; false if not. 
     */
    public synchronized boolean rename(String oldName, String newName) {
        boolean found = patterns.containsKey(makePatternsKey(oldName)); 
        if (found) {
            Pattern pat = get(oldName);
            remove(oldName);
            put(newName, pat);
        }
        
        if (parentLibrary != null)
            found |= parentLibrary.rename(oldName, newName);
            
        return found;
    }

    /**
     * Remove named pattern from this library and all of its ancestors.
     * @param name Canonical name of pattern to remove
     * @effects Removes pattern from this library and all its ancestors, if found.
     */
    public synchronized void remove (String name) {
        // First delete name -> pattern mapping
        if (patterns.remove (makePatternsKey(name)) != null) {

            // Then delete basename -> name mapping from names hash
            if (!isHidden(name)) {
                CaselessString key = makeNamesKey(name);
                Object obj = names.get(key);
                if (obj == null); // name name not found 
                else {
                    if (obj instanceof String)
                        names.remove(key);
                    else {
                        Vector v = (Vector) obj;
                        v.removeElement(name);
                        if (v.isEmpty())
                            names.remove(key);
                    }
                }
            }
    
            fireRemovedEvent(name);
        }
        
        // finally, delete from ancestors
        if (parentLibrary != null)
            parentLibrary.remove (name);
    }

    /**
     * Test whether this library or one of its ancestors contains a named pattern.
     * @param name Canonical name of pattern to test for
     * @return true if name is found in this library or one of its ancestors; false if not.
     */
    public synchronized boolean contains(String name) {
        return patterns.containsKey(makePatternsKey(name))
            || (parentLibrary != null && parentLibrary.contains(name));
        //getDefaultLibrary().patterns.containsKey (makePatternsKey (name));	    
    }

    /**
     * Convert a pattern name into a canonical name, relative to the top-level namespace.  See {@link #lookup(String,String) lookup(String,String)} for
     * more details.   
     * @param name Relative or canonical name for pattern, e.g. "Word".
     * @return canonical name for the pattern (e.g. ".English.Word"), assuming only one named pattern matches in this
     * library or any of its ancestors.
     * @throws LabelNotFoundException if no names in the pattern library or its ancestors match the short pattern name.
     * @throws LabelAmbiguousException if more than one name in the pattern library or its ancestors matches the short name.
     */
    public synchronized String lookup(String name)
        throws LabelNotFoundException, LabelAmbiguousException {
        return lookup(name, "");
    }

    /**
     * Convert a short pattern name into a canonical name, given a preferred namespace.
     * <p>
     * A canonical pattern name begins with a period (e.g. ".English.Word.AllCapsWord").
     * A relative pattern name is any proper suffix (on period boundaries) of a canonical name, e.g.  "AllCapsWord", 
     * "Word.AllCapsWord", and "English.Word.AllCapsWord".
     * <p>
     * This method resolves a name relative to the preferred namespace, using the following rules:
     * <ol>
     * <li>If the name is already canonical, and it appears in this library or its ancestors, it is returned unchanged.
     * <li>If namespace+"."+name is found in this library or its ancestors, then namespace+"."+name is returned.
     * <li>Otherwise, all canonical names in this library or its ancestors that have name as a proper suffix are returned.
     * </ol>
     * If exactly one canonical name matches step #3, then it is returned.  If fewer than one or more than one canonical names 
     * match, then an exception is thrown.   
     * @param name Relative or canonical name for pattern, e.g. "Word".
     * @param namespace Preferred namespace for relative name lookups.  Must be canonical, i.e. starting with ".".
     * May be null if no namespace should be preferred (in which case step #2 is skipped above). 
     * @return canonical name for the pattern (e.g. ".English.Word"), assuming only one named pattern matches in this
     * library or any of its ancestors.
     * @throws LabelNotFoundException if no names in the pattern library or its ancestors match the short pattern name.
     * @throws LabelAmbiguousException if more than one name in the pattern library or its ancestors matches the short name.
     */
    public synchronized String lookup(String name, String namespace)
        throws LabelNotFoundException, LabelAmbiguousException {

        // Names starting with . are already canonical -- strip the period 
        // and look it up in the patterns hash
        if (name.startsWith(".")) {
            String canonical = name.substring(1);
            //            if (patterns.containsKey (makePatternsKey (canonical)))
            if (this.contains(canonical))
                return canonical;
            else {
                //	throw new LabelNotFoundException (name);

                if (parentLibrary == null)
                    throw new LabelNotFoundException(name);
                else
                    return parentLibrary.lookup(name, namespace);
            }

        }

        // First look for name in specified namespace.
        String specific =
            (namespace != null && namespace.length() > 0)
                ? namespace + "." + name
                : name;
        //        if (patterns.containsKey (makePatternsKey (specific)))
        if (this.contains(specific))
            return specific;

        // Then search for the name in all namespaces.
        // If it appears exactly once, return that one; otherwise
        // throw LabelAmbiguous or LabelNotFound.

        // Look it up in names mapping to get 0 or more canonical names 
        // ending with that base

        Object obj = names.get(makeNamesKey(name));

        /**
        
           TOTAL HACK!
        
            Object docSpecObj = names.get (makeNamesKey (name));
        Object genObj     = getDefaultLibrary().names.get(makeNamesKey (name));
        if (this == getDefaultLibrary()) genObj = null;
        
        Object obj = null;
        if (docSpecObj instanceof String && genObj == null) { 
            obj = docSpecObj;
            }
        if (genObj instanceof String && docSpecObj == null) {
        obj = genObj;
            }
        if (docSpecObj instanceof Vector &&
        genObj instanceof Vector) {
            obj = new Vector((Vector)docSpecObj);
            ((Vector)obj).addAll((Vector)genObj);
            }
        if (docSpecObj instanceof Vector && !(genObj instanceof Vector)) {
        obj = new Vector((Vector)docSpecObj);
            if (genObj instanceof String) {
            ((Vector)obj).add(genObj);
        	}
            }
        if (genObj instanceof Vector && !(docSpecObj instanceof Vector)) {
        obj = new Vector((Vector)genObj);
            if (docSpecObj instanceof String) {
            ((Vector)obj).add(docSpecObj);
        	}
            }
        */

        if (obj instanceof String) {
            String canonical = (String) obj;

            if (keyEndsWith(canonical, name))
                return canonical;
            else {
                //throw new LabelNotFoundException (name);
                if (parentLibrary == null)
                    throw new LabelNotFoundException(name);
                else
                    return parentLibrary.lookup(name, namespace);
            }
        } else if (obj instanceof Vector) {
            Vector v = (Vector) obj;
            String match = null;
            Vector multipleMatches = null;

            for (Enumeration e = v.elements(); e.hasMoreElements();) {
                String canonical = (String) e.nextElement();
                if (keyEndsWith(canonical, name)) {
                    if (multipleMatches != null || match != null) {
                        if (multipleMatches == null) {
                            multipleMatches = new Vector();
                            multipleMatches.addElement(match);
                        }
                        multipleMatches.addElement(canonical);
                    } else if (match == null)
                        match = canonical;
                }
            }

            if (match == null) {
                // throw new LabelNotFoundException (name);
                if (parentLibrary == null)
                    throw new LabelNotFoundException(name);
                else
                    return parentLibrary.lookup(name, namespace);
            } else if (multipleMatches != null)
                throw new LabelAmbiguousException(name, multipleMatches);
            else
                return match;
        } else {
            //throw new LabelNotFoundException (name);
            if (parentLibrary == null)
                throw new LabelNotFoundException(name);
            else
                return parentLibrary.lookup(name, namespace);
        }
    }

    /**
      * Get the visible names in the library (i.e., names that don't
      * begin with an underscore).
      * @return Enumeration of Strings, the canonical names of the
      * visible identifiers in this library and its ancestors
      */
    public Enumeration visibleNames() {
        Enumeration e = null;
        if (parentLibrary == null)
            e = patterns.keys();
        else
            e =
                new ConcatEnumeration(
                    patterns.keys(),
                    parentLibrary.visibleNames());

        return new FilteredEnumeration(e) {
            public void transform(Object o) {
                String name = o.toString();
                if (!isHidden(name))
                    yield(name);
            }
        };
    }

    /**
      * Get all names in this library and its ancestors (both visible and hidden).
      * @return Enumeration of Strings, the canonical names of the
      * identifiers in this library and its ancestors
      */
    public Enumeration allNames() {
        Enumeration e = null;
        if (parentLibrary == null)
            e = patterns.keys();
        else
            e =
                new ConcatEnumeration(
                    patterns.keys(),
                    parentLibrary.visibleNames());

        return new FilteredEnumeration(e) {
            public void transform(Object o) {
                yield(o.toString());
            }
        };
    }

    /**
     * Get number of named patterns in this library and its ancestors.
     * @return number of names that would be enumerated by {@link #allNames}.
     */
    public int getSize() {
        if (parentLibrary == null)
            return patterns.size();
        else
            return patterns.size() + this.parentLibrary.getSize();
    }

    /**
     * Get shortest unambiguous name for a canonical name.
     * @param canonical canonical pattern name, e.g. ".English.Word.AllCapsWord"
     * @return shortest unambiguous relative name for this pattern, e.g. "AllCapsWord" 
     */
    public synchronized String getShortestName(String canonical) {
        if (canonical.startsWith("."))
            canonical = canonical.substring(1);

        for (int i = canonical.lastIndexOf('.');
            i != -1;
            i = canonical.lastIndexOf('.', i - 1)) {
            try {
                String abbr = canonical.substring(i + 1);
                if (lookup(abbr).equalsIgnoreCase(canonical))
                    return abbr;
            } catch (LookupException e) {
                //added - MN
                if (parentLibrary != null)
                    parentLibrary.getShortestName(canonical);
            }
        }

        return canonical;
    }

    /**
     * Extract the last component of a name, i.e.,
     * the string following the last period.
     * @param name Pattern name (either canonical or relative), e.g. "English.Word"
     * @return component of name following last period, e.g. "Word"
     */
    public static String getBasename(String name) {
        if (name.startsWith("."))
            name = name.substring(1);

        int lastPeriod = name.lastIndexOf('.');
        return (lastPeriod == -1) ? name : name.substring(lastPeriod + 1);
    }

    /**
     * Test whether name is a hidden identifier.
     * @param name Identifier to test
     * @return true iff name's last component (its base name)
     * starts with an underscore
     */
    public static boolean isHidden(String name) {
        return getBasename(name).startsWith("_");
    }

    static boolean keyEndsWith(String key, String suffix) {
        int keyLen = key.length();
        int suffixLen = suffix.length();
        int pos = keyLen - suffixLen;
        return pos >= 0
            && key.substring(pos).equalsIgnoreCase(suffix)
            && (pos == 0 || key.charAt(pos - 1) == '.');
    }

    static CaselessString makePatternsKey(String name) {
        if (name.startsWith("."))
            name = name.substring(1);
        return new CaselessString(name);
    }

    static CaselessString makeNamesKey(String name) {
        return new CaselessString(getBasename(name));
    }

    /**
     * Base class for exceptions thrown by lookup(). 
     */
    public static class LookupException extends Exception {
        String name;
        public LookupException(String name, String message) {
            super(message);
            this.name = name;
        }
        public String getName() {
            return name;
        }
    }

    /**
     * Exception thrown when lookup() finds no matching pattern names.  
     */
    public static class LabelNotFoundException extends LookupException {
        public LabelNotFoundException(String name) {
            super(name, "there is no pattern named " + name.toUpperCase());
        }
    }

    /**
     * Exception thrown when lookup() finds multiple matching pattern names.  
     */
    public static class LabelAmbiguousException extends LookupException {
        Vector choices;
        public LabelAmbiguousException(String name, Vector choices) {
            super(name, "pattern name " + name.toUpperCase() + " is ambiguous");
            this.choices = choices;
        }
        public String[] getChoices() {
            String[] result = new String[choices.size()];
            choices.copyInto(result);
            return result;
        }
    }

    //
    // Event listeners
    //

    /**
     * Listener notified when patterns are added and removed from library.  
     */
    public static interface LibraryListener {
        /**
         * Called when a pattern is bound into the pattern library using put().
         * @param lib PatternLibrary that changed
         * @param name name that was bound
         */
        public void patternAdded(PatternLibrary lib, String name);
        /**
         * Called when a pattern name is removed from the pattern library by remove() or clear().
         * @param lib PatternLibrary that changed
         * @param name name that was removed
         */
        public void patternRemoved(PatternLibrary lib, String name);
        /**
        * Called when a parser is added to the pattern library using addParser().
        * @param lib PatternLibrary that changed
        * @param parser Parser that was added
        */
       public void parserAdded (PatternLibrary lib, Parser parser);
   }

    /**
     * Add a listener to this library.  The listener receives events that affect this library and all its ancestors.
     * @param listener Listener to receive library events
     * @effects Registers listener for future events from this library and its ancestors.
     */
    public synchronized void addLibraryListener(LibraryListener listener) {
        debug.println("adding library listener " + listener.getClass());
        listeners.add (listener);
        if (parentLibrary != null)
            parentLibrary.addLibraryListener (listener);
    }

   /**
    * Removes a listener from this library and all its ancestors.
    * @param listener Listener to stop receiving library events
    * @effects Listener no longer receives events from this library or any of its ancestors.
    */
   public synchronized void removeLibraryListener(LibraryListener listener) {
        listeners.remove (listener);
        if (parentLibrary != null)
            parentLibrary.removeLibraryListener (listener);
    }

    private synchronized void fireAddedEvent(String name) {
        debug.println("**** firing added event: " + name);
        if (name.startsWith("."))
            name = name.substring(1);

        for (Iterator g = listeners.iterator (); g.hasNext (); )
             ((LibraryListener) g.next ()).patternAdded (this, name);
    }

    private synchronized void fireRemovedEvent(String name) {
        debug.println("PATTERN REMOVE:" + name);
        if (name.startsWith("."))
            name = name.substring(1);

        for (Iterator g = listeners.iterator (); g.hasNext (); )
             ((LibraryListener) g.next ()).patternRemoved (this, name);
    }

    private synchronized void fireAddedParserEvent (Parser parser) {
        for (Iterator g = listeners.iterator (); g.hasNext (); )
             ((LibraryListener) g.next ()).parserAdded (this, parser);
    }

    private synchronized void fireClearEvent() {
        debug.println("*** in fire clear event...");
        Iterator names = patterns.keySet().iterator();
        while (names.hasNext()) {
            CaselessString temp = (CaselessString) names.next();
            debug.println("*** removing: " + temp);
            fireRemovedEvent(temp.toString());
        }
    }
}
