/*
 * LAPIS lightweight structured text processing system
 * Copyright (C) 1998-2002 Carnegie Mellon University. All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Library
 * General Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://www.cs.cmu.edu/~rcm/lapis/
 */
package lapis;

import lapis.tc.*;
import lapis.trees.*;
import lapis.swing.*;
import java.util.*;
import java.text.*;
import lapisx.util.*;

public class Perf {

    final static String PREFIX = "lapis.Perf$"; 

    static boolean DEBUG = false;

    static int INCREMENT = 10;

    static long MIN_TEST_DURATION = 1000;

    static long MIN_ITERATIONS = 3;

    static int N;

    static Random random;

    static NumberFormat nf = NumberFormat.getInstance ();
    static {
        nf.setMaximumFractionDigits (2);
    }

    static TC pattern;

    public static void main (String[] argv) throws Exception {
        int start = 1000;
        int pow = 6;
        double multiple = 2;
        long seed = System.currentTimeMillis ();
        Vector classes = new Vector ();
        boolean display = false;

        // parse command-line arguments
        for (int i=0; i<argv.length; ++i) {
            if (argv[i].startsWith ("-")) {
                String opt = argv[i].substring (1);
                if (opt.equals ("n")) {
                    start = Integer.parseInt (argv[++i]);
                } else if (opt.equalsIgnoreCase ("pow")) {
                    pow = Integer.parseInt (argv[++i]);
                } else if (opt.equals ("seed")) {
                    seed = Long.parseLong (argv[++i]);
                } else if (opt.equals ("pattern")) {
                    pattern = new TC (argv[++i]);
                } else if (opt.equals ("display")) {
                    display = true;
                } else if (opt.equals ("increment")) {
                    INCREMENT = Integer.parseInt (argv[++i]);
                } else if (opt.equals ("mintest")) {
                    MIN_TEST_DURATION = Integer.parseInt (argv[++i]);
                } else if (opt.equals ("miniter")) {
                    MIN_ITERATIONS = Integer.parseInt (argv[++i]);
                } else if (opt.equals ("multiple")) {
                    multiple = Double.parseDouble (argv[++i]);
                } else
                    println ("unknown arg: " + opt);
            } else
                classes.addElement (Class.forName (PREFIX + argv[i]));
        }

        println ("seed = " + seed);
        random = new Random (seed);
            
        for (Enumeration e = classes.elements (); e.hasMoreElements (); ) {
            Class cls = (Class) e.nextElement ();

            // run once to load classes and run JIT, without issuing a report
            println ("priming the pump -- may take a while");
            N = start;
            for (int i = 0; i < pow; ++i) {
                println ("N=" + N);
                ((Timeable) cls.newInstance ()).run ();
                N = (int) (N * multiple);
            }

            System.gc ();
            System.gc ();

            // now run for real
            N = start;
            for (int i = 0; i < pow; ++i) {
                RegionSet s = runTimed ((Timeable) cls.newInstance ());
                if (display) {
                    popup (s);
                    display = false;
                }
                N = (int) (N * multiple);
            }
        }
    }


    //
    // building region sets
    //
    public static abstract class Build implements Timeable {
        public void report (RegionSet result, double time, long space) {
            println (Str.after (getClass ().getName (), PREFIX) + ": "
                     + "N=" + N + ", "
                     + nf.format (time) + " msec, "
                     + space + " bytes, "
                     + result.getSize () + " rects, "
                     + nf.format ((time * 1000) / result.getSize ()) + " upr, "
                     + nf.format (space / (double) result.getSize ()) + " bpr");
        }
    }

    public static class BuildRegionArrayInOrder extends Build {
        int[] c = randomFlat (N);
        public RegionSet run () {
            return makeRegionSet (c);
        }
    }

    public static class BuildRBTreeInOrder extends Build {
        int[] c = randomFlat (N);
        public RegionSet run () {
            return makeRBTree (c);
        }
    }

    public static class BuildRegionArrayOutOfOrder extends Build {
        int[] c = randomOrder (randomFlat (N));
        public RegionSet run () {
            return makeRegionSet (c);
        }
    }

    public static class BuildRBTreeOutOfOrder extends Build {
        int[] c = randomOrder (randomFlat (N));
        public RegionSet run () {
            return makeRBTree (c);
        }
    }

    public static class BuildFuzzyRegionArrayInOrder extends Build {
        int[] c = randomFlat (2*N);
        public RegionSet run () {
            return makeFuzzyRegionArray (c);
        }
    }

    public static class BuildFuzzyRBTreeInOrder extends Build {
        int[] c = randomFlat (2*N);
        public RegionSet run () {
            return makeFuzzyRBTree (c);
        }
    }


    //
    // querying region sets
    //
    public static abstract class Query implements Timeable {
        RegionSet s;
        int[] q = randomFlat (N*2);
        int matches = 0;
        public RegionSet run () {
            matches = query (s, q);
            return s;
        }
        public void report (RegionSet result, double time, long space) {
            println (Str.after (getClass ().getName (), PREFIX) + ": "
                     + "N=" + N + ", "
                     + nf.format (time) + " msec, "
                     + s.getSize () + " rects, "
                     + matches + " matches, "
                     + nf.format ((time * 1000) / N) + " upr");
        }
    }

    public static class QueryRegionArray extends Query {
        public QueryRegionArray () {
            s = makeRegionArray (randomFlat (N));
        }
    }

    public static class QueryRBTree extends Query {
        public QueryRBTree () {
            s = makeRBTree (randomNested (N));
        }
    }

    public static class QueryFuzzyRegionArray extends Query {
        public QueryFuzzyRegionArray () {
            s = makeFuzzyRegionArray (randomFlat (2*N));
        }
    }

    public static class QueryFuzzyRBTree extends Query {
        public QueryFuzzyRBTree () {
            s = makeFuzzyRBTree (randomFlat (2*N));
        }
    }

    static int query (RegionSet s, int[] q) {
        int matches = 0;
        for (int i = 0; i < q.length; i += 4) {
            RegionEnumeration re = 
                s.regions (new FuzzyRegion (q[i], q[i+1], q[i+2], q[i+3], doc));
            if (re.first () != null) {
                ++matches;
                while (re.next () != null)
                    ++matches;
            }
        }
        return matches;
    }

    //
    // running patterns
    //

    static Document doc = BasicDocument.defaultFactory.make ("");

    public static class RunPattern implements Timeable {
        long opnds = 0;
        public RunPattern () {
            String pat = pattern.getExpression ();

            if (pat.indexOf ("any") != -1)
                bind ("any", 
                	new FuzzyRegion (0, N * INCREMENT, 0, N * INCREMENT, doc));
            if (pat.indexOf ("all") != -1)
                bind ("all", new Region (0, N * INCREMENT, doc));
            if (pat.indexOf ("F1") != -1)
                bind ("F1", makeRegionSet (randomFlat (N)));
            if (pat.indexOf ("F2") != -1)
                bind ("F2", makeRegionSet (randomFlat (N)));
            if (pat.indexOf ("F3") != -1)
                bind ("F3", makeRegionSet (decimate (100, randomFlat (N))));
            if (pat.indexOf ("N1") != -1)
                bind ("N1", makeRegionSet (randomNested (N)));
            if (pat.indexOf ("N2") != -1)
                bind ("N2", makeRegionSet (randomNested (N)));
            if (pat.indexOf ("N3") != -1)
                bind ("N3", makeRegionSet (decimate (100, randomNested (N))));
            if (pat.indexOf ("A1") != -1)
                bind ("A1", makeRegionSet (randomAny (N)));
            if (pat.indexOf ("A2") != -1)
                bind ("A2", makeRegionSet (randomAny (N)));
            if (pat.indexOf ("A3") != -1)
                bind ("A3", makeRegionSet (decimate (100, randomAny (N))));
        }
        void bind (String name, RegionSet s) {
            PatternLibrary lib = PatternLibrary.getGlobalLibrary ();
            //lib.put (name, new TC (new NamedRegionSet (doc, s)));
			lib.put (name, new TC (new NamedRegionSet (s)));
            opnds += s.getSize ();
        }
        public RegionSet run () {
            RegionSet s = pattern.match (doc);
            RegionEnumeration re = s.regions ();
            if (re.first () != null) {
                while (re.next () != null)
                    ;
            }
            return s;
        }
        public void report (RegionSet result, double time, long space) {
            long totalRects = opnds + result.getSize ();
            println (Str.after (getClass ().getName (), PREFIX) + ": "
                     + "N=" + N + ", "
                     + nf.format (time) + " msec, "
                     + opnds + " input rects, "
                     + result.getSize () + " output rects in "
                     + Str.before (result.toString (),
                                  "[") + ", "
                     + totalRects + " rects, "
                     + nf.format ((time * 1000) / N) + " upr");
        }
    }
    

    //
    // generating random regions
    //

    static int[] randomFlat (int n) {
        return randomFlat (n, INCREMENT);
    }

    static int[] randomFlat (int n, int inc) {
        int m = n*2;
        int[] a = new int[m];
        int p = 0;
        int i = 0;
        while (i < m) {
            p += random.nextInt (inc);
            a[i++] = p;
            p += random.nextInt (inc);
            a[i++] = p;
        }
        return a;
    }

    static final float POP_THRESHOLD = 0.33f;
    static final float PUSH_THRESHOLD = 0.5f;
    
    static int[] randomNested (int n) {
        return randomNested (n, INCREMENT);
    }

    static int[] randomNested (int n, int inc) {
        int[] coords = randomFlat (n, inc);
        Stack s = new Stack ();

        int m = coords.length;
        int i = 0; // index into coords
        int[] a = new int[m];
        int j = 0; // index into a
        while (i < m) {
            if (i + s.size () == m
                || (!s.empty () && random.nextFloat () < POP_THRESHOLD)) {
                // pop
                a[j++] = ((Integer) s.pop ()).intValue ();
                a[j++] = coords[i++];
            } else if (random.nextFloat () < PUSH_THRESHOLD) {
                // push
                s.push (new Integer (coords[i++]));
            } else {
                // sibling
                a[j++] = coords[i++];
                a[j++] = coords[i++];
            }
        }

        return a;
    }

    static int[] randomAny (int n) {
        return randomAny (n, INCREMENT);
    }

    static int[] randomAny (int n, int inc) {
        int[] coords = randomFlat (n, inc);
        int m = coords.length;

        // randomly permute coords
        for (int i = 0; i < m; ++i) {
            int j = random.nextInt (m);
            int x = coords[i];
            coords[i] = coords[j];
            coords[j] = x;
        }

        // make sure every pair is in order
        for (int i = 0; i < m; i+=2)
            if (coords[i] > coords[i+1]) {
                int x = coords[i];
                coords[i] = coords[i+1];
                coords[i+1] = x;
            }
            
        return coords;
    }


    static int[] randomOrder (int[] coords) {
        // randomly permute coords
        int m = coords.length;
        for (int i = 0; i < m; i+=2) {
            int j = random.nextInt (m/2) * 2;
            int x = coords[i];
            coords[i] = coords[j];
            coords[j] = x;
            x = coords[i+1];
            coords[i+1] = coords[j+1];
            coords[j+1] = x;
        }
        return coords;
    }

    static int[] decimate (int frac, int[] coords) {
        int m = coords.length / frac;
        if ((m & 1) == 1)
            --m;
        int[] a = new int[m];
        for (int i = 0; i < m; i += 2) {
            a[i] = coords[i*frac];
            a[i+1] = coords[i*frac + 1];
        }
        return a;
    }

    //
    // constructing region sets
    // 
    
    static RegionSet makeRegionSet (int[] a) {
        MutableRegionSet t = new MutableRegionSet ();
        for (int i = 0; i < a.length; i+= 2)
            t.insert (a[i], a[i+1], doc);

        if (DEBUG)
            assertSame (t, a);
        return t;
    }

    static RBTree makeRBTree (int[] a) {
        RBTree t = new RBTree (doc.getVersion());
        for (int i = 0; i < a.length; i+= 2)
            t.insert (new Region (a[i], a[i+1], doc));

        if (DEBUG)
            assertSame (t, a);
        return t;
    }

    static RBTree makeFuzzyRBTree (int[] a) {
        RBTree t = new RBTree (doc.getVersion());
        for (int i = 0; i < a.length; i+= 4)
            t.insert (new FuzzyRegion (a[i], a[i+1], a[i+2], a[i+3], doc));

        if (DEBUG)
            assertSame (t, a);
        return t;
    }

    static RegionArray makeRegionArray (int[] a) {
        RegionArray t = new RegionArray ();
        try {
            for (int i = 0; i < a.length; i+= 2)
                t.insert (a[i], a[i+1], doc.getVersion());
            t.validate ();
        } catch (RepresentationChangeException e) {
            e.printStackTrace ();
        }

        if (DEBUG)
            assertSame (t, a);
        return t;
    }


    static RegionArray makeFuzzyRegionArray (int[] a) {
        RegionArray t = new RegionArray ();
        try {
            for (int i = 0; i < a.length; i+= 4)
                t.insert (new FuzzyRegion (a[i], a[i+1], a[i+2], a[i+3], doc));
        } catch (RepresentationChangeException e) {
            e.printStackTrace ();
        }

        if (DEBUG)
            assertSame (t, a);
        return t;
    }

    static void assertSame (RegionSet t, int[] a) {
        int[] d = new int[a.length];
        System.arraycopy (a, 0, d, 0, a.length);

        RegionEnumeration re = t.regions ();
        for (Region r = re.first (); r != null; r = re.next ()) {
            int s = r.getStart ();
            int e = r.getEnd ();
            boolean found = false;
            for (int i = 0; i < d.length; i+=2)
                if (s == d[i] && e == d[i+1]) {
                    found = true;
                    d[i] = -1;
                    d[i+1] = -1;
                }
            if (!found)
                println ("*** " + t + " shouldn't include " + r);
        }
        for (int i = 0; i < d.length; i+=2)
            if (d[i] != -1)
                println ("*** " + t + " doesn't include "
                         + new Region (d[i], d[i+1], doc));
    }



    //
    // timing
    //

    static interface Timeable {
        RegionSet run ();
        void report (RegionSet result, double time, long space);
    }

    static RegionSet runTimed (Timeable r) {
        // mark memory usage
        Mem.gc ();
        long memStart = Mem.used ();

        // run the test enough times to bring it up to MIN_TEST_DURATION
        long start = System.currentTimeMillis ();
        long duration;
        RegionSet result;
        int iterations = 0;
        do {
            result = r.run ();
            duration = System.currentTimeMillis () - start;
            ++iterations;
        } while (duration < MIN_TEST_DURATION
                 || iterations < MIN_ITERATIONS);

        // compute memory used by result
        Mem.gc ();
        long mem = Mem.used () - memStart;

        // report and return
        r.report (result, ((double) duration) / iterations, mem);
        return result;
    }



    //
    // output and debugging
    //

    static void println (Object obj) {
        System.out.println (obj.toString ());
    }

    static void dump (String msg, int[] a) {
        println (msg);
        for (int i = 0; i < a.length; i+= 2)
            println ("\t" + a[i] + "-" + a[i+1]);
    }

    static void dump (String msg, RegionSet s) {
        println (msg);
        RegionEnumeration re = s.regions ();
        for (Region r = re.first (); r != null; r = re.next ())
            println ("\t" + r);
    }

    static void popup (RegionSet s) {
        RegionSetDisplay rsd = new RegionSetDisplay ();
        rsd.setLength (N * INCREMENT);
        //rsd.setSelection (new NamedRegionSet (null, s, DocumentVersion.getOriginal(doc) /*DocumentVersion.ORIGINAL*/));
		//rsd.setSelection (new NamedRegionSet (doc, s));
		rsd.setSelection (new NamedRegionSet (s));
        rsd.show ();
    }

        
}
