/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis;

import lapisx.util.Debug;

/**
 * Region interval in a document [start, end].
 * Start is the zero-based index of the first character in the
 * region.  End is the zero-based index of the first character
 * after the region.
 *
 */
public class Region extends SingleDocRegionSet {

	public static Debug debug = Debug.QUIET;

	public final static EmptyRegion EMPTY = new EmptyRegion();

	protected int start;
	protected int end;

	protected DocumentVersion version;

	protected Region() {
	}

	public Region(int start, int end, DocumentVersion ver) {
		this.start = start;
		this.end = end;
		this.version = ver;
	}

	public Region(int start, int end, Document doc) {
		this(start, end, doc.getVersion());
	}

	//
	// Accessor methods
	//

	public int getStartLow() {
		return start;
	}

	public int getStartHigh() {
		return start;
	}

	public int getEndLow() {
		return end;
	}

	public int getEndHigh() {
		return end;
	}

	public int getStart() {
		return start;
	}

	public int getEnd() {
		return end;
	}

	public int getLength() {
		return getEnd() - getStart();
	}

	//
	// Converting Regions to strings
	//

	//     public String extract (Document doc) {
	//         return doc.getText (this);
	//     }

	public String extract(String content) {
		int l = content.length();
		int s = Math.min(l, Math.max(0, getStart()));
		int e = Math.min(l, Math.max(0, getEnd()));
		return content.substring(s, e);
	}

	public String toString() {
		return "["
			+ rangeToString(getStartLow(), getStartHigh())
			+ ","
			+ rangeToString(getEndLow(), getEndHigh())
			+ "]"
			+ " ver("
			+ getVersion()
			+ ")";
	}

	static String rangeToString(int low, int high) {
		if (low == high)
			return String.valueOf(low);
		else
			return low + "-" + high;
	}

	//
	// Region identity
	//

	public boolean equals(Object o) {
		if (!(o instanceof Region))
			return super.equals(o);

		Region r = (Region) o;

		if (this.isEmpty() && r.isEmpty())
			return true;
		else if (!(this.isEmpty()) && !(r.isEmpty()))
			return getStartLow() == r.getStartLow()
				&& getStartHigh() == r.getStartHigh()
				&& getEndLow() == r.getEndLow()
				&& getEndHigh() == r.getEndHigh()
				&& this.version.equals(r.version);
		else
			return false;
	}

	//
	// Computing superset or subset of two Regions
	//
	// no client calls for max and min yet!
	//
	public Region max(Region r) {
		if (!(this.version.equals(r.version)))
			return null;

		if (mightEnclose(r))
			return this;
		else if (r.mightEnclose(this))
			return r;
		else
			return null;
	}

	public Region min(Region r) {
		if (!(this.version.equals(r.version)))
			return null;

		if (r.mightEnclose(this))
			return this;
		else if (mightEnclose(r))
			return r;
		else
			return null;
	}

	//
	// Enumerating regions
	//

	public int getSize() {
		return 1;
	}

	public boolean isEmpty() {
		return false;
	}

	public boolean isAllPoints() {
		return isPoint();
	}

	public boolean isPoint() {
		return getStartLow() == getEndHigh();
	}

	public boolean isCrisp() {
		return getStartLow() == getStartHigh() && getEndLow() == getEndHigh();
	}

	public boolean isFlat() {
		return isCrisp();
	}

	public RegionEnumeration regions() {
		return new RegionEnumeration() {
			public Region first() {
				return Region.this;
			}

			public Region next() {
				return null;
			}

			public Region firstFast() {
				return Region.this;
			}

			public Region nextFast() {
				return null;
			}
		};
	}

	public RegionEnumeration regions(Region bbox) {
		return mightIntersect(bbox) ? regions() : EMPTY.regions();
	}

	public RegionEnumeration regions(RegionSet bboxs) {
		SingleDocRegionSet s = bboxs.getSingleDocRegionSet(this.getDocument());
		return mightIntersect(s.getBoundingBox()) ? regions() : EMPTY.regions();
	}

	public RegionSet freeze() {
		return this;
	}

	public MutableRegionSet intersect(
		IntersectOp op,
		RegionSet u,
		MutableRegionSet result) {
		if (!mightIntersect(u))
			// no intersection possible
			;
		else if (u instanceof Region)
			op.combine(this, (Region) u, result);
		else {
			// expand u into regions overlapping me
			RegionEnumeration e = u.regions(this);
			for (Region s = e.firstFast(); s != null; s = e.nextFast())
				op.combine(this, s, result);
		}
		return result;
	}

	/**
	 * Fast test for whether this Region actually intersects a given Region.
	 * @return true iff r is a member of this
	 */
	public boolean intersects(Region r) {
		return mightIntersect(r);
	}

	//
	// Region set operations
	//

	public static Region make(
		int startLow,
		int startHigh,
		int endLow,
		int endHigh,
		DocumentVersion ver) {
		// canonicalize the interval endpoints
		startHigh = Math.min(startHigh, endHigh);
		endLow = Math.max(endLow, startLow);

		// test whether the set of regions contains at least 
		// one region
		if (startLow > startHigh || endLow > endHigh || startLow > endHigh)
			return EMPTY;

		if (startLow > endLow)
			endLow = startLow;

		if (startHigh > endHigh)
			startHigh = endHigh;

		// Construct a region of the most specific class possible
		if (startLow == startHigh && endLow == endHigh)
			return new Region(startLow, endHigh, ver);
		else
			return new FuzzyRegion(startLow, startHigh, endLow, endHigh, ver);
	}

	public RegionSet and(RegionSet rs) {
		if (rs instanceof Region) {
			if (!rs.isEmpty())
				debug.assertion(this.sameDocument((Region) rs));
			if (this.sameDocument((Region) rs)) {
				Region r = (Region) rs;
				return make(
					Math.max(getStartLow(), r.getStartLow()),
					Math.min(getStartHigh(), r.getStartHigh()),
					Math.max(getEndLow(), r.getEndLow()),
					Math.min(getEndHigh(), r.getEndHigh()),
					version);
			} else {
				return EMPTY;
			}
		} else
			return rs.and(this);
	}

	public RegionSet fromTo(RegionSet rs) {
		if (rs instanceof Region) {
			if (!rs.isEmpty())
				debug.assertion(this.sameDocument((Region) rs));
			if (this.sameDocument((Region) rs)) {
				Region r = (Region) rs;
				return make(
					getStartLow(),
					getStartHigh(),
					r.getEndLow(),
					r.getEndHigh(),
					version);
			} else {
				return EMPTY;
			}
		} else
			return super.fromTo(rs);
	}

	public RegionSet equalTo(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			currbkgd.fuzzify(getStartLow(), FUZZY_LOW),
			currbkgd.fuzzify(getStartHigh(), FUZZY_HIGH),
			currbkgd.fuzzify(getEndLow(), FUZZY_LOW),
			currbkgd.fuzzify(getEndHigh(), FUZZY_HIGH),
			version);
	}

	public RegionSet justBefore(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			MIN_INDEX,
			getStartHigh(),
			currbkgd.fuzzify(getStartLow(), FUZZY_LOW),
			getStartHigh(),
			version);
	}

	public RegionSet before() {
		return make(
			MIN_INDEX,
			getStartHigh(),
			MIN_INDEX,
			getStartHigh(),
			version);
	}

	public RegionSet beforeIn(RegionSet rs) {
		if (rs instanceof Region) {
			if (!rs.isEmpty())
				debug.assertion(this.sameDocument((Region) rs));
			if (this.sameDocument((Region) rs)) {
				Region r = (Region) rs;
				return make(
					r.getStartLow(),
					getStartHigh(),
					r.getStartLow(),
					getStartHigh(),
					version);
			} else {
				return EMPTY;
			}
		} else
			return super.beforeIn(rs);
	}

	public RegionSet justAfter(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			getEndLow(),
			currbkgd.fuzzify(getEndHigh(), FUZZY_HIGH),
			getEndLow(),
			MAX_INDEX,
			version);
	}

	public RegionSet after() {
		return make(getEndLow(), MAX_INDEX, getEndLow(), MAX_INDEX, version);
	}

	public RegionSet afterIn(RegionSet rs) {
		if (rs instanceof Region) {
			if (!rs.isEmpty())
				debug.assertion(this.sameDocument((Region) rs));
			if (this.sameDocument((Region) rs)) {
				Region r = (Region) rs;
				return make(
					getEndLow(),
					r.getEndHigh(),
					getEndLow(),
					r.getEndHigh(),
					version);
			} else {
				return EMPTY;
			}
		} else
			return super.afterIn(rs);
	}

	public RegionSet contains() {
		return make(MIN_INDEX, getStartHigh(), getEndLow(), MAX_INDEX, version);
	}

	public RegionSet starting(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			currbkgd.fuzzify(getStartLow(), FUZZY_LOW),
			Math.min(
				currbkgd.fuzzify(getStartHigh(), FUZZY_HIGH),
				getEndHigh() - 1),
			getStartLow() + 1,
			MAX_INDEX,
			version);
	}

	public RegionSet ending(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			MIN_INDEX,
			Math.max(MIN_INDEX, getEndHigh() - 1),
			Math.max(
				currbkgd.fuzzify(getEndLow(), FUZZY_LOW),
				getStartLow() + 1),
			currbkgd.fuzzify(getEndHigh(), FUZZY_HIGH),
			version);
	}

	// startingIn and endingIn are fuzzified inward to be compatible with in()
	public RegionSet startingIn(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			getStartLow(),
			Math.min(
				currbkgd.fuzzify(getStartHigh(), FUZZY_HIGH),
				getEndHigh()),
			getStartLow(),
			getEndHigh(),
			version);
	}

	public RegionSet endingIn(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			getStartLow(),
			getEndHigh(),
			Math.max(currbkgd.fuzzify(getEndLow(), FUZZY_LOW), getStartLow()),
			getEndHigh(),
			version);
	}

	// startingContains and endingContains are fuzzified outward to be compatible with contains()
	public RegionSet startingContains(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			Math.max(MIN_INDEX, currbkgd.fuzzify(getStartLow(), FUZZY_LOW)),
			getStartHigh(),
			getEndLow(),
			MAX_INDEX,
			version);
	}

	public RegionSet endingContains(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			MIN_INDEX,
			getStartHigh(),
			getEndLow(),
			Math.min(currbkgd.fuzzify(getEndHigh(), FUZZY_HIGH), MAX_INDEX),
			version);
	}

	public RegionSet in() {
		return make(
			getStartLow(),
			getEndHigh(),
			getStartLow(),
			getEndHigh(),
			version);
	}

	public RegionSet overlapsStart() {
		return make(
			MIN_INDEX,
			getStartHigh(),
			getStartLow(),
			getEndHigh(),
			getVersion());
	}

	public RegionSet overlapsEnd() {
		return make(
			getStartLow(),
			getEndHigh(),
			getEndLow(),
			MAX_INDEX,
			version);
	}

	public RegionSet overlaps() {
		return make(MIN_INDEX, getEndHigh(), getStartLow(), MAX_INDEX, version);
	}

	public RegionSet trim(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		int sl = currbkgd.fuzzify(getStartLow(), FUZZY_HIGH);
		int eh = currbkgd.fuzzify(getEndHigh(), FUZZY_LOW);
		return make(
			sl,
			Math.max(sl, getStartHigh()),
			Math.min(eh, getEndLow()),
			eh,
			version);
	}

	public RegionSet include(RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return new Region(
			currbkgd.fuzzify(getStartLow(), FUZZY_LOW),
			currbkgd.fuzzify(getEndHigh(), FUZZY_HIGH),
			version);
	}

	public RegionSet crispen() {
		return new Region(getStartLow(), getEndHigh(), version);
	}

	public RegionSet flatten() {
		return crispen();
	}

	public RegionSet melt() {
		return crispen();
	}

	public RegionSet convert(CoordinateMap coords) {
		return this.convert(coords, version);
	}

	public RegionSet convert(CoordinateMap coords, DocumentVersion ver) {
		if (isPoint()) {
			int p = coords.mapStart(getStartLow());
			return make(p, p, p, p, ver);
		} else
			return make(
				coords.mapStart(getStartLow()),
				coords.mapStart(getStartHigh()),
				coords.mapEnd(getEndLow()),
				coords.mapEnd(getEndHigh()),
				ver);
	}
	public RegionSet convert(CoordinateMap coords, Document doc) {
		return this.convert(coords, doc.getVersion());
	}

    public RegionSet convert (Versions vers) {
        DocumentVersion targetVer = vers.getVersion(this.getDocument());
        return targetVer.convert(this);
    }
    
    public RegionSet convertToLatest() {
        DocumentVersion targetVer = this.getDocument().getVersion();
        return targetVer.convert(this); 
    }

	public int fuzzify(int x, int direction) {
		if (getStartLow() <= x && x <= getEndHigh())
			return (direction == FUZZY_LOW) ? getStartLow() : getEndHigh();
		else
			return x;
	}

	//
	// Delimiter operators
	//

	public RegionSet beforeDelimited(int lastEndHigh) {
		return make(
			lastEndHigh,
			getStartLow(),
			lastEndHigh,
			getStartLow(),
			version);
	}

	public RegionSet afterDelimited(int nextStartLow) {
		return make(
			getEndHigh(),
			nextStartLow,
			getEndHigh(),
			nextStartLow,
			version);
	}

	public RegionSet justBeforeDelimited(int lastEndHigh, RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			lastEndHigh,
			getStartLow(),
			currbkgd.fuzzify(getStartLow(), FUZZY_LOW),
			getStartLow(),
			version);
	}

	public RegionSet justAfterDelimited(int nextStartLow, RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			getEndHigh(),
			currbkgd.fuzzify(getEndHigh(), FUZZY_HIGH),
			getEndHigh(),
			nextStartLow,
			version);
	}

	public RegionSet startingDelimited(int nextStartLow, RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			currbkgd.fuzzify(getStartLow(), FUZZY_LOW),
			getStartHigh(),
			getEndLow(),
			nextStartLow,
			version);
	}

	public RegionSet endingDelimited(int lastEndHigh, RegionSet bkgd) {
		SingleDocRegionSet currbkgd = bkgd.getSingleDocRegionSet(getDocument());

		return make(
			lastEndHigh,
			getStartHigh(),
			getEndLow(),
			currbkgd.fuzzify(getEndHigh(), FUZZY_HIGH),
			version);
	}

	public SubsetEnumeration singleDocize() {
		return new SubsetEnumeration() {
			public RegionSet firstFast() {
				return Region.this;
			}
			public RegionSet nextFast() {
				return null;
			}
		};
	}
	// Bounding box 
	public Region getBoundingBox() {
		if (this.isEmpty())
			return Region.EMPTY;
		else
			return new FuzzyRegion(
				getStartLow(),
				getStartHigh(),
				getEndLow(),
				getEndHigh(),
				version);
	}
	public boolean lessThan(RegionSet r) {
		if (r.isEmpty())
			return true;

		debug.assertion(r instanceof SingleDocRegionSet);
		Region s = ((SingleDocRegionSet) r).getBoundingBox();
		debug.assertion(this.sameDocument(s));

		int x, sx;

		x = getStartLow();
		sx = s.getStartLow();
		if (x != sx)
			return x < sx;

		x = getEndLow();
		sx = s.getEndLow();
		if (x != sx)
			return x < sx;

		x = getStartHigh();
		sx = s.getStartHigh();
		if (x != sx)
			return x < sx;

		return getEndHigh() < s.getEndHigh();
	}

	// document-related
	public DocumentVersion getVersion() {
		return this.version;
	}

	public Document getDocument() {
		//TODO: change to debug.assertion(this.version != null)
		if (this.version == null)
			return null;
		else
			return this.version.getDocument();
	}

	public SingleDocRegionSet getSingleDocRegionSet(Document doc) {
		if (this.version == null)
			return Region.EMPTY;

		if (this.version.getDocument() == doc)
			return this;
		else
			return Region.EMPTY;
	}

    public Versions getVersions() {
        return new SingleDocVersions(this.version);
    }

	private boolean sameDocumentVersion(Region r) {
		DocumentVersion ver1 = this.getVersion();
		DocumentVersion ver2 = r.getVersion();
		if (ver1 == null || ver2 == null)
			return false;
		else
			return ver1.equals(ver2);
	}
	private boolean sameDocument(Region r) {
		return (this.getDocument() == r.getDocument());
	}
    
    
}
