/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */



package lapis;

/**
 * Enumerates the regions in a RegionSet, in lexicographic order.
 * (See RegionSet.lessThan () for a definition of lexicographic
 * order -- it sorts first by startLow, then startHigh, then
 * endLow, then endHigh.)
 * <P>
 * The correct calling protocol for RegionEnumeration: call first()
 * to get the first Region, then next () to get subsequent regions.
 * The enumeration ends by returning null.
 * <P>
 * Here is the standard idiom:
 * <pre>
 * RegionSet set; // initialized somehow
 * RegionEnumeration e = set.regions ();
 * for (Region r = e.first (); r != null; r = e.next ()) {
 *    // use r
 * }
 * </pre>
 * <P>
 * The first/next methods always return an immutable Region, which you
 * can store in other data structures and local variables without fear that
 * it will change.   However, for some RegionSet implementations (such
 * as RegionArray), this requires allocating a fresh Region object.
 * To avoid the allocation, you can use the firstFast/nextFast methods instead.
 * These methods may return a Region object internal to 
 * the RegionEnumeration which is overwritten on the next iteration of the
 * enumeration.  If you use these methods and you want to reference the
 * returned Region later (in a different iteration of the enumeration, or
 * after the enumeration has ended), call freeze () to get an immutable copy.
 * <P>
 * Example of firstFast/nextFast:
 * <pre>
 * RegionSet set; // initialized somehow
 * RegionEnumeration e = set.regions ();
 * Region last = null; // region returned by previous iteration
 * for (Region r = e.firstFast (); r != null; r = e.nextFast ()) {
 *    // use r here.  It's safe to call any Region methods on r and
 *    // insert r into other RegionSets or Relations.
 * 
 *    // In case you need to save r for the next iteration, call freeze ():
 *    last = r.freeze ();
 * }
 * </pre>
 */
public interface RegionEnumeration {
    /**
     * Get the first region in the enumeration.
     * Can be called only once, just after the enumeration is created.
     * @return immutable Region, or null if enumeration is exhausted
     */
    public abstract Region first ();

    /**
     * Get the next region in the enumeration.
     * @return immutable Region, or null if enumeration is exhausted
     */
    public abstract Region next ();

    /**
     * Get the first region in the enumeration without allocating a new
     * Region ojbect.
     * Can be called only once, just after the enumeration is created.
     * @return potentially-mutable Region, or null if enumeration is exhausted
     */
    public abstract Region firstFast ();

    /**
     * Get the next region in the enumeration without allocating a new
     * Region object.
     * @return potentially-mutable Region, or null if enumeration is exhausted
     */
    public abstract Region nextFast ();
}
