/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis;

import java.util.*;
import java.io.*;
import lapisx.util.Str;
import lapisx.util.Debug;

public abstract class RegionSet implements Serializable {
	public static Debug debug = Debug.QUIET;

	public final static int MIN_INDEX = 0;
	public static int MAX_INDEX = Integer.MAX_VALUE / 2;

	/**
	 * Canonical empty region set.
	 */

	public final static EmptyRegionSet EMPTY = new EmptyRegionSet();

	/**
	 * Canonical universal region set, containing every possible region.
	 */
	public static Region getUniversal(Document doc) {
		if (doc == null)
			return Region.EMPTY;
		else
			return new FuzzyRegion(
				MIN_INDEX,
				MAX_INDEX,
				MIN_INDEX,
				MAX_INDEX,
				doc);
	}

	public String toString() {
		StringBuffer result = new StringBuffer();
		result.append(getClass().getName()).append("\n");
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet r = se.firstFast(); r != null; r = se.nextFast()) {
			result.append('[').append(
				((SingleDocRegionSet) r).getStartLow()).append(
				'-');
			result.append(((SingleDocRegionSet) r).getStartHigh()).append(',');
			result.append(((SingleDocRegionSet) r).getEndLow()).append('-');
			result.append(((SingleDocRegionSet) r).getEndHigh()).append(']');
			result.append(((SingleDocRegionSet) r).getVersion()).append("\n");
		}
		return result.toString();
	}

	/**
	 * Test whether two RegionSets' bounding boxes intersect, indicating
	 * that there's a possibility that the RegionSets themselves intersect.
	 * @return true if bounding boxes intersect
	 */

	public boolean mightIntersect(RegionSet rs) {
		// TODO: enumerate RegionSet that contains less documents 
		//		 by checking the size of DocumentVersion list
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet thisr = se.firstFast();
			thisr != null;
			thisr = se.nextFast()) {
			SingleDocRegionSet rsr =
				rs.getSingleDocRegionSet(
					((SingleDocRegionSet) thisr).getDocument());
			if (rsr.getStartLow()
				<= ((SingleDocRegionSet) thisr).getStartHigh()
				&& ((SingleDocRegionSet) thisr).getStartLow()
					<= rsr.getStartHigh()
				&& rsr.getEndLow() <= ((SingleDocRegionSet) thisr).getEndHigh()
				&& ((SingleDocRegionSet) thisr).getEndLow() <= rsr.getEndHigh())
				return true;
		}
		return false;
	}

	public boolean mightEnclose(RegionSet rs) {
		// TODO: enumerate RegionSet that contains less documents 
		//		 by checking the size of DocumentVersion list
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet thisr = se.firstFast();
			thisr != null;
			thisr = se.nextFast()) {
			SingleDocRegionSet rsr =
				rs.getSingleDocRegionSet(
					((SingleDocRegionSet) thisr).getDocument());
			if (((SingleDocRegionSet) thisr).getStartLow() <= rsr.getStartLow()
				&& rsr.getStartHigh()
					<= ((SingleDocRegionSet) thisr).getStartHigh()
				&& ((SingleDocRegionSet) thisr).getEndLow() <= rsr.getEndLow()
				&& rsr.getEndHigh() <= ((SingleDocRegionSet) thisr).getEndHigh())
				return true;
		}
		return false;
	}

	//
	// Region set identity
	//
	public boolean equals(Object o) {
		if (!(o instanceof RegionSet))
			return false;

		RegionSet s = (RegionSet) o;

		if (s.getSize() != getSize())
			return false;

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet thisr = se.firstFast();
			thisr != null;
			thisr = se.nextFast()) {
			SingleDocRegionSet sr =
				s.getSingleDocRegionSet(
					((SingleDocRegionSet) thisr).getDocument());
			if (!sr
				.getBoundingBox()
				.equals(((SingleDocRegionSet) thisr).getBoundingBox()))
				return false;
		}

		RegionEnumeration re1 = regions();
		RegionEnumeration re2 = s.regions();
		Region r1, r2;
		for (r1 = re1.firstFast(), r2 = re2.firstFast();
			r1 != null && r2 != null;
			r1 = re1.nextFast(), r2 = re2.nextFast()) {
			if (!r1.equals(r2))
				return false;
		}

		return (r1 == null && r2 == null);
	}

	public boolean slowEquals(Object o) {
		if (!(o instanceof RegionSet))
			return false;

		RegionSet s = (RegionSet) o;

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet currR = se.firstFast();
			currR != null;
			currR = se.nextFast()) {
			RegionSet currS =
				(RegionSet) s.getSingleDocRegionSet(
					((SingleDocRegionSet) currR).getDocument());
			if (currS == null)
				return false;
			if (currR.isEmpty() && currS.isEmpty())
				continue;
			if (!currR.isEmpty() && !currS.isEmpty()) {
				StandardRectangles srR = new StandardRectangles(currR);
				ArrayList lR = srR.standardize();
				StandardRectangles srS = new StandardRectangles(currS);
				ArrayList lS = srS.standardize();

				int sizeR = lR.size();
				if (lS.size() != sizeR)
					return false;
				for (int i = 0; i < sizeR; i++)
					if (!((StandardRectangles.Rect) lR.get(i))
						.equals((StandardRectangles.Rect) lS.get(i)))
						return false;
			} else {
				return false;
			}
		}

		return true;
	}

	public int hashCode() {
		int result = 17;
		RegionEnumeration re = regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
			result = 37 * result + r.getStartLow();
			result = 37 * result + r.getStartHigh();
			result = 37 * result + r.getEndLow();
			result = 37 * result + r.getEndHigh();
		}
		return result;
	}

	//
	// Enumerating primitive region rectangles
	//

	public abstract int getSize();

	public boolean isEmpty() {
		return getSize() == 0;
	}

	public boolean isAllPoints() {
		RegionEnumeration re = regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast())
			if (!r.isPoint())
				return false;
		return true;
	}

	public boolean isCrisp() {
		RegionEnumeration re = regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast())
			if (!r.isCrisp())
				return false;
		return true;
	}

	public boolean isFlat() {
		int max = 0;
		RegionEnumeration re = regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
			if (!r.isCrisp() || r.getStart() < max)
				return false;
			max = r.getEnd();
		}
		return true;
	}

	public abstract RegionEnumeration regions(Region bbox);

    /**
     * enumerate regions in each bounding box from each document 
     * @param bboxs region set consisting of bounding boxes, one from each document
     * @return region enumeration
     */
	public abstract RegionEnumeration regions(RegionSet bboxs);

	public RegionEnumeration regions() {
        Iterator idoc = this.getDocuments();
        MutableRegionSet bboxs = new MutableRegionSet();
        while (idoc.hasNext())
            bboxs.insert(new FuzzyRegion(MIN_INDEX, MAX_INDEX, MIN_INDEX, MAX_INDEX, (Document)idoc.next()));
        
		return this.regions(bboxs);
	}

	//
	// Enumerating subsets
	//

	public SubsetEnumeration subsets() {
		return new SubsetEnumeration() {
			RegionEnumeration e = regions();
			public RegionSet firstFast() {
				return e.firstFast();
			}
			public RegionSet nextFast() {
				return e.nextFast();
			}
		};
	}

	/**
	 * Make a copy of this region set.
	 * @return an instance of this RegionSet that nobody else can modify (may be the same object
	 * reference, if this RegionSet subclass is immutable)
	 */
	public abstract RegionSet freeze();

	/**
	 * Combine two RegionSets by an operation.
	 */

	public MutableRegionSet intersect(
		final IntersectOp op,
		RegionSet u,
		MutableRegionSet result) {
		if (!mightIntersect(u)) {
			// no intersection possible
		} else if (u instanceof Region) {
			// expand me into regions overlapping u
			Region s = (Region) u;
			RegionEnumeration e = regions(s);
			for (Region r = e.firstFast(); r != null; r = e.nextFast()) {
				op.combine(r, s, result);
			}
		} else if (mightEnclose(u)) {
			// expand me into subsets
			SubsetEnumeration children = subsets();
			for (RegionSet child = children.firstFast();
				child != null;
				child = children.nextFast())
				result = child.intersect(op, u, result);
		} else {
			// expand u into subsets
			SubsetEnumeration children = u.subsets();
			for (RegionSet child = children.firstFast();
				child != null;
				child = children.nextFast())
				result = intersect(op, child, result);
		}

		return result;
	}

	/**
	 * Operation for intersect().
	 */
	public static interface IntersectOp {
		public void combine(Region r, Region s, MutableRegionSet result);
	}

	/**
	 * Test whether two RegionSets actually intersect.
	 * @return true iff some region is a member of both this and r
	 */
	public boolean intersects(RegionSet r) {
		try {
			intersect(INTERSECT_TEST, r, null);

			// if intersect didn't throw an exception, then INTERSECT_TEST never ran,
			// so no intersection was found
			return false;
		} catch (IntersectedException e) {
			return true;
		}
	}

	static IntersectOp INTERSECT_TEST = new IntersectOp() {
		public void combine(Region r, Region s, MutableRegionSet result) {
			throw new IntersectedException();
		}
	};

	static class IntersectedException extends RuntimeException {
	}

	/**
	 * Fast test for whether this RegionSet actually intersects a given Region.
	 * @return true iff r is a member of this
	 */
	public boolean intersects(Region r) {
		return regions(r).firstFast() != null;
	}

	/**
	 * Count number of Regions in this region set that intersect the given Region.
	 * @return number of Regions in this region set that intersect r
	 */
	public int countIntersection(Region r) {
		int n = 0;
		RegionEnumeration re = regions(r);
		for (Region t = re.firstFast(); t != null; t = re.nextFast())
			++n;
		return n;
	}

	/*
	 * Comparing two region sets.
	 */

	public static final int EQUAL = 0;
	public static final int SUBSET = 1;
	public static final int SUPERSET = 2;
	public static final int DISJOINT = 3;
	public static final int UNRELATED = 4;

	public int compareTo(RegionSet u) {
		RegionEnumeration e = regions();
		RegionEnumeration eu = u.regions();

		Region r, s;
		boolean skippedR = false;
		boolean skippedS = false;
		int cmp = DISJOINT;
		for (r = e.firstFast(), s = eu.firstFast(); r != null && s != null;) {
			if (r.mightIntersect(s)) {
				boolean rEnclosesS = r.mightEnclose(s);
				boolean sEnclosesR = s.mightEnclose(r);
				if (rEnclosesS && sEnclosesR) {
					if (cmp == DISJOINT)
						cmp = EQUAL;
					r = e.nextFast();
					s = eu.nextFast();
				} else if (rEnclosesS) {
					if (cmp == SUBSET || skippedS)
						return UNRELATED;
					else
						cmp = SUPERSET;
					do s = eu.nextFast();
					while (s != null && r.mightEnclose(s));
					r = e.nextFast();
				} else if (sEnclosesR) {
					if (cmp == SUPERSET || skippedR)
						return UNRELATED;
					else
						cmp = SUBSET;
					do r = e.nextFast();
					while (r != null && s.mightEnclose(r));
					s = eu.nextFast();
				} else {
					return UNRELATED;
				}
			} else if (r.lessThan(s)) {
				if (cmp == SUBSET)
					return UNRELATED;
				else if (cmp == EQUAL)
					cmp = SUPERSET;
				r = e.nextFast();
				skippedR = true;
			} else {
				// s less than r
				if (cmp == SUPERSET)
					return UNRELATED;
				else if (cmp == EQUAL)
					cmp = SUBSET;
				s = eu.nextFast();
				skippedS = true;
			}
		}

		if (r != null) {
			if (cmp == SUBSET)
				return UNRELATED;
			else if (cmp == EQUAL)
				cmp = SUPERSET;
		}
		if (s != null) {
			if (cmp == SUPERSET)
				return UNRELATED;
			else if (cmp == EQUAL)
				cmp = SUBSET;
		}

		return cmp;
	}

	/*
	 * Set operations
	 */

	public RegionSet or(RegionSet u) {
		if (u.isEmpty())
			return this;
		else if (isEmpty())
			return u;
		else if (getSize() > u.getSize()) {
			MutableRegionSet m = new MutableRegionSet(this);
			m.insert(u);
			return m;
		} else {
			MutableRegionSet m = new MutableRegionSet(u);
			m.insert(this);
			return m;
		}
	}

	public RegionSet and(RegionSet u) {
		if (u.isEmpty() || isEmpty())
			return EMPTY;
		else {
			MutableRegionSet m = new MutableRegionSet();
			if (getSize() > u.getSize())
				intersect(AND, u, m);
			else
				u.intersect(AND, this, m);
			return m;
		}
	}

	static IntersectOp AND = new IntersectOp() {
		public void combine(Region r, Region s, MutableRegionSet result) {
			result.insert(r.and(s));
		}
	};

	public RegionSet butnot(RegionSet u) {
		if (u.isEmpty())
			return this;

		MutableRegionSet m = new MutableRegionSet(this);
		m.delete(u);
		return m;
	}

	/*
	 * RegionSet relations
	 */

	public RegionSet equalTo(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.equalTo(bkgd));
		return result;
	}

	public RegionSet before() {
		int rightmost = MIN_INDEX - 1;
		MutableRegionSet ms = new MutableRegionSet();
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			RegionEnumeration e = rs.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast())
				rightmost = Math.max(rightmost, r.getStartHigh());
			if (rightmost >= MIN_INDEX)
				ms.insert(
					new FuzzyRegion(
						MIN_INDEX,
						rightmost,
						MIN_INDEX,
						rightmost,
						((SingleDocRegionSet) rs).getVersion()));
		}
		return ms;
	}

	public RegionSet beforeIn(RegionSet context) {
		return intersect(BEFORE, context, new MutableRegionSet());
	}

	static IntersectOp BEFORE = new IntersectOp() {
		public void combine(Region r, Region s, MutableRegionSet result) {
			result.insert(r.beforeIn(s));
		}
	};

	public RegionSet justBefore(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.justBefore(bkgd));
		return result;
	}

	public RegionSet after() {
		int leftmost = MAX_INDEX + 1;
		MutableRegionSet ms = new MutableRegionSet();
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			RegionEnumeration e = rs.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast())
				leftmost = Math.min(leftmost, r.getEndLow());
			if (leftmost <= MAX_INDEX)
				ms.insert(
					new FuzzyRegion(
						leftmost,
						MAX_INDEX,
						leftmost,
						MAX_INDEX,
						((SingleDocRegionSet) rs).getVersion()));
		}
		return ms;
	}

	public RegionSet afterIn(RegionSet context) {
		return intersect(AFTER, context, new MutableRegionSet());
	}

	static IntersectOp AFTER = new IntersectOp() {
		public void combine(Region r, Region s, MutableRegionSet result) {
			result.insert(r.afterIn(s));
		}
	};

	public RegionSet justAfter(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.justAfter(bkgd));
		return result;
	}

	public RegionSet contains() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.contains());
		return result;
	}

	public RegionSet starting(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.starting(bkgd));
		return result;
	}

	public RegionSet ending(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.ending(bkgd));
		return result;
	}

	public RegionSet startingIn(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) (r.startingIn(bkgd)));
		return result;
	}

	public RegionSet endingIn(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) (r.endingIn(bkgd)));
		return result;
	}

	public RegionSet startingContains(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) (r.startingContains(bkgd)));
		return result;
	}

	public RegionSet endingContains(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) (r.endingContains(bkgd)));
		return result;
	}

	public RegionSet descendant() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next()) {
			if (!r.isPoint()) {
				int start = r.getStartLow();
				int end = r.getEndHigh();
				Region r1 = new Region(start, end - 1, r.getVersion());
				result.insert((Region) (r1.startingIn(EMPTY)));
				Region r2 = new Region(start + 1, end, r.getVersion());
				result.insert((Region) (r2.in()));
			}
		}
		return result;
	}

	public RegionSet ancestor() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next()) {
			if (!r.equals((Region) RegionSet.getUniversal(r.getDocument()))) {
				int start = r.getStartHigh();
				int end = r.getEndLow();
				if (start == MIN_INDEX) {
					Region r1 = new Region(MIN_INDEX, end + 1, r.getVersion());
					result.insert((Region) (r1.contains()));
				} else if (end == MAX_INDEX) {
					Region r1 =
						new Region(start - 1, MAX_INDEX, r.getVersion());
					result.insert((Region) (r1.contains()));
				} else if (start <= end) {
					Region r1 = new Region(start - 1, end, r.getVersion());
					result.insert((Region) (r1.endingContains(EMPTY)));
					Region r2 = new Region(start, end + 1, r.getVersion());
					result.insert((Region) (r2.contains()));
				} else {
					int exclusivePointStart = Math.max(r.getStartLow(), end);
					int exclusivePointEnd = Math.min(r.getEndHigh(), start);
					result.insert(
						new FuzzyRegion(
							0,
							exclusivePointStart - 1,
							exclusivePointStart,
							MAX_INDEX,
							r.getVersion()));
					for (int i = exclusivePointStart;
						i <= exclusivePointEnd;
						i++)
						result.insert(
							new FuzzyRegion(
								i,
								i,
								i + 1,
								MAX_INDEX,
								r.getVersion()));
				}
			}
		}
		return result;
	}

	public RegionSet in() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next()) {
			result.insert((Region) r.in());
		}
		return result;
	}

	public RegionSet overlapsStart() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.overlapsStart());
		return result;
	}

	public RegionSet overlapsEnd() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.overlapsEnd());
		return result;
	}

	public RegionSet overlaps() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.overlaps());
		return result;
	}

	//
	// Delimiter operators
	//

	public RegionSet beforeDelimited() {
		MutableRegionSet result = new MutableRegionSet();

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			int lastEndHigh = MIN_INDEX;
			RegionEnumeration e = rs.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast()) {
				result.insert(r.beforeDelimited(lastEndHigh));
				lastEndHigh = r.getEndHigh();
			}
		}

		return result;
	}

	public RegionSet afterDelimited() {
		MutableRegionSet result = new MutableRegionSet();

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			Region last = null;
			RegionEnumeration e = rs.regions();
			for (Region r = e.first(); r != null; r = e.next()) {
				if (last != null)
					result.insert(last.afterDelimited(r.getStartLow()));
				last = r;
			}
			if (last != null)
				result.insert(last.afterDelimited(MAX_INDEX));
		}

		return result;
	}

	public RegionSet justBeforeDelimited(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			int lastEndHigh = MIN_INDEX;
			RegionEnumeration e = rs.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast()) {
				result.insert(r.justBeforeDelimited(lastEndHigh, bkgd));
				lastEndHigh = r.getEndHigh();
			}
		}

		return result;
	}

	public RegionSet justAfterDelimited(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			Region last = null;
			RegionEnumeration e = rs.regions();
			for (Region r = e.first(); r != null; r = e.next()) {
				if (last != null)
					result.insert(
						last.justAfterDelimited(r.getStartLow(), bkgd));
				last = r;
			}
			if (last != null)
				result.insert(last.justAfterDelimited(MAX_INDEX, bkgd));
		}

		return result;
	}

	public RegionSet startingDelimited(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			Region last = null;
			RegionEnumeration e = rs.regions();
			for (Region r = e.first(); r != null; r = e.next()) {
				if (last != null)
					result.insert(
						last.startingDelimited(r.getStartLow(), bkgd));
				last = r;
			}
			if (last != null)
				result.insert(last.startingDelimited(MAX_INDEX, bkgd));
		}

		return result;
	}

	public RegionSet endingDelimited(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			int lastEndHigh = MIN_INDEX;
			RegionEnumeration e = rs.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast()) {
				result.insert(r.endingDelimited(lastEndHigh, bkgd));
				lastEndHigh = r.getEndHigh();
			}
		}

		return result;
	}

	//
	// Other operators
	//

	public RegionSet startOf() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next()) {
			int start = r.getStart();
			result.insert(start, start, r.getVersion());
		}
		return result;
	}

	public RegionSet endOf() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next()) {
			int end = r.getEnd();
			result.insert(end, end, r.getVersion());
		}
		return result;
	}

	public RegionSet nextTo(RegionSet us, RegionSet bkgd) {
		if (this.isEmpty() || us.isEmpty())
			return EMPTY;

		// do nextTo operation based on SingleDocBased region set
		MutableRegionSet result = new MutableRegionSet();

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			RegionSet ru =
				(RegionSet) us.getSingleDocRegionSet(
					((SingleDocRegionSet) rs).getDocument());
			MutableRegionSet x = new MutableRegionSet();
			int size = rs.getSize();
			int usize = ru.getSize();

			// convert the smaller of this and u into a predicate region
			// set, and remember the original regions corresponding to each
			// predicate.
			RegionSet y = null;
			final Relation originalRegions = new Relation();
			boolean reversed;

			if (size > usize) {
				// x = u.justBefore ()
				// y = this
				RegionEnumeration e = ru.regions();
				for (Region r = e.firstFast(); r != null; r = e.nextFast()) {
					Region jbefR = (Region) r.justBefore(bkgd);
					if (!originalRegions.add(jbefR, r))
						x.insert(jbefR);
				}
				y = rs;
				reversed = true;
			} else {
				// x = this.justAfter ()
				// y = u
				RegionEnumeration e = rs.regions();
				for (Region r = e.firstFast(); r != null; r = e.nextFast()) {
					Region jaftR = (Region) r.justAfter(bkgd);
					if (!originalRegions.add(jaftR, r))
						x.insert(jaftR);
				}
				y = ru;
				reversed = false;
			}

			final boolean isReversed = reversed;

			// intersect x and y to find all x's that are adjacent to y.
			IntersectOp op = new IntersectOp() {
				public void combine(Region r, Region s, MutableRegionSet ms) {
						// map r back to its original region (or set of regions)
		// and construct a region spanning both regions
	RegionEnumeration re = originalRegions.get(r).regions();
					for (Region t = re.first(); t != null; t = re.next())
						ms.insert(isReversed ? s.fromTo(t) : t.fromTo(s));
				}
			};
			result.insert(x.intersect(op, y, new MutableRegionSet()));
		}
		return result;
	}

	public RegionSet separatedBy() {
		MutableRegionSet result = new MutableRegionSet();

		SubsetEnumeration se = this.singleDocize();
		for (RegionSet s = se.firstFast(); s != null; s = se.nextFast()) {
			int lastEndHigh = MIN_INDEX;
			RegionEnumeration e = s.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast()) {
				result.insert((Region) r.beforeDelimited(lastEndHigh));
				lastEndHigh = r.getEndHigh();
			}
			result.insert(
				new FuzzyRegion(
					lastEndHigh,
					MAX_INDEX,
					lastEndHigh,
					MAX_INDEX,
					((SingleDocRegionSet) s).getVersion()));
		}
		return result;
	}

	public RegionSet balancedWith(RegionSet closeDelimiter) {
		if (closeDelimiter == null)
			closeDelimiter = this;

		MutableRegionSet result = new MutableRegionSet();

		// do balancedWith operation based on SingleDocBased region set
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			RegionSet close =
				(RegionSet) closeDelimiter.getSingleDocRegionSet(
					((SingleDocRegionSet) rs).getDocument());
			Stack stack = new Stack();

			RegionEnumeration eOpen = rs.regions();
			Region rOpen = eOpen.first();
			if (rOpen == null)
				continue;

			RegionEnumeration eClose = close.regions();
			Region rClose = eClose.first();
			if (rClose == null)
				continue;

			while (rOpen != null || rClose != null) {
				//RegionSet r;
				Region r;
				boolean open;

				if (rOpen == null || rClose == null)
					open = (rOpen != null);
				else if (rOpen.lessThan(rClose))
					open = true;
				else if (rClose.lessThan(rOpen))
					open = false;
				else
					open = stack.isEmpty();

				if (open) {
					r = rOpen;
					stack.push(rOpen);
				} else {
					r = rClose;
					if (!stack.isEmpty()) {
						Region rMatchingOpen = (Region) stack.pop();
						Region spanner = (Region) rMatchingOpen.fromTo(rClose);
						result.insert(spanner);
					}
				}

				// advance rOpen and rClose past end of current region
				int end = r.getEndLow();
				while (rOpen != null && rOpen.getStartLow() < end)
					rOpen = eOpen.next();
				while (rClose != null && rClose.getStartLow() < end)
					rClose = eClose.next();
			}
		}
		return result;
	}

	public RegionSet fromTo(RegionSet closeDelimiter) {
		MutableRegionSet result = new MutableRegionSet();

		// do fromTo operation based on SingleDocBased region set
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			RegionSet close =
				(RegionSet) closeDelimiter.getSingleDocRegionSet(
					((SingleDocRegionSet) rs).getDocument());

			RegionEnumeration eOpen = rs.regions();
			Region rOpen = eOpen.firstFast();
			if (rOpen == null)
				continue;

			RegionEnumeration eClose = close.regions();
			Region rClose = eClose.firstFast();
			if (rClose == null)
				continue;

			int p;
			while (rOpen != null && rClose != null) {
				// advance rClose past rOpen
				p = rOpen.getEnd();
				while (rClose != null && rClose.getStart() < p)
					rClose = eClose.nextFast();
				if (rClose == null)
					break;

				// add [rOpen..rClose] to result
				result.insert((Region) rOpen.fromTo(rClose));

				// advance rOpen at least once
				rOpen = eOpen.nextFast();

				// advance rOpen  past rClose
				p = rClose.getEnd();
				while (rOpen != null && rOpen.getStart() < p)
					rOpen = eOpen.nextFast();
			}
		}
		return result;
	}

	public RegionSet trim(RegionSet bkgd) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert(r.trim(bkgd));
		return result;
	}

	public RegionSet include(RegionSet units) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.include(units));
		return result;
	}

	public RegionSet nth(int n, RegionSet context) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration eContext = context.regions();
		for (Region c = eContext.first(); c != null; c = eContext.next()) {
			RegionEnumeration e = regions(c);
			Region r;

			if (n >= 0) {
				int i = 0;
				for (r = e.firstFast(), ++i;
					i < n && r != null;
					r = e.nextFast(), ++i);
				if (r != null)
					result.insert(r);
			} else if (n == -1) {
				RegionSet last = null;
				for (r = e.first(); r != null; r = e.next())
					last = r;
				if (last != null)
					result.insert(last);
			} else {
				// n < -1: we want to count -n regions from
				// the end of the region set. 
				// Since we can't iterate region sets backward, 
				// we'll keep track of the last N regions seen
				// (N == -n) as we enumerate forwards.
				int N = -n;
				RegionSet[] window = new RegionSet[N];
				// circular queue
				int pos = 0;
				// Next entry in window to fill.  

				// Once we've iterated at least N regions, then
				// window[pos] will always contain the oldest
				// region in window -- in other words, the region
				// that was seen N iterations ago.
				for (r = e.first(); r != null; r = e.next()) {
					window[pos] = r;
					if (++pos == N)
						pos = 0;
				}
				if (window[pos] != null)
					result.insert(window[pos]);
			}
		}

		return result;
	}

	public RegionSet crispen() {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration e = regions();
		for (Region r = e.first(); r != null; r = e.next())
			result.insert((Region) r.crispen());
		return result;
	}

	public RegionSet flatten() {
		MutableRegionSet result = new MutableRegionSet();
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			MutableRegionSet ms =
				MutableRegionSet.makeFlattener(
					rs.getSize(),
					false,
					((SingleDocRegionSet) rs).getVersion());
			RegionEnumeration re = rs.regions();
			for (Region r = re.firstFast(); r != null; r = re.nextFast())
				ms.insert(r);

			result.insert(ms);
		}
		return result;
	}

	public RegionSet melt() {
		MutableRegionSet result = new MutableRegionSet();
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			MutableRegionSet ms =
				MutableRegionSet.makeFlattener(
					rs.getSize(),
					true,
					((SingleDocRegionSet) rs).getVersion());
			RegionEnumeration re = this.regions();
			for (Region r = re.firstFast(); r != null; r = re.nextFast())
				ms.insert(r);

			result.insert(ms);
		}
		return result;
	}

	public RegionSet stripEmptyRegions() {
		RegionEnumeration e = regions();
		for (Region r = e.firstFast(); r != null; r = e.nextFast())
			if (r.isPoint()) {
				MutableRegionSet result = new MutableRegionSet();
				e = regions();
				for (r = e.firstFast(); r != null; r = e.nextFast())
					if (!r.isPoint())
						result.insert(r);
				return result;
			}
		// no point regions were found, so nothing to strip
		return this;
	}

	//
	// Debugging reps
	//

	/**
	 * Convert region set to a multi-line string, one region per line.
	 * Each region is emitted as [x,y] followed by the content spanned
	 * by [x,y], taken from doc, abbreviated to at most 70 characters
	 * using lapisx.util.Str.abbreviate().
	 * @param doc Document to use for content.  May be null, in which case
	 * only region coordinates are printed to the string (with no content).
	 */
	public String toStringWithContent() {
		StringBuffer out = new StringBuffer();
		RegionEnumeration e = regions();
		for (Region r = e.firstFast(); r != null; r = e.nextFast()) {
			Document doc = r.getDocument();
			out.append(r.toString());
			if (doc != null)
				out.append(Str.abbreviate(doc.getText(r), 70));
			out.append('\n');
		}
		return out.toString();
	}

	/**
	 * Get the text of a document with all the regions in this region
	 * set surrounded by square brackets [...].  Designed for debugging,
	 * slow implementation, only works for flat region sets at the moment.
	 */
	public String highlight(Document doc) {
		String content = doc.getText();
		StringBuffer out = new StringBuffer(content);
		int l = content.length();
		int adjust = 0; // number of brackets already inserted
		RegionEnumeration e = regions();
		for (Region r = e.firstFast(); r != null; r = e.nextFast()) {
			int start = Math.min(l, Math.max(0, r.getStart()));
			int end = Math.min(l, Math.max(0, r.getEnd()));
			out.insert(start + adjust, '[');
			++adjust;
			out.insert(end + adjust, ']');
			++adjust;
		}
		return out.toString();
	}

	public abstract SubsetEnumeration singleDocize();

	public abstract SingleDocRegionSet getSingleDocRegionSet(Document doc);

    public abstract Versions getVersions ();
    
    public Iterator getDocuments () {
        return this.getVersions().documents();
    }
     
    public RegionSet convert (Versions vers) {
        if (this.getVersions().compatible(vers))
            return this;
        
        MutableRegionSet result = new MutableRegionSet();
        SubsetEnumeration se = this.singleDocize();
        for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
            Document sourceDoc = ((SingleDocRegionSet)rs).getDocument();
            DocumentVersion targetVer = vers.getVersion(sourceDoc);
            if (targetVer == null)
                result.insert(rs);
            else
                result.insert(targetVer.convert((SingleDocRegionSet)rs));
        }
        return result;
    }

	public RegionSet convert(RegionSet other) {
        return this.convert(other.getVersions());
	}

	public RegionSet convertToLatest() {
		MutableRegionSet result = new MutableRegionSet();
		SubsetEnumeration se = this.singleDocize();
		for (RegionSet rs = se.firstFast(); rs != null; rs = se.nextFast()) {
			DocumentVersion targetVer = ((SingleDocRegionSet)rs).getDocument().getVersion();
			result.insert(targetVer.convert((SingleDocRegionSet) rs));
		}
		return result;
	}
}
