/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis;

/**
 * SingleDocRegionSet is RegionSet that contains regions from single document.
 */
public abstract class SingleDocRegionSet extends RegionSet {

    /**
     * @return lower bound for starting offset
     */
    public abstract int getStartLow ();
    
    /**
     * @return upper bound for starting offset
     */
    public abstract int getStartHigh ();
    
    /**
     * @return lower bound for ending offset
     */
    public abstract int getEndLow ();
    
    /**
     * @return upper bound for ending offset
     */
    public abstract int getEndHigh ();
    
    /**
     * @return document
     */
    public abstract Document getDocument ();
    
    /**
     * @return document version 
     */
    public abstract DocumentVersion getVersion ();

    /**
     * @return bound for this region set
     */
    public Region getBoundingBox () {
        if (this.isEmpty())
            return Region.EMPTY;
        else
            return new FuzzyRegion (getStartLow(), 
                                    getStartHigh(), 
                                    getEndLow(), 
                                    getEndHigh(),
                                    getVersion());
        
    }
    
    /**
     * test if this is less than r in lexicographic order
     * @param r is a instance of SingleDocRegionSet 
     * and has the same document as this
     * @return true is this is less than r, false otherwise
     */
    public boolean lessThan (RegionSet r) {
        if (r.isEmpty())
            return true;
            
        debug.assertion(r instanceof SingleDocRegionSet);
        Region s = ((SingleDocRegionSet)r).getBoundingBox();
        debug.assertion(this.getDocument() == s.getDocument());
        
        int x, sx;

        x = getStartLow ();
        sx = s.getStartLow ();
        if (x != sx)
            return x < sx;

        x = getEndLow ();
        sx = s.getEndLow ();
        if (x != sx)
            return x < sx;

        x = getStartHigh ();
        sx = s.getStartHigh ();
        if (x != sx)
            return x < sx;

        return getEndHigh () < s.getEndHigh ();
    }

    public static final int FUZZY_LOW = 0;    // used by fuzzify()
    public static final int FUZZY_HIGH = 1;

    public int fuzzify (int x, int direction) {
        return this.melt().getSingleDocRegionSet(this.getDocument()).fuzzify (x, direction);
    }

    public RegionSet convert (CoordinateMap coords, DocumentVersion ver) {
        if (coords == IdentityMap.IDENTITY && ver.equals(this.getVersion()))
            return this;
            
        MutableRegionSet result = new MutableRegionSet ();
        RegionEnumeration e = regions ();
        for (Region r = e.first(); r != null; r = e.next ())
            result.insert ((Region) r.convert (coords, ver));
        return result;
    }
    
    public RegionSet convert (CoordinateMap coords) {
        return this.convert(coords, this.getVersion());
    }
    
    public RegionSet convert (CoordinateMap coords, Document doc) {
        return this.convert (coords, doc.getVersion());
    }
}
