/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


/**
procedure: STANDARDIZE OF UNION OF N RECTANGLES

1   X[1:2N] := sorted sequence of abscissae of vertical sides
2   construct and initialize interval sequence S for ordinates of rectangle sides at X[1]
3   for i := 2 until 2N 
4       yinterval[] := merge (S)
5       if (X[i] is a left-side abscissa)
6           insert ([X[i-1],(X[i]-1)]*yinterval[], BUFFER) only if X[i-1]<=X[i]-1
7           insert ([bi, ei], S)
8       else
9           insert ([X[i-1],X[i]]*yinterval[], BUFFER) only if X[i-1]<=X[i]
10          X[i]++
11          delete ([bi, ei], S)
12      insert (BUFFER, RESULT)
13  end
14  return RESULT



procedure: MERGE OF UNION OF M INTERVALS

1   Y[1:2M] := sorted sequence of abscissae of intervals
2   Y[0] := Y[1]
3   C := 0
4   START := END := null
5   for i := 1 until 2M
6       if (Y[i] is a low-side abscissa) 
7           if (START == null) START := Y[i]
8           C := C+1
9       else C := C-1
10      if (C==0)
11          END := Y[i]
12          insert ([START, END], BUFFER, RESULT)
13          START:= END := null
14  end
15  return RESULT
**/

package lapis;

import java.util.*;

public class StandardRectangles {
    ArrayList standardRects;

    Buffer buffer;
    private Object[] xSlides;

    private ArrayList resultRects = new ArrayList();

    public class Rect {
	int left;
	int right;
	int low;
	int high;

	public Rect(int left, int right, int low, int high) {
	    this.left = left;
	    this.right = right;
	    this.low = low;
	    this.high = high;
	}

	public String toString() {
	    return "Rectangle: x[" + left + ", " + right + "], y[" + low + ", " + high + "]";
	}

	public boolean equals(Object other) {
	    if (other instanceof Rect)
		if ( (this.left == ((Rect)other).left) &&
		     (this.right == ((Rect)other).right) &&
		     (this.low == ((Rect)other).low) &&
		     (this.high == ((Rect)other).high) )
		    return true;
		else
		    return false;
	    else
		return false;
	}
    }

    public class XAbscissa implements Comparable {
	private boolean startFlag;
	private YAbscissa ystart;
	private YAbscissa yend;
	private int value;

	public XAbscissa(int value, boolean startFlag, YAbscissa ystart, YAbscissa yend) {
	    this.startFlag = startFlag;
	    this.value = value;
	    this.ystart = ystart;
	    this.yend = yend;
	}

	public boolean isStart() {
	    return startFlag;
	}

	public YAbscissa getYstart() {
	    return ystart;
	}

	public YAbscissa getYend () {
	    return yend;
	}

	public int getValue () {
	    return value;
	}

	protected void setValue(int value) {
	    this.value = value;
	}

	public String toString() {
	    String status = startFlag ? "start" : "end";
	    return "x: " + value + " {" + status + "} " + "y: [" + ystart.getValue() + ", " + yend.getValue() + "]";
	}

	public int compareTo(Object other) throws ClassCastException {
	    if (other instanceof XAbscissa) {
		if (this.value < ((XAbscissa)other).getValue())
		    return -1;
		else if (this.value > ((XAbscissa)other).getValue())
		    return 1;
		else if (this.isStart() && !((XAbscissa)other).isStart())
		    return -1;
		else if (!this.isStart() && ((XAbscissa)other).isStart())
		    return 1;
		else
		    return 0;
	    } else
		throw new ClassCastException("XAbscissa uncomparable");
	}
    }

    public class YAbscissa implements Comparable {
	private boolean startFlag;
	private int value;
	private int regionIndex;

	public YAbscissa(int value, boolean startFlag, int regionIndex) {
	    this.startFlag = startFlag;
	    this.value = value;
	    this.regionIndex = regionIndex;
	}

	public boolean isStart() {
	    return startFlag;
	}

	public int getValue() {
	    return value;
	}
    
	public boolean equals(Object other) {
	    if (other instanceof YAbscissa)
		if ( (this.value == ((YAbscissa)other).getValue()) && 
		     (this.startFlag == ((YAbscissa)other).isStart()) )
		    return true;
		else
		    return false;
	    else
		return false;
	}

	public String toString() {
	    String status = startFlag ? "start" : "end";
	    return "y: " + status + " " + value;
	}

	public int compareTo(Object other) throws ClassCastException {
	    if (other instanceof YAbscissa) {
		if (this.value < ((YAbscissa)other).getValue())
		    return -1;
		else if (this.value > ((YAbscissa)other).getValue())
		    return 1;
		else if (this.isStart() && !((YAbscissa)other).isStart())
		    return -1;
		else if (!this.isStart() && ((YAbscissa)other).isStart())
		    return 1;
		else if (this.regionIndex < ((YAbscissa)other).regionIndex)
		    return -1;
		else if (this.regionIndex > ((YAbscissa)other).regionIndex)
		    return 1;
		else
		    return 0;
	    } else
		throw new ClassCastException("YAbscissa uncomparable");
	}

    }

    class Interval {
	int start;
	int end;

	Interval(int start, int end) {
	    this.start = start;
	    this.end = end;
	}
    }

    class Buffer {
	int xStart;
	int xEnd;
	ArrayList yIntervals;

	Buffer(int xStart, int xEnd, ArrayList yIntervals) {
	    this.xStart = xStart;
	    this.xEnd = xEnd;
	    this.yIntervals = yIntervals;
	}

	boolean sameYIntervals(ArrayList intervs) {
	    int size = this.yIntervals.size();
	    if (size != intervs.size()) 
		return false;
	    
	    for (int i = 0; i < size; i++)
		if  ( ( ((Interval)this.yIntervals.get(i)).start != ((Interval)intervs.get(i)).start ) ||
		      ( ((Interval)this.yIntervals.get(i)).end != ((Interval)intervs.get(i)).end ) )
		    return false;
	    
	    return true;
	}
    }

    // used for test only
    
    public StandardRectangles (ArrayList regions) {
	ArrayList verticals = new ArrayList();
	for (int i = 0; i < regions.size(); i++) {
	    Region r = (Region) (regions.get(i));

	    int startLow = r.getStartLow();
	    int startHigh = r.getStartHigh();
	    int endLow = r.getEndLow();
	    int endHigh = r.getEndHigh();
	    
	    if (startLow > endHigh) {
		// ignore the rectangle under diagonal
	    } else {
		if ( (startLow >= endLow) || (startHigh >= endHigh) )
		    endLow = RegionSet.MIN_INDEX;
		
		verticals.add(new XAbscissa(startLow, true,
					    new YAbscissa(endLow, true, i),
					    new YAbscissa(endHigh, false, i)));
		verticals.add(new XAbscissa(startHigh, false,
					    new YAbscissa(endLow, true, i),
					    new YAbscissa(endHigh, false, i)));
	    }
	}

	standardRects = new ArrayList();
	buffer = new Buffer(-1, -1, null);
	xSlides = verticals.toArray();
	Arrays.sort(xSlides);
    }
    
    // end

    public StandardRectangles (RegionSet rs) {
	ArrayList verticals = new ArrayList();
	RegionEnumeration e = rs.regions();
	int regionIndex = 0;
	for (Region r = e.first(); r != null; r = e.next()) {
	    regionIndex++;
	    int startLow = r.getStartLow();
	    int startHigh = r.getStartHigh();
	    int endLow = r.getEndLow();
	    int endHigh = r.getEndHigh();
	    
	    if (startLow > endHigh) {
		// ignore the rectangle under diagonal
	    } else {
		if ( (startLow >= endLow) || (startHigh >= endHigh) )
		    endLow = RegionSet.MIN_INDEX;
		
		verticals.add(new XAbscissa(startLow, true,
					    new YAbscissa(endLow, true, regionIndex),
					    new YAbscissa(endHigh, false, regionIndex)));
		verticals.add(new XAbscissa(startHigh, false,
					    new YAbscissa(endLow, true, regionIndex),
					    new YAbscissa(endHigh, false, regionIndex)));
	    }
	}
	
	standardRects = new ArrayList();
	buffer = new Buffer(-1, -1, null);

	xSlides = verticals.toArray();
	Arrays.sort(xSlides);
    }

    public ArrayList standardize() {
	if (xSlides.length == 0)
	    return standardRects;

	TreeSet horizontals = new TreeSet();
 
	horizontals.add(((XAbscissa)xSlides[0]).getYstart());
	horizontals.add(((XAbscissa)xSlides[0]).getYend());

	for (int i = 1; i < xSlides.length; i++) {
	    ArrayList yinterval = mergeInterval(horizontals);

	    int prevX = ((XAbscissa)xSlides[i-1]).getValue();
	    int currX = ((XAbscissa)xSlides[i]).getValue();

	    if (((XAbscissa)xSlides[i]).isStart()) {
		insertRects(prevX, (currX-1), yinterval);
		horizontals.add(((XAbscissa)xSlides[i]).getYstart());
		horizontals.add(((XAbscissa)xSlides[i]).getYend());
	    } else {
		insertRects(prevX, currX, yinterval);
		((XAbscissa)xSlides[i]).setValue(currX+1);
		horizontals.remove( ((XAbscissa)xSlides[i]).getYstart() );
		horizontals.remove( ((XAbscissa)xSlides[i]).getYend() );
	    }
	}

	flushRects();

	return standardRects;
    }

    ArrayList mergeInterval(TreeSet segments) {
	ArrayList yintervals = new ArrayList();
	Interval ybuffer = null;

	Iterator it = segments.iterator();

	int start = -1;
	int end = -1;
	int count = 0;

	while (it.hasNext()) {
	    YAbscissa current = (YAbscissa)it.next();
	    if (current.isStart()) {
		if (start == -1)
		    start = current.getValue();
		count++;
	    } else
		count--;

	    if ( (count == 0) && (start != -1) ) {
		end = current.getValue();

		if (ybuffer == null)
		    ybuffer = new Interval(start, end);
		else if (ybuffer.end == start-1) 
		    ybuffer = new Interval(ybuffer.start, end);
		else {
		    yintervals.add(ybuffer);
		    ybuffer = new Interval(start, end);
		}
		start = -1;
		end = -1;
	    }
	}
	
	if (ybuffer != null)
	    yintervals.add(ybuffer);

	return yintervals;
    }

    void insertRects(int xStart, int xEnd, ArrayList yIntervals) {
	if (xStart > xEnd)
	    return;

	if (buffer.yIntervals == null) {
	    buffer.xStart = xStart;
	    buffer.xEnd = xEnd;
	    buffer.yIntervals = yIntervals;
	} else {
	    if (buffer.sameYIntervals(yIntervals))
		buffer.xEnd = xEnd;
	    else {
		flushRects();
		buffer.xStart = xStart;
		buffer.xEnd = xEnd;
		buffer.yIntervals = yIntervals;
	    }
	}
    }

    void flushRects() {
	if (buffer == null)
	    return;

	int size = buffer.yIntervals.size();
	for (int i = 0; i < size; i++)
	    standardRects.add(new Rect(buffer.xStart, buffer.xEnd,
			      ((Interval)buffer.yIntervals.get(i)).start,
			      ((Interval)buffer.yIntervals.get(i)).end));
    }

    public String toString() {
	String result = "";
	for (int i = 0; i < xSlides.length; i++)
	    result += ((XAbscissa)xSlides[i]).toString() + "\n";
	return result;
    }

    void printHorizontals(TreeSet horizon) {
	Iterator it = horizon.iterator();
	while (it.hasNext())
	    System.out.println((YAbscissa)it.next());
    }

}
