/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */



package lapis;

import java.util.*;
import java.io.*;
import lapisx.util.*;

public class Toolbox {
    public static Debug debug = Debug.QUIET; 
    Vector toolinfo = new Vector (); // Vector of Info

    public Toolbox () {
    }

    static Toolbox defaultToolbox;

    public static Toolbox getDefaultToolbox () {
        if (defaultToolbox == null)
            defaultToolbox = new Toolbox ();
        return defaultToolbox;
    }
    
    public boolean contains (Info info) {

	return (toolinfo.contains(info));
    } 
    
    /*
     * Adds named tool to Toolbox. If a tool by that name already
     * exists in the Toolbox, that tool is replaced (in place), else
     * new tool is appended to the Toolbox
     */
    
    public void add (String name, Tool tool) {
	Info info = new Info (tool, name, null, null, true, true); 
	add(info);


    }
    /*
     * Adds named tool to Toolbox. If a tool by that name already
     * exists in the Toolbox, that tool is replaced (in place), else
     * new tool is appended to the Toolbox
     */
    public void add (Info info) {
	debug.println("" + this + " " + toolinfo.size());
	if (toolinfo.size() > 0) {
	    Iterator i = toolinfo.iterator();
	    while (i.hasNext()){
		Info in = (Info)i.next();
		debug.println(in.getName());
	    }
	}
	int index = toolinfo.indexOf(info);
	if (index != -1) {
	    debug.println("tool name " + info.getName() + " exists at index " + index);
	    toolinfo.setElementAt(info, index);
	     fireChangedEvent(info);
	    // toolbox has been changed, a tool has been modified
	}
	else {
	    debug.println("tool name " + info.getName() + " does not exist " + index);
	    toolinfo.addElement (info);
	    fireAddedEvent(info); 
	}
	if (toolinfo.size() > 0) {
	    Iterator i = toolinfo.iterator();
	    while (i.hasNext()){
		Info in = (Info)i.next();
		debug.println(in.getName());
	    }
	}

    }
    
    public void remove (String name) {
        for (int i = toolinfo.size () - 1; i >= 0; --i) {
            Info info = (Info) toolinfo.elementAt (i);
            if (name.equals (info.getName ())){
                toolinfo.removeElementAt (i);
		fireRemovedEvent(info); 
	    }
        }
    }
    
    public void remove (Info inf) {
        for (int i = toolinfo.size () - 1; i >= 0; --i) {
            Info info = (Info) toolinfo.elementAt (i);
            if (inf.equals (info)){
                toolinfo.removeElementAt (i);
		fireRemovedEvent(info); 
	    }
        }

    }

    public Enumeration tools () {
        return toolinfo.elements ();
    }

    public static class Info {
        Tool tool;
        String name;
        String description;
        File icon;
        boolean onMenu;
        boolean onToolbar;

        public Info (Tool tool, String name, String description, 
                     File icon, boolean onMenu, boolean onToolbar) {
            this.tool = tool;
            this.name = name;
            this.description = description;
            this.icon = icon;
            this.onMenu = onMenu;
            this.onToolbar = onToolbar;
        }

        public Tool getTool () {
            return tool;
        }
        public String getName () {
            return name;
        }
        public String getDescription () {
            return description;
        }
        public File getIcon () {
            return icon;
        }
        public boolean getOnMenu () {
            return onMenu;
        }
        public boolean getOnToolbar () {
            return onToolbar;
        }
	public boolean equals(Object info) { 
	    if (info instanceof Info){ 
		debug.println(((Info)info).getName() + " vs. " + this.name);
		return (((Info)info).getName().equals(this.name));
	    }
	    else
		return false;
	}
    }

    //
    // Event listeners
    //

    Vector listeners = new Vector ();

    public static interface ToolBoxListener {
        public void toolAdded (Toolbox toolBox, Info info);
        public void toolRemoved (Toolbox toolBox, Info info);
	public void toolChanged (Toolbox toolBox, Info info);
    }

    public synchronized void addToolBoxListener (ToolBoxListener listener) {
        if (!listeners.contains (listener))
            listeners.addElement (listener);
    }

    public synchronized void removeToolBoxListener (ToolBoxListener listener) {
        listeners.removeElement (listener);
    }

    protected synchronized void fireAddedEvent (Info info) {
	debug.println("added " + info.getName());
        ToolBoxListener[] l = new ToolBoxListener[listeners.size ()];
        listeners.copyInto (l);
        for (int i = 0; i < l.length; ++i)
            l[i].toolAdded (this, info);
    }
    protected synchronized void fireChangedEvent (Info info) {
	debug.println("changed " + info.getName());
        ToolBoxListener[] l = new ToolBoxListener[listeners.size ()];
        listeners.copyInto (l);
        for (int i = 0; i < l.length; ++i)
            l[i].toolChanged (this, info);
    }

    protected synchronized void fireRemovedEvent (Info info) {

        ToolBoxListener[] l = new ToolBoxListener[listeners.size ()];
        listeners.copyInto (l);
        for (int i = 0; i < l.length; ++i)
            l[i].toolRemoved (this, info);
    }
}
