/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.maps;

import lapis.*;
import lapisx.util.Str;
import lapisx.util.BinarySearch;

public class PiecewiseLinearMonotonicMap implements CoordinateMap {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;
 
    int[] domain;
    int[] range;
    int nPoints;

    public PiecewiseLinearMonotonicMap () {
        this (10);
    }

    public PiecewiseLinearMonotonicMap (int initialSize) {
        if (initialSize < 1)
            initialSize = 1;
        domain = new int[initialSize];
        range = new int[initialSize];
        nPoints = 0;
    }

    protected PiecewiseLinearMonotonicMap (int[] domain, int[] range, int nPoints) {
        this.domain = domain;
        this.range = range;
        this.nPoints = nPoints;
    }

    public void add (int x, int y) {
        debug.println ("add " + x + " -> " + y);
        if (nPoints > 1) {
            int x1 = domain[nPoints-1];
            int y1 = range[nPoints-1];
            int x2 = domain[nPoints-2];
            int y2 = range[nPoints-2];

            if ((x - x1) * (y1 - y2) == (x1 - x2) * (y - y1)) {
                // new point lies on an extension of the last line segment;
                // just extend the previous line segment to reach x,y
                domain[nPoints-1] = x;
                range[nPoints-1] = y;
                return;
            }
        }

        if (nPoints >= domain.length)
            expand ();
        domain[nPoints] = x;
        range[nPoints] = y;
        ++nPoints;
    }

    private void expand () {
        int[] newDomain = new int[domain.length * 2 + 1];
        int[] newRange = new int[newDomain.length];
        System.arraycopy (domain, 0, newDomain, 0, domain.length);
        System.arraycopy (range, 0, newRange, 0, range.length);
        domain = newDomain;
        range = newRange;

    }

    public CoordinateMap invert () {
        return new PiecewiseLinearMonotonicMap (range, domain, nPoints);
    }

    public int mapStart (int x) {
        if (nPoints == 0)
            return x;

        if (x == RegionSet.MAX_INDEX)
            return x;

        int i = BinarySearch.search (domain, 0, nPoints, x);

        if (i == nPoints)
            return (x - domain[nPoints-1]) + range[nPoints-1];
        
        if (x == domain[i]) {
            if (i < nPoints-1 && x == domain[i+1])
                return range[i+1];
            else
                return range[i];
        }
        
        if (i == 0)
            return x;

        // domain[i-1] < x < domain[i]
        if (range[i-1] == range[i])
            return range[i];

        return range[i-1] + (x - domain[i-1]);
    }

    public int mapEnd (int x) {
        if (nPoints == 0)
            return x;

        if (x == RegionSet.MAX_INDEX)
            return x;

        int i = BinarySearch.search (domain, 0, nPoints, x);

        if (i == nPoints)
            return (x - domain[nPoints-1]) + range[nPoints-1];
        
        if (x == domain[i])
            return range[i];

        if (i == 0)
            return x;

        // domain[i-1] < x < domain[i]
        if (range[i-1] == range[i])
            return range[i];

        return range[i-1] + (x - domain[i-1]);
    }

    public void dump (String docA, String docB) {
        if (!debug.printEnabled ())
            debug = lapisx.util.Debug.VERBOSE;

        try {
            int xA, xB, yA, yB;
            xA = domain[0];
            xB = range[0];
            for (int i = 1; i < nPoints; ++i) {
                yA = domain[i];
                yB = range[i];

                String strA;
                try {
                    strA = docA.substring (xA, yA);
                } catch (StringIndexOutOfBoundsException e) {
                    strA = e.toString ();
                }

                String strB;
                try {
                    strB = docB.substring (xB, yB);
                } catch (StringIndexOutOfBoundsException e) {
                    strB = e.toString ();
                }

                debug.println ("A[" + xA + "," + yA + "]:\t" 
                             + Str.abbreviate (strA, 50));
                debug.println ("B[" + xB + "," + yB + "]:\t"
                           + Str.abbreviate (strB, 50));

                xA = yA;
                xB = yB;
            }
        } catch (Exception e) {
            debug.report (e);
        }
    }
}
