/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.ml;

//import lapis.*;
import java.util.*;
import java.io.*;
//import java.net.*;
import lapisx.util.*;

public class CommonSubstrings {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    Trie root; // substring trie
    String base; // string from which Trie was built

    public CommonSubstrings () {
        root = new Trie ("", 0);
    }

    public CommonSubstrings (String s) {
        this ();
        add (s);
    }

    public CommonSubstrings (String[] strings) {
        this ();

        // find shortest string to build trie
        String shortest = null;
        int shortestLen = Integer.MAX_VALUE;
        for (int i = 0; i < strings.length; ++i) {
            int len = strings[i].length ();
            if (len < shortestLen) {
                shortest = strings[i];
                shortestLen = len;
            }
        }
        if (shortest == null)
            return; // strings is empty
        add (shortest);

        // update trie from all other strings
        for (int j = 0; j < strings.length; ++j) {
            String s = strings[j];
            if (s != shortest)
                add (s);
        }
        
        root.dump ();
    }

    public void add (String s) {
        add (s, s.length ());
    }

    public void add (String s, int n) {
        if (base == null) {
            base = s;
            root.insert (base, n);
        } else {
            root.update (s, n);
        }
        //root.dump ();
    }


    public CommonSubstring[] getMaximalCommonSubstrings () {
        Vector v = new Vector ();
        Trie t = root.copy ();
        Trie[] nodes = new Trie[base.length ()];
        boolean firstIteration = true;

        int k = t.getMinimumRegularity ();
        while (k != Integer.MAX_VALUE) {
            debug.println ("getting " + k + "-regular nodes");
            t.getRegularNodes (k, nodes);
            int m = 0;
            for (int i = 0; i < nodes.length; ++i) {
                Trie node = nodes[i];
                if (node != null) {
                    int j = i + node.getDepth ();
                    if (j > m) {
                        debug.println (node);
                        String s = base.substring (i, j);
                        
                        Trie u = root.getNode (s);
                        v.addElement (new CommonSubstring (s, 
                                                           u.minOccurrences,
                                                           u.maxOccurrences));
                        m = j;
                        
                        // reduce occurrence counts of all substrings
                        // of s.  This step enables prune() to delete
                        // some of these nodes from tree.
                        t.removeOccurrences (s, k);
                    }

                    nodes[i] = null;
                }
            }

            t.prune ();
            t.dump ();

            k = t.getMinimumRegularity ();
        }

        // now get all varying substrings from *original* tree
        Vector varying = new Vector ();
	root.getVaryingSubstrings ("", varying);

        for (Enumeration e = varying.elements (); e.hasMoreElements (); ) {
            String s = (String) e.nextElement ();
            Trie u = root.getNode (s);
            v.addElement (new CommonSubstring (s,
                                               u.minOccurrences,
                                               u.maxOccurrences));
        }

        CommonSubstring[] cs = new CommonSubstring[v.size ()];
        v.copyInto (cs);
        return cs;
    }

    public CommonSubstring[] getLeaves () {
        Vector v = new Vector ();
        Trie[] leaves = new Trie[base.length ()];
        root.getLeaves (leaves);

        for (int i = 0; i < leaves.length; ++i) {
            Trie leaf = leaves[i];
            if (leaf != null) {
                int j = i + leaf.getDepth ();
                String s = base.substring (i, j);
                v.addElement (new CommonSubstring (s, 
                                                   leaf.minOccurrences,
                                                   leaf.maxOccurrences));
            }
        }

        CommonSubstring[] cs = new CommonSubstring[v.size ()];
        v.copyInto (cs);
        return cs;
    }

    public static class CommonSubstring {
        public String s;
        public int minOccurrences;
        public int maxOccurrences;
        public CommonSubstring (String s, int min, int max) {
            this.s = s;
            this.minOccurrences = min;
            this.maxOccurrences = max;
        }
        public String toString () {
            return s + "[" + minOccurrences + "-" + maxOccurrences + "]";
        }
    }


    public boolean isUnique (String s) {
        return s.length () > 0 
            ? root.isUnique (s) 
            : false;
    }

    public String getShortestUniquePrefix (String s) {
        // FIX: find i by scanning down suffix tree
        for (int i = 1, n = s.length (); i <= n; ++i) {
            String t = s.substring (0, i);
            if (isUnique (t))
                return t;
        }
        return null;
//         try {
//             return s.length () > 0
//                 ? s.substring (0, root.getShortestUniquePrefixLength (s))
//                 : null;
//         } catch (NotUniqueException e) {
//             return null;
//         }
    }

    public String getShortestUniqueSuffix (String s) {
        // FIX: find i by binary search,
        // or build a suffix tree on the reverses of the strings
        // (this should be derivable from the suffix tree, right?)
        for (int i = s.length ()-1; i >= 0; --i) {
            String t = s.substring (i);
            if (isUnique (t))
                return t;
        }
        return null;
    }

    //
    // for each node in the suffix tree that is a prefix of s, return the
    // corresponding prefix.
    //
    public String[] getAllPrefixes (String s) {
        Vector v = new Vector ();
        root.getAllPrefixes (s, "", v);
        String[] result = new String[v.size ()];
        v.copyInto (result);
        return result;
    }

    static class Trie implements Cloneable {
        String label; 
        // string that labels the link to this node.
        //
        // This implementation depends for efficiency 
        // on the fact that java.lang.String.substring() 
        // does not copy characters, but simply returns
        // a [start,end] pointer.
        //
        // If java.lang.String didn't work that way, we'd have to
        // put a [start,end] interval here instead of a
        // String.

        int start;
        // start position (in base) of any match to this
        // node.

        Hashtable children;
        // Child nodes, indexed by first character of child's label.
        // Every child node label starts with a distinct character.

        Trie parent; // parent node

        int occurrences;
        // number of times this substring occurs in the current string.

        int minOccurrences = Integer.MAX_VALUE;
        int maxOccurrences = Integer.MIN_VALUE;
        // minimum and maximum number of times this substring occurs in
        // any string processed so far.


        public Trie (String label, int start) {
            this.label = label;
            this.start = start;
        }

        public Trie copy () {
            try {
                Trie t = (Trie) clone ();
                t.parent = null;
                if (t.children != null) {
                    t.children = new Hashtable ();
                    for (Enumeration e = children.elements (); e.hasMoreElements (); )
                        t.addChild (((Trie) e.nextElement ()).copy ());
                }
                return t;
            } catch (CloneNotSupportedException e) {
                throw new RuntimeException (e.toString ());
            }
        }

        public boolean isRoot () {
            return parent == null;
        }

        public boolean isLeaf () {
            return children == null || children.size () == 0;
        }
        
        public String getLabel () {
            return label;
        }

        public String getPath () {
            return isRoot () ? "" : parent.getPath () + label;
        }
 
        public void addChild (Trie child) {
            if (children == null)
                children = new Hashtable ();
            children.put (getKey (child.label), child);
            child.parent = this;
        }

        public void removeChild (Trie child) {
            if (children == null)
                return;
            Trie t = (Trie) children.remove (getKey (child.label));
            debug.assertion (child == t);
            child.parent = null;
        }

        public Trie getChild (String s) {
            if (children == null)
                return null;
            return (Trie) children.get (getKey (s));
        }

        static Object getKey (String s) {
            return new Character (s.charAt (0));
        }

        public void insert (String s, int n) {
            for (int i = 0; i < n; ++i)
                insertSuffix (s.substring (i), i);
            updateMinMax ();
        }

        public void update (String s, int n) {
            for (int i = 0; i < n; ++i)
                insertSuffix (s.substring (i), -1);
            updateMinMax ();
            prune ();
        }

        void insertSuffix (String s, int start) {
            if (s.length () == 0) {
                ++occurrences;
                return;
            }
            
            Trie c = getChild (s);
            if (c == null) {
                if (start >= 0) {
                    Trie t = new Trie (s, start);
                    addChild (t);
                    ++t.occurrences;
                } else
                    ++occurrences;
                return;
            }

            int k = Str.longestCommonPrefixLength (s, c.label);
            if (k == c.label.length ()) {
                c.insertSuffix (s.substring (k), start);
                return;
            }

            // introduce a new node and split c.label
            Trie t = new Trie (s.substring (0, k), c.start);
            t.minOccurrences = c.minOccurrences;
            t.maxOccurrences = c.maxOccurrences;
            
            removeChild (c);
            c.label = c.label.substring (k);
            
            t.addChild (c);
            addChild (t);
            t.insertSuffix (s.substring (k), start);
        }

        // update minOccurrences/maxOccurrences at every node
        int updateMinMax () {
            int n = occurrences;

            if (children != null)
                for (Enumeration e = children.elements ();
                     e.hasMoreElements (); )
                    n += ((Trie) e.nextElement ()).updateMinMax ();

            minOccurrences = Math.min (minOccurrences, n);
            maxOccurrences = Math.max (maxOccurrences, n);

            occurrences = 0;
            return n;
        }

        // prune off nodes with minOccurrences == 0
        void prune () {
            if (!isRoot () && minOccurrences == 0)
                parent.removeChild (this);
            else if (children != null)
                for (Enumeration e = children.elements ();
                     e.hasMoreElements (); )
                    ((Trie) e.nextElement ()).prune ();
        }

        public void getLeaves (Trie[] leaves) {
            if (isLeaf ()) {
                if (!isRoot ()) // don't treat the root as a leaf
                    leaves[start] = this;
            } else
                for (Enumeration e = children.elements ();
                     e.hasMoreElements (); )
                    ((Trie) e.nextElement ()).getLeaves (leaves);
        }

        // return length of substring to this node
        public int getDepth () {
            return 
                label.length ()
                + (isRoot () ? 0 : parent.getDepth ());
        }

        public boolean isUnique (String s) {
            if (s.length () == 0)
                return maxOccurrences == 1;

            Trie c = getChild (s);
            if (c == null)
                return false;

            int k = Str.longestCommonPrefixLength (s, c.label);
            if (k == c.label.length ())
                return c.isUnique (s.substring (k));

            return c.maxOccurrences == 1;
        }
        
        public int getShortestUniquePrefixLength (String s) throws NotUniqueException {
            if (s.length () == 0)
                return 0;

            Trie c = getChild (s);
            if (c == null)
                throw new NotUniqueException ();

            int k = Str.longestCommonPrefixLength (s, c.label);
            if (k == c.label.length ())
                return k + c.getShortestUniquePrefixLength (s.substring (k));

            if (c.maxOccurrences > 1)
                throw new NotUniqueException ();

            return k;
        }
        
        public boolean contains (String s) {
            Trie c = getChild (s);
            if (c == null)
                return false;
            else {
                int k = Str.longestCommonPrefixLength (s, c.label);
                if (k == s.length ())
                    return true;
                else if (k == c.label.length ())
                    return c.contains (s.substring (k));
                else
                    return false;
            }
        }

        // Finds deepest nodes such that minOccurrences and maxOccurrences
        // both equal k, and no deeper node does.  Returns number of nodes
        // added by this subtree.
        public int getRegularNodes (int k, Trie[] nodes) {
            int n = 0;
            if (children != null)
                for (Enumeration e = children.elements ();
                     e.hasMoreElements (); )
                    n += ((Trie) e.nextElement ()).getRegularNodes (k, nodes);
            if (n == 0
                && minOccurrences == k
                && maxOccurrences == k
                && !isRoot ()) {
                nodes[start] = this;
                ++n;
            }
            return n;
        }

        // Finds minimum k such that there is a node with
        // minOccurrences == maxOccurrences == k.  
        // Returns Integer.MAX_VALUE if no such node.
        public int getMinimumRegularity () {
            int k = 
                (minOccurrences == maxOccurrences && !isRoot ())
                ? minOccurrences
                : Integer.MAX_VALUE;
                
            if (children != null)
                for (Enumeration e = children.elements ();
                     e.hasMoreElements (); )
                    k = Math.min (k, 
                                  ((Trie) e.nextElement ())
                                  .getMinimumRegularity ());

            return k;
        }

        // Finds all nodes such that minOccurrences != maxOccurrences.
        // Returns number of nodes added to nodes array by this subtree.
        public int getVaryingNodes (Trie[] nodes) {
            int n = 0;
            if (children != null)
                for (Enumeration e = children.elements ();
                     e.hasMoreElements (); )
                    n += ((Trie) e.nextElement ()).getVaryingNodes (nodes);
            if (n == 0
                && minOccurrences != maxOccurrences
                && !isRoot ()) {
                nodes[start] = this;
                ++n;
            }
            return n;
        }


        public void removeOccurrences (String s, int ct) {
            for (int i = 0, n = s.length (); i < n; ++i)
                removeOcc (s.substring (i), ct);
        }

        void removeOcc (String s, int ct) {
            minOccurrences = Math.max (0, minOccurrences - ct);
            maxOccurrences = Math.max (0, maxOccurrences - ct);

            if (s.length () == 0)
                return;
            
            Trie c = getChild (s);
            if (c == null)
                return;

            int k = Str.longestCommonPrefixLength (s, c.label);
            if (k == c.label.length ())
                c.removeOcc (s.substring (k), ct);
        }

        public void getAllSubstrings (String prefix, Vector v) {
            String me = prefix + label;
            if (!isRoot ())
                v.addElement (me);
            if (children != null)
                for (Enumeration e = children.elements ();
                     e.hasMoreElements (); )
                    ((Trie) e.nextElement ()).getAllSubstrings (me, v);
        }

        public void getVaryingSubstrings (String prefix, Vector v) {
            String me = prefix + label;
            if (minOccurrences != maxOccurrences
                && !isRoot ())
                v.addElement (me);
            if (children != null)
                for (Enumeration e = children.elements ();
                     e.hasMoreElements (); )
                    ((Trie) e.nextElement ()).getVaryingSubstrings (me, v);
        }

        public Trie getNode (String s) {
            if (s.length () == 0)
                return this;
            
            Trie c = getChild (s);
            if (c == null)
                return null;

            int k = Str.longestCommonPrefixLength (s, c.label);
            if (k == c.label.length ())
                return c.getNode (s.substring (k));
            else
                return null;
        }

        public void getAllPrefixes (String s, String prefix, Vector v) {
            String me = prefix + label;
            if (!isRoot ())
                v.addElement (me);

            if (s.length () == 0)
                return;
            
            Trie c = getChild (s);
            if (c == null)
                return;

            int k = Str.longestCommonPrefixLength (s, c.label);
            if (k == c.label.length ())
                c.getAllPrefixes (s.substring (k), me, v);
        }


        //
        // debugging
        //

        public String toString () {
            return getPath () + "(" + start + ") [" 
                + minOccurrences + "-" + maxOccurrences + "]";
        }

        public void dump () {
            dump (0);
        }
    
        public void dump (int indent) {
            debug.println (Str.repeat (" ", indent) 
                           + label + "(" + start + ") [" 
                           + minOccurrences + "-" + maxOccurrences + "]");
            if (children != null)
                for (Enumeration e = children.elements ();
                     e.hasMoreElements (); )
                    ((Trie) e.nextElement ()).dump (indent + 2);
        }
    }

    static class NotUniqueException extends Exception {
        public NotUniqueException () {
        }
    }


    //
    // debugging
    //

    public static void main (String[] argv) throws Exception {
        String[] strings = readStrings (System.in);
        CommonSubstrings cs = 
            new CommonSubstrings (strings);

        Trie[] nodes = new Trie[cs.base.length ()];
        System.out.println ("Right-maximal regular substrings:");
        for (int k = 0; k < 10; ++k) {
            cs.root.getRegularNodes (k, nodes);
            for (int i = 0; i < nodes.length; ++i) {
                if (nodes[i] != null)
                    System.out.println (nodes[i]);
                nodes[i] = null;
            }
        }
        System.out.println ("------------------");

        Vector unique = new Vector ();
        Vector regular = new Vector ();
        Vector varying = new Vector ();
        CommonSubstring[] subs = cs.getMaximalCommonSubstrings ();
        System.out.println ("Common substrings:");
        for (int i = 0; i < subs.length; ++i)
            System.out.println (subs[i]);


        for (int i = 0; i < argv.length; ++i) {
            String s = argv[i];
            System.out.println (s);
            System.out.println ("\tunique=" + cs.isUnique (s));
            System.out.println ("\tsup=" + cs.getShortestUniquePrefix (s));
            System.out.println ("\tsus=" + cs.getShortestUniqueSuffix (s));
        }
    }

    static String[] readStrings (InputStream inp) throws IOException {
        StreamTokenizer in = new StreamTokenizer (new InputStreamReader (inp));
        in.resetSyntax ();
        in.wordChars (Character.MIN_VALUE, Character.MAX_VALUE);
        in.whitespaceChars ('\n', '\n');
        Vector v = new Vector ();
        while (in.nextToken () != StreamTokenizer.TT_EOF)
            if (in.ttype == StreamTokenizer.TT_WORD)
                v.addElement (in.sval);
        String[] result = new String[v.size ()];
        v.copyInto (result);
        return result;
    }
}

