/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.ml;

import lapis.*;
import java.util.*;
import lapisx.util.*;

public class DataSet extends NamedRegionSet {
    public static Debug debug = Debug.QUIET;

    // fields
    Feature[] features;
    double[] sum;
    double[] mean; 
    double[] median;
    double[] variance;

    Hashtable featureValues = new Hashtable (); 
    // maps each region in examples to a double[] feature vector

    /**
     * Compute features for a set of example regions.
     */
    public DataSet (NamedRegionSet ns, Feature[] features) {
        super (ns);
        this.features = features;

        int n = features.length;
        sum = new double[n];
        mean = new double[n];
        variance = new double[n];
        median = new double[n];

        //RegionSet s = getRegions ();

        // associate an empty feature vector with each region
        RegionEnumeration re = this.regions ();
        for (Region r = re.first (); r != null; r = re.next ())
            featureValues.put (r, new double[n]);

        int m = this.getSize ();
        for (int i = 0; i < n; ++i) {
            Main.getProgress ().fireAddProgress (1);
            // Intersect feature with our region set.
            // SetFeature (defined below) sets the feature to 1
            // for every region found in both sets (i.e., regions
            // which match the pattern).
            this.intersect (new SetFeature(i), features[i], null);

            // Compute statistics.
            mean[i] = m > 0 ? sum[i] / m : 0.0;
            variance[i] = mean[i] * (1-mean[i]);
            median[i] = (mean[i] >= .5) ? 1 : 0;
        }
    }

    public Feature[] getFeatures () {
        return features;
    }

    public double[] getSum () {
        return sum;
    }

    public double[] getMean () {
        return mean;
    }

    public double[] getMedian () {
        return median;
    }

    public double[] getVariance () {
        return variance;
    }

    class SetFeature implements RegionSet.IntersectOp {
        int i; // index of feature to set
        public SetFeature (int i) {
            this.i = i;
        }
        public void combine (Region r, Region s, MutableRegionSet result) {
            getFeatureValues (r) [i] = 1;
            ++sum[i];
        }
    }

    //
    // accessors
    //
    
    public double[] getFeatureValues (Region r) {
        return (double[]) featureValues.get (r);
    }

    //
    // computing statistics on subsets of examples
    //

    public double[] computeMean (RegionSet s) {
        int n = features.length;
        double[] sum = new double[n];
        RegionEnumeration re = s.regions ();
        for (Region r = re.firstFast (); r != null; r = re.nextFast ()) {
            double[] v = getFeatureValues (r);
            for (int i = 0; i < n; ++i)
                sum[i] += v[i];
        }

        int m = s.getSize ();
        if (m > 0) {
            for (int i = 0; i < n; ++i)
                sum[i] /= m;
        }
        
        return sum;
    }

    public double[] computeVariance (RegionSet s, double[] mean) {
        int n = mean.length;
        double[] sum = new double[n];
        RegionEnumeration re = s.regions ();
        for (Region r = re.firstFast (); r != null; r = re.nextFast ()) {
            double[] v = getFeatureValues (r);
            for (int i = 0; i < n; ++i) {
                double d = v[i] - mean[i];
                sum[i] += d * d;
            }
        }

        int m = s.getSize ();
        if (m > 0) {
            for (int i = 0; i < n; ++i)
                sum[i] /= m;
        }
        
        return sum;
    }
    
    public double[][] computeCovariance (RegionSet s, double[] mean) {
        int n = mean.length;
        double[][] cov = new double[n][n];
        RegionEnumeration re = s.regions ();
        for (Region r = re.firstFast (); r != null; r = re.nextFast ()) {
            double[] v = getFeatureValues (r);
            for (int i = 0; i < n; ++i) {
                double d = v[i] - mean[i];
                cov[i][i] += d * d;
                for (int j = i + 1; j < n; ++j) {
                    cov[i][j] += d * (v[j] - mean[j]);
                }
            }
        }

        int m1 = s.getSize () - 1;
        if (m1 > 0) {
            for (int i = 0; i < n; ++i) {
                cov[i][i] /= m1;
                for (int j = i+1; j < n; ++j) {
                    cov[i][j] /= m1;
                    cov[j][i] = cov[i][j];
                }
            }
        }

        return cov;
    }
    
    public double computeEnergy (RegionSet s, double[] mean, DistanceMetric metric) {
        double sum = 0.0f;
        RegionEnumeration re = s.regions ();
        for (Region r = re.firstFast (); r != null; r = re.nextFast ()) {
            sum += metric.getDistance (getFeatureValues (r), mean);
        }
        int m = s.getSize ();
        if (m > 0)
            sum /= m;
        return sum;
    }

            
    //
    // Pruning unimportant features
    //

    /**
     * Prune any features that are homogeneous (same value for all 
     * examples).
     */
    public void pruneFeatures () {
        // find all features that are homogeneous (same value for all
        // regions) and mark them for deletion
        int m = getSize ();
        int n = features.length;
        int deletedIndexes[] = new int[n];
        int nDeletions = 0;
        for (int i = 0; i < n; ++i)
            // if sum of all values of this feature is 0, then all are 0;
            // similarly, if the sum is the number of examples, then all are 1.
            if (sum[i] == 0 || sum[i] == m) {
                debug.println ("pruning " 
                               + features[i]
                               + " because " 
                               + sum[i]);
                deletedIndexes[nDeletions++] = i;
            }

        if (nDeletions == 0)
            return; // no pruning needs to be done

        // remove deleted features from features array and statistics
        features = arrayDelete (features, new Feature[n - nDeletions], 
                                deletedIndexes, nDeletions);
        sum = arrayDelete (sum, new double[n - nDeletions], 
                           deletedIndexes, nDeletions);
        mean = arrayDelete (mean, new double[n - nDeletions], 
                            deletedIndexes, nDeletions);
        variance = arrayDelete (variance, new double[n - nDeletions], 
                                deletedIndexes, nDeletions);
        median = arrayDelete (median, new double[n - nDeletions], 
                              deletedIndexes, nDeletions);

        // remove deleted features from each region's feature values,
        // so that the feature vectors correspond
        RegionEnumeration re = regions ();
        for (Region r = re.firstFast (); r != null; r = re.nextFast ()) {
            featureValues.put (r, arrayDelete (getFeatureValues (r), 
                                               new double[n - nDeletions],
                                               deletedIndexes, 
                                               nDeletions));
        }
    }

    Feature[] arrayDelete (Feature[] src, Feature[] dest, int[] deletions, int nDeletions) {
        int i = 0; // index into src, next element to consider
        int j = 0; // index into deletions, next index to delete
        int k = 0; // index into dest, next slot to put an element
        for (i = j = k = 0; j < nDeletions; ++i) {
            if (i == deletions[j])
                ++j; // and don't copy src[i]
            else
                dest[k++] = src[i];
        }
        // copy remaining undeleted elements
        System.arraycopy (src, i, dest, k, src.length - i);
        return dest;
    }

    double[] arrayDelete (double[] src, double[] dest, int[] deletions, int nDeletions) {
        int i = 0; // index into src, next element to consider
        int j = 0; // index into deletions, next index to delete
        int k = 0; // index into dest, next slot to put an element
        for (i = j = k = 0; j < nDeletions; ++i) {
            if (i == deletions[j])
                ++j; // and don't copy src[i]
            else
                dest[k++] = src[i];
        }
        // copy remaining undeleted elements
        System.arraycopy (src, i, dest, k, src.length - i);
        return dest;
    }
                

    /**
     * Dump feature statistics to debugging output.
     */
    public void dumpFeatures () {
        for (int i = 0; i < features.length; ++i) {
            debug.println (features[i]
                                   + ": mean = " + mean[i]
                                   + ", median = " + median[i]
                                   + ", var = " + variance[i]);
        }
    }
}
