/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.ml;

import lapis.*;
import lapisx.sort.*;

public class DistanceMetric {
    double[] weight;

    public void pruneWeights (final double[] weight, NamedRegionSet[] features) {
        int n = features.length;

        // rank features by weight
        Integer[] featuresByWeight = new Integer [n]; 
        for (int i = 0; i < n; ++i)
            featuresByWeight[i] = new Integer (i);
        Quicksort.sort (new SortableArray (featuresByWeight) {
                public int compare (Object o1, Object o2) {
                    int i1 = ((Integer)o1).intValue ();
                    int i2 = ((Integer)o2).intValue ();
                    double w1 = weight[i1];
                    double w2 = weight[i2];
                    if (w1 < w2)
                        return 1;
                    else if (w1 > w2)
                        return -1;
                    else
                        return i1 - i2;
                }
            });

        for (int f = 0; f < n ; ++f) {
            int i = featuresByWeight[f].intValue ();
            if (weight[i] == 0.0)
                break;

            // test whether feature is a generalization or specialization
            // of a previously-selected feature; if so, skip it.
        searchForRelatedFeature:
            for (int g = f-1; g >= 0; --g) {
                int j = featuresByWeight[g].intValue ();
                if (weight[j] == 0)
                    continue; // j is already discarded

                // cheap test: do features have the same weight?
                if (weight[j] > weight[i])
                    break; // no, and all earlier weights will differ too

                // more expensive test: is one feature a subset of the
                // other across the entire document?

                switch (features[i].compareTo (features[j])) {
                case RegionSet.EQUAL:
                    // FIX: decide which to prefer
                case RegionSet.SUPERSET:
                    // discard feature i by zeroing its weight
//                     debug.println ("discarded " + features[i]);
                    weight[i] = 0;
                    break searchForRelatedFeature;
                case RegionSet.SUBSET:
                    // discard feature j by zeroing its weight
//                     debug.println ("discarded " + features[j]);
                    weight[j] = 0;
                    break;
                case RegionSet.DISJOINT:
                    // don't know how to handle this case yet
//                     debug.println (features[i] + " and " 
//                                    + features[j] + " are disjoint");
                    break;
                default:
                    break;
                }
            }
        }

        for (int f = 0; f < n ; ++f) {
            int i = featuresByWeight[f].intValue ();
            if (weight[i] != 0)
                Similarity.debug.println ("weight[" 
                               + features[i]
                               + "]=" 
                               + weight[i]);
        }
    }

    public double getDistance (double[] v1, double[] v2) {
        double sum = 0.0f;
        int diffs = 0;
        for (int i = 0, n = v1.length; i < n; ++i) {
            double d = v1[i] - v2[i];
            if (d != 0)
                ++diffs;
            sum += weight[i] * d * d;
        }
        return sum;
    }

    public double getWeight (int i) {
        return weight[i];
    }
}
