/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.ml;

import lapis.*;
import lapisx.enum.*;
import java.util.*;

public class HybridInference implements InferenceAlgorithm {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    NamedRegionSet range;
    SimulEdit lastSimulEdit;
    List allSimulEdits;
    Inference inference;

    public HybridInference (NamedRegionSet range) {
        this.range = range;

        allSimulEdits = new LinkedList ();
        if (range.getSize () > 1)
            allSimulEdits.add (new SimulEdit (range));

        // split range into record sets with RecordSetGuesser
        RecordSetGuesser rsg = new RecordSetGuesser (range);
        allSimulEdits.addAll (rsg.getGuesses ());

        // default to first simuledit in list
        lastSimulEdit = 
            (allSimulEdits.size () > 0)
            ? (SimulEdit) allSimulEdits.get (0)
            : null;

        // create Inference object
        //NamedRegionSet empty = new NamedRegionSet (range, RegionSet.EMPTY);
		NamedRegionSet empty = new NamedRegionSet (RegionSet.EMPTY);
        inference = new Inference (empty, empty, range);
    }

    abstract class DeferredEnumeration implements Enumeration {
        Enumeration e;
        protected abstract Enumeration realize ();
        public boolean hasMoreElements () {
            if (e == null)
                e = realize ();
            return e.hasMoreElements ();
        }
        public Object nextElement () {
            if (e == null)
                e = realize ();
            return e.nextElement ();
        }
    }

    class Generalizer extends DeferredEnumeration {
        InferenceAlgorithm alg;
        NamedRegionSet positiveExamples;
        NamedRegionSet negativeExamples;
        public Generalizer (InferenceAlgorithm alg,
                            NamedRegionSet positiveExamples,
                            NamedRegionSet negativeExamples) {
            this.alg = alg;
            this.positiveExamples = positiveExamples;
            this.negativeExamples = negativeExamples;
        }
        protected Enumeration realize () {
            return alg.generalize (positiveExamples,
                                  negativeExamples);
        }
    }

    public RestartableEnumeration generalize (NamedRegionSet positiveExamples,
                                              NamedRegionSet negativeExamples) {
        long start = System.currentTimeMillis ();
        try {
            Enumeration hyp;
            boolean foundOne = false;
            List allHypotheses = new ArrayList (allSimulEdits.size() + 2);

            if (lastSimulEdit != null) {
                hyp = new Generalizer (lastSimulEdit,
                                       positiveExamples, 
                                       negativeExamples);
                if (hyp.hasMoreElements ())
                    foundOne = true;
                allHypotheses.add (hyp);
            }
            
            // if lastSimulEdit couldn't make a hypothesis,
            // try all the other simuledits
            for (Iterator g = allSimulEdits.iterator ();
                 g.hasNext (); ) {
                SimulEdit se = (SimulEdit) g.next ();
                if (se.equals (lastSimulEdit))
                    continue; // already tried it above
                
                hyp = new Generalizer (se, 
                                       positiveExamples,
                                       negativeExamples);
                if (!foundOne && hyp.hasMoreElements ()) {
                    foundOne = true;
                    debug.println ("switched to " + se);
                    lastSimulEdit = se;
                }
                allHypotheses.add (hyp);
            }
            
            // use Inference as a last-ditch fallback
            Enumeration fallback =
                new Generalizer (inference,
                                 positiveExamples,
                                 negativeExamples);

            // return an enumeration that picks the top hypothesis
            // from each of allHypotheses.  
            return new MemoizedEnumeration (new ConcatEnumeration 
                (new FirstOnly (allHypotheses), fallback));
        } finally {
            debug.println ("generalization time: " 
                           + (System.currentTimeMillis () - start)
                           + " ms");
        }
    }

    /**
     * Takes a list of enumerations and returns only the first element
     * of each (nonempty) enumeration.
     */
    static class FirstOnly extends FilteredEnumeration {
        public FirstOnly (List enums) {
            super (Collections.enumeration (enums));
        }
        public void transform (Object o) {
            Enumeration e = (Enumeration) o;
            if (e.hasMoreElements ())
                yield (e.nextElement ());
        }
    }
    
    public NamedRegionSet getRecords () {
        return 
            (lastSimulEdit != null)
            ? lastSimulEdit.getRecords ()
            : range;
    }
    
    public void addEditLocation (NamedRegionSet endpoints, int len) {
        for (Iterator g = allSimulEdits.iterator (); g.hasNext (); ) {
            SimulEdit simulEdit = (SimulEdit) g.next ();
            if (endpoints.getSize () == simulEdit.getRecords ().getSize ()) {
                simulEdit.addEditLocation (endpoints, len);
            }
        }
        inference.addEditLocation (endpoints, len);
    }

    public void addPasteLocation (NamedRegionSet targetEnds, NamedRegionSet sourceSelection) {
        for (Iterator g = allSimulEdits.iterator (); g.hasNext (); ) {
            SimulEdit simulEdit = (SimulEdit) g.next ();
            if (targetEnds.getSize () == simulEdit.getRecords ().getSize ()) {
                simulEdit.addPasteLocation (targetEnds, sourceSelection);
            }
        }
        inference.addPasteLocation (targetEnds, sourceSelection);
    }


    public void addCopyLocation (NamedRegionSet sel) {
        for (Iterator g = allSimulEdits.iterator (); g.hasNext (); ) {
            SimulEdit simulEdit = (SimulEdit) g.next ();
            if (sel.getSize () == simulEdit.getRecords ().getSize ()) {
                simulEdit.addCopyLocation (sel);
            }
        }
        inference.addCopyLocation (sel);
    }

    public void undoTo (DocumentVersion ver) {
        for (Iterator g = allSimulEdits.iterator (); g.hasNext (); ) {
            SimulEdit simulEdit = (SimulEdit) g.next ();
            simulEdit.undoTo (ver);
        }
        inference.undoTo (ver);
    }
    
    public void close () {
        for (Iterator g = allSimulEdits.iterator (); g.hasNext (); ) {
            SimulEdit simulEdit = (SimulEdit) g.next ();
            simulEdit.close ();
        }
        inference.close ();
    }
}
