/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.ml;

import lapis.*;
import lapis.tc.*;
import java.util.*;
import lapisx.util.*;
import lapisx.enum.*;
import lapisx.sort.*;

public class Inference implements InferenceAlgorithm {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    // maximum length of generated literals, in characters
    static final int MAX_LITERAL = 100;


    Document doc;
    DocumentVersion ver;
    RegionSet positiveExamples = RegionSet.EMPTY; // positive examples
    RegionSet negativeExamples = RegionSet.EMPTY; // negative examples

    // inferences are limited to this range
    NamedRegionSet range;
    RegionSet inRange; // range.in ()

    // all (nonempty) pattern features in library
    Vector libraryFeatures;

    // Features consistent with all the positive examples.
    // These vectors consist of library features or literal features
    // that bear the specified relationship to all the positive examples.
    Vector startsContainsNoBackgroundPattern = new Vector ();
    Vector startsNoBackgroundPattern = new Vector ();
    Vector startsBackgroundPattern = new Vector ();

    Vector endsContainsNoBackgroundPattern = new Vector ();
    Vector endsNoBackgroundPattern = new Vector ();
    Vector endsBackgroundPattern = new Vector ();

    Vector beforeNoBackgroundPattern = new Vector ();
    Vector beforeBackgroundPattern = new Vector ();

    Vector afterNoBackgroundPattern = new Vector ();
    Vector afterBackgroundPattern = new Vector ();

    Vector containsRegularPattern = new Vector ();
    Vector containsVaryingPattern = new Vector ();

    Vector inRegularPattern = new Vector ();
    Vector inVaryingPattern = new Vector ();

    Vector startsContainsNoBackgroundLiteral = new Vector ();
    Vector startsContainsBackgroundLiteral = new Vector ();

    Vector endsContainsNoBackgroundLiteral = new Vector ();
    Vector endsContainsBackgroundLiteral = new Vector ();

    Vector beforeNoBackgroundLiteral = new Vector ();
    Vector beforeBackgroundLiteral = new Vector ();

    Vector afterNoBackgroundLiteral = new Vector ();
    Vector afterBackgroundLiteral = new Vector ();

    Vector containsRegularLiteral = new Vector ();
    Vector containsVaryingLiteral = new Vector ();

    Vector beforeAfterRegularFeatures = new Vector ();

    // memoized results of various feature enumerations.
    // Cleared when features are generated or updated.
    Vector startFeatures;
    Vector endFeatures;
    Vector beforeFeatures;
    Vector afterFeatures;
    Vector allConstraints;

    // substrings common to all positive examples
    CommonSubstrings cs;

    // the RegularFeatures (both patterns and literals) that predict
    // the likeliest number of records
    Vector likelyRegularFeatures = new Vector ();

    // memoized result of getScore ().  Cleared when likelyRegularFeatures
    // changes.
    Hashtable scores = new Hashtable ();

    // partitions (flat region sets consistent with the positive examples)
    Vector partitions = new Vector ();

    // hypotheses consistent with all positive and negative examples
    Vector hypotheses = new Vector ();


    public Inference (NamedRegionSet positiveExamples, 
                      NamedRegionSet negativeExamples,
                      NamedRegionSet range) {
		// TODO: support multiple documents
        
        // minwu
		//this.doc = (Document)(range.getDocuments().toArray())[0];
		//this.ver = (DocumentVersion)(range.getVersions().toArray())[0];
        
        Iterator iver = range.getVersions().versions();
        while (iver.hasNext()) {
            this.ver = (DocumentVersion)iver.next();
            this.doc = this.ver.getDocument();
        }
        
        this.positiveExamples = positiveExamples;
        this.negativeExamples = negativeExamples;
        this.range = range;

        changedDocument ();
    }

    public synchronized RestartableEnumeration getPartitions () {
        return new MemoizedEnumeration (partitions);
    }

    public synchronized RestartableEnumeration getHypotheses () {
        return new MemoizedEnumeration (hypotheses);
    }

    public synchronized RegionSet getPartitionUnion () {
        MutableRegionSet result = new MutableRegionSet ();
        for (int i = 0, n = partitions.size (); i < n; ++i) {
            Feature partition = (Feature) partitions.elementAt (i);
            result.insert (partition);
        }
        return result;
    }

    public synchronized RestartableEnumeration generalize (NamedRegionSet pos,
                                                           NamedRegionSet neg) {
        setExamples (pos, neg);
        return getHypotheses ();
    }

    public synchronized NamedRegionSet getRecords () {
		return new NamedRegionSet (RegionSet.EMPTY);
    }

    public void addEditLocation (NamedRegionSet endpoints, int len) {
    }

    public void addPasteLocation (NamedRegionSet targetEnds, NamedRegionSet sourceSelection) {
    }

    public void addCopyLocation (NamedRegionSet sel) {
    }

    public void undoTo (DocumentVersion ver) {
    }

    public void close () {        
    }
    
    public synchronized void setExamples (NamedRegionSet selPos, NamedRegionSet selNeg) {
        RegionSet newPos = selPos;
        RegionSet newNeg = selNeg;

        debug.println ("positive examples:\n" 
					   + newPos.toStringWithContent ()); 
        debug.println ("negative examples:\n"
		 			   + newNeg.toStringWithContent ()); 

        positiveExamples = newPos;
        negativeExamples = newNeg;

        // TODO: support multiple documents
        
        // minwu
		//debug.assertion(selPos.getDocuments().size() == 1);
		//debug.assertion(selPos.getVersions().size() == 1);
		//Document newdoc = (Document)(selPos.getDocuments().toArray())[0];
		//DocumentVersion newver = (DocumentVersion)(selPos.getVersions().toArray())[0];
        
        debug.assertion(selPos.getVersions().getSize() == 1);
        Document newdoc = null;
        DocumentVersion newver = null;

        Iterator iver = selPos.getVersions().versions();
        while (iver.hasNext()) {
            newver = (DocumentVersion)iver.next();
            newdoc = newver.getDocument();
        }        
		if (newdoc != doc
		   || newver != ver) {
		   doc = newdoc;
		   ver = newver;
		   changedDocument ();
	   } else
		   changedExamples ();
    }


    void changedDocument () {
        debug.println ("updating for new document");
        libraryFeatures = null;
        changedExamples ();
    }

    void changedExamples () {
        debug.println ("updating for all examples");
        inRange = range.convertToLatest ().in ();
        debug.println ("in range = " + inRange);

        partitions.removeAllElements ();
        hypotheses.removeAllElements ();
        
        if (positiveExamples.isEmpty ())
            return;

        generateFeatures ();
        generatePartitions ();
        generateHypotheses ();
    }

    void addedPositiveExamples () {
        debug.println ("updating for new positive examples");
        updateFeatures (positiveExamples);
        generatePartitions ();
        generateHypotheses ();
    }

    void addedNegativeExamples () {
        debug.println ("updating for new negative examples");
        generateHypotheses ();
    }


    //////////////////////////
    ////////////////////////// Feature generation
    //////////////////////////

    // generate features (both library and literals) that are
    // starting/ending/jbef/jaft all positive examples.
    // store them in the vectors above.
    void generateFeatures () {
        startFeatures = endFeatures = 
            beforeFeatures = afterFeatures = 
            allConstraints = null;

        if (libraryFeatures == null) {
            Main.getProgress ().fireStartProgress 
                ("Generating features", 
                 PatternLibrary.getDocumentLibrary (doc).getSize ());
            try {
                libraryFeatures = accumulate 
                    (null, constrainToRange (new LibraryGenerator (doc)));
            
            } finally {
                Main.getProgress ().fireDoneProgress ();
            }
        }

        generatePatternFeatures ();
        generateLiteralFeatures ();
        generateBeforeAfterRegularFeatures ();
        generateLikelyRegularFeatures ();
        dumpAllFeatures ();
    }

    Enumeration constrainToRange (Enumeration features) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;

                    // fix f so that it only matches inside range
                    RegionSet s = f.and (inRange).crispen ();

                    if (!s.isEmpty ())
                        yield (new Feature (f, s, f.toString ()));
                }
            };
    }

    // filter or regenerate the feature vectors when new positive
    // examples are given.
    void updateFeatures (RegionSet newPos) {
        startFeatures = endFeatures = 
            beforeFeatures = afterFeatures = 
            allConstraints = null;
        
        updatePatternFeatures (newPos);
        updateLiteralFeatures (newPos);
        generateBeforeAfterRegularFeatures ();
        generateLikelyRegularFeatures ();
        dumpAllFeatures ();
    }

    void dumpAllFeatures () {
        dumpFeatures ("startsContainsNoBackground ", 
                      startsContainsNoBackgroundPattern.elements ());
        dumpFeatures ("startsNoBackground ", 
                      startsNoBackgroundPattern.elements ());
        dumpFeatures ("startsBackground ", 
                      startsBackgroundPattern.elements ());
        dumpFeatures ("endsContainsNoBackground ", 
                      endsContainsNoBackgroundPattern.elements ());
        dumpFeatures ("endsNoBackground ", 
                      endsNoBackgroundPattern.elements ());
        dumpFeatures ("endsBackground ", 
                      endsBackgroundPattern.elements ());
        dumpFeatures ("beforeNoBackground ", 
                      beforeNoBackgroundPattern.elements ());
        dumpFeatures ("beforeBackground ", 
                      beforeBackgroundPattern.elements ());
        dumpFeatures ("afterNoBackground ", 
                      afterNoBackgroundPattern.elements ());
        dumpFeatures ("afterBackground ", 
                      afterBackgroundPattern.elements ());
        dumpFeatures ("containsRegular ", 
                      containsRegularPattern.elements ());
        dumpFeatures ("containsVarying ", 
                      containsVaryingPattern.elements ());
        dumpFeatures ("startsContainsNoBackground ", 
                      startsContainsNoBackgroundLiteral.elements ());
        dumpFeatures ("startsContainsBackground ", 
                      startsContainsBackgroundLiteral.elements ());
        dumpFeatures ("endsContainsNoBackground ", 
                      endsContainsNoBackgroundLiteral.elements ());
        dumpFeatures ("endsContainsBackground ", 
                      endsContainsBackgroundLiteral.elements ());
        dumpFeatures ("beforeNoBackground ", 
                      beforeNoBackgroundLiteral.elements ());
        dumpFeatures ("beforeBackground ", 
                      beforeBackgroundLiteral.elements ());
        dumpFeatures ("afterNoBackground ", 
                      afterNoBackgroundLiteral.elements ());
        dumpFeatures ("afterBackground ", 
                      afterBackgroundLiteral.elements ());
        dumpFeatures ("containsRegular ", 
                      containsRegularLiteral.elements ());
        dumpFeatures ("containsVarying ", 
                      containsVaryingLiteral.elements ());
        dumpFeatures ("inRegular ", 
                      inRegularPattern.elements ());
        dumpFeatures ("inVarying ", 
                      inVaryingPattern.elements ());
        dumpFeatures ("beforeAfterRegular ", 
                      beforeAfterRegularFeatures.elements ());
        dumpFeatures ("likelyRegularFeatures ", 
                      likelyRegularFeatures.elements ());
    }

    //////////////////////////
    ////////////////////////// Pattern feature generation
    //////////////////////////

    void generatePatternFeatures () {
        startsContainsNoBackgroundPattern.removeAllElements ();
        startsNoBackgroundPattern.removeAllElements ();
        startsBackgroundPattern.removeAllElements ();
        endsContainsNoBackgroundPattern.removeAllElements ();
        endsNoBackgroundPattern.removeAllElements ();
        endsBackgroundPattern.removeAllElements ();
        beforeNoBackgroundPattern.removeAllElements ();
        beforeBackgroundPattern.removeAllElements ();
        afterNoBackgroundPattern.removeAllElements ();
        afterBackgroundPattern.removeAllElements ();
        containsRegularPattern.removeAllElements ();
        containsVaryingPattern.removeAllElements ();
        inRegularPattern.removeAllElements ();
        inVaryingPattern.removeAllElements ();

        RegionSet bkgd = getBackground ();

        // generate "starts" and "ends" pattern features
        if (!positiveExamples.isAllPoints ()) {
            // use starts or ends features only if positive examples 
            // aren't points

            accumulate
                (startsContainsNoBackgroundPattern,
                 filterFeatures
                 (filterFeatures
                  (filterFeatures 
                   (allLibraryFeatures (),
                    positiveExamples.starting (bkgd),
                    null),
                   positiveExamples.starting (RegionSet.EMPTY),
                   startsBackgroundPattern),
                  positiveExamples.startingIn (RegionSet.EMPTY),
                  startsNoBackgroundPattern));

            accumulate
                (endsContainsNoBackgroundPattern,
                 filterFeatures
                 (filterFeatures
                  (filterFeatures 
                   (allLibraryFeatures (),
                    positiveExamples.ending (bkgd),
                    null),
                   positiveExamples.ending (RegionSet.EMPTY),
                   endsBackgroundPattern),
                  positiveExamples.endingIn (RegionSet.EMPTY),
                  endsNoBackgroundPattern));

        }

        // generate "just before" and "just after" pattern features
        accumulate
            (beforeNoBackgroundPattern,
             filterFeatures
             (filterFeatures 
              (allLibraryFeatures (),
               positiveExamples.justAfter (bkgd),
               null),
              positiveExamples.justAfter (RegionSet.EMPTY),
              beforeBackgroundPattern));

        accumulate
            (afterNoBackgroundPattern,
             filterFeatures
             (filterFeatures 
              (allLibraryFeatures (),
               positiveExamples.justBefore (bkgd),
               null),
              positiveExamples.justBefore (RegionSet.EMPTY),
              afterBackgroundPattern));

        // generate "contains" patterns, both regular and varying
        generateContainsPattern (allLibraryFeatures (),
                                 positiveExamples.in (),
                                 containsRegularPattern,
                                 containsVaryingPattern);

        // generate "in" patterns, both regular and varying
        generateInPattern ();
    }

    void updatePatternFeatures (RegionSet newPos) {
        RegionSet bkgd = getBackground ();

        filterFeatureVector 
            (startsContainsNoBackgroundPattern,
             newPos.startingIn (RegionSet.EMPTY),
             startsNoBackgroundPattern);
        filterFeatureVector 
            (startsNoBackgroundPattern,
             newPos.starting (RegionSet.EMPTY),
             startsBackgroundPattern);
        filterFeatureVector 
            (startsBackgroundPattern,
             newPos.starting (bkgd),
             null);

        filterFeatureVector 
            (endsContainsNoBackgroundPattern,
             newPos.startingIn (RegionSet.EMPTY),
             endsNoBackgroundPattern);
        filterFeatureVector 
            (endsNoBackgroundPattern,
             newPos.starting (RegionSet.EMPTY),
             endsBackgroundPattern);
        filterFeatureVector 
            (endsBackgroundPattern,
             newPos.starting (bkgd),
             null);

        filterFeatureVector 
            (beforeNoBackgroundPattern,
             newPos.justAfter (RegionSet.EMPTY),
             beforeBackgroundPattern);
        filterFeatureVector 
            (beforeBackgroundPattern,
             newPos.justAfter (bkgd),
             null);

        filterFeatureVector 
            (afterNoBackgroundPattern,
             newPos.justBefore (RegionSet.EMPTY),
             afterBackgroundPattern);
        filterFeatureVector 
            (afterBackgroundPattern,
             newPos.justBefore (bkgd),
             null);

        updateContainsPattern (newPos.in (),
                               containsRegularPattern,
                               containsVaryingPattern);

        generateInPattern ();
    }

    RegionSet getBackground () {
        PatternLibrary lib = PatternLibrary.getDocumentLibrary (doc);
        Pattern background = lib.get (lapis.parsers.SystemParser.BACKGROUND);
        PatternCache cache = doc.getPatternCache ();
        RegionSet bkgd = cache.get (background);
        if (bkgd == null) {
            bkgd = background.match (doc);
            cache.put (background, bkgd);
        }
        return bkgd;
    }



    //////////////////////////
    ////////////////////////// Literal feature generation
    //////////////////////////

    void generateLiteralFeatures () {
    // 
    // Removed generation of jb/ja/st/en literals with background,
    // because it was generating way too many literals for repeated
    // strings of background chars like "----" and "     "
    //
        //        final RegionSet bkgd = getBackground ();
        final RegionSet bkgd = RegionSet.EMPTY;

        startsContainsNoBackgroundLiteral.removeAllElements ();
        startsContainsBackgroundLiteral.removeAllElements ();
        endsContainsNoBackgroundLiteral.removeAllElements ();
        endsContainsBackgroundLiteral.removeAllElements ();
        beforeNoBackgroundLiteral.removeAllElements ();
        beforeBackgroundLiteral.removeAllElements ();
        afterNoBackgroundLiteral.removeAllElements ();
        afterBackgroundLiteral.removeAllElements ();

        // generate starts/ends/jbef/jaft literal features.

        // Only maximal literals are generated, in the following
        // sense.  A prefix/suffix is nonmaximal if every occurrence
        // of it in the document is found in some longer prefix/suffix.

        Enumeration e;
        final String text = doc.getText ();
        final String revText = Str.reverse (text);
        Vector v;


        // find "just before L" features
        e = new JavaEnumeration (positiveExamples.regions (), false);
        e = new FilteredEnumeration (e) {
                public void transform (Object o) {
                    Region r = (Region) o;
                    yield (extract (text, (Region) r.justAfter (bkgd), true));
                }
            };
        e = getLongestCommonPrefix (e);
        e = maximalPrefixes (e, text);
        accumulate
            (beforeNoBackgroundLiteral,
             filterFeatures
             (literalFeatures (e),
              positiveExamples.justAfter (RegionSet.EMPTY),
              beforeBackgroundLiteral));


        // find "just after L" features
        e = new JavaEnumeration (positiveExamples.regions (), false);
        e = new FilteredEnumeration (e) {
                public void transform (Object o) {
                    Region r = (Region) o;
                    yield (extract (revText, (Region) r.justBefore (bkgd), false));
                }
            };
        e = getLongestCommonPrefix (e);
        e = maximalPrefixes (e, revText);
        e = reverse (e);
        accumulate
            (afterNoBackgroundLiteral,
             filterFeatures
             (literalFeatures (e),
              positiveExamples.justBefore (RegionSet.EMPTY),
              afterBackgroundLiteral));

        // generate "starts" and "ends" pattern features
        if (!positiveExamples.isAllPoints ()) {
            // use starts or ends features only if positive examples 
            // aren't points

            // find "starting L" features
            e = new JavaEnumeration (positiveExamples.regions (), false);
            e = new FilteredEnumeration (e) {
                    public void transform (Object o) {
                        Region r = (Region) o;
                        yield (extract (text, (Region) r.startingIn (bkgd), true));
                    }
                };
            e = getLongestCommonPrefix (e);
            e = maximalPrefixes (e, text);
            accumulate
                (startsContainsNoBackgroundLiteral,
                 filterFeatures
                 (literalFeatures (e),
                  positiveExamples.startingIn (RegionSet.EMPTY),
                  startsContainsBackgroundLiteral));

            // find "ending L" features
            e = new JavaEnumeration (positiveExamples.regions (), false);
            e = new FilteredEnumeration (e) {
                    public void transform (Object o) {
                        Region r = (Region) o;
                        yield (extract (revText, (Region) r.endingIn (bkgd), false));
                    }
                };
            e = getLongestCommonPrefix (e);
            e = maximalPrefixes (e, revText);
            e = reverse (e);
            accumulate
                (endsContainsNoBackgroundLiteral,
                 filterFeatures
                 (literalFeatures (e),
                  positiveExamples.endingIn (RegionSet.EMPTY),
                  endsContainsBackgroundLiteral));

        }


        // find regular and varying common substrings
        generateContainsLiteral ();

    }

    void updateLiteralFeatures (RegionSet newPos) {
        final RegionSet bkgd = getBackground ();

        filterFeatureVector 
            (startsContainsNoBackgroundLiteral,
             newPos.startingIn (RegionSet.EMPTY),
             startsContainsBackgroundLiteral);
        filterFeatureVector 
            (startsContainsBackgroundLiteral,
             newPos.startingIn (bkgd),
             null);

        filterFeatureVector 
            (endsContainsNoBackgroundLiteral,
             newPos.endingIn (RegionSet.EMPTY),
             endsContainsBackgroundLiteral);
        filterFeatureVector 
            (endsContainsBackgroundLiteral,
             newPos.endingIn (bkgd),
             null);


        filterFeatureVector 
            (beforeNoBackgroundLiteral,
             newPos.justAfter (RegionSet.EMPTY),
             beforeBackgroundLiteral);
        filterFeatureVector 
            (beforeBackgroundLiteral,
             newPos.justAfter (bkgd),
             null);

        filterFeatureVector 
            (afterNoBackgroundLiteral,
             newPos.justBefore (RegionSet.EMPTY),
             afterBackgroundLiteral);
        filterFeatureVector 
            (afterBackgroundLiteral,
             newPos.justBefore (bkgd),
             null);

        updateContainsLiteral (newPos);
    }

    static class Prefixes {
        int startLow;
        int startHigh;
        int end;
        String text;
        public Prefixes (int sl, int sh, int e, String t) {
            startLow = sl;
            startHigh = sh;
            end = e;
            text = t;
        }
        public String toString () {
            return "[" + startLow + "-" + startHigh + "," + end + "]";
        }
    }
        
    public Prefixes extract (String text, Region r, boolean forward) {
        Prefixes p;
        int len = text.length ();

        // check whether we want a forward or reversed string
        if (forward) {
            // forward string
            p = new Prefixes (r.getStartLow (), 
                              r.getStartHigh (),
                              r.getEnd (), 
                              text);
        } else{
            p = new Prefixes (len - r.getEndHigh (),
                              len - r.getEndLow (),
                              len - r.getStart (),
                              text);
        }

        p.startLow = Math.min (len, Math.max (0, p.startLow));
        p.startHigh = Math.min (len, Math.max (0, p.startHigh));
        p.end = Math.min (len, Math.max (0, p.end));

        p.end = Math.min (p.end, p.startLow + MAX_LITERAL);
        p.startHigh = Math.min (p.startHigh, p.end);

        debug.println ("extracted " + p + " for " + r.toStringWithContent ());
        return p;
    }

    Enumeration allEqual (Enumeration e) {
        String s = (String) e.nextElement ();
        if (s.length () == 0)
            return new EmptyEnumeration ();
        while (e.hasMoreElements ())
            if (!s.equals (e.nextElement ()))
                return new EmptyEnumeration ();
        return new ArrayEnumeration (new String[] { s } );
    }  

    Enumeration maximalPrefixes (Enumeration e, final String r) {
        return new FilteredEnumeration (e) {
                public void transform (Object o) {
                    String s = (String) o;
                    
                    if (s.length () == 0)
                        return;

                    debug.println ("longest common prefix:");
                    debug.println (escapeNewlines (s));

                    // find token boundaries
                    boolean[] endsToken = new boolean[s.length ()];
                    int type = getCharType (s.charAt (0));
                    for (int i = 1, n = s.length (); i < n; ++i) {
                        int type2 = getCharType (s.charAt(i));
                        endsToken[i-1] = (type != type2);
                        type = type2;
                    }
                    endsToken[endsToken.length - 1] = true;
                        
                    // search r for maximal prefixes of s
                    // maximal[i] is true if s.substring (0, i+1) is a 
                    // maximal prefix.
                    boolean[] maximal = new boolean[s.length ()];
                    char firstChar = s.charAt (0);
                    for (int p = 0, q = 0, i = 0; (i = r.indexOf (firstChar, p)) != -1; p = i+1) {
                        int lcp = Str.longestCommonPrefixLength 
                            (r.substring (i), s);
                        if (lcp > 0)
                            maximal[lcp-1] = true;
                    }

                    debug.println ("maximal strings:");
                    for (int j = 0; j < maximal.length; ++j)
                        if (maximal[j])
                            debug.println (escapeNewlines (s.substring (0, j+1)));

                    // check s for periodic prefixes.
                    int nMaxCycles = 1; // maximum number of cycles
                    int periodMaxCycles = 0; // length of period 
                                              // corresponding to nMaxCycles
                    for (int period = s.length ()/2; period >= 1; --period) {
                        int nCycles = countCycles (s, period);
                        if (nCycles > nMaxCycles) {
                            nMaxCycles = nCycles;
                            periodMaxCycles = period;
                        }
                    }
                    debug.println ("best period is " 
                                   + escapeNewlines (s.substring (0, periodMaxCycles))
                                   + " repeating " + nMaxCycles + " times");

                    // if s has a periodic prefix x, then we only want
                    // to yield prefixes less than or equal to |x|
                    // and greater than or equal to |x^n| (where n is the
                    // number of cycles of x in s).
                    int minSkipped = periodMaxCycles;
                    int maxSkipped = periodMaxCycles * nMaxCycles - 2;

                    boolean emitted = false;
                    debug.println ("yielded strings:");
                    for (int j = 0; j < maximal.length; ++j) {
                        if (!emitted && (maximal[j] || endsToken[j])) {
                            String t = s.substring (0, j+1);
                            if (j < minSkipped || j > maxSkipped) {
                                debug.println (escapeNewlines (t));
                                yield (t);
                            }
                            emitted = !maximal[j];
                        } else if (maximal[j])
                            emitted = false;
                    }
                }
            };
    }

    int countCycles (String s, int period) {
        if (period == 0)
            return 0;
        else
            return 
                (Str.longestCommonPrefixLength (s, s.substring (period))
                 / period) + 1;
    }

    String escapeNewlines (String s) {
        return Str.escape (s, "\n\t", '\\', "nt");
    }

    static final int LETTER = 0x10000;
    static final int DIGIT = 0x20000;

    static int getCharType (char c) {
        if (Character.isLetter (c))
            return LETTER;
        else if (Character.isDigit (c))
            return DIGIT;
        else
            return c;
    }

    Enumeration getLongestCommonPrefix (Enumeration e) {
        CommonSubstrings cs = new CommonSubstrings ();

        while (e.hasMoreElements ()) {
            Prefixes p = (Prefixes) e.nextElement ();
            for (int i = p.startLow; i <= p.startHigh; ++i)
                cs.add (p.text.substring (p.startLow, p.end), 
                        p.startHigh - p.startLow + 1);
        }

        CommonSubstrings.CommonSubstring[] subs = cs.getLeaves ();
        Vector v = new Vector ();
        for (int i = 0; i < subs.length; ++i)            
            v.addElement (subs[i].s);
        return v.elements ();
    }  

    
    Enumeration reverse (Enumeration strings) {
        return new FilteredEnumeration (strings) {
                public void transform (Object o) {
                    yield (Str.reverse ((String) o));
                }
            };
    }
    
    Enumeration literalFeatures (Enumeration strings) {
        return new FilteredEnumeration (strings) {
                public void transform (Object o) {
                    String s = (String) o;
                    yield (makeCaseSensitiveLiteral (s));
                }
            };
    }

    //////////////////////////
    ////////////////////////// Regular and varying features
    //////////////////////////

    // regular features occur a fixed number of times in every 
    // positive example; varying features occur a varying number
    // of times, but at least once, in each positive example

    void generateContainsPattern (Enumeration features,
                                  RegionSet transformedPosExamples,
                                  Vector regular,
                                  Vector varying) {
        featureIteration:
        while (features.hasMoreElements ()) {
            Feature f = (Feature) features.nextElement ();

            int min = Integer.MAX_VALUE;
            int max = Integer.MIN_VALUE;

            RegionEnumeration re = transformedPosExamples.regions ();
            for (Region pos = re.firstFast (); 
                 pos != null; 
                 pos = re.nextFast ()) {
                int nMatch = ((RegionSet)f).countIntersection(pos);
                if (nMatch == 0)
                    continue featureIteration;

                min = Math.min (min, nMatch);
                max = Math.max (max, nMatch);
            }

            if (min == max && f.isFlat ())
                regular.addElement (new RegularFeature (f, min, f.getSize ()));
            else
                varying.addElement (f);
        }
    }

    void updateContainsPattern (RegionSet transformedNewPos,
                                Vector regular, 
                                Vector varying) {
        // update the varying patterns first, since the
        // regular pattern update below might add elements
        // to varying that we don't want to check twice
        filterFeatureVector (varying,
                             transformedNewPos,
                             null);

        // now update regular patterns
        featureIteration:
        for (int i = 0, n = regular.size (); i < n; ++i) {
            RegularFeature f = (RegularFeature) regular.elementAt (i);

            RegionEnumeration re = transformedNewPos.regions ();
            for (Region pos = re.firstFast (); 
                 pos != null; 
                 pos = re.nextFast ()) {
                int nMatch = ((RegionSet)f).countIntersection (pos);
                if (nMatch != f.nInPos) {
                    regular.removeElementAt (i);
                    --i;
                    --n;
                    if (nMatch == 0)
                        continue featureIteration;
                    else
                        varying.addElement (f);
                }
            }
        }
    }

    void generateContainsLiteral () {
        cs = new CommonSubstrings ();
        updateContainsLiteral (positiveExamples);
    }

    void updateContainsLiteral (RegionSet newPos) {
        RegionEnumeration re = newPos.regions ();
        for (Region r = re.firstFast (); r != null; r = re.nextFast ())
            cs.add (doc.getText (r));

        containsRegularLiteral.removeAllElements ();
        containsVaryingLiteral.removeAllElements ();

        CommonSubstrings.CommonSubstring[] subs = 
            cs.getMaximalCommonSubstrings ();
        for (int i = 0; i < subs.length; ++i) {
            Feature f = makeCaseSensitiveLiteral (subs[i].s);
           	if (subs[i].minOccurrences == subs[i].maxOccurrences)
               	containsRegularLiteral.addElement 
                   	(new RegularFeature (f, 
                       	                 subs[i].minOccurrences,
                           	             f.getSize ()));
           	else
               	containsVaryingLiteral.addElement (f);
        }
    }


    void generateInPattern () {
        inRegularPattern.removeAllElements ();
        inVaryingPattern.removeAllElements ();

        features:
        for (Enumeration e = allLibraryFeatures ();
             e.hasMoreElements (); ) {
            Feature f = (Feature) e.nextElement ();
            RegionEnumeration re = positiveExamples.regions ();
            Region last = null;
            boolean isRegular = f.isFlat ();
            for (Region r = re.firstFast (); 
                 r != null; 
                 r = re.nextFast ()) {
                Region next = f.regions ((Region) r.contains ())
                    .first ();
                if (next == null)
                    continue features;
                else if (next.equals (last))
                    isRegular = false;
                else
                    last = next;
            }
            if (isRegular)
                inRegularPattern.addElement (new InRegularFeature (f));
            else
                inVaryingPattern.addElement (f);
        }
    }

    void generateBeforeAfterRegularFeatures () {
        beforeAfterRegularFeatures.removeAllElements ();

        final RegionSet inPosEx = positiveExamples.in ();

        accumulate 
            (beforeAfterRegularFeatures,
             new FilteredEnumeration (new ConcatEnumeration (new Enumeration[]
                     {
                         beforeNoBackgroundPattern.elements (),
                         beforeNoBackgroundLiteral.elements (),
                         beforeBackgroundPattern.elements (),
                         beforeBackgroundLiteral.elements (),
                     })) {
                     public void transform (Object o) {
                         Feature f = (Feature) o;
                         debug.println ("testing regular feature before " + f);
                         if (f.isFlat() && noneMatch (inPosEx, f))
                             yield (new BeforeRegularFeature (f));
                     }
                 });

        accumulate 
            (beforeAfterRegularFeatures,
             new FilteredEnumeration (new ConcatEnumeration (new Enumeration[]
                     {
                         afterNoBackgroundPattern.elements (),
                         afterNoBackgroundLiteral.elements (),
                         afterBackgroundPattern.elements (),
                         afterBackgroundLiteral.elements (),
                     })) {
                     public void transform (Object o) {
                         Feature f = (Feature) o;
                         debug.println ("testing regular feature after " + f);
                         if (noneMatch (inPosEx, f))
                             yield (new AfterRegularFeature (f));
                     }
                 });
    }

    class RegularFeature extends Feature {
        public int nInPos;
        public int nInRange;
        public double estimate;
        RegionSet constraint;
 
        public RegularFeature (Feature f, int nInPos, int nInRange) {
            super (f, f, f.getPattern ());
            this.nInPos = nInPos;
            this.nInRange = nInRange;
            estimate = ((double) nInRange) / nInPos;
        }

        public RegionSet getConstraint () {
        	if (constraint == null) {
        		RegionEnumeration re = this.regions();
        		MutableRegionSet s = new MutableRegionSet ();
        		
        		Region r = re.first ();
        		
                SingleDocRegionSet currentInRange = 
        						inRange.getSingleDocRegionSet(r.getDocument());
        		
        		int a = currentInRange.getStartLow();
        		int b = r.getStart();
        		
				Document curdoc = currentInRange.getVersion().getDocument();

        		while (r != null) {
        			if (r.getDocument() != curdoc) {
        				curdoc = r.getDocument();
        				currentInRange = 
        						inRange.getSingleDocRegionSet(curdoc);
        			}
        			
        			Region last = r;
        			for (int i = 0; 
        				 i < nInPos && r != null && r.getDocument() == curdoc;
        				 ++i) {
        				last = r;
        				r = re.next();
        			}
        			int c = last.getEnd ();
        			int d = (r != null && r.getDocument() == curdoc) 
        					? r.getStart () 
        					: currentInRange.getEndHigh();
        			s.insert (new FuzzyRegion (a, b, c, d, curdoc));
        			a = c;
        			b = d;
        		}
        		constraint = s;
        	}
        	return constraint;
        }

        public int countIntersection (RegionSet partition) {
            int matches = 0;
            RegionEnumeration re = getConstraint ().regions ();
            for (Region r = re.first (); r != null; r = re.next ())
                if (partition.countIntersection (r) == 1)
                    ++matches;
            return matches;
        }

        public String toDumpString () {
            return super.toString ()
                + "(" + nInRange + "/" + nInPos + "=" + estimate + ")";
        }
    }

    class InRegularFeature extends RegularFeature {
        public InRegularFeature (Feature f) {
            super (f, 1, f.getSize ());
        }

        public RegionSet getConstraint () {
            if (constraint == null)
                constraint = this.in ();
            return constraint;
        }
    }

    class BeforeRegularFeature extends RegularFeature {
        public BeforeRegularFeature (Feature f) {
            super (f, 1, f.getSize ());
        }

        public RegionSet getConstraint () {
            if (constraint == null)
                constraint = this.justBefore (getBackground ());
            return constraint;
        }
    }

    class AfterRegularFeature extends RegularFeature {
        public AfterRegularFeature (Feature f) {
            super (f, 1, f.getSize ());
        }

        public RegionSet getConstraint () {
            if (constraint == null)
                constraint = this.justAfter (getBackground ());
            return constraint;
        }
    }

    //////////////////////////
    ////////////////////////// Enumerating various features
    //////////////////////////

    Enumeration allLibraryFeatures () {
        return libraryFeatures.elements ();
    }

    Enumeration startFeatures () {
        if (startFeatures == null) {
            startFeatures = accumulate
                (null,
                 starts
                 (new ConcatEnumeration (new Enumeration[]
                     {
                         startsContainsNoBackgroundPattern.elements (),
                         startsContainsNoBackgroundLiteral.elements (),
                         startsNoBackgroundPattern.elements (),
                         startsBackgroundPattern.elements (),
                         startsContainsBackgroundLiteral.elements (),
                     })));
        }
        return startFeatures.elements ();
    }

    Enumeration endFeatures () {
        if (endFeatures == null) {
            endFeatures = accumulate
                (null,
                 ends
                 (new ConcatEnumeration (new Enumeration[]
                     {
                         endsContainsNoBackgroundPattern.elements (),
                         endsContainsNoBackgroundLiteral.elements (),
                         endsNoBackgroundPattern.elements (),
                         endsBackgroundPattern.elements (),
                         endsContainsBackgroundLiteral.elements (),
                     })));
        }
        return endFeatures.elements ();
    }

    Enumeration beforeFeatures () {
        if (beforeFeatures == null) {
            beforeFeatures = accumulate
                (null,
                 before
                 (new ConcatEnumeration (new Enumeration[]
                     {
                         beforeNoBackgroundPattern.elements (),
                         beforeNoBackgroundLiteral.elements (),
                         beforeBackgroundPattern.elements (),
                         beforeBackgroundLiteral.elements (),
                     })));
        }
        return beforeFeatures.elements ();
    }

    Enumeration afterFeatures () {
        if (afterFeatures == null) {
            afterFeatures = accumulate
                (null,
                 after
                 (new ConcatEnumeration (new Enumeration[]
                     {
                         afterNoBackgroundPattern.elements (),
                         afterNoBackgroundLiteral.elements (),
                         afterBackgroundPattern.elements (),
                         afterBackgroundLiteral.elements (),
                     })));
        }
        return afterFeatures.elements ();
    }

    Enumeration startDelimiters () {
        return new ConcatEnumeration (new Enumeration[]
        {
            // P (where starts & contains P ignoring nothing is a feature)
            startsContainsNoBackgroundPattern.elements (),
                
                // L (where starts & contains L ignoring nothing is a feature)
                startsContainsNoBackgroundLiteral.elements (),

                // start of P (where starts P ignoring nothing is a feature)
                startOf (startsNoBackgroundPattern.elements ()),

                // end of P (where jaft P ignoring nothing is a feature)
                endOf (afterNoBackgroundPattern.elements ()),

                // end of L (where jaft L ignorning nothing is a feature)
                endOf (afterNoBackgroundLiteral.elements ())
                });
    }

    Enumeration endDelimiters () {
        return new ConcatEnumeration (new Enumeration[]
        {
            // P (where ends & contains P ignoring nothing is a feature)
            endsContainsNoBackgroundPattern.elements (),
                
                // L (where ends & contains L ignoring nothing is a feature)
                endsContainsNoBackgroundLiteral.elements (),

                // end of P (where ends P ignoring nothing is a feature)
                endOf (endsNoBackgroundPattern.elements ()),

                // start of P (where jaft P ignoring nothing is a feature)
                startOf (beforeNoBackgroundPattern.elements ()),

                // start of L (where jaft L ignorning nothing is a feature)
                startOf (beforeNoBackgroundLiteral.elements ())
                });
    }

    Enumeration regularFeatures () {
        return new ConcatEnumeration (new Enumeration[] 
                                      {containsRegularPattern.elements (),
                                      containsRegularLiteral.elements (),
                                      inRegularPattern.elements (),
                                           beforeAfterRegularFeatures.elements ()});
    }

    Enumeration allConstraints () {
        if (allConstraints == null) {
            allConstraints = accumulate
                (null,
                 new ConcatEnumeration (new Enumeration[]
                     {
                         startFeatures (),
                         endFeatures (),
                         afterFeatures (),
                         beforeFeatures (),
                         contains (containsRegularPattern.elements ()),
                         contains (containsRegularLiteral.elements ()),
                         contains (containsVaryingPattern.elements ()),
                         contains (containsVaryingLiteral.elements ()),
                         in (inRegularPattern.elements ()),
                         in (inVaryingPattern.elements ()),
                     }));
        }
        return allConstraints.elements ();
    }

    //////////////////////////
    ////////////////////////// Partition generation
    //////////////////////////

    void generatePartitions () {
        debug.println ("generating partitions");

        partitions.removeAllElements ();
        accumulate (partitions, simplePartitions ()); 
        accumulate (partitions, delimitedPartitions ()); 
             
        clipPartitionsToRange (partitions);

        // sort partitions by score
        rankPartitions (partitions);

        // combine partitions that generate the same region set
        uniq (partitions, true);
    }

    Enumeration simplePartitions () {
        Enumeration library = 
            consistentWithExamples (allLibraryFeatures (), 
                                    positiveExamples, 
                                    RegionSet.EMPTY);

        Vector flat = new Vector (), nonflat = new Vector ();
        while (library.hasMoreElements ()) {
            Feature hyp = (Feature) library.nextElement ();
            if (hyp.isFlat ())
                flat.addElement (hyp);
            else
                nonflat.addElement (hyp);
        }

        Enumeration flattened = flatten (nonflat.elements ());

        Enumeration literals =
            consistentWithExamples 
            (literalFeatures 
             (allEqual 
              (new FilteredEnumeration 
               (new JavaEnumeration 
                (positiveExamples.regions (), true)) {
                      public void transform (Object o) {
                          Region r = (Region) o;
                          yield (doc.getText (r));
                      }
                  })),
             positiveExamples, RegionSet.EMPTY);

        return new ConcatEnumeration 
            (new Enumeration[] {
                flat.elements (), 
                    literals,
                    flattened,
                    });
    }

    Enumeration flatten (Enumeration hyps) {
        return new FilteredEnumeration (hyps) {
                public void transform (Object o) {
                    Feature hyp = (Feature) o;
                    for (Enumeration e = startFeatures ();
                         e.hasMoreElements (); ) {
                        Feature f = (Feature) e.nextElement ();
                        Feature h = constrain (hyp, f);
                        if (h.isFlat ())
                            yield (h);
                    }
                    for (Enumeration e = endFeatures ();
                         e.hasMoreElements (); ) {
                        Feature f = (Feature) e.nextElement ();
                        Feature h = constrain (hyp, f);
                        if (h.isFlat ())
                            yield (h);
                    }
                    for (Enumeration e = beforeFeatures ();
                         e.hasMoreElements (); ) {
                        Feature f = (Feature) e.nextElement ();
                        Feature h = constrain (hyp, f);
                        if (h.isFlat ())
                            yield (h);
                    }
                    for (Enumeration e = afterFeatures ();
                         e.hasMoreElements (); ) {
                        Feature f = (Feature) e.nextElement ();
                        Feature h = constrain (hyp, f);
                        if (h.isFlat ())
                            yield (h);
                    }
                }
            };
    }

                    
    Enumeration delimitedPartitions () {
        Enumeration e;
        if (positiveExamples.isAllPoints ())
            e = new ConcatEnumeration 
                (startDelimiters (),
                 endDelimiters ());
        else
            e = fromTo (startDelimiters (),
                        endDelimiters ());

        return consistentWithExamples (e, positiveExamples, RegionSet.EMPTY);

    }

    Enumeration fromTo (Enumeration starts, Enumeration ends) {
        return new PairEnumeration (starts, ends) {
                public void transform (Object o1, Object o2) {
                    Feature st = (Feature) o1;
                    Feature en = (Feature) o2;
                    TC.Node tcStart = st.getPattern ().getRoot ();
                    TC.Node tcEnd = en.getPattern ().getRoot ();
                    TC.Node tc = 
                        new TC.FromTo (tcStart, tcEnd);
                    String expr = "from " + st + "\n  to " + en;
                    Feature f = makeFeature (new TC (tc, expr));
                    if (!f.isEmpty ())
                        yield (f);
                }
            };
    }

    void clipPartitionsToRange (Vector partitions) {
        for (int i = 0, n = partitions.size (); i < n; ++i) {
            Feature partition = (Feature) partitions.elementAt (i);
            //RegionSet parts = partition.getRegions ();
            Feature clippedPartition = 
                new Feature (partition, partition.and (inRange), partition.toString ());
            partitions.setElementAt (clippedPartition, i);
        }
    }

    //////////////////////////
    ////////////////////////// Constraining hypotheses to exclude negative examples
    //////////////////////////

    void generateHypotheses () {
        debug.println ("generating hypotheses");
        if (negativeExamples.isEmpty ())
            hypotheses = partitions;
        else {
            // add constraints to partitions
            // in order to exclude negative examples 
            hypotheses = accumulate 
                (null,
                 constrainHypotheses (partitions.elements ()));

            // combine hypotheses that generate the same region set
            uniq (hypotheses, false);
        }
        debug.println ("done");
    }

    Enumeration constrainHypotheses (Enumeration hyps) {
        return new FilteredEnumeration (hyps) {
                public void transform (Object o) {
                    Feature hyp = (Feature) o;
                    RegionSet neg = negativeExamples.and (hyp);

                    while (!neg.isEmpty ()) {
                        int min = neg.getSize ();
                        debug.println (hyp + " still needs to exclude " + min
                                           + " neg exs");
                        Feature bestConstraint = null;
                        for (Enumeration e = allConstraints ();
                             e.hasMoreElements (); ) {
                            Feature f = (Feature) e.nextElement ();
                            int n = countIntersection (neg, f);
                            debug.println (f + " excludes " + (neg.getSize () - n)
                                           + " neg exs");
                            if (n < min) {
                                min = n;
                                bestConstraint = f;
                            }
                        }
            
                        if (bestConstraint == null)
                            return;
                        hyp = constrain (hyp, bestConstraint);
                        if (min == 0)
                            break;
                        else
                            neg = neg.and (hyp);
                    }

                    yield (hyp);
                }
            };
    }

    //////////////////////////
    ////////////////////////// Ranking hypotheses 
    //////////////////////////

    void rankPartitions (Vector partitions) {
        MergeSort.sort (new SortableVector (partitions) {
                public int compare (Object o1, Object o2) {
                    double s1 = getScore ((Feature) o1);
                    double s2 = getScore ((Feature) o2);
                    if (s1 < s2)
                        return 1;
                    else if (s1 > s2)
                        return -1;
                    else
                        return 0;
                }
            });
    }

    void generateLikelyRegularFeatures () {
        likelyRegularFeatures.removeAllElements ();
        scores.clear ();

        // accumulate all regular features into one vector
        Vector v = accumulate (null, regularFeatures ());

        // sort by estimated # of records
        MergeSort.sort (new SortableVector (v) {
                public int compare (Object o1, Object o2) {
                    double e1 = ((RegularFeature)o1).estimate;
                    double e2 = ((RegularFeature)o2).estimate;
                    if (e1 < e2)
                        return -1;
                    else if (e1 > e2)
                        return 1;
                    else
                        return 0;
                }
            });

        // find mode (most frequent estimate) and the features that
        // predicted it.
        double last = 0;
        int nLast = 0;
        int iLast = 0;
        Vector predictors = new Vector ();
        double mode = 0;
        int nMode = 0;
        int iMode = 0;
        int peaks = 0;
        for (int i = 0, n = v.size (); i < n; ++i) {
            RegularFeature ct = (RegularFeature) v.elementAt (i);
            if (ct.estimate == last) {
                ++nLast;
            } else {
                if (last > 1) {
                    if (nLast > nMode) {
                        mode = last;
                        nMode = nLast;
                        iMode = iLast;
                        peaks = 1;
                    } else if (nLast == nMode)
                        ++peaks;
                }
                last = ct.estimate;
                nLast = 1;
                iLast = i;
            }
        }
        if (nLast > nMode) {
            mode = last;
            nMode = nLast;
            iMode = iLast;
            peaks = 1;
        }

        // update likelyRegularFeatures vector
        if (peaks > 1) {
            debug.println ("multiple peaks, no likely regular features chosen");
        } else {
            for (int i = iMode, n = iMode + nMode; i < n; ++i) {
                RegularFeature ct = (RegularFeature) v.elementAt (i);
                likelyRegularFeatures.addElement (ct);
            }
        }
    }

    public double getScore (Feature p) {
        if (p.getSize () == 1)
            return 0.0; // p matches only one thing -- not a great partition 

        if (likelyRegularFeatures.isEmpty ()
            || p.isEmpty ())
            return 1 / (1 + getComplexity (p));

        // check memoizing
        Double d = (Double) scores.get (p);
        if (d != null)
            return d.doubleValue ();

        int nP = p.getSize ();

        double score = 0;
        for (Enumeration e = likelyRegularFeatures.elements (); 
             e.hasMoreElements (); ) {
            RegularFeature ct = (RegularFeature) e.nextElement ();
            double nF = ct.getSize ();
            double nPF = ct.countIntersection (p);
            score += (nPF * nPF) / (nP * nF);
        }

        double avgScore = score / likelyRegularFeatures.size ();

        scores.put (p, new Double (avgScore));
        return avgScore;
    }


    //////////////////////////
    ////////////////////////// Combining equivalent hypotheses 
    //////////////////////////

    void uniq (Vector v, boolean replace) {
        for (int i = 0, n = v.size (); i < n; ++i) {
            Feature f = (Feature) v.elementAt (i);
            
            Feature simplest = f;
            double minComplexity = getComplexity (simplest);

            if (replace)
                f.clearEquivalentFeatures ();

            for (int j = i+1; j < n; ++j) {
                Feature g = (Feature) v.elementAt (j);
                if (((RegionSet)f).equals ((RegionSet)g)) {
                    double complexity = getComplexity (g);
                    if (complexity < minComplexity) {
                        simplest = g;
                        minComplexity = complexity;
                    }

                    f.addEquivalentFeature (g);
                    v.removeElementAt (j);
                    --n;
                    --j;
                }
            }

            if (simplest != f) {
                debug.println (simplest + " should be preferred over " + f);
                
                for (Enumeration e = f.equivalentFeatures ();
                     e.hasMoreElements (); ) {
                    Feature g = (Feature) e.nextElement ();
                    if (g != simplest)
                        simplest.addEquivalentFeature (g);
                }
                simplest.addEquivalentFeature (f);
                f.clearEquivalentFeatures ();

                v.setElementAt (simplest, i);
                f = simplest;
            }
        }       
    }

    double getComplexity (Feature f) {
        ComplexityTraversal t = new ComplexityTraversal ();
        f.getPattern ().getRoot ().accept (t);
        return t.complexity;
    }

    static class ComplexityTraversal extends Traversal {
        double complexity = 0;
        public Object forAnd (TC.And n) {
            ++complexity;
            return super.forAnd (n);
        }
        public Object forOr (TC.Or n) {
            ++complexity;
            return super.forOr (n);
        }
        public Object forButNot (TC.ButNot n) {
            ++complexity;
            return super.forButNot (n);
        }
        public Object forOrButNot (TC.OrButNot n){
            ++complexity;
            return super.forOrButNot (n);
        }
        public Object forEquals (TC.Equals n) {
            ++complexity;
            return super.forEquals (n);
        }
        public Object forBefore (TC.Before n) {
            ++complexity;
            return super.forBefore (n);
        }
        public Object forAfter (TC.After n) {
            ++complexity;
            return super.forAfter (n);
        }
        public Object forJustBefore (TC.JustBefore n) {
            ++complexity;
            return super.forJustBefore (n);
        }
        public Object forJustAfter (TC.JustAfter n) {
            ++complexity;
            return super.forJustAfter (n);
        }
        public Object forBeforeDelimited (TC.BeforeDelimited n) {
            ++complexity;
            return super.forBeforeDelimited (n);
        }
        public Object forAfterDelimited (TC.AfterDelimited n) {
            ++complexity;
            return super.forAfterDelimited (n);
        }
        public Object forJustBeforeDelimited (TC.JustBeforeDelimited n) {
            ++complexity;
            return super.forJustBeforeDelimited (n);
        }
        public Object forJustAfterDelimited (TC.JustAfterDelimited n) {
            ++complexity;
            return super.forJustAfterDelimited (n);
        }
        public Object forIn (TC.In n) {
            ++complexity;
            return super.forIn (n);
        }
        public Object forContains (TC.Contains n) {
            ++complexity;
            return super.forContains (n);
        }
        public Object forStarting (TC.Starting n) {
            ++complexity;
            return super.forStarting (n);
        }
        public Object forEnding (TC.Ending n) {
            ++complexity;
            return super.forEnding (n);
        }
        public Object forStartingDelimited (TC.StartingDelimited n) {
            ++complexity;
            return super.forStartingDelimited (n);
        }
        public Object forEndingDelimited (TC.EndingDelimited n) {
            ++complexity;
            return super.forEndingDelimited (n);
        }
        public Object forOverlapsStart (TC.OverlapsStart n) {
            ++complexity;
            return super.forOverlapsStart (n);
        }
        public Object forOverlapsEnd (TC.OverlapsEnd n) {
            ++complexity;
            return super.forOverlapsEnd (n);
        }
        public Object forOverlaps (TC.Overlaps n) {
            ++complexity;
            return super.forOverlaps (n);
        }
        public Object forSeparatedBy (TC.SeparatedBy n) {
            ++complexity;
            return super.forSeparatedBy (n);
        }
        public Object forStartOf (TC.StartOf n) {
            ++complexity;
            return super.forStartOf (n);
        }
        public Object forEndOf (TC.EndOf n) {
            ++complexity;
            return super.forEndOf (n);
        }
        public Object forBalancedUnary (TC.BalancedUnary n) {
            ++complexity;
            return super.forBalancedUnary (n);
        }
        public Object forBalancedBinary (TC.BalancedBinary n){
            ++complexity;
            return super.forBalancedBinary (n);
        }
        public Object forFromTo (TC.FromTo n){
            ++complexity;
            return super.forFromTo (n);
        }
        public Object forConstant (TC.Constant n) {
            ++complexity;
            return null;
        }
        public Object forExternalPattern (TC.ExternalPattern n) {
            ++complexity;
            return null;
        }
        public Object forLiteral (TC.Literal n) {
            complexity += 2 - 1.0 / (n.getString ().length ()+1);
            return null;
        }
        public Object forRegexp (TC.Regexp n) {
            complexity += 1 - 1.0 / (n.getString ().length ()+1);
            return null;
        }
        public Object forId (TC.Id n) {
            ++complexity;
            return null;
        }
        public Object forAny (TC.Any n) {
            ++complexity;
            return null;
        }
        public Object forNothing (TC.Nothing n) {
            ++complexity;
            return null;
        }
        public Object forNextTo (TC.NextTo n) {
            ++complexity;
            return super.forNextTo (n);
        }
        public Object forTrim (TC.Trim n) {
            ++complexity;
            return super.forTrim (n);
        }
        public Object forInclude (TC.Include n) {
            ++complexity;
            return super.forInclude (n);
        }
        public Object forAssign (TC.Assign n) {
            ++complexity;
            return super.forAssign (n);
        }
        public Object forSeq (TC.Seq n) {
            ++complexity;
            return super.forSeq (n);
        }
        public Object forIgnoreBinary (TC.IgnoreBinary n) {
            ++complexity;
            return super.forIgnoreBinary (n);
        }
        public Object forNth (TC.Nth n){
            ++complexity;
            return super.forNth (n);
        }
        public Object forFlatten (TC.Flatten n) {
            ++complexity;
            return super.forFlatten (n);
        }
        public Object forMelt (TC.Melt n) {
            ++complexity;
            return super.forMelt (n);
        }
        public Object forCrispen (TC.Crispen n) {
            ++complexity;
            return super.forCrispen (n);
        }
        public Object forNonempty (TC.Nonempty n) {
            ++complexity;
            return super.forNonempty (n);
        }
        public Object forPrefix (TC.Prefix n) {
            ++complexity;
            return super.forPrefix (n);
        }
    }

    //////////////////////////
    ////////////////////////// Utility routines
    //////////////////////////

    Enumeration dumpFeatures (final String prefix, Enumeration e) {
        return accumulate 
            (null, 
             new FilteredEnumeration (e) {
                     public void transform (Object o) {
                         String s =
                             (o instanceof RegularFeature)
                             ? ((RegularFeature)o).toDumpString ()
                             : o.toString ();
                         debug.println (prefix + s);
                         yield (o);
                     }
                 }).elements ();
    }
            
    Enumeration consistentWithExamples (Enumeration e,
                                        final RegionSet pos,
                                        final RegionSet neg) {
        return new FilteredEnumeration (e) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    if (allMatch (pos, f) && noneMatch (neg, f))
                        yield (f);
                }
            };
    }

    boolean allMatch (RegionSet ex, Feature f) {
        //RegionSet constraint = f.getRegions ();
        RegionEnumeration re = ex.regions ();
        for (Region r = re.firstFast (); r != null; r = re.nextFast ()) {
            if (!f.intersects (r)) {
                return false;
            }
        }
        return true;
    }

    boolean noneMatch (RegionSet ex, Feature f) {
        //RegionSet constraint = f.getRegions ();
        RegionEnumeration re = ex.regions ();
        for (Region r = re.firstFast (); r != null; r = re.nextFast ()) {
            if (f.intersects (r)) {
                return false;
            }
        }
        return true;
    }

    Vector accumulate (Vector v, Enumeration enum) {
        if (v == null)
            v = new Vector ();
        while (enum.hasMoreElements ())
            v.addElement (enum.nextElement ());
        return v;
    }

    int countIntersection (RegionSet partition, RegionSet constraint) {
        int matches = 0;
        RegionEnumeration re = partition.regions ();
        for (Region r = re.first (); r != null; r = re.next ())
            if (constraint.intersects (r))
                ++matches;
        return matches;
    }


    Feature makeFeature (TC pat) {
        //return new Feature (doc, pat.match (doc), pat);
		return new Feature (pat.match (doc), pat);
    }

    Feature makeFeature (Feature base, TC pat) {
        return new Feature (base, pat.match (doc), pat);
    }

    Feature makeCaseSensitiveLiteral (String s) {
        TC tcInsensitive = new TC (new TC.Literal (s),
                                   TC.makeLiteral (s));
        TC tcSensitive = new TC (new TC.CaseSensitivity (true,
                                                         tcInsensitive.getRoot ()),
                                 "case-sensitive " + tcInsensitive.getExpression ());
        
        RegionSet rInsensitive = tcInsensitive.match (doc);
        RegionSet rSensitive = tcSensitive.match (doc);

		return
            (rInsensitive.equals (rSensitive))
            ? new Feature (rInsensitive, tcInsensitive) //new Feature (doc, rInsensitive, tcInsensitive)
            : new Feature (rSensitive, tcSensitive); //new Feature (doc, rSensitive, tcSensitive);
    }

    Feature constrain (Feature hyp, Feature constraint) {
        Feature chyp = makeFeature 
            (new TC (new TC.And (hyp.getPattern ().getRoot (),
                                 constraint.getPattern ().getRoot ()),
                     hyp + "\n  " + constraint));

        Enumeration e = hyp.equivalentFeatures ();
        while (e.hasMoreElements ()) {
            Feature equiv = (Feature) e.nextElement ();
            Feature cequiv = new Feature (equiv, chyp,
                                          new TC (new TC.And (equiv.getPattern ().getRoot (),
                                                              constraint.getPattern ().getRoot ()),
                                                  equiv + "\n  " + constraint));
            chyp.addEquivalentFeature (cequiv);
        }

        return chyp;
    }



    Enumeration starts (Enumeration features) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    yield 
                        (makeFeature 
                         (f, 
                          new TC (new TC.Starting (f.getPattern ()
                                                   .getRoot ()), 
                                  "starting " + f)));
                }
            };
    }

    Enumeration ends (Enumeration features) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    yield 
                        (makeFeature 
                         (f, 
                          new TC (new TC.Ending (f.getPattern ()
                                                   .getRoot ()), 
                                  "ending " + f)));
                }
            };
    }

    Enumeration before (Enumeration features) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    yield 
                        (makeFeature 
                         (f, 
                          new TC (new TC.JustBefore (f.getPattern ()
                                                   .getRoot ()), 
                                  "just before " + f)));
                }
            };
    }

    Enumeration after (Enumeration features) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    yield 
                        (makeFeature 
                         (f, 
                          new TC (new TC.JustAfter (f.getPattern ()
                                                   .getRoot ()), 
                                  "just after " + f)));
                }
            };
    }

    Enumeration in (Enumeration features) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    yield 
                        (makeFeature 
                         (f, 
                          new TC (new TC.In (f.getPattern ()
                                                   .getRoot ()), 
                                  "in " + f)));
                }
            };
    }

    Enumeration contains (Enumeration features) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    yield 
                        (makeFeature 
                         (f, 
                          new TC (new TC.Contains (f.getPattern ()
                                                   .getRoot ()), 
                                  "contains " + f)));
                }
            };
    }


    Enumeration startOf (Enumeration features) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    TC.Node tc = 
                        new TC.StartOf (f.getPattern ().getRoot ());
                    String expr = "start of " + f;
                    yield (makeFeature (new TC (tc, expr)));
                }
            };
    }

    Enumeration endOf (Enumeration features) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    TC.Node tc = 
                        new TC.EndOf (f.getPattern ().getRoot ());
                    String expr = "end of " + f;
                    yield (makeFeature (new TC (tc, expr)));
                }
            };
    }


    Enumeration filterFeatures (Enumeration features,
                                final RegionSet transformedPosExamples,
                                final Vector failures) {
        return new FilteredEnumeration (features) {
                public void transform (Object o) {
                    Feature f = (Feature) o;
                    if (allMatch (transformedPosExamples, f))
                        yield (f);
                    else if (failures != null)
                        failures.addElement (f);
                }
            };
    }

    void filterFeatureVector (Vector features,
                              RegionSet transformedPosExamples,
                              Vector failures) {
        for (int i = 0, n = features.size (); i < n; ++i) {
            Feature f = (Feature) features.elementAt (i);
            if (!allMatch (transformedPosExamples, f)) {
                features.removeElementAt (i);
                --i;
                --n;
                if (failures != null)
                    failures.addElement (f);
            }
        }
    }


}
