/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */
package lapis.ml;
import lapis.*;
import java.util.*;

import lapis.maps.LexMap;
import lapisx.sort.*;
/**
 * @author alisa_m
 * Lexical Similarity, much like Similarity but just used for Lexical Attraction.
 */
public class LexicalSimilarity extends Similarity {
	LexMap lm;
	public LexicalSimilarity(NamedRegionSet matches, NamedRegionSet misses, LexMap lm) {
		super(matches, misses);
		this.lm = lm;
	}
	//
	// Similarity algorithm
	//
	/**
	 * Compute and return the clusters of regions.
	 * @return clusters, ordered by distance from the typical match
	 * (closest cluster first).  Both positive-example clusters and
	 * negative-example clusters are included in the array.
	 */
	public Cluster[] getClusters() {
		if (clusters != null)
			return clusters;
        
        Document doc = (Document)matches.getDocuments().next();
		if (features == null) {
			features = generateFeatures(matches, misses, lm);
			featuresByWeight = null;
			pos = null;
			neg = null;
		}
		Main.getProgress().fireStartProgress(
			"Testing features",
			misses.isEmpty() ? features.length : features.length * 2);
		if (pos == null || neg == null) {
			long start = System.currentTimeMillis();
			try {
				pos = new DataSet(matches, features);
				neg = new DataSet(misses, features);
				clusters = null;
			}
			finally {
				debug.println("feature testing: " + + (System.currentTimeMillis() - start) + " ms");
				Main.getProgress().fireDoneProgress();
			}
		}
		if (clusters == null) {
			long start = System.currentTimeMillis();
			try {
				// compute weights for each feature
				metric =
					(!misses.isEmpty())
						? (DistanceMetric) new MutualInformationMetric(pos, neg)
						: (DistanceMetric) new InverseVarianceMetric(pos);
			}
			finally {
				debug.println("feature weighting: " + + (System.currentTimeMillis() - start) + " ms");
			}
			start = System.currentTimeMillis();
			try {
				double[] typicalPosExample = pos.median;
				// find distance of each example from typicalPositiveExample
				Vector points = new Vector();
				RegionEnumeration re = pos.regions();
				for (Region r = re.first(); r != null; r = re.next()) {
					double[] fv = pos.getFeatureValues(r);
					double d = metric.getDistance(fv, typicalPosExample);
					debug.println("pos " + r + " at distance " + d);
					//points.addElement(new Cluster(doc, r, fv, d, true));
					points.addElement(new Cluster(r, fv, d, true));
				}
				re = neg.regions();
				for (Region r = re.first(); r != null; r = re.next()) {
					double[] fv = neg.getFeatureValues(r);
					double d = metric.getDistance(fv, typicalPosExample);
					debug.println("neg " + r + " at distance " + d);
					//points.addElement(new Cluster(doc, r, fv, d, false));
					points.addElement(new Cluster(r, fv, d, false));
				}
				// sort by distances
				Quicksort.sort(new IncreasingDistance(points));
				// merge points with identical vectors together
				for (int i = 0, n = points.size(); i < n; ++i) {
					Cluster ptI = (Cluster) points.elementAt(i);
					for (int j = i + 1; j < n; ++j) {
						Cluster ptJ = (Cluster) points.elementAt(j);
						if (ptI.dist != ptJ.dist)							// j and all later examples are farther than i,
							// so they can't possibly have identical vectors
							break;
						if (ptI.isPosExample == ptJ.isPosExample && equalVectors(ptI.fv, ptJ.fv)) {
							// merge j into i
							ptI = new Cluster(ptI, ptJ);
							points.setElementAt(ptI, i);
							points.removeElementAt(j);
							--j;
							--n;
						}
					}
				}
				clusters = new Cluster[points.size()];
				points.copyInto(clusters);
			}
			finally {
				debug.println("example sorting: " + + (System.currentTimeMillis() - start) + " ms");
			}
		}
		return clusters;
	}
	static Feature[] generateFeatures(NamedRegionSet positive, NamedRegionSet negative, LexMap lm) {
        Document doc = (Document)positive.getDocuments().next();
		long start = System.currentTimeMillis();
		Main.getProgress().fireStartProgress("Generating features", PatternLibrary.getDocumentLibrary(doc).getSize());
		try {
			Vector features = new Vector();
			for (Enumeration e = lm.getFeatures(positive); e.hasMoreElements();) {
				features.addElement(e.nextElement());
			}
			// convert Vector into an array
			Feature[] f = new Feature[features.size()];
			features.copyInto(f);
			debug.println(f.length + " features generated");
			for (int i = 0; i < f.length; ++i)
				debug.println(f[i].toString());
			return f;
		}
		finally {
			debug.println("feature generation: " + + (System.currentTimeMillis() - start) + " ms");
			Main.getProgress().fireDoneProgress();
		}
	}
}
