/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.ml;

import lapis.*;
import lapis.tc.*;
import java.util.*;
import lapisx.util.*;
import lapisx.sort.*;

public class LiteralGenerator implements Enumeration {
	public static Debug debug = Debug.QUIET;

	Document doc;
	NamedRegionSet positive;
	NamedRegionSet negative;
	double minMatches;
	int maxLiteral;
	int ops;
	Vector features;
	Enumeration output;

	public static final int STARTING = 0x01;
	public static final int ENDING = 0x02;
	public static final int JUST_BEFORE = 0x04;
	public static final int JUST_AFTER = 0x08;
	public static final int EQUALS = 0x10;
	public static final int OMIT_OPERATOR = 0x100;

	public LiteralGenerator(
		NamedRegionSet positive,
		NamedRegionSet negative,
		double minMatches,
		int maxLiteral,
		int ops) {
		// TODO: support multiple documents
		debug.assertion(positive.getVersions().getSize() == 1);
		Iterator idoc = positive.getDocuments();
		while (idoc.hasNext())
			doc = (Document) idoc.next();

		this.positive = positive;
		this.negative = negative;
		this.minMatches = minMatches;
		this.maxLiteral = maxLiteral;
		this.ops = ops;
		features = new Vector();
	}

	public boolean hasMoreElements() {
		if (output == null)
			output = generate();
		return output.hasMoreElements();
	}

	public Object nextElement() {
		if (output == null)
			output = generate();
		return output.nextElement();
	}

	//
	// generating literal features
	//
	Enumeration generate() {
		// copy positive and negative examples into an array
		int nPos = positive.getSize();
		int nNeg = negative.getSize();
		LiteralRegion[] examples = new LiteralRegion[nPos + nNeg];
		int n = 0;
		RegionEnumeration re = positive.regions();
		for (Region r = re.first(); r != null; r = re.next())
			examples[n++] = new LiteralRegion(r, true);
		re = negative.regions();
		for (Region r = re.first(); r != null; r = re.next())
			examples[n++] = new LiteralRegion(r, false);

		Main.getProgress().fireStartProgress(
			"Generating literal features",
			n * countBits(ops));

		int minPosHits = Math.max((int) (minMatches * nPos), 1);
		int minNegHits = Math.max((int) (minMatches * nNeg), 1);
		//         debug.println ("minPosHits=" + minPosHits + ", minNegHits=" + minNegHits);

		if ((ops & STARTING) != 0) {
			// starting "literal": extract maxLiteral-character prefix 
			// of each example, then compute longest common prefixes.
			for (int i = 0; i < n; ++i)
				examples[i].extract(
					doc,
					examples[i].r.getStart(),
					examples[i].r.getEnd());
			for (Enumeration e =
				getLongestCommonPrefixes(examples, minPosHits, minNegHits);
				e.hasMoreElements();
				) {
				String literal = (String) e.nextElement();
				Main.getProgress().fireAddProgress(1);

				// FIX: handle embedded quotes
				if ((ops & OMIT_OPERATOR) != 0)
					yield(new TC.Literal(literal), "\"" + literal + "\"");
				else
					yield(
						new TC.Starting(new TC.Literal(literal)),
						"starts \"" + literal + "\"");
			}
		}

		if ((ops & ENDING) != 0) {
			// ending "literal": extract maxLiteral-character suffix of
			// each example, then reverse it and compute longest common prefixes.
			for (int i = 0; i < n; ++i)
				examples[i].extract(
					doc,
					examples[i].r.getEnd(),
					examples[i].r.getStart());
			for (Enumeration e =
				getLongestCommonPrefixes(examples, minPosHits, minNegHits);
				e.hasMoreElements();
				) {
				String literal = Str.reverse((String) e.nextElement());
				Main.getProgress().fireAddProgress(1);
				if ((ops & OMIT_OPERATOR) != 0)
					yield(new TC.Literal(literal), "\"" + literal + "\"");
				else
					yield(
						new TC.Ending(new TC.Literal(literal)),
						"ends \"" + literal + "\"");
			}
		}

		if ((ops & JUST_BEFORE) != 0) {
			// just before "literal": extract maxLiteral-character prefix 
			// of content after each example, then compute longest common prefixes.
			for (int i = 0; i < n; ++i)
				examples[i].extract(
					doc,
					examples[i].r.getEnd(),
					RegionSet.MAX_INDEX);
			for (Enumeration e =
				getLongestCommonPrefixes(examples, minPosHits, minNegHits);
				e.hasMoreElements();
				) {
				String literal = (String) e.nextElement();
				Main.getProgress().fireAddProgress(1);
				if ((ops & OMIT_OPERATOR) != 0)
					yield(new TC.Literal(literal), "\"" + literal + "\"");
				else
					yield(
						new TC.JustBefore(new TC.Literal(literal)),
						"just before \"" + literal + "\"");
			}
		}

		if ((ops & JUST_AFTER) != 0) {
			// just after "literal": extract maxLiteral-character suffix of
			// of content before each example, then reverse it and compute
			// longest common prefixes.
			for (int i = 0; i < n; ++i)
				examples[i].extract(
					doc,
					examples[i].r.getStart(),
					RegionSet.MIN_INDEX);
			for (Enumeration e =
				getLongestCommonPrefixes(examples, minPosHits, minNegHits);
				e.hasMoreElements();
				) {
				String literal = Str.reverse((String) e.nextElement());
				Main.getProgress().fireAddProgress(1);
				if ((ops & OMIT_OPERATOR) != 0)
					yield(new TC.Literal(literal), "\"" + literal + "\"");
				else
					yield(
						new TC.JustAfter(new TC.Literal(literal)),
						"just after \"" + literal + "\"");
			}
		}

		if ((ops & EQUALS) != 0) {
			// equals "literal": extract each example in its entirety 
			// (but not more than maxLiteral characters).
			for (int i = 0; i < n; ++i)
				examples[i].extract(
					doc,
					examples[i].r.getStart(),
					examples[i].r.getEnd());
			for (Enumeration e =
				getEqualLiterals(examples, minPosHits, minNegHits);
				e.hasMoreElements();
				) {
				String literal = (String) e.nextElement();
				Main.getProgress().fireAddProgress(1);

				// FIX: handle embedded quotes
				yield(new TC.Literal(literal), "\"" + literal + "\"");
			}
		}

		Main.getProgress().fireDoneProgress();

		return features.elements();
	}

	int countBits(int x) {
		int n = 0;
		while (x != 0) {
			if ((x & 1) != 0)
				++n;
			x >>>= 1;
		}
		return n;
	}

	protected void yield(TC.Node tc, String expr) {
		// Find matches to the pattern
		TC pattern = new TC(tc, expr);
		RegionSet matches = pattern.match(doc);

		// If no matches to pattern, skip it
		if (matches.isEmpty())
			return;

		// create a feature
		//features.addElement (new Feature (doc, matches, pattern));
		features.addElement(new Feature(matches, pattern));
	}

	protected String canonicalize(String s) {
		s = s.trim();

		// don't allow embedded newlines; strip off the newline
		// and everything after it
		int nl = s.indexOf("\n");
		if (nl >= 0)
			s = s.substring(0, nl).trim();

		s = s.toLowerCase();

		return s;
	}

	class LiteralRegion {
		public Region r;
		public String s;
		public boolean positive;
		public LiteralRegion(Region r, boolean positive) {
			this.r = r;
			this.positive = positive;
		}

		public void extract(Document doc, int p0, int p1) {
			// check whether we want a forward or reversed string
			if (p0 <= p1) {
				// forward string
				p1 = Math.min(p1, p0 + maxLiteral);
				s = doc.getText(new Region(p0, p1, doc));
			} else {
				// reversed string
				p1 = Math.max(p1, p0 - maxLiteral);
				s = Str.reverse(doc.getText(new Region(p1, p0, doc)));
			}

			// canonicalize the string
			s = canonicalize(s);
		}
	}

	static Enumeration getLongestCommonPrefixes(
		LiteralRegion[] examples,
		int minPosHits,
		int minNegHits) {
		// sort examples so that longest common prefixes are adjacent
		Quicksort.sort(new SortableArray(examples) {
			public int compare(Object o1, Object o2) {
				return ((LiteralRegion) o1).s.compareTo(((LiteralRegion) o2).s);
			}
		});

		// find the longest common prefix of each adjacent pair in the
		// sorted examples, and store them in a hashtable to omit duplicates.
		Hashtable lcp = new Hashtable();
		int n = examples.length;
		for (int i = 1; i < n; ++i) {
			Main.getProgress().fireAddProgress(1);
			String prefix =
				Str.longestCommonPrefix(examples[i - 1].s, examples[i].s);
			if (prefix.length() != 0 && !lcp.containsKey(prefix))
				lcp.put(prefix, new Integer(i));
		}

		// Discard useless literals.
		// A literal must match at least minPosHits positive examples
		// or minNegHits negative examples to be considered useful.
		for (Enumeration e = lcp.keys(); e.hasMoreElements();) {
			String prefix = (String) e.nextElement();
			int i = ((Integer) lcp.get(prefix)).intValue();

			// count how many positive or negative examples this prefix
			// matches.  Since examples is sorted, all of them have to
			// be contiguous, so it suffices to just scan forward and backward
			// from examples[i].
			int nPosHits = 0;
			int nNegHits = 0;
			// scan backward first:
			for (int j = i; j >= 0; --j) {
				if (examples[j].s.startsWith(prefix)) {
					if (examples[j].positive)
						++nPosHits;
					else
						++nNegHits;
				} else
					break;
			}

			// now scan forward
			for (int j = i + 1; j < n; ++j) {
				if (examples[j].s.startsWith(prefix)) {
					if (examples[j].positive)
						++nPosHits;
					else
						++nNegHits;
				} else
					break;
			}

			// test against threshold
			//             debug.println (prefix + " has " + nPosHits + " poshits, "
			//                            + nNegHits + " neghits");
			if (nPosHits < minPosHits && nNegHits < minNegHits)
				lcp.remove(prefix);
		}

		return lcp.keys();
	}

	static Enumeration getEqualLiterals(
		LiteralRegion[] examples,
		int minPosHits,
		int minNegHits) {
		// sort examples
		Quicksort.sort(new SortableArray(examples) {
			public int compare(Object o1, Object o2) {
				return ((LiteralRegion) o1).s.compareTo(((LiteralRegion) o2).s);
			}
		});

		// find the longest common prefix of each adjacent pair in the
		// sorted examples, and store them in a hashtable to omit duplicates.
		Vector result = new Vector();
		int n = examples.length;
		String curr = null;
		int nPosHits = 0;
		int nNegHits = 0;
		for (int i = 0; i < n; ++i) {
			Main.getProgress().fireAddProgress(1);
			if (curr == null || !curr.equals(examples[i].s)) {
				if (curr != null
					&& (nPosHits >= minPosHits || nNegHits >= minNegHits)) {
					debug.println(
						curr
							+ " has "
							+ nPosHits
							+ " pos hits, "
							+ nNegHits
							+ " neg hits");
					result.addElement(curr);
				}
				curr = examples[i].s;
				nPosHits = nNegHits = 0;
			}
			if (examples[i].positive)
				++nPosHits;
			else
				++nNegHits;
		}
		if (curr != null
			&& (nPosHits >= minPosHits || nNegHits >= minNegHits)) {
			debug.println(
				curr
					+ " has "
					+ nPosHits
					+ " pos hits, "
					+ nNegHits
					+ " neg hits");
			result.addElement(curr);
		}

		return result.elements();
	}

}
