/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.ml;

import lapis.*;
import lapis.tc.*;
import java.util.*;
import java.io.*;

public class RecordSetGuesser {
	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

	/**
	 * NamedRegionSet to be split.  All hypotheses must be inside 
	 * this selection.
	 */
	NamedRegionSet range;

	/**
	 * Record sets must cover at least this fraction of the characters in the
	 * selection.
	 */
	private static final double THRESHOLD = 0.33;

	/**
	 * Make a RecordSetGuesser object.
	 * @param range NamedRegionSet to split.  All guessed record sets are inside this
	 * selection.
	 * @requires range is flat
	 */
	public RecordSetGuesser(NamedRegionSet range) {
		this.range = range;
		debug.assertion(range.isFlat());
	}

	/**
	 * Get the hypothesized record sets.
	 * @return enumeration of SimulEdit objects, sorted by goodness score
	 * so that the most preferred record set is returned first.
	 */
	public List getGuesses() {
		//Document doc = range.getDocument (); // rcm: dead code

		Vector features = generateFeatures();

		RegionSet inRange = range.in();
		int lenRange = length(range);
		List hypotheses = new ArrayList();
		Map alreadyGenerated = new HashMap();

		for (Iterator g = features.iterator(); g.hasNext();) {
			Feature f = (Feature) g.next();

			// find matches to f that are inside the selection 
			RegionSet fInRange = f.and(inRange).flatten();

			// filter out if fewer than 2 such matches 
			if (fInRange.getSize() < 2) {
				debug.println(
					"rejected "
						+ f
						+ ": only "
						+ fInRange.getSize()
						+ " matches");
				continue;
			}

			// filter out if another hypothesis had the same region set
			if (alreadyGenerated.containsKey(fInRange)) {
				debug.println("rejected " + f + ": redundant");
				continue;
			}
			alreadyGenerated.put(fInRange, fInRange);

			// filter out if matches to f cover less than a
			// fraction THRESHOLD of the characters in selection
			//             int l = length (fInRange);
			//             if (l < THRESHOLD * lenRange) {
			//                 debug.println ("rejected " + f + ": covered only " 
			//                                + l + " chars out of " + lenRange);
			//                 continue;
			//             }

			// make a selection corresponding to fInRange
			NamedRegionSet hyp = new NamedRegionSet(fInRange, f.getPattern());

			// make a SimulEdit from it
			SimulEdit se = new SimulEdit(hyp);

			// filter out if score <= 1
			double score = getScore(se);
			if (score <= 1)
				continue;

			hypotheses.add(se);
		}

		// sort by score
		Collections.sort(hypotheses, new HypothesisSorter());

		// dump for debugging
		if (debug.printEnabled())
			for (Iterator g = hypotheses.iterator(); g.hasNext();) {
				SimulEdit se = (SimulEdit) g.next();
				debug.println(se.getRecords() + ": " + getScore(se));
			}

		return hypotheses;
	}

	int length(RegionSet s) {
		int l = 0;
		RegionEnumeration re = s.regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast())
			l += r.getLength();
		return l;
	}

	/**
	 * Get the score for a record set.
	 * @param se SimulEdit object
	 * @return "goodness" score computed for se
	 */
	public double getScore(SimulEdit se) {
		return se.getRegularFeatureCount();
	}

	/**
	 * Feature generation.
	 */
	Vector generateFeatures() {
		Vector result = new Vector();
		int size = 0;
		Iterator idoc = range.getDocuments();
		while (idoc.hasNext())
			size
				+= PatternLibrary
					.getDocumentLibrary((Document) idoc.next())
					.getSize();

		Main.getProgress().fireStartProgress("Generating features", size);
		try {
			idoc = range.getDocuments();
			while (idoc.hasNext())
				result.addAll(
					accumulate(
						null,
						new LibraryGenerator((Document) idoc.next())));
			return result;
		} finally {
			Main.getProgress().fireDoneProgress();
		}
	}

	Vector accumulate(Vector v, Enumeration enum) {
		if (v == null)
			v = new Vector();
		while (enum.hasMoreElements())
			v.addElement(enum.nextElement());
		return v;
	}

	/**
	 * Hypothesis scoring and sorting.
	 */
	class HypothesisSorter implements Comparator {
		public int compare(Object o1, Object o2) {
			SimulEdit s1 = (SimulEdit) o1;
			SimulEdit s2 = (SimulEdit) o2;
			double score1 = getScore(s1);
			double score2 = getScore(s2);
			if (score1 < score2)
				return 1;
			else if (score1 > score2)
				return -1;
			else
				return 0;
		}
	}

	public static void main(String[] args) throws Exception {
		debug = lapisx.util.Debug.VERBOSE;
		SimulEdit.debug = lapisx.util.Debug.VERBOSE;
		Main.init();

		String fn = args.length > 0 ? args[0] : "-";
		String pat = args.length > 1 ? args[1] : "all";

		Document doc =
			!fn.equals("-")
				? BasicDocument.defaultFactory.make(new File(args[0]))
				: BasicDocument.defaultFactory.make(System.in);
		NamedRegionSet sel = new NamedRegionSet(new TC(pat).match(doc));

		RecordSetGuesser s = new RecordSetGuesser(sel);
		s.getGuesses();
	}
}
