/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.ml;

import lapis.*;
import lapis.tc.TC;
import java.util.*;
import lapisx.sort.*;
import lapisx.util.*;
import java.text.NumberFormat;

/**
 * Implementation of the outlier-finding algorithm, which ranks a by
 * similarity to the typical region in the set.
 */
public class Similarity {
	public static Debug debug = Debug.QUIET;

	// maximum length of generated literals, in characters
	static final int MAX_LITERAL = 40;

	// generated literals must match at least this fraction of positive 
	// examples or negative examples
	static final double MIN_MATCHES = 0.5;

	NamedRegionSet matches;
	NamedRegionSet misses;

	Feature[] features;
	DataSet pos;
	DataSet neg;
	DistanceMetric metric;
	Integer[] featuresByWeight;
	// indexes of features in decreasing order of weight

	Cluster[] clusters;

	/**
	 * Make a Similarity object.  No computation is done in the constructor.
	 * @param matches Positive examples ("matches" to the pattern)
	 * @param misses Negative examples (nonmatches).  Often an empty set.
	 */
	public Similarity(NamedRegionSet matches, NamedRegionSet misses) {
		this.matches = matches;
		this.misses = misses;
	}

	//
	// Similarity algorithm
	//

	/**
	 * Compute and return the clusters of regions.
	 * @return clusters, ordered by distance from the typical match
	 * (closest cluster first).  Both positive-example clusters and
	 * negative-example clusters are included in the array.
	 */
	public Cluster[] getClusters() {
		if (clusters != null)
			return clusters;

		// rcm: dead code
		//Document doc = matches.getDocument ();

		if (features == null) {
			features = generateFeatures(matches, misses);
			featuresByWeight = null;
			pos = null;
			neg = null;
		}

		Main.getProgress().fireStartProgress(
			"Testing features",
			misses.isEmpty() ? features.length : features.length * 2);
		if (pos == null || neg == null) {
			long start = System.currentTimeMillis();
			try {
				pos = new DataSet(matches, features);
				neg = new DataSet(misses, features);
				clusters = null;
			} finally {
				debug.println(
					"feature testing: "
						+ 
						+ (System.currentTimeMillis() - start)
						+ " ms");
				Main.getProgress().fireDoneProgress();
			}
		}

		if (clusters == null) {
			long start = System.currentTimeMillis();
			try {
				// compute weights for each feature
				metric =
					(!misses.isEmpty())
						? (DistanceMetric) new MutualInformationMetric(pos, neg)
						: (DistanceMetric) new InverseVarianceMetric(pos);
			} finally {
				debug.println(
					"feature weighting: "
						+ 
						+ (System.currentTimeMillis() - start)
						+ " ms");
			}

			start = System.currentTimeMillis();
			try {
				double[] typicalPosExample = pos.median;

				// find distance of each example from typicalPositiveExample
				Vector points = new Vector();
				RegionEnumeration re = pos.regions();
				for (Region r = re.first(); r != null; r = re.next()) {
					double[] fv = pos.getFeatureValues(r);
					double d = metric.getDistance(fv, typicalPosExample);
					debug.println("pos " + r + " at distance " + d);
					points.addElement(new Cluster(r, fv, d, true));
				}

				re = neg.regions();
				for (Region r = re.first(); r != null; r = re.next()) {
					double[] fv = neg.getFeatureValues(r);
					double d = metric.getDistance(fv, typicalPosExample);
					debug.println("neg " + r + " at distance " + d);
					points.addElement(new Cluster(r, fv, d, false));
				}

				// sort by distances
				Quicksort.sort(new IncreasingDistance(points));

				// merge points with identical vectors together
				for (int i = 0, n = points.size(); i < n; ++i) {
					Cluster ptI = (Cluster) points.elementAt(i);

					for (int j = i + 1; j < n; ++j) {
						Cluster ptJ = (Cluster) points.elementAt(j);
						if (ptI.dist != ptJ.dist)
							// j and all later examples are farther than i,
							// so they can't possibly have identical vectors
							break;

						if (ptI.isPosExample == ptJ.isPosExample
							&& equalVectors(ptI.fv, ptJ.fv)) {
							// merge j into i
							ptI = new Cluster(ptI, ptJ);
							points.setElementAt(ptI, i);
							points.removeElementAt(j);
							--j;
							--n;
						}
					}
				}

				clusters = new Cluster[points.size()];
				points.copyInto(clusters);
			} finally {
				debug.println(
					"example sorting: "
						+ 
						+ (System.currentTimeMillis() - start)
						+ " ms");
			}
		}
		return clusters;
	}

	/**
	 * Test two feature vectors for equality.
	 * @param x one feature vector
	 * @param y another feature vector
	 * @return true iff every feature value in x equals the
	 * corresponding feature value in y, ignoring features that have
	 * zero weight.
	 */
	boolean equalVectors(double[] x, double[] y) {
		if (x.length != y.length)
			return false;

		for (int i = 0, n = x.length; i < n; ++i)
			if (metric.getWeight(i) != 0 && x[i] != y[i])
				return false;

		return true;
	}

	//
	// generating features
	//

	static Class[] RELOPS = new Class[] {
		//         TC.Equals.class,
		TC.In.class,
			TC.Contains.class,
			TC.Starting.class,
			TC.Ending.class,
			TC.JustAfter.class,
			TC.JustBefore.class };

	/**
	 * Generate a list of features for discriminating a set of examples
	 */
	static Feature[] generateFeatures(
		NamedRegionSet positive,
		NamedRegionSet negative) {
		int size = 0;
		Iterator idoc = positive.getDocuments();
		while (idoc.hasNext())
			size
				+= PatternLibrary
					.getDocumentLibrary((Document) idoc.next())
					.getSize();

		long start = System.currentTimeMillis();
		Main.getProgress().fireStartProgress("Generating features", size);
		try {
			Vector features = new Vector();

			//it = docs.iterator();
			idoc = positive.getVersions().documents();
			while (idoc.hasNext()) {
				Document doc = (Document) idoc.next();
				// pattern features: every library pattern P
				for (Enumeration e = new LibraryGenerator(doc);
					e.hasMoreElements();
					)
					features.addElement(e.nextElement());
			}

			// relop features: op P for every pattern feature P
			for (Enumeration e =
				new RelopGenerator(
					RELOPS,
					((Vector) features.clone()).elements());
				e.hasMoreElements();
				)
				features.addElement(e.nextElement());

			// literal features: derived from examples
			for (Enumeration e =
				new LiteralGenerator(
					positive,
					negative,
					MIN_MATCHES,
					MAX_LITERAL,
					LiteralGenerator.STARTING
						| LiteralGenerator.ENDING
						| LiteralGenerator.JUST_BEFORE
						| LiteralGenerator.JUST_AFTER);
				e.hasMoreElements();
				)
				features.addElement(e.nextElement());

			// convert Vector into an array
			Feature[] f = new Feature[features.size()];
			features.copyInto(f);
			debug.println(f.length + " features generated");
			for (int i = 0; i < f.length; ++i)
				debug.println(f[i].toString());
			return f;
		} finally {
			debug.println(
				"feature generation: "
					+ 
					+ (System.currentTimeMillis() - start)
					+ " ms");
			Main.getProgress().fireDoneProgress();
		}
	}

	/**
	 * A cluster of regions with identical feature vectors.
	 */
	public class Cluster extends NamedRegionSet {
		double[] fv;
		double dist;
		boolean isPosExample;
		Explanation[] explanation;

		/**
		 * Make a Cluster from a region set.
		 */
		public Cluster(
			Region r,
			double[] fv,
			double dist,
			boolean isPosExample) {
			super(r);
			this.fv = fv;
			this.dist = dist;
			this.isPosExample = isPosExample;
		}

		/**
		 * Make a Cluster by merging two other Clusters.
		 */
		public Cluster(Cluster c1, Cluster c2) {
			super(mergeRegionSets(c1, c2));
			fv = c1.fv;
			dist = c1.dist;
			isPosExample = c1.isPosExample;
		}

		/**
		 * Test whether cluster consists of matches (positive examples)
		 * or misses (negative examples).
		 * @return true iff cluster consists of positive examples
		 */
		public boolean isPositive() {
			return isPosExample;
		}

		/**
		 * Get distance from typical positive example.
		 * @return distance from typical positive example, using weighted
		 * Euclidean metric
		 */
		public double getDistance() {
			return dist;
		}

		/**
		 * Get similarity to typical positive example.
		 * @return similarity to typical positive example, ranging from
		 * 1.0 (identical feature vectors) to 0.0 (different in all features).
		 */
		public double getSimilarity() {
			return 1 - (dist / features.length);
		}

		/**
		 * Get explanation for cluster.
		 * @return a short list of features that are most important
		 * for explaining the differences between this cluster and the
		 * typical positive example.
		 */
		public Explanation[] getExplanation() {
			if (explanation != null)
				return explanation;
			explanation = Similarity.this.getExplanation(this, isPosExample);
			return explanation;
		}
	}

	/**
	 * If u is mutable, inserts v into u and returns u.  If u isn't
	 * mutable, creates a new mutable region set and inserts both u
	 * and v into it.
	 */
	static RegionSet mergeRegionSets(RegionSet u, RegionSet v) {
		MutableRegionSet m =
			(u instanceof MutableRegionSet)
				? (MutableRegionSet) u
				: new MutableRegionSet(u);
		m.insert(v);
		return m;
	}

	/**
	 * Get an explanation for a region set.
	 * @return a short list of features.  If isPosExample is true,
	 * then the explanation consists of the most important differences
	 * between the given region set and the typical positive example.
	 * If isPosExample is false, the explanation consists of the most
	 * important similarities between the given region set and the typical
	 * positive example.
	 */
	// FIX: this is broken for multiple clusters.
	public Explanation[] getExplanation(
		RegionSet regions,
		boolean isPosExample) {
		int n = features.length;
		double[] group = (isPosExample ? pos : neg).computeMean(regions);

		// rank features by weight
		if (featuresByWeight == null) {
			Vector fbw = new Vector();
			for (int i = 0; i < n; ++i)
				if (metric.getWeight(i) != 0)
					fbw.addElement(new Integer(i));
			Quicksort.sort(new SortableVector(fbw) {
				public int compare(Object o1, Object o2) {
					int i1 = ((Integer) o1).intValue();
					int i2 = ((Integer) o2).intValue();
					double w1 = metric.weight[i1];
					double w2 = metric.weight[i2];
					if (w1 < w2)
						return 1;
					else if (w1 > w2)
						return -1;
					else
						return i1 - i2;
				}
			});

			featuresByWeight = new Integer[fbw.size()];
			fbw.copyInto(featuresByWeight);
		}

		// Look for relevant features.  If p is a positive example,
		// a relevant feature is one where group differs from pos.median;
		// if p is a negative example, a relevant features is one where
		// group is identical to pos.median and different from neg.median.
		//
		// Start with high-weight features and work downward.  Also, 
		// skip relevant features that are related to other relevant 
		// features we've already selected.  Continue until we've checked
		// all the features or found enough for the explanation
		// (MAX_EXPLANATION).

		Explanation[] relevant = new Explanation[MAX_EXPLANATION];
		int r = 0;
		featureSearch : for (
			int f = 0;
				r < MAX_EXPLANATION && f < featuresByWeight.length;
				++f) {
			int i = featuresByWeight[f].intValue();

			// test whether feature i is relevant to this example
			if (isPosExample) {
				if (pos.median[i] == group[i])
					continue featureSearch; // not relevant
			} else {
				if (pos.median[i] != group[i])
					continue featureSearch; // not relevant
			}

			boolean isPosFeature = (group[i] > 0.5);

			// test whether feature i is redundant with previously selected
			// features
			Feature baseFeature = features[i].getBaseFeature();
			for (int j = 0; j < r; ++j) {
				Feature explanationBase =
					features[relevant[j].index].getBaseFeature();
				debug.println(
					"considering "
						+ baseFeature
						+ " against "
						+ explanationBase);

				if (baseFeature.equals(explanationBase)) {
					debug.println("base features equal, skipping");
					continue featureSearch;
				}

				switch (features[i].compareTo(features[relevant[j].index])) {
					case RegionSet.EQUAL :
					case RegionSet.SUBSET :
					case RegionSet.SUPERSET :
						if ((isPosFeature && relevant[j].positive)
							|| (!isPosFeature && !relevant[j].positive))
							continue featureSearch;
						break;
					case RegionSet.DISJOINT :
						if (isPosFeature || relevant[j].positive)
							continue featureSearch;
						break;
					default :
						break;
				}
			}
			relevant[r++] = new Explanation(i, isPosFeature);
		}

		// return the relevant features
		Explanation[] explanation = new Explanation[r];
		NumberFormat fmt = NumberFormat.getPercentInstance();
		for (int i = 0; i < r; ++i) {
			Explanation fv = relevant[i];
			fv.posMean =
				(fv.positive) ? pos.mean[fv.index] : 1 - pos.mean[fv.index];
			fv.negMean =
				(fv.positive) ? neg.mean[fv.index] : 1 - neg.mean[fv.index];
			fv.feature = (fv.positive ? "" : "not ") + features[fv.index];
			explanation[i] = fv;
		}
		return explanation;
	}

	public static final int MAX_EXPLANATION = 5;

	/**
	 * A (feature, value) pair in an explanation.
	 * For example, ("contains URL", true).
	 */
	public static class Explanation {
		/** Feature number (index into feature vector). */
		public int index;
		/** Feature value -- all features are binary, so either true or false */
		public boolean positive;
		/** Feature name, such as "contains URL" or "AllCapsWord" */
		public String feature;
		/** Mean value of this feature across all positive examples.
		 * Since features are binary, ranges from 0.0 to 1.0. */
		public double posMean;
		/** Mean value of this feature across all positive examples.
		 * Since features are binary, ranges from 0.0 to 1.0.  */
		public double negMean;
		public Explanation(int index, boolean positive) {
			this.index = index;
			this.positive = positive;
		}
	}

	static class IncreasingDistance extends SortableVector {
		public IncreasingDistance(Vector v) {
			super(v);
		}
		public int compare(Object o1, Object o2) {
			double f1 = ((Cluster) o1).dist;
			double f2 = ((Cluster) o2).dist;
			if (f1 > f2)
				return 1;
			else if (f1 < f2)
				return -1;
			else
				return 0;
		}
	}

}
