/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.ml;

import lapis.*;
import lapis.maps.*;
import lapis.tc.*;
import java.util.*;
import lapisx.util.*;
import lapisx.enum.*;
import lapisx.sort.*;

public class SimulEdit
	implements InferenceAlgorithm, PatternLibrary.LibraryListener {
	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

	// document
	Document doc;

	// record set
	NamedRegionSet records; // kept up to date with current doc version
	NamedRegionSet inRecords; // stays fixed at original doc version

	// feature lists
	Vector uniqueLiterals = new Vector();
	Vector uniquePatterns = new Vector();
	Vector anchoredLiterals = new Vector();
	Vector regularPatterns = new Vector();
	Vector regularLiterals = new Vector();
	Vector varyingPatterns = new Vector();
	Vector varyingLiterals = new Vector();
	Vector editLocations = new Vector();

	// hash map to avoid redundant patterns
	//Map alreadyGenerated = new HashMap (); // Map[RegionSet->Feature]
	Set alreadyGenerated = new HashSet();

	CommonSubstrings cs;

	int editCount = 1;

	public SimulEdit(NamedRegionSet records) {
		// TODO: support multiple documents
        debug.assertion(records.getVersions().getSize() == 1);
        Iterator idoc = records.getDocuments();
        while (idoc.hasNext())
            this.doc = (Document)idoc.next();
            
		this.records = records;
		inRecords = new NamedRegionSet(records.in());

		generateFeatures();
	}

	public NamedRegionSet getRecords() {
		return records;
	}

	public void setRecords(NamedRegionSet records) {
		this.records = records;
	}

	public int getFeatureCount() {
		return editLocations.size()
			+ uniqueLiterals.size()
			+ uniquePatterns.size()
			+ anchoredLiterals.size()
			+ regularPatterns.size()
			+ regularLiterals.size()
			+ varyingPatterns.size();
	}

	public int getRegularFeatureCount() {
		return uniqueLiterals.size()
			+ uniquePatterns.size()
			+ anchoredLiterals.size()
			+ regularPatterns.size()
			+ regularLiterals.size();
	}

	public String toString() {
		return "SimulEdit(" + records + ")";
	}

	//
	// Preprocess document to find useful features
	//
	void generateFeatures() {
		long start = System.currentTimeMillis();
		Main.getProgress().fireStartProgress(
			"Generating features",
			PatternLibrary.getDocumentLibrary(doc).getSize());
		try {
			generateLiterals();
			generatePatterns();
		} finally {
			debug.println(
				"feature generation: "
					+ 
					+ (System.currentTimeMillis() - start)
					+ " ms");
			Main.getProgress().fireDoneProgress();
		}

		dump("unique patterns:", uniquePatterns);
		dump("unique literals:", uniqueLiterals);
		dump("anchored literals:", anchoredLiterals);
		dump("regular patterns:", regularPatterns);
		dump("regular literals:", regularLiterals);
		dump("varying patterns:", varyingPatterns);
		dump("varying literals:", varyingLiterals);

		// listen for future changes in library
		PatternLibrary.getDocumentLibrary(doc).addLibraryListener(this);
	}

	static void dump(String title, Vector features) {
		debug.println(title);
		for (Enumeration e = features.elements(); e.hasMoreElements();) {
			Feature f = (Feature) e.nextElement();
			debug.println(f.toString());
		}
	}

	/**
	 * Generalize a selection to patterns matching an equivalent region
	 * in every record.
	 * @param r NamedRegionSet to generalize
	 * @return Enumeration of Features representing generalizations,
	 * best generalization first.
	 */
	public RestartableEnumeration generalize(
		NamedRegionSet pos,
		NamedRegionSet neg) {
		Examples ex = new Examples(pos, neg);
		Enumeration hyps = generalize(ex);
		hyps = addRecordSet(hyps);
		return new MemoizedEnumeration(hyps);
	}

	Enumeration addRecordSet(Enumeration e) {
		return new FilteredEnumeration(e) {
			public void transform(Object o) {
				Feature f = (Feature) o;
				debug.println(
					"hypothesis from simuledit(" + records + "): " + f);
				yield(new Feature(f, f + "\n  in " + records));
			}
		};
	}

	Enumeration generalize(Examples ex) {
		debug.println("original examples: " + ex);

		if (ex.pos.isEmpty())
			return new EmptyEnumeration();
		else if (!isInRecords(ex))
			return new EmptyEnumeration();
		else if (ex.isAllPoints())
			return new ConcatEnumeration(
				new Enumeration[] {
					EDITS.generalize(ex),
					STRONG.generalize(ex),
					WEAK.generalize(ex)});
		else {
			PairEnumeration ep0 = generalizeEndpoints(ex, EDITS, null);
			PairEnumeration ep1 = generalizeEndpoints(ex, STRONG, ep0);
			PairEnumeration ep2 = generalizeEndpoints(ex, WEAK, ep1);
			return new ConcatEnumeration(
				new Enumeration[] {
					EDITS.generalize(ex),
					consistentHypotheses(ep0, (Examples) ex.clone()),
					STRONG.generalize(ex),
					consistentHypotheses(ep1, (Examples) ex.clone()),
					WEAK.generalize(ex),
					consistentHypotheses(ep2, (Examples) ex.clone())});
		}
	}

	boolean isInRecords(Examples ex) {
		//RegionSet inSomeRecord = inRecords.getRegions ();
		RegionEnumeration re = ex.getPositives().regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast())
			if (!inRecords.intersects(r.convert(inRecords)))
				return false;
		return true;
	}

	interface Generalizer {
		public Enumeration generalize(Examples ex);
	}

	Generalizer EDITS = new Generalizer() {
		public Enumeration generalize(Examples ex) {
			Enumeration hyp = generalizeEditLocation(ex);
			hyp = dumpHypotheses(ex, " edit guess: ", hyp);
			hyp = consistentHypotheses(hyp, ex);
			hyp = dumpHypotheses(ex, " OK: ", hyp);
			return hyp;
		}
	};

	Generalizer STRONG = new Generalizer() {
		public Enumeration generalize(Examples ex) {
			Enumeration hyp =
				new ConcatEnumeration(
					new Enumeration[] {
						generalizeUniquePattern(ex),
						generalizeUniqueLiteral(ex),
						generalizeAnchoredLiteral(ex),
						generalizeCountedPattern(regularPatterns, ex),
						generalizeCountedLiteral(regularLiterals, ex)});

			hyp = dumpHypotheses(ex, " strong guess: ", hyp);
			hyp = consistentHypotheses(hyp, ex);
			hyp = dumpHypotheses(ex, " OK: ", hyp);
			return hyp;
		}
	};

	Generalizer WEAK = new Generalizer() {
		public Enumeration generalize(Examples ex) {
			Enumeration hyp =
				new ConcatEnumeration(
					new Enumeration[] {
						generalizeCountedPattern(varyingPatterns, ex),
						generalizeCountedLiteral(varyingLiterals, ex)});

			// sort hypotheses so that smallest counts are first
			// first X < last X < second X < second from last X < ...
			hyp = dumpHypotheses(ex, " weak guess: ", hyp);
			hyp = consistentHypotheses(hyp, ex);
			hyp = dumpHypotheses(ex, " OK: ", hyp);
			return hyp;
		}
	};

	static Enumeration consistentHypotheses(Enumeration e, final Examples ex) {
		final Examples originalEx = (Examples) ex.clone();
		return new FilteredEnumeration(e) {
			public void transform(Object o) {
				Feature f = (Feature) o;
				f = ex.convertBack(f);
				if (isConsistent(f, originalEx))
					yield(f);
			}
		};
	}

	static boolean isConsistent(Feature f, Examples ex) {
		DocumentVersion curr = ex.getVersion();
        
        // minwu
		//RegionSet currMatches = f.convert(Collections.singleton(curr));
        
        RegionSet currMatches = f.convert(new SingleDocVersions(curr));
		// TODO: to handle multidocs, currMatches must be made version-compatible with
		// both ex.getPositives() and ex.getNegatives ().  One way to do it:
		// currMatches = f.convert (ex.getPositives ()).convert (ex.getNegatives ());

		debug.println("testing feature " + currMatches.regions().firstFast());
		RegionSet pos = ex.getPositives();
		RegionEnumeration re = pos.regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
			if (!currMatches.intersects(r)) {
				debug.println("pos example " + r + " failed to match");
				return false;
			}
		}

		RegionSet neg = ex.getNegatives();
		re = neg.regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
			if (currMatches.intersects(r)) {
				debug.println("neg example " + r + " matched");
				return false;
			}
		}

		return true;
	}

	static Enumeration dumpHypotheses(
		final Examples ex,
		final String label,
		Enumeration e) {
		return new FilteredEnumeration(e) {
			public void transform(Object o) {
				debug.println(ex + label + o + o.getClass());
				yield(o);
			}
		};
	}

	public RegionSet getSubregions(RegionSet matches, int start, int end) {
		MutableRegionSet result = new MutableRegionSet();
		RegionEnumeration re = matches.regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast())
			result.insert(
				r.getStartLow() + start,
				r.getStartLow() + end,
				r.getVersion());
		return result;
	}

	class Examples implements Cloneable {
		RegionSet pos;
		RegionSet neg;

		Region oneExample;

		Hashtable oneExampleByVersion = new Hashtable();
		Hashtable startsOneExampleByVersion = new Hashtable();
		Hashtable endsOneExampleByVersion = new Hashtable();

		boolean allPoints;

		Examples startPoints;
		Examples endPoints;

		CoordinateMap map; // composed map followed back through pastes 
		DocumentVersion mapSourceVer;

		public Examples(RegionSet pos, RegionSet neg) {
			debug.assertion(
				pos.getVersions().getSize() == 0 || (pos.getVersions().getSize() == 1));
			this.pos = pos;
			this.neg = neg;

			oneExample = pos.regions().first();
			oneExampleByVersion.put(oneExample.getVersion(), oneExample);

            // minwu
			//mapSourceVer = (DocumentVersion) (pos.getVersions().toArray())[0];
            
            Iterator iver = pos.getVersions().versions();
            while (iver.hasNext())
                mapSourceVer = (DocumentVersion)iver.next();
                
			map = IdentityMap.IDENTITY;

			// test positive examples for various properties
			allPoints = pos.isAllPoints();
		}

		public Object clone() {
			try {
				Examples ex = (Examples) super.clone();
				ex.oneExampleByVersion = (Hashtable) ex.oneExampleByVersion;
				ex.startsOneExampleByVersion =
					(Hashtable) ex.startsOneExampleByVersion;
				ex.endsOneExampleByVersion =
					(Hashtable) ex.endsOneExampleByVersion;
				return ex;
			} catch (CloneNotSupportedException e) {
				throw new RuntimeException();
			}
		}

		public String toString() {
			String s = "";
			RegionEnumeration re = pos.regions();
			for (Region r = re.firstFast(); r != null; r = re.nextFast())
				s += r + " ";
			if (!neg.isEmpty()) {
				s += "/";
				re = neg.regions();
				for (Region r = re.firstFast(); r != null; r = re.nextFast())
					s += r + " ";
			}
			return s + getVersion();
		}

		public boolean isAllPoints() {
			return allPoints;
		}

		public DocumentVersion getVersion() {
			return oneExample.getVersion();
		}

		public RegionSet getPositives() {
			return pos;
		}

		public RegionSet getNegatives() {
			return neg;
		}

		/**
		 * Get one positive example. 
		 * @param f Region set that positive example will be compared with, to determine document versions.
		 * Use RegionSet.EMPTY if you don't care what document version is returned.
		 * @return one positive example region, relative to the same document version as f.  
		 */
		public Region getOneExample(RegionSet f) {
			DocumentVersion ver =
				f.getSingleDocRegionSet(oneExample.getDocument()).getVersion();
			if (ver == null)
				ver = oneExample.getVersion();

			Region r = (Region) oneExampleByVersion.get(ver);
			if (r == null) {
				r = (Region) oneExample.convert(f).flatten();
				oneExampleByVersion.put(ver, r);
			}
			return r;
		}

		/**
		 * Get the set of regions that are just before one positive example. 
		 * @param f Region set that positive example will be compared with, to determine document versions.
		 * Use RegionSet.EMPTY if you don't care what document version is returned.
		 * @return justBefore one positive example region, relative to the same document version as f.  
		 */
		public Region getStartsOneExample(RegionSet f) {
			DocumentVersion ver =
				f.getSingleDocRegionSet(oneExample.getDocument()).getVersion();
			if (ver == null)
				ver = oneExample.getVersion();

			Region r = (Region) startsOneExampleByVersion.get(ver);
			if (r == null) {
				r = (Region) getOneExample(f).justAfter(RegionSet.EMPTY);
				startsOneExampleByVersion.put(ver, r);
			}
			return r;
		}

		/**
		 * Get the set of regions that are just after one positive example. 
		 * @param f Region set that positive example will be compared with, to determine document versions.
		 * Use RegionSet.EMPTY if you don't care what document version is returned.
		 * @return justAfter one positive example region, relative to the same document version as f.  
		 */
		public Region getEndsOneExample(RegionSet f) {
			DocumentVersion ver =
				f.getSingleDocRegionSet(oneExample.getDocument()).getVersion();
			if (ver == null)
				ver = oneExample.getVersion();

			Region r = (Region) endsOneExampleByVersion.get(ver);
			if (r == null) {
				r = (Region) getOneExample(f).justBefore(RegionSet.EMPTY);
				endsOneExampleByVersion.put(ver, r);
			}
			return r;
		}

		public Examples startOf() {
			if (startPoints == null)
				startPoints = new Examples(pos.startOf(), RegionSet.EMPTY);
			return startPoints;
		}

		public Examples endOf() {
			if (endPoints == null)
				endPoints = new Examples(pos.endOf(), RegionSet.EMPTY);
			return endPoints;
		}

		public void convert(CoordinateMap map, DocumentVersion mapVer) {
			// TODO: support multiple document
			debug.assertion(mapVer.getDocument() == oneExample.getDocument());
			pos =
				pos.getSingleDocRegionSet(mapVer.getDocument()).convert(
					map,
					mapVer);
			neg =
				neg.getSingleDocRegionSet(mapVer.getDocument()).convert(
					map,
					mapVer);
			oneExample = pos.regions().first();

			this.map = new CompositeCoordinateMap(map, this.map);

			// clear caches
			oneExampleByVersion.clear();
			startsOneExampleByVersion.clear();
			endsOneExampleByVersion.clear();
			startPoints = null;
			endPoints = null;

			// reload some of the caches
			oneExampleByVersion.put(oneExample.getVersion(), oneExample);
		}

		public Feature convertBack(Feature f) {
			// TODO: support multiple documents
			debug.assertion(f.getVersions().getSize() == 1);
            
            // minwu
			//Set vers = new HashSet();
			//vers.add(getVersion());
            
			if (map != IdentityMap.IDENTITY) {
				SingleDocRegionSet rs =
					f.convert(new SingleDocVersions(getVersion())).getSingleDocRegionSet(doc);
				Feature g =
					new Feature(
						rs.convert(map.invert(), mapSourceVer),
						f.toString());

				debug.println(
					"converted from "
						+ f.regions().firstFast()
						+ (DocumentVersion) (f.getVersions().versions().next())
						+ " to "
						+ g.regions().firstFast()
						+ (DocumentVersion) (g.getVersions().versions().next()));
				return g;
			} else
				return f;
		}
	}

	//
	// Editing locations:
	// in P  for every deleted, inserted, or pasted region P
	// eq P  for every copied region P
	//

	public void addEditLocation(NamedRegionSet endpoints, int len) {
		MutableRegionSet matches = new MutableRegionSet();
		RegionEnumeration re = endpoints.regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
			int s = r.getEndHigh() - len;
			int e = r.getEndHigh();
			matches.insert(new FuzzyRegion(s, e, s, e, r.getVersion()));
		}

		Feature f = new EditFeature(matches, "edit" + (editCount++));
		editLocations.insertElementAt(f, 0);
		debug.println("added feature " + f);
		// FIX: append instead, but need to generalize in LIFO order
	}

	public void addPasteLocation(
		NamedRegionSet targetEnds,
		NamedRegionSet sourceSelection) {
		// TODO: support multiple documents
		debug.assertion(targetEnds.getVersions().getSize() == 1);
		debug.assertion(sourceSelection.getVersions().getSize() == 1);

        // minwu
		//Document sourceDoc =
		//	(Document) (sourceSelection.getDocuments().toArray())[0];
		//DocumentVersion sourceVer =
		//	(DocumentVersion) (sourceSelection.getVersions().toArray())[0];
		//DocumentVersion ver =
		//	(DocumentVersion) (targetEnds.getVersions().toArray())[0];
        
        Document sourceDoc = null;
        DocumentVersion sourceVer = null;
        DocumentVersion targetVer = null;
        Iterator isourcever = sourceSelection.getVersions().versions();
        while (isourcever.hasNext()) {
            sourceVer = (DocumentVersion)isourcever.next();
            sourceDoc = sourceVer.getDocument();
        }
        Iterator itargetver = targetEnds.getVersions().versions();
        while (itargetver.hasNext()) {
            targetVer = (DocumentVersion)itargetver.next();
        }

		if (doc != sourceDoc)
			// copied from different document; can't infer from it
			// with this SimulEdit object.
			// FIX: check whether sourceDoc has an associated SimulEdit
			// object; if so, refer to that one.
			return;

		// figure out all the target regions, and compute a coordinate map
		PiecewiseLinearMonotonicMap map =
			new PiecewiseLinearMonotonicMap(sourceSelection.getSize());
		MutableRegionSet target1 = new MutableRegionSet();
		MutableRegionSet target2 = new MutableRegionSet();
		RegionEnumeration reSource = sourceSelection.regions();
		RegionEnumeration reTarget = targetEnds.regions();

		for (Region src = reSource.firstFast(), targ = reTarget.firstFast();
			src != null && targ != null;
			src = reSource.nextFast(), targ = reTarget.nextFast()) {
			int s = targ.getEndHigh() - (src.getEndHigh() - src.getStartLow());
			int e = targ.getEndHigh();
			target1.insert(s, e, targetVer);
			target2.insert(Region.make(s + 1, e - 1, s + 1, e - 1, targetVer));
			map.add(s, src.getStartLow());
			map.add(e, src.getEndHigh());
		}

		String name = "paste" + (editCount++);
		Feature f1 = new Feature(target1, name);
		Feature f2 = new PasteFeature(target2, map, sourceVer, name);
		editLocations.insertElementAt(f2, 0);
		debug.println("added feature " + f2);
		editLocations.insertElementAt(f1, 0);
		debug.println(
			"added feature "
				+ f1
				+ " "
				+ f1.regions().firstFast()
				+ " "
				+ (DocumentVersion) (f1.getVersions().versions().next()));
		// FIX: append instead, but need to generalize in LIFO order
	}

	public void addCopyLocation(NamedRegionSet sel) {
		Feature f = new Feature(sel, "copy" + (editCount++));
		editLocations.insertElementAt(f, 0);
		debug.println("added feature " + f);
		// FIX: append instead, but need to generalize in LIFO order
	}

	Enumeration generalizeEditLocation(final Examples ex) {
		return new FilteredEnumeration(editLocations.elements()) {
			public void transform(Object o) {
				if (o instanceof EditFeature) {
					((EditFeature) o).yield(ex, this);
				} else {
					Feature f = (Feature) o;
					if (!ex.isAllPoints())
						yield(f);
					else {
						Region startsR = ex.getStartsOneExample(f);
						Region endsR = ex.getEndsOneExample(f);
						if (f.intersects(startsR))
							yield(new Feature(f, f.startOf(), "start of " + f));
						if (f.intersects(endsR))
							yield(new Feature(f, f.endOf(), "end of " + f));
					}
				}
			}
		};
	}

	class EditFeature extends Feature {
		public EditFeature(RegionSet matches, String description) {
			super(matches, description);
		}

		public void yield(Examples ex, FilteredEnumeration enum) {
			// TODO: EditFeature support multiple documents
			Region r = ex.getOneExample(this);
			Region m = this.regions(r).firstFast();
			if (m != null) {
				Feature g =
					new Feature(
						this,
						getSubregions(
							this,
							r.getStartLow() - m.getStartLow(),
							r.getEndHigh() - m.getStartLow()),
						"somewhere in " + this);
				enum.yield(g);
			}
		}
	}

	class PasteFeature extends EditFeature {
		CoordinateMap map; // maps coordinates from target back to source
		CoordinateMap inverse; // maps target to source
		DocumentVersion sourceVer;

		public PasteFeature(
			RegionSet target,
			CoordinateMap map,
			DocumentVersion sourceVer,
			String description) {
			super(target, description);
			this.map = map;
			this.inverse = map.invert();
			this.sourceVer = sourceVer;
		}

		public void yield(Examples ex, FilteredEnumeration enum) {
			// test whether ALL positive examples lie in pasted targets
			// TODO: PasteFeature support multiple documents

			RegionSet pos = ex.getPositives();
			if (pos.isEmpty())
				return;
			//RegionSet targets = getRegions ();
			RegionEnumeration re = pos.regions();
			for (Region r = re.firstFast(); r != null; r = re.nextFast())
				if (!this.intersects(r.convert(this)))
					return;

			// map examples back to source coordinate space
            
            // minwu
			//DocumentVersion ver =
			//	(DocumentVersion) (this.getVersions().toArray())[0];
            
            DocumentVersion ver = null;
            Iterator iver = this.getVersions().versions();
            while (iver.hasNext())
                ver = (DocumentVersion)iver.next();
                
			CoordinateMap examplesToTarget = ver.diff(ex.getVersion());

			CoordinateMap examplesToSource =
				new CompositeCoordinateMap(map, examplesToTarget);
			ex.convert(examplesToSource, sourceVer);
			SimulEdit.debug.println("mapped back through " + this +" to " + ex);
		}
	}

	//
	// Handling undo: remove all editLocations with versions later than ver
	//
	public void undoTo(DocumentVersion ver) {
		// TODO: support multiple documents
		for (int i = editLocations.size() - 1; i >= 0; --i) {
			Feature f = (Feature) editLocations.elementAt(i);
            
            // minwu
			//debug.assertion(f.getVersions().size() == 1);
			//DocumentVersion fver =
			//	(DocumentVersion) (f.getVersions().toArray())[0];
            
            debug.assertion(f.getVersions().getSize() == 1);
            DocumentVersion fver = null;
            Iterator iver = f.getVersions().versions();
            while (iver.hasNext())
                fver = (DocumentVersion)iver.next();
                
			if (ver.lessThan(fver)) {
				debug.println("removing future edit " + f);
				editLocations.removeElementAt(i);
			}
		}
	}

	public void close() {
		PatternLibrary.getDocumentLibrary(doc).removeLibraryListener(this);
	}

	//
	// Literal features:
	// in "lit" for every unique or regular common substring <lit>
	//

	void generateLiterals() {
		String[] recordContent = new String[records.getSize()];
		RegionEnumeration re = records.regions();
		int m = 0;
		for (Region r = re.firstFast(); r != null; r = re.nextFast())
			recordContent[m++] = doc.getText(r);

		cs = new CommonSubstrings(recordContent);
		CommonSubstrings.CommonSubstring[] subs =
			cs.getMaximalCommonSubstrings();

		for (int i = 0; i < subs.length; ++i) {
			CommonSubstrings.CommonSubstring c = subs[i];
			if (c.minOccurrences == c.maxOccurrences) {
				if (c.minOccurrences == 1) {
					uniqueLiterals.addElement(new LiteralFeature(c.s));
				} else {
					regularLiterals.addElement(new RegularLiteralFeature(c.s));
				}
			} else {
				varyingLiterals.addElement(
					new VaryingLiteralFeature(c.s, c.minOccurrences));
			}
		}

		// sort unique and regular literals by length.  Longer literals are preferred
		// to shorter ones.
		Quicksort.sort(new LongestFirst(uniqueLiterals));
		Quicksort.sort(new LongestFirst(regularLiterals));
	}

	class LongestFirst extends SortableVector {
		public LongestFirst(Vector v) {
			super(v);
		}
		public int compare(Object o1, Object o2) {
			String l1 = ((LiteralFeature) o1).lit;
			String l2 = ((LiteralFeature) o2).lit;
			return l2.length() - l1.length();
		}
	}

	Enumeration generalizeUniqueLiteral(final Examples ex) {
		return new FilteredEnumeration(uniqueLiterals.elements()) {
			public void transform(Object o) {
				LiteralFeature f = (LiteralFeature) o;
				Region r = ex.getOneExample(f);
				Region m = f.regions(r).firstFast();
				if (m != null) {
					debug.println(
						"example "
							+ r
							+ " occurs in unique literal "
							+ f
							+ " at "
							+ m);
					yield(
						f.getInnerFeature(
							r.getStartLow() - m.getStartLow(),
							r.getEndHigh() - m.getStartLow()));
				}
			}
		};
	}

	class LiteralFeature extends Feature {
		public String lit;
		Feature entireLiteral;

		public LiteralFeature(String lit) {
			this(
				lit,
				TC.makeLiteral(lit),
				new TC(
					new TC.And(
						new TC.CaseSensitivity(true, new TC.Literal(lit)),
						new TC.Constant(inRecords))).match(
					doc));
		}

		public LiteralFeature(String lit, String pattern, RegionSet matches) {
			super(matches.in(), pattern);
			this.lit = lit;
			this.entireLiteral = new Feature(matches, pattern);
		}

		public String describe(int start, int end) {
			// suppose literal is X[Y]Z, where [Y] is the 
			// substring identified by [start,end],
			String x = lit.substring(0, start);
			String y = lit.substring(start, end);
			String z = lit.substring(end);

			String pattern;
			if (cs.isUnique(y))
				// Y is unique, so Y itself is an unambigious description
				pattern = TC.makeLiteral(y);
			else {
				String x2 = cs.getShortestUniqueSuffix(x + y);
				String z2 = cs.getShortestUniquePrefix(y + z);
				if (x2 != null && (z2 == null || x2.length() <= z2.length())) {
					String x3 = x2.substring(0, x2.length() - y.length());

					pattern = "just after " + TC.makeLiteral(x3);
				} else if (z2 != null) {
					String z3 = z2.substring(y.length());

					pattern = "just before " + TC.makeLiteral(z3);
				} else {

					String w = cs.getShortestUniquePrefix(lit);
					String v = cs.getShortestUniqueSuffix(lit);

					if (x.length() + w.length() <= z.length() + v.length()) {
						pattern = "starting " + TC.makeLiteral(w);

						if (x.length() > 0)
							pattern =
								"just after "
									+ TC.makeLiteral(x)
									+ " "
									+ pattern;
					} else {
						pattern = "ending " + TC.makeLiteral(v);

						if (z.length() > 0)
							pattern =
								"just before "
									+ TC.makeLiteral(z)
									+ " "
									+ pattern;
					}
				}

				// prepend y to the pattern, either as itself or as point
				if (start == end)
					pattern = "point " + pattern;
				else
					pattern = TC.makeLiteral(y) + " " + pattern;
			}

			return pattern;
		}

		public Feature getInnerFeature(int start, int end) {
			return new Feature(
				this,
				getSubregions(this, start, end),
				describe(start, end));
		}

	}

	Enumeration generalizeCountedLiteral(Vector literals, final Examples ex) {
		return new SmallNFirstEnumeration(new FilteredEnumeration(
			literals
			.elements()) {
			public void transform(Object o) {
				RegularLiteralFeature f = (RegularLiteralFeature) o;
				// TODO: support multiple documents
				//debug.assertion(f.getDocumentVersions().size() == 1);
				Region r = ex.getOneExample(f);
				Region m = f.regions(r).first();
				if (m != null)
					f.yield(this, r, m);
			}
		});
	}

	static interface Counted {
		int getN();
	}

	class SmallNFirstEnumeration extends SortedEnumeration {
		public SmallNFirstEnumeration(Enumeration hyp) {
			super(hyp);
		}

		public int compare(Object o1, Object o2) {
			if (!(o1 instanceof Counted))
				debug.println("o1 = " + o1 + " instanceof " + o1.getClass());
			int n1 = ((Counted) o1).getN();
			int n2 = ((Counted) o2).getN();
			int diff = Math.abs(n1) - Math.abs(n2);
			if (diff != 0)
				return diff;
			else
				return n2 - n1;
		}
	}

	class RegularLiteralFeature extends LiteralFeature {
		public RegularLiteralFeature(String lit) {
			super(lit);
		}

		public void yield(FilteredEnumeration enum, Region r, Region m) {
			int[] result = new int[2];
			if (!getIndexOf(r, this, result)) {
				SimulEdit.debug.println("Failed to find ordinum of " + r + "!");
				return;
			}

			int start = r.getStartLow() - m.getStartLow();
			int end = r.getEndHigh() - m.getStartLow();

			Feature ref = new NthFeature(entireLiteral, result[0]);
			enum.yield(
				new NthRegularLiteralFeature(
					this,
					getSubregions(ref, start, end),
					describe(start, end) + ref,
					result[0]));
		}

		public String describe(int start, int end) {
			// suppose literal is X[Y]Z, where [Y] is the 
			// substring identified by [start,end],
			int len = lit.length();
			if (start == 0) {
				if (end == len)
					return "";
				else if (start == end)
					return "start of ";
				else
					return TC.makeLiteral(lit.substring(start, end))
						+ " starting ";
			} else if (end == len) {
				if (start == end)
					return "end of ";
				else
					return TC.makeLiteral(lit.substring(start, end))
						+ " ending ";
			} else
				return (
					(start == end)
						? "end of "
						: (TC.makeLiteral(lit.substring(start, end))
							+ " just after "))
					+ TC.makeLiteral(lit.substring(0, start))
					+ " starting ";
		}
	}

	class NthRegularLiteralFeature extends Feature implements Counted {
		int n;
		public NthRegularLiteralFeature(
			Feature ref,
			RegionSet regions,
			String description,
			int n) {
			super(ref, regions, description);
			this.n = n;
		}
		public int getN() {
			return n;
		}
	}

	class VaryingLiteralFeature extends RegularLiteralFeature {
		int min; // min number of occurences in records
		public VaryingLiteralFeature(String lit, int min) {
			super(lit);
			this.min = min;
		}

		public void yield(FilteredEnumeration enum, Region r, Region m) {
			int[] result = new int[2];
			if (!getIndexOf(r, this, result)) {
				SimulEdit.debug.println(
					r
						+ " matched varying pattern "
						+ this
						+ ", with "
						+ result[1]
						+ " hits in this record, but couldn't determine its index");
				return;
			}

			int start = r.getStartLow() - m.getStartLow();
			int end = r.getEndHigh() - m.getStartLow();

			int i = result[0];
			int n = result[1];

			// first try "ith"
			if (i <= min) {
				Feature ref = new NthFeature(entireLiteral, i);
				enum.yield(
					new NthFeature(
						getSubregions(ref, start, end),
						i,
						describe(start, end) + ref));
			}

			// now try "n-i+1th from last"
			int k = n - i + 1;
			if (k <= min) {
				Feature ref = new NthFeature(entireLiteral, -k);
				enum.yield(
					new NthFeature(
						getSubregions(ref, start, end),
						-k,
						describe(start, end) + ref));
			}
		}
	}

	//
	// Pattern features
	//

	void generatePatterns() {
		// get all nontrivial patterns from pattern library,
		// qualify all features with "in Record", and separate
		// unique patterns from repeated patterns
		for (Enumeration e = new LibraryGenerator(doc); e.hasMoreElements();) {
			Feature f = (Feature) e.nextElement();
			addPatternFeature(f, false);
		}

		// make sure we have a feature for the entire record
		addPatternFeature(new Feature(records, "record"), false);
	}

	void addPatternFeature(Feature f, boolean important) {
		Vector v;
		int[] minmax = new int[2];

		countOccurrences(f, 1, Integer.MAX_VALUE, minmax);
		if (minmax[0] > 0) {
			// pattern occurs at least once in every record

			// fix f so that it only matches inside records
			f = new Feature(f, f.and(inRecords).crispen(), f.toString());

			// check whether we've already generated an equivalent feature
			if (!alreadyGenerated.contains(f)) {
				alreadyGenerated.add(f);

				if (minmax[1] == 1) {
					// pattern matches at most once in every record
					debug.println(f + " unique");
					v = uniquePatterns;
					addAnchoredLiterals(f);
				} else if (minmax[0] == minmax[1]) {
					debug.println(
						f + " repeated exactly " + minmax[0] + " times");
					v = regularPatterns;
					f = new RegularFeature(f);
				} else {
					debug.println(
						f
							+ " repeated "
							+ minmax[0]
							+ "-"
							+ minmax[1]
							+ " times");
					v = varyingPatterns;
					f = new VaryingFeature(f, minmax[0]);
				}

				if (important)
					v.insertElementAt(f, 0);
				else
					v.addElement(f);
			}
		}
	}

	//
	// Library listener
	//
	public void patternAdded(final PatternLibrary lib, final String name) {
		Thread t = new Thread() {
			public void run() {
				RegionSet matches = lib.get(name).match(doc);
				if (!matches.isEmpty()) {
					Feature f = new Feature(matches, lib.getShortestName(name));
					addPatternFeature(f, true);
				}
			}
		};
		t.setPriority(3);
		t.start();
	}

	public void patternRemoved(PatternLibrary lib, String name) {
	}

	public void parserAdded(PatternLibrary lib, Parser parser) {
	}

	void countOccurrences(
		RegionSet matches,
		int minAllowed,
		int maxAllowed,
		int[] minmax) {
		int min = Integer.MAX_VALUE;
		int max = Integer.MIN_VALUE;

		RegionEnumeration re = inRecords.regions();
		for (Region rec = re.firstFast(); rec != null; rec = re.nextFast()) {
			RegionEnumeration featuresInRec = matches.regions(rec);
			int n = 0;
			for (Region r = featuresInRec.firstFast();
				r != null;
				r = featuresInRec.nextFast())
				++n;
			min = Math.min(min, n);
			max = Math.max(max, n);

			if (n < minAllowed || n > maxAllowed)
				break;
		}

		minmax[0] = min;
		minmax[1] = max;
	}

	Enumeration generalizeUniquePattern(final Examples ex) {
		Enumeration hypotheses = uniquePatterns.elements();

		if (ex.isAllPoints())
			hypotheses = new FilteredEnumeration(hypotheses) {
			public void transform(Object o) {
				Feature f = (Feature) o;
				Region startsR = ex.getStartsOneExample(f);
				Region endsR = ex.getEndsOneExample(f);
				if (f.intersects(startsR))
					yield(new Feature(f, f.startOf(), "start of " + f));
				if (f.intersects(endsR))
					yield(new Feature(f, f.endOf(), "end of " + f));
			}
		};

		return hypotheses;
	}

	//
	// Anchored literal features:
	// in "lit" jb/ja/st/en/eq unique pattern f
	//

	void addAnchoredLiterals(Feature f) {
		String eq = null;
		String st = null;
		String end = null;
		String jbef = null;
		String jaft = null;
		RegionEnumeration re = f.regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
			eq = updateEquality(eq, r.getStartLow(), r.getEndHigh());
			st = updatePrefix(st, r.getStartLow());
			end = updateSuffix(end, r.getEndHigh());
			jaft = updatePrefix(jaft, r.getEndHigh());
			jbef = updateSuffix(jbef, r.getStartLow());
			if (st.length() == 0
				&& end.length() == 0
				&& jaft.length() == 0
				&& jbef.length() == 0)
				return;
		}

		debug.println(
			f
				+ ": eq="
				+ eq
				+ ", st="
				+ st
				+ ", end="
				+ end
				+ ", jbef="
				+ jbef
				+ ", jaft="
				+ jaft);

		if (eq.length() > 0) {
			// use only the equals feature; skip the starting 
			// and ending features (which should be identical)
			if (!cs.isUnique(eq))
				anchoredLiterals.addElement(new LiteralEqualsFeature(eq, f));
		} else {
			if (st.length() > 0 && !cs.isUnique(st))
				anchoredLiterals.addElement(new LiteralStartsFeature(st, f));
			if (end.length() > 0 && !cs.isUnique(end))
				anchoredLiterals.addElement(new LiteralEndsFeature(end, f));
		}

		if (jbef.length() > 0 && !cs.isUnique(jbef))
			anchoredLiterals.addElement(new LiteralJustBeforeFeature(jbef, f));

		if (jaft.length() > 0 && !cs.isUnique(jaft))
			anchoredLiterals.addElement(new LiteralJustAfterFeature(jaft, f));

		Main.getProgress().fireAddProgress(1);
	}

	String updateEquality(String eq, int start, int end) {
		if (eq != null && eq.length() != end - start)
			return "";
		String s = doc.getText(new Region(start, end, doc));
		return (eq == null || eq.equals(s)) ? s : "";
	}

	String updatePrefix(String prefix, int start) {
		if (prefix != null && prefix.length() == 0)
			return prefix;
		String s = doc.getText(new Region(start, RegionSet.MAX_INDEX, doc));
		return (prefix != null) ? Str.longestCommonPrefix(s, prefix) : s;
	}

	String updateSuffix(String suffix, int end) {
		if (suffix != null && suffix.length() == 0)
			return suffix;
		String s = doc.getText(new Region(RegionSet.MIN_INDEX, end, doc));
		return (suffix != null) ? Str.longestCommonSuffix(s, suffix) : s;
	}

	Enumeration generalizeAnchoredLiteral(final Examples ex) {
		return new FilteredEnumeration(anchoredLiterals.elements()) {
			public void transform(Object o) {
				LiteralFeature f = (LiteralFeature) o;
				Region r = ex.getOneExample(f);
				Region m = f.regions(r).firstFast();
				if (m != null)
					yield(
						f.getInnerFeature(
							r.getStartLow() - m.getStartLow(),
							r.getEndHigh() - m.getStartLow()));
			}
		};
	}

	class AnchoredLiteralFeature extends LiteralFeature {
		public Feature relop;

		public AnchoredLiteralFeature(String lit, Feature relop) {
			super(
				lit,
				TC.makeLiteral(lit) + " " + relop,
				new TC(
					new TC.And(
						new TC.And(
							new TC.CaseSensitivity(true, new TC.Literal(lit)),
							new TC.Constant(inRecords)),
						new TC.Constant(relop))).match(
					doc));
			this.relop = relop;
		}
	}

	class LiteralJustBeforeFeature extends AnchoredLiteralFeature {
		public LiteralJustBeforeFeature(String lit, Feature f) {
			super(
				lit,
				new Feature(
					f,
					f.justBefore(RegionSet.EMPTY),
					"just before " + f));
		}

		public String describe(int start, int end) {
			String pattern = relop.toString();
			if (end < lit.length())
				pattern =
					"just before "
						+ TC.makeLiteral(lit.substring(end))
						+ " "
						+ pattern;
			if (start == end)
				pattern = "point " + pattern;
			else
				pattern =
					TC.makeLiteral(lit.substring(start, end)) + " " + pattern;
			return pattern;
		}
	}

	class LiteralJustAfterFeature extends AnchoredLiteralFeature {
		public LiteralJustAfterFeature(String lit, Feature f) {
			super(
				lit,
				new Feature(
					f,
					f.justAfter(RegionSet.EMPTY),
					"just after " + f));
		}

		public String describe(int start, int end) {
			String pattern = relop.toString();
			if (start > 0)
				pattern =
					"just after "
						+ TC.makeLiteral(lit.substring(0, start))
						+ " "
						+ pattern;
			if (start == end)
				pattern = "point " + pattern;
			else
				pattern =
					TC.makeLiteral(lit.substring(start, end)) + " " + pattern;
			return pattern;
		}
	}

	class LiteralStartsFeature extends AnchoredLiteralFeature {
		public LiteralStartsFeature(String lit, Feature f) {
			super(
				lit,
				new Feature(f, f.starting(RegionSet.EMPTY), "starting " + f));
		}

		public String describe(int start, int end) {
			String pattern = relop.toString();
			if (start > 0)
				pattern =
					"just after "
						+ TC.makeLiteral(lit.substring(0, start))
						+ " "
						+ pattern;
			if (start == end)
				pattern = "point " + pattern;
			else
				pattern =
					TC.makeLiteral(lit.substring(start, end)) + " " + pattern;
			return pattern;
		}
	}

	class LiteralEndsFeature extends AnchoredLiteralFeature {
		public LiteralEndsFeature(String lit, Feature f) {
			super(
				lit,
				new Feature(f, f.ending(RegionSet.EMPTY), "ending " + f));
		}

		public String describe(int start, int end) {
			String pattern = relop.toString();

			if (end < lit.length())
				pattern =
					"just before "
						+ TC.makeLiteral(lit.substring(end))
						+ " "
						+ pattern;

			if (start == end)
				pattern = "point " + pattern;
			else
				pattern =
					TC.makeLiteral(lit.substring(start, end)) + " " + pattern;

			return pattern;
		}
	}

	class LiteralEqualsFeature extends AnchoredLiteralFeature {
		Feature baseFeature;

		public LiteralEqualsFeature(String lit, Feature f) {
			super(lit, new Feature(f, (RegionSet) f, "equals " + f));
			baseFeature = f;
		}

		public String describe(int start, int end) {
			String pattern = baseFeature.toString();

			// suppose literal is X[Y]Z, where [Y] is the substring 
			// identified by [start,end],
			if (start == 0 && end == lit.length())
				// literal = [Y]; just return P
				return pattern;
			else if (start <= lit.length() - end)
				// |X| <= |Z|
				// return "Y jaft X starting P"
				pattern =
					"just after "
						+ TC.makeLiteral(lit.substring(0, start))
						+ " starting "
						+ pattern;
			else
				// |X| > |Z|
				// return "Y jbef Z ending P"
				pattern =
					"just before "
						+ TC.makeLiteral(lit.substring(end))
						+ " ending "
						+ pattern;

			if (start == end)
				pattern = "point " + pattern;
			else
				pattern =
					TC.makeLiteral(lit.substring(start, end)) + " " + pattern;

			return pattern;
		}
	}

	//
	// Regular patterns (repeated the same number of times in every record)
	//

	Enumeration generalizeCountedPattern(Vector patterns, final Examples ex) {
		if (!ex.isAllPoints()) {
			return new SmallNFirstEnumeration(new FilteredEnumeration(
				patterns
				.elements()) {
				public void transform(Object o) {
					RegularFeature f = (RegularFeature) o;
					Region r = ex.getOneExample(f);
					if (f.intersects(r))
						f.yield(this, r);
				}
			});
		} else {
			return new SmallNFirstEnumeration(new FilteredEnumeration(
				patterns
				.elements()) {
				boolean start;
				public void yield(Object o) {
					NthFeature g = (NthFeature) o;
					g =
						(start)
							? new NthFeature(
								g.startOf(),
								g.n,
								"point just before " + g)
							: new NthFeature(
								g.endOf(),
								g.n,
								"point just after " + g);

					super.yield(g);
				}
				public void transform(Object o) {
					RegularFeature f = (RegularFeature) o;
					Region startsR = ex.getStartsOneExample(f);
					Region endsR = ex.getEndsOneExample(f);
					if (f.intersects(startsR)) {
						start = true;
						f.yield(this, startsR);
					}

					if (f.intersects(endsR)) {
						start = false;
						f.yield(this, endsR);
					}
				}
			});
		}
	}

	class RegularFeature extends Feature {
		public RegularFeature(Feature f) {
			super(f, f.toString());
		}

		public void yield(FilteredEnumeration enum, Region r) {
			int[] result = new int[2];
			if (!getIndexOf(r, this, result)) {
				SimulEdit.debug.println("Failed to find ordinum of " + r + "!");
				return;
			}

			enum.yield(new NthFeature(this, result[0]));
		}
	}

	class NthFeature extends Feature implements Counted {
		public int n;
		public NthFeature(Feature f, int n) {
			this(f.nth(n, inRecords), n, toNth(n) + " " + f + " in " + records);
		}
		public NthFeature(RegionSet matches, int n, String description) {
			super(matches, description);
			this.n = n;
		}
		public int getN() {
			return n;
		}
	}

	static String toNth(int i) {
		if (i == -1)
			return "last";

		if (i < 0)
			return toNth(-i) + " from last";

		if (((i / 10) % 10) != 1)
			switch (i % 10) {
				case 1 :
					return i + "st";
				case 2 :
					return i + "nd";
				case 3 :
					return i + "rd";
			}

		return i + "th";
	}

	boolean getIndexOf(Region r, RegionSet matches, int[] iAndN) {
		// get record where r occurs
		RegionEnumeration re = inRecords.regions(r);
		Region inRec = re.firstFast();
		if (inRec == null)
			return false;

		if (r.getStartLow() == inRec.getEndHigh()) {
			Region next = re.nextFast();
			if (next != null)
				inRec = next;
		}

		// r is the ith occurrence of the pattern in this
		// record; find i
		int i = 0;
		int n = 0;
		re = matches.regions(inRec);
		for (Region s = re.firstFast(); s != null; s = re.nextFast()) {
			++n;
			if (s.mightIntersect(r))
				i = n;
		}

		iAndN[0] = i;
		iAndN[1] = n;
		return i > 0;
	}

	//
	// Varying patterns (repeated a varying number of times in every record)
	//

	class VaryingFeature extends RegularFeature {
		int min; // min number of occurences in records
		public VaryingFeature(Feature f, int min) {
			super(f);
			this.min = min;
		}

		public void yield(FilteredEnumeration enum, Region r) {
			int[] result = new int[2];
			if (!getIndexOf(r, this, result)) {
				SimulEdit.debug.println(
					r
						+ " matched varying pattern "
						+ this
						+ ", with "
						+ result[1]
						+ " hits in this record, but couldn't determine its index");
				return;
			}

			int i = result[0];
			int n = result[1];

			// first try "ith"
			if (i <= min)
				enum.yield(new NthFeature(this, i));

			// now try "n-i+1th from last"
			int k = n - i + 1;
			if (k <= min)
				enum.yield(new NthFeature(this, -k));
		}
	}

	//
	// Generalizing a selection by describing its two endpoints separately
	//

	PairEnumeration generalizeEndpoints(
		final Examples ex,
		Generalizer g,
		PairEnumeration lastEndpoints) {
		return new PairEnumeration(
			g.generalize(ex.startOf()),
			g.generalize(ex.endOf()),
			lastEndpoints) {
			int[] minmax = new int[2];
			public void transform(Object o1, Object o2) {
				Feature fStart = (Feature) o1;
				Feature fEnd = (Feature) o2;

				// TODO: support multiple documents
                
                // minwu
				//debug.assertion(fStart.getVersions().size() == 1);
				//DocumentVersion fStartver =
				//	(DocumentVersion) (fStart.getVersions().toArray())[0];
				//debug.assertion(fEnd.getVersions().size() == 1);
				//DocumentVersion fEndver =
				//	(DocumentVersion) (fEnd.getVersions().toArray())[0];
                
                debug.assertion(fStart.getVersions().getSize() == 1);
                debug.assertion(fEnd.getVersions().getSize() == 1);
                
                DocumentVersion fStartver = null;
                DocumentVersion fEndver = null;
                
                Iterator istartver = fStart.getVersions().versions();
                while (istartver.hasNext())
                    fStartver = (DocumentVersion)istartver.next();
                Iterator iendver = fEnd.getVersions().versions();
                while (iendver.hasNext())
                    fEndver = (DocumentVersion)iendver.next();

				String pattern = "from " + fStart + "\n  to " + fEnd;
				debug.println("considering " + pattern);

				// figure out what document version to return
				DocumentVersion ver;
				CoordinateMap stmap, endmap;
				if (fStartver == fEndver) {
					ver = fStartver;
					stmap = endmap = IdentityMap.IDENTITY;
				} else {
					ver = ex.getVersion();
					stmap = ver.diff(fStartver);
					endmap = ver.diff(fEndver);
				}
				MutableRegionSet matches = new MutableRegionSet();
				RegionEnumeration starts = fStart.regions();
				RegionEnumeration ends = fEnd.regions();
				debug.println(
					fStart.getSize() + " starts, " + fEnd.getSize() + " ends");

				for (Region st = starts.firstFast(), en = ends.firstFast();
					st != null && en != null;
					st = starts.nextFast(), en = ends.nextFast()) {
					int s = stmap.mapStart(st.getStartLow());
					int e = endmap.mapEnd(en.getEndHigh());
					if (s <= e)
						matches.insert(s, e, ver);
					else {
						debug.println(
							"doesn't occur in every record: " + s + " > " + e);
						return;
					}
				}
				yield(new Feature(matches, pattern));
			}
		};
	}

}
