/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.parsers;

import lapis.*;

/**
 * Character class parser. Generates the following character classes:
 * <UL>
 * <LI> Whitespace
 *     <UL>
 *      <LI>  Spaces
 *      <LI>  Linebreak
 *      <LI>  Tab
 *     </UL>
 * <LI> Token 
 *      <UL>
 *     <li>Alphanumeric
 *      <UL>
 *      <LI> Letters
 *           <UL>
 *           <LI> UppercaseLetters
 *           <LI> LowercaseLetters
 *           </UL>
 *      <LI> Digits
 *      </UL>
 *    <LI> Punctuation
 *    </UL>
 * </UL>
 */

public class CharacterParser implements Parser {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    private MutableRegionSet uppercaseLetters;
    private MutableRegionSet lowercaseLetters;
    private MutableRegionSet digits;
    private MutableRegionSet punctuation;
    private MutableRegionSet spaces;
    private MutableRegionSet linebreak;
    private MutableRegionSet tab;
    private MutableRegionSet letters;
    private MutableRegionSet alphanumeric;
    private MutableRegionSet token;
    private MutableRegionSet whitespace;

    private final static int UPPERCASE_CODE  = 0;
    private final static int LOWERCASE_CODE  = 1;
    private final static int LETTER_CODE     = 2;
    private final static int DIGIT_CODE      = 3;
    private final static int PUNCTUATION_CODE= 4;
    private final static int SPACE_CODE      = 5;
    private final static int LINEBREAK_CODE  = 6;
    private final static int TAB_CODE  = 7;

    private static final int PROGRESS_INTERVAL = 1000;

    private static final char NON_BREAKING_SPACE = 160;
    
    // the document that this parser is parsing
    private Document pdoc;

    private void parse (String content) {
    	debug.assertion(pdoc != null);
    	
        int len = content.length ();

        uppercaseLetters = MutableRegionSet.makeFlattener (len / 10, true, pdoc);
        lowercaseLetters = MutableRegionSet.makeFlattener (len / 4, true, pdoc);
        digits = MutableRegionSet.makeFlattener (len / 10, true, pdoc);
        punctuation = MutableRegionSet.makeFlattener (len / 10, true, pdoc);
        spaces = MutableRegionSet.makeFlattener (len / 4, true, pdoc);
        linebreak = MutableRegionSet.makeFlattener (len / 10, false, pdoc);
        tab = MutableRegionSet.makeFlattener (len / 10, false, pdoc);
        letters = MutableRegionSet.makeFlattener (len / 4, true, pdoc);
        alphanumeric = MutableRegionSet.makeFlattener (len / 4, true, pdoc);
        token = MutableRegionSet.makeFlattener (len / 4, true, pdoc);
        whitespace = MutableRegionSet.makeFlattener (len / 4, true, pdoc);

        if (len > 0) {
            int start = 0;
            int charType = charClassOf (content.charAt (0));
            
            for (int i=1; i < len; ++i) {
                if (i % PROGRESS_INTERVAL == 0)
                    Main.getProgress ().fireAddProgress (PROGRESS_INTERVAL);
            
                char c = content.charAt (i);
                int newType = charClassOf (c);
                if (charType != newType 
                    || newType == LINEBREAK_CODE
                    || newType == TAB_CODE) {
                    add (charType, start, i);
                    start = i;
                    charType = newType;
                }
            }

            // add last run
            add (charType, start, len);
        }
    }

    private static int charClassOf (char c) {
        switch (Character.getType (c)) {
        case Character.UPPERCASE_LETTER:
            return UPPERCASE_CODE;
        case Character.LOWERCASE_LETTER:
            return LOWERCASE_CODE;
        case Character.DECIMAL_DIGIT_NUMBER:
            return DIGIT_CODE;
        default:
            if (Character.isLetter (c))
                return LETTER_CODE;
            else if (Character.isWhitespace (c) || c == NON_BREAKING_SPACE)
                switch (c) {
                case '\n':
                    return LINEBREAK_CODE;
                case '\t':
                    return TAB_CODE;
                default:
                    return SPACE_CODE;
                }
            else if (Character.isDigit (c))
                return DIGIT_CODE;
            else
                return PUNCTUATION_CODE;
        }
    }

    private void add (int charType, int start, int end) {
    	debug.assertion (pdoc != null);
        switch (charType) {
        case UPPERCASE_CODE:
            uppercaseLetters.insert (start, end, pdoc);
            letters.insert (start, end, pdoc);
            alphanumeric.insert (start, end, pdoc);
            token.insert (start, end, pdoc);
            break;
        case LOWERCASE_CODE:
            lowercaseLetters.insert (start, end, pdoc);
            letters.insert (start, end, pdoc);
            alphanumeric.insert (start, end, pdoc);
            token.insert (start, end, pdoc);
            break;
        case LETTER_CODE:
            letters.insert (start, end, pdoc);
            alphanumeric.insert (start, end, pdoc);
            token.insert (start, end, pdoc);
            break;
        case DIGIT_CODE:
            digits.insert (start, end, pdoc);
            alphanumeric.insert (start, end, pdoc);
            token.insert (start, end, pdoc);
            break;
        case LINEBREAK_CODE:
            linebreak.insert (start, end, pdoc);
            whitespace.insert (start, end, pdoc);
            break;
        case TAB_CODE:
            tab.insert (start, end, pdoc);
            whitespace.insert (start, end, pdoc);
            break;
        case SPACE_CODE:
            spaces.insert (start, end, pdoc);
            whitespace.insert (start, end, pdoc);
            break;
        case PUNCTUATION_CODE:
            punctuation.insert (start, end, pdoc);
            token.insert (start, end, pdoc);
            break;
        default:
            throw new RuntimeException ("illegal character type: " + charType);
        }
    }


    //
    // Pattern stubs for each name defined by this parser.
    //
    static class PatternStub implements Pattern {
        String name;

        public PatternStub (String name) {
            this.name = name;
        }

        public RegionSet match (Document doc) {
            PatternCache cache = doc.getPatternCache ();
            if (!cache.wasParsedBy ("lapis.parsers.CharacterParser"))
                new CharacterParser ().parse (doc);
            RegionSet s = cache.get (name);
            if (s == null) {
                debug.println ("CharacterParser ran but didn't create " + this);
                s = RegionSet.EMPTY;
            }
            return s;
        }

        public String toString () {
            return name;
        }
    }

    static PatternStub TOKEN = new PatternStub ("Characters.Token");
    static PatternStub ALPHANUMERIC = new PatternStub ("Characters.Token.Alphanumeric");
    static PatternStub DIGITS = new PatternStub ("Characters.Token.Alphanumeric.Digits");
    static PatternStub LETTERS = new PatternStub ("Characters.Token.Alphanumeric.Letters");
    static PatternStub UPPERCASELETTERS = new PatternStub ("Characters.Token.Alphanumeric.Letters.UpperCaseLetters");
    static PatternStub LOWERCASELETTERS = new PatternStub ("Characters.Token.Alphanumeric.Letters.LowerCaseLetters");
    static PatternStub PUNCTUATION = new PatternStub ("Characters.Token.Punctuation");
    static PatternStub WHITESPACE = new PatternStub ("Characters.Whitespace");
    static PatternStub SPACES = new PatternStub ("Characters.Whitespace.Spaces");
    static PatternStub LINEBREAK = new PatternStub ("Characters.Whitespace.Linebreak");
    static PatternStub TAB = new PatternStub ("Characters.Whitespace.Tab");
    static PatternStub RAWWHITESPACE = new PatternStub ("Characters._RawWhitespace");

    //
    // Bind pattern stubs to names.
    //
    public void bind (PatternLibrary lib) {
        debug.println ("binding Character patterns");
        lib.put (TOKEN.name, TOKEN);
        lib.put (ALPHANUMERIC.name, ALPHANUMERIC);
        lib.put (DIGITS.name, DIGITS);
        lib.put (LETTERS.name, LETTERS);
        lib.put (UPPERCASELETTERS.name, UPPERCASELETTERS); 
        lib.put (LOWERCASELETTERS.name, LOWERCASELETTERS);
        lib.put (PUNCTUATION.name, PUNCTUATION);
        lib.put (WHITESPACE.name, WHITESPACE);
        lib.put (SPACES.name, SPACES);
        lib.put (LINEBREAK.name, LINEBREAK);
        lib.put (TAB.name, TAB);
        lib.put (WHITESPACE.name, WHITESPACE);
        lib.put (RAWWHITESPACE.name, RAWWHITESPACE);
    }

    public void bind (PatternLibrary lib, Document doc) {
        // no doc-specific patterns for this parser
    }
    
    //
    // Scan document for character classes, returning the character
    // class represented by pat (which must be one of the stub patterns
    // listed above).  As long as we're scanning the document, we cache
    // all the other stub patterns in the document's pattern cache, so that
    // one scan suffices for any of the patterns.
    //
    public void parse (Document doc) {
    	pdoc = doc;
        long start = System.currentTimeMillis ();
        try {
            PatternCache cache = doc.getPatternCache ();

            DocumentView cookedView = doc.getCookedView();
            CoordinateMap cookedToRaw = doc.getVersion().diff (cookedView.getVersion());

            if (cookedToRaw != IdentityMap.IDENTITY) {
                String cooked = cookedView.getText();
                String raw = doc.getText ();

                Main.getProgress ().fireStartProgress ("Characters", 
                                       cooked.length () + raw.length ());
        
                parse (cooked);

                cache.put (TOKEN.name, token.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (ALPHANUMERIC.name, alphanumeric.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (DIGITS.name, digits.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (LETTERS.name, letters.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (UPPERCASELETTERS.name, uppercaseLetters.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (LOWERCASELETTERS.name, lowercaseLetters.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (PUNCTUATION.name, punctuation.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (WHITESPACE.name, whitespace.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (SPACES.name, spaces.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (LINEBREAK.name, linebreak.getSingleDocRegionSet(doc).convert (cookedToRaw));
                cache.put (TAB.name, tab.getSingleDocRegionSet(doc).convert (cookedToRaw));
            
                parse (raw);

                cache.put (RAWWHITESPACE.name, whitespace);
            }
            else {
                String content = doc.getText ();

                Main.getProgress ().fireStartProgress ("Characters", 
                                                       content.length ());
        
                parse (content);
  
                cache.put (TOKEN.name, token);
                cache.put (ALPHANUMERIC.name, alphanumeric);
                cache.put (DIGITS.name, digits);
                cache.put (LETTERS.name, letters);
                cache.put (UPPERCASELETTERS.name, uppercaseLetters);
                cache.put (LOWERCASELETTERS.name, lowercaseLetters);
                cache.put (PUNCTUATION.name, punctuation);
                cache.put (WHITESPACE.name, whitespace);
                cache.put (RAWWHITESPACE.name, whitespace);
                cache.put (SPACES.name, spaces);
                cache.put (LINEBREAK.name, linebreak);
                cache.put (TAB.name, tab);

            }
			
            cache.setParsedBy ("lapis.parsers.CharacterParser");
        } finally {
            Main.getProgress ().fireDoneProgress ();
            debug.println ("CharacterParser ran in "
                           + (System.currentTimeMillis () - start)
                           + " ms");
        }
    }
}
