/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

 
package lapis.parsers;

import lapis.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.io.*;
import lapisx.io.TeeStream;
import lapisx.io.NullOutputStream;

import org.apache.xerces.parsers.*;
import org.xml.sax.*;
import org.w3c.dom.*;


/**
 * ExternalParser
 *
 */
public class ExternalParser implements lapis.Parser {

    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;
    //public static java.io.PrintStream debug = debug;

    public static void loadExternalParser(String commandStr, String patternCommandStr, String osName) {
        loadExternalParser(commandStr, patternCommandStr, osName, "ExternalParser");
    }

    public static void loadExternalParser(String commandStr, String patternCommandStr, String osName, String type) {
	debug.println("loading external parser for " + osName + " ... " + commandStr);

	//disable external parsers for Mac (for now)
	if (osName.indexOf("Mac") != -1)
	    return;

	try {
	    PatternLibrary lib = PatternLibrary.getGlobalLibrary();
	    Command patternCmd = getCommand(patternCommandStr);

	    //get patterns from XML
	    HashMap patternMap = getPatterns(patternCmd);
	    Command command = getCommand(commandStr);
	    ExternalParser externalParser = null;
	    if ("SentenceCooker".equals(type)) {
		externalParser = new SentenceCooker(commandStr, command, patternMap);
	    }
	    else {
		externalParser = new ExternalParser(commandStr, command, patternMap);
	    }
	    lib.addParser (externalParser);
	}
	catch (Exception e) {
	    debug.report (e);
	}
    }
    
    private static Command getCommand(String commandStr) {
	StringTokenizer st = new StringTokenizer(commandStr, " ");
	ArrayList tokens = new ArrayList();
	while (st.hasMoreTokens()) {
	    tokens.add(st.nextToken());
	}

	String[] args = new String[tokens.size()];
	for (int i=0; i<tokens.size(); i++) {
	    args[i] = (String)tokens.get(i);
	}
	return new Command(args);

    }

    ///////////////////////////// INSTANCE DATA ///////////////////////////

    private String name = null;
    private HashMap sets = null; //map of MutableRegionSet objects
    private ArrayList patternStubs = null; //list of PatternStub objects
    private HashMap patternMap = null; //state => Business.Address.State
    private Command command = null;

	// the document that this parser is parsing
	private lapis.Document pdoc;

    ///////////////////////////// CONSTRUCTORS  ///////////////////////////

    public ExternalParser (String name, Command command, HashMap patternMap) {
	this.name = name;
	this.command = command;
	this.patternMap = patternMap;

	patternStubs = new ArrayList();
	
	String tempAbbreviation = null;
	String tempPatternName = null;
	Iterator patterns = patternMap.keySet().iterator();
	while (patterns.hasNext()) {
	    tempAbbreviation = (String)patterns.next();
	    tempPatternName = (String)patternMap.get(tempAbbreviation);
	    PatternStub p = new PatternStub(name, tempAbbreviation, tempPatternName, this);
	    patternStubs.add(p);
	}

    }




    ///////////////////////////// METHODS  ///////////////////////////


    static class PatternStub implements Pattern {
	private String name = null;
	private String abbr = null;	
	private String patternName = null;
	private ExternalParser externalParser = null;
	
	public PatternStub (String name, String abbr, String patternName, ExternalParser externalParser) {
	    this.name = name;
	    this.abbr = abbr;
	    this.patternName = patternName;
	    this.externalParser = externalParser;
	}

	public RegionSet match (lapis.Document doc) {
	    PatternCache cache = doc.getPatternCache();
	    if (!cache.wasParsedBy(name)) {
	        externalParser.parse(doc);
	    }
	    RegionSet s = cache.get(patternName);
	    if (s == null) {
		debug.println("ExternalParser ran but didn't create patterns for " + patternName);
		s = RegionSet.EMPTY;
	    }
	    return s;
	}	
	public String toString() {
	    return name;
	}
    }


    public void bind(PatternLibrary lib) {       
	for (int i=0; i<patternStubs.size(); i++) {
	    PatternStub p = (PatternStub)patternStubs.get(i);
	    lib.put(p.patternName, p);
	}
	debug.println("done binding...");
    }

    public void bind (PatternLibrary lib, lapis.Document doc) {
        // no doc-specific patterns for this parser
    }


    /**
     * Parses XML document passed as InputSource.  Returns a DOM document.
     */
    private static org.w3c.dom.Document getDocument(InputSource is) {
	org.w3c.dom.Document doc = null;
	try {
	    DOMParser p = new DOMParser();

	    //THESE features and properties are used to turn on VALIDATION OF XML SCHEMA
	    //p.setFeature("http://xml.org/sax/features/validation", true);  
	    //p.setFeature("http://apache.org/xml/features/validation/schema", true );
	    //p.setFeature("http://apache.org/xml/features/validation/schema-full-checking", true );
	    //String dir = Main.getLapisDir() + "/lib/lapis.xsd";
	    //debug.println("DIR: " + dir);
	    //p.setProperty("http://apache.org/xml/properties/schema/external-noNamespaceSchemaLocation",
	    //		  "lapis.xsd");

	    long sys = System.currentTimeMillis();
	    p.parse(is);
	    doc = p.getDocument();
	    debug.println("XML parsing ran in " + (System.currentTimeMillis() - sys));
	}
	catch (Exception e) {
	    debug.println("could not parse XML.");
	    //	    e.printStackTrace();
	}
	return doc;
    }



    private static HashMap getPatterns(Command cmd) {
	HashMap map = new HashMap();
	try {
	    org.w3c.dom.Document doc = 
		getDocument(new InputSource(cmd.exec()));
	    if (doc != null) {
		NodeList nodeList = doc.getElementsByTagName("pattern"); //<regionset>
		for (int i=0; i<nodeList.getLength(); i++) {
		    org.w3c.dom.Node temp = nodeList.item(i);
		    NamedNodeMap nnmap = temp.getAttributes();
		    org.w3c.dom.Node attr = nnmap.getNamedItem("class");
		    NodeList children = temp.getChildNodes();
		    org.w3c.dom.Node child = children.item(0);
		    map.put(attr.getNodeValue(), child.getNodeValue());		
		}
	    }
	}
	catch(Exception e) {
	    debug.println("Failed to Load " + cmd);
        debug.report (e);
	}
	return map;
    }


    private void parse(String content) {
	InputStream output = new BufferedInputStream(command.exec(content), 8192);
	
	TeeStream tee = null;
	if (debug.printEnabled()) {
	    tee = new TeeStream(output, debug.getOutputStream());	    
	}
	else {
	    tee = new TeeStream(output, NullOutputStream.getInstance());
	}

	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	pipe(output, baos);
	ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray());

	debug.println("************ XML ***********");
	try {
	    debug.println("creating document...");
	    org.w3c.dom.Document doc = 
		getDocument(new InputSource(bais)); //(tee/bais)
	    
	    debug.println("document created.");
	    if (doc != null) {
		NodeList nodeList = doc.getElementsByTagName("regionset"); //<regionset>
		
		for (int i=0; i<nodeList.getLength(); i++) {
		    org.w3c.dom.Node temp = nodeList.item(i);
		    NamedNodeMap nnmap = temp.getAttributes();
		    org.w3c.dom.Node attr = nnmap.getNamedItem("class");
		    
		    NodeList children = temp.getChildNodes();
		    org.w3c.dom.Node child = children.item(0);
		    
		    String patternStr = (String)patternMap.get(attr.getNodeValue());
		    
		    if (patternStr != null) {
			MutableRegionSet set = 
			    (MutableRegionSet)sets.get(patternStr);
			if (set == null)
			set = new MutableRegionSet();
			putHelper(set, child.getNodeValue());
			sets.put(patternStr, set);
		    }
		}
	    }
	}
	catch (Exception e) {
	    debug.println("EXCEPTION: " + e.getMessage());
	    //	    e.printStackTrace();
	}

    }

    private void putHelper(MutableRegionSet set, String regionsets) {
    debug.assertion(pdoc!=null);
	StringTokenizer st = new StringTokenizer(regionsets,"[], \n\t");
	while (st.hasMoreTokens()) {
	    set.insert(Integer.parseInt(st.nextToken()),
		       Integer.parseInt(st.nextToken()), pdoc);
	}
    }


    public void parse(lapis.Document doc) {
    pdoc=doc;
	long start = System.currentTimeMillis();

	sets = new HashMap(); //state => RegionSet

	try {
	    PatternCache cache = doc.getPatternCache();
	    DocumentView view = cook (doc);
        CoordinateMap map = view.getVersion().diff (doc.getVersion());
        String content = view.getText();
	    Main.getProgress().fireStartProgress("External Parser", content.length());		
	    parse(content);		
	    String patternStr = null;
	    Iterator iterator = sets.keySet().iterator();
	    while (iterator.hasNext()) {
    	   patternStr = (String)iterator.next();
    	   RegionSet converted = ((RegionSet)sets.get(patternStr)).getSingleDocRegionSet(doc).convert(map.invert());
    	   cache.put(patternStr, converted);
	    }
	    
	    
	    /**
	    if (cdoc != null) {
		debug.println("*** COOKED VERSION...");
		CoordinateMap cookedToRaw = doc.getVersion().diff(cdoc.getVersion());
		lapis.Document doc2 = cook(cdoc);
		CoordinateMap map = cdoc.getVersion().diff(doc2.getVersion());
		String content = doc2.getText();
		Main.getProgress().fireStartProgress("External Parser", content.length());		
		parse(content);		
		String patternStr = null;
		Iterator iterator = sets.keySet().iterator();
		while (iterator.hasNext()) {
		    patternStr = (String)iterator.next();
		    RegionSet converted = 
			(((RegionSet)sets.get(patternStr)).convert(map)).convert(cookedToRaw);
		    cache.put(patternStr, converted);
		}
	    }
	    else {
		lapis.Document doc2 = cook(doc);
		CoordinateMap map = doc.getVersion().diff(doc2.getVersion());
		String content = doc2.getText();
		Main.getProgress().fireStartProgress("External Parser", content.length());		
		parse(content);
		String patternStr = null;
		Iterator iterator = sets.keySet().iterator();
		while (iterator.hasNext()) {
		    patternStr = (String)iterator.next();
		    debug.println("PATTERN: " + patternStr);
		    debug.println("converting ...");
		    RegionSet converted = ((RegionSet)sets.get(patternStr)).convert(map);
		    debug.println("putting in pattern cache ...");
		    cache.put(patternStr, converted);
		}
	    }
	    **/
            cache.setParsedBy (name);//"lapis.parsers.ListParser");
	}
	finally {
	    Main.getProgress().fireDoneProgress();
	    debug.println("ExternalParser ran in " + 
			       (System.currentTimeMillis() - start)
			       + " ms");
	}
    }//end parse(Document)



    public lapis.DocumentView cook(lapis.Document doc) {
	return doc.getDefaultView();
    }



    public static class Command {
	String[] cmd = null;

	public Command(String[] cmd) {
	    this.cmd = cmd;
	}

	public InputStream exec() {
	    return this.exec("");
	}

	public InputStream exec(String content) {
	    String output = "";
	    InputStream is = null;
	    try {
		final Process process = Runtime.getRuntime().exec(cmd);

	    final ByteArrayOutputStream baos = new ByteArrayOutputStream();
	    
	    Thread stdoutThread = new Thread(new Runnable() {
		    public void run() {
			try {
			    InputStream br_in = new BufferedInputStream(process.getInputStream());
			    String buff = null;
			    int amtRead = -1;
			    byte[] buffer = new byte[128000];
			    while ((amtRead = br_in.read(buffer,0,buffer.length)) != -1) {
				debug.println("Process out >>>" + amtRead);
				baos.write(buffer, 0, amtRead);
				//				try {Thread.sleep(0); } catch(Exception e) {}
			    }
			    br_in.close();
			} catch (IOException ioe) {
			    debug.println("Exception caught printing javac result");
			    ioe.printStackTrace();
			}
		    }
		});
	    stdoutThread.start();

	    Thread stderrThread = new Thread(new Runnable() {
		    public void run() {
			try {
			    InputStream br_err = new BufferedInputStream(process.getErrorStream());
			    int amtRead = -1;
			    byte[] buffer = new byte[128000];
			    while ((amtRead = br_err.read(buffer,0,buffer.length)) != -1) {
                                if (debug.printEnabled ())
                                    debug.getOutputStream ().write (buffer, 0, amtRead);
				//				debug.println("Process err >>>" + buff);
				//				try {Thread.sleep(100); } catch(Exception e) {}
			    }
			    br_err.close();
			} catch (IOException ioe) {
			    debug.println("Exception caught printing javac result");
			    ioe.printStackTrace();
			}
		    }
		});
	    stderrThread.start();

	    OutputStream os = process.getOutputStream();
	    byte[] bytes = content.getBytes();
	    BufferedOutputStream bos = new BufferedOutputStream(os);
	    bos.write(bytes, 0, bytes.length);
	    bos.flush();
	    bos.close();

	    debug.println("waiting...\n");

	    try {
		stdoutThread.join();
	    }
	    catch (Exception e) {
		debug.println("EXCEPTION: " + e.getMessage());
	    }
	    int result = process.waitFor();

	    ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray());
	    //	    baos.close();
	    debug.println("parsing...\n");
	    return bais;
	    }
	    catch (Exception e) {
		debug.println("EXCEPTION: " + e.getMessage());
		//		e.printStackTrace();
	    }
	    return is;
	}

    }





    private static final int _BUFFER_SIZE = 100000;
    private static byte[] _buffer = new byte[_BUFFER_SIZE];
    
    public static void pipe(InputStream in_, OutputStream out_) {
	debug.println("piping...");
        if ( in_ != null && out_ != null ){
	    try {
		int amountRead = 0;
		while ( amountRead != -1 ){
		    synchronized(_buffer) {
			amountRead = in_.read(_buffer);
			debug.println("amount read...");
			if ( amountRead != -1 ){
			    out_.write(_buffer, 0, amountRead);
			}
		    }
		}
	    } catch ( IOException ex ){
		ex.printStackTrace();
	    }
        }
    }




}//end class

