/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.parsers;

import lapis.*;

/**
 * HTMLElement in an HTML page.  An element runs from a start tag
 * (like &lt;ul&gt;) to its matching end tag (&lt;/ul&gt;),
 * inclusive.
 * An element may have an optional end tag (like &lt;p&gt;),
 * in which case the element runs up to (but not including)
 * the tag that implicitly closes it.  For example:
 * <PRE>&lt;p&gt;Paragraph 1&lt;p&gt;Paragraph 2</PRE>
 * contains two elements, <PRE>&lt;p&gt;Paragraph 1</PRE>
 * and <PRE>&lt;p&gt;Paragraph 2</PRE>.
 */
public class HTMLElement extends Region {

    public HTMLTag startTag;
    public HTMLTag endTag;
    public boolean empty;

    public HTMLElement sibling; // next sibling
    public HTMLElement parent;
    public HTMLElement child;   // first child

    /**
     * Make an HTMLElement from a start tag and an end position.  Used
     * when the end tag has been omitted (like &lt;p&gt;, frequently).
     * @param startTag Start tag of element
     * @param end Ending offset of element
     * @param ver Document version
     */
    /*
    public HTMLElement (HTMLTag startTag, int end) {
        super (startTag.getStart (), end);
        this.startTag = startTag;
        this.endTag = null;
        this.empty = false;
    }
    */
    public HTMLElement (HTMLTag startTag, int end, DocumentVersion ver) {
    	super (startTag.getStart (), end, ver);
    	this.startTag = startTag;
    	this.endTag = null;
    	this.empty = false;
    }
    
    public HTMLElement (HTMLTag startTag, int end, Document doc) {
    	this(startTag, end, doc.getVersion()); 
    }

    public void setEnd (int end) {
        this.end = end;
    }

    /**
     * Get tag name.
     * @return tag name (like "P"), in upper-case, String.intern()'ed form.
     * Thus you can compare tag names with ==, as in:
     * <CODE>getTagName() == HTMLTag.IMG</CODE>.
     */
    public String getTagName () {
        return startTag.getTagName();
    }

    /**
     * Get start tag.
     * @return start tag of element
     */
    public HTMLTag getStartTag () {
        return startTag;
    }

    /**
     * Get end tag.
     * @return end tag of element, or null if element has no end tag.
     */
    public HTMLTag getEndTag () {
        return endTag;
    }

    /**
     * Get element's parent.
     * @return element that contains this element, or null if at top-level.
     */
    public HTMLElement getParent () {
        return parent;
    }

    /**
     * Get element's next sibling.
     * @return element that follows this element, or null if at end of
     * parent's children.
     */
    public HTMLElement getSibling () {
        return sibling;
    }

    /**
     * Get element's first child.
     * @return first element contained by this element, or null if no children.
     */
    public HTMLElement getChild () {
        return child;
    }

    /**
     * Return next element in an inorder walk of the tree,
     * assuming this element and its children have been visited.
     * @return next element
     */
    public HTMLElement getNext () {
        if (sibling != null)
            return sibling;
        else if (parent != null)
            return parent.getNext ();
        else
            return null;
    }

    /**
     * Test if tag has an HTML attribute.
     * @param name Label of HTML attribute (e.g. "HREF").  Doesn't have to be
     * converted with toHTMLAttributeName().
     * @return true if tag has the attribute, false if not
     */
    public boolean hasHTMLAttribute (String name) {
        return startTag.hasHTMLAttribute (name);
    }

    /**
     * Get an HTML attribute's value.
     * @param name Label of HTML attribute (e.g. "HREF").  Doesn't have to be
     * converted with toHTMLAttributeName().
     * @return value of attribute if it exists, TRUE if the attribute exists but has no value, or null if tag lacks the attribute.
     */
    public String getHTMLAttribute (String name) {
        return startTag.getHTMLAttribute (name);
    }

    /**
     * Get an HTML attribute's value, with a default value if it doesn't exist.
     * @param name Label of HTML attribute (e.g. "HREF").  Doesn't have to be
     * converted with toHTMLAttributeName().
     * @param defaultValue default value to return if the attribute
     * doesn't exist
     * @return value of attribute if it exists, TRUE if the attribute exists but has no value, or defaultValue if tag lacks the attribute.
     */
    public String getHTMLAttribute (String name, String defaultValue) {
        return startTag.getHTMLAttribute (name, defaultValue);
    }

    public void changeHTMLAttribute (String name, String newValue) {
        startTag.changeHTMLAttribute (name, newValue);
    }
}
