/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.parsers;


import lapis.*;
import java.io.IOException;
import java.io.Reader;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Stack;
import lapisx.progress.ProgressReader;

/**
 * HTML parser.  Parses an input stream or String and
 * converts it to a sequence of tags and a tree of elements.
 */
public class HTMLParser implements Parser {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    // size of parser's read buffer
    static final int BUFFER_SIZE = 1024;

    // parameter for automatic HTML type detection.  If !alwaysParse
    // and the parser doesn't encounter at least one known HTML 4.0 tag
    // in the first VALID_HTML_PREFIX bytes of the stream, then parser concludes
    // that the stream isn't HTML and stops parsing it (putting no region sets in
    // the document's pattern cache). Use multiples of BUFFER_SIZE.
    static final int VALID_HTML_PREFIX = BUFFER_SIZE;


    // when alwaysParse is true, then the parser always parses the entire stream,
    // regardless of whether it finds any valid tags or not.
    boolean alwaysParse = false;

    public HTMLParser () {
    }

    public boolean getAlwaysParse () {
        return alwaysParse;
    }

    public void setAlwaysParse (boolean f) {
        alwaysParse = f;
    }

    //
    // Pattern stubs for each name defined by this parser.
    //
    static class PatternStub implements Pattern {
        String name;

        public PatternStub (String name) {
            this.name = name;
        }

        public RegionSet match (Document doc) {
            PatternCache cache = doc.getPatternCache ();
            if (!cache.wasParsedBy ("lapis.parsers.HTMLParser"))
                new HTMLParser ().parse (doc);
            RegionSet s = cache.get (name);
            if (s == null) {
                debug.println ("HTMLParser ran but didn't create " + this);
                s = RegionSet.EMPTY;
            }
            return s;
        }

        public String toString () {
            return name;
        }
    }

    static PatternStub TEXT = new PatternStub ("HTML.Text");
    static PatternStub TAG = new PatternStub ("HTML.Tag");
    static PatternStub START_TAG = new PatternStub ("HTML.Tag.StartTag");
    static PatternStub END_TAG = new PatternStub ("HTML.Tag.EndTag");
    static PatternStub ELEMENT = new PatternStub ("HTML.Element");
    static PatternStub ATTRIBUTE = new PatternStub ("HTML.Attribute");
    static PatternStub ATTRIBUTE_NAME = new PatternStub ("HTML.Attribute.Name");
    static PatternStub ATTRIBUTE_VALUE = new PatternStub ("HTML.Attribute.Value");

    static Hashtable startTagPatterns;
    static Hashtable endTagPatterns;
    static Hashtable elementPatterns;
    static Hashtable attributePatterns;

    static {
        startTagPatterns = makeTagPatterns ("HTML.Tag.StartTag.<", ">");
        endTagPatterns = makeTagPatterns ("HTML.Tag.EndTag.</", ">");
        elementPatterns = makeTagPatterns ("HTML.Element.[", "]");
        attributePatterns = makeAttributePatterns ("HTML.Attribute.@", "");
    }

    static Hashtable makeTagPatterns (String prefix, String suffix) {
        Hashtable h = new Hashtable ();
        for (int i = 0; i < HTMLTag.ALL_TAGS.length; ++i) {
            String name = HTMLTag.ALL_TAGS[i];
            h.put (name, new PatternStub (prefix + name + suffix));
        }
        return h;
    }

    static Hashtable makeAttributePatterns (String prefix, String suffix) {
        Hashtable h = new Hashtable ();
        for (int i = 0; i < HTMLTag.ALL_ATTRIBUTES.length; ++i) {
            String name = HTMLTag.ALL_ATTRIBUTES[i];
            h.put (name, new PatternStub (prefix + name + suffix));
        }
        return h;
    }

    //
    // Bind names to pattern stubs in a pattern library.
    //
    public void bind (PatternLibrary lib) {
        lib.put (TEXT.name, TEXT);
        lib.put (TAG.name, TAG);
        lib.put (START_TAG.name, START_TAG);
        lib.put (END_TAG.name, END_TAG);
        lib.put (ELEMENT.name, ELEMENT);
        lib.put (ATTRIBUTE.name, ATTRIBUTE);
        lib.put (ATTRIBUTE_NAME.name, ATTRIBUTE_NAME);
        lib.put (ATTRIBUTE_VALUE.name, ATTRIBUTE_VALUE);

        bindPatterns (lib, startTagPatterns);
        bindPatterns (lib, endTagPatterns);
        bindPatterns (lib, elementPatterns);
        bindPatterns (lib, attributePatterns);
    }

    void bindPatterns (PatternLibrary lib, Hashtable patterns) {
        for (Enumeration e = patterns.elements (); e.hasMoreElements (); ) {
            PatternStub pat = (PatternStub) e.nextElement ();
            lib.put (pat.name, pat);
        }
    }

    public void bind (PatternLibrary lib, Document doc) {
        // no doc-specific patterns for this parser
    }

    /**
     * Parse a Document.
     * @param doc Document to parse
     */
    public void parse (Document doc) {
        this.doc = doc;

        long start = System.currentTimeMillis ();
        try {
            tokenize (doc.getReader (), null);
        } catch (IOException e) {
            throw new Error ("unexpected " + e);
        } finally {
            debug.println ("HTMLParser ran in "
                           + (System.currentTimeMillis () - start)
                           + " ms");
        }
    }

    /**
     * Parse a Document from a stream, simultaneously accumulating its content to a buffer.
     * @param doc Document to receive parsed HTML
     * @param reader stream of HTML to parse
     * @param content StringBuffer into which document contents are accumulated.
     */

    public void parse (Document doc, Reader reader, StringBuffer content) throws IOException {
        this.doc = doc;
        tokenize (reader, content);
    }


    /*
     * Fields describing document being parsed
     */
    protected Document doc;

    /*
     *  HTML tokenizer state machine
     */

    // state takes on one of the following values:
    private static final int START = 0;
    private static final int INWORD = 1;
    private static final int ENTITY = 2;
    private static final int LT = 4;
    private static final int BANG = 5;
    private static final int BANG_DASH = 6;
    private static final int CMT = 7;
    private static final int CMT_DASH = 8;
    private static final int CMT_DASHDASH = 9;
    private static final int DIRECTIVE = 10;
    private static final int STAG = 11;
    private static final int ETAG = 12;
    private static final int ATTR = 13;
    private static final int ATTRLABEL = 14;
    private static final int EQ = 15;
    private static final int AFTEREQ = 16;
    private static final int ATTRVAL = 17;
    private static final int ATTRVAL_SQ = 18;
    private static final int ATTRVAL_DQ = 19;
    private static final int DONE = 20;
    private static final int ENTNUM = 21;
    private static final int ENTREF = 22;
    private static final int PCDATA_ENTITY = 23;

    char[] buf = new char[BUFFER_SIZE];
    Vector attrs = new Vector ();
    StringBuffer entity = new StringBuffer ();
    boolean isHTML;

    private void tokenize (Reader stream, StringBuffer content) throws IOException {
        int nRead = 0;

        int state = START;
        boolean space = true;
        boolean inPreformatted = false;

        int bufptr = 0;
        int buflen = 0;
        int bufbase = 0;

        int tagStart = -1;
        int textStart = -1;
        String tagName = null;     // last tag name parsed
        HTMLTag tag = makeTag (0, 0, "", false);

        int attrNameStart = -1;
        int attrNameEnd = -1;
        int attrValStart = -1;
        String attrName = null;    // last attribute name parsed

        int entnum = 0;
        int postEntityState = 0;

        clearWord ();

        isHTML = alwaysParse;

        handleInitParse ();

        Main.getProgress ().fireStartProgress ("HTML", 100000);
        stream = new ProgressReader (stream, 1000, Main.getProgress ());
        try {

            parsing:
            while (true) {
                while (bufptr >= buflen) {
                    bufptr = 0;
                    bufbase += buflen;

                    if (!isHTML && bufbase >= VALID_HTML_PREFIX)
                        // we didn't see any known HTML tags in the first VALID_HTML_PREFIX bytes,
                        // so assume the document isn't HTML and stop parsing it.
                        return;

                    buflen = stream.read (buf);
                    if (buflen == -1)
                        break parsing;
                    Main.getProgress ().fireAddProgress (1);

                    buflen = removeCRs (buf, buflen);

                    if (content != null)
                        content.append (buf, 0, buflen);
                }

                char c = (char)buf[bufptr];

                //System.err.println ("%% state == " + state + ", ptr == " + (bufbase+bufptr) + ", c == " + c);

                switch (state) {
                case START:
                    // after whitespace or tag
                    switch (c) {
                    case '<':
                        endWord (bufbase + bufptr);
                        if (textStart != -1) {
                            addTextRun (textStart, bufbase + bufptr);
                            textStart = -1;
                        }
                        tagStart = bufbase+bufptr;
                        ++bufptr;
                        state = LT;
                        break;
                    case '&':
                        endWord (bufbase + bufptr);
                        handleStartWord (bufbase + bufptr);
                        ++bufptr;
                        postEntityState = PCDATA_ENTITY;
                        state = ENTITY;
                        break;
                    case ' ':
                    case '\t':
                    case '\r':
                    case '\n':
                        if (inPreformatted)
                            addWordChar (c);
                        else if (!space)
                            addWordChar (' ');
                        else
                            endWord (bufbase + bufptr);
                        space = true;
                        ++bufptr;
                        break;
                    default:
                        if (textStart == -1)
                            textStart = bufbase + bufptr;
                        if (wordptr == 0)
                            handleStartWord (bufbase + bufptr);
                        addWordChar (c);
                        space = false;
                        ++bufptr;
                        break;
                    }
                    break;

                case PCDATA_ENTITY:
                    endWord (bufbase + bufptr);
                    space = false;
                    state = START;
                    break;

                case LT:
                    switch (c) {
                    case '/':
                        ++bufptr;
                        clearWord (); // for tag name
                        state = ETAG;
                        break;
                    case '!':
                        ++bufptr;
                        state = BANG;
                        break;
                    default:
                        if (Character.isLetter (c)) {
                            clearWord (); // for tag name
                            state = STAG;
                        }
                        else {
                            addWordChar ('<');
                            space = false;
                            clearWord ();
                            state = START;
                        }
                        break;
                    }
                    break;

                    // Comments and directives.
                    // Implements the (broken, but easy) Netscape rule:
                    // <!-- starts a comment, --> closes.
                    // All other directives <!foo> are also returned as comments.
                case BANG:
                    if (c == '-') {
                        ++bufptr;
                        state = BANG_DASH;
                    }
                    else {
                        state = DIRECTIVE;
                    }
                    break;

                case BANG_DASH:
                    if (c == '-') {
                        ++bufptr;
                        state = CMT;
                    }
                    else {
                        state = DIRECTIVE;
                    }
                    break;

                case CMT:
                    if (c == '-') {
                        ++bufptr;
                        state = CMT_DASH;
                    }
                    else {
                        ++bufptr;
                    }
                    break;

                case CMT_DASH:
                    if (c == '-') {
                        ++bufptr;
                        state = CMT_DASHDASH;
                    }
                    else {
                        ++bufptr;
                        state = CMT;
                    }
                    break;

                case CMT_DASHDASH:
                    if (c == '>') {
                        ++bufptr;
                        tag = makeTag (tagStart, bufbase + bufptr, HTMLTag.COMMENT, true);
                        handleTag (tag);
                        clearWord ();
                        state = START;
                    }
                    else if (c == '-') {
                        ++bufptr;
                        state = CMT_DASHDASH;
                    }
                    else {
                        ++bufptr;
                        state = CMT;
                    }
                    break;

                case DIRECTIVE:
                    if (c == '>') {
                        ++bufptr;
                        tag = makeTag (tagStart, bufbase + bufptr, HTMLTag.COMMENT, true);
                        handleTag (tag);
                        clearWord ();
                        state = START;
                    }
                    else {
                        ++bufptr;
                    }
                    break;

                    // Tags
                case STAG:
                    if (c == '>' || Character.isWhitespace(c)) {
                        tagName = getWord ();
                        attrs.setSize (0);
                        state = ATTR;
                    }
                    else {
                        addWordChar (Character.toLowerCase (c));
                        ++bufptr;
                        // state == STAG;
                    }
                    break;

                case ETAG:
                    if (c == '>') {
                        ++bufptr;

                        tagName = getWord ();
                        tag = makeTag (tagStart, bufbase + bufptr, tagName, false);
                        handleTag (tag);
                        if (tag.isBlockTag ())
                            space = true;
                        if (inPreformatted &&
                            tag.isPreformatted ())
                            inPreformatted = false;
                        clearWord ();
                        state = START;
                    }
                    else {
                        addWordChar (Character.toLowerCase (c));
                        ++bufptr;
                        // state == ETAG
                    }
                    break;

                    // Attributes
                case ATTR:
                    if (Character.isWhitespace(c))
                        ++bufptr;
                    else if (c == '>') {
                        ++bufptr;

                        tag = makeTag (tagStart, bufbase + bufptr, tagName, true);
                        if (attrs.size() > 0) {
                            ((HTMLTag)tag).htmlAttributes = new String[attrs.size()];
                            attrs.copyInto (((HTMLTag)tag).htmlAttributes);
                        }
                        handleTag (tag);
                        if (tag.isBlockTag ())
                            space = true;
                        if (tag.isPreformatted ())
                            inPreformatted = true;
                        clearWord ();
                        state = START;
                    }
                    else {
                        clearWord ();   // for attribute name
                        attrNameStart = bufbase+bufptr;
                        state = ATTRLABEL;
                    }
                    break;

                case ATTRLABEL:
                    if (c == '>' || c == '=' || Character.isWhitespace(c)) {
                        attrName = HTMLTag.toHTMLAttributeName (getWord ());
                        attrNameEnd = bufbase+bufptr;
                        state = EQ;
                    }
                    else {
                        addWordChar (Character.toLowerCase (c));
                        ++bufptr;
                        // state == ATTRLABEL;
                    }
                    break;

                case EQ:
                    if (Character.isWhitespace(c))
                        ++bufptr;
                    else if (c == '=') {
                        ++bufptr;
                        state = AFTEREQ;
                    }
                    else {
                        handleAttribute (attrName, attrNameStart, attrNameEnd,
                                         null, attrNameEnd, attrNameEnd);
                        state = ATTR;
                    }
                    break;

                case AFTEREQ:
                    if (Character.isWhitespace (c))
                        ++bufptr;
                    else
                        switch (c) {
                        case '>':
                            {
                                handleAttribute (attrName, 
                                                 attrNameStart, attrNameEnd,
                                                 null, 
                                                 bufbase+bufptr, bufbase+bufptr);
                                state = ATTR;
                                break;
                            }
                        case '\'':
                            attrValStart = bufbase+bufptr;
                            ++bufptr;
                            clearWord ();   // for attribute value
                            state = ATTRVAL_SQ;
                            break;
                        case '"':
                            attrValStart = bufbase+bufptr;
                            ++bufptr;
                            clearWord ();   // for attribute value
                            state = ATTRVAL_DQ;
                            break;
                        default:
                            attrValStart = bufbase+bufptr;
                            clearWord ();   // for attribute value
                            state = ATTRVAL;
                            break;
                        }
                    break;

                case ATTRVAL:
                    if (c == '>' || Character.isWhitespace(c)) {
                        handleAttribute (attrName, attrNameStart, attrNameEnd,
                                         getWord(), attrValStart, bufbase+bufptr);
                        state = ATTR;
                    }
                    else if (c == '&') {
                        ++bufptr;
                        postEntityState = ATTRVAL;
                        state = ENTITY;
                    }
                    else {
                        ++bufptr;
                        addWordChar (c);
                        // state == ATTRVAL;
                    }
                    break;

                case ATTRVAL_SQ:
                    if (c=='\'') {
                        ++bufptr;
                        handleAttribute (attrName, attrNameStart, attrNameEnd,
                                         getWord(), attrValStart, bufbase+bufptr);
                        state = ATTR;
                    }
                    else if (c == '&') {
                        ++bufptr;
                        postEntityState = ATTRVAL_SQ;
                        state = ENTITY;
                    }
                    else {
                        ++bufptr;
                        addWordChar (c);
                        // state == ATTRVAL_SQ;
                    }
                    break;

                case ATTRVAL_DQ:
                    if (c=='"') {
                        ++bufptr;
                        handleAttribute (attrName, attrNameStart, attrNameEnd,
                                         getWord(), attrValStart, bufbase+bufptr);
                        state = ATTR;
                    }
                    else if (c == '&') {
                        ++bufptr;
                        postEntityState = ATTRVAL_DQ;
                        state = ENTITY;
                    }
                    else {
                        ++bufptr;
                        addWordChar (c);
                        // state == ATTRVAL_DQ;
                    }
                    break;

                    //  Entities
                case ENTITY:
                    if (c == '#') {
                        ++bufptr;
                        entnum = 0;
                        state = ENTNUM;
                    }
                    else if ((c >= 'A' && c <= 'Z')
                             || (c >= 'a' && c <= 'z')) {
                        entity.setLength (0);
                        state = ENTREF;
                    }
                    else {
                        addWordChar ('&');
                        state = postEntityState;
                    }
                    break;

                case ENTREF:
                    if (!Character.isLetterOrDigit(c)) {
                        Character ent = lookupEntityRef (entity.toString ());
                        if (ent != null) {
                            addWordChar (ent.charValue());
                            if (c == ';')
                                ++bufptr;
                        }
                        else {
                            // unrecognized entity -- leave
                            // as-is
                            addWordChar ('&');
                            for (int i=0, len=entity.length(); i < len; ++i)
                                addWordChar (entity.charAt (0));
                        }
                        state = postEntityState;
                    }
                    else {
                        ++bufptr;
                        entity.append (c);
                        // state == ENTREF;
                    }
                    break;

                case ENTNUM:
                    if (c==';' || !Character.isDigit(c)) {
                        addWordChar ((char) entnum);
                        if (c == ';')
                            ++bufptr;
                        state = postEntityState;
                    }
                    else {
                        entnum = 10*entnum + (c - '0');
                        ++bufptr;
                    }
                    break;

                default:
                    throw new RuntimeException ("HTMLParser entered illegal state " + state);
                }
            }

            endWord (bufbase + bufptr);
            if (textStart != -1)
                addTextRun (textStart, bufbase + bufptr);

            // close any remaining open elements
            closeAll (bufbase + bufptr);

            handleDoneParse (bufbase + bufptr);

        } finally {
            Main.getProgress ().fireDoneProgress ();
            doc.getPatternCache ().setParsedBy ("lapis.parsers.HTMLParser");
        }
    }

    private int removeCRs (char[] buf, int len) {
        int i, j;

        // find first CR
        for (i = 0; i < len && buf[i] != '\r'; ++i)
            ;

        // start compacting buffer, eliminating CRs
        for (j = i++; i < len; ++i) {
            char c = buf[i];
            if (c != '\r')
                buf[j++] = c;
        }

        return j;   // new length of buffer
    }


    private static Hashtable entities = new Hashtable2()
          .add ("quot", new Character ((char)34))
          .add ("amp", new Character ((char)38))
          .add ("lt", new Character ((char)60))
          .add ("gt", new Character ((char)62))
          .add ("nbsp", new Character ((char)160))
          .add ("iexcl", new Character ((char)161))
          .add ("cent", new Character ((char)162))
          .add ("pound", new Character ((char)163))
          .add ("curren", new Character ((char)164))
          .add ("yen", new Character ((char)165))
          .add ("brvbar", new Character ((char)167))
          .add ("sect", new Character ((char)167))
          .add ("uml", new Character ((char)168))
          .add ("copy", new Character ((char)169))
          .add ("ordf", new Character ((char)170))
          .add ("laquo", new Character ((char)171))
          .add ("not", new Character ((char)172))
          .add ("shy", new Character ((char)173))
          .add ("reg", new Character ((char)174))
          .add ("macr", new Character ((char)175))
          .add ("deg", new Character ((char)176))
          .add ("plusmn", new Character ((char)177))
          .add ("sup2", new Character ((char)178))
          .add ("sup3", new Character ((char)179))
          .add ("acute", new Character ((char)180))
          .add ("micro", new Character ((char)181))
          .add ("para", new Character ((char)182))
          .add ("middot", new Character ((char)183))
          .add ("cedil", new Character ((char)184))
          .add ("sup1", new Character ((char)185))
          .add ("ordm", new Character ((char)186))
          .add ("raquo", new Character ((char)187))
          .add ("frac14", new Character ((char)188))
          .add ("frac12", new Character ((char)189))
          .add ("frac34", new Character ((char)190))
          .add ("iquest", new Character ((char)191))
          .add ("Agrave", new Character ((char)192))
          .add ("Aacute", new Character ((char)193))
          .add ("Acirc", new Character ((char)194))
          .add ("Atilde", new Character ((char)195))
          .add ("Auml", new Character ((char)196))
          .add ("Aring", new Character ((char)197))
          .add ("AElig", new Character ((char)198))
          .add ("Ccedil", new Character ((char)199))
          .add ("Egrave", new Character ((char)200))
          .add ("Eacute", new Character ((char)201))
          .add ("Ecirc", new Character ((char)202))
          .add ("Euml", new Character ((char)203))
          .add ("Igrave", new Character ((char)204))
          .add ("Iacute", new Character ((char)205))
          .add ("Icirc", new Character ((char)206))
          .add ("Iuml", new Character ((char)207))
          .add ("ETH", new Character ((char)208))
          .add ("Ntilde", new Character ((char)209))
          .add ("Ograve", new Character ((char)210))
          .add ("Oacute", new Character ((char)211))
          .add ("Ocirc", new Character ((char)212))
          .add ("Otilde", new Character ((char)213))
          .add ("Ouml", new Character ((char)214))
          .add ("times", new Character ((char)215))
          .add ("Oslash", new Character ((char)216))
          .add ("Ugrave", new Character ((char)217))
          .add ("Uacute", new Character ((char)218))
          .add ("Ucirc", new Character ((char)219))
          .add ("Uuml", new Character ((char)220))
          .add ("Yacute", new Character ((char)221))
          .add ("THORN", new Character ((char)222))
          .add ("szlig", new Character ((char)223))
          .add ("agrave", new Character ((char)224))
          .add ("aacute", new Character ((char)225))
          .add ("acirc", new Character ((char)226))
          .add ("atilde", new Character ((char)227))
          .add ("auml", new Character ((char)228))
          .add ("aring", new Character ((char)229))
          .add ("aelig", new Character ((char)230))
          .add ("ccedil", new Character ((char)231))
          .add ("egrave", new Character ((char)232))
          .add ("eacute", new Character ((char)233))
          .add ("ecirc", new Character ((char)234))
          .add ("euml", new Character ((char)235))
          .add ("igrave", new Character ((char)236))
          .add ("iacute", new Character ((char)237))
          .add ("icirc", new Character ((char)238))
          .add ("iuml", new Character ((char)239))
          .add ("eth", new Character ((char)240))
          .add ("ntilde", new Character ((char)241))
          .add ("ograve", new Character ((char)242))
          .add ("oacute", new Character ((char)243))
          .add ("ocirc", new Character ((char)244))
          .add ("otilde", new Character ((char)245))
          .add ("ouml", new Character ((char)246))
          .add ("divide", new Character ((char)247))
          .add ("oslash", new Character ((char)248))
          .add ("ugrave", new Character ((char)249))
          .add ("uacute", new Character ((char)250))
          .add ("ucirc", new Character ((char)251))
          .add ("uuml", new Character ((char)252))
          .add ("yacute", new Character ((char)253))
          .add ("thorn", new Character ((char)254))
          .add ("yuml", new Character ((char)255))
          ;

    private static Character lookupEntityRef (String name) {
        return (Character) entities.get (name);
    }

    /*
     *  Parser (constructs a canonical tree of elements)
     *
     */

    private void initParse () {
    }

    protected void handleInitParse () {
        elems.setSize (0);
        openPtr = 0;
        initTagRuns ();
    }

    protected void handleTag (HTMLTag tag) {
        String tagName = tag.getTagName();

        // now handled by doneTagRuns()
        //doc.add ("Tag", tag);
        //doc.add ((tag.isStartTag() ? "<" : "</") + tagName + ">", tag);

        //addTextRun (tag.startHigh);
        addTagRun (tag);

        if (tag.isStartTag()) {
            // start tag <X>

            // check if <X> forces closure of an open element
            if (forcesClosed.containsKey (tagName)) {
                int i = findOpenElement ((Hashtable)context.get (tagName));
                if (i != -1) {
                    HTMLElement e = (HTMLElement)elems.elementAt (openElems[i]);
                    if (((Hashtable)forcesClosed.get (tagName)).containsKey (e.getTagName()))
                        close (e, tag.getStart ());
                }
            }

            // create the element and push it on the elems stack
            HTMLElement e = makeElement (tag);

            if (empty.containsKey (tagName)) {
                // element has no content -- close it off right now
                e.setEnd (tag.getEnd ());
                handleEmptyElement (e);
            }
            else
                handleStartElement (e);
        }
        else {
            // end tag </X>
            // find matching start tag <X>
            int i = findOpenElement (tagName);
            if (i != -1)
                close (i, tag);
        }
    }

    protected void handleAttribute (String attrName, 
                                    int attrNameStart, int attrNameEnd,
                                    String attrValue,
                                    int attrValueStart, int attrValueEnd) {
        attrs.addElement (attrName);
        attrs.addElement (attrValue != null ? attrValue : HTMLTag.TRUE);

        addAttributeRun (attrName, attrNameStart, attrNameEnd,
                         attrValueStart, attrValueEnd);
    }

    protected void handleDoneParse (int pos) {
        // link together the top-level elements
        if (!elems.empty()) {
            int nElems = elems.size ();
            HTMLElement c = (HTMLElement)elems.elementAt (0);
            for (int j=1; j<nElems; ++j) {
                HTMLElement d = (HTMLElement)elems.elementAt (j);
                c.sibling = d;
                c = d;
            }
        }

        doneTagRuns ();
    }

    protected HTMLTag makeTag (int start, int end, String tagName, boolean startTag) {
		//return new HTMLTag (start, end, tagName, startTag);
        return new HTMLTag (start, end, tagName, startTag, doc);
    }

    protected HTMLElement makeElement (HTMLTag tag) {
    	debug.assertion(doc.getVersion()==tag.getVersion());
		//HTMLElement e = new HTMLElement (tag, tag.getEnd ());    	
        HTMLElement e = new HTMLElement (tag, tag.getEnd (), tag.getVersion());
        String tagName = tag.getTagName ();
        tag.element = e;
        return e;
    }

    // Stack management

    Stack elems = new Stack();
        // stack of Elements appearing before the current element in
        // a preorder traversal, except that completely-visited subtrees
        // are represented by their root.
    int[] openElems = new int[20];
    int openPtr = 0;
        // stack of indices of open elements in elems

    protected void handleEmptyElement (HTMLElement e) {
        if (openPtr > 0)
            e.parent = (HTMLElement)elems.elementAt (openElems[openPtr-1]);
        else
            e.parent = null;

        // Push on stack of e's children
        elems.push (e);
        e.empty = true;

        // Put element in the labeling
        putElementInLabeling (e);
    }

    protected void handleStartElement (HTMLElement e) {
        if (openPtr > 0)
            e.parent = (HTMLElement)elems.elementAt (openElems[openPtr-1]);
        else
            e.parent = null;

        elems.push (e);

        if (openPtr == openElems.length) {
            int[] newarr = new int[openElems.length + 10];
            System.arraycopy (openElems, 0, newarr, 0, openElems.length);
            openElems = newarr;
        }
        openElems[openPtr] = elems.size()-1;
        ++openPtr;
    }

    protected void handleEndElement (HTMLElement e) {
        putElementInLabeling (e);
    }

    protected int findOpenElement (String tagname) {
        for (int i=openPtr-1; i >= 0; --i) {
            HTMLElement e = (HTMLElement)elems.elementAt (openElems[i]);
            if (tagname == e.getTagName ())
                return i;
        }
        return -1;
    }

    private int findOpenElement (Hashtable tags) {
        for (int i=openPtr-1; i >= 0; --i) {
            HTMLElement e = (HTMLElement)elems.elementAt (openElems[i]);
            if (tags.containsKey (e.getTagName ()))
                return i;
        }
        return -1;
    }

    private void close (HTMLElement elem, int end) {
        int v;
        HTMLElement e;
        do {
            v = openElems[--openPtr];
            e = (HTMLElement)elems.elementAt (v);

            e.setEnd (end);

            handleEndElement (e);

            int firstChild = v+1;
            int nElems = elems.size();
            if (firstChild < nElems) {
                HTMLElement c = (HTMLElement)elems.elementAt (firstChild);
                e.child = c;
                for (int j=firstChild+1; j<nElems; ++j) {
                    HTMLElement d = (HTMLElement)elems.elementAt (j);
                    c.sibling = d;
                    c = d;
                }
                elems.setSize (firstChild);
            }
        } while (e != elem);
    }

    // NIY: stack up unclosed flow tags (like <B> and <A>) and reopen them
    // when the next element is opened
    private void close (int i, HTMLTag tag) {
        HTMLElement elem = (HTMLElement)elems.elementAt (openElems[i]);

        if (i < openPtr-1)
            // close any open elements above elem
            close ((HTMLElement)elems.elementAt (openElems[i+1]), tag.getStart ());

        elem.endTag = tag;
        tag.element = elem;
        close (elem, tag.getEnd ());
    }

/*
    private void close (HTMLElement elem) {
        HTMLTag tag = elem.startTag;
        close (elem, tag.getEnd ());
    }
*/

    private void closeAll (int end) {
        if (openPtr > 0)
            close ((HTMLElement)elems.elementAt (openElems[0]), end);
    }

    /*
     * Element namespace
     *
     */

    void putElementInLabeling (HTMLElement e) {
        if (e instanceof HTMLFakeElement)
            return; // don't want to label fake elements

        String tagName = e.getTagName ();

        PatternCache cache = doc.getPatternCache ();
        cache.add (ELEMENT.name, e);
        PatternStub pat = (PatternStub) elementPatterns.get (tagName);
        if (pat != null)
            cache.add (pat.name, e);

//         if (tagName == HTMLTag.INPUT
//             || tagName == HTMLTag.SELECT
//             || tagName == HTMLTag.TEXTAREA) {
//             String fieldName = e.getHTMLAttribute ("name");
//             if (fieldName != null) {
//                 labels.add ("Layout.Form.Field." 
//                             + fieldName.toLowerCase (), e);
//             }
//         }
    }

/*
          .add (HTMLTag.A, "Anchor")
          .add (HTMLTag.B, "Bold")
          .add (HTMLTag.BR, "Break")
          .add (HTMLTag.COMMENT, "Comment")
          .add (HTMLTag.DD, "Definition")
          .add (HTMLTag.DIV, "Division")
          .add (HTMLTag.DL, "DefinitionList")
          .add (HTMLTag.DT, "Term")
          .add (HTMLTag.H1, "Heading1")
          .add (HTMLTag.H2, "Heading2")
          .add (HTMLTag.H3, "Heading3")
          .add (HTMLTag.H4, "Heading4")
          .add (HTMLTag.H5, "Heading5")
          .add (HTMLTag.H6, "Heading6")
          .add (HTMLTag.HR, "HorizontalRule")
          .add (HTMLTag.I, "Italic")
          .add (HTMLTag.IMG, "Image")
          .add (HTMLTag.LI, "ListItem")
          .add (HTMLTag.OL, "OrderedList")
          .add (HTMLTag.P, "Paragraph")
          .add (HTMLTag.SUB, "Subscript")
          .add (HTMLTag.SUP, "Superscript")
          .add (HTMLTag.TD, "TableCell")
          .add (HTMLTag.TH, "TableHeader")
          .add (HTMLTag.TR, "TableRow")
          .add (HTMLTag.TT, "Typewriter")
          .add (HTMLTag.U, "Underlined")
          .add (HTMLTag.UL, "BulletedList")
          ;
*/


    /*
     * Faster alternative to StringBuffer for accumulating
     * tag names, attribute names, attribute values, and content words
     */
    char[] wordbuf = new char[1024];
    int wordptr;

    private void addWordChar (char c) {
        if (wordptr == wordbuf.length) {
            char[] wb = new char[wordbuf.length + 200];
            System.arraycopy (wordbuf, 0, wb, 0, wordptr);
            wordbuf = wb;
        }
        wordbuf[wordptr++] = c;
    }

    private String getWord () {
        return new String (wordbuf, 0, wordptr);
    }

    protected void handleStartWord (int pos) {
    }

    private void endWord (int pos) {
        if (wordptr > 0) {
            handleEndWord (wordbuf, 0, wordptr, pos);
            wordptr = 0;
        }
    }

    protected void handleEndWord (char[] text, int offset, int len, int pos) {
    }

    private void clearWord () {
        wordptr = 0;
    }


    /*
     * Creating tag, attribute, and text region sets
     */
    MutableRegionSet textRuns;

    MutableRegionSet tags;
    MutableRegionSet startTags;
    MutableRegionSet endTags;
    Hashtable startTagHash; // maps tag name (e.g. "img") to MutableRegionSet
    Hashtable endTagHash; // maps tag name (e.g. "img") to MutableRegionSet

    MutableRegionSet attributes;
    MutableRegionSet attrNames;
    MutableRegionSet attrValues;
    Hashtable attrHash;  // maps attribute name (e.g. "src") to MutableRegionSet

    private void initTagRuns () {
        int nExpectedTags = doc.toRegion ().getLength () / 5;
        textRuns = new MutableRegionSet (nExpectedTags);
        tags = new MutableRegionSet (nExpectedTags);
        startTags = new MutableRegionSet (nExpectedTags/2);
        endTags = new MutableRegionSet (nExpectedTags/2);
        startTagHash = new Hashtable ();
        endTagHash = new Hashtable ();

        attributes = new MutableRegionSet (nExpectedTags);
        attrNames = new MutableRegionSet (nExpectedTags);
        attrValues = new MutableRegionSet (nExpectedTags);
        attrHash = new Hashtable ();
    }

    private void addTagRun (HTMLTag tag) {
        String tagName = tag.getTagName ();
        Hashtable tagHash;
        MutableRegionSet tagKind;
        if (tag.isStartTag ()) {
            tagHash = startTagHash;
            tagKind = startTags;
        }
        else {
            tagHash = endTagHash;
            tagKind = endTags;
        }
        
        // Look up tag's region set, creating it if not present
        MutableRegionSet sameTag = (MutableRegionSet) tagHash.get (tagName);
        if (sameTag == null)
            tagHash.put (tagName, sameTag = new MutableRegionSet ());

        // Add tag to all region sets it belongs in
        debug.assertion (doc != null);
        sameTag.insert (tag.getStart (), tag.getEnd (), doc);
        tagKind.insert (tag.getStart (), tag.getEnd (), doc);
        tags.insert (tag.getStart (), tag.getEnd (), doc);

        // If we're trying to autodetect HTML, see if this is a
        // known HTML tag
        if (!isHTML && startTagPatterns.containsKey (tagName))
            isHTML = true;
    }

    private void addAttributeRun (String attrName, 
                                  int attrNameStart, int attrNameEnd,
                                  int attrValueStart, int attrValueEnd) {
        // Look up attribute's region set, creating it if not present
        MutableRegionSet sameAttr = (MutableRegionSet) attrHash.get (attrName);
        if (sameAttr == null)
            attrHash.put (attrName, sameAttr = new MutableRegionSet ());

        // Add tag to all region sets it belongs in
        debug.assertion(doc != null);
        sameAttr.insert (attrNameStart, attrValueEnd, doc);
        attributes.insert (attrNameStart, attrValueEnd, doc);
        attrNames.insert (attrNameStart, attrNameEnd, doc);
        if (attrValueStart != attrValueEnd)
            attrValues.insert (attrValueStart, attrValueEnd, doc);
    }

    protected void addTextRun (int start, int end) {
    	debug.assertion(doc != null);
        textRuns.insert (start, end, doc);
    }

    private void doneTagRuns () {
        PatternCache cache = doc.getPatternCache ();

        cache.put (TEXT.name, textRuns);
        cache.put (TAG.name, tags);
        cache.put (START_TAG.name, startTags);
        cache.put (END_TAG.name, endTags);
        cacheTags (startTagHash, startTagPatterns, "HTML.Tag.StartTag.<");
        cacheTags (endTagHash, endTagPatterns, "HTML.Tag.EndTag.</");

        cache.put (ATTRIBUTE.name, attributes);
        cache.put (ATTRIBUTE_NAME.name, attrNames);
        cache.put (ATTRIBUTE_VALUE.name, attrValues);
        cacheAttributes (attrHash, attributePatterns, "HTML.Attribute.");

    }

    private void cacheTags (Hashtable hash, Hashtable patterns, String prefix) {
        PatternCache cache = doc.getPatternCache ();
        for (Enumeration e = hash.keys (); e.hasMoreElements (); ) {
            String tagName = (String)e.nextElement ();
            RegionSet tags = (RegionSet)hash.get (tagName);
            PatternStub pat = (PatternStub) patterns.get (tagName);
            if (pat == null)
                debug.println ("no pattern for unknown tag '" + tagName + "'");
            else
                cache.put (pat.name, tags);
        }
    }

    private void cacheAttributes (Hashtable hash, Hashtable patterns, String prefix) {
        PatternCache cache = doc.getPatternCache ();
        for (Enumeration e = hash.keys (); e.hasMoreElements (); ) {
            String attrName = (String)e.nextElement ();
            RegionSet attrs = (RegionSet)hash.get (attrName);
            PatternStub pat = (PatternStub) patterns.get (attrName);
            if (pat == null)
                debug.println ("no pattern for unknown attribute '" + attrName + "'");
            else
                cache.put (pat.name, attrs);
        }
    }

    /*
     * HTMLTag classes used by the parser
     *
     */

    // elements with no content: e.g., IMG, BR, HR.  End tags for these elements
    // are simply ignored.
    private static Hashtable empty = new Hashtable2 ()
          .add (HTMLTag.AREA)
          .add (HTMLTag.BASE)
          .add (HTMLTag.BASEFONT)
          .add (HTMLTag.BGSOUND)
          .add (HTMLTag.BR)
          .add (HTMLTag.COL)
          .add (HTMLTag.COLGROUP)
          .add (HTMLTag.COMMENT) // actually <!-- ... -->
          .add (HTMLTag.FRAME)
          .add (HTMLTag.HR)
          .add (HTMLTag.IMG)
          .add (HTMLTag.INPUT)
          .add (HTMLTag.ISINDEX)
          .add (HTMLTag.LINK)
          .add (HTMLTag.META)
          .add (HTMLTag.NEXTID)
          .add (HTMLTag.PARAM)
          .add (HTMLTag.SPACER)
          .add (HTMLTag.WBR)
          ;

    // elements that close <P> (correspond to "%block" entity in HTML 3.2 DTD)
    static Hashtable blocktag = new Hashtable2()
          .add (HTMLTag.P)
          .add (HTMLTag.UL)
          .add (HTMLTag.OL)
          .add (HTMLTag.DIR)
          .add (HTMLTag.MENU)
          .add (HTMLTag.PRE)
          .add (HTMLTag.XMP)
          .add (HTMLTag.LISTING)
          .add (HTMLTag.DL)
          .add (HTMLTag.DIV)
          .add (HTMLTag.CENTER)
          .add (HTMLTag.BLOCKQUOTE)
          .add (HTMLTag.FORM)
          .add (HTMLTag.ISINDEX)
          .add (HTMLTag.HR)
          .add (HTMLTag.TABLE)
          .add (HTMLTag.H1)
          .add (HTMLTag.H2)
          .add (HTMLTag.H3)
          .add (HTMLTag.H4)
          .add (HTMLTag.H5)
          .add (HTMLTag.H6)
          .add (HTMLTag.ADDRESS)

          // not strictly block tags, but need to be
          .add (HTMLTag.TD)
          .add (HTMLTag.TH)
          .add (HTMLTag.TR)
          ;

    // maps elements which force closure to the elements that they close, e.g.,
    // LI maps to LI, DT maps to DD,DT, and all block-level tags map to P.
    private static Hashtable forcesClosed = new Hashtable2 ()
          .add (HTMLTag.DD, new Hashtable2 () .add (HTMLTag.DD) .add (HTMLTag.DT))
          .add (HTMLTag.DT, new Hashtable2 () .add (HTMLTag.DD) .add (HTMLTag.DT))
          .add (HTMLTag.LI, new Hashtable2 () .add (HTMLTag.LI))
          .add (HTMLTag.OPTION, new Hashtable2 () .add (HTMLTag.OPTION))
          .add (HTMLTag.TR, new Hashtable2 () .add (HTMLTag.TR))
          .add (HTMLTag.TD, new Hashtable2 () .add (HTMLTag.TD) .add (HTMLTag.TH))
          .add (HTMLTag.TH, new Hashtable2 () .add (HTMLTag.TD) .add (HTMLTag.TH))
          ;
    static {
        Hashtable p = new Hashtable2 () .add (HTMLTag.P);
        Enumeration enum = blocktag.keys ();
        while (enum.hasMoreElements ())
            union (forcesClosed, enum.nextElement(), p);
    }

    // union of forcesClosed plus the tag's possible containers.  For instance,
    // LI maps to LI, OL, UL, MENU, DIR.  When a forcesClosed tag like LI is
    // encountered, the parser looks upward for the first context tag.
    // Having the tag's container element included in the search ensures that
    // LI in a nested list won't close its parent LI.
    static Hashtable context = new Hashtable2 ()
          .add (HTMLTag.DD, new Hashtable2 () .add (HTMLTag.DL))
          .add (HTMLTag.DT, new Hashtable2 () .add (HTMLTag.DL))
          .add (HTMLTag.LI, new Hashtable2 () .add (HTMLTag.OL) .add (HTMLTag.UL) .add (HTMLTag.MENU) .add (HTMLTag.DIR))
          .add (HTMLTag.OPTION, new Hashtable2 () .add (HTMLTag.SELECT))
          .add (HTMLTag.TR, new Hashtable2 () .add (HTMLTag.TABLE))
          .add (HTMLTag.TD, new Hashtable2 () .add (HTMLTag.TR) .add (HTMLTag.TABLE))
          .add (HTMLTag.TH, new Hashtable2 () .add (HTMLTag.TR) .add (HTMLTag.TABLE))
          ;
    static {
        Enumeration enum = forcesClosed.keys ();
        while (enum.hasMoreElements ()) {
            Object tagname = enum.nextElement();
            union (context, tagname, (Hashtable)forcesClosed.get (tagname));
        }
    }

    // NIY: handle literal and semi-literal elements (XMP, LISTING, TEXTAREA, OPTION)
    // elements whose content should be treated as plain text
    static Hashtable literal = new Hashtable2()
            ;

    // elements found in <HEAD>
    static Hashtable headtag = new Hashtable2()
          .add (HTMLTag.META)
          .add (HTMLTag.TITLE)
          .add (HTMLTag.BASE)
          .add (HTMLTag.LINK)
          .add (HTMLTag.ISINDEX)
          ;


    private static void union (Hashtable map, Object tagname, Hashtable tagset) {
        Hashtable2 currset = (Hashtable2)map.get (tagname);
        if (currset == null)
            map.put (tagname, tagset);
        else
            map.put (tagname, currset.union (tagset));
    }

    public static void main (String[] args) throws Exception {
        HTMLParser parser = new HTMLParser ();
        for (int i = 0; i < args.length; ++i) {
            Document doc = BasicDocument.defaultFactory.make (lapisx.net.URLUtil.HrefToURL(args[i]));

            System.err.print (parser.getClass() + " parsing...");
            long t = System.currentTimeMillis ();
            //Runtime.getRuntime().traceMethodCalls(true);  // turns tracing on
            parser.parse (doc);
            //Runtime.getRuntime().traceMethodCalls(false);  // turns tracing off
            System.err.println ((System.currentTimeMillis() - t) + " ms");

            System.err.println (parser.tags.toStringWithContent());
        }
    }
}

class Hashtable2 extends Hashtable {
    public Hashtable2 () {
    }

    public Hashtable2 add (Object key) {
        put (key, key);
        return this;
    }

    public Hashtable2 add (Object key, Object val) {
        put (key, val);
        return this;
    }

    public Hashtable2 union (Hashtable map) {
        Enumeration enum = map.keys ();
        while (enum.hasMoreElements ()) {
            Object key = enum.nextElement ();
            put (key, map.get (key));
        }

        return this;
    }


}

