/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.parsers;

import lapis.*;

/**
 * HTMLTag in an HTML page.
 */
public class HTMLTag extends Region {

    public static String TRUE = "TRUE";

    public String tagName;
    public boolean startTag;

    public String[] htmlAttributes;
        // HTML attribute name-value pairs, alternating
        // Names are upper case and interned

    public HTMLElement element; // HTML element to which this tag belongs

    /**
     * Make a HTMLTag.
     * @param start Starting offset of tag in page
     * @param end Ending offset of tag
     * @param tagName Label of tag (like "p")
     * @param startTag true for start tags (like "&lt;p&gt;"), false for end tags ("&lt;/p&gt;")
     * @param ver Document version
     */    
/*    
    public HTMLTag (int start, int end, String tagName, boolean startTag) {
        super (start, end);
        this.tagName = tagName.intern ();
        this.startTag = startTag;
        this.htmlAttributes = null;
    }
*/
    
    public HTMLTag (int start, int end, String tagName, 
    				boolean startTag, DocumentVersion ver) {
    	super (start, end, ver);
    	this.tagName = tagName.intern();
    	this.startTag = startTag;
    	this.htmlAttributes = null;
    }
    
    public HTMLTag (int start, int end, String tagName,
    				boolean startTag, Document doc) {
    	this (start, end, tagName, startTag, doc.getVersion());
    }

    /**
     * Get tag name.
     * @return tag name (like "p"), in upper-case, String.intern()'ed form.
     */
    public String getTagName () {
        return tagName;
    }

    /**
     * Get element to which this tag is the start or end tag.
     * @return element, or null if tag has no element.
     */
    public HTMLElement getElement () {
        return element;
    }

    /**
     * Convert a String to a tag name.  HTMLTag names are upper-case, intern()'ed
     * Strings.  Thus you can compare tag names with ==, as in:
     * <CODE>getTagName() == HTMLTag.IMG</CODE>.
     * @param name Label to convert (e.g., "P")
     * @return tag name (e.g. "p"), in upper-case, String.intern()'ed form.
     */
    public static String toTagName (String name) {
        return name.intern ();
    }

    /**
     * Test if tag is a start tag.  Equivalent to !isEndTag().
     * @return true if and only if tag is a start tag (like "&lt;P&gt;")
     */
    public boolean isStartTag () {
        return startTag;
    }

    /**
     * Test if tag is an end tag.  Equivalent to !isStartTag().
     * @return true if and only if tag is a start tag (like "&lt;/P&gt;")
     */
    public boolean isEndTag () {
        return !startTag;
    }

    /**
     * Test if tag is a block-level tag.  Equivalent to !isFlowTag().
     * @return true if and only if tag is a block-level tag (like "&lt;P&gt;")
     */
    public boolean isBlockTag () {
        return HTMLParser.blocktag.containsKey (tagName);
    }

    public boolean isPreformatted () {
        return tagName == PRE;
    }

    /**
     * Test if tag is a flow-level tag.  Equivalent to !isBlockTag().
     * @return true if and only if tag is a block-level tag (like "&lt;A&gt;")
     */
    public boolean isFlowTag () {
        return !isBlockTag ();
    }

    /**
     * Test if tag belongs in the <HEAD> element.
     * @return true if and only if tag is a HEAD-level tag (like "&lt;TITLE&gt;")
     */
    public boolean isHeadTag () {
        return HTMLParser.headtag.containsKey (tagName);
    }

    /**
     * Test if tag belongs in the <BODY> element.
     * @return true if and only if tag is a BODY-level tag (like "&lt;A&gt;")
     */
    public boolean isBodyTag () {
        return !isHeadTag()
                && tagName != HTML
                && tagName != HEAD
                && tagName != BODY;
    }


    /**
     * Convert a String to an HTML attribute name.  Attribute names are
     * upper-case, intern()'ed
     * Strings.  Thus you can compare attribute names with ==.
     * @param name Label to convert (e.g., "HREF")
     * @return tag name (e.g. "href"), in upper-case, String.intern()'ed form.
     */
    public static String toHTMLAttributeName (String name) {
        return name.intern ();
    }

    /**
     * Test if tag has an HTML attribute.
     * @param name Label of HTML attribute (e.g. "HREF").  Doesn't have to be
     * converted with toHTMLAttributeName().
     * @return true if tag has the attribute, false if not
     */
    public boolean hasHTMLAttribute (String name) {
        if (htmlAttributes == null)
            return false;
        name = toHTMLAttributeName (name);
        for (int i=0; i<htmlAttributes.length; i+=2)
            if (htmlAttributes[i] == name)
                return true;
        return false;
    }

    /**
     * Get an HTML attribute's value.
     * @param name Label of HTML attribute (e.g. "HREF").  Doesn't have to be
     * converted with toHTMLAttributeName().
     * @return value of attribute if it exists, TRUE if the attribute exists but has no value, or null if tag lacks the attribute.
     */
    public String getHTMLAttribute (String name) {
        if (htmlAttributes == null)
            return null;
        name = toHTMLAttributeName (name);
        for (int i=0; i<htmlAttributes.length; i += 2)
            if (htmlAttributes[i] == name)
                return htmlAttributes[i+1];
        return null;
    }

    /**
     * Get an HTML attribute's value, with a default value if it doesn't exist.
     * @param name Label of HTML attribute (e.g. "HREF").  Doesn't have to be
     * converted with toHTMLAttributeName().
     * @param defaultValue default value to return if the attribute
     * doesn't exist
     * @return value of attribute if it exists, TRUE if the attribute exists but has no value, or defaultValue if tag lacks the attribute.
     */
    public String getHTMLAttribute (String name, String defaultValue) {
        String val = getHTMLAttribute (name);
        return val != null ? val : defaultValue;
    }

    /**
     * Get number of HTML attributes on this tag.
     * @return number of HTML attributes
     */
    public int countHTMLAttributes () {
        if (htmlAttributes == null)
            return 0;
        return htmlAttributes.length / 2;
    }

    /**
     * Get all the HTML attributes found on this tag.
     * @return array of name-value pairs, alternating between
     * names and values.  Thus array[0] is a name, array[1] is a value,
     * array[2] is a name, etc.
     */
    public String[] getHTMLAttributes () {
        if (htmlAttributes == null)
            return new String[0];
        return htmlAttributes;
    }

    public void changeHTMLAttribute (String name, String newValue) {
        if (htmlAttributes == null)
            return;
        name = toHTMLAttributeName (name);
        for (int i=0; i<htmlAttributes.length; i += 2)
            if (htmlAttributes[i] == name) {
                htmlAttributes[i+1] = newValue;
                break;
            }
    }

    /**
     * Commonly useful tag names.
     * Derived from <a href="http://www.sandia.gov/sci_compute/elements.html">HTML Elements</a>
     * at Sandia National Labs.
     */

    public static final String A = "a";
    public static final String ABBREV = "abbrev";
    public static final String ACRONYM = "acronym";
    public static final String ADDRESS = "address";
    public static final String APPLET = "applet";
    public static final String AREA = "area";
    public static final String B = "b";
    public static final String BASE = "base";
    public static final String BASEFONT = "basefont";
    public static final String BDO = "bdo";
    public static final String BGSOUND = "bgsound";
    public static final String BIG = "big";
    public static final String BLINK = "blink";
    public static final String BLOCKQUOTE = "blockquote";
    public static final String BODY = "body";
    public static final String BR = "br";
    public static final String CAPTION = "caption";
    public static final String CENTER = "center";
    public static final String CITE = "cite";
    public static final String CODE = "code";
    public static final String COL = "col";
    public static final String COLGROUP = "colgroup";
    public static final String COMMENT = "!";
    public static final String DD = "dd";
    public static final String DEL = "del";
    public static final String DFN = "dfn";
    public static final String DIR = "dir";
    public static final String DIV = "div";
    public static final String DL = "dl";
    public static final String DT = "dt";
    public static final String EM = "em";
    public static final String EMBED = "embed";
    public static final String FIELDSET = "fieldset";
    public static final String FONT = "font";
    public static final String FORM = "form";
    public static final String FRAME = "frame";
    public static final String FRAMESET = "frameset";
    public static final String H1 = "h1";
    public static final String H2 = "h2";
    public static final String H3 = "h3";
    public static final String H4 = "h4";
    public static final String H5 = "h5";
    public static final String H6 = "h6";
    public static final String HEAD = "head";
    public static final String HR = "hr";
    public static final String HTML = "html";
    public static final String I = "i";
    public static final String IFRAME = "iframe";
    public static final String IMG = "img";
    public static final String INPUT = "input";
    public static final String INS = "ins";
    public static final String ISINDEX = "isindex";
    public static final String KBD = "kbd";
    public static final String LABEL = "label";
    public static final String LEGEND = "legend";
    public static final String LI = "li";
    public static final String LINK = "link";
    public static final String LISTING = "listing";
    public static final String MAP = "map";
    public static final String MARQUEE = "marquee";
    public static final String MENU = "menu";
    public static final String META = "meta";
    public static final String NEXTID = "nextid";
    public static final String NOBR = "nobr";
    public static final String NOEMBED = "noembed";
    public static final String NOFRAMES = "noframes";
    public static final String NOSCRIPT = "noscript";
    public static final String OBJECT = "object";
    public static final String OL = "ol";
    public static final String OPTION = "option";
    public static final String P = "p";
    public static final String PARAM = "param";
    public static final String PLAINTEXT = "plaintext";
    public static final String PRE = "pre";
    public static final String Q = "q";
    public static final String S = "s";
    public static final String SAMP = "samp";
    public static final String SCRIPT = "script";
    public static final String SELECT = "select";
    public static final String SMALL = "small";
    public static final String SPACER = "spacer";
    public static final String SPAN = "span";
    public static final String STRIKE = "strike";
    public static final String STRONG = "strong";
    public static final String STYLE = "style";
    public static final String SUB = "sub";
    public static final String SUP = "sup";
    public static final String TABLE = "table";
    public static final String TBODY = "tbody";
    public static final String TD = "td";
    public static final String TEXTAREA = "textarea";
    public static final String TFOOT = "tfoot";
    public static final String TH = "th";
    public static final String TITLE = "title";
    public static final String TR = "tr";
    public static final String TT = "tt";
    public static final String U = "u";
    public static final String UL = "ul";
    public static final String VAR = "var";
    public static final String WBR = "wbr";
    public static final String XMP = "xmp";

    public static final String[] ALL_TAGS = {
        A, ABBREV, ACRONYM, ADDRESS, APPLET, AREA, B, BASE,
        BASEFONT, BDO, BGSOUND, BIG, BLINK, BLOCKQUOTE, BODY, BR,
        CAPTION, CENTER, CITE, CODE, COL, COLGROUP, COMMENT, DD,
        DEL, DFN, DIR, DIV, DL, DT, EM, EMBED, FIELDSET,
        FONT, FORM, FRAME, FRAMESET, H1, H2, H3, H4,
        H5, H6, HEAD, HR, HTML, I, IFRAME, IMG, INPUT, INS,
        ISINDEX, KBD, LABEL, LEGEND, LI, LINK, LISTING, MAP, MARQUEE, MENU,
        META, NEXTID, NOBR, NOEMBED, NOFRAMES, NOSCRIPT, OBJECT, OL, OPTION,
        P, PARAM, PLAINTEXT, PRE, Q, S, SAMP, SCRIPT, SELECT, SMALL,
        SPACER, SPAN, STRIKE, STRONG, STYLE, SUB, SUP, TABLE, TBODY, TD,
        TEXTAREA, TFOOT, TH, TITLE, TR, TT, U, UL, VAR,
        WBR, XMP
    };

    // FIX: add all other HTML 4 attribute names
    public static final String ACCEPT_CHARSET_ATTRIB = "accept-charset";
    public static final String ACCEPT_ATTRIB = "accept";
    public static final String ACCESSKEY_ATTRIB = "accesskey";
    public static final String ACTION_ATTRIB = "action";
    public static final String ALIGN_ATTRIB = "align";
    public static final String ALINK_ATTRIB = "alink";
    public static final String ALT_ATTRIB = "alt";
    public static final String ARCHIVE_ATTRIB = "archive";
    public static final String AXES_ATTRIB = "axes";
    public static final String AXIS_ATTRIB = "axis";
    public static final String BACKGROUND_ATTRIB = "background";
    public static final String BGCOLOR_ATTRIB = "bgcolor";
    public static final String BORDER_ATTRIB = "border";
    public static final String CELLPADDING_ATTRIB = "cellpadding";
    public static final String CELLSPACING_ATTRIB = "cellspacing";
    public static final String CHAR_ATTRIB = "char";
    public static final String CHAROFF_ATTRIB = "charoff";
    public static final String CHARSET_ATTRIB = "charset";
    public static final String CHECKED_ATTRIB = "checked";
    public static final String CITE_ATTRIB = "cite";
    public static final String CLASS_ATTRIB = "class";
    public static final String CLASSID_ATTRIB = "classid";
    public static final String CLEAR_ATTRIB = "clear";
    public static final String CODE_ATTRIB = "code";
    public static final String CODEBASE_ATTRIB = "codebase";
    public static final String CODETYPE_ATTRIB = "codetype";
    public static final String COLOR_ATTRIB = "color";
    public static final String COLS_ATTRIB = "cols";
    public static final String COLSPAN_ATTRIB = "colspan";
    public static final String COMPACT_ATTRIB = "compact";
    public static final String CONTENT_ATTRIB = "content";
    public static final String COORDS_ATTRIB = "coords";
    public static final String DATA_ATTRIB = "data";
    public static final String DATETIME_ATTRIB = "datetime";
    public static final String DECLARE_ATTRIB = "declare";
    public static final String DIR_ATTRIB = "dir";
    public static final String DISABLED_ATTRIB = "disabled";
    public static final String ENCTYPE_ATTRIB = "enctype";
    public static final String FACE_ATTRIB = "face";
    public static final String FOR_ATTRIB = "for";
    public static final String FRAME_ATTRIB = "frame";
    public static final String FRAMEBORDER_ATTRIB = "frameborder";
    public static final String HEIGHT_ATTRIB = "height";
    public static final String HREF_ATTRIB = "href";
    public static final String HTTP_EQUIV_ATTRIB = "http-equiv";
    public static final String ID_ATTRIB = "id";
    public static final String ISMAP_ATTRIB = "ismap";
    public static final String LANG_ATTRIB = "lang";
    public static final String LANGUAGE_ATTRIB = "language";
    public static final String LINK_ATTRIB = "link";
    public static final String MARGINHEIGHT_ATTRIB = "marginheight";
    public static final String MARGINWIDTH_ATTRIB = "marginwidth";
    public static final String MAXLENGTH_ATTRIB = "maxlength";
    public static final String MEDIA_ATTRIB = "media";
    public static final String METHOD_ATTRIB = "method";
    public static final String MULTIPLE_ATTRIB = "multiple";
    public static final String NAME_ATTRIB = "name";
    public static final String NOHREF_ATTRIB = "nohref";
    public static final String NORESIZE_ATTRIB = "noresize";
    public static final String NOSHADE_ATTRIB = "noshade";
    public static final String NOWRAP_ATTRIB = "nowrap";
    public static final String OBJECT_ATTRIB = "object";
    public static final String ONBLUR_ATTRIB = "onblur";
    public static final String ONCHANGE_ATTRIB = "onchange";
    public static final String ONCLICK_ATTRIB = "onclick";
    public static final String ONDBLCLICK_ATTRIB = "ondblclick";
    public static final String ONFOCUS_ATTRIB = "onfocus";
    public static final String ONKEYDOWN_ATTRIB = "onkeydown";
    public static final String ONKEYPRESS_ATTRIB = "onkeypress";
    public static final String ONKEYUP_ATTRIB = "onkeyup";
    public static final String ONLOAD_ATTRIB = "onload";
    public static final String ONMOUSEDOWN_ATTRIB = "onmousedown";
    public static final String ONMOUSEMOVE_ATTRIB = "onmousemove";
    public static final String ONMOUSEOUT_ATTRIB = "onmouseout";
    public static final String ONMOUSEOVER_ATTRIB = "onmouseover";
    public static final String ONMOUSEUP_ATTRIB = "onmouseup";
    public static final String ONRESET_ATTRIB = "onreset";
    public static final String ONSELECT_ATTRIB = "onselect";
    public static final String ONSUBMIT_ATTRIB = "onsubmit";
    public static final String ONUNLOAD_ATTRIB = "onunload";
    public static final String PROFILE_ATTRIB = "profile";
    public static final String PROMPT_ATTRIB = "prompt";
    public static final String READONLY_ATTRIB = "readonly";
    public static final String REL_ATTRIB = "rel";
    public static final String REV_ATTRIB = "rev";
    public static final String ROWS_ATTRIB = "rows";
    public static final String ROWSPAN_ATTRIB = "rowspan";
    public static final String RULES_ATTRIB = "rules";
    public static final String SCHEME_ATTRIB = "scheme";
    public static final String SCROLLING_ATTRIB = "scrolling";
    public static final String SELECTED_ATTRIB = "selected";
    public static final String SHAPE_ATTRIB = "shape";
    public static final String SHAPES_ATTRIB = "shapes";
    public static final String SIZE_ATTRIB = "size";
    public static final String SPAN_ATTRIB = "span";
    public static final String SRC_ATTRIB = "src";
    public static final String STANDBY_ATTRIB = "standby";
    public static final String START_ATTRIB = "start";
    public static final String STYLE_ATTRIB = "style";
    public static final String TABINDEX_ATTRIB = "tabindex";
    public static final String TARGET_ATTRIB = "target";
    public static final String TEXT_ATTRIB = "text";
    public static final String TITLE_ATTRIB = "title";
    public static final String TYPE_ATTRIB = "type";
    public static final String USEMAP_ATTRIB = "usemap";
    public static final String VALIGN_ATTRIB = "valign";
    public static final String VALUE_ATTRIB = "value";
    public static final String VALUETYPE_ATTRIB = "valuetype";
    public static final String VERSION_ATTRIB = "version";
    public static final String VLINK_ATTRIB = "vlink";
    public static final String WIDTH_ATTRIB = "width";

    public static final String[] ALL_ATTRIBUTES = {
        ACCEPT_CHARSET_ATTRIB, ACCEPT_ATTRIB, ACCESSKEY_ATTRIB, ACTION_ATTRIB, ALIGN_ATTRIB, 
        ALINK_ATTRIB, ALT_ATTRIB, ARCHIVE_ATTRIB, AXES_ATTRIB, AXIS_ATTRIB, BACKGROUND_ATTRIB, BGCOLOR_ATTRIB, 
        BORDER_ATTRIB, CELLPADDING_ATTRIB, CELLSPACING_ATTRIB, CHAR_ATTRIB, CHAROFF_ATTRIB, 
        CHARSET_ATTRIB, CHECKED_ATTRIB, CITE_ATTRIB, CLASS_ATTRIB, CLASSID_ATTRIB, CLEAR_ATTRIB, CODE_ATTRIB, 
        CODEBASE_ATTRIB, CODETYPE_ATTRIB, COLOR_ATTRIB, COLS_ATTRIB, COLSPAN_ATTRIB, COMPACT_ATTRIB, 
        CONTENT_ATTRIB, COORDS_ATTRIB, DATA_ATTRIB, DATETIME_ATTRIB, DECLARE_ATTRIB, DIR_ATTRIB, DISABLED_ATTRIB, 
        ENCTYPE_ATTRIB, FACE_ATTRIB, FOR_ATTRIB, FRAME_ATTRIB, FRAMEBORDER_ATTRIB, HEIGHT_ATTRIB, HREF_ATTRIB, 
        HTTP_EQUIV_ATTRIB, ID_ATTRIB, ISMAP_ATTRIB, LANG_ATTRIB, LANGUAGE_ATTRIB, LINK_ATTRIB, MARGINHEIGHT_ATTRIB, 
        MARGINWIDTH_ATTRIB, MAXLENGTH_ATTRIB, MEDIA_ATTRIB, METHOD_ATTRIB, MULTIPLE_ATTRIB, NAME_ATTRIB, 
        NOHREF_ATTRIB, NORESIZE_ATTRIB, NOSHADE_ATTRIB, NOWRAP_ATTRIB, OBJECT_ATTRIB, ONBLUR_ATTRIB, 
        ONCHANGE_ATTRIB, ONCLICK_ATTRIB, ONDBLCLICK_ATTRIB, ONFOCUS_ATTRIB, ONKEYDOWN_ATTRIB, 
        ONKEYPRESS_ATTRIB, ONKEYUP_ATTRIB, ONLOAD_ATTRIB, ONMOUSEDOWN_ATTRIB, ONMOUSEMOVE_ATTRIB, 
        ONMOUSEOUT_ATTRIB, ONMOUSEOVER_ATTRIB, ONMOUSEUP_ATTRIB, ONRESET_ATTRIB, ONSELECT_ATTRIB, 
        ONSUBMIT_ATTRIB, ONUNLOAD_ATTRIB, PROFILE_ATTRIB, PROMPT_ATTRIB, READONLY_ATTRIB, REL_ATTRIB, 
        REV_ATTRIB, ROWS_ATTRIB, ROWSPAN_ATTRIB, RULES_ATTRIB, SCHEME_ATTRIB, SCROLLING_ATTRIB, SELECTED_ATTRIB, 
        SHAPE_ATTRIB, SHAPES_ATTRIB, SIZE_ATTRIB, SPAN_ATTRIB, SRC_ATTRIB, STANDBY_ATTRIB, START_ATTRIB, STYLE_ATTRIB, 
        TABINDEX_ATTRIB, TARGET_ATTRIB, TEXT_ATTRIB, TITLE_ATTRIB, TYPE_ATTRIB, USEMAP_ATTRIB, VALIGN_ATTRIB, 
        VALUE_ATTRIB, VALUETYPE_ATTRIB, VERSION_ATTRIB, VLINK_ATTRIB, WIDTH_ATTRIB
    };

}
