/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.parsers;

import lapis.*;
import lapis.tc.*;
import java.util.Collection;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.List;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;

/**
 * Parser that reads files that are lists of words and/or phrases and
 * forms classes that represent them (e.g. Universities, Presidents)
 *
 * NOTE: When using the TRIE, the searches default to finding ALL
 * OVERLAPPING SETS (not greedy like REGEX), ONLY CASE-SENSITIVE
 * MATCHES, and DISREGARDING WHETHER MATCHES OCCUR AT WORD BOUNDARY OR
 * NOT.  These would interesting options to be able to specify in the
 * file.
 */
public class ListParser implements Parser {

    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;
    //    public static java.io.PrintStream debug = System.out;

    //The following constants are used to specify the algorithm to use
    //to search for words in the text.  The TRIE algorithm uses a trie
    //data structure to index the words; the REGEX algorithm uses a
    //JAVA regex package; the BRUTE_FORCE algorithm simply scans the
    //content for each word to find occurances.
    public static final int TRIE = 0;
    public static final int REGEX = 1;
    public static final int BRUTE_FORCE = 2;
    public static final int TYPE = ListParser.TRIE;

    /**
     * specifies the max number of words per leaf node.
     */
    private static final int TRIE_BRANCH_FACTOR = 20;

    //    private static HashMap TCs = new HashMap();
    //private static ArrayList patternStubs = new ArrayList(); //PatternStub
    private static final int PROGRESS_INTERVAL = 1000;

	// the document that this parser is parsing
	private lapis.Document pdoc; 

    public static void loadList(String fileStr) throws IOException {
	loadList(fileStr, "CASE-INSENSITIVE", "NO-WORD-BOUNDARY");
    }

    public static void loadList(String fileStr, 
				String caseSensitiveStr, String wordBoundaryStr) throws IOException {
// 	try {
	    PatternLibrary lib = PatternLibrary.getGlobalLibrary();
	    File file = new File(fileStr);
	    if (file == null || !file.exists() || file.isDirectory()) {
		throw new IOException("Cannot load file for ListParser: file " + 
				    fileStr + " does not exist.");
	    }
	    boolean caseSensitive = "CASE-SENSITIVE".equals(caseSensitiveStr);
	    boolean wordBoundary  = "WORD-BOUNDARY".equals(wordBoundaryStr);
	    List words = getListFromFile(file, caseSensitive, wordBoundary);
	    debug.println(file.getName());
	    String[] wordsStr = (String[])(words.toArray(new String[0]));
	    Arrays.sort(wordsStr);
	    Node[] nodes = getArray(wordsStr, 0, words.size(), 0);
	    for (int j=0; j<nodes.length; j++) {
		//debug.println(nodes[j]);
	    }
	    ListParser listParser = new ListParser(file.getName(), words, nodes, 
						   caseSensitive, wordBoundary);
	    lib.addParser (listParser);
// 	}
// 	catch (Exception e) {
// 	    debug.println(e.getMessage());
// 	}
    }

    // static initialization of lists of universities, presidents, etc
    public static void loadFromDirectory(String dir) {
	try {
	    PatternLibrary lib = PatternLibrary.getGlobalLibrary();
	    File listDir = new File(dir);
	    //	URLUtil.HrefToFile("parsers/lists",
	    //			   Main.getLapisDir());
	    File[] files = listDir.listFiles();
	    if (files == null) {
		throw new Exception("Cannot load files for ListParser: directory " 
				    + dir + " does not exist.");
	    }
	    for (int i=0; i<files.length; i++) {
		// ignore directory (e.g. CVS directory), 
		// and ignore temp files terminated by ~
		// assume remaining files are valid lists
		if (!files[i].isDirectory() && !files[i].getName().endsWith("~")) {
		    List words = getListFromFile(files[i], true, false);
		    debug.println(files[i].getName());
		    String[] wordsStr = (String[])(words.toArray(new String[0]));
		    Arrays.sort(wordsStr);
		    Node[] nodes = getArray(wordsStr, 0, words.size(), 0);
		    for (int j=0; j<nodes.length; j++) {
			//debug.println(nodes[j]);
		    }
		    
		    ListParser listParser = new ListParser(files[i].getName(), 
							   words, nodes);
		    lib.addParser (listParser);
			
		}
	    }
	}
	catch (Exception e) {
	    debug.println(e.getMessage());
	    e.printStackTrace();
	}
    }//end static initialization



    ///////////////////////////// INSTANCE DATA ///////////////////////////

    /**
     * fully-qualified name (e.g. Business.Address.Countries)
     */
    private String name = null;

    /**
     * set of all regions found after parsing the list for word
     * occurances.
     */
    private MutableRegionSet set = null;

    /**
     * list of words for this pattern.
     */
    private List wordList = null;

    /**
     * trie data structure.
     */
    private Node[] trie = null;

    private boolean caseSensitive = false;
    private boolean greedy = false; //NOT implemented
    private boolean wordBoundary = false;

    private String tc = null;
    private PatternStub p = null;


    ///////////////////////////// CONSTRUCTORS  ///////////////////////////

    /**
    public ListParser() {
	//MUST KEEP no-arg constructor to allow Reflection to
	//correctly bind patterns!
    }
    **/

    /**
     * Create new list parser with specified name, list of words, and
     * trie structure.
     *
     */
    public ListParser(String name, List wordList, Node[] trie) {
	tc = getStringFromList(wordList);
	p = new PatternStub(name, this);

	this.name = name;
	this.wordList = wordList;
	this.trie = trie;
    }

    public ListParser(String name, List wordList, Node[] trie, 
		      boolean caseSensitive, boolean wordBoundary) {
	this(name, wordList, trie);
	this.caseSensitive = caseSensitive;
	this.wordBoundary = wordBoundary;
    }

    public ListParser(String name, Collection wordList, 
		      boolean caseSensitive, boolean wordBoundary) {
	this(name, toList (wordList), buildTrie (wordList), caseSensitive, wordBoundary);
    }

    private static List toList (Collection c) {
        if (c instanceof List)
            return (List) c;
        else
            return new ArrayList (c);
    }

    private static Node[] buildTrie (Collection words) {
        String[] wordsStr = (String[])(words.toArray(new String[0]));
        Arrays.sort(wordsStr);
        return getArray(wordsStr, 0, words.size(), 0);
    }

    ///////////////////////////// METHODS  ///////////////////////////


    private static Node[] getArray(String[] s, int st, int end, int loc) {
	List nodes = new ArrayList();
	boolean word = false;

	int size = 1;
	int start = st;

	if (loc >= s[st].length()) {
	    debug.println("s[st]: " + s[st]);
	    debug.println("loc: " + loc);
	}
	char c = s[st].charAt(loc);
	for (int i=st+1; i<end; i++) {
	    if (c == s[i].charAt(loc)) {
		size++;
	    }
	    else {
		if (size > TRIE_BRANCH_FACTOR) { 
		    if (loc + 1 == s[start].length()) { word = true; }
		    Node nonLeaf = 
			new TrieNode(c, word, 
				     getArray(s, 
					      (word) ? (start + 1) : start, 
					      start + size, loc + 1));
		    nodes.add(nonLeaf);
		    word = false;
		}
		else {
		    String[] list = new String[size];
		    //		    debug.println("size: " + size);
		    System.arraycopy(s, start, list, 0, size);
		    Node node = new LeafNode(c, word, list );	
		    nodes.add(node);
		}
		//reset values
		c = s[i].charAt(loc);
		size = 1;
		start = i;		
	    }
	}

	if (size > TRIE_BRANCH_FACTOR) {
	    //	    debug.println("222: st: " + s[st] + " end: " + s[end]);
	    debug.println("222: loc + 1: " + (loc + 1) + 
			  " length(): " + s[start].length() + 
			  " s[start]=" + s[start]);
	    debug.println("222: start + 1 word: " + s[start+1]);
	    if (loc + 1 == s[start].length()) { word = true; }
	    Node nonLeaf = new TrieNode(c, word, 
					getArray(s, 
						 (word) ? (start + 1) : start, 
						 start + size, loc + 1));
	    nodes.add(nonLeaf);
	}
	else {
	    String[] list = new String[size];
	    debug.println("222: size: " + size);
	    debug.println("start: " + start);
	    debug.println("s len: " + s.length);
	    System.arraycopy(s, start, list, 0, size);
	    Node node = new LeafNode(c, word, list );	
	    nodes.add(node);
	    debug.println("222: added LeafNode");
	}
	debug.println("return array ...");
	return ((Node[])(nodes.toArray(new Node[1])));
    }


    private static String getStringFromList (List words) {
	/** gen string like /apple|pear/ */
	StringBuffer sb = new StringBuffer("/");
	for (int i=0; i<words.size(); i++) {
	    sb.append(escape((String)words.get(i)));
	    if (i != (words.size()-1)) sb.append("|");
	}
	sb.append("/");
	//debug.println("REGEX: " + sb);
	return sb.toString();
    }

    
    private static String escape (String s) {
	StringBuffer sb = new StringBuffer(s);
	char temp;
	for (int i=0; i<sb.length(); i++) {
	    temp = sb.charAt(i);
	    if (temp == '\\' ||
		temp == '/' ||
		temp == '|' ||
		temp == '(' ||
		temp == ')' ||
		temp == '[' ||
		temp == '{' ||
		temp == '^' ||
		temp == '$' ||
		temp == '*' ||
		temp == '+' ||
		temp == '?' ||
		temp == '.') {
		sb.insert(i, '\\');
		i++;
	    }
	}	    
	return sb.toString();
    }


    private static List getListFromFile (File file, 
					 boolean caseSensitive, 
					 boolean wordBoundary) {
	String temp = null;
	List list = new ArrayList();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(file));
	    while ((temp = br.readLine()) != null) {
		if (temp != null && !temp.equals("")) { // && temp.length() > 1) {
		    if (!caseSensitive) {
			String tempStr = temp.trim().toLowerCase();
			list.add(tempStr);
		    }
		    else {
			list.add(temp.trim());
		    }
		}
	    }
	}
	catch (IOException e) {
	    debug.println("error building list for " + file);
	    e.printStackTrace();
	}
	return list;
    }


    private void parseBruteForce(String content) {
	/**
	 * ALGORITHM:
	 * for each list
	 *  for each word
	 *   find all instances of word in document
	 *   add to mutable region set
	 *
	 * runtime is O(w*m*n)
	 *  w is average words per list
	 *  m is average length of word
	 *  n is length of document
	 *
	 * This is too slow.  Programmer needs more help from
	 * theoreticians.
	 */
	debug.assertion (pdoc != null);
	int len = content.length();
	
	String token = null;
	for (int i=0; i<wordList.size(); i++) {
	    token = (String)wordList.get(i);
	    for (int n=0; n<content.length(); n++) {
		int s = content.indexOf(token, n);
		if (s > 0) {
		    set.insert(s, s + token.length(), pdoc);
		    n = s + token.length();
		}
	    }
	}
    }


   static class PatternStub implements Pattern {
	private String name = null;	
	private ListParser listParser = null;

	public PatternStub (String name, ListParser listParser) {
	    this.name = name;
	    this.listParser = listParser;
	}

	public RegionSet match (Document doc) {
	    PatternCache cache = doc.getPatternCache();
	    if (!cache.wasParsedBy(name)) { //"lapis.parsers.ListParser")) {		
	        listParser.parse(doc);
	    }
	    RegionSet s = cache.get(name);
	    if (s == null) {
		debug.println("ListParser ran but didn't create " + this);
		s = RegionSet.EMPTY;
	    }
	    return s;
	}	
	public String toString() {
	    return name;
	}
    }


    /**
     * binds patterns that have been statically initialized.
     */
    public void bind(PatternLibrary lib) {
	debug.println("binding List patterns");
	if (TYPE == REGEX) { //if REGEX then add TC objects
	    debug.println("iterator binding...");
	    try {
		lib.put(name, new TC(tc));
	    }
	    catch (Exception e) {
		e.printStackTrace();
	    }
	}	    
	else { //if not REGEX, then use the PatternStub objects
	    lib.put(p.name, p);
	}
	debug.println("done binding...");
    }

    public void bind (PatternLibrary lib, Document doc) {
        // no doc-specific patterns for this parser
    }

    //
    // Scan document for instances of words that are in either of the
    // lists, returning the class represented by the word.
    //
    public void parse(Document doc) {
    pdoc = doc;
	long start = System.currentTimeMillis();

	set = new MutableRegionSet();

	try {
	    PatternCache cache = doc.getPatternCache();
	    DocumentView cookedView = doc.getCookedView ();
	    CoordinateMap cookedToRaw = doc.getVersion().diff(cookedView.getVersion());

		/*	    
	    if (cookedToRaw != IdentityMap.IDENTITY) {
			String cooked = cookedView.getText();
			Main.getProgress().fireStartProgress("Lists", cooked.length());
		
			if (TYPE == TRIE) {
		    	parse(cooked);
			}
			else {
		    	parseBruteForce(cooked);
			}		
			cache.put(name, set.convert(cookedToRaw));
	    }
	    else {
			String content = doc.getText();
			Main.getProgress().fireStartProgress("Lists", content.length());
			if (TYPE == TRIE) {
		    	parse(content);
			}
			else {
		    	parseBruteForce(content);
			}		
			cache.put(name, set);		
	    }
	    */
	    String content = cookedView.getText();
	    Main.getProgress().fireStartProgress("Lists", content.length());
	    if (TYPE == TRIE) {
	    	parse (content);
	    }
	    else {
	    	parseBruteForce (content);
	    }
	    cache.put(name, set.getSingleDocRegionSet(doc).convert(cookedToRaw));
        cache.setParsedBy (name);//"lapis.parsers.ListParser");
	}
	finally {
	    Main.getProgress().fireDoneProgress();
	    debug.println("ListParser ran in " + 
			       (System.currentTimeMillis() - start)
			       + " ms");
	}
    }//end parse(Document)


    static class Node implements Comparable {

	public Node(char c, boolean word) {
	    this.c = c;
	    this.word = word;
	}
	
	//character at node in trie
	public char c;

	//boolean used to indicate whether the node is ALSO a word
	//boundary. (e.g. a LeafNode could contain Apple, Apricot ...,
	//but also 'A'.  in such a case, word should be true.  note
	//this is NOT IMPLEMENTED YET)
	public boolean word = false;

	public int compareTo(Object o) {
	    if (! (o instanceof Node) )
		throw new ClassCastException();
	    if (this.c == ((Node)o).c)
		return 0;
	    if (this.c < ((Node)o).c)
		return -1;
	    else return 1;
	}
    }



    static class TrieNode extends Node {
	public Node[] children;
	
	public TrieNode (char c, boolean word, Node[] children) {
	    super(c, word);
	    this.children = children;
	}

	private String print(Node[] children) {
	    StringBuffer sb = new StringBuffer();
	    for (int i=0; i<children.length; i++) 
		sb.append("\t" + children[i] + "\n");
	    return sb.toString();
	}

	public String toString() {
	    return "[ c=" + c + ", word=" + word + ", children=" + print(children);
	}
    }



    static class LeafNode extends Node {
	public String[] list;

	public LeafNode (char c, boolean word, String[] list) {
	    super(c, word);
	    this.list = list;
	}

	private String print(String[] list) {
	    StringBuffer sb = new StringBuffer();
	    for (int i=0; i<list.length; i++) 
		sb.append(list[i] + ", ");
	    return sb.toString();
	}

	public String toString() {
	    return "[ c=" + c + ", word=" + word + ", list=" + print(list) + " ]";
	}
    }



    private void search(Node[] children, int start, char c, String content, int offset, MutableRegionSet set, String path) {
	debug.assertion (pdoc != null);
	int index = binarySearch(children, 0, children.length, c);
	if (index >= 0) {
	    Node temp = children[index];

	    if (temp instanceof LeafNode) {
		LeafNode leaf = (LeafNode)temp;
		for (int i=0; i<leaf.list.length; i++) {
		    int end = start + leaf.list[i].length();
		    String ctemp = content.substring(start, ((end > content.length()) ? content.length() : end));
		    if (leaf.list[i].equals(ctemp)) {
			if (wordBoundary) {
                            if (isWordBoundary (content, start)
                                && isWordBoundary (content, end))

                                set.insert(start, end, pdoc);	 
			}
			else {
			    set.insert(start, end, pdoc);
			}
						 
		    }

		}
	    }
	    else {
		String newPath = path + c;
		int end = start + newPath.length();
		String ctemp = content.substring(start, ((end > content.length()) ? content.length() : end));
		if (((TrieNode)temp).word && (newPath.equals(ctemp))) {
		    if (wordBoundary) {
			if ((start == 0 || 
			     content.charAt(start-1) == ' ' ||
			     content.charAt(start-1) == '\n' ||
			     content.charAt(start-1) == '\t') &&
			    (end == content.length() ||
			     content.charAt(end) == ' ' ||
			     content.charAt(end) == '\n' ||
			     content.charAt(end) == '\t')) {				 
			    set.insert(start, end, pdoc);				 
			}
		    }
		    else {
			set.insert(start, end, pdoc);
		    }
		}

		offset++;
		if (content.length() == (start + offset)) return;
		search(((TrieNode)temp).children, start, content.charAt(start+offset), content, offset, set, path+c);
	    }
	}
    }

    private static boolean isWordBoundary (String content, int offset) {
        if (offset == 0 || offset == content.length ())
            return true;

        char charBefore = content.charAt (offset-1);
        char charAfter = content.charAt (offset);

        return Character.isLetterOrDigit (charBefore)
            != Character.isLetterOrDigit (charAfter);
    }


    // TRIE ALGORITHM
    //  for all categories
    //
    private void parse(String content) {
	String data = content;
	if (!caseSensitive) {
	    data = content.toLowerCase();
	}
	for (int j=0; j<data.length(); j++) {
	    search(trie, j, data.charAt(j), data, 0, set, "");
	}
    }//end parse


    /**
     * Search a sorted array of integers.
     * @param array Array of integers
     * @param offset Starting offset of subarray to search
     * @param length Length of subarray to search
     * @param x Value to search for
     * @return largest index i in subarray (offset &lt;= i &lt;= offset+length)
     * such that all elements below i in the subarray are strictly less
     * than x.  If x is found in the subarray, then array[i] == x (and i is
     * the first occurence of x in the subarray).  If x is not found, 
     * then array[i] is where x should be inserted in the sort order.
     */
    public static int binarySearch (Node[] array, int offset, int length, char c) {
        // handle 0-length subarray case right away
        if (length <= 0)
            return -1;

        int low = offset;
        int high = offset+length-1;
        // since length > 0, array[low] and array[high] are valid indices

	if (c < array[low].c)
            return -1;
        if (c > array[high].c)
            return -1;
	if (c == array[low].c) {
	    return low;
	}
	if (c == array[high].c)
	    return high;
        
        while (low+1 < high) {
            // loop invariant: array[low] < x <= array[high],
            //                 offset <= low < high < offset+length
            int mid = (low + high)/2;
            if (c <= array[mid].c)
                high = mid;
            else
                low = mid;
        }

        // now we have array[low] < x <= array[high]
        //             && (low+1 == high || low == high)
        //  implies low+1 == high
        //debug.assertion (low+1 == high);
	if (c == array[high].c)
	    return high;
	else return -1;
    }

}//end class


//performance benchmarks

//                           (parse time in milliseconds)
// Filename       (wc -l)    BRUTE_FORCE    REGEX    TRIE
// TestFile1.txt       12            655      312     262
// TestFile2.txt       72           7288      534     234
// TestFile3.txt     1080            -       2577     231
// TestFile4.txt     4327            -      10007     678
