/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.parsers;

import lapis.*;
import lapis.maps.*;
import java.util.StringTokenizer;
import java.io.*;

/**
 * LinkGrammar parser extends ExternalParser by overriding the "cook"
 * method.  The LinkGrammar separates the content along sentences, and
 * then for each region removes the newlines and separates the content
 * by appending newlines.  This simplifies the logic of the adapter.
 */
public class SentenceCooker extends ExternalParser {

    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    public SentenceCooker (String name, Command command, java.util.HashMap patternMap) {
	super(name, command, patternMap);
	debug.println("******************************" + command.toString() + "*****************");
    }

    
    /**
     * This method coalesces whitespace while eliminating newlines
     * within sentences and placing newlines between sentences.
     *
     */
    public lapis.DocumentView cook(lapis.Document doc) {
	DocumentView cookedView = doc.getCookedView();
	CoordinateMap rawToHtml = cookedView.getVersion().diff (doc.getVersion());
	
	//go into pattern library
	//get ".English.Sentence" regions
	//for each region, r
	// remove all '\n' from r
	// separate tokens "';.?...\n" from sentence
	// append '\n'

	PatternLibrary lib = PatternLibrary.getGlobalLibrary();
	Pattern sentence = lib.get(".English.Sentence");
	RegionSet regionSet = sentence.match(doc);
	
	//if (rawToHtml != IdentityMap.IDENTITY)
	regionSet = regionSet.getSingleDocRegionSet(doc).convert(rawToHtml);
		
	RegionEnumeration regions = regionSet.regions();

	PiecewiseLinearMonotonicMap map = new PiecewiseLinearMonotonicMap();
	map.add(0,0);

	int loc = 0; //how far we've moved along original string
	String oldContent = cookedView.getText();
	debug.println("DOC TEXT: " + cookedView.getText());

	StringBuffer newContent = new StringBuffer();
	String text = null;
	String chars = ",:;'\"!?";
	String delims = chars + " \n\r";
	boolean previousSpace = false;
	for (Region r = regions.firstFast(); r != null; r = regions.nextFast()) {
	    loc = r.getStart();
	    map.add(loc, newContent.length());
	    text = r.extract(oldContent);

	    debug.println("TEXT: " + text);
	    int last = text.lastIndexOf('.');
	    String end = "";
	    if (last != -1) {
		end = text.substring(last);
		text = text.substring(0,last);
	    }

	    StringTokenizer st = new StringTokenizer(text, delims, true);
	    String tempToken = null;
	    while (st.hasMoreTokens()) {
		tempToken = st.nextToken();
		if (tempToken.equals(" ")) {
		    if (!previousSpace) {
			newContent.append(tempToken);
			previousSpace = true;
		    }
		    else {

		    }					
		    loc += 1;
		    map.add(loc, newContent.length());
		}
		else if (chars.indexOf(tempToken) != -1) {
		    //INVARIANT: must be of length "1"
		    if (previousSpace) {
			debug.println("previous space, therefore no need to append one");
			newContent.append(tempToken);
			loc += 1;
			map.add(loc, newContent.length());
			previousSpace = false;
		    }
		    else {
			debug.println("no space, therefore need to append one.");
			newContent.append(" ").append(tempToken);
			loc += 1;
			map.add(loc, newContent.length());
		    }
		}
		else if ("\n".equals(tempToken) || "\r".equals(tempToken)) {
		    if (previousSpace) {
			debug.println("space before newline, therefore no need to add another");
			loc += 1;
			map.add(loc, newContent.length());
		    }
		    else {
			debug.println("no space, therefore need to add before removing newline");
			newContent.append(" ");
			loc += 1;
			map.add(loc, newContent.length());
			previousSpace = true;
		    }
		}
		else {
		    newContent.append(tempToken);
		    loc += tempToken.length();
		    map.add(loc, newContent.length());
		    previousSpace = false;
		}
	    }//end while
	    if (newContent.charAt(newContent.length() - 1) == ' ') {
		newContent.append(end + "\n");
	    }
	    else {
		newContent.append(" " + end + "\n");
	    }
	    map.add(loc, newContent.length());
	}//end for
	debug.println("NEW CONTENT: " + newContent.toString());
	//	debug.println("DUMP:");
	//	map.dump(doc.getText(), newContent.toString());
	
	return new SentenceView(doc, cookedView.getVersion(), map, newContent.toString()); 
    }



    public static void main(String[] args) {
	String content = args[0];
	debug.println("CONTENT: " + content);
	SentenceCooker lgp = new SentenceCooker("lgpTest", new Command(args), new java.util.HashMap());
	Document doc = new BasicDocument(content, "content-type");
	DocumentView cookedView = lgp.cook(doc);

	debug.println("DOC: " + doc.getText());
	debug.println("NEWDOC: " + cookedView.getText());

	PiecewiseLinearMonotonicMap map = 
	    (PiecewiseLinearMonotonicMap)cookedView.getVersion().diff(doc.getVersion());

    }

	public class SentenceView implements DocumentView {
		Document doc;
		DocumentVersion version = DocumentVersion.getOriginal(this);
		String content = null;
		
		SentenceView (Document doc, DocumentVersion lastVersion,
					  CoordinateMap map, String content) {
			this.doc = doc;
			this.version = new DocumentVersion(map, lastVersion, this);
			this.content = content;
		}
		
		public String getText() {
			return content;
		}
		
		public String getText (Region r) {
			if (content == null)
				return null;
				
			int l = content.length ();
			int s = Math.min (l, Math.max(0, r.getStart ()));
			int e = Math.min (l, Math.max(0, r.getEnd ()));
			
			return (s < e) ? content.substring(s,e) : "";
		}
		
		public Region toRegion () {
			return new Region (0,
						content != null ? content.length () : 0,
						this.version);
		}
		
		public Reader getReader () {
			return new StringReader (getText ());
		}
		
		public DocumentVersion getVersion () {
			return this.version;
		}
		
		public Document getDocument () {
			return this.doc;
		}
	}
}
