/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.parsers;

import lapis.*;

public class SystemParser implements Parser {
    /**
     * A single region representing the entire document. 
     */
    public final static String ALL = "System.All";
    
    /**
     * A region set representing all the regions in the document.
     */
    public final static String ANY = "System.Any";
    
    /**
     * The current set of background regions (typically whitespace).
     */
    public final static String BACKGROUND = "System.Background";

    
    public void bind (PatternLibrary lib) {
        lib.put (ANY, ANY_PATTERN);
        lib.put (ALL, ALL_PATTERN);
        lib.put (BACKGROUND, BACKGROUND_PATTERN);
        lib.put ("Layout.Paragraph", PARAGRAPH_PATTERN);
    }

    public void bind (PatternLibrary lib, Document doc) {
        // no doc-specific patterns for this parser
    }

    static Pattern ANY_PATTERN = new Pattern () {
        public RegionSet match (Document doc) {
            PatternCache cache = doc.getPatternCache ();
            RegionSet s = cache.get (ANY);
            if (s == null) {
                s = doc.toRegion ().in ();
                cache.put (ANY, s);
            }
            return s;
        }
    };

    static Pattern ALL_PATTERN = new Pattern () {
        public RegionSet match (Document doc) {
            PatternCache cache = doc.getPatternCache ();
            RegionSet s = cache.get (ALL);
            if (s == null) {
                s = doc.toRegion ();
                cache.put (ALL, s);
            }
            return s;
        }
    };

    static Pattern BACKGROUND_PATTERN = new Pattern () {
        public RegionSet match (Document doc) {
            PatternCache cache = doc.getPatternCache ();
            RegionSet s = cache.get (BACKGROUND);
            if (s == null) {
                PatternLibrary lib = PatternLibrary.getGlobalLibrary ();
                String contentType = 
                    (String) doc.getProperty (doc.MIMEProperty);
                if ("text/html".equals (contentType))
                    s = lib.get ("Characters._RawWhitespace").match (doc)
                        .or (lib.get ("HTML.Tag").match (doc))
                        .or (lib.get ("Characters.Token.Punctuation").match (doc))
                        .or (lib.get ("Characters.Whitespace").match (doc));
                else
                    s = lib.get ("Characters._RawWhitespace").match (doc)
                        .or (lib.get ("Characters.Token.Punctuation").match (doc));
                
                s = s.melt ();

                cache.put (BACKGROUND, s);
            }
            return s;
        }
    };

    static Pattern PARAGRAPH_PATTERN = new Pattern () {
        public RegionSet match (Document doc) {
            PatternCache cache = doc.getPatternCache ();
            RegionSet s = cache.get ("Layout.Paragraph");
            if (s == null) {
                PatternLibrary lib = PatternLibrary.getGlobalLibrary ();
                String contentType = 
                    (String) doc.getProperty (doc.MIMEProperty);
                if ("text/html".equals (contentType))
                    s = lib.get ("Layout.Paragraph._htmlParagraph").match (doc);
                else
                    s = lib.get ("Layout.Paragraph._textParagraph").match (doc);
                cache.put ("Layout.Paragraph", s);
            }
            return s;
        }
    };
}
