/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.parsers;

import lapis.*;

import java.util.*;
import java.io.*;
import org.apache.xerces.parsers.*;
import org.apache.xml.serialize.*;
import org.xml.sax.*;
import org.xml.sax.ext.*;

/**
 *
 * Parser that uses classes from the Xerces XML parser to create an
 * XML parser that integrates into LAPIS.
 *
 * NOTE: Remaining issues include ordering of attributes and empty
 * elements.
 * 
 */
public class XMLParser implements lapis.Parser {

    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;


    // the document that this parser is parsing
    private lapis.Document pdoc;
    private String content = null;

    // region sets found by the parser
    private MutableRegionSet elements = new MutableRegionSet();
    private MutableRegionSet attributes = new MutableRegionSet();
    private MutableRegionSet attributeValues = new MutableRegionSet();
    private MutableRegionSet attributeNames = new MutableRegionSet();
    //private MutableRegionSet elementTexts = new MutableRegionSet();
    private MutableRegionSet tags = new MutableRegionSet();
    private MutableRegionSet startTags = new MutableRegionSet();
    private MutableRegionSet endTags = new MutableRegionSet();
    private MutableRegionSet comments = new MutableRegionSet();

    // doc-specific region sets
    private HashMap docSpecificSets = new HashMap();

    // stack of start points for elements
    int[] elementStack = new int[8]; // grown automatically as needed
    
    // index of top of elementStack
    int currElem = -1;

    // Pattern stubs for each name defined by this parser.
    static class PatternStub implements Pattern {
        String name;

        public PatternStub(String name) {
            this.name = name;
        }

        public RegionSet match(lapis.Document doc) {
            PatternCache cache = doc.getPatternCache();
            if (!cache.wasParsedBy("lapis.parsers.LapisXMLParser"))
                new XMLParser().parse(doc);
            RegionSet s = cache.get(name);
            if (s == null) {
                debug.println("LapisXMLParser ran but didn't create " + this);
                s = RegionSet.EMPTY;
            }
            return s;
        }

        public String toString() {
            return name;
        }
    }

    static PatternStub ELEMENT = new PatternStub("XML.Element");
    static PatternStub ATTRIBUTE = new PatternStub("XML.Attribute");
    static PatternStub ATTRIBUTE_NAME =
        new PatternStub("XML.Attribute.Name");
    static PatternStub ATTRIBUTE_VALUE =
        new PatternStub("XML.Attribute.Value");
    static PatternStub ELEMENT_TEXT = new PatternStub("XML.Element.Text");
    static PatternStub TAG = new PatternStub("XML.Tag");
    static PatternStub START_TAG = new PatternStub("XML.Tag.StartTag");
    static PatternStub END_TAG = new PatternStub("XML.Tag.EndTag");
    static PatternStub COMMENT = new PatternStub("XML.Comment");

    //
    // Bind pattern stubs to names.
    //
    public void bind(PatternLibrary lib) {
        debug.println("binding XML patterns");
        lib.put (ELEMENT.name, ELEMENT);
        lib.put (ATTRIBUTE.name, ATTRIBUTE);
        lib.put(ATTRIBUTE_NAME.name, ATTRIBUTE_NAME);
        lib.put(ATTRIBUTE_VALUE.name, ATTRIBUTE_VALUE);
        //	lib.put (ELEMENT_TEXT.name, ELEMENT_TEXT);
        lib.put(TAG.name, TAG);
        lib.put(START_TAG.name, START_TAG);
        lib.put(END_TAG.name, END_TAG);
        lib.put(COMMENT.name, COMMENT);
    }

    public void bind(PatternLibrary lib, Document doc) {
        new XMLParser().parse(doc);
    }

    public void parse(lapis.Document doc) {
        pdoc = doc;
        long start = System.currentTimeMillis();
        try {
            PatternCache cache = doc.getPatternCache();

            String content = doc.getText();

            Main.getProgress().fireStartProgress("XML", content.length());
            parse(content);

            cache.put(ELEMENT.name, elements);
            cache.put(ATTRIBUTE.name, attributes);
            cache.put(ATTRIBUTE_NAME.name, attributeNames);
            cache.put(ATTRIBUTE_VALUE.name, attributeValues);
            //cache.put(ELEMENT_TEXT.name, elementTexts);
            cache.put(TAG.name, tags);
            cache.put(START_TAG.name, startTags);
            cache.put(END_TAG.name, endTags);
            cache.put(COMMENT.name, comments);

            PatternLibrary lib = PatternLibrary.getDocumentLibrary(doc);
 
            Iterator regionSets = docSpecificSets.keySet().iterator();
            while (regionSets.hasNext()) {
                String patternStr = (String) regionSets.next();
                MutableRegionSet set =
                    (MutableRegionSet) docSpecificSets.get(patternStr);
                lib.put(patternStr, new PatternStub(patternStr));
                cache.put(patternStr, set);
            }

            cache.setParsedBy("lapis.parsers.LapisXMLParser");
        } finally {
            Main.getProgress().fireDoneProgress();
            debug.println(
                "XML parser ran in "
                    + (System.currentTimeMillis() - start)
                    + " ms");
        }
    }

    private void parse(String content) {
        this.content = content;
        try {
            // create canonical content
            StringWriter stringWriter = new StringWriter();

            SAXParser saxParser = new SAXParser();
            saxParser.setProperty("LapisXMLParser", this);
            saxParser.parse(new InputSource(new StringReader(content)));
        } catch (Exception e) {
            debug.println("INVALID XML content");
            //debug.printStackTrace();
        }
    }

    private String fixAndTrim(int start, int end) {
        String temp = content.substring(start + 1, end - 1).trim();
        debug.println(
            "BEFORE: " + content.substring(start, end) + " TRIM: " + temp);
        return temp;
    }

    private String removeSlash(String raw) {
        //remove '/'
        StringBuffer sb = new StringBuffer(raw);
        int loc = raw.indexOf('/');
        if (loc != -1) {
            sb.deleteCharAt(loc);
        }
        return sb.toString();
    }

    private void insertDocSpecificPattern(
        String patternStr,
        int start,
        int end) {
        debug.println(
            "ADDING region (P: " + patternStr + " " + start + "," + end + ")");
        if (docSpecificSets.get(patternStr) == null) {
            docSpecificSets.put(patternStr, new MutableRegionSet());
        }
        MutableRegionSet set =
            (MutableRegionSet) docSpecificSets.get(patternStr);
        debug.assertion(pdoc != null);
        set.insert(start, end, pdoc);
    }

    public void startTag(String tagname, int start, int end) {
        debug.println ("startTag(" + tagname + "," + start + "," + end + "): " + content.substring (start, end));
        debug.assertion(pdoc != null);
        startTags.insert(start, end, pdoc);
        tags.insert(start, end, pdoc);
        insertDocSpecificPattern(
            "XML.Tag.StartTag.<" + tagname + ">",
            start,
            end);
        
        // push the start point on the element stack     
        ++currElem;
        if (currElem == elementStack.length) {
            int[] newStack = new int[elementStack.length * 2];
            System.arraycopy (elementStack, 0, newStack, 0, elementStack.length);
            elementStack = newStack;
        }
        elementStack[currElem] = start;
    }
    public void endTag(String tagname, int start, int end) {
        debug.println ("endTag(" + tagname + "," + start + "," + end + "): " + content.substring (start, end));
        debug.assertion(pdoc != null);
        endTags.insert(start, end, pdoc);
        tags.insert(start, end, pdoc);
        insertDocSpecificPattern(
            "XML.Tag.EndTag.</" + tagname + ">",
            start,
            end);
            
        // pop the whole element's start point
        debug.assertion (currElem >= 0);
        int elementStart = elementStack[currElem--];
        elements.insert (elementStart, end, pdoc);
        insertDocSpecificPattern ("XML.Element.[" + tagname + "]", elementStart, end);
    }
    public void attributeName(int start, int end) {
        debug.println ("attributeName(" + start + "," + end + "): " + content.substring (start, end));
        debug.assertion(pdoc != null);
        attributeNames.insert(start, end, pdoc);
    }

    public void attributeValue(
        int start,
        int end,
        int startName,
        int endName) {
        String name = content.substring(startName, endName).trim();
        
        debug.println ("attributeValue(" + start + "," + end + "," + startName + "," + endName + "): " + content.substring (start, end));
        debug.assertion(pdoc != null);
        
        attributeValues.insert(start, end, pdoc);
        attributes.insert (startName, end, pdoc);
        insertDocSpecificPattern("XML.Attribute.@" + name, startName, end);
        
    }

    public void comment(int start, int end) {
        debug.println ("comment(" + start + "," + end + "): " + content.substring (start, end));
        debug.assertion(pdoc != null);
        comments.insert(start, end, pdoc);
    }

    public static void main(String[] args) throws Exception {
        SAXParser saxParser = new SAXParser();

        OutputFormat of = new OutputFormat(Method.XML, "UTF-8", false);
        of.setOmitXMLDeclaration(true);
        of.setPreserveEmptyAttributes(true);
        of.setPreserveSpace(true);
        of.setOmitComments(false);

        XMLSerializer xmlSerializer = new XMLSerializer(System.out, of);

        saxParser.setContentHandler(xmlSerializer.asContentHandler());
        saxParser.setProperty(
            "http://xml.org/sax/properties/lexical-handler",
            (LexicalHandler) xmlSerializer);

        saxParser.parse(
            new InputSource(new FileInputStream(new File(args[0]))));
    }

} //end static ExternalParserSAXHandler class

// ISSUES:
//
// X aggregate Text regions
// - eliminate? whitespace
// - figure out empty elements
// - EMPTYTag should be "StartTag contains EndTag"
