/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.swing;

import lapis.*;
import lapis.Document; // override ambiguity with javax.swing.text.Document
import lapis.tc.*;
import lapis.swing.cluster.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import lapisx.swing.*;
import java.io.*;
import java.net.*;
import lapisx.net.*;
import java.util.*;
import lapisx.util.*;
import lapisx.progress.*;
import lapisx.progress.swing.*;


public class Browser extends JFrame implements Tool{ 
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 


    ////////////////////
    //////////////////// Static initialization
    ////////////////////

    static {
        SwingCompatibility.disableBuggyFocusManager ();
        SwingConfig.restoreLookAndFeel ();
    }


    ////////////////////
    //////////////////// UI components 
    ////////////////////

    /** menu and toolbar */
    private JMenuBar menubar;
    private JToolBar toolbar;

    /** busy animation (a spinning L) */
    private Animation animation;

    /** Tools menu (stored here because new tools can be added dynamically) */
    private JMenu toolMenu;

    /** NamedRegionSet mode switches: menu items and toolbar buttons */
    private JRadioButtonMenuItem manualRadioButtonMenuItem;
    private JRadioButtonMenuItem simulEditRadioButtonMenuItem;
    private ButtonGroup selectionMenuGroup;

    private JToggleButton manualToggleButton;
    private JToggleButton simulEditToggleButton;
    private ButtonGroup selectionButtonGroup;

    /** Command (URL, filename, Tcl command) */
    private JComboBox command;

    /** Content type (text, HTML, Java, etc.) */
    private JComboBox view;
    
    /** Browser panes */
    private BrowserPane stdoutPane;
    private BrowserPane stderrPane;
    private BrowserScrollPane stdoutScrollPane;
    private BrowserScrollPane stderrScrollPane;

    /** Horizontal splitter between browser panes */
    private JSplitPane browserSplitter;

    /** Sidebar */
    private DisappearingTabbedPane sidebar;

    /** Sidebar pane containing pattern box and pattern library tree */
    private PatternPane patternPane;

    /** Vertical splitter between sidebar and browserSplitter */
    private JSplitPane sidebarSplitter;

    /** Sidebar tab containing pattern editor (null until first requested) */
    private TCFileEditor tcEditor;

    /** Sidebar tab containing script editor (null until first requested) */
    private TclFileEditor tclEditor;

    /** Status bar components: progress, status message, feedback button */
    private SwingProgressBar progressBar;
    private FlashingLabel statusBar;
    private JButton feedbackButton;


    ////////////////////
    //////////////////// Auxilliary windows
    ////////////////////

    /** Outlier finder (null until first requested) */
    private OutlierFinder outlierFinder;

    /** Debugging windows (null until first requested) */
    private DocumentProperties docProps;
    private RegionSetDisplay regionSetDisplay;


    ////////////////////
    //////////////////// Unchanging, non-UI objects
    ////////////////////

    /** Manual selection mode */
    private final ManualMode manualSelection = new ManualMode (this);

    /** Progress manager */
    private final BasicProgressGenerator progress = 
        new BasicProgressGenerator (this, true);

    /** Background activity manager */
    private final BackgroundThreads backgroundThreads =
        new BackgroundThreads (this, progress);

    /** Listeners attached to this browser */
    private final Set browserListeners = new SafeIteratorSet ();
    private final Set selectionListeners = new SafeIteratorSet ();

    ////////////////////
    //////////////////// Mutable state
    ////////////////////

    /** Page history (History[Document]) */
    private History pageHistory;

    /** Command history (History[String]) */
    private HistoryComboBoxModel commandHistory;

    /** Interpreter used to evaluate commands. */
    private Interpreter interpreter;

    /** Browser pane with the focus (either stdoutPane or stderrPane) */
    private BrowserPane activePane;

    /** Current selection */
    private NamedRegionSet selection;

    /** Current selection mode */
    private SelectionMode selectionMode = manualSelection;

    /** Enumeration through selection, or null if none in progress */
    private RegionEnumeration findEnumerator;

    /** reference counts for setBusy() and setHourglass */
    private int busyCount;
    private int hourglassCount;

    ////////////////////
    //////////////////// Help viewer (shared among all browsers)
    ////////////////////

    private static Browser helpViewer;

    ////////////////////
    //////////////////// Actions & listeners defined elsewhere
    ////////////////////

    private final Action newAction = new FileAction.New (this);
    private final Action newBrowserAction = new FileAction.NewBrowser (this);
    private final Action openAction = new FileAction.Open (this);
    private final Action saveAction = new FileAction.Save (this);
    private final Action saveAsAction = new FileAction.SaveAs (this);
    private final Action closeAction = new FileAction.Close (this);
    private final Action exitAction = new FileAction.Exit (this);

    private final Action extractAction = new ToolAction.Extract (this);
    private final Action sortAction = new ToolAction.Sort (this);
    private final Action keepAction = new ToolAction.Keep (this);
    private final Action omitAction = new ToolAction.Omit (this);
    private final Action replaceAction = new ToolAction.Replace (this);
    private final Action calcAction = new ToolAction.Calc (this);

    private final HyperlinkListener hyperlinkListener = 
	new BrowserAction.Hyperlink (this);
    private final Toolbox.ToolBoxListener toolboxListener =  
	new ToolAction.ToolboxChange(this);

    private final Action commandAction = 
	new BrowserAction.Command (this);
    private final Action homeAction = 
	new BrowserAction.Home (this);
    private final BrowserAction.Back backAction = 
	new BrowserAction.Back (this);
    private final BrowserAction.Forward forwardAction = 
	new BrowserAction.Forward (this);
    private final Action reloadAction = 
	new BrowserAction.Reload (this);
    private final PopupMenuListener goMenuListener = 
	new BrowserAction.HistoryMenuListener (this);

    private final EditAction.Cut cutAction = 
        new EditAction.Cut (this);
    private final EditAction.Copy copyAction = 
        new EditAction.Copy (this);
    private final EditAction.Paste pasteAction = 
        new EditAction.Paste (this);
    private final EditAction.Delete deleteAction = 
        new EditAction.Delete (this);
    private final EditAction.SelectAll selectAllAction = 
        new EditAction.SelectAll (this);
    private final EditAction.Undo undoAction = 
        new EditAction.Undo (this);
    private final EditAction.Redo redoAction = 
        new EditAction.Redo (this);
    private final MenuListener editMenuListener = 
        new EditAction.EditMenuListener (this,
                                         cutAction, copyAction, pasteAction,
                                         deleteAction, undoAction, redoAction);
    
    private final Action downcaseAction = 
	new TransformAction.Downcase (this);
    private final Action upcaseAction = 
	new TransformAction.Upcase (this);
    private final Action sentenceCaseAction = 
	new TransformAction.SentenceCase (this);
    private final Action titleCaseAction = 
	new TransformAction.TitleCase (this);
    private final Action toggleCaseAction = 
	new TransformAction.ToggleCase (this);

    private final Action preferencesAction = 
	new PreferencesDialog.Preferences (this); 

    private final Action manualAction = 
	new ManualMode.ManualAction (this);
    private final Action simulEditAction = 
	new SimulEditMode.SimulEditAction (this);

    private final SimulEditMode.IgnoreOutlierAction ignoreOutlierAction
        = new SimulEditMode.IgnoreOutlierAction (this);

    private final ViewAsOutlierAction viewAsOutlierAction
        = new ViewAsOutlierAction ();

    private final UnselectAction unselectAction = 
	new UnselectAction ();


    private final Action findReplaceAction = 
	new FindReplacePane.FindReplaceAction(this);
    private final Action clusterFindReplaceAction = 
	new ClusterFindReplacePane.ClusterFindReplaceAction(this);
//     private final Action miniBrowserAction = 
// 	new MiniBrowser.MiniBrowserAction(this);

    ////////////////////
    //////////////////// Constructor
    ////////////////////

    /**
     * Make a new Browser.
     */
    public Browser () {
        super ("LAPIS");



	debug.println ("created new browser window");
	Main.windowCreated (this);

	// set window icon
	setIconImage (SwingConfig.makeImage ("lapis-icon.gif"));

        // Handle window-close operation
        setDefaultCloseOperation (DO_NOTHING_ON_CLOSE);
        addWindowListener (new WindowAdapter () {
            public void windowClosing (WindowEvent evt) {
                dispose ();
            }
        });

	Toolbox toolbox = Toolbox.getDefaultToolbox();
	toolbox.addToolBoxListener(toolboxListener);

        // Create and lay out the window's components
        makeGUI ();

        // Set default window size and position
        Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
        d.width -= 100;
        d.height -= 100;
        setSize (d);
        Win.center (this, null);

        // Restore saved window size and position, if any 
        restoreWindowPositions ();

        // Activate stdout pane
        activatePane (stdoutPane, false);

        // Clear the selection
        clearSelection ();

        // Reset progress UI 
        clearBusy ();
        clearHourglass ();
        clearStatus ();
    }

    ////////////////////
    //////////////////// Making UI components
    ////////////////////

    private void makeGUI () {
        // Set layout manager
        Layout mainLayout = Layout.vertical ((JComponent)getContentPane ());

        // Create the page history
        int pageHistorySize = Integer.parseInt 
            (Main.getConfig ().getProperty ("shell.page-history-size"));
        pageHistory = new History (pageHistorySize);

        // Create the menubar
        menubar = makeMenuBar ();
        setJMenuBar (menubar);

        // Create the toolbar and busy animation
        toolbar = makeToolBar ();

        animation = new Animation 
            (SwingConfig.makeImage ("busy-animated.gif"), 
             SwingConfig.makeImage ("busy-frozen.gif"),
             false);

        mainLayout.add (Layout.horizontal ()
                        .add (toolbar)
                        .addGlue ()
                        .add (animation, Layout.FIXED_SIZE, 
			      new Insets (5, 2, 5, 2))
                        .done (),
                        Layout.FIXED_HEIGHT);


        // add Toolbox tools to menu and toolbar
        for (Enumeration tools = Toolbox.getDefaultToolbox ().tools ();
             tools.hasMoreElements (); ) {
            Toolbox.Info info = (Toolbox.Info) tools.nextElement ();
	    addToToolBar(info);
        }


	
        // Create the Command/View As controls
        int cmdHistorySize = Integer.parseInt 
            (Main.getConfig ().getProperty ("browser.command-history-size"));
        commandHistory = new HistoryComboBoxModel (cmdHistorySize);
        command = new JComboBox (commandHistory);
        command.setEditable (true);
        CommandEditor editor = new CommandEditor ();
        command.setEditor (editor);
        editor.addActionListener (commandAction);

        view = new JComboBox ();
        view.addItem (new ViewType ("Plain Text", "text/plain"));
        view.addItem (new ViewType ("Web Page", "text/html"));
        view.addActionListener (changeViewAction);

        mainLayout.add (Layout.gridbag ()
                        .setContainerInsets (1, 1, 1, 1)
                        .add (new JLabel ("Command: "), Layout.FIXED_WIDTH)
                        .add (command)
                        .add (new JLabel (" View As: "), Layout.FIXED_WIDTH)
                        .add (view, Layout.FIXED_WIDTH)
                        .done (),
                        Layout.FIXED_HEIGHT);

        // Create the browser pane and its decorations
        stdoutPane = makeBrowserPane ();
        stdoutScrollPane = new BrowserScrollPane (stdoutPane, "Output:");
        stdoutScrollPane.setTitled (false);

        // Create the error pane and its decorations
        stderrPane = makeBrowserPane ();
        stderrScrollPane = new BrowserScrollPane (stderrPane, "Error:");
        stderrScrollPane.setTitled (true);

        // Horizontal splitter separating stdout & stderr panes
        browserSplitter = 
            new JSplitPane (JSplitPane.VERTICAL_SPLIT,
                            stdoutScrollPane, 
                            stderrScrollPane);
        browserSplitter.setOneTouchExpandable (true);
        browserSplitter.setContinuousLayout (false);
        browserSplitter.setResizeWeight (1.0);

        // Sidebar, with initially one tab: the pattern pane
        patternPane = new PatternPane (this);
        ((JComponent)getContentPane ()).registerKeyboardAction 
            (nextMatchAction, 
             KeyStroke.getKeyStroke (KeyEvent.VK_F3, 0),
             JComponent.WHEN_IN_FOCUSED_WINDOW);

        JTabbedPane tabs = new JTabbedPane (SwingConstants.TOP);
        SwingCompatibility.setScrollTabLayoutPolicy (tabs);
        sidebar = new DisappearingTabbedPane (tabs);
        sidebar.addTab ("Patterns", patternPane);

        // Vertical splitter separating sidebar from browser panes
        sidebarSplitter = new JSplitPane (JSplitPane.HORIZONTAL_SPLIT,
                                      browserSplitter, sidebar);
        sidebarSplitter.setOneTouchExpandable (true);
        sidebarSplitter.setContinuousLayout (false);
        sidebarSplitter.setResizeWeight (0.5);

        mainLayout.add (sidebarSplitter);

        // Create the status panel
        progressBar = new BrowserProgressBar (this);
        progress.addProgressListener (progressBar);

        statusBar = new FlashingLabel ("Initializing...");

        feedbackButton = new JButton ("Feedback");
        feedbackButton.addActionListener (feedbackAction);

        mainLayout.add (Layout.horizontal ()
                        .setContainerInsets (1, 1, 1, 1)
                        .add (progressBar, Layout.FIXED_SIZE)
                        .add (statusBar, new Insets (0, 3, 0, 0))
                        .addGlue ()
                        .add (feedbackButton, Layout.FIXED_WIDTH)
                        .done (),
                        Layout.FIXED_HEIGHT);
    }

    private JMenuBar makeMenuBar () {
        JMenuBar menubar = new JMenuBar ();
        JMenuItem item;

        JMenu fileMenu = new AcceleratedMenu ("&File");
        fileMenu.add (ActionAdapters.toMenuAction (newAction));
        fileMenu.add (ActionAdapters.toMenuAction (openAction));
        fileMenu.add (ActionAdapters.toMenuAction (saveAction));
        fileMenu.add (saveAsAction);
        fileMenu.addSeparator ();
        fileMenu.add (newBrowserAction);
        fileMenu.add (closeAction);
        fileMenu.add (exitAction);
        menubar.add (fileMenu);

        JMenu editMenu = new AcceleratedMenu ("&Edit");
        editMenu.add (undoAction);
        editMenu.add (redoAction);
        editMenu.addSeparator ();
        editMenu.add (ActionAdapters.toMenuAction (cutAction));
        editMenu.add (ActionAdapters.toMenuAction (copyAction));
        editMenu.add (ActionAdapters.toMenuAction (pasteAction));
        editMenu.add (deleteAction);
        editMenu.add (selectAllAction);
        
        JMenu caseMenu = new AcceleratedMenu ("C&hange Case");
        caseMenu.add (downcaseAction);
        caseMenu.add (upcaseAction);
        caseMenu.add (sentenceCaseAction);
        caseMenu.add (titleCaseAction);
        caseMenu.add (toggleCaseAction);
        editMenu.add (caseMenu);

        editMenu.addMenuListener (editMenuListener);
                
	editMenu.addSeparator ();
	editMenu.add (findReplaceAction);
	editMenu.add (clusterFindReplaceAction);
        
        editMenu.addSeparator ();
        editMenu.add (preferencesAction);
        menubar.add (editMenu);

        JMenu goMenu = new AcceleratedMenu ("&Go");

        backAction.setEnabled (false);
        forwardAction.setEnabled (false);
        goMenu.add (ActionAdapters.toMenuAction (backAction));
        goMenu.add (ActionAdapters.toMenuAction (forwardAction));
        goMenu.add (ActionAdapters.toMenuAction (reloadAction));
        goMenu.add (ActionAdapters.toMenuAction (homeAction));
        goMenu.add (ActionAdapters.toMenuAction (stopAction));
        goMenu.addSeparator ();
        goMenu.getPopupMenu ().addPopupMenuListener (goMenuListener);

        menubar.add (goMenu);

        JMenu selectionMenu = new AcceleratedMenu ("&Selection");
        selectionMenuGroup = new ButtonGroup ();
        selectionMenu.add (manualRadioButtonMenuItem = 
                           ActionAdapters.toRadioButtonMenuItem (manualAction, 
								 selectionMenuGroup));
        selectionMenu.add (simulEditRadioButtonMenuItem = 
                           ActionAdapters.toRadioButtonMenuItem (simulEditAction, 
								 selectionMenuGroup));
        manualRadioButtonMenuItem.setSelected (true);

        selectionMenu.addSeparator ();
        selectionMenu.add (findOutliersAction);
        selectionMenu.add (patternEditorAction);
        menubar.add (selectionMenu);

        JMenu scriptMenu = new AcceleratedMenu ("S&cripts");
        scriptMenu.add (scriptEditorAction);
        scriptMenu.add (historyScriptAction);
        menubar.add (scriptMenu);

        toolMenu = new AcceleratedMenu ("&Tools");
        toolMenu.add (extractAction);
        toolMenu.add (sortAction);
        toolMenu.add (keepAction);
        toolMenu.add (omitAction);
        toolMenu.add (replaceAction);
        toolMenu.add (calcAction);
        toolMenu.addSeparator ();
        menubar.add (toolMenu);

        JMenu debugMenu = new AcceleratedMenu ("&Debug");
        debugMenu.add (documentPropertiesAction);
        debugMenu.add (regionSetDisplayAction);
        debugMenu.add (highlightOptionAction);
        menubar.add (debugMenu);

        JMenu helpMenu = new AcceleratedMenu ("&Help");
        helpMenu.add (new HelpAction ("Quick Start Tutorial",
                                      "quickstart.html"));
        helpMenu.add (new HelpAction ("Pattern Language Reference",
                                      "doc/patterns/reference.html"));
        helpMenu.add (new HelpAction ("Command Reference",
                                      "doc/commands/reference.html"));
        helpMenu.addSeparator ();
        helpMenu.add (new HelpAction ("LAPIS Website",
                                      "http://graphics.lcs.mit.edu/lapis/"));
        helpMenu.add (feedbackAction);
        helpMenu.addSeparator ();
        helpMenu.add (aboutAction);
        menubar.add (helpMenu);

        return menubar;
    }

    private JToolBar makeToolBar () {
        JToolBar toolbar = new JToolBar (JToolBar.HORIZONTAL);
        toolbar.setFloatable (false);

        JButton backButton = 
            toolbar.add (ActionAdapters.toToolbarAction (backAction));
        backAction.attachMenu (backButton);

        JButton forwardButton = 
            toolbar.add (ActionAdapters.toToolbarAction (forwardAction));
        forwardAction.attachMenu (forwardButton);

        toolbar.add (ActionAdapters.toToolbarAction (reloadAction));
        toolbar.add (ActionAdapters.toToolbarAction (homeAction));
        
        JButton stopButton = toolbar.add (ActionAdapters.toToolbarAction (stopAction));
        stopButton.setIcon (SwingConfig.makeIcon ("stop.gif"));
        stopButton.setDisabledIcon (SwingConfig.makeIcon ("gray-stop.gif"));

        toolbar.addSeparator ();

        toolbar.add (ActionAdapters.toToolbarAction (newAction));
        toolbar.add (ActionAdapters.toToolbarAction (openAction));
        toolbar.add (ActionAdapters.toToolbarAction (saveAction));

        toolbar.addSeparator ();

        toolbar.add (ActionAdapters.toToolbarAction (cutAction));
        toolbar.add (ActionAdapters.toToolbarAction (copyAction));
        toolbar.add (ActionAdapters.toToolbarAction (pasteAction));

        toolbar.addSeparator ();

        selectionButtonGroup = new ButtonGroup ();
        toolbar.add (manualToggleButton = 
                     ActionAdapters.toToggleButton (manualAction, 
						    selectionButtonGroup));
        toolbar.add (simulEditToggleButton =
                     ActionAdapters.toToggleButton (simulEditAction, 
						    selectionButtonGroup));
        manualToggleButton.setSelected (true);

        return toolbar;
    }

    // adds new tool to ToolBar and ToolMenu
    public void addToToolBar (Toolbox.Info info) {
	    debug.println("Adding " + info.getName() + " to toolbar.");
            Action a = new ToolAction.ToolboxAction (this, info);
            if (info.getOnMenu ())
                toolMenu.add (a);
            if (info.getOnToolbar ())
                toolbar.add (a);         
    }

    // removes specified tool from ToolBar and Tool Menu
    public void removeFromTools(Toolbox.Info info) {

	debug.println("removing " + info.getName() + " from toolbar.");
	if (info.getOnMenu ()){
	    removeFromToolMenu(info);
	}
	if (info.getOnToolbar ()){
	    removeFromToolBar(info);
	}
    }

    /**
     * removes specified Tool from ToolMenu 
     * returns the index of the Tool removed if tool exists in Menu 
     * else returns -1 
     **/

    private int removeFromToolMenu(Toolbox.Info info) { 
	Component[] components = toolMenu.getMenuComponents();
	debug.println("" + components.length + " components on toolMenu");
	for (int i = 0; i < components.length ; i++ ) {
	    
	    if (components[i] instanceof JMenuItem) {
		debug.println(((JMenuItem)components[i]).getText());
		if (info.getName().equals(((JMenuItem)components[i]).getText())){
		    toolMenu.remove(i);
		    return i;
		}
	    }
	    else {
		debug.println(components[i].getClass().getName());
	    }
	}
	return -1;
    }

    /**
     * @removes specified Tool from ToolBar 
     * @returns the index of the Tool removed if tool exists in Menu 
     * else returns -1
     **/

    private int removeFromToolBar(Toolbox.Info info){
	Component[] components = toolbar.getComponents();
	for (int i = 0; i < components.length ; i++ ) {
	    if (components[i] instanceof JButton) {
		// debug.println(((JButton)components[i]).getText());
		if (info.getName().equals(((JButton)components[i]).getText())){
		    toolbar.remove(i);
		    return i;
		}
	    }
	}
	return -1;
	
    }

    public void undo (ActionEvent e) {
	undoAction.actionPerformed(e);
    } 

    /** 
    * changes the action associated with the Tool name. Tool should
    * remain in the same place on both the ToolMenu and the ToolBar
    **/

    public void changeInTool(Toolbox.Info info){
	debug.println("changing " + info.getName() + " on toolbar.");
	Action a = new ToolAction.ToolboxAction (this, info);
	if (info.getOnMenu ()){
	    int i = removeFromToolMenu(info);
	    if (i != -1)
		toolMenu.insert(a,i);
	    else 
		toolMenu.add(a);
	}
	if (info.getOnToolbar ()){
	    int i = removeFromToolBar(info);
	    if (i != -1)
		toolbar.add(new JButton(a), i);
	    else 
		toolbar.add(a);
	}

    }



    private BrowserPane makeBrowserPane () {
        BrowserPane bp = new BrowserPane (this);
        bp.addHyperlinkListener (hyperlinkListener);
        bp.addMouseListener (rightClickListener);
        bp.addFocusListener (paneFocusListener);
        return bp;
    }


    ////////////////////
    //////////////////// Configuration
    ////////////////////

    /**
     * Save this browser's window location, size, and splitter positions.
     * @effects stores location and size information in Main.config
     * and then saves the config to disk.
     */
    public void saveWindowPositions () {
        if (SwingConfig.inDemoMode ())
            return; // don't overwrite existing settings

        SwingConfig.saveLocation ("browser.location", this);
        SwingConfig.saveSize ("browser.size", this);
        SwingConfig.saveSplitter ("browser.midSplitter", sidebarSplitter);
        SwingConfig.saveSplitter ("browser.errorSplitter", browserSplitter);
        patternPane.saveWindowPositions ();
        Main.getConfig ().save ();
    }

    /**
     * Load window location, size, and splitter positions from configuration.
     * @effects moves and resizes window according to saved location and size
     * found in Main.config.
     */
    public void restoreWindowPositions () {
        SwingConfig.restoreLocation ("browser.location", this);
        SwingConfig.restoreSize ("browser.size", this);
        SwingConfig.restoreSplitter ("browser.midSplitter", sidebarSplitter);
        SwingConfig.restoreSplitter ("browser.errorSplitter", browserSplitter);
        patternPane.restoreWindowPositions ();
    }


    ////////////////////
    //////////////////// Closing the browser
    ////////////////////

    public void dispose () {
        if (!FileAction.okToClear (this, getPageHistory ().elements ()))
            return;

        saveWindowPositions ();

        disposeIfNonNull (outlierFinder);
        disposeIfNonNull (docProps);
        disposeIfNonNull (regionSetDisplay);
        
        if (this == helpViewer)
            helpViewer = null;
        
        super.dispose ();

        Main.windowDisposed (Browser.this);
    }

    private void disposeIfNonNull (Window w) {
        if (w != null)
            w.dispose ();
    }


    ////////////////////
    //////////////////// Debugging
    ////////////////////

    private Action documentPropertiesAction = new AcceleratedAction ("&Document Properties") {
        public void actionPerformed (ActionEvent evt) {
            if (docProps == null) {
                docProps = new DocumentProperties ();
                addBrowserListener (docProps);
                Win.position (docProps, Browser.this, 1.5, 0.1);
            }
            docProps.setVisible (true);
            docProps.setDocument (getDocument ());
        }
    };

    private Action regionSetDisplayAction = new AcceleratedAction ("&Region Set") {
        public void actionPerformed (ActionEvent evt) {
            if (regionSetDisplay == null) {
                regionSetDisplay = new RegionSetDisplay ();
                addSelectionListener (regionSetDisplay);
                Win.position (regionSetDisplay, Browser.this, -1.5, 0.5);
            }
            regionSetDisplay.setVisible (true);
            regionSetDisplay.setSelection (getSelection ());
        }
    };

    private Action highlightOptionAction = new AcceleratedAction ("&Highlight Options") {
        public void actionPerformed (ActionEvent evt) {
            getActivePane ().getFindAndReplaceHighlighter().showOptionDialog (Browser.this);
        }
    };

    ////////////////////
    //////////////////// Help & feedback
    ////////////////////

    class HelpAction extends AcceleratedAction {
	URL url;

	public HelpAction (String caption, String href) {
	    super (caption);
            url = SwingConfig.makeURL (href);
	}

        public void actionPerformed (ActionEvent evt) {
            if (helpViewer == null) {
                helpViewer = new Browser ();
                Win.position (helpViewer, Browser.this, 1.15, 1.15);
            }

            helpViewer.setVisible (true);
            helpViewer.eval (url.toString ());
        }
    }

    private Action aboutAction = new AcceleratedAction ("&About LAPIS") {
        public void actionPerformed (ActionEvent evt) {
            try {
                URL aboutURL = SwingConfig.makeURL ("doc/messages/about.html");
                JEditorPane aboutMsg = new JEditorPane (aboutURL);
                aboutMsg.setOpaque (false);
                aboutMsg.setEditable (false);
                JScrollPane aboutScroll = new JScrollPane (aboutMsg);
                aboutScroll.setPreferredSize (new Dimension (500, 400));

                switch (JOptionPane.showOptionDialog
                        (Browser.this, 
                         aboutScroll, 
                         "About LAPIS",
                         JOptionPane.DEFAULT_OPTION, 
                         JOptionPane.PLAIN_MESSAGE, 
                         null,
                         new String[] { "OK", "Feedback" },
                         "OK")) {
                case JOptionPane.CLOSED_OPTION:
                case 0:     // OK
                    break;
                case 1:     // Send Feedback
                    FeedbackForm.showFeedbackForm (Browser.this);
                    break;
                }

            } catch (IOException e) {
                FeedbackForm.showExceptionDialog (Browser.this, e);
            }
        }
    };


    private Action feedbackAction = new AcceleratedAction ("&Feedback and Bug Reports") {
        public void actionPerformed (ActionEvent evt) {
            FeedbackForm.showFeedbackForm (Browser.this);
        }
    };

    ////////////////////
    //////////////////// Status bar
    ////////////////////

    /**
     * Set status bar message.
     * @param message Message to display in status bar
     * @effects Displays <em>message</em> in status bar.
     */
    public void setStatus (String message) {
        if (statusBar != null)
            statusBar.setText (message);
    }

    /**
      * Flash the status bar message.
      * @effects Causes status bar to flash briefly to call attention to itself.
      */
     public void flashStatus () {
         statusBar.startFlashing ();
     }

    /**
     * Clears status bar message.
     * @effects Displays no message in status bar.
     */
    public void clearStatus () {
        if (SwingUtilities.isEventDispatchThread ()) {
            setStatus (" ");
            progressBar.setValue (0);
        } else 
            SwingUtilities.invokeLater (new Runnable () {
                    public void run () {
                        setStatus (" ");
                        progressBar.setValue (0);
                    }
                });
    }

    ////////////////////
    //////////////////// Jumping to a URL
    ////////////////////

    /**
      * Jump to a URL in the background.
      * @param url URL to visit
      * @effects fetches URL in a background thread.  If the fetch
      * succeeds, the resulting page becomes the new current document.
      */
    public void gotoURL (URL url) {
        String cmd = getInterpreter ().quote (url.toString ());
        eval (cmd);
    }


    ////////////////////
    //////////////////// Evaluating Tcl commands
    ////////////////////

    /**
      * Evaluate a script in the background.
      * @param script script code to evaluate (may be multiple lines)
      * @effects evaluates script on browser's current document in a
      * background thread.  If the script returns a result, that result
      * becomes the new current document.
      */
    public void eval (final String script) {
        stop ();

        setStatus ("Evaluating command...");

        final Document input = getDocument ();
        run (new RunnableWithExceptions () {
                public void run () throws Exception {
                    final Document output = evalNow (script, input);
                    clearStatus ();
                    SwingUtilities.invokeLater (new Runnable () {
                            public void run () {
                                if (output != null) {
                                    setDocument (output);
                                } else {
                                    setStatus ("Command returned no data.");
                                    flashStatus ();
                                }
                            }
                        });
                }
            });
    }

    /**
     * Evaluate a script immediately (not in the background).
     *
     * @param script  script code to evaluate
     * @param input   Document to use as input for the script
     * @effects       evaluates <em>script</em> on <em>input</em>.  
     *                If the script returns a result, that result 
     *                becomes the new current document.
     */
    public Document evalNow (String script, Document input) 
	throws Exception {
        Interpreter interp = getInterpreter ();

        debug.println ("eval " + script);
        
        // Set the interpreter's input document
        interp.setVar (Interpreter.DOCUMENT_VAR, input);

        // Evaluate the script and get its output 
        Object obj = interp.eval (script);
        Thr.check ();
        
        Document output = null;
        if (obj != null) {
            if (obj instanceof Document)
                output = (Document) obj;
            else {
                String s = obj.toString ();
                if (s.length () > 0)
                    output = getFactory ().make (s);
                else
                    output = null;
            }
            
            if (output != null) {
                // Attach script to output document, so that
                // Reload can rerun this later
                if (output.getProperty (output.CommandProperty) == null)
                    output.putProperty 
                        (Document.CommandProperty, 
                         Str.replace (script.trim (), "\n", " ; "));
            }
        }

        return output;
    }


    ////////////////////
    //////////////////// Getting browser state
    ////////////////////

    /**
     * Get the current document.
     */
    public SwingDocument getDocument () {
        BrowserPane bp = getActivePane ();
        if (bp == null)
            return null;

        return (SwingDocument) bp.getDocument ();
    }

    /**
     * Get the browser's document factory.
     */
    public DocumentFactory getFactory () {
        return SwingDocument.defaultFactory;
    }

    /**
     * Get the browser's interpreter.
     */
    public Interpreter getInterpreter () {
        if (interpreter == null) {
            try {
                String interpreterClass = 
                    Main.getConfig ().getProperty 
                    ("interpreter");
                Interpreter interp = 
                    (Interpreter) Class.forName (interpreterClass).newInstance ();
                setInterpreter (interp);
            } catch (Exception e) {
                debug.report (e);
            }
        }
        return interpreter;
    }

    /**
     * Set the browser's interpreter.
     */
    public void setInterpreter (Interpreter interpreter) {
        if (this.interpreter == interpreter)
            return;

        this.interpreter = interpreter;
        interpreter.setVar (Interpreter.FACTORY_VAR, getFactory ());

        // Copy history from interpreter to browser
        setPageHistory ((History) interpreter.getVar (Interpreter.HISTORY_VAR));
 
        // Bind browser-specific tools
        interpreter.bindTool ("select", new Select (this));
        interpreter.bindTool ("scrollto", new ScrollTo (this));
        interpreter.bindTool ("view", new ViewNow (this));

        // Hacks for find & replace experiment
        interpreter.bindTool ("save-clusters", new SaveClusters (this));
        interpreter.bindTool ("highlight-replacements", 
			      new HighlightReplacements (this));

    }


    /**
     * Get the browser's page history.
     */
    public History getPageHistory () {
        return pageHistory;
    }

    /**
     * Set the browser's history.
     */
    public void setPageHistory (History history) {
        pageHistory = new History (history);
        getInterpreter ().setVar (Interpreter.HISTORY_VAR, pageHistory);
    }

    /**
     * Get the pattern library.
     */
    public PatternLibrary getPatternLibrary () {
        return PatternLibrary.getDocumentLibrary (getDocument());
    }

    /**
     * Get command that generated browser contents.
     *
     * @return command  (actually a URL, filename, or command), 
     *                  or null if browser is showing nothing.
     */
    public String getCommand () {
        SwingDocument doc = getDocument ();
        return (doc != null) 
            ? (String)doc.getProperty (SwingDocument.CommandProperty) 
            : null;
    }

    /**
     * Set the command that generated the current browser contents.
     *
     * @param cmd           a command, URL, or filename
     * @param addToHistory  true if cmd should be added to the end 
     *                      of the command history too
     * @effects             sets content of command box to <em>cmd</em>, 
     *                      and if <em>addToHistory</em> is true, 
     *                      also adds it to the drop-down
     *                      history.
     */
    public void setCommand (String cmd, boolean addToHistory) {
        command.getEditor ().setItem (cmd);
        if (addToHistory && !commandHistory.contains (cmd))
            commandHistory.add (cmd);
    }

    /**
     * Get URL that generated browser contents.
     * @return URL or null if contents not generated by a URL.
     */
    public URL getURL () {
        SwingDocument doc = getDocument ();
        return (doc != null) 
            ? (URL)doc.getProperty (SwingDocument.URLProperty) 
            : null;
    }

    /**
     * Get active browser pane.
     * @return BrowserPane that most recently had the focus.
     */
    public BrowserPane getActivePane () {
        return activePane;
    }

    /**
     * Get last pattern entered in pattern box.
     * @return contents of pattern box as a string
     */
    public String getPattern () {
        return patternPane.getPattern ();
    }

    /**
     * Set pattern in pattern box.
     * @param pattern String that should appear in pattern box.
     * @param addToHistory true if pattern should be added to the
     * end of the pattern history too.
     * @effects sets content of pattern box to <em>pattern</em>, and
     * if <em>addToHistory</em> is true, also adds it to the drop-down
     * history.
     */
    public void setPattern (String pattern, boolean addToHistory) {
        patternPane.setPattern (pattern, addToHistory);
    }

    ////////////////////
    //////////////////// Visiting new pages in browser
    ////////////////////


    /**
     * Set the current document.  Unlike most methods on Browser, this
     * one can be safely called from any thread.
     * @param newdoc Document to set
     * @effects Browser shows <em>newdoc</em> and adds it to the page
     * history.  If <em>newdoc</em> is not a SwingDocument, it is copied
     * into a new SwingDocument and the SwingDocument is used instead.
     */
    public void setDocument (Document newdoc) {
        // Convert to a SwingDocument if necessary
        final SwingDocument doc = convertToSwingDocument (newdoc);

        getInterpreter ().setVar (Interpreter.DOCUMENT_VAR, doc);

        debug.println ("show " + doc);

        SwingDocument olddoc = getDocument ();
        if (doc == olddoc)
            return;

        if (SwingUtilities.isEventDispatchThread ())
            doSetDocument (doc);
        else
            try {
                debug.println ("invokeAndWait");
                SwingUtilities.invokeAndWait (new Runnable () {
                        public void run () {
                            doSetDocument (doc);
                        }
                    });
            } catch (InterruptedException e) {
                debug.report (e);
            } catch (java.lang.reflect.InvocationTargetException e) {
                debug.report (e);
            }
    }

    private void doSetDocument (SwingDocument doc) {
        // Prepare to change current document
        loadingDocument (doc);

        SwingDocument error = 
            (SwingDocument) doc.getProperty (doc.StderrProperty);
        if (error != null) {
            // Load both stdout and error panes, and activate whichever
            // one corresponds
            SwingDocument stdout = 
                (SwingDocument) doc.getProperty (doc.StdoutProperty);
            stdoutPane.setDocument (stdout);
            load (stdoutPane, stdout);

            stderrPane.setDocument (error);
            load (stderrPane, error);

            activatePane ((doc == error) ? stderrPane : stdoutPane, true);

        } else {
            // No error document, so just load and activate stdout pane
            stdoutPane.setDocument (doc);
            load (stdoutPane, doc);
            activatePane (stdoutPane, false);
        }
    }

    /**
     * Makes sure document content is loaded.  Can be called from
     * any thread.
     */
    private void load (final BrowserPane pane, final SwingDocument doc) {
        if (doc.isLoaded ()) {
            // document is already loaded
            debug.println ("already loaded " + doc);
            if (pane == stdoutPane)
                loadedDocument (doc);
        } else {
            // load document in background
            debug.println ("loading " + doc);
            run 
                (new RunnableWithExceptions () {
                        public void run () {
                            doc.load (pane.getEditorKit ());
                            SwingUtilities.invokeLater (new Runnable () {
                                    public void run () {
                                        pane.loadedDocument (doc);
                                        if (pane == stdoutPane)
                                            loadedDocument (doc);
                                    }
                                });
                        }
                    });
        }
    }

    private SwingDocument convertToSwingDocument (Document doc) {
        if (doc == null || doc instanceof SwingDocument)
            return (SwingDocument)doc;

        SwingDocument sdoc = (SwingDocument) getFactory ().make (doc);

        // convert stdout & error document as well
        Document stdout = 
            (Document) sdoc.getProperty (doc.StdoutProperty);
        if (stdout != null) {
            sdoc.putProperty 
                (doc.StdoutProperty,
                 (stdout == doc)
                 ? sdoc
                 : getFactory ().make (stdout));
        }
        Document error =
            (Document) sdoc.getProperty (doc.StderrProperty);
        if (error != null) {
            sdoc.putProperty 
                (doc.StderrProperty,
                 (error == doc)
                 ? sdoc
                 : getFactory ().make (error));
        }
            

        // Replace document in history, if it's found there.
        History history = getPageHistory ();
        if (history.get () == doc) {
            debug.println ("replacing " + doc + " with " + sdoc);
            history.replace (sdoc);
        }

        return sdoc;
    }

    protected void loadingDocument (SwingDocument doc) {
        Thr.check ();

        // update status bar
        setStatus ("Loading " + doc + "...");

        // Put new doc in the history, if necessary
        History history = getPageHistory ();
        if (!history.contains (doc)) {
            Enumeration droppedDocs = history.forwardElements ();
            
            if (droppedDocs.hasMoreElements ()) {
                // newer documents are going to be removed
                // from the history.  If any are dirty,
                // prompt user to save them.
                if (!FileAction.okToClear (this, droppedDocs))
                    history.add (doc);
                else
                    history.put (doc);
            } else {              
                if (history.isFull ()) {
                    // oldest document is going to disappear.
                    // check if it's dirty, and expand the history if so.
                    Document oldest = 
                        (Document) history.elements ().nextElement ();
                    if (FileAction.isDirty (oldest))
                        history.expand ();
                }
                
                history.put (doc);
            } 
        }

        // Enable/disable navigation actions
        backAction.setEnabled (history.canBack ());
        forwardAction.setEnabled (history.canForward ());

        // Update content-type
        updateContentType (doc);

        // Update title bar (no TitleProperty available yet, but URLProperty
        // is)
        setTitle (doc);

        // Update command bar
        setCommand (doc.toString (), true);

        // Switch to manual selection mode 
//         getSelectionMode ().loadingDocument (doc);

        // Clear the selection
        clearSelection ();

        // fire event to document listeners
        for (Iterator g = browserListeners.iterator (); g.hasNext (); )
            ((BrowserListener) g.next ()).loadingDocument (this, doc);
    }

    /**
     * Called after the current document has been fully loaded.
     * Always called from event-handling thread.
     */
    protected void loadedDocument (SwingDocument doc) {
        Thr.check ();

        // Jump to # anchor specified by URL
        URL url = (URL) doc.getProperty (doc.URLProperty);
        if (url != null) {
            final String ref = url.getRef();
            if (ref != null)
                SwingUtilities.invokeLater (new Runnable () {
                        public void run () {
                            stdoutPane.scrollToReference (ref);
                        }
                    });
        }

        // Change directory to document's directory
        File f = FileAction.getFile (doc);
        if (f != null) {
            if (!f.isDirectory ())
                f = new File (f.getParent ());
            debug.println ("changing directory to " + f);
            getInterpreter ().setVar (Interpreter.CURRENT_DIRECTORY_VAR, f);
        }

        // Update title bar -- TitleProperty (if any) is available now
        // since document is loaded.
        setTitle (doc);

        // Clear the selection again
        clearSelection ();

        // Finish status bar
        clearStatus ();
        clearHourglass ();
        setStatus ("Document loaded.");

        // fire event to document listeners
        for (Iterator g = browserListeners.iterator (); g.hasNext (); )
            ((BrowserListener) g.next ()).loadedDocument (this, doc);
    }

    /**
     * Set window title bar appropriately for a document.
     * @param doc Document to use
     * @effects Sets window title to the HTML title, URL, or filename
     * of <em>doc</em>.
     */
    public void setTitle (Document doc) {
        String title = 
            (String) doc.getProperty (SwingDocument.TitleProperty);
        if (title == null)
           title = doc.toString ();
        setTitle ((title != null) ? "LAPIS - " + title : "LAPIS");
    }

    /**
     * Sleep until document contents have been fully loaded.
     * @effects Blocks until the current document has been fully
     * read from disk, network, or process.  Returns immediately if
     * current document is already loaded.
     */
    public void waitUntilLoaded () {
        SwingDocument doc = getDocument ();
        while (!doc.isLoaded ())
            try {
                Thread.sleep (100);
            } catch (InterruptedException e) {
                throw new ThreadDeath ();
            }
    }

    /**
     * Attach a listener for document-loading events.
     * @param listener listener to add
     * @effects listener is notified of future document-loading events
     */
    public void addBrowserListener (BrowserListener listener) {
        browserListeners.add (listener);
    }

    /**
     * Remove a listener that was listening for document-loading events.
     * @param listener listener to remove
     * @effects listener is no longer notified of document-loading events
     */
    public void removeBrowserListener (BrowserListener listener) {
        browserListeners.remove (listener);
    }

    ////////////////////
    //////////////////// Switching between browser and error pane
    ////////////////////

    private FocusListener paneFocusListener = new FocusAdapter () {
            public void focusGained (FocusEvent event) {
                paneGotFocus ((BrowserPane)event.getSource ());
            }
        };


    private void paneGotFocus (BrowserPane pane) {
        if (pane == activePane)
            return;

        debug.println ("focus on " + 
                       ((pane == stdoutPane) ? "stdout" : "error"));

        // Clear the old pane's selection
        clearSelection ();
        
        // Activate the new pane
        activatePane (pane, stderrScrollPane.isVisible ());
        
        // Update the UI to match new pane
        SwingDocument doc = getDocument ();
        updateContentType (doc);
        getPageHistory ().replace (doc);
    }

    private void activatePane (BrowserPane pane, boolean showErrorPane) {
        // show or hide error panel, adjusting error splitter accordingly
        int[] locations = getErrorSplitterLocations ();

        stdoutScrollPane.setTitled (showErrorPane);
        stdoutScrollPane.setActivated (pane == stdoutPane);
        stderrScrollPane.setActivated (pane == stderrPane);
        stderrScrollPane.setVisible (showErrorPane);
        
        browserSplitter.setDividerLocation (locations[showErrorPane ? 0 : 1]);
        browserSplitter.setLastDividerLocation (locations[0]);

        activePane = pane;
    }

    private int[] getErrorSplitterLocations () {
        int bottom = browserSplitter.getHeight ()
            - browserSplitter.getInsets ().bottom
            - browserSplitter.getDividerSize ();

        int middle = browserSplitter.getDividerLocation ();

        if (Math.abs (middle - bottom) < 15)
            middle = browserSplitter.getLastDividerLocation();

        browserSplitter.setLastDividerLocation (middle);

        return new int[] { middle, bottom };
    }


    ////////////////////
    //////////////////// Running and stopping background activities
    ////////////////////

    /**
     * Run a background task.
     * @param r Task to run
     * @effects Calls r.run() in a background thread.
     */
    public synchronized void run (RunnableWithExceptions r) {
        backgroundThreads.run (r);
    }


    /**
     * Stop all background tasks.
     * @effects Kills all tasks started with {@link #run}.  Also kills
     * the external process generating the current document, if any.
     */
    public synchronized void stop () {
        // Kill process associated with current document, if any
        Document doc = getDocument ();
        if (doc != null) {
            Process process = (Process) doc.getProperty (doc.ProcessProperty);
            if (process != null)
                process.destroy ();
        }

        // Interrupt all background threads, forcing I/O operations to throw
        // errors and Thr.check() to throw ThreadDeath
        backgroundThreads.stop ();

        // Reset progress UI 
        clearBusy ();
        clearStatus ();
    }

    /**
     * Action invoked by Stop command.  Stops downloading or parsing.
     */
    private Action stopAction = new AcceleratedAction 
        ("&Stop", "Stop the current page load") {
        public void actionPerformed (ActionEvent evt) {
            stop ();
            setStatus ("Stopped.");
        }
    };


    ////////////////////
    //////////////////// Starting and stopping busy animation and hourglass
    ////////////////////

    /**
     * Turn the busy animation (the spinning L icon) on and off.
     * Multiple callers may request the busy animation.  This method
     * uses a reference count to keep the icon spinning as long as
     * at least one caller still wants it enabled.
     * @param enabled true to start the animation, false to stop it
     * @effects If <em>enabled</em> is true, starts the busy animation
     * and increments a request count.  If <em>enabled</em> is false,
     * decrements the request count, and stops the busy animation if 
     * the count reaches zero.
     */
    public void setBusy (boolean enabled) {
        if (enabled)
            ++busyCount;
        else if (--busyCount > 0)
            return; // somebody still wants the busy showing
        else
            busyCount = 0;

        animation.setAnimating (enabled);
        stopAction.setEnabled (enabled);
    }

    /**
     * Stops the busy animation (the spinning L icon) and clears its
     * reference count.
     */
    public void clearBusy () {
        busyCount = 1;
        setBusy (false);
    }

    /**
     * Turn the wait mouse cursor (an hourglass or watch) on and off.
     * Multiple callers may request the wait cursor.  This method
     * uses a reference count to keep the wait cursor set as long as
     * at least one caller still wants it enabled.
     * @param enabled true for the wait cursor, false for the normal cursor
     * @effects If <em>enabled</em> is true, sets the wait cursor and
     * increments a request count.  If <em>enabled</em> is false,
     * decrements the request count, and reverts to the normal cursor
     * if the count reaches zero.
     */
    public void setHourglass (boolean enabled) {
        if (enabled)
            ++hourglassCount;
        else if (--hourglassCount > 0)
            return; // somebody still wants the hourglass showing
        else
            hourglassCount = 0;

        Cursor c = 
            enabled
            ? Cursor.getPredefinedCursor (Cursor.WAIT_CURSOR)
            : Cursor.getPredefinedCursor (Cursor.TEXT_CURSOR);
        stdoutPane.setCursor (c);
        stderrPane.setCursor (c);
    }

    /**
     * Reverts to the normal cursor and clears its reference count.
     */
    public void clearHourglass () {
        hourglassCount = 1;
        setHourglass (false);
    }


    ////////////////////
    //////////////////// Changing the view (text, html, etc.)
    ////////////////////

    static class ViewType {
        public String uiName;       // name for UI purposes (e.g. "Text")
        public String contentType;  // content type (e.g. "text/plain")

        public ViewType (String uiName, String contentType) {
            this.uiName = uiName;
            this.contentType = contentType;
        }

        public String toString () {
            return uiName;
        }
    }

    private void updateContentType (SwingDocument doc) {
        setViewContentType (view, 
                            (String) doc.getProperty 
                            (SwingDocument.MIMEProperty));
    }

    /**
     * Action invoked when user selects a different view from the drop-down list.
     */
    private Action changeViewAction = new AbstractAction () {
        public void actionPerformed (ActionEvent evt) {
            if (!view.isPopupVisible ())
                // if popup menu isn't visible, then
                // this action was triggered by code, not
                // by the user.  Don't do anything.
                return;

            ViewType vt = (ViewType) view.getSelectedItem ();
            if (vt == null)
                return;

            setContentType (vt.contentType);
        }
    };

    /**
     * Get content type of current document.
     * @return MIME content type, e.g. "text/plain" or "text/html".
     * May be null if content type is unknown.
     */
    public String getContentType () {
        SwingDocument doc = getDocument ();
        if (doc == null)
            return null;
        return (String) doc.getProperty (SwingDocument.MIMEProperty);
    }

    /**
     * Set content type of current document.
     * @param contentType MIME content type, e.g. "text/plain" or
     * "text/html".
     * @effects if current document's content type is the same as
     * <em>contentType</em>, nothing happens.  Otherwise, a new
     * document is created of the desired <em>contentType</em>, the 
     * the curent document is copied into it, and the new document
     * replaces the current document in the browser and in the page
     * history.
     */
    public void setContentType (String contentType) {
        SwingDocument olddoc = getDocument ();
        if (olddoc == null)
            return;

        // Check whether content type is actually changing
        String oldContentType = 
            (String) olddoc.getProperty (SwingDocument.MIMEProperty);
        if (contentType.equals (oldContentType))
            return;

        // Re-create document with desired content type
        olddoc.putProperty (SwingDocument.MIMEProperty, contentType);
        SwingDocument newdoc = 
            (SwingDocument) getFactory ().make (olddoc);
        debug.println ("converted " + olddoc.getClass () + " to " + contentType
                       + " to get " + newdoc.getClass ());

        // If olddoc was generated by a program, fix up its
        // stdout and stderr properties
        SwingDocument stdout = 
            (SwingDocument) olddoc.getProperty (olddoc.StdoutProperty);
        SwingDocument stderr = 
            (SwingDocument) olddoc.getProperty (olddoc.StderrProperty);
        if (stdout == olddoc) {
            newdoc.putProperty (olddoc.StdoutProperty, newdoc);
            stderr.putProperty (olddoc.StdoutProperty, newdoc);
        } else if (stderr == olddoc) {
            newdoc.putProperty (olddoc.StderrProperty, newdoc);
            stdout.putProperty (olddoc.StderrProperty, newdoc);
            newdoc = stdout;
        }
                
        // Replace document in history.
        getPageHistory ().replace (newdoc);

        // Hang on to the selection
        NamedRegionSet oldSelection = getSelection ();

        // Set the converted document
        setDocument (newdoc);
        getActivePane ().requestFocus ();
        
        // Restore the original selection
        //setSelection (oldSelection);
        transformSelection (oldSelection);
    }

    /*
     * Set the view drop-down list to match a loaded document's content type.
     */
    private static void setViewContentType (JComboBox view, 
					    String contentType) {
        int selection = 0;  // use Text as default if we can't find contentType

        try {
            ComboBoxModel model = view.getModel ();
            int n = model.getSize ();
            for (int i = 0; i < n; ++i) {
                ViewType vt = (ViewType)model.getElementAt (i);
                if (vt.contentType.equals (contentType)) {
                    selection = i;
                    break;
                }
            }
        } catch (NullPointerException e) {
        }

        view.setSelectedIndex (selection);
    }


    ////////////////////
    //////////////////// Tabs on sidebar
    ////////////////////

    /**
     * Add a tab to the sidebar.
     * @param name Name to display in tab
     * @param comp Component for tab
     * @effects adds <em>comp</em> as a new tab in the sidebar
     */
    public void addSidebarTab (String name, JComponent comp) {
	if (!sidebar.containsComp(comp))
	    sidebar.addTab (name, comp);
        sidebar.setSelectedComponent (comp);
	resetSidebarSplitter();
    }

    /**
     * Remove a tab from the sidebar.
     * @param comp Component to remove
     * @effects if f<em>comp</em> is a tab in the sidebar, removes it.
     * Otherwise does nothing.
     */
    public void removeSidebarTab (JComponent comp) {
        sidebar.remove (comp);
    }

    /** 
     * resets the sidebar Splitter to the width of the largest pane in
     * the sidebar if the width is currently too small, else leaves
     * the splitter alone.
     *
     * @effects sidebarSplitter
     **/

    private void resetSidebarSplitter(){
	if ((sidebar.getMinimumSize().width+15) > (sidebarSplitter.getWidth() - sidebarSplitter.getDividerLocation())){
	    sidebarSplitter.setDividerLocation(sidebarSplitter.getWidth() - (sidebar.getMinimumSize().width+15));
	}
    }


    ////////////////////
    //////////////////// NamedRegionSet
    ////////////////////

    /**
     * Set the current selection.
     * @param sel new selection
     * @effects sets selection in active pane to <em>sel</em>
     */
    public void setSelection (NamedRegionSet ns) {
    	// TODO: support multiple documents
		
        // minwu
        //DocumentVersion currVer = getDocument ().getVersion ();
		//HashSet vers = new HashSet ();
		//vers.add(currVer);
		//ns = (NamedRegionSet)ns.convert(vers);
        
        ns = (NamedRegionSet)ns.convert(new SingleDocVersions(getDocument().getVersion()));
		this.selection = ns;
		changedSelection ();
    }
    
    private void transformSelection (NamedRegionSet sel) {
		this.selection = new NamedRegionSet (RegionSet.EMPTY, sel.getPattern());
    }

    /**
     * Clears the selection.
     * @effects sets current selection to empty region sets
     */
    public void clearSelection () {
        debug.println ("clear");
		setSelection (new NamedRegionSet (RegionSet.EMPTY));
    }

    /**
     * Set selection to the result of a pattern match.
     * @param tc TC pattern to search for
     * @effects searches for TC pattern in active pane and sets selection
     * to the matching region set
     */
    public void find (TC tc) {
        patternPane.find (tc);
    }

    /**
     * Get the current selection.
     * @return current selection
     */
    public NamedRegionSet getSelection () {
        return selection;
    }

	protected Set getSelectionNames() {
		TC pattern = selection.getPattern();
		
		if (pattern == null)
			return null;

		Set set = pattern.getNames();
		if (set == null)
			return null;
		debug.println(set.toString());
		Set set2 = new HashSet();
		Iterator i = set.iterator();
		while (i.hasNext()){
			Object test = i.next();
			if (test instanceof String){
				String name = (String)test;
				debug.println(name); 
				String name2;
				try {
					name2 = PatternLibrary.getDocumentLibrary (this.getDocument()).lookup (name);
				} catch (PatternLibrary.LookupException e) {
					name2 = null;
				}
				if (name2 == null)
					return null;
				else
					set2.add(name2);
			}
		}
		return set2;
	}

    /**
     * Called when the selection changes.
     */
	protected void changedSelection () {
		// debugging
		if (debug.printEnabled ()
			&& !selection.isEmpty ())
			debug.print ("selection changed:\n"
						 + selection.highlight (getDocument ()));

		patternPane.tree.setSelectedNames (this.getSelectionNames());
		debug.println("selection Names: " + this.getSelectionNames());
		// Reset the next-match search
		if (findEnumerator != null) {
			findEnumerator = null;
			getActivePane ().getFinderHighlighter().setSelection (null);
		}

		// Highlight the matches in the browser pane
		getActivePane ().selectionHighlighter.setSelection (selection);

		// Display number of matches in status bar
		int n = selection.getSize ();
		if (getActivePane ().feedbackHighlighter.getSelection () != null)
			++n;
		patternPane.setMatchCount (n);

		// Enable or disable tools
		boolean nonempty = !selection.isEmpty ();
		keepAction.setEnabled (nonempty);
		omitAction.setEnabled (nonempty);
		sortAction.setEnabled (nonempty);
		replaceAction.setEnabled (nonempty);
		extractAction.setEnabled (nonempty);
		calcAction.setEnabled (nonempty);

		downcaseAction.setEnabled (nonempty);
		upcaseAction.setEnabled (nonempty);
		sentenceCaseAction.setEnabled (nonempty);
		titleCaseAction.setEnabled (nonempty);
		toggleCaseAction.setEnabled (nonempty);

		// fire event to selection listeners
		for (Iterator g = selectionListeners.iterator (); g.hasNext (); )
			((SelectionListener) g.next ()).changedSelection (this, selection);
	}

    /**
     * Action invoked when user clicks on Unselect on context menu.
     * Unselects the region under the cursor.
     */
    private class UnselectAction extends AcceleratedAction {
        /** Region that was right-clicked on */
        private Region clickedRegion = null;

        public UnselectAction () {
            super ("Unselect");
            setEnabled (false);
        }

        public void setRegion (Region clickedRegion) {
            this.clickedRegion = clickedRegion;
            setEnabled (clickedRegion != null);
        }

        public void actionPerformed (ActionEvent evt) {
            if (clickedRegion == null)
                return;

            getSelectionMode ().unselectRegion (clickedRegion);
        }
    }


    /**
     * Action invoked when user clicks on Clear NamedRegionSet on context menu.
     * Clears entire selection.
     */
    private Action clearSelectionAction = new AcceleratedAction ("Unselect All") {
        public void actionPerformed (ActionEvent evt) {
            clearSelection ();
        }
    };

    /**
     * Attach a listener for selection-change events.
     * @param listener listener to add
     * @effects listener is notified of future selection changes in this browser
     */
    public void addSelectionListener (SelectionListener listener) {
        selectionListeners.add (listener);
    }

    /**
     * Remove a listener that was listening for selection-change events.
     * @param listener listener to remove
     * @effects listener is no longer notified of selection changes in
     * this browser
     */
    public void removeSelectionListener (SelectionListener listener) {
        selectionListeners.remove (listener);
    }

    ////////////////////
    //////////////////// NamedRegionSet mode
    ////////////////////

    /**
     * Get current selection mode.
     * @return selection mode
     */
    public SelectionMode getSelectionMode () {
        return selectionMode;
    }

    /**
     * Get this browser's manual selection mode.
     * @return selection mode to use for manual selection
     */
    public ManualMode getManualSelectionMode () {
        return manualSelection;
    }

    /**
     * Set current selection mode.
     * @param newMode new selection mode
     * @effects sets selection mode to newMode
     */
    public void setSelectionMode (SelectionMode newMode) {
        SelectionMode oldMode = selectionMode;

        oldMode.leavingMode (newMode);

        if (oldMode instanceof BrowserListener)
            removeBrowserListener ((BrowserListener) oldMode);
        if (oldMode instanceof SelectionListener)
            removeSelectionListener ((SelectionListener) oldMode);

        selectionMode = newMode;

        if (newMode instanceof BrowserListener)
            addBrowserListener ((BrowserListener) newMode);
        if (newMode instanceof SelectionListener)
            addSelectionListener ((SelectionListener) newMode);

        newMode.enteringMode (oldMode);

        getActivePane ().updateFocus (getActivePane ().haveFocus);

        if (newMode instanceof SimulEditMode) {
            simulEditRadioButtonMenuItem.setSelected (true);
            simulEditToggleButton.setSelected (true);
        } else {
            manualRadioButtonMenuItem.setSelected (true);
            manualToggleButton.setSelected (true);
        }
    }


    ////////////////////
    //////////////////// Stepping through selection
    ////////////////////

    /**
     * Step to next match in selection.
     */
    private Action nextMatchAction = new AcceleratedAction 
        ("&Next Match", KeyEvent.VK_F3, 0,
         "Jump to the next highlight in selection [F3]") {
            public void actionPerformed (ActionEvent event) {
                findNext ();
            }
        };

    private void findNext () {
        NamedRegionSet sel = getSelection ();
        if (sel.isEmpty ()) {
            getToolkit().beep ();
            setStatus ("No regions are selected.");
            return;
        }

        Region r;
        if (findEnumerator == null) {
            findEnumerator = sel.regions ();
            r = findEnumerator.first ();
        } else
            r = findEnumerator.next ();

        debug.println ("next match = " + r);
        if (r == null){
            getToolkit().beep ();
            setStatus ("No more regions in selection.");
        } else {
            // highlight match
			getActivePane ().getFinderHighlighter().setSelection (new NamedRegionSet (r));
            getActivePane ().scrollRegionToVisible (r);
        }
    }


    ////////////////////
    //////////////////// Right-click menu
    ////////////////////

    /**
     * Action invoked when user right-clicks on the browser pane.
     * Pops up a menu showing region set names that coincide with
     * the current selection.
     */
    private MouseListener rightClickListener = new MouseAdapter () {
            public void mousePressed (MouseEvent evt) {
                if (SwingUtilities.isRightMouseButton (evt))
                    showPopupMenu ((Component)evt.getSource (), 
				   evt.getX (), evt.getY ());
            }
        };

    private void showPopupMenu (Component source, int x, int y) {
        JPopupMenu menu = new JPopupMenu ();

        Region containsClick = (Region) getActivePane ().pointToRegion (new Point (x, y)).contains ();

        // test whether user clicked on an outlier record
        NamedRegionSet outlierRecords = getActivePane ().outlierRecordHighlighter.getSelection ();
        if (outlierRecords != null && !outlierRecords.isEmpty ()) {
            Region clickedRecord = outlierRecords.regions (containsClick).first ();
            ignoreOutlierAction.setRecord (clickedRecord);
            menu.add (ignoreOutlierAction);
        }

        // test whether user clicked on an outlier example
        if (outlierFinder != null && outlierFinder.isVisible ()) {
            Region clickedExample = outlierFinder.allExamples.regions (containsClick).first ();
            viewAsOutlierAction.setExample (clickedExample);
            menu.add (viewAsOutlierAction);
        }

        // test whether clicked on a selected region
        Region clickedRegion = getSelection ().regions (containsClick).first ();
        unselectAction.setRegion (clickedRegion);
        menu.add (unselectAction);

        menu.add (clearSelectionAction);

        menu.addSeparator ();
        menu.add (cutAction);
        menu.add (copyAction);
        menu.add (pasteAction);
        menu.add (deleteAction);

        menu.show (source, x, y);
    }


    ////////////////////
    //////////////////// Outliers
    ////////////////////

    /**
     * Action invoked when user clicks on Find Unusual Matches menu item.
     * Pops up a FindOutliers window
     */
    private Action findOutliersAction = new AcceleratedAction 
        ("&Unusual Matches", 
         KeyEvent.VK_F9, 0) {
            public void actionPerformed (ActionEvent evt) {
                if (outlierFinder == null) {
                    outlierFinder = new OutlierFinder (Browser.this);
                    addBrowserListener (outlierFinder);
                    addSelectionListener (outlierFinder);
                }
                outlierFinder.setVisible (true);
            }
        };
    
    /**
     * Action invoked when user clicks on Find Unusual Matches menu item.
     * Pops up a FindOutliers window
     */
    private class ViewAsOutlierAction extends AcceleratedAction {
        /** Positive or negative example that was right-clicked on */
        private Region clickedExample = null;
        
        public ViewAsOutlierAction () {
            super ("Locate in Unusual Matches Window");
            setEnabled (false);
        }

        public void setExample (Region clickedExample) {
            this.clickedExample = clickedExample;
            setEnabled (clickedExample != null);
        }

        public void actionPerformed (ActionEvent evt) {
            if (outlierFinder == null  
                || !outlierFinder.isVisible () 
                || clickedExample == null)
                return;
            outlierFinder.jumpTo (clickedExample);
        }
    }


    ////////////////////
    //////////////////// Pattern Editing
    ////////////////////

    /**
     * Action invoked when user selects Pattern Editor menu option.
     * Pops up a new TCFileEditor window.
     */
    private Action patternEditorAction = 
	new AcceleratedAction ("Pattern &Editor") {
	    public void actionPerformed (ActionEvent evt) {
		if (tcEditor == null) {
		    showTCEditor();
		}
		else {
		    sidebar.setSelectedComponent (tcEditor);
		}
	    }
	};


    private void showTCEditor () {
	tcEditor = new TCFileEditor (Browser.this);
	tcEditor.setText ("# \n# This is a pattern editor.\n#\n\n");
	tcEditor.setDirty (false);  
	addSidebarTab ("Pattern Editor", tcEditor);
    }


    public void closeTCEditor() {
	tcEditor = (TCFileEditor) sidebar.getSelectedComponent();
	if (tcEditor.okToClear ()) {
	    tcEditor.clear ();
	    removeSidebarTab (tcEditor);
	    tcEditor = null;
	}
    }


    ////////////////////
    //////////////////// Scripting
    ////////////////////

    /**
     * Action invoked when user selects Script Editor menu item.
     * Pops up a new TclFileEditor window.
     */
    private Action scriptEditorAction = 
	new AcceleratedAction ("Script &Editor") {
	    public void actionPerformed (ActionEvent evt) {
		if (tclEditor == null) {
		    showTclEditor();
		}
		else {
		    sidebar.setSelectedComponent (tclEditor);
		}
	    }
	};

    /**
     * A Tcl comment to help differentiate the script editor panel from 
     * a pattern editor panel.
     */
    private String tclComment = "# \n# This is a Tcl script editor.\n#\n\n";

    private void showTclEditor () {
          tclEditor = new TclFileEditor (Browser.this);
	  tclEditor.setText (tclComment);
	  addSidebarTab ("Script Editor", tclEditor);
    }


    /**
     * Close the sidebar tab containing TclEditor.
     */
    public void closeTclEditor() {
	tclEditor = (TclFileEditor) sidebar.getSelectedComponent();
	if (tclEditor.okToClear ()) {
	    tclEditor.clear ();
	    removeSidebarTab (tclEditor);
	    tclEditor = null;
	}
    }


    /**
     * Action invoked by History to Script command.
     */
    private Action historyScriptAction = new AcceleratedAction 
	("Convert &History to Script") {
        public void actionPerformed (ActionEvent evt) {
	    if (tclEditor == null) {
                popupScriptEditor (tclComment + convertHistoryToScript ());
	    }
	    else {
		tclEditor.setText (tclComment + convertHistoryToScript ());
		tclEditor.setFile (null);
		tclEditor.setDirty (false); 
		sidebar.setSelectedComponent (tclEditor);
	    }
        }
    };


    /**
     * Convert bookmark history to script.
     */
    String convertHistoryToScript () {
        History history = getPageHistory ();
        String script = "";
        for (Enumeration e = history.elements (); e.hasMoreElements (); ) {
            Document doc = (Document) e.nextElement ();
            String cmd = doc.toString ();
            script += cmd + "\n";
        }
        return script;
    }


    /**
     * Make a new sidebar panel pop up containing a TclEditor with 
     * the specified contents.
     */
    void popupScriptEditor (String contents) {
        showTclEditor ();
	tclEditor.setText (contents);
	tclEditor.setFile (null);
	tclEditor.setDirty (false);          
    }
            

    ////////////////////
    //////////////////// Tool interface for Browser itself
    ////////////////////

    /**
     * Invoke browser with command-line arguments.
     * @param args arguments for browser
     * @effects uses the arguments to initialize this browser, then
     * makes this browser visible.
     */
    public Document invoke (Arguments args) throws Exception {
        if (interpreter == null)
            setInterpreter (args.getInterpreter ());
        setVisible (true);

        String href = null;
        String script = null;

        args.setUsage (
        "Usage: Browser [options]\n"
        + "\n"
        + "Options:\n"
        + "    -help     Display this message\n"
        );
        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("e".equals (name)) {
                script = args.nextString ();
            } else if ("f".equals (name)) {
                script = "source {" + args.nextString () + "}";
            } else if ("href".equals (name)) {
                href = URLUtil.HrefToURL (args.nextString ()).toString ();
            } else 
                args.consume (name);
        }

        // prepend href to script
        if (href != null)
            script =
                (script != null)
                ? href + "\nview\n" + script
                : href;
            
        Document doc = args.getInput ();

        if (doc != null)
            setDocument (doc);
        else if (script == null)
            script = SwingConfig.makeURLFromProperty ("browser.homepage")
                .toString ();

        if (script != null)
            eval (script);

        getActivePane ().requestFocus ();
        return getDocument ();
    }
    // retrival of cluster pane for testing
    public ClusterFindReplacePane getClusterPane() {
	ClusterFindReplacePane cfrp = new ClusterFindReplacePane(this);
	if (sidebar.containsComp(cfrp))
	    sidebar.add(cfrp);
	return cfrp;


    }


}
