/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.swing;

import lapis.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.JTextComponent;
import lapisx.swing.*;

import java.net.*;

import java.util.*;
import java.util.List;
import lapisx.util.*;

public abstract class BrowserAction {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    /**
     * Listener invoked when user mouses over or clicks on a hyperlink
     * in a Browser.
     */
    public static class Hyperlink implements HyperlinkListener {
        Browser browser;

        public Hyperlink (Browser browser) {
            this.browser = browser;
        }

        public void hyperlinkUpdate(HyperlinkEvent e) {
            HyperlinkEvent.EventType type = e.getEventType();

            URL url = e.getURL ();
            if (url == null)
                return;

            if (type == HyperlinkEvent.EventType.ACTIVATED)
                clickedHyperlink (url);
            else if (type == HyperlinkEvent.EventType.ENTERED)
                browser.setStatus (url.toString ());
            else if (type == HyperlinkEvent.EventType.EXITED)
                browser.clearStatus ();
        }

        void clickedHyperlink (URL url) {
            debug.println ("clicked on hyperlink " + url);
        
            if ("file".equals (url.getProtocol ())
                && "localhost".equals (url.getHost ())
                && !url.getFile ().startsWith ("//")) {
                // delete the localhost for readability
                String file = url.getFile ();
                String ref = url.getRef ();
                try {
                    url = new URL ("file", "", 
                                   file + ((ref != null && ref.length () > 0) 
                                           ? "#" + ref 
                                           : ""));
                } catch (MalformedURLException e) {
                    debug.report (e);
                }
            }
        
            URL currURL = browser.getURL ();
            if (currURL != null && url.sameFile (currURL)) {
                // just scroll to the desired reference
                String ref = url.getRef ();
                if (ref != null)
                    browser.getActivePane ().scrollToReference (ref);
                return;
            }

            browser.gotoURL (url);
        }
    }

    /**
     * Action that runs a URL or command typed into the Command box.
     */
    public static class Command extends AbstractAction {
        Browser browser;
        public Command (Browser browser) {
            this.browser = browser;
        }
        public void actionPerformed (ActionEvent evt) {
            JTextComponent editor = (JTextComponent) evt.getSource ();
            String cmd = editor.getText ();
            browser.setCommand (cmd, true);
            editor.selectAll ();
            browser.eval (cmd);
        }

    }

    /**
     * Action that jumps to home page.
     */
    public static class Home extends AcceleratedAction {
        Browser browser;

        public Home (Browser browser) {
            super ("&Home",
                   SwingConfig.makeIcon ("home.gif"),
                   "Go to home page");
            this.browser = browser;
        } 

        public void actionPerformed (ActionEvent evt) {
            URL homepage = 
                SwingConfig.makeURLFromProperty ("browser.homepage");
            browser.gotoURL (homepage);
        }
    }

    /**
     * Action that moves back to previous document in history.
     */
    public static class Back extends Navigate {
        public Back (Browser browser) {
            super (browser, 
                   "&Back", 
                   SwingConfig.makeIcon ("back.gif"),
                   KeyEvent.VK_LEFT, KeyEvent.ALT_MASK,
                   "Go back one page");
        }
        
        public Document go () {
            return (Document) browser.getPageHistory ().back ();
        }

        public Iterator choices () {
            List l = new ArrayList ();
            History h = browser.getPageHistory ();
            for (Enumeration e = h.backElements (); e.hasMoreElements (); )
                l.add (e.nextElement ());
            Collections.reverse (l);
            return l.iterator ();
        }
    }

    /**
     * Action that moves forward to next document in history.
     */
    public static class Forward extends Navigate {
        public Forward (Browser browser) {
            super (browser, 
                   "&Forward", 
                   SwingConfig.makeIcon ("fwd.gif"),
                   KeyEvent.VK_RIGHT, KeyEvent.ALT_MASK,
                   "Go forward one page");
        }
        
        public Document go () {
            return (Document) browser.getPageHistory ().forward ();
        }

        public Iterator choices () {
            List l = new ArrayList ();
            History h = browser.getPageHistory ();
            for (Enumeration e = h.forwardElements (); e.hasMoreElements (); )
                l.add (e.nextElement ());
            return l.iterator ();
        }
    }

    /**
     * Base class of Back and Forward, providing a common framework.
     */
    public static abstract class Navigate extends AcceleratedAction {
        Browser browser;
        DelayedPopupMenu menu;

        public Navigate (Browser browser, String name, Icon icon, 
                         int keycode, int modifiers, String tooltip) {
            super (name, icon, keycode, modifiers, tooltip);
            this.browser = browser;
        }

        public abstract Document go ();
        public abstract Iterator choices ();

        public void actionPerformed (ActionEvent evt) {
            // don't do default action if choice menu is popped up
            if (menu != null && menu.isVisible ())
                return;

            Document doc = go ();

            if (doc == null)
                Toolkit.getDefaultToolkit ().beep ();
            else {
                browser.stop ();
                browser.setDocument (doc);
                browser.getActivePane ().requestFocus ();
            }
        }

        public void attachMenu (JButton button) {
            menu = new DelayedPopupMenu (browser, button, 500) {
                    public JPopupMenu makeMenu () {
                        Iterator g = choices ();
                        if (!g.hasNext ())
                            return null;

                        JPopupMenu menu = super.makeMenu ();
                        while (g.hasNext ()) {
                            Document doc = (Document) g.next ();
                            Action a = new GotoPage (browser, doc);
                            menu.add (a);
                        }
                        return menu;
                    }
                };
        }
    }

    /**
     * Action that jumps to a particular page.  Used for menu items
     * on the back and forward drop-down menus.
     */
    public static class GotoPage extends AbstractAction {
        Browser browser;
        Document doc;
        public GotoPage (Browser browser, Document doc) {
            super (Str.abbreviate (doc.toString (), 50));
            this.browser = browser;
            this.doc = doc;
        }
        public void actionPerformed (ActionEvent event) {
            browser.stop ();
            browser.getPageHistory ().jumpTo (doc);
            browser.setDocument (doc);
            browser.getActivePane ().requestFocus ();
        }
    }


    /**
     * Action that reloads the current page.
     */
    public static class Reload extends AcceleratedAction {
        Browser browser;

        public Reload (Browser browser) {
            super ("&Reload", 
                   SwingConfig.makeIcon ("reload.gif"),
                   'R', KeyEvent.CTRL_MASK,
                   "Reload the current page");
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            Document doc = browser.getDocument ();
            final String command = getReloadCommand ();
            if (command == null) {
                Toolkit.getDefaultToolkit ().beep ();
                return;
            }
            
            browser.setStatus ("Reloading...");
            
            final Document input = 
                (Document) browser.getPageHistory ().peekBack ();
            
            browser.run (new RunnableWithExceptions () {
                    public void run () throws Exception {
                        final Document stdout = 
                            browser.evalNow (command, input);
                        
                        browser.clearStatus ();
                        if (stdout != null) {
                            browser.getPageHistory ().replace (stdout);
                            SwingUtilities.invokeLater (new Runnable () {
                                    public void run () {
                                        browser.setDocument (stdout);
                                        browser.getActivePane ().requestFocus ();
                                    }
                                });
                        }
                    }
                });
        }

        String getReloadCommand () {
            String command = browser.getCommand ();
            if (command != null)
                return command;
            
            URL url = browser.getURL ();
            if (url != null)
                return url.toString ();
            
            return null;
        }
    }


    /** 
     * Listener that displays the page history on a menu.  When the menu
     * appears, this listener adds the history contents as menu items.
     * When it disappears, the items are removed.
     */
    public static class HistoryMenuListener implements PopupMenuListener {
        Browser browser;

        public HistoryMenuListener (Browser browser) {
            this.browser = browser;
        }

        public void popupMenuWillBecomeVisible (PopupMenuEvent event) {
            JPopupMenu menu = (JPopupMenu) event.getSource ();

            // enumerate history into a list, so we can reverse it
            History history = browser.getPageHistory ();
            Object curr = history.get ();
            List l = new ArrayList (history.size ());
            for (Enumeration e = history.elements (); e.hasMoreElements (); )
                l.add (e.nextElement ());

            // add history to menu in reverse order
            for (int i = l.size () - 1; i >= 0; --i) {
                Document doc = (Document) l.get (i);
                Action action = new GotoPage (browser, doc);
                String name = (i+1) + ". " + action.getValue (Action.NAME);
                boolean isCurrent = (doc == curr);
                JMenuItem item = new HistoryItem (name, action, isCurrent);
                menu.add (item);
            }
        }

        public void popupMenuWillBecomeInvisible (PopupMenuEvent event) {
            JPopupMenu menu = (JPopupMenu) event.getSource ();

            // remove all items we added
            MenuElement[] items = menu.getSubElements ();
            for (int i = 0, n = items.length; i < n; ++i)
                if (items[i] instanceof HistoryItem)
                    menu.remove (menu.getComponentIndex ((HistoryItem) items[i]));
        }

        public void popupMenuCanceled (PopupMenuEvent e) {
        }

        static class HistoryItem extends JRadioButtonMenuItem {
            public HistoryItem (String name, Action action, boolean isCurrent) {
                super (name, isCurrent);
                addActionListener (action);
            }
        }

    }
}
