/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.net.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.undo.*;
import javax.swing.plaf.TextUI;
import lapisx.swing.*;
import java.awt.datatransfer.*;

public class BrowserPane extends JEditorPane {
	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

	Browser browser;
	RegionHighlighter selectionHighlighter;
	RegionHighlighter feedbackHighlighter;
	RegionHighlighter positiveExamplesHighlighter;
	RegionHighlighter negativeExamplesHighlighter;
	private RegionHighlighter finderHighlighter;
	RegionHighlighter outlierRecordHighlighter;
	RegionHighlighter recordHighlighter;
	private RegionHighlighter findAndReplaceHighlighter;

	static Color wheat = new Color(255, 255, 170);
	static Color red = new Color(255, 153, 153);
	static Color lightRed = new Color(255, 200, 200);
	static Color veryLightRed = new Color(255, 230, 230);

	static Color selectionWithFocus = new Color(180, 180, 255);
	static Color selectionWithoutFocus = new Color(220, 220, 255);
	static Color positiveExampleSelection = new Color(160, 160, 255);
	static Color inferredSelection = new Color(200, 200, 255);

	static Color caretWithFocus = new Color(0, 0, 255);
	static Color caretWithoutFocus = new Color(180, 180, 255);
	static Color positiveExampleCaret = new Color(0, 0, 255);
	static Color inferredCaret = new Color(0, 200, 255);

	static Color selectionScrollbar = new Color(0, 0, 178);
	static Color finderScrollbar = Color.red;
	static Color outlierRecordScrollbar = finderScrollbar;

	boolean antialias = Boolean.valueOf(Main.getConfig().getProperty("browser.antialias")).booleanValue();

	boolean pretendToBeReadOnly = false;

	boolean haveFocus; // true whenever we have keyboard focus
	boolean ignoreFocusClick;
	// set true when pane loses focus when it has an active selection,
	// so that user can click on BrowserPane to give focus back without
	// clearing the selection.

	public BrowserPane(Browser browser) {
		this.browser = browser;

		// selection: displays the primary selection in blue
		selectionHighlighter = new RegionHighlighter();
		selectionHighlighter.setGradientColor(selectionWithoutFocus);
		selectionHighlighter.setScrollBarHighlighted(true);
		selectionHighlighter.setScrollBarHighlightColor(selectionScrollbar);
		selectionHighlighter.setCaretColor(caretWithoutFocus);

		// feedback: displays a region being dragged out by the mouse, same color
		// as selection
		feedbackHighlighter = new RegionHighlighter();
		feedbackHighlighter.setGradientColor(selectionWithFocus);
		feedbackHighlighter.setCaretColor(caretWithFocus);

		// positive examples: displays the positive examples during inference
		positiveExamplesHighlighter = new RegionHighlighter();
		positiveExamplesHighlighter.setGradientColor(positiveExampleSelection);
		positiveExamplesHighlighter.setCaretColor(positiveExampleCaret);

		// negative examples: displays the negative examples during inference
		negativeExamplesHighlighter = new RegionHighlighter();
		negativeExamplesHighlighter.setGradientColor(veryLightRed);
		negativeExamplesHighlighter.setCaretColor(lightRed);
		negativeExamplesHighlighter.setHandlesVisible(false);

		// finder: highlights a subset of the selection in red
		finderHighlighter = new RegionHighlighter();
		finderHighlighter.setGradientColor(red);
		finderHighlighter.setCaretColor(red);
		//         finderHighlighter.setCaretGradientVisible (true);
		finderHighlighter.setScrollBarHighlighted(true);
		finderHighlighter.setScrollBarHighlightColor(finderScrollbar);
		//         finderHighlighter.setHandlesVisible (false);
		//         finderHighlighter.setOutlining (true);
		//         finderHighlighter.setShrinkage (-2);

		// record: displays simultaneous editing records in yellow
		recordHighlighter = new RegionHighlighter();
		recordHighlighter.setHandlesVisible(false);
		recordHighlighter.setGradientColor(wheat);

		// outlier records: displays records with weird selections in red
		outlierRecordHighlighter = new RegionHighlighter();
		outlierRecordHighlighter.setHandlesVisible(false);
		outlierRecordHighlighter.setGradientColor(lightRed);
		outlierRecordHighlighter.setScrollBarHighlighted(true);
		outlierRecordHighlighter.setScrollBarHighlightColor(outlierRecordScrollbar);

		// find & replace: displays matches left to replace
		findAndReplaceHighlighter = new RegionHighlighter();
		findAndReplaceHighlighter.setHandlesVisible(false);
		findAndReplaceHighlighter.setCaretVisible(false);
		findAndReplaceHighlighter.setGradientColor(new Color(255, 255, 185));
		findAndReplaceHighlighter.setScrollBarHighlighted(true);
		findAndReplaceHighlighter.setScrollBarHighlightColor(Color.yellow);

		// highlight ordering
		// record highlight should be under all the others, so put it first
		// on the rendering list.  Feedback should be on top of all others, so
		// put it last on the list.
		finderHighlighter.setNextHighlighter(feedbackHighlighter);
		positiveExamplesHighlighter.setNextHighlighter(finderHighlighter);
		selectionHighlighter.setNextHighlighter(positiveExamplesHighlighter);
		negativeExamplesHighlighter.setNextHighlighter(selectionHighlighter);
		outlierRecordHighlighter.setNextHighlighter(negativeExamplesHighlighter);
		findAndReplaceHighlighter.setNextHighlighter(outlierRecordHighlighter);
		recordHighlighter.setNextHighlighter(findAndReplaceHighlighter);
		setHighlighter(recordHighlighter);

		// scrollbar highlight ordering.
		// only selection, outlier, and finder are highlighted in the scrollbar.
		// Put outlier and finder on top of selection.
		outlierRecordHighlighter.setNextScrollBarHighlighter(finderHighlighter);
		selectionHighlighter.setNextScrollBarHighlighter(outlierRecordHighlighter);
		findAndReplaceHighlighter.setNextScrollBarHighlighter(selectionHighlighter);

		setEditable(true);
		setCaret(new BrowserCaret());

		setFont(new Font("Monospaced", Font.PLAIN, 12));
		try {
			int size = SwingConfig.getFontSize();
			FontUtil.resizeFont(this, size);
			FontUtil.resizeHTMLFont(this, size);
		}
		catch (NumberFormatException e) {}

		setEditorKit(createDefaultEditorKit()); // why doesn't JEditorPane do this, so that editor kit is always valid?
		//	setOpaque(false);
		// change appearance of selection highlight when we lose focus
		addFocusListener(new FocusListener() {
			public void focusGained(FocusEvent event) {
				updateFocus(true);
			}
			public void focusLost(FocusEvent event) {
				updateFocus(false);
			}
		});
	}

	public void updateFocus(boolean gainingFocus) {
		//         SelectionMode mode = browser.getSelectionMode ();

		//         if (mode.showExampleHighlight ()) {
		//             if (gainingFocus) {
		//                 positiveExamplesHighlighter.setGradientColor (positiveExampleSelection);
		//                 positiveExamplesHighlighter.setCaretColor (positiveExampleCaret);
		//             } else {
		//                 positiveExamplesHighlighter.setGradientColor (selectionWithoutFocus);
		//                 positiveExamplesHighlighter.setCaretColor (caretWithoutFocus);
		//             }

		//             if (gainingFocus) {
		//                 selectionHighlighter.setGradientColor (inferredSelection);
		//                 selectionHighlighter.setCaretColor (inferredCaret);
		//             } else {
		//                 selectionHighlighter.setGradientColor (selectionWithoutFocus);
		//                 selectionHighlighter.setCaretColor (caretWithoutFocus);
		//             }

		//             positiveExamplesHighlighter.damageAll ();
		//             selectionHighlighter.damageAll ();

		//         } else {
		if (gainingFocus) {
			selectionHighlighter.setGradientColor(selectionWithFocus);
			selectionHighlighter.setCaretColor(caretWithFocus);
		}
		else {
			selectionHighlighter.setGradientColor(selectionWithoutFocus);
			selectionHighlighter.setCaretColor(caretWithoutFocus);
		}
		selectionHighlighter.damageAll();
		//         }

		haveFocus = gainingFocus;
		debug.println("browser focus " + haveFocus);
	}

	public void paintComponent(Graphics g) {
		if (antialias && "text/html".equals(getContentType())) {
			((Graphics2D)g).setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		}
		super.paintComponent(g);
		updateIgnoreFocusClick();
		debug.println("repaint updated ignoreFocusClick to " + ignoreFocusClick);
	}

	public void setAntialias(boolean f) {
		antialias = f;
	}

	public boolean getAntialias() {
		return antialias;
	}

	public boolean isEditable() {
		if (pretendToBeReadOnly)
			return false;
		else
			return super.isEditable();
	}

	public ScrollBarHighlighter getScrollBarHighlighter() {
		return findAndReplaceHighlighter;
	}

	public boolean getScrollableTracksViewportWidth() {
		SwingDocument doc = (SwingDocument)getDocument();
		if (doc instanceof lapis.swing.PlainDocument)
			return true;
		else
			return super.getScrollableTracksViewportWidth();
	}

	public void setDocument(SwingDocument doc) {
		debug.println("BrowserPane.setDocument ()");

		selectionHighlighter.setSelection(null);
		feedbackHighlighter.setSelection(null);
		finderHighlighter.setSelection(null);
		positiveExamplesHighlighter.setSelection(null);
		negativeExamplesHighlighter.setSelection(null);
		outlierRecordHighlighter.setSelection(null);
		recordHighlighter.setSelection(null);

		// Set the editor's content type to match the document
		// (a side-effect is selecting the right EditorKit for the content type).
		String contentType = (String)doc.getProperties().get(doc.MIMEProperty);
		setContentType(contentType);

		// Make sure the document's StreamDescriptionProperty is set
		URL url = (URL)doc.getProperty(doc.URLProperty);
		if (url != null)
			doc.putProperty(doc.StreamDescriptionProperty, url);

		// Make the editor editable iff content type is text/plain
		//         setEditable (doc instanceof PlainDocument);

		// Set the editor's document to the new document
		debug.println("JEditorPane.setDocument (" + doc.getClass() + ")");
		super.setDocument(doc);

		// reset the scrollbar
		debug.println("scrolling");
		scrollRectToVisible(new Rectangle(0, 0, 1, 1));

		debug.println("done setting document");
	}

	public void loadedDocument(SwingDocument doc) {
		CoalescingUndo.setUndoableEditListener(doc, new UndoManager());
	}

	//
	// Undo
	//

	boolean undoing = false;

	class MultipleSelectionEdit extends CompoundEdit {
		DocumentVersion oldVersion;
		DocumentVersion oldCookedVersion;
		NamedRegionSet oldRecords;
		NamedRegionSet oldFindRecords;
		
		String rawText;
		
		public MultipleSelectionEdit() {
			SwingDocument doc = (SwingDocument)getDocument();
			oldVersion = doc.getVersion();
			oldCookedVersion = doc.getCookedView().getVersion();

			oldRecords = recordHighlighter.getSelection();
			oldFindRecords = getFindAndReplaceHighlighter().getSelection();
			
			if (doc instanceof HTMLDocument)
				rawText = doc.getText();
		}

		public void undo() throws CannotUndoException {
			SwingDocument doc = (SwingDocument)getDocument();

			//debug.println ("before undo, doc ver is " + doc.getVersion ());
			undoing = true;
			super.undo();
			undoing = false;
			//debug.println ("after undo, doc ver is now " + doc.getVersion ());

			debug.println("undo reverting to " + oldVersion);

			if (doc instanceof PlainDocument) {
				((PlainDocument)doc).setVersion(oldVersion);
			}
			else
				if (doc instanceof HTMLDocument) {
					((HTMLDocument)doc).setVersions(oldVersion, oldCookedVersion);
					((HTMLDocument)doc).setText(rawText);
				}
				else
					debug.println("can't set version of " + doc.getClass());

			// TODO: support multiple documents
            
            // minwu
			//HashSet oldVersions = new HashSet();
			//oldVersions.add(oldVersion);
            
            Versions oldVersions = new SingleDocVersions(oldVersion);
			NamedRegionSet currRecords = recordHighlighter.getSelection();
			if (currRecords != null) {
				if (oldRecords == null)
					oldRecords = (NamedRegionSet)currRecords.convert(oldVersions);
				recordHighlighter.setSelection(oldRecords);
			}
			NamedRegionSet currFindRecords = getFindAndReplaceHighlighter().getSelection();
			if (currFindRecords != null) {
				if (oldFindRecords == null)
					oldFindRecords = (NamedRegionSet)currRecords.convert(oldVersions);
				getFindAndReplaceHighlighter().setSelection(oldFindRecords);
			}
			browser.getSelectionMode().undidTo(oldVersion);

			((BrowserCaret)getCaret()).updateBrowser();
		}
	}

	//
	// Multi-selection editing
	//

	boolean replacing = false;

	public void replaceSelection(String content) {
		debug.println("replaceSelection \"" + content + "\"");
		if (content != null && content.length() == 1 && content.charAt(0) == 255) {
			debug.println("replaceSelection() called with DEL argument");
			return;
		}
		replace(browser.getSelection(), null, content, false);

	}

	public void replaceSelection(TransferableSelection mss) {
		replace(browser.getSelection(), mss, null, false);
	}



	void replace(NamedRegionSet selection, TransferableSelection mss, String content, boolean leaveSelected) {
		//RegionSet selectedRegions = selection.getRegions();

		Vector v = null;

		if (mss != null) {
			v = mss.getStrings();
			if (v.size() != 1 && selection.getSize() != 1 && v.size() != selection.getSize()) {
				FeedbackForm.showMessageDialog(
					browser,
					"Can't paste the " + v.size() + " clipboard items to the " + selection.getSize() + " selected regions.\n",
					"Wrong Number of Selected Regions",
					JOptionPane.ERROR_MESSAGE);
				return;
			}
		}

		int len = 0;
		if (content != null && (len = content.length()) == 0)
			content = null;

		SwingDocument doc = (SwingDocument)getDocument();
		RegionSet selectedRegions = doc.getCookedView().getVersion().convert(selection.getSingleDocRegionSet(doc));

		MultipleSelectionEdit msedit = new MultipleSelectionEdit();
		UndoableEditListener ul = CoalescingUndo.getUndoableEditListener(doc);
		ul.undoableEditHappened(new UndoableEditEvent(this, msedit));

		replacing = true;
		try {
			try {
				int[] points = new int[selectedRegions.getSize()*2];
				int pindex = 0;
				DocumentVersion oldver = doc.getVersion();
				//MutableRegionSet afterwards = new MutableRegionSet();
				RegionEnumeration re = selectedRegions.regions();
				int adjust = 0;
				int i = 0;
				int after = 0;
				for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
					if (mss != null) {
						content = (String)v.elementAt(i);
						++i;
						if (i >= v.size())
							i = 0;
						len = content.length();
					}

					int before = r.getStartLow() + adjust;

					if (r.getStartLow() < r.getEndHigh())
						doc.remove(before, r.getEndHigh() - r.getStartLow());
					if (len > 0)
						doc.insertString(before, content, null);

					adjust += len - (r.getEndHigh() - r.getStartLow());
					after = r.getEndHigh() + adjust;

					if (leaveSelected) {
						//afterwards.insert(before, after, doc);
						points[pindex] = before;
						points[pindex+1] = after;
					}
					else {
						//afterwards.insert(after, after, doc);
						points[pindex] = after;
						points[pindex+1] = after;
					}
					pindex += 2;
				}

				int[] points_2 = null;
				if (mss != null && i == 1) {
					points_2 = new int[(v.size()-1)*2];
					pindex = 0; 
					// user pasted N items to 1 location:
					// we've already pasted the first item,
					// now finish pasting the other N-1, 
					// leaving a cursor after each paste
					doc.insertString(after, "\n", null);
					++adjust;
					++after;
					for (int n = v.size(); i < n; ++i) {
						content = (String)v.elementAt(i);
						len = content.length() + 1;
						doc.insertString(after, content + "\n", null);
						adjust += len;
						after += len;
						//afterwards.insert(after - 1, after - 1, doc);
						points_2[pindex] = after-1;
						points_2[pindex+1] = after-1;
					}
				}

				MutableRegionSet afterwards = new MutableRegionSet();
				for (pindex = 0; pindex < points.length; pindex += 2) 				
					afterwards.insert(points[pindex], points[pindex+1], 
									  doc.getCookedView().getVersion());
				if (points_2 != null)
					for (pindex = 0; pindex < points_2.length; pindex += 2)
						afterwards.insert(points_2[pindex], points_2[pindex], 
										  doc.getCookedView().getVersion());

				RegionSet rsAfterwards = afterwards;
				rsAfterwards = doc.getVersion().convert(rsAfterwards.getSingleDocRegionSet(doc));

				//NamedRegionSet newSelection = new NamedRegionSet(doc, rsAfterwards);
				NamedRegionSet newSelection = new NamedRegionSet(rsAfterwards);
				browser.setSelection(newSelection);

				browser.getSelectionMode().replaced(selection, newSelection, mss, content);

			}
			catch (BadLocationException e) {
				debug.report(e);
				getToolkit().beep();
			}
		}
		finally {
			msedit.end();
			replacing = false;
		}
	}

	public void cut() {
		copy();
		replaceSelection((String) null);
	}

	static final Clipboard lapisClipboard = new Clipboard("LAPIS");
	public static Clipboard getLapisClipboard() {
		return lapisClipboard;
	}

	public void copy() {
		NamedRegionSet selection = browser.getSelection();
		//RegionSet regions = selection.getRegions();
		if (!selection.isAllPoints()) {
			SwingDocument doc = (SwingDocument)getDocument();

			// use TransferableSelection only if we have to, because
			// MS Java won't copy TransferableSelections to other apps --
			// or even other JTextFields in LAPIS
			if (selection.getSize() > 1) {
				Transferable content = new TransferableSelection(selection);
				getToolkit().getSystemClipboard().setContents(new StringSelection(""), clipboardOwner);
				getLapisClipboard().setContents(content, clipboardOwner);
				debug.println("copied " + content);
			}
			else {
				RegionSet regions = doc.getCookedView().getVersion().convert(selection.getSingleDocRegionSet(doc));
				String text = doc.getCookedView().getText(regions.regions().first());
				Transferable content = new StringSelection(text);
				getToolkit().getSystemClipboard().setContents(content, clipboardOwner);
				getLapisClipboard().setContents(content, clipboardOwner);
			}

			browser.getSelectionMode().copied(selection);
		}

	}

	private static ClipboardOwner clipboardOwner = new ClipboardOwner() {
		public void lostOwnership(Clipboard clipboard, Transferable contents) {}
	};

	public void paste() {
		if (!pasteFrom(getToolkit().getSystemClipboard()) && !pasteFrom(getLapisClipboard())) {
			debug.println("can't paste");
			getToolkit().beep();
		}
	}

	boolean pasteFrom(Clipboard clipboard) {
		Transferable content = clipboard.getContents(this);
		if (content == null)
			return false;

		debug.println("pasting " + content);
		try {
			try {
				TransferableSelection mss = (TransferableSelection)content.getTransferData(TransferableSelection.multiStringFlavor);
				debug.println("pasting multiple strings " + mss.getStrings());
				replaceSelection(mss);
				return true;
			}
			catch (UnsupportedFlavorException e1) {
				try {
					String s = (String)content.getTransferData(DataFlavor.stringFlavor);
					if (s != null && s.length() > 0) {
						debug.println("pasting string " + s);
						replaceSelection(s);
						return true;
					}
				}
				catch (UnsupportedFlavorException e2) {}
			}
		}
		catch (Exception e) {
			debug.report(e);
		}

		return false;
	}

	public void scrollToReference(String ref) {
		super.scrollToReference(ref);
	}

	public void firePropertyChange(String propertyName, Object oldObj, Object newObj) {
		super.firePropertyChange(propertyName, oldObj, newObj);
	}

	public URL getPage() {
		return (URL) ((SwingDocument)getDocument()).getProperty(SwingDocument.URLProperty);
	}

    /**
     * Scrolls browser pane to make at least one region from a region set visible.
     * @param s RegionSet to make visible
     * @effects If s can fit entirely into the browser pane, scrolls browser pane to make all of s visible.
     * Otherwise, if a region of s is already visible, nothing happens.  Otherwise, scrolls the browser pane
     * so as to center the first region in s in the pane.
     */
	public void scrollRegionsToVisible(RegionSet s) {
		if (s.isEmpty())
			return;

		try {
			Rectangle visibleRect = getVisibleRect();
			Region visible = rectangleToRegion(visibleRect);

			// shrink the visibleRect by 15 pixels on each side, to make sure we have some context
			visibleRect.x += 15;
			visibleRect.y += 15;
			visibleRect.width = Math.max(visibleRect.width - 30, 1);
			visibleRect.height = Math.max(visibleRect.height - 30, 1);

			// check whether s could fit entirely in the window
			// Region bbox = s.singleDocize ().getBoundingBox();
			// TODO: change later to fit MultiDoc
			Region bbox = 
				s.getSingleDocRegionSet(((SwingDocument)getDocument())).getBoundingBox();
			Rectangle bboxRect = regionToRectangle(bbox);
			if (visibleRect.contains(bboxRect.x, bboxRect.y) && visibleRect.contains(bboxRect.x + bboxRect.width, bboxRect.y + bboxRect.height))
				return; // entire s is already visible

			if (bboxRect.height < visibleRect.height) {
				// s can fit entirely into the window, so scroll it all into view
				scrollRegionToVisible (bbox);
				return;
			}

			// check whether any regions in s are already visible
			if (bboxRect.intersects(visibleRect)) {
				RegionEnumeration re = s.regions((Region)visible.overlaps());
				for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
					// make sure r really intersects the screen
					Rectangle rect = regionToRectangle(r);
					debug.println("testing " + r + " at " + rect + " against " + visibleRect);
					if (rect.intersects(visibleRect)) {
						debug.println(r + " is already visible on screen");
						return; // no need to scroll
					}
				}
			}

			// failing everything, just scroll to first region in s
			scrollRegionToVisible (s.regions().first());

		}
		catch (BadLocationException e) {
			debug.report(e);
		}

	}

    /**
     * Scrolls browser pane to make a region visible, ideally centered.
     * @param r Region to make visible
     * @effects Scrolls browser pane to center r in the pane.  If r is bigger than
     * the pane, scrolls to put the beginning of r at the top of the pane.
     */
	public void scrollRegionToVisible(Region r) {
        scrollRegionToVisible (r, true);
    }
    
    /**
     * Scrolls browser pane to make a region visible.
     * @param r Region to make visible
     * @param centered true if r should be centered, false if r should be at top or bottom (whichever requires less scrolling)
     * @effects If centered=true, scrolls browser pane to center r in the pane (if r is bigger than pane, the start of r is scrolled
     * to the start of the pane). 
     * If centered=false, scrolls pane as little as possible until as much of r as will fit is displayed, at the top or bottom of the pane.
     */
    public void scrollRegionToVisible (Region r, boolean centered) {
		try {
			Rectangle rect = regionToRectangle(r);
            Rectangle visibleRect = getVisibleRect();

            if (centered) {
                // center the region in the window by enlarging
                // its rectangle to the window's width and height
                if (rect.width < visibleRect.width)
                    rect.x = Math.max(0, rect.x - (visibleRect.width - rect.width) / 2);
                if (rect.height < visibleRect.height)
                    rect.y = Math.max(0, rect.y - (visibleRect.height - rect.height) / 2);
                
                rect.width = visibleRect.width;
                rect.height = visibleRect.height;
            } else {
                // enlarge rectangle by 15 pixels on each side, to
                // get a little context
                rect.x = Math.max (rect.x - 15, 0);
                rect.y = Math.max (rect.y - 15, 0);
                rect.width += 30;
                rect.height += 30;              
            }

			// scroll it to visible
			scrollRectToVisible(rect);
		}
		catch (BadLocationException e) {
			debug.report(e);
		}
	}

	// make scrollRectToVisible public
	public void scrollRectToVisible(Rectangle r) {
		super.scrollRectToVisible(r);
	}

	// returns an interval representing the visible portion of the document
	public Region rectangleToRegion(Rectangle rect) {
		TextUI ui = getUI();

		// start and end Swing document positions corresponding to upper-left
		// and lower-right corners of rect
		int s = ui.viewToModel(this, new Point(rect.x, rect.y));
		int e = ui.viewToModel(this, new Point(rect.x + rect.width, rect.y + rect.height));

		SwingDocument doc = (SwingDocument)getDocument();
		CoordinateMap cookedToRaw = doc.getVersion().diff(doc.getCookedView().getVersion());
		s = cookedToRaw.mapStart (s);
		e = cookedToRaw.mapEnd (e);
		return new Region(s, e, doc);
	}

	// returns screen rectangle corresponding to Region (in raw coordinates).
	public Rectangle regionToRectangle(Region r) throws BadLocationException {
		int[] pos = rawToCooked(r);

		// Convert cooked coordinates to 2D view coordinates
		TextUI ui = getUI();
		Rectangle rectStart = ui.modelToView(this, pos[0], Position.Bias.Forward);
		debug.println("rectStart: " + rectStart.toString());
		Rectangle rectEnd = ui.modelToView(this, pos[1], Position.Bias.Backward);
		debug.println("rectEnd: " + rectEnd.toString());
		return rectStart.union(rectEnd);
	}

	public Rectangle regionToExpandedRectangle(Region r) throws BadLocationException {
		int[] pos = rawToCooked(r);

		// Convert cooked coordinates to 2D view coordinates
		TextUI ui = getUI();
		Rectangle rectStart = ui.modelToView(this, pos[0], Position.Bias.Forward);
		debug.println("rectStart: " + rectStart.toString());
		Rectangle rectEnd = ui.modelToView(this, pos[1], Position.Bias.Backward);
		debug.println("rectEnd: " + rectEnd.toString());

		if (rectStart.x <= rectEnd.x) {
			if ((rectStart.y + rectStart.height) != (rectEnd.y + rectEnd.height)) {
				return new Rectangle(0, rectStart.y, (int)this.getWidth(), rectStart.union(rectEnd).height);
			}
			else
				return rectStart.union(rectEnd);
		}
		else
			return new Rectangle(0, rectStart.y, (int)this.getWidth(), rectStart.union(rectEnd).height);
	}

	// returns Region (in raw coordinates) nearest to x,y Point
	public Region pointToRegion(Point pt) {
		TextUI ui = getUI();
		int p = ui.viewToModel(this, pt);
		SwingDocument doc = (SwingDocument)getDocument();

		// in COOKED coordinate
		Region r = new Region(p, p, doc.getCookedView().getVersion());
		// convert to RAW coordinate
		r = (Region)doc.getVersion().convert(r);
		return r;
	}

	public Region cookedToRaw(int s, int e) {
		SwingDocument doc = (SwingDocument)getDocument();
		Region r = new Region (s, e, doc.getCookedView().getVersion());
		
		r = (Region)doc.getVersion().convert(r);
		return r;
	}

	public int[] rawToCooked(Region r) {
		SwingDocument doc = (SwingDocument)getDocument();
		int len = doc.getCookedView().toRegion().getLength();
		r = (Region)doc.getCookedView().getVersion().convert(r);
		return new int[] { Math.min(r.getStartLow(), len), Math.min(r.getEndHigh(), len)};
	}

	/*
	 * Override MIME type to editor kit mapping behavior of JEditorPane,
	 * which is painfully broken in Swing 1.1
	 */
	Hashtable editorKits;
	static final String DEFAULT_MIME_TYPE = "text/plain";

	protected void initEditorKits() {
		editorKits = new Hashtable(); // maps MIME type to editor kit
		editorKits.put("text/plain", new lapis.swing.PlainEditorKit());
		editorKits.put("text/html", new lapis.swing.HTMLEditorKit());
	}

	protected EditorKit createDefaultEditorKit() {
		return getEditorKitForContentType("text/plain");
	}

	public EditorKit getEditorKitForContentType(String type) {
		if (editorKits == null)
			initEditorKits();
		EditorKit kit = (EditorKit)editorKits.get(type);
		if (kit == null)
			kit = (EditorKit)editorKits.get(DEFAULT_MIME_TYPE);
		return kit;
	}

	public void setEditorKitForContentType(String type, EditorKit kit) {
		if (editorKits == null)
			initEditorKits();
		editorKits.put(type, kit);
	}

	/**
	 * Override caret so it only responds to left clicks (not all clicks)
	 */
	class BrowserCaret extends DefaultCaret {
		boolean usingMouse = false;
		// true if currently dragging mouse

		boolean selecting = false;
		// true if mouse drag is creating a selection

		Region clickedRegion;
		// selected region that was under mouse when it was first pressed

		public BrowserCaret() {
			addCaretListener(new CaretListener() {
				public void caretUpdate(CaretEvent evt) {
					if (!usingMouse && !selecting && !replacing && !undoing && !arrowKeyEventsQueued())
						updateBrowser();
				}
			});
		}

		public void updateBrowser() {
			SwingDocument doc = (SwingDocument)getDocument();
			if (doc != null && !doc.isLoaded())
				return;

			int dot = getDot();
			int mark = getMark();
			debug.println("dot=" + dot + ", mark=" + mark);
			Region r = cookedToRaw(Math.min(dot, mark), Math.max(dot, mark));
			browser.clearSelection();
			browser.getSelectionMode().selectRegion(r);
			scrollRegionToVisible(cookedToRaw(dot, dot), false);
		}

		public void setVisible(boolean f) {
			// prevent caret from becoming visible -- RegionHighlighter
			// does it all
		}

		protected void adjustVisibility(final Rectangle r) {
			// don't let caret control scrolling, except when it
			// actually changes the region set.  updateBrowser() 
			// takes care of scrolling in that case.
		}

		public void mousePressed(MouseEvent evt) {
			JTextComponent c = getComponent();
			if (c.isEnabled())
				c.requestFocus();

			if (evt.isShiftDown()) {
				super.mousePressed(evt);
				return;
			}

			usingMouse = true;

			SwingDocument doc = (SwingDocument)getDocument();
			Point pt = new Point(evt.getX(), evt.getY());
			boolean controlDown = evt.isControlDown();

			debug.println("ignoreFocusClick = " + ignoreFocusClick);
			if (!controlDown && ignoreFocusClick)
				return;

			if (SwingUtilities.isLeftMouseButton(evt)) {

				browser.getSelectionMode().startMouseSelection();
				selecting = true;

				if (!controlDown && browser.getSelectionMode().clearSelectionOnClick()) {
					browser.clearSelection();
				}
				else {
					Region click = pointToRegion(pt);
					clickedRegion = browser.getSelection().regions((Region)click.contains()).first();
				}

				RegionHighlighter.Handle h;

				if (controlDown && (h = selectionHighlighter.handleAtPoint(pt)) != null) {
					// move handle of highlighted region
					Region r = h.getRegion();
					int[] pos = rawToCooked(r);
					if (h.isStart()) {
						setDot(pos[1]);
						moveDot(pos[0]);
					}
					else {
						setDot(pos[0]);
						moveDot(pos[1]);
					}
					//feedbackHighlighter.setSelection(new NamedRegionSet(doc, r));
					feedbackHighlighter.setSelection(new NamedRegionSet(r));

					browser.getSelectionMode().unselectRegion(r);
					Browser.debug.println("adjust handle of " + r);
				}
				else {
					// start dragging it out
					TextUI ui = getUI();
					int dot = ui.viewToModel(BrowserPane.this, pt);
					setDot(dot);
					moveDot(dot);
					Region r = cookedToRaw(dot, dot);

					if (clickedRegion == null)
						feedbackHighlighter.setSelection(new NamedRegionSet(r));
				}
			}
		}

		public void mouseReleased(MouseEvent evt) {
			if (selecting) {
				NamedRegionSet sel = feedbackHighlighter.getSelection();
				feedbackHighlighter.setSelection(null);

				if (sel != null)
					browser.getSelectionMode().selectRegion(sel);
				else
					if (clickedRegion != null)
						browser.getSelectionMode().unselectRegion(clickedRegion);

				browser.getSelectionMode().stopMouseSelection();
				selecting = false;
				clickedRegion = null;
			}
			usingMouse = false;
		}

		public void mouseClicked(MouseEvent evt) {}

		public void mouseDragged(MouseEvent evt) {
			if (SwingUtilities.isLeftMouseButton(evt)) {
				if (!selecting)
					mousePressed(evt);

				if (selecting) {
					Point pt = new Point(evt.getX(), evt.getY());
					TextUI ui = getUI();
					int dot = ui.viewToModel(BrowserPane.this, pt);
					moveDot(dot);

					int mark = getMark();
					Region r = cookedToRaw(Math.min(dot, mark), Math.max(dot, mark));
					SwingDocument doc = (SwingDocument)getDocument();

					//NamedRegionSet sel = (mark != dot || clickedRegion == null) ? new NamedRegionSet(doc, r) : null;
					NamedRegionSet sel = (mark != dot || clickedRegion == null) 
									? new NamedRegionSet(r) 
									: null;
					feedbackHighlighter.setSelection(sel);

					scrollRegionToVisible(cookedToRaw(dot, dot), false);
				}
			}
		}
	}

	void updateIgnoreFocusClick() {
		NamedRegionSet sel;

		ignoreFocusClick =
			!haveFocus && (sel = selectionHighlighter.getSelection()) != null && !sel.isEmpty() && !(sel.getSize() == 1 && sel.isAllPoints());
	}

	// return true if the next input event waiting in the system event queue
	// is an arrow key press
	boolean arrowKeyEventsQueued() {
		EventQueue queue = getToolkit().getSystemEventQueue();

		if (queue.peekEvent(MouseEvent.MOUSE_MOVED) != null)
			// there's a mouse move in the queue, which may or may not
			// come before the arrow key -- be conservative
			return false;

		KeyEvent keyEvent = (KeyEvent)queue.peekEvent(KeyEvent.KEY_PRESSED);
		if (keyEvent == null)
			return false;

		switch (keyEvent.getKeyCode()) {
			case KeyEvent.VK_LEFT :
			case KeyEvent.VK_RIGHT :
			case KeyEvent.VK_UP :
			case KeyEvent.VK_DOWN :
			case KeyEvent.VK_HOME :
			case KeyEvent.VK_END :
				return true;
			default :
				return false;
		}
	}

	public RegionHighlighter getFinderHighlighter() {
		return finderHighlighter;
	}

	public RegionHighlighter getFindAndReplaceHighlighter() {
		return findAndReplaceHighlighter;
	}

}
