/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.swing;

import javax.swing.text.*;
import java.awt.*;
import lapis.*;

public class ColoringView extends WrappedPlainView {

  public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;
  
  boolean isColored;
  RegionSetStyle styleSpec;

  public ColoringView(Element elem) {
    this(elem, false);
  }
  
  public ColoringView(Element elem, boolean colored) {
    super(elem);
    debug.println("Constructing a ColoringView");
    this.isColored = colored;
  }

  public void applyStyle(RegionSetStyle rs) {
    styleSpec = rs;
  }

  public void setColoring(boolean isColored) {
    this.isColored = isColored;
    getContainer().repaint();
  }

  /**
   * Renders the given range in the model as normal unselected
   * text.
   *
   * @param g the graphics context
   * @param x the starting X coordinate
   * @param y the starting Y coordinate
   * @param p0 the beginning position in the model
   * @param p1 the ending position in the model
   * @returns the location of the end of the range
   * @exception BadLocationException if the range is invalid
   */
  protected int drawUnselectedText(Graphics g, int x, int y,
				   int p0, int p1) throws BadLocationException {
    
    // debug.println("Entered drawUnselectedText...");
    //  |---------x----------x---------|
    // p0         <-colored-->        p1
    
    if(isColored) {
      int mark = p0;
      Color last = null;
      javax.swing.text.Document doc = getDocument();
	  // same as ColoringContext class. need debug.assertion
	  // This Region is only used to determine the points for drawing.
      Region drawRegion = new Region(p0, p1, lapis.Document.DEFAULT);
      // may need to use melt or flatten here - but how?
      RegionEnumeration regions = styleSpec.r.regions((Region) drawRegion.overlaps());

      for(Region r = regions.first(); r != null; r = regions.next()) {
	// if mark < r.start
	// draw(p, r.start, normal); mark <- r.start
	if(mark < r.getStart()) {
	  Segment snippet = getLineBuffer();
	  doc.getText(mark, r.getStart() - mark, snippet);
	  javax.swing.text.Utilities.drawTabbedText(snippet, x, y, g, this, mark);
	  mark = r.getStart();
	}
	
	// get color of text to be drawn
	Color fg = StyleConstants.getForeground(styleSpec.getStyle());
	last = g.getColor(); // save last color
	
	// draw(max(p0, r.start), min(r.end, p1), colored)
	g.setColor(fg);
	int r0 = Math.max(p0, r.getStart()),
	  r1 = Math.min(r.getEnd(), p1);
	Segment styled = getLineBuffer();
	doc.getText(r0, r1 - r0, styled);
	javax.swing.text.Utilities.drawTabbedText(styled, x, y, g, this, r0);
	mark = r1;
      }
      
      // draw(mark, p1, normal)
      g.setColor(last);
      Segment normal = getLineBuffer();
      doc.getText(mark, p1 - mark, normal);
      javax.swing.text.Utilities.drawTabbedText(normal, x, y, g, this, mark);
      mark = p1; // not really necessary, but do it for the sake of completeness
      
      return x;
    }
    return super.drawUnselectedText(g, x, y, p0, p1);
  }

  
  protected int drawSelectedText(Graphics g, int x, int y,
				 int p0, int p1) throws BadLocationException {
    debug.println("drawSelectedText: p0 =  " + p0 + "; p1 = " + p1 + "; x = " + x + "; y = " + y);
    return super.drawSelectedText(g, x, y, p0, p1);
  }
  
}
