/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.swing;

import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.plaf.basic.BasicComboBoxEditor;
import lapisx.swing.*;
import javax.swing.undo.*;

/**
 * Text field of Command combo box in Browser. Two purposes:
 * <ul>
 * <li> Overrides the standard combo box editor so that losing focus
 * doesn't automatically send an Action event; only pressing Enter does.</li>
 * <li> Next-word and previous-word only use whitespace as word boundaries,
 * not punctuation.  That makes URLs easier to select and edit.</li>
 * </ul>
 */
class CommandEditor extends BasicComboBoxEditor {
    public  CommandEditor () {
        super ();
        editor.setCaret (new CommandCaret ());

        Keymap keymap = editor.getKeymap ("lapis.Browser.CommandEditor");
        if (keymap == null) {
            keymap = editor.addKeymap ("lapis.Browser.CommandEditor",
                                       editor.getKeymap ());
            keymap.addActionForKeyStroke 
                (KeyStroke.getKeyStroke (KeyEvent.VK_RIGHT,
                                         KeyEvent.CTRL_MASK),
                 new MoveCaretByWord ("next-word", true, false));
            keymap.addActionForKeyStroke 
                (KeyStroke.getKeyStroke (KeyEvent.VK_LEFT,
                                         KeyEvent.CTRL_MASK),
                 new MoveCaretByWord ("previous-word", false, false));
            keymap.addActionForKeyStroke 
                (KeyStroke.getKeyStroke (KeyEvent.VK_RIGHT,
                                         KeyEvent.CTRL_MASK | KeyEvent.SHIFT_MASK),
                 new MoveCaretByWord ("select-next-word", true, true));
            keymap.addActionForKeyStroke 
                (KeyStroke.getKeyStroke (KeyEvent.VK_LEFT,
                                         KeyEvent.CTRL_MASK | KeyEvent.SHIFT_MASK),
                 new MoveCaretByWord ("select-previous-word", false, true));
        }
        editor.setKeymap (keymap);

        CoalescingUndo.setUndoableEditListener 
            (((JTextComponent) getEditorComponent ())
             .getDocument (), new UndoManager ());
    }

    public void focusGained(FocusEvent e) {
        selectAll ();
    }

    public void focusLost(FocusEvent e) {
        JTextField f = (JTextField)getEditorComponent ();
        Caret caret = f.getCaret ();
        caret.setDot (caret.getDot ());
    }

    static class CommandCaret extends DefaultCaret {
        public void mouseClicked(MouseEvent e) {
            if ((e.getModifiers() & InputEvent.BUTTON1_MASK) != 0
                && e.getClickCount() == 2) {
                try {
                    JTextComponent c = getComponent ();
                    Document doc = c.getDocument ();
                    String s = doc.getText (0, doc.getLength ());
                    setDot (backToWordStart (s, getDot ()));
                    moveDot (forwardToWordEnd (s, getDot ()));
                } catch (BadLocationException ex) {
                }
            } else
                super.mouseClicked (e);
        }
    }

    static class MoveCaretByWord extends TextAction {
        boolean forward;
        boolean select;
        public  MoveCaretByWord (String name, boolean forward, boolean select) {
            super (name);
            this.forward = forward;
            this.select = select;
        }
        public void actionPerformed (ActionEvent event) {
            JTextComponent c = getTextComponent (event);
            if (c != null) {
                try {
                    Document doc = c.getDocument ();
                    String s = doc.getText (0, doc.getLength ());
                    int pos = c.getCaretPosition ();
                    if (forward)
                        pos = forwardToWordEnd (s, pos);
                    else
                        pos = backToWordStart (s, pos);

                    if (select)
                        c.moveCaretPosition (pos);
                    else
                        c.setCaretPosition (pos);
                } catch (BadLocationException ex) {
                }
            }                    
        }
    }
    static int backToWordStart (String s, int i) {
        while (i > 0 
               && Character.isWhitespace (s.charAt (i-1)))
            --i;
        while (i > 0 
               && !Character.isWhitespace (s.charAt (i-1)))
            --i;
        return i;
    }
    static int forwardToWordEnd (String s, int i) {
        int n = s.length ();
        while (i < n
               && Character.isWhitespace (s.charAt (i)))
            ++i;
        while (i < n
               && !Character.isWhitespace (s.charAt (i)))
            ++i;
        return i;
    }
}
