/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.swing;

import lapis.*;
import lapis.Document; // override ambiguity with javax.swing.text.Document
import lapis.tc.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.util.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import javax.swing.event.*;
import lapisx.swing.*;

public class Demonstrator extends Browser {

    /**
     * Action that pops up a Demonstrate window.
     */
    public static class Demonstrate extends AcceleratedAction {
        Browser browser;
        public Demonstrate (Browser browser) {
            super ("&Demonstrate Script...");
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            Demonstrator d = new Demonstrator (browser);
            d.setVisible (true);
            Point p = browser.getLocation ();
            p.translate (50, 50);
            d.setLocation (p);
        }
    }


    Browser browser;          // associated browser showing input page
    
    FilledOutDocument fdoc = null;

    JToolBar demonstrateToolbar;
   
    Hashtable modelToElement; // maps Swing model for a form field to
    // the document element representing form field.  Changes with every
    // new document.

    boolean automaticControl = false;
    // when true, form fields are being set programmatically.
    // FIX: need a better way to detect whether field events are coming
    // from user or from program

    public Demonstrator () {
        super ();

        demonstrateToolbar = new JToolBar (JToolBar.HORIZONTAL);
        demonstrateToolbar.setFloatable (false);
        demonstrateToolbar.add (doneAction);
        
        Layout.vertical ((JComponent) getContentPane ())
            .add (demonstrateToolbar, Layout.FIXED_SIZE);

        bind (getActivePane ().getKeymap ());
        bind (new JTextField ().getKeymap ());
        bind (new JTextArea ().getKeymap ());
    }

    void bind (Keymap map) {
        map.addActionForKeyStroke 
            (KeyStroke.getKeyStroke ('C', KeyEvent.CTRL_MASK),
             copyAction);
        map.addActionForKeyStroke 
            (KeyStroke.getKeyStroke ('V', KeyEvent.CTRL_MASK),
             pasteAction);
    }

    public Browser make () {
        return new Demonstrator ();
    }

    public Demonstrator (Browser browser) {
        this ();
        this.browser = browser;
        setInputPage ();
    }

    
    public Action doneAction = new AbstractAction ("Done Demonstration") {
        public void actionPerformed (ActionEvent event) {
            String prolog = "set " 
                + Interpreter.PARAMETERS_VAR 
                + " [doc]\n";

            popupScriptEditor (prolog + convertHistoryToScript ());
        }
    };

    public void loadedDocument (SwingDocument doc) {
        super.loadedDocument (doc);
        if (fdoc != null)
            setCommand (fdoc.toString (), false);
        snoop (doc);
    }

    /**
     * Change the current document to a Document object.
     */
    public void setDocument (Document doc) {
        if (doc instanceof FilledOutDocument) {
            fdoc = (FilledOutDocument)doc;
            fdoc.fillout ();
            super.setDocument (fdoc.doc);//fdoc.getDocument ());
        } else {
            fdoc = null;
            if (doc instanceof SwingDocument)
                resetForm ((SwingDocument)doc);
            super.setDocument (doc);
        }
    }

    /**
      * Store input page as a Tcl variable
      */
    void setInputPage () {
        if (browser != null) {
            debug.println ("setting input page");
            getInterpreter ().setVar (Interpreter.PARAMETERS_VAR, 
                                      browser.getDocument ());
        }
    }        
        

    /**
      * Start watch all form fields in current document for user events.
      */
    void snoop (SwingDocument doc) {
        if (doc == null)
            return;

        modelToElement = findFieldModels (doc);
        addFieldListeners (getActivePane ());
    }

    /**
      * Find all the form field elements in the document (INPUT, SELECT, TEXTAREA), 
      * extract the Swing model representing each one (e.g. ButtonModel for buttons,
      * Document for text fields), and make a hashtable mapping the model to the
      * element.
      */
    Hashtable findFieldModels (SwingDocument doc) {
        Hashtable modelToElement = new Hashtable ();

	ElementIterator iter = 
            new ElementIterator(doc.getDefaultRootElement());
        Element elem;
             
        while ((elem = iter.next ()) != null) {
            AttributeSet attr = elem.getAttributes ();
            HTML.Tag tag = (HTML.Tag)attr.getAttribute (StyleConstants.NameAttribute);
            Object model = attr.getAttribute(StyleConstants.ModelAttribute);
            
            if (model != null
                && (tag == HTML.Tag.INPUT
                    || tag == HTML.Tag.TEXTAREA
                    || tag == HTML.Tag.SELECT))
                modelToElement.put (model, elem);
        }

        return modelToElement;
    }

    Element getElement (java.util.EventObject event) {
        return getElement ((Component)event.getSource ());
    }

    Element getElement (Component comp) {
        Object model;

        if (comp instanceof AbstractButton)
            model = ((AbstractButton)comp).getModel ();
        else if (comp instanceof JTextComponent)
            model = ((JTextComponent)comp).getDocument ();
        else if (comp instanceof JList)
            model = ((JList)comp).getModel ();
        else if (comp instanceof JComboBox)
            model = ((JComboBox)comp).getModel ();
        else
            return null;

        if (model == null)
            return null;

        return (Element)modelToElement.get (model);
    }

    /**
      * Add a listener to every form field in the browser pane, so that
      * user events in these fields can be translated into history events.
      */
    void addFieldListeners (Component comp) {
        if (comp instanceof JButton) {
            ((JButton)comp).removeActionListener (buttonFieldListener);
            ((JButton)comp).addActionListener (buttonFieldListener);
        } else if (comp instanceof JTextField
                      || comp instanceof JTextArea) {
            comp.removeFocusListener (textFieldListener);
            comp.addFocusListener (textFieldListener);
        } else if (comp instanceof JToggleButton) {
            ((JToggleButton)comp).removeActionListener (checkFieldListener);
            ((JToggleButton)comp).addActionListener (checkFieldListener);
        } else if (comp instanceof JList) {
            ((JList)comp).removeListSelectionListener (listFieldListener);
            ((JList)comp).addListSelectionListener (listFieldListener);
        } else if (comp instanceof JComboBox) {
            ((JComboBox)comp).removeActionListener (comboFieldListener);
            ((JComboBox)comp).addActionListener (comboFieldListener);
        }

        if (comp instanceof Container) {
            Container container = (Container)comp;
            for (int i = 0, n = container.getComponentCount (); i < n; ++i)
                addFieldListeners (container.getComponent (i));
        }
    }

    public ActionListener buttonFieldListener = new ActionListener () {
        public void actionPerformed (ActionEvent event) {
            if (automaticControl)
                return;

            Element elem = getElement (event);
            if (elem == null)
                return;

            String name = (String)elem.getAttributes()
                .getAttribute (HTML.Attribute.NAME);
            renameNextDocument ("submit");
            // FIX: handle non-submit buttons
        }
    };

    String oldValue = "";
    String selectionCopy = null;
    String selectionPattern = null;

    public FocusListener textFieldListener = new FocusListener () {
        public void focusGained (FocusEvent event) {
            if (automaticControl)
                return;

            oldValue = ((JTextComponent)event.getSource ()).getText ();
            selectionCopy = null;
            selectionPattern = null;
        }
        public void focusLost (FocusEvent event) {
            if (automaticControl)
                return;

            Element elem = getElement (event);
            if (elem == null)
                return;
            String name = (String)elem.getAttributes()
                .getAttribute (HTML.Attribute.NAME);

            String newValue = ((JTextComponent)event.getSource ()).getText ();

            if (selectionCopy != null && selectionCopy.equals (newValue))
                addEventToHistory ("enter Field." + name + 
                                   " [extract -as text "
                                   + getInterpreter ().quote (selectionPattern)
                                   + " $source]");
            else if (!newValue.equals (oldValue))
                addEventToHistory ("enter Field." + name 
                                   + " " + getInterpreter ().quote (newValue));
            oldValue = newValue;
            selectionCopy = null;
            selectionPattern = null;
        }
    };

    public ActionListener checkFieldListener = new ActionListener () {
        public void actionPerformed (ActionEvent event) {
            if (automaticControl)
                return;

            Element elem = getElement (event);
            if (elem == null)
                return;

            String name = (String)elem.getAttributes()
                .getAttribute (HTML.Attribute.NAME);
            boolean value = ((AbstractButton)event.getSource ())
                .isSelected ();
            addEventToHistory ("enter " 
                                + name + "Field " 
                                + (value ? "on" : "off"));
        }
    };

    public ListSelectionListener listFieldListener = new ListSelectionListener () {
        public void valueChanged (ListSelectionEvent event) {
            if (automaticControl)
                return;

            Element elem = getElement (event);
            if (elem == null)
                return;
            String name = (String)elem.getAttributes()
                .getAttribute (HTML.Attribute.NAME);
            Object value = ((JList)event.getSource())
                .getSelectedValue (); // FIX: deal with multiple selections too
            // FIX: convert value to Tcl string
            addEventToHistory ("enter " 
                                + name + "Field {" 
                                + value + "}");
            
        }
    };

    public ActionListener comboFieldListener = new ActionListener () {
        public void actionPerformed (ActionEvent event) {
            if (automaticControl)
                return;

            Element elem = getElement (event);
            if (elem == null)
                return;
            String name = (String)elem.getAttributes()
                .getAttribute (HTML.Attribute.NAME);
            Object value = ((JComboBox)event.getSource())
                .getSelectedItem (); // FIX: deal with multiple selections too
            // FIX: convert value to Tcl string
            addEventToHistory ("enter " 
                                + name + "Field {" 
                                + value + "}");
        }
    };


    void addEventToHistory (final String cmd) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                debug.println (cmd);
                FilledOutDocument fdoc = new FilledOutDocument (getDocument ());
                fdoc.putProperty (fdoc.CommandProperty, cmd);
                getPageHistory ().put (fdoc);
            }
        });
    }

    void renameNextDocument (final String cmd) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                SwingDocument doc = getDocument ();
                doc.putProperty (SwingDocument.CommandProperty, cmd);
                setCommand (cmd, false);
            }
        });
    }

    void resetForm (SwingDocument doc) {
	ElementIterator it = new ElementIterator(doc.getDefaultRootElement());
	Element next;
	boolean startedReset = false;

	while((next = it.next()) != null) {
	    AttributeSet elemAttr = next.getAttributes();
            Object m = elemAttr.getAttribute(StyleConstants.ModelAttribute);
            // FIX: handle TEXTAREA and SELECT properly
//             if (m instanceof TextAreaDocument) {
//                 TextAreaDocument doc = (TextAreaDocument)m;
//                 doc.reset();
//             } else	
                if (m instanceof javax.swing.text.PlainDocument) {
                try {
                    javax.swing.text.PlainDocument pdoc =  (javax.swing.text.PlainDocument)m;
                    pdoc.remove(0, pdoc.getLength());
                    if (elemAttr.getAttribute (StyleConstants.NameAttribute)
                        == HTML.Tag.INPUT) {
                        String value = (String)elemAttr.getAttribute(HTML.Attribute.VALUE);
                        if (value != null) {
                            pdoc.insertString(0, value, null);
                        }
                    }
                } catch (BadLocationException e) {
                }
//             } else	if (m instanceof OptionListModel) {
//                 OptionListModel model = (OptionListModel) m;
//                 int size = model.getSize();
//                 for (int i = 0; i < size; i++) {
//                     model.removeIndexInterval(i, i);
//                 }
//                 BitSet selectionRange = model.getInitialSelection();
//                 for (int i = 0; i < selectionRange.size(); i++) {
//                     if (selectionRange.get(i)) {
//                         model.addSelectionInterval(i, i);
//                     }
//                 }
//             } else 	if (m instanceof OptionComboBoxModel) {
//                 OptionComboBoxModel model = (OptionComboBoxModel) m;
//                 Option option = model.getInitialSelection();
//                 if (option != null) {
//                     model.setSelectedItem(option);
//                 }
            } else 	if (m instanceof JToggleButton.ToggleButtonModel) {
                boolean checked = ((String)elemAttr.getAttribute(HTML.Attribute.CHECKED) != null);
                JToggleButton.ToggleButtonModel model = (JToggleButton.ToggleButtonModel)m;
                model.setSelected(checked);
            }
            startedReset = true;
	}
    }

    void setFormField (HTML.Tag tag, String name, String value) {
        if (value == null)
            value = "";

        Element e = findElement (getDocument ().getDefaultRootElement (),
                                 tag, HTML.Attribute.NAME, name);

        if (e == null) {
            debug.println ("Demonstrator: no field called " + name);
            return;
        }
        //debug.println ("Found element " + e);

        Object model = e.getAttributes ()
                        .getAttribute(StyleConstants.ModelAttribute);

        if (model == null) {
            debug.println ("Demonstrator: field " + name + " has no model");
            return;
        }
        //debug.println ("Found model " + model);

        // Replace old contents with value
	if (model instanceof javax.swing.text.Document) {
            // for INPUT type=text and TEXTAREA
            javax.swing.text.Document field = (javax.swing.text.Document) model;
            try {
                field.remove (0, field.getLength ());
            } catch (BadLocationException ex) {}
            
            try {
                field.insertString (0, value, null);
            } catch (BadLocationException ex) {}
            //debug.println ("Set " + name + " to " + model);
        }
        else if (model instanceof ComboBoxModel) {
            // for SELECT (single selection, only 1 line visible)
            ComboBoxModel items = (ComboBoxModel)model;
        found: {
                for (int i = 0, n = items.getSize (); i < n; ++i) {
                    Object item = items.getElementAt (i);
                    if (item != null && item.toString ().equals (value)) {
                        items.setSelectedItem (item);
                        break found;
                    }
                }
                debug.println ("Field " + name + 
                                    " doesn't contain an option " + value);
            }
        }

        // NIY:
        // for SELECT (with multiple lines visible or multi select)
        //      ListModel/ListSelectionModel (defines both interfaces)
        // for INPUT type=submit | reset | image
        //      ButtonModel
        // for INPUT=checkbox | radio
        //      ToggleButtonModel

        else {
            debug.println ("Field " + name + " has unknown model: " + model);
        }
    }

    /**
     * Action invoked when user selects Save Tool.
     */
    public Action saveToolAction = new AbstractAction ("Save Tool") {
        public void actionPerformed (ActionEvent evt) {
        }
    };


    Element findElement (Element e, javax.swing.text.html.HTML.Tag tag, 
                         javax.swing.text.html.HTML.Attribute attr, String value) {
        if (e == null)
            return null;

        // Check current element
        javax.swing.text.html.HTML.Tag myTag =
            (javax.swing.text.html.HTML.Tag)
            e.getAttributes ()
            .getAttribute (StyleConstants.NameAttribute);
        if (myTag == tag) {
            String myValue = (String)e.getAttributes ().getAttribute (attr);
            if (value.equals (myValue))
                return e;
        }

        // Search through children
        for (int i = 0, n = e.getElementCount (); i < n; ++i) {
            Element found = findElement (e.getElement (i), tag, attr, value);
            if (found != null)
                return found;
        }

        return null;
    }

    class FilledOutDocument extends PlainDocument {
        protected SwingDocument doc;
        Hashtable fields = new Hashtable ();  
        // maps Element representing a form field -> String value to fill in

        public FilledOutDocument (SwingDocument doc) {
            super (doc.getFactory (), doc, (String) doc.getProperty (MIMEProperty));
            this.doc = doc;
            extract ();
        }

		/*
        public SwingDocument getDocument () {
            return doc;
        }
		*/
		
        public void extract () {
            fields.clear ();

            ElementIterator iter = 
                new ElementIterator(doc.getDefaultRootElement());
            Element elem;
             
            while ((elem = iter.next ()) != null) {
                AttributeSet attr = elem.getAttributes ();
                HTML.Tag tag = (HTML.Tag)attr.getAttribute (StyleConstants.NameAttribute);
                String name = (String)attr.getAttribute (HTML.Attribute.NAME);
                Object model = attr.getAttribute(StyleConstants.ModelAttribute);
            
                if (model != null
                    && (tag == HTML.Tag.INPUT
                        || tag == HTML.Tag.TEXTAREA
                        || tag == HTML.Tag.SELECT)) {

                    if (model instanceof javax.swing.text.Document) {
                        // for INPUT type=text and TEXTAREA
                        javax.swing.text.Document field = 
                            (javax.swing.text.Document) model;
                        try {
                            fields.put (elem, field.getText (0, field.getLength ()));
                        } catch (BadLocationException e) {
                            throw new RuntimeException (e.toString());
                        }
                    } else if (model instanceof ButtonModel) {
                        fields.put (elem, String.valueOf (((ButtonModel)model).isSelected ()));
                    } else if (model instanceof ListSelectionModel) {
                        ListModel items = (ListModel)model;
                        ListSelectionModel selection = (ListSelectionModel)model;
                        int i = selection.getMinSelectionIndex ();
                        // FIX: handle no-selection and multiple-selection case
                        if (i == -1)
                            fields.put (elem, elem);
                        else
                            fields.put (elem, items.getElementAt (i).toString ());
                    } else if (model instanceof ComboBoxModel) {
                        // for SELECT (single selection, only 1 line visible)
                        ComboBoxModel items = (ComboBoxModel)model;
                        fields.put (elem, items.getSelectedItem ().toString ());
                    }
                    else {
                        Demonstrator.debug.println ("Field " + name + " has unknown model: " + model);
                    }
                }
            }
        }

        public void fillout () {
            automaticControl = true;

            try {
                for (Enumeration e = fields.keys (); e.hasMoreElements (); ) {
                    Element elem = (Element)e.nextElement ();
                    String value = (String)fields.get (elem);
                    AttributeSet attr = elem.getAttributes ();
                    String name = (String)attr.getAttribute (HTML.Attribute.NAME);
                    Object model = attr.getAttribute(StyleConstants.ModelAttribute);

                    if (model == null) {
                        Demonstrator.debug.println ("Demonstrator: field " + name + " has no model");
                        return;
                    }

                    if (model instanceof javax.swing.text.Document) {
                        // for INPUT type=text and TEXTAREA
                        javax.swing.text.Document field = (javax.swing.text.Document) model;
                        try {
                            field.remove (0, field.getLength ());
                        } catch (BadLocationException ex) {}
                
                        try {
                            field.insertString (0, value, null);
                        } catch (BadLocationException ex) {}
                        //debug.println ("Set " + name + " to " + model);
                    } else if (model instanceof ButtonModel) {
                        ((ButtonModel)model).setSelected (Boolean.valueOf (value).booleanValue ());
                    } else if (model instanceof ListSelectionModel) {
                        ListModel items = (ListModel)model;
                        ListSelectionModel selection = (ListSelectionModel)model;
                        selection.clearSelection ();
                        // FIX: handle multiple selections
                    found: {
                            for (int i = 0, n = items.getSize (); i < n; ++i) {
                                Object item = items.getElementAt (i);
                                if (item != null && item.toString ().equals (value)) {
                                    selection.setSelectionInterval (i, i);
                                    break found;
                                }
                            }
                            Demonstrator.debug.println ("Field " + name + 
                                                " doesn't contain an option " + value);
                        }
                    } else if (model instanceof ComboBoxModel) {
                        // for SELECT (single selection, only 1 line visible)
                        ComboBoxModel items = (ComboBoxModel)model;
                    found: {
                            for (int i = 0, n = items.getSize (); i < n; ++i) {
                                Object item = items.getElementAt (i);
                                if (item != null && item.toString ().equals (value)) {
                                    items.setSelectedItem (item);
                                    break found;
                                }
                            }
                            Demonstrator.debug.println ("Field " + name + 
                                                " doesn't contain an option " + value);
                        }
                    }
                    else {
                        Demonstrator.debug.println ("Field " + name + " has unknown model: " + model);
                    }
                }
            } finally {
                automaticControl = false;
            }
        }
    }

    /**
      * Override copy action (for ALL JTextComponents) to mark 
      * which JTextComponent generated the clipboard contents.
      */
    public Action copyAction = new DefaultEditorKit.CopyAction () {
        public void actionPerformed(ActionEvent event) {
            JTextComponent target = getTextComponent(event);
            debug.println ("Copy from Browser");
            if (target != null) {
                try {
                    Clipboard clipboard = 
                        Toolkit.getDefaultToolkit().getSystemClipboard();
                    Caret caret = target.getCaret();
                    int p0 = Math.min(caret.getDot(), caret.getMark());
                    int p1 = Math.max(caret.getDot(), caret.getMark());
                    if (p0 != p1) {
                        javax.swing.text.Document doc = 
                            (javax.swing.text.Document)target.getDocument();
                        String srcData = doc.getText(p0, p1 - p0);
                        StringSelection contents = 
                            new TextSelection (target, p0, p1, srcData);
                        clipboard.setContents(contents, contents);
                    }
                } catch (BadLocationException e) {
                }
            }
        }
    };

    public class TextSelection extends StringSelection {
        public JTextComponent origin;
        public int start, end;
        public String data;

        public TextSelection (JTextComponent origin, int p0,
                              int p1, String data) {
            super (data);
            this.origin = origin;
            start = p0;
            end = p1;
            this.data = data;
        }
    }


    /**
      * Override paste action to detect pastes in form fields.
      */
    public static Action pasteAction = new DefaultEditorKit.PasteAction () {
        public void actionPerformed(ActionEvent event) {
            debug.println ("Paste to Browser");
            super.actionPerformed (event);

            Clipboard clipboard = 
                Toolkit.getDefaultToolkit().getSystemClipboard();
            Transferable content = clipboard.getContents(this);
            if (! (content instanceof TextSelection))
                return;

            JTextComponent origin = ((TextSelection)content).origin;
            Browser originBrowser = getBrowser (origin);
            if (originBrowser == null)
                return;

            JTextComponent target = getTextComponent (event);
            Browser targetBrowser = getBrowser (target);
            if (! (targetBrowser instanceof Demonstrator))
                return;

            ((Demonstrator)targetBrowser).pasted (target, originBrowser, 
                                                  (TextSelection)content);
        }

        Browser getBrowser (Component comp) {
            while (comp != null) {
                if (comp instanceof BrowserPane)
                    return ((BrowserPane)comp).browser;
                comp = comp.getParent ();
            }
            return null;
        }
    };

    void pasted (JTextComponent field, Browser originBrowser, 
                 TextSelection selection) {
        if (originBrowser != browser)
            return;
        
        debug.println ("Pasted from Input Browser to Demonstrator");

        Element elem = getElement (field);
        if (elem == null)
            return;

        selectionCopy = selection.data;
        selectionPattern = "selection"; // by default

        // FIX: assumes the user is copying only the first region in the
        // region set

        // FIX: this code belongs in CopyAction, since highlighted
        // regions may change between copy and paste
        NamedRegionSet sel = originBrowser.getSelection ();
        Region r = sel.regions ().firstFast ();

        if (r != null) {
            int[] pos = originBrowser.getActivePane ().rawToCooked (r);
            debug.println ("first highlighted region: " +
                                pos[0] + "," + pos[1]);
            debug.println ("copied region: " +
                                selection.start + "," + selection.end);
            if (pos[0] == selection.start
                && pos[1] == selection.end) {
                TC pat = sel.getPattern ();
                if (pat != null)
                    selectionPattern = pat.getExpression ();
                debug.println ("Matches! pattern =  " + selectionPattern);
            }
        }
    }
}

