/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.swing;

import lapis.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.undo.*;
import lapisx.swing.*;
import java.util.*;
import java.util.List;

public abstract class EditAction extends AcceleratedAction {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    Browser browser;
    String actionName;

    static Hashtable actions;

    public EditAction (Browser browser,
                       String name, 
                       int keycode, int modifiers, 
                       String tooltip,
                       String actionName) {
        super (name,  keycode, modifiers, tooltip);
        this.browser = browser;
        this.actionName = actionName;
    }

    public EditAction (Browser browser,
                       String name, 
                       Icon icon,
                       int keycode, int modifiers, 
                       String tooltip,
                       String actionName) {
        super (name, icon, keycode, modifiers, tooltip);
        this.browser = browser;
        this.actionName = actionName;
    }

    public void actionPerformed (ActionEvent event) {
        if (actions == null) {
            actions = new Hashtable();
            Action[] actionsArray = browser.getActivePane ().getActions();
            for (int i = 0; i < actionsArray.length; i++) {
                Action a = actionsArray[i];
                actions.put(a.getValue(Action.NAME), a);
            }
        }

        Action action = (Action) actions.get (actionName);
        if (action == null) {
            debug.println ("JEditorPane has no action named " + actionName);
            Toolkit.getDefaultToolkit ().beep ();
            return;
        }

        action.actionPerformed (event);

        JTextComponent c = TextUtil.getFocusedTextComponent ();
        if (c != null)
            c.requestFocus ();
    }


    /**
     * Cut-to-clipboard action.
     */
    public static class Cut extends EditAction {
        public Cut (Browser browser) {
            super (browser,
                   "Cu&t", 
                   SwingConfig.makeIcon ("cut.gif"),
                   'X', KeyEvent.CTRL_MASK,
                   "Cut text to clipboard",
                   DefaultEditorKit.cutAction);
        }
    }

    /**
     * Copy-to-clipboard action.
     */
    public static class Copy extends EditAction {
        public Copy (Browser browser) {
            super (browser,
                   "&Copy",
                   SwingConfig.makeIcon ("copy.gif"),
                   'C', KeyEvent.CTRL_MASK,
                   "Copy text to clipboard",
                   DefaultEditorKit.copyAction);
        }
    }

    /**
     * Paste from clipboard action.
     */
    public static class Paste extends EditAction {
        public Paste (Browser browser) {
            super (browser,
                   "&Paste", 
                   SwingConfig.makeIcon ("paste.gif"),
                   'V', KeyEvent.CTRL_MASK,
                   "Paste text from clipboard",
                   DefaultEditorKit.pasteAction);
        }
    }

    /**
     * Delete text action.
     */
    public static class Delete extends EditAction {
        public Delete (Browser browser) {
            super (browser,
                   "&Delete", 
                   KeyEvent.VK_DELETE, 0,
                   "Delete text",
                   DefaultEditorKit.deleteNextCharAction);
        }
    }

    /**
     * Select-all action.
     */
    public static class SelectAll extends EditAction {
        public SelectAll (Browser browser) {
            super (browser,
                   "Select &All", 
                   'A', KeyEvent.CTRL_MASK,
                   "Select all text in document", 
                   DefaultEditorKit.selectAllAction);
        }
    }


    /**
     * Action that undoes one editing action.
     */
    public static class Undo extends AcceleratedAction {
        Browser browser;
        public Undo (Browser browser) {
            super ("&Undo", 
                   'Z', KeyEvent.CTRL_MASK);
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            UndoManager undo = getUndoManager (TextUtil.getTextComponent (evt));
            if (undo == null) {
                Toolkit.getDefaultToolkit ().beep ();
                return;
            }
            
            try {
                undo.undo ();
            } catch (CannotUndoException e) {
                Toolkit.getDefaultToolkit ().beep ();
            }
        }
    }

    /**
     * Action that redoes one editing action.
     */
    public static class Redo extends AcceleratedAction {
        Browser browser;
        public Redo (Browser browser) {
            super ("&Redo",
                   'Y', KeyEvent.CTRL_MASK);
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            UndoManager undo = getUndoManager (TextUtil.getTextComponent (evt));
            if (undo == null) {
                Toolkit.getDefaultToolkit ().beep ();
                return;
            }
            
            try {
                undo.redo ();
            } catch (CannotRedoException e) {
                Toolkit.getDefaultToolkit ().beep ();
            }
        }
    }

    /**
     * Listener for Edit menu that enables/disables edit actions
     * depending on whether they're possible.
     */
    public static class EditMenuListener implements MenuListener {
        Browser browser;
        Cut cut;
        Copy copy;
        Paste paste;
        Delete delete;
        Undo undo;
        Redo redo;

        public EditMenuListener (Browser browser,
                                 Cut cut, Copy copy, Paste paste,
                                 Delete delete, Undo undo, Redo redo) {
            this.browser = browser;
            this.cut = cut;
            this.copy = copy;
            this.paste = paste;
            this.delete = delete;
            this.undo = undo;
            this.redo = redo;
        }

        public void menuSelected (MenuEvent evt) {
            JTextComponent c = TextUtil.getFocusedTextComponent ();
            if (c == null)
                return;
            
            boolean editable = c.isEditable ();
            boolean nonpoints = 
                (c == browser.getActivePane ())
                ? (!browser.getSelection ().isAllPoints ())
                : (c.getSelectionStart () < c.getSelectionEnd ());
            
            cut.setEnabled (editable && nonpoints);
            copy.setEnabled (nonpoints);
            delete.setEnabled (editable && nonpoints);
            
            UndoManager um = getUndoManager (c);
            undo.setEnabled (editable && um != null && um.canUndo ());
            redo.setEnabled (editable && um != null && um.canRedo ());
        }
        
        public void menuDeselected (MenuEvent evt) {
            cut.setEnabled (true);
            copy.setEnabled (true);
            delete.setEnabled (true);
            undo.setEnabled (true);
            redo.setEnabled (true);
        }
        
        public void menuCanceled (MenuEvent evt) {
            cut.setEnabled (true);
            copy.setEnabled (true);
            delete.setEnabled (true);
            undo.setEnabled (true);
            redo.setEnabled (true);
        }
    }


    /**
     * Get the undo manager installed in a text component's document.
     * @param c JTextComponent to look at
     * @return UndoManager attached to <em>c</em>'s document by {@link
     * lapisx.swing.CoalescingUndo#setUndoableEditListener}, or null
     * if no undo manager found.
     */
    private static UndoManager getUndoManager (JTextComponent c) {
    	if (c == null || c.getDocument () == null)
    		return null;
        UndoableEditListener uel = 
            CoalescingUndo.getUndoableEditListener (c.getDocument ());
        if (uel instanceof UndoManager)
            return (UndoManager) uel;
        else
            return null;
    }

    public static class DownAction extends ExtendedAction {
        public DownAction (Action a) {
            super (a);
        }
        public void actionPerformed (ActionEvent evt) {
            JTextComponent target = TextUtil.getTextComponent(evt);
            if (target == null || target.isEditable ()) {
                super.actionPerformed (evt);
                return;
            }

            int scrollOffset;
            int selectedIndex;
            Rectangle visible = new Rectangle();
            target.computeVisibleRect(visible);
            visible.y += target.getScrollableUnitIncrement (visible, 
                                                     SwingConstants.VERTICAL,
                                                     1);
            if((visible.y+visible.height) > target.getHeight())
                visible.y = (target.getHeight() - visible.height);
            target.scrollRectToVisible(visible);
        }
    }

    public static class UpAction extends ExtendedAction {
        public UpAction (Action a) {
            super (a);
        }
        public void actionPerformed (ActionEvent evt) {
            JTextComponent target = TextUtil.getTextComponent(evt);
            if (target == null || target.isEditable ()) {
                super.actionPerformed (evt);
                return;
            }

            int scrollOffset;
            int selectedIndex;
            Rectangle visible = new Rectangle();
            target.computeVisibleRect(visible);
            visible.y -= target.getScrollableUnitIncrement (visible, 
                                                     SwingConstants.VERTICAL,
                                                     -1);
            if((visible.y+visible.height) < 0)
                visible.y = 0;
            target.scrollRectToVisible(visible);
        }
    }

    public static class PageUpAction extends ExtendedAction {
        public PageUpAction (Action a) {
            super (a);
        }
        public void actionPerformed (ActionEvent evt) {
            JTextComponent target = TextUtil.getTextComponent(evt);
            if (target == null || target.isEditable ()) {
                super.actionPerformed (evt);
                return;
            }

            int scrollOffset;
            int selectedIndex;
            Rectangle visible = new Rectangle();
            Rectangle r;
            target.computeVisibleRect(visible);
            scrollOffset = visible.y;
            visible.y -= visible.height;
            if(visible.y < 0)
                visible.y = 0;
            scrollOffset = scrollOffset - visible.y;
            target.scrollRectToVisible(visible);
        }
    }

    public static class PageDownAction extends ExtendedAction {
        public PageDownAction (Action a) {
            super (a);
        }
        public void actionPerformed (ActionEvent evt) {
            JTextComponent target = TextUtil.getTextComponent(evt);
            if (target == null || target.isEditable ()) {
                super.actionPerformed (evt);
                return;
            }

            int scrollOffset;
            int selectedIndex;
            Rectangle visible = new Rectangle();
            Rectangle r;
            target.computeVisibleRect(visible);
            scrollOffset = visible.y;
            visible.y += visible.height;
            int maxHeight = target.getHeight();
            if((visible.y+visible.height) > maxHeight)
                visible.y = (maxHeight - visible.height);
            scrollOffset = visible.y - scrollOffset;
            target.scrollRectToVisible(visible);
        }
    }

    public static class BeginAction extends ExtendedAction {
        public BeginAction (Action a) {
            super (a);
        }
        public void actionPerformed (ActionEvent evt) {
            debug.println ("begin action");
            JTextComponent target = TextUtil.getTextComponent(evt);
            if (target == null || target.isEditable ()) {
                super.actionPerformed (evt);
                return;
            }

            debug.println ("keeping selection");
            target.scrollRectToVisible (new Rectangle (0, 0, 1, 1));
        }
    }

    public static class EndAction extends ExtendedAction {
        public EndAction (Action a) {
            super (a);
        }
        public void actionPerformed (ActionEvent evt) {
            debug.println ("end action");
            JTextComponent target = TextUtil.getTextComponent(evt);
            if (target == null || target.isEditable ()) {
                super.actionPerformed (evt);
                return;
            }

            debug.println ("keeping selection");
            Dimension d = target.getSize ();
            target.scrollRectToVisible (new Rectangle (0, d.height - 1, 
                                                       1, 1));
        }
    }

    public static class SelectAllAction extends ExtendedAction {
        public SelectAllAction (Action a) {
            super (a);
        }
        public void actionPerformed (ActionEvent evt) {
            JTextComponent target = TextUtil.getTextComponent(evt);
            if (target == null
                || !(target instanceof BrowserPane)) {
                super.actionPerformed (evt);
                return;
            }
                
            BrowserPane bp = (BrowserPane) target;
            bp.browser.find (new lapis.tc.TC (new lapis.tc.TC.Id (lapis.parsers.SystemParser.ALL)));
            bp.browser.setPattern (PatternLibrary.getDocumentLibrary ((lapis.Document)bp.getDocument())
                                   .getShortestName (lapis.parsers.SystemParser.ALL),
                                   false);
        }
    }

    public static class DefaultKeyTypedAction extends ExtendedAction {
        public DefaultKeyTypedAction (Action a) {
            super (a);
        }
        public void actionPerformed(ActionEvent e) {
            // Pass keystroke on only if it lacks Control & Alt modifiers
            int mod = e.getModifiers ();
            if ((mod & (ActionEvent.ALT_MASK | ActionEvent.CTRL_MASK)) == 0)
                super.actionPerformed (e);
        }
    }

    public static class DeleteCharAction extends ExtendedAction {
        boolean previous;  
        // true for delete-previous-char, false for delete-next-char
        public DeleteCharAction (Action a, boolean previous) {
            super (a);
            this.previous = previous;
        }
        public void actionPerformed (ActionEvent evt) {
            JTextComponent target = TextUtil.getTextComponent(evt);
            if (!(target instanceof BrowserPane)) {
                super.actionPerformed (evt);
                return;
            }
                
            BrowserPane bp = (BrowserPane) target;
            NamedRegionSet selection = bp.browser.getSelection ();
            if (target == null
                || !target.isEditable ()
                || !target.isEnabled ()
                || selection.isEmpty ()) {
                Toolkit.getDefaultToolkit ().beep ();
                return;
            }

            //RegionSet s = selection.getRegions ();
            if (selection.isAllPoints ()) {
                // selection is just a set of caret positions.
                // change selection to either char-before or char-after
                // each cursor position (depending on whether this action
                // is Backspace or Delete).
                RegionEnumeration re = selection.regions ();
                MutableRegionSet newSelection = new MutableRegionSet ();

                if (previous)
                    for (Region r = re.firstFast (); r != null; r = re.nextFast ()) {
                        int p = r.getStartLow ();
                        newSelection.insert 
                            (new Region ((p > 0) ? p-1 : p, p, r.getVersion()));
                    }
                else {
                    int len = ((SwingDocument)bp.getDocument ())
                        .toRegion ().getLength ();
                    for (Region r = re.firstFast (); r != null; r = re.nextFast ()) {
                        int p = r.getStartLow ();
                        newSelection.insert 
                            (new Region (p, (p < len) ? p+1 : p, r.getVersion()));
                    }
                }
                
				selection = new NamedRegionSet (newSelection);
            }

            if (selection.isAllPoints ()) {
                // cursors are all at start or end of document
                bp.getToolkit ().beep ();
                return;
            }

            // delete the selection
            bp.replace (selection, null, null, false);
        }
    }

    public static class ActionMap {
        /** Map[String->Action] */
        private Map map;

        /** default actions + new actions */
        private Action[] actions;

        public ActionMap (Action[] defaultActions) {
            map = new HashMap ();
            List actionList  = new LinkedList ();

            for (int i = 0; i < defaultActions.length; ++i) {
                Action a = defaultActions[i];
                String name = a.getValue (Action.NAME).toString ();

                if (name.equals (DefaultEditorKit.selectAllAction))
                    a = new EditAction.SelectAllAction (a);
                else if (name.equals (DefaultEditorKit.deletePrevCharAction))
                    a = new EditAction.DeleteCharAction (a, true);
                else if (name.equals (DefaultEditorKit.deleteNextCharAction))
                    a = new EditAction.DeleteCharAction (a, false);
                else if (name.equals (DefaultEditorKit.defaultKeyTypedAction))
                    a = new EditAction.DefaultKeyTypedAction (a);

                if (a != defaultActions[i])
                    map.put (name, a);

                debug.println (name + ": " + a);
                actionList.add (a);
            }
        
            actions = new Action[actionList.size ()];
            actionList.toArray (actions);
        }

        public Action[] getActions () {
            return actions;
        }

        public void fixKeymap (Keymap keymap) {
            debug.println ("fixing keymap");

            while (keymap != null) {
                KeyStroke[] keystrokes = keymap.getBoundKeyStrokes ();
                for (int i = 0; i < keystrokes.length; ++i) {
                    KeyStroke k = keystrokes[i];

                    Action a = keymap.getAction (k);
                    String name = (String) a.getValue (Action.NAME);
                    debug.println (name + " on " + k);
                    Action b = (Action) map.get (name);

                    if (b != null && b != a) {
                        debug.println ("rebinding to " + b);
                        keymap.addActionForKeyStroke (k, b);
                    }
                }
                keymap = keymap.getResolveParent ();
            }
        }



    }

}

