/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapis.swing;

import lapis.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import java.net.*;
import java.beans.*;
import javax.swing.*;
import lapisx.swing.*;
import lapisx.net.URLUtil;
import lapisx.net.FormURL;
import lapisx.util.Str;
import lapisx.util.Mem;
import lapisx.swing.Win;
import lapisx.io.FileUtil;

public class FeedbackForm extends OkCancelDialog {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    /**
     * Limit on uploaded page content.  Longer pages
     * are truncated to this limit.
     */
    public static final int CONTENT_LIMIT = 50000;

    // Relevant context information grabbed from browser
    Throwable exception;      // exception that caused feedback, if any
    String command;              // URL of current page
    String content;           // content of current page
    String contentType;        // e.g. text/plain, text/html
    
    String patternHistory;    // history of patterns (terminated by newlines)

    Browser browser;          // associated browser

    // Controls

    JEditorPane form;
    JScrollPane scroller;
    
    public FeedbackForm (Browser browser, Throwable exception) {
        super (browser, "Send Feedback");

        // Grab context information now, because user may keep working
        // while waiting for feedback form to appear.
        this.browser = browser;
        this.exception = exception;

        if (browser != null) {
            SwingDocument doc = browser.getDocument ();
            if (doc != null) {
                this.command = browser.getCommand ();
                this.content = doc.getText ();
                if (this.content.length () > CONTENT_LIMIT)
                    this.content = this.content.substring (0, CONTENT_LIMIT);
                this.contentType = browser.getContentType ();
            }
            
            patternHistory = browser.getPattern ();
        }

        // construct dialog
        URL url = SwingConfig.makeURLFromProperty ("browser.feedback");

        try {
            form = new JEditorPane (url);
        } catch (IOException e) {
            form = new JEditorPane ("text/html", 
                                    "LAPIS is misconfigured.  Can't send feedback.");
        }

        form.setOpaque (false);
        form.setEditable (false);

        JScrollPane scroller = new JScrollPane (form);
        scroller.setPreferredSize (new Dimension (600, 600));

        Layout.border (getControlsPane ())
            .add (scroller);

        getOKButton ().setText ("Send Feedback");

        pack ();
    }

    public void ok () {
        // extract user's feedback from the HTML form
        HTMLFormAccess access = new HTMLFormAccess (form);
        final String comment = access.getStringValue ("Comment");
        final boolean includeSystemInfo = access.getBooleanValue ("IncludeSystemInfo");
        final boolean includeContent = access.getBooleanValue ("IncludeContent");
        final String name = access.getStringValue ("Name");
        final String email = access.getStringValue ("Email");

        // assemble feedback report into a vector of name/value pairs
        Vector report = new Vector ();
        report.add ("Comment"); report.add (comment);
        report.add ("Name"); report.add (name);
        report.add ("Email"); report.add (email);

        if (includeSystemInfo) {
            addSystemProperty (report, "java.version");
            addSystemProperty (report, "java.vendor");
            addSystemProperty (report, "os.name");
            addSystemProperty (report, "os.version");
            addSystemProperty (report, "os.arch");
            addConfigProperty (report, "lapis.version");
            report.add ("Memory"); report.add ((Mem.used ()/1000) + "KB");

            if (exception != null) {
                StringWriter stringWriter = new StringWriter ();
                PrintWriter printWriter = new PrintWriter (stringWriter);
                FeedbackForm.this.exception.printStackTrace (printWriter);
                printWriter.close ();

                report.add ("StackTrace"); report.add (stringWriter.toString ());
            }
        }

        if (includeContent) {
            report.add ("URL"); report.add (command);
            if (contentType != null) {
                report.add ("ContentType"); 
                report.add (contentType);
            }
            report.add ("PageContent"); report.add (content);
        }

        // make submission URL
        URL url = SwingConfig.makeURLFromProperty ("browser.feedback-submit");

        try {
            // if url is file:, append report to the file
            File file = URLUtil.URLToFile (url);
            
            try {
                fileReport (file, report);
            } catch (IOException e) {
                debug.report (e);
                showExceptionDialog (browser, e);
                return;
            }
        } catch (MalformedURLException e) {
            // otherwise post report to net
            try {
                postReport (url, report);
            } catch (IOException e2) {
                debug.report (e2);

                // couldn't contact server; put up a page asking user 
                // to email the comment instead
                URL errorURL = SwingConfig.makeURLFromProperty ("browser.feedback-failed");

                try {
                    form.setPage (errorURL);
                } catch (IOException e3) {
                    debug.report (e3);
                    showExceptionDialog (browser, e3);
                    return;
                }

                form.addPropertyChangeListener (new PropertyChangeListener () {
                        public void propertyChange (PropertyChangeEvent evt) {
                            if ("page".equals (evt.getPropertyName ()))
                                new HTMLFormAccess (form)
                                    .setStringValue ("Comment", comment);
                        }
                    });

                getOKButton ().setEnabled (false);
                return;
            }
        }

        // thank the user
        JOptionPane.showMessageDialog 
            (browser, 
             "Your feedback has been received successfully. Thank you!",
             "Thank You",
             JOptionPane.PLAIN_MESSAGE);
        
        super.ok ();
    }

    void addSystemProperty (Vector report, String propertyName) {
        report.add (propertyName);
        report.add (System.getProperty (propertyName));
    }

    void addConfigProperty (Vector report, String propertyName) {
        report.add (propertyName);
        report.add (Main.getConfig ().getProperty (propertyName));
    }

    void fileReport (File file, Vector report) throws IOException {
        PrintWriter out = new PrintWriter (new FileWriter (file.toString (), 
                                                           true));
        out.println ("----------------------------------------------------");
        for (int i = 0; i < report.size (); i+=2) { 
            out.print (report.get (i));
            out.print ("=");
            out.println (report.get (i+1));
        }
        out.close ();
    }

    void postReport (URL url, Vector report) throws IOException {
        FormURL furl = new FormURL ("POST", url, report);
        
        // make the connection to upload the report
        URLConnection conn = furl.openConnection ();
        InputStream in = conn.getInputStream ();
        String result = FileUtil.load (in);
        debug.println (result);
        in.close ();
    }


    /**
     * Open a new window containing the feedback form.  Returns immediately
     * (without waiting for user to submit form).
     */
    public static void showFeedbackForm () {
        showFeedbackForm (null);
    }

    /**
     * Open a new window containing the feedback form.  Returns immediately
     * (without waiting for user to submit form).
     */
    public static void showFeedbackForm (Browser browser) {
        showFeedbackForm (browser, null);
    }

    /**
     * Open a new window containing the feedback form.  Returns immediately
     * (without waiting for user to submit form).
     */
    public static void showFeedbackForm (Browser browser, Throwable exception) {
        FeedbackForm f = new FeedbackForm (browser, exception);
        Win.center (f, null);
        f.show ();
    }

    static final String[] FEEDBACK_OPTIONS = { "OK", "Send Feedback" };
    static final String[] NO_FEEDBACK_OPTIONS = { "OK" };

    /**
     * Pop up an error message dialog.  Basically like 
     * JOptionPane.showMessageDialog (),
     * but adds a "Send Feedback" button which triggers a feedback form.
     */
    public static void showMessageDialog (Browser browser, 
                                          Object message,
                                          String title, int messageType) {
        switch (showOptionDialog(browser, message, 
                                 title, JOptionPane.DEFAULT_OPTION, 
                                 messageType, null, 
                                 FEEDBACK_OPTIONS, FEEDBACK_OPTIONS[0])) {
        case JOptionPane.CLOSED_OPTION:
        case 0:     // OK
            break;
        case 1:     // Send Feedback
            showFeedbackForm (browser);
            break;
        }
    }

    /**
     * Pop up an error message dialog with a "don't-show-this-again" option.
     */
    public static void showBlockableMessageDialog (Browser browser, 
                                          Object message,
                                          String title, 
                                          int messageType,
                                          String enabledProperty) {
        if (!Boolean.valueOf (Main.getConfig ().getProperty (enabledProperty))
            .booleanValue ())
            return; // user blocked this error

        switch (showOptionDialog(browser, message, 
                                 title, JOptionPane.DEFAULT_OPTION, 
                                 messageType, null, 
                                 FEEDBACK_OPTIONS, FEEDBACK_OPTIONS[0])) {
        case JOptionPane.CLOSED_OPTION:
        case 0:     // OK
            break;
        case 1:     // Send Feedback
            showFeedbackForm (browser);
            break;
        }
    }

    /**
     * Pop up an error message dialog describing an exception, including
     * a "Send Feedback" button which triggers a feedback form.
     */
    public static void showExceptionDialog (Browser browser, 
                                            Throwable exception) {
        showExceptionDialog (browser, exception, true);
    }

    public static void showExceptionDialog (Browser browser, 
                                            Throwable exception,
                                            boolean offerFeedbackOption) {
        String s = exception.toString ();

        // split s into classname: message parts
        String name = Str.before (s, ":").trim ();
        String message = Str.after (s, ":").trim ();

        // strip off package name, if any
        int p = name.lastIndexOf ('.');
        if (p != -1)
            name = name.substring (p+1);

        // strip off outer class name, if any
        p = name.lastIndexOf ('$');
        if (p != -1)
            name = name.substring (p+1);
        
        // if there's no message, make it the same as the class name
        if (message.length () == 0)
            message = name;

        switch (showOptionDialog
                (browser, 
                 message, 
                 name,
                 JOptionPane.DEFAULT_OPTION, 
                 JOptionPane.ERROR_MESSAGE, 
                 null, 
                 offerFeedbackOption ? FEEDBACK_OPTIONS : NO_FEEDBACK_OPTIONS, 
                 FEEDBACK_OPTIONS[0])) {
        case JOptionPane.CLOSED_OPTION:
        case 0:     // OK
            break;
        case 1:     // Send Feedback
            showFeedbackForm (browser, exception);
            break;
        }
    }

    static int showOptionDialog(Component parentComponent,
                                Object message,
                                String title,
                                int optionType,
                                int messageType,
                                Icon icon,
                                Object[] options,
                                Object initialValue) {
        JOptionPane pane = 
            new JOptionPane (message, messageType,
                             optionType, icon,
                             options, initialValue);
        pane.setInitialValue(initialValue);

        JDialog dialog = pane.createDialog(parentComponent, title);
        pane.selectInitialValue();

        if (SwingConfig.inDemoMode ())
            new DialogTimeout (dialog, 60).start ();
        dialog.show();

        Object selectedValue = pane.getValue();

        if (selectedValue == null)
            return JOptionPane.CLOSED_OPTION;

        if (options == null) {
            if (selectedValue instanceof Integer)
                return ((Integer)selectedValue).intValue();
            return JOptionPane.CLOSED_OPTION;
        }

        for (int counter = 0, maxCounter = options.length;
             counter < maxCounter; 
             counter++) {
            if (options[counter].equals(selectedValue))
                return counter;
        }

        return JOptionPane.CLOSED_OPTION;
    }

    static class DialogTimeout extends Thread {
        Dialog dialog;
        int seconds;

        public DialogTimeout (Dialog dialog, int seconds) {
            this.dialog = dialog;
            this.seconds = seconds;
            setPriority (MIN_PRIORITY);
        }

        public void run () {
            try {
                Thread.sleep (seconds * 1000); 
            } catch (InterruptedException e) {
            }
            if (dialog.isShowing ())
                dialog.dispose ();
        }
    }
}
