/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import lapis.tools.*;

import java.awt.event.*;
import javax.swing.*;
import lapisx.swing.*;
import java.io.*;
import java.net.*;
import java.util.*;
import lapisx.util.*;
import lapisx.net.*;

public abstract class FileAction {

    /**
     * Action that creates a blank text document.
     */
    public static class New extends AcceleratedAction {
        Browser browser;

        public New (Browser browser) {
            super ("&New File", 
                   SwingConfig.makeIcon ("new.gif"),
                   'N', KeyEvent.CTRL_MASK,
                   "Make a new document");
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            browser.setDocument (browser.getFactory ().make ("", "text/plain"));
            browser.getActivePane ().requestFocus ();
        }
    }


    /**
     * Action that pops up a file-open dialog to open a file into the
     * browser.
     */
    public static class Open extends AcceleratedAction {
        Browser browser;

        public Open (Browser browser) {
            super ("&Open File...", 
                   SwingConfig.makeIcon ("open.gif"),
                   'O', KeyEvent.CTRL_MASK,
                   "Open file");
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            JFileChooser chooser = makeFileChooser (browser, null);
            
            if (chooser.showOpenDialog (browser) != JFileChooser.APPROVE_OPTION)
                return;

            String filename = 
                URLUtil.FileToURL (chooser.getSelectedFile ()).toString ();
            browser.eval (filename);
        }
    }



    /**
     * Action invoked by File/Save command.  Saves the contents of the browser.
     */
    public static class Save extends AcceleratedAction {
        Browser browser;

        public Save (Browser browser) {
            super ("&Save",
                   SwingConfig.makeIcon ("save.gif"),
                   'S', KeyEvent.CTRL_MASK,
                   "Save file");
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            save (browser, browser.getDocument ());
        }
    }
    
    /**
     * Action invoked by File/Save As command.  Pops up a save-as dialog.
     */
    public static class SaveAs extends AcceleratedAction {
        Browser browser;

        public SaveAs (Browser browser) {
            super ("Save &As...");
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            saveAs (browser, browser.getDocument ());
        }
    }


    /**
     * Make a JFileChooser for a browser.
     * @param browser Browser to use for current directory
     * @param filename Default selected file, or null if no default
     * should be selected
     * @return If <em>filename</em> is non-null, returns a
     * JFileChooser that starts with that filename selected.  Otherwise,
     * returns a JFileChooser that starts showing the current directory
     * of <em>browser</em>.
     */
    public static JFileChooser makeFileChooser (Browser browser, File filename) {
        JFileChooser chooser = new JFileChooser();

        if (filename != null)
            chooser.setSelectedFile (filename);
        else {
            File dir = (File) browser.getInterpreter ()
                .getVar (Interpreter.CURRENT_DIRECTORY_VAR);
            if (dir == null)
                dir = new File (System.getProperty ("user.dir", "."));
            chooser.setCurrentDirectory (dir);
        }

        return chooser;
    }

    
    /**
     * Save a document, popping up a save-as dialog if the document doesn't
     * have a local filename yet.
     * @param browser Browser to use for feedback
     * @param doc Document to save
     * @effects If <em>doc</em> was loaded from disk or has been previously
     * saved, then saves its content of <em>doc</em> to that filename.
     * If not, then uses {@link saveAs} to pop up a save-as dialog box
     * asking the user for a name.  If an error occurs during saving, 
     * pops up an error message dialog.
     * @returns true if file was successfully saved (no cancellation
     * or error).
     */
    /**
     */
    public static boolean save (Browser browser, Document doc) {
        File filename = getFile (doc);

        return (filename != null && !isReadOnly (doc))
            ? save (browser, filename, doc)
            : saveAs (browser, doc);
    }

    /**
     * Pop up a save-as dialog box to save a document.
     * @param browser Browser to use for feedback
     * @param doc Document to save
     * @effects If user chooses a filename from the dialog box,
     * then the content of <em>doc</em> is saved to that filename.
     * If an error occurs, pops up an error message dialog.
     * @returns true if file was successfully saved (no cancellation
     * or error).
     */
    public static boolean saveAs (Browser browser, Document doc) {
        JFileChooser chooser = makeFileChooser (browser, getFile (doc));

        if (chooser.showSaveDialog (browser) != JFileChooser.APPROVE_OPTION)
            return false;

        File f = chooser.getSelectedFile ();
        if (f.exists ()) 
            switch (JOptionPane.showConfirmDialog 
                    (browser, 
                     f + " already exists. Overwrite?\n",
                     "Overwrite?", 
                     JOptionPane.YES_NO_CANCEL_OPTION)) {
            case JOptionPane.YES_OPTION:
                break;
            case JOptionPane.NO_OPTION:
                // Pop up dialog box again
                return saveAs (browser, doc);
            case JOptionPane.CANCEL_OPTION:
                return false;
            }
        
        if (!save (browser, f, doc))
            return false;

        URL url = URLUtil.FileToURL (f);
        doc.putProperty (SwingDocument.URLProperty, url);
        doc.putProperty (SwingDocument.CommandProperty, url.toString ());

        if (doc == browser.getDocument ()) {
            browser.setTitle (doc);
            browser.setCommand (url.toString (), true);
        }

        return true;
    }

    /**
     * Save browser contents to a local file.
     * @param browser Browser to use for feedback
     * @param f Filename for saved document
     * @param doc Document to save
     * @effects Saves content of <em>doc</em> as filename <em>f</em>.
     * If an error occurs, pops up an error message dialog.
     */
    public static boolean save (Browser browser, File f, Document doc) {
        try {
            new lapis.tools.Save ()
                .invoke (new Arguments (browser.getInterpreter ())
                         .add ("filename", f.toString ())
                         .add (null, doc));
            browser.setStatus ("Document saved.");
            return true;
        } catch (Exception e) {
            FeedbackForm.showExceptionDialog (browser, e);
            return false;
        }
    }


    /**
     * Get local filename corresponding to a document.
     * @param doc Document
     * @return doc's filename, or null if document was not obtained from
     * local disk.
     */
    public static File getFile (Document doc) {
        URL url = (URL)doc.getProperty (SwingDocument.URLProperty);
        if (url == null)
            return null;

        try {
            return URLUtil.URLToFile (url);
        } catch (MalformedURLException e) {
            return null;
        }
    }


    /**
     * Action that closes a browser window.
     */
    public static class Close extends AcceleratedAction {
        Browser browser;
        public Close (Browser browser) {
            super ("&Close Window");
            this.browser = browser;
        }
        public void actionPerformed (ActionEvent evt) {
            browser.dispose ();
        }
    }

    /**
     * Check dirty flags on a set of documents and asks the user
     * if the dirty ones should be saved.
     * @param browser Browser to use for popping up dialog boxes
     * @param docs Enumeration of Documents that should be checked
     * @effects For each document in <em>docs</em>, if the document
     * has been edited since it was last saved, a dialog box is popped
     * up asking the user if it should be saved.
     */
    public static boolean okToClear (Browser browser, Enumeration docs) {
        while (docs.hasMoreElements ()) {
            Document doc = (Document) docs.nextElement ();

            if (isDirty (doc)) {
                switch (JOptionPane.showOptionDialog 
                        (browser, 
                         doc
                         + " has been changed.\n"
                         +"Do you want to save it?",
                         "Save First?", 
                         JOptionPane.YES_NO_CANCEL_OPTION,
                         JOptionPane.INFORMATION_MESSAGE,
                         null,
                         new String[] { "Yes", "No", "No to All", "Cancel" },
                         "Yes")) {
                case 0: // Yes
                    if (!save (browser, doc))
                        return false;
                    break;

                case 1: // No
                    break;
                    
                case 2: // No to All
                    return true;

                case 3: // Cancel
                    return false;
                }
            }
        }

        return true;
    }
        
    /**
     * Get a document's dirty flag.
     * @param doc Document to test
     * @return true if document has been edited since it was last saved.
     */
    public static boolean isDirty (Document doc) {
        if (SwingConfig.inDemoMode ())
            return false; // don't encourage user to overwrite demo examples 

        if (isReadOnly (doc))
            // don't encourage user to overwrite tutorial pages, which
            // are marked read-only using a META tag
            return false; 

        DocumentVersion savedVersion = 
            (DocumentVersion) doc.getProperty (SwingDocument.SavedVersionProperty);
        if (savedVersion == null)
            savedVersion = DocumentVersion.getOriginal(doc); //DocumentVersion.ORIGINAL;
        
        //return (savedVersion != doc.getVersion ());
        return (!savedVersion.equals(doc.getVersion()));
    }

    /**
     * Test whether a document is marked read-only.
     * @param doc Document to test
     * @return true if document has the "read-only" property.  An HTML
     * page has this property if it had a read-only META tag (&lt;META
     * name="read-only" content="true"&gt;) when it was loaded.
     * Tutorial and help pages are marked this way to discourage the
     * user from overwriting them by accident.
     */
    public static boolean isReadOnly (Document doc) {
        Boolean f = 
            (Boolean) doc.getProperty (SwingDocument.ReadOnlyProperty);
        return (f != null && f.booleanValue ());
    }

    /**
     * Action that exits LAPIS.
     */
    public static class Exit extends AcceleratedAction {
        Browser browser;
        public Exit (Browser browser) {
            super ("E&xit", 'Q', KeyEvent.CTRL_MASK);
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            if (lapis.Main.getWindowCount () == 1
                || JOptionPane.showConfirmDialog (browser, 
                           "Close all windows and exit Lapis?", "Exit Lapis?", 
                           JOptionPane.YES_NO_OPTION)
                == JOptionPane.YES_OPTION) {

                if (okToClear (browser, browser.getPageHistory ().elements ())) {
                    browser.saveWindowPositions ();
		    LibraryLoader.writeToFile();
                    System.exit (0);
                }
            }
        }
    }


    /**
     * Action that creates a new Browser window by cloning the current
     * window.
     */
    public static class NewBrowser extends AcceleratedAction {
        Browser browser;
        public NewBrowser (Browser browser) {
            super ("New &Window", 'N', KeyEvent.CTRL_MASK | KeyEvent.SHIFT_MASK);
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            Browser b = new Browser ();
            b.getInterpreter (); // make sure interpreter has been created
            
            // Copy history to new browser window
            History h = browser.getPageHistory ();
            b.setPageHistory (h);
            Document d = (Document) h.get ();
            if (d != null)
                b.setDocument (d);
            
            // Offset the new browser window down and to the right
            Win.position (b, browser, 1.15, 1.15);
            b.setVisible (true);
        }
    }



}

