/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */
package lapis.swing;

import lapis.*;
import lapis.tc.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import lapisx.swing.*;
import javax.swing.undo.*;
import java.util.*;
/**
 * A Simple Find and Replace Pane meant to emulate MS notepad as close
 * as possible but as a side panel in LAPIS
 *
 * <p>
 *
 * All Find and Replace Panes are equal to themselves for each
 * instance of a browser (same hashCode and equals returns true)
 **/

public class FindReplacePane extends JPanel implements StateEditable {

	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;
	Browser browser;
	private JButton closeButton, findButton, replaceButton, replaceAllButton, undoButton;
	private JTextField findText;
	private JTextField replaceText;
	private JCheckBox wholeWord, matchCase;
	private boolean caseSensitive = false;
	//    private boolean showHTMLWarning = true;
	private boolean wordMatch = false;
	private JLabel replacementsMade;
	private UndoManager undoManager;
	/** 
	 * 
	 * @effects constructs a new blank Find and Replace Dialog with all buttons, except for the 
	 * close button, disabled
	 **/

	public FindReplacePane(Browser browser) {
		undoManager = new UndoManager();
		this.browser = browser;
		Document doc = browser.getDocument();
		NamedRegionSet sel = browser.getSelection();
		Region r;
		RegionEnumeration findEnumerator = sel.regions();
		r = findEnumerator.first();
		String textToFind = "";
		if (r != null) {
			textToFind = doc.getText(r);
			browser.setSelection(new NamedRegionSet(r));
		}

		// Make GUI components
		Layout.vertical(this).setContainerInsets(5, 5, 5, 5).setAlignment(Layout.NORTHWEST).add(
			Layout
				.vertical()
				.add(
					Layout
						.horizontal()
						.setAlignment(Layout.SOUTH)
						.add(new JLabel("Find What:     "), Layout.FIXED_SIZE)
						.add(findText = new JTextField(textToFind), Layout.FIXED_HEIGHT)
						.done())
				.add(
					Layout
						.horizontal()
						.setAlignment(Layout.NORTH)
						.add(new JLabel("Replace With: "), Layout.FIXED_SIZE)
						.add(
							Layout
								.vertical()
								.setAlignment(Layout.WEST)
								.add(replaceText = new JTextField(), Layout.FIXED_HEIGHT)
								.add(
									Layout
										.horizontal()
										.add(matchCase = ActionAdapters.toCheckBox(caseAction))
										.add(wholeWord = ActionAdapters.toCheckBox(wordAction))
										.done())
								.done())
						.done())
				.done())

			//			.add(
		//				Layout
		//					.horizontal()
		//					.setSize(Layout.FIXED_SIZE)
		//					.setContainerInsets(2, 2, 2, 2)
		//					.add(matchCase = ActionAdapters.toCheckBox(caseAction))
		//					.add(wholeWord = ActionAdapters.toCheckBox(wordAction))
		//					.done(),
		//				Layout.FIXED_HEIGHT)
	.add(
		Layout
			.gridbag()
			.setSize(Layout.FIXED_SIZE)
			.setAlignment(Layout.SOUTH)
			.add(findButton = ActionAdapters.toButton(findAction))
			.add(replaceButton = ActionAdapters.toButton(replaceAction))
			.add(replaceAllButton = ActionAdapters.toButton(replaceAllAction))
			.add(undoButton = ActionAdapters.toButton(undoAction))
			.add(closeButton = ActionAdapters.toButton(closeAction))
			.done(),
		Layout.FIXED_HEIGHT)
	.add(replacementsMade = new JLabel(""))
	.addGlue();

		setupListeners();

		if (findText.getText().equals("")) {
			findButton.setEnabled(false);
			replaceButton.setEnabled(false);
			replaceAllButton.setEnabled(false);
		}
		if (!undoManager.canUndo())
			undoButton.setEnabled(false);
	}

	private void setupListeners() {
		findText.getDocument().addDocumentListener(new FindTextListener(findButton));
		findText.getDocument().addDocumentListener(new FindTextListener(replaceAllButton));

		findText.getInputMap().put(KeyStroke.getKeyStroke("ENTER"), "Find Next");
		findText.getActionMap().put("Find Next", findAction);
		replaceText.getInputMap().put(KeyStroke.getKeyStroke("ENTER"), "Replace");
		replaceText.getActionMap().put("Replace", replaceAction);

		findButton.getInputMap().put(KeyStroke.getKeyStroke("ENTER"), "Find Next");
		findButton.getActionMap().put("Find Next", findAction);

		replaceButton.getInputMap().put(KeyStroke.getKeyStroke("ENTER"), "Replace");
		replaceButton.getActionMap().put("Replace", replaceAction);

		replaceAllButton.getInputMap().put(KeyStroke.getKeyStroke("ENTER"), "ReplaceAll");
		replaceAllButton.getActionMap().put("ReplaceAll", replaceAllAction);

		closeButton.getInputMap().put(KeyStroke.getKeyStroke("ENTER"), "Close");
		closeButton.getActionMap().put("Close", closeAction);

		browser.addSelectionListener(new SelectionListener() {
			public void changedSelection(Browser browser, NamedRegionSet sel) {
				if (sel.getSize() == 1) {
					replaceButton.setEnabled(true);
				}
				else
					replaceButton.setEnabled(false);
			}
		});
	}

	/** 
	 * Listener attached to a JTextField
	 * 
	 **/
	private class FindTextListener implements DocumentListener {
		JButton attachedButton;
		public FindTextListener() {}
		public FindTextListener(JButton b) {
			attachedButton = b;
		}

		public void changedUpdate(DocumentEvent e) {
			debug.println("action performed on Find text");
			clearReplacementsMadeText();

		}
		public void insertUpdate(DocumentEvent e) {
			debug.println(findText.getText() + ":");
			attachedButton.setEnabled(true);
			clearReplacementsMadeText();
		}
		public void removeUpdate(DocumentEvent e) {
			clearReplacementsMadeText();
			debug.println(findText.getText() + ".");
			if (e.getDocument().getLength() == 0) {
				attachedButton.setEnabled(false);
			}
		}

	}

	private void clearReplacementsMadeText() {
		replacementsMade.setText("");
	}

	/** 
	 * Invoked when Match Case box is clicked
	 *
	 * @effects flips the case sensitivity
	 **/
	public Action caseAction = new AcceleratedAction("&Match Case") {
		public void actionPerformed(ActionEvent evt) {
			caseSensitive = !caseSensitive;
			debug.println("case sensitive " + caseSensitive);

		}
	};

	public Action wordAction = new AcceleratedAction("&Whole Word Only") {
		public void actionPerformed(ActionEvent evt) {
			wordMatch = !wordMatch;
			debug.println("matchWord " + wordMatch);
		}
	};
	/**
	 * Invoked when the Close button is pressed
	 * 
	 * @effects closes and disposes the Dialog
	 **/
	public Action closeAction = new AcceleratedAction("&Close") {
		public void actionPerformed(ActionEvent evt) {
			browser.clearSelection();
			browser.getActivePane().getFinderHighlighter().setSelection(null);
			browser.removeSidebarTab(FindReplacePane.this);
		}
	};
	/**
	 * Invoked with the Find Next Button is pressed
	 * @effects calls findNextMatch()
	 **/

	public Action findAction = new AcceleratedAction("&Find Next") {
		public void actionPerformed(ActionEvent evt) {
			clearReplacementsMadeText();
			findNextMatch();
		}
	};

	/** 
	 * finds the next instance of the text in findText in the browser 
	 *          pane after the cursor and highlights it
	 * @effects current NamedRegionSet
	 * 
	 **/
	private void findNextMatch() {

		Document doc = browser.getDocument();
		NamedRegionSet sel = browser.getSelection();

		TC pat = new TC(new TC.CaseSensitivity(caseSensitive, new TC.Literal(findText.getText())));
		if (wordMatch) {
			TC.Literal word = new TC.Literal(findText.getText());
			TC.Ending endingWord = new TC.Ending(new TC.Id("word"));
			TC.Starting startingWord = new TC.Starting(new TC.Id("word"));

			pat = new TC(new TC.CaseSensitivity(caseSensitive, new TC.And(endingWord, new TC.And(startingWord, word))));

		}
		RegionSet matches = ((TC)pat).match(doc);

		RegionSet after = browser.getSelection().after();

		if (after.isEmpty()) {
			debug.println("nothing after cursor, or no selection marked");
			after = RegionSet.getUniversal(doc);
		}

		RegionSet matchesAfter = matches.and(after);
		sel = new NamedRegionSet(matchesAfter);

		if (sel.isEmpty()) {
			browser.getToolkit().beep();
			browser.setStatus("No matches were found");
			JOptionPane.showMessageDialog(FindReplacePane.this, "Cannot Find '" + findText.getText() + "'", null, JOptionPane.INFORMATION_MESSAGE);
			browser.setSelection(new NamedRegionSet(RegionSet.EMPTY));
			return;
		}

		Region r;
		RegionEnumeration findEnumerator = sel.regions();
		r = findEnumerator.first();

		debug.println("next match = " + r);

		debug.println("setting selection in findNextMatch()");
		browser.setSelection(new NamedRegionSet(r));
		browser.repaint ();
		debug.println("selection : " + browser.getSelection().getSize());
		browser.getActivePane().scrollRegionToVisible(r);
		debug.println("scrolling in findnext");
		debug.println(doc.getText(r) + " " + r.getStart() + " " + r.getEnd());
		// 	}

	}

	/** 
	 * Invoked when the Replace button is pressed
	 * @effects replaces selected text with text in replaceText
	 * JTextfield and highlights the next match
	 * 
	 **/
	public Action replaceAction = new AcceleratedAction("&Replace") {
		public void actionPerformed(ActionEvent evt) {
			clearReplacementsMadeText();
			Document doc = browser.getDocument();
			if (doc instanceof HTMLDocument) {

				debug.println("HTML doc, semi-editable");

				// 		    if (showHTMLWarning){
				// 			HTMLWarningDialog dialog = new  HTMLWarningDialog (browser);
				// 			dialog.show ();
				// 			debug.println("choice: " + dialog.getChoice());
				// 			int choice = dialog.getChoice();
				// 			showHTMLWarning = dialog.getWarningLevel();
				// 			if(choice == HTMLWarningDialog.YES){ // yes change to txt mode
				// 			    browser.setContentType ("text/plain");
				// 			    debug.println("Changed to Text mode");
				// 			}
				// 			if (choice == HTMLWarningDialog.NO) { // no don't change to text mode, do the replace
				// 			    doReplace();
				// 			}

				// 		    }
				// 		    else 
				doReplace(evt);
			}

			if (doc instanceof PlainDocument) {
				debug.println("Plain doc, editable");
				doReplace(evt);

			}
		}
	};

	private void doReplace(ActionEvent evt) {
		StateEdit newEdit = new StateEdit(FindReplacePane.this);
		Document doc = browser.getDocument();
		NamedRegionSet sel = browser.getSelection();
		debug.println("selection : " + sel.getSize());

		// check first that something is selected for replacement,
		// else find the next match after cursor

		if ((sel.isEmpty()) || (sel.regions().first().getLength() == 0)) {
			debug.println("empty selection, finding next");
			findNextMatch();
			sel = browser.getSelection();
			debug.println("empty selection, return");
			return;
		}

		Region r;
		RegionEnumeration findEnumerator = sel.regions();
		r = findEnumerator.first();
		debug.println("doc.getText: " + doc.getText(r) + " findText.getText " + findText.getText());

		// Check that what is currently highlighted matches contents
		// of Find Box else, go to next match

		if (caseSensitive) {
			if (!doc.getText(r).equals(findText.getText())) {
				findNextMatch();
				sel = browser.getSelection();
				debug.println("text didn't match, return");
				return;
			}
		}
		else {
			if (!doc.getText(r).equalsIgnoreCase(findText.getText())) {
				findNextMatch();
				sel = browser.getSelection();
				debug.println("text didn't match, return");
				return;
			}
		}

		debug.println("before replace " + doc.getText(r) + " " + r.getStart() + " " + r.getEnd() + " " + r.getLength());

		// do replacement 
		if (caseSensitive || !replaceText.getText().equals(replaceText.getText().toLowerCase()))
			browser.getActivePane().replaceSelection(replaceText.getText());
		else {
			PreserveCase transform = new PreserveCase(browser, replaceText.getText());
			transform.actionPerformed(evt);
		}
		int newCursorPosition = r.getEnd() + replaceText.getText().length() - 1;
		r = new Region(newCursorPosition, newCursorPosition, doc);
		browser.setSelection(new NamedRegionSet(r));
		debug.println("selection after: " + browser.getSelection().getSize());

		debug.println("after replace " + doc.getText(r) + " " + r.getStart() + " " + r.getEnd() + " " + r.getLength());

		findNextMatch();
		newEdit.end();
		undoManager.addEdit(newEdit);
		if (undoManager.canUndo())
			undoButton.setEnabled(true);
	}

	/** 
	 * Invoked when the replace all button is pressed
	 * @effects replaces all examples of selected text with text in
	 * replaceText JTextfield
	 **/
	public Action replaceAllAction = new AcceleratedAction("&Replace All") {
		public void actionPerformed(ActionEvent evt) {
			clearReplacementsMadeText();
			// only plain documents should be editable.. 
			Document doc = browser.getDocument();
			Cursor bCurrent = browser.getCursor();
			Cursor rCurrent = replaceText.getCursor();
			browser.setCursor(new Cursor(Cursor.WAIT_CURSOR));
			replaceText.setCursor(new Cursor(Cursor.WAIT_CURSOR));

			if (doc instanceof HTMLDocument) {
				debug.println("HTML doc, semi editable");
				// 		    if (showHTMLWarning){
				// 			HTMLWarningDialog dialog = new  HTMLWarningDialog (browser);
				// 			dialog.show ();
				// 			debug.println("choice: " + dialog.getChoice());
				// 			int choice = dialog.getChoice();
				// 			showHTMLWarning = dialog.getWarningLevel();			
				// 			if(choice == HTMLWarningDialog.YES){ // yes change to txt mode
				// 			    browser.setContentType ("text/plain");
				// 			    debug.println("Changed to Text mode");
				// 			}
				// 			if (choice == HTMLWarningDialog.NO) { // no don't change to text mode, do the replace
				// 			    doReplaceAll();
				// 			}
				// 		    }
				// 		    else 

				doReplaceAll(evt);
			}
			if (doc instanceof PlainDocument) {
				debug.println("Plain doc, editable");
				doReplaceAll(evt);
			}
			browser.setCursor(bCurrent);
			replaceText.setCursor(rCurrent);
		}
	};

	private void doReplaceAll(ActionEvent evt) {
		browser.clearSelection();
		StateEdit newEdit = new StateEdit(FindReplacePane.this);
		TC pat = new TC(new TC.CaseSensitivity(caseSensitive, new TC.Literal(findText.getText())));
		if (wordMatch) {
			TC.Literal word = new TC.Literal(findText.getText());
			TC.Ending endingWord = new TC.Ending(new TC.Id("word"));
			TC.Starting startingWord = new TC.Starting(new TC.Id("word"));

			pat = new TC(new TC.CaseSensitivity(caseSensitive, new TC.And(endingWord, new TC.And(startingWord, word))));

		}

		RegionSet matches = ((TC)pat).match(browser.getDocument());

		NamedRegionSet sel = browser.getSelection();

		sel = new NamedRegionSet(matches);
		if (sel.isEmpty()) {}
		else {
			browser.setSelection(sel);
			if (caseSensitive || !replaceText.getText().equals(replaceText.getText().toLowerCase()))
				browser.getActivePane().replaceSelection(replaceText.getText());
			else {
				PreserveCase transform = new PreserveCase(browser, replaceText.getText());
				transform.actionPerformed(evt);
			}
		}

		replacementsMade.setText(sel.getSize() + " replacements made");
		browser.clearSelection();
		newEdit.end();
		undoManager.addEdit(newEdit);
		if (undoManager.canUndo())
			undoButton.setEnabled(true);

	}

	//undoable stuff
	public Action undoAction = new AcceleratedAction("&Undo") {
		public void actionPerformed(ActionEvent evt) {
			browser.undo(new ActionEvent(browser.getActivePane(), evt.getID(), evt.getActionCommand(), evt.getModifiers()));
			try {
				undoManager.undo();
			}
			catch (CannotUndoException cue) {}
			if (!undoManager.canUndo())
				undoButton.setEnabled(false);
		}

	};
	public void restoreState(Hashtable state) {
		debug.println("savedState: " + state.toString());
		if (state.get("selectionHighlighted") != null)
			browser.setSelection((NamedRegionSet)state.get("selectionHighlighted"));
		debug.println("restored selection: " + browser.getSelection().toString());

	}
	public void storeState(Hashtable state) {
		state.put("selectionHighlighted", new NamedRegionSet(browser.getSelection()));
		debug.println("saving selection: " + browser.getSelection().toString());
	}

	/**
	 * All instances of Find and Replace Panes for each browser are the
	 * same
	 **/

	public boolean equals(Object obj) {

		if (obj instanceof FindReplacePane) {
			debug.println("another FindReplacePane");
			if (((FindReplacePane)obj).browser.equals(this.browser))
				return true;
		}
		return false;

	}

	/**
	 * All instances of Find and Replace Panes for each browser are the
	 * same
	 **/

	public int hashCode() {
		return this.browser.hashCode() + 12705783;
	}

	/**
	 * adds A Find and Replace Pane to the sidebar (if one already
	 * exists, then the find and replace pane is cleared of all
	 * input). Also resets the browser splitter to the preferred size
	 * of the largest pane in the sidebar. 
	 * @effects browser splitter
	 **/

	public static class FindReplaceAction extends AcceleratedAction {
		Browser browser;

		public FindReplaceAction(Browser browser) {
			super("&Find & Replace", 'F', KeyEvent.CTRL_MASK);
			this.browser = browser;
		}
		public void actionPerformed(ActionEvent evt) {
			browser.addSidebarTab("Find and Replace", new FindReplacePane(browser));

		}
	}

	/**
	 * simple dialog box to warn users of the risks of editing HTML
	 * documents.
	 * 
	 * Will only show itself once per Find and Replace session if
	 * users ask not to see it again. Should probably be moved up and
	 * more global?
	 **/

	public class HTMLWarningDialog extends JDialog {
		//	FindReplacePane frp;
		private JCheckBox showWarning;
		private String warning1 = "Editing HTML document may cause some formatting to be lost";
		private String warning2 = "Switch to Text mode?";
		private JButton yesButton, noButton, cancelButton;
		private boolean show = true;
		Browser browser;
		private int choice;
		public final static int YES = 1;
		public final static int NO = 2;
		public final static int CANCEL = 3;
		/**
		 * constructs a new HTML warning dialog
		 **/
		public HTMLWarningDialog(Browser browser) {
			super(browser, "Warning", true);
			this.browser = browser;

			// GUI 
			Layout.vertical(this.getContentPane()).setContainerInsets(5, 5, 5, 5).setAlignment(Layout.NORTHWEST).add(
				Layout
					.gridbag()
					.add(new JLabel(warning1))
					.nextRow()
					.add(new JLabel(warning2))
					.nextRow()
					.add(
						Layout
							.horizontal()
							.add(yesButton = ActionAdapters.toButton(yesAction))
							.add(noButton = ActionAdapters.toButton(noAction))
							.add(cancelButton = ActionAdapters.toButton(cancelAction))
							.done(),
						Layout.FIXED_SIZE)
					.nextRow()
					.add(showWarning = ActionAdapters.toCheckBox(showAction))
					.done(),
				Layout.FIXED_SIZE);
			// How to handle window-close
			setDefaultCloseOperation(DISPOSE_ON_CLOSE);

			// Set default window size and position
			pack();
		}

		/** 
		 * @effects closes and disposes the Dialog
		 **/
		private Action cancelAction = new AcceleratedAction("&Cancel") {
			public void actionPerformed(ActionEvent evt) {
				choice = CANCEL;
				dispose();
			}
		};
		/** 
		 * @effects sets choice to YES
		 **/
		private Action yesAction = new AcceleratedAction("&Yes") {
			public void actionPerformed(ActionEvent evt) {
				choice = YES;
				dispose();
			}
		};
		/** 
		 * @effects sets choice to NO
		 **/
		private Action noAction = new AcceleratedAction("&No") {
			public void actionPerformed(ActionEvent evt) {
				choice = NO;
				dispose();
			}
		};

		/** 
		 * @effects flips the show again
		 **/
		private Action showAction = new AcceleratedAction("&Don't show this warning again") {
			public void actionPerformed(ActionEvent evt) {
				show = !show;
				debug.println("showWarning " + show);

			}
		};
		/**
		 * returns warning level 
		 * @return true if warning should be shown again, false if
		 * user does not want to see warning again
		 **/
		public boolean getWarningLevel() {
			return show;
		}

		/**
		 * return button clicked
		 **/
		public int getChoice() {
			return choice;
		}

		/**
		 * displays dialog box
		 * @effects centers and displays dialog box
		 **/
		public void show() {
			Win.center(this, FindReplacePane.this);
			super.show();
		}
	}
}
