/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.Enumeration;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import lapisx.net.*;

public class FormView extends javax.swing.text.html.FormView {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    /**
     * Creates a new FormView object.
     *
     * @param elem the element to decorate
     */
    public FormView(Element elem) {
	super(elem);
    }

    public void actionPerformed(ActionEvent evt) {
        debug.println ("FormView.actionPerformed");
        super.actionPerformed (evt);
    }

    protected void submitData(String data) {
        debug.println ("FormView.submitData(" + data + ")");

        Browser browser = getBrowser ();

        if (SwingConfig.inDemoMode ()) {
            // don't submit feedback; just record it in log
            browser.debug.println (data);
            browser.dispose ();
            return;
        }

        try {
            String method = getMethod ();
            String action = getAction ();
            HTMLDocument doc = (HTMLDocument)getElement().getDocument();
            URL base = doc.getBase();
            URL url = (action != null) ? new URL (base, action) : base;

            debug.println ("method = " + method);
            debug.println ("action = " + action);
            debug.println ("base = " + base);
            debug.println ("form url = " + url);
            FormURL formURL = 
                "http".equals (url.getProtocol ())
                ? new FormURL (method, url, data)
                : (FormURL) new LocalFormURL (method, url, data);

            if (browser != null)
                browser.setDocument ((SwingDocument) 
                                     doc.getFactory ()
                                     .make (formURL));
        } catch (IOException e) {
            debug.report (e);
            FeedbackForm.showExceptionDialog (browser, e);
        }
    }

    Browser getBrowser () {
        Container container = getContainer ();
        return
            (container instanceof BrowserPane)
            ? ((BrowserPane) container).browser
            : null;
    }

	/**
	 * Get the value of the action attribute.
	 */
    public String getAction() {
	    AttributeSet formAttr = getFormAttributes();
	    if (formAttr == null) { 
		return null;
	    }
	    String action = (String)formAttr.getAttribute(HTML.Attribute.ACTION);
	    return action;
	}
	
	/**
	 * Get the form's method parameter.
	 */
	String getMethod() {
	    AttributeSet formAttr = getFormAttributes();
	    if (formAttr != null) {
		String method = (String)formAttr.getAttribute(HTML.Attribute.METHOD);
		if (method != null) {
		    return method.toLowerCase();
		}
	    }
	    return null;
	} 

    AttributeSet getFormAttributes() {
        // JDK 1.4
        Element elem = getElement();
        while (elem != null) {
            if (elem.getAttributes().getAttribute
                (StyleConstants.NameAttribute) == HTML.Tag.FORM) {
                debug.println ("jdk1.4: elem = " + elem);
                return elem.getAttributes ();
            }
            elem = elem.getParentElement();
        }

        // JDK 1.2 or 1.3
        elem = getElement ();
        debug.println ("jdk1.3: elem = " + elem);
        AttributeSet attr = elem.getAttributes ();
        debug.println ("attr = " + attr);
	Enumeration names = attr.getAttributeNames();
	while (names.hasMoreElements()) {
	    Object name = names.nextElement();
            debug.println ("attrname = " + name);
	    if (name instanceof HTML.Tag) {
		HTML.Tag tag = (HTML.Tag)name;
		if (tag == HTML.Tag.FORM) {
		    Object o = attr.getAttribute(tag);
		    if (o != null && o instanceof AttributeSet) {
			return (AttributeSet)o;
		    }
		}
	    }
	}

	return null;
    }

    Element getFormElement() {
        return null;
    }

    public void paint(Graphics g, Shape a) {
        JTextComponent c = (JTextComponent)getContainer();
        Highlighter h = c.getHighlighter();
        if (h instanceof LayeredHighlighter) {
            ((LayeredHighlighter)h).paintLayeredHighlights
                (g, getStartOffset(), getEndOffset(), a, c, this);
        }
        
        super.paint (g, a);
    }

    protected Component createComponent() {
        Component c = super.createComponent ();

        // Make sure component is transparent
        c = makeTransparent (c);
        return c;
    }

    static Component makeTransparent (Component c) {
        if (c != null && c instanceof JComponent) {
            ((JComponent)c).setOpaque (false);
            
            // If it's a JScrollPane, make sure
            // its contents are also transparent
            if (c instanceof JScrollPane)
                makeTransparent (((JScrollPane)c).getViewport ());
            else if (c instanceof JViewport)
                makeTransparent (((JViewport)c).getView ());
            else if (c instanceof JList)
                ((JList)c).setCellRenderer 
                    (new TransparentListCellRenderer ());
            else if (c instanceof JComboBox)
                c = new TransparentComboBox (((JComboBox)c).getModel ()); 
        }
        return c;
    }

    static class TransparentListCellRenderer extends DefaultListCellRenderer {
        public Component getListCellRendererComponent(
        JList list,
	Object value,
        int index,
        boolean isSelected,
        boolean cellHasFocus) {
            JComponent c = 
                (JComponent) super.getListCellRendererComponent 
                (list, value, index, isSelected, cellHasFocus);
            c.setOpaque (isSelected);
            return c;
        }
    }

    static class TransparentComboBox extends JComboBox {
        public TransparentComboBox (ComboBoxModel model) {
            super (model);
            setOpaque (false);
            setRenderer (new TransparentListCellRenderer ());
        }

        public Component add (Component comp) {
            return super.add (makeTransparent (comp));
        }
    }
}
