/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import lapis.tc.TC;
import lapis.maps.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import javax.swing.event.*;
import lapisx.util.Str;

public class HTMLDocument extends javax.swing.text.html.HTMLDocument implements SwingDocument {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    // stream from which content should be read.  
    // null if content has already been read.
    Reader in;              

    // loading state
    int state = NOT_LOADED;
    static final int NOT_LOADED = 0;
    static final int LOADING = 1;
    static final int LOADED = 2;

    // raw text of HTML document exactly as retrieved (including tags).
    //String rawText;     

    // versions
    // DocumentVersion version = DocumentVersion.getOriginal(this); //DocumentVersion.ORIGINAL;
    // DocumentVersion cookedVersion = DocumentVersion.getOriginal(this); //DocumentVersion.ORIGINAL;
    // CoordinateMap cookedMap = IdentityMap.IDENTITY;

    // pattern cache
    PatternCache patternCache = null;

    // document factory
    DocumentFactory factory;

	// document views
	RawView rawView;
	CookedView cookedView;

    public HTMLDocument (DocumentFactory factory, Reader in) {
        super (defaultKit.getStyleSheet ());

        this.factory = factory;
        this.in = in;

    	setTokenThreshold(1000);
        putProperty (MIMEProperty, "text/html");
        //putProperty (CookedProperty, new CookedDocument ());
        
        rawView = new RawView(this);
        cookedView = new CookedView(this);
    }

    public HTMLDocument (DocumentFactory factory, lapis.Document doc) {
        this (factory, doc.getReader ());

        BasicDocument.copyMissingProperties (this, doc);
    }


    private static HTMLEditorKit defaultKit = new HTMLEditorKit ();

    public String toString () {
        String cmd = (String) getProperty (CommandProperty);
        if (cmd != null && cmd.length () > 0)
            return cmd;
//             return Str.abbreviate (cmd, 200);

        URL url = (URL) getProperty (URLProperty);
        if (url != null)
            return url.toString ();

        return super.toString ();
    }

    public URL getBase () {
        URL base = (URL) getProperty (BaseProperty);
        if (base != null) {
            super.setBase (base);
            return base;
        } 

        base = super.getBase ();
        if (base != null) {
            putProperty (BaseProperty, base);
            return base;
        }

        base = (URL) getProperty (URLProperty);
        return base;
    }

    public void setBase (URL base) {
        debug.println ("setBase(" + base + ")");
        super.setBase (base);
        putProperty (BaseProperty, base);
    }

    /*
     * Methods implementing SwingDocument interface
     */

    public String getText () {
    /*
        if (!load (defaultKit))
            return "";
        return ((RawView)rawView).rawText;
    */
    	return this.getDefaultView().getText();
    }

	void setText (String rawText) {
		this.rawView.rawText = rawText;
	}

    public String getText (Region r) {
    /*
        String content = getText ();
        int l = content.length ();
        int s = Math.min (l, Math.max (0, r.getStart ()));
        int e = Math.min (l, Math.max (0, r.getEnd ()));
        return (s < e) ? content.substring (s, e) : "";
    */
    	return this.getDefaultView().getText(r);
    }

    public Region toRegion () {
        // return new Region (0, getText ().length (), getVersion());
        return this.getDefaultView().toRegion();
    }

    public Reader getReader () {
        // return new StringReader (getText ());
        return this.getDefaultView().getReader();
    }

    public Dictionary getProperties () {
        return getDocumentProperties ();
    }

    public PatternCache getPatternCache () {
        if (patternCache == null)
            patternCache = new PatternCache ();
        return patternCache;
    }

    public DocumentFactory getFactory () {
        return factory;
    }
/*
    class CookedDocument implements lapis.Document {
        public String getText () {
            if (!load (defaultKit))
                return "";

            try {
                return HTMLDocument.this.getText (0, getLength ());
            } catch (BadLocationException e) {
                debug.report (e);
                return "";
            }
        }

        public String getText (Region r) {
            if (!load (defaultKit))
                return "";

            try {
                int l = getLength ();
                int s = Math.min (l, Math.max (0, r.getStart ()));
                int e = Math.min (l, Math.max (0, r.getEnd ()));

                return HTMLDocument.this.getText (s, e - s);
            } catch (BadLocationException e) {
                debug.report (e);
                return "";
            }
        }

        public Region toRegion () {
            return new Region (0, getLength (), getVersion());
        }

        public Reader getReader () {
            return new StringReader (getText ());
        }

        public Dictionary getProperties () {
            return HTMLDocument.this.getProperties ();
        }

        public Object getProperty (Object key) {
            return HTMLDocument.this.getProperty (key);
        }

        public void putProperty (Object key, Object value) {
            HTMLDocument.this.putProperty (key, value);
        }

        public PatternCache getPatternCache () {
            return HTMLDocument.this.getPatternCache ();
        }

        public DocumentFactory getFactory () {
            return HTMLDocument.this.getFactory ();
        }

        public DocumentVersion getVersion () {
            return cookedVersion;
        }
    }
*/
    void setCookedMap (CoordinateMap map) {
    /*
        cookedMap = map;
        cookedVersion = new DocumentVersion (map, version, this);
    */
    	this.cookedView.setCookedMap(map, this.rawView.getVersion());
    }

    public boolean isLoaded () {
        return state == LOADED;
    }

    public boolean load (EditorKit kit) {
        if (in == null) {
            if (rawView.rawText == null)
                rawView.rawText = "";
            state = LOADED;
            return true;
        }

        synchronized (this) {
            if (state == LOADING)
                return false;
            state = LOADING;
        }

        debug.printStackTrace ();

        try {
            try {
                try {
                    // Fetch the document content, simultaneously
                    // constructing the view
                    kit.read(in, this, 0);
                } catch (BadLocationException e) {
                    throw new IOException (e.toString ());
                }
                in.close ();

                addDocumentListener (new DocumentListener () {
                        public void changedUpdate (DocumentEvent e) {
                            debug.println ("changedUpdate " + e);
                            
                            if (patternCache != null)
                                patternCache.clear ();
                        }
                        public void insertUpdate (DocumentEvent e) {
                            debug.println ("insertUpdate " + e);
                            int cookedStart =  e.getOffset ();
                            int len = e.getLength ();
                            
                            int rawStart = cookedView.cookedMap.invert ().mapStart (cookedStart);
                            try {
                                replaceInRaw (rawStart, 
                                              rawStart, 
                                              getText (cookedStart, len));
                            } catch (BadLocationException ex) {
                                debug.report (ex);
                            }

                            CoordinateMap rawChange = 
                                new EditCoordinateMap.Insertion (rawStart, rawStart, rawStart + len);
                            rawView.setVersion (new DocumentVersion (rawChange, rawView.getVersion(), 
                            							   rawView));
                            cookedView.cookedMap = new SplicedCoordinateMap (cookedView.cookedMap, IdentityMap.IDENTITY,
                                                                  rawStart, rawStart + len,
                                                                  cookedStart, cookedStart + len);
                            cookedView.setVersion (new DocumentVersion (cookedView.cookedMap, rawView.getVersion(),
                            									 cookedView));
                            
                            if (patternCache != null)
                                patternCache.clear ();
                        }
                        public void removeUpdate (DocumentEvent e) {
                            debug.println ("removeUpdate " + e);
                            int cookedStart =  e.getOffset ();
                            int cookedEnd = cookedStart + e.getLength ();
                            
                            int rawStart = cookedView.cookedMap.invert ().mapStart (cookedStart);
                            int rawEnd = cookedView.cookedMap.invert ().mapEnd (cookedEnd);
                            replaceInRaw (rawStart, rawEnd, "");

                            CoordinateMap rawChange = 
                                new EditCoordinateMap.Deletion (rawStart, rawEnd, rawStart);
                            rawView.setVersion (new DocumentVersion (rawChange, rawView.getVersion(),
                            							   rawView));
                            cookedView.cookedMap = new SlicedCoordinateMap (cookedView.cookedMap,
                                                                 rawStart, rawEnd,
                                                                 cookedStart, cookedEnd);
                            cookedView.setVersion (new DocumentVersion (cookedView.cookedMap, rawView.getVersion(),
                            									 cookedView));

                            if (patternCache != null)
                                patternCache.clear ();
                        }
                    });
                    
            } catch (IOException e) {
                debug.report (e);
            }
        } finally {
            synchronized (this) {
                state = LOADED;
                in = null;
            }
        }
            
        // Find <TITLE>, if any
        Region titleMatch = tcTitle.match (this).regions ().firstFast ();
        if (titleMatch != null)
            putProperty (TitleProperty, 
                         Str.compressWhitespace (getText (titleMatch)));
            
        // Find read-only META tag, if any
        Region readOnlyMatch = tcReadOnly.match (this).regions ().firstFast ();
        if (readOnlyMatch != null)
            putProperty (ReadOnlyProperty, Boolean.TRUE);
            
        // Make sure the Base property is set (since our base class
        // tends to set its private base variable instead of politely
        // calling setBase())
        setBase (getBase ());

        return true;
    }

    private static TC tcTitle = TC.make (".HTML.Text in .HTML.Element.[Title]");
    private static TC tcReadOnly = TC.make ("view source 'true' in .HTML.Element.[meta] contains .HTML.Attribute.@name contains 'read-only'");


    /*
     * Methods overriding javax.swing.text.html.HTMLDocument behavior
     */
    public javax.swing.text.html.HTMLEditorKit.ParserCallback getReader(int pos) {
        setBase (getBase ());
        return new MyHTMLReader(pos);
    }
    
    public javax.swing.text.html.HTMLEditorKit.ParserCallback getReader(int pos, int popDepth,
						  int pushDepth,
						  HTML.Tag insertTag) {
        setBase (getBase ());
        return new MyHTMLReader(pos, popDepth, pushDepth, insertTag);
    }

    public class MyHTMLReader extends HTMLReader {
    	public MyHTMLReader(int offset) {
    	    super(offset);
    	}

        public MyHTMLReader(int offset, int popDepth, int pushDepth,
			  HTML.Tag insertTag) {
            super(offset, popDepth, pushDepth, insertTag);
        }

        public HTMLDocument getDocument () {
            return HTMLDocument.this;
        }

        public int getCurrentLength () {
            int len = getLength ();
            if (parseBuffer != null) {
                    // add in lengths of elements waiting in buffer
                    int n = parseBuffer.size ();
                    for (int i = 0; i < n; ++i) {
                            ElementSpec es = (ElementSpec)parseBuffer.elementAt (i);
                            len += es.getLength ();
                    }
            }
            return len;
        }

    }


    public DocumentVersion getVersion () {
        // return version;
        return this.getDefaultView().getVersion();
    }

    void setVersions (DocumentVersion ver, DocumentVersion cookedVer) {
    /*
        this.version = ver;
        this.cookedVersion = cookedVer;
        this.cookedMap = cookedVer.diff (ver);
    */
    	this.rawView.setVersion(ver);
    	this.cookedView.setVersion(cookedVer, cookedVer.diff(ver));
    }

    void replaceInRaw (int start, int end, String content) {
        rawView.rawText = 
            rawView.rawText.substring (0, start) 
            + content 
            + rawView.rawText.substring (end);

    }

	class RawView implements DocumentView {
		lapis.Document doc;
		DocumentVersion version = DocumentVersion.getOriginal(this); 
		
		String rawText = null;
		
		RawView (lapis.Document doc){
			this.doc = doc;
		}
		
		public String getText () {
			if (!load (defaultKit))
				return "";
			return rawText;
		}

		public String getText (Region r) {
			String content = getText ();
			int l = content.length ();
			int s = Math.min (l, Math.max (0, r.getStart ()));
			int e = Math.min (l, Math.max (0, r.getEnd ()));
			return (s < e) ? content.substring (s, e) : "";
		}

		public Region toRegion () {
			return new Region (0, getText ().length (), this.getVersion());
		}
		
		public Reader getReader () {
			return new StringReader (getText ());
		}
		
		public DocumentVersion getVersion () {
			return this.version;
		}
		
		public void setVersion (DocumentVersion ver) {
			this.version = ver;
		}
		
		public lapis.Document getDocument () {
			return this.doc;
		}
	}
	
	class CookedView implements DocumentView {
		lapis.Document doc;
		DocumentVersion version = DocumentVersion.getOriginal(this); 
		CoordinateMap cookedMap = IdentityMap.IDENTITY;
		
		CookedView (lapis.Document doc) {
			this.doc = doc;
		}
		
		public String getText () {
			if (!load (defaultKit))
				return "";

			try {
				return HTMLDocument.this.getText (0, getLength ());
			} catch (BadLocationException e) {
				debug.report (e);
				return "";
			}
		}

		public String getText (Region r) {
			if (!load (defaultKit))
				return "";

			try {
				int l = getLength ();
				int s = Math.min (l, Math.max (0, r.getStart ()));
				int e = Math.min (l, Math.max (0, r.getEnd ()));

				return HTMLDocument.this.getText (s, e - s);
			} catch (BadLocationException e) {
				debug.report (e);
				return "";
			}
		}

		public Region toRegion () {
			return new Region (0, getLength (), this.getVersion());
		}
		
		public Reader getReader () {
			return new StringReader (getText ());
		}

		public DocumentVersion getVersion () {
			return this.version;
		}
		
		public void setVersion (DocumentVersion ver, CoordinateMap map) {
			this.version = ver;
			this.cookedMap = map;		
		}
		
		public void setVersion (DocumentVersion ver) {
			this.version = ver;
		}
		
		public void setCookedMap (CoordinateMap map, DocumentVersion ver) {
			this.cookedMap = map;
			this.version = new DocumentVersion (map, ver, this);
		}
		
		public lapis.Document getDocument () {
			return this.doc;
		}
	}
	
	public DocumentView getDefaultView () {
		return this.rawView;
	}
	
	public DocumentView getCookedView () {
		return this.cookedView; 
	}
	
}
