/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.text.html.*;

public class HTMLEditorKit extends javax.swing.text.html.HTMLEditorKit {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    public HTMLEditorKit () {
        super ();
        setDefaultCursor (Cursor.getPredefinedCursor (Cursor.TEXT_CURSOR));
        setLinkCursor (Cursor.getPredefinedCursor (Cursor.HAND_CURSOR));
    }

    public Object clone() {
    	return new HTMLEditorKit();
    }    

    //
    // Overridden edit actions
    //
    
    private static EditAction.ActionMap map;

    public Action[] getActions () {
        if (map == null)
            map = new EditAction.ActionMap (super.getActions ());
        return map.getActions ();
    }

    //
    // Overridden HTML parser
    //

    javax.swing.text.html.HTMLEditorKit.Parser parser;

    protected javax.swing.text.html.HTMLEditorKit.Parser getParser() {
        if (parser == null) {
            if (Boolean.valueOf (lapis.Main.getConfig ().getProperty ("browser.use-swing-parser"))
                .booleanValue ())
                parser = super.getParser ();
            else
                parser = new lapis.swing.HTMLParser ().getSwingParser ();
        }
        return parser;
    }

    public ViewFactory getViewFactory() {
	return new lapis.swing.HTMLEditorKit.HTMLFactory();
    }

    public static class HTMLFactory extends javax.swing.text.html.HTMLEditorKit.HTMLFactory {
        public View create(Element elem) {
            Object o = elem.getAttributes().getAttribute(StyleConstants.NameAttribute);
            if (o instanceof HTML.Tag) {
                HTML.Tag kind = (HTML.Tag) o;
                if ((kind == HTML.Tag.INPUT) ||
                    (kind == HTML.Tag.SELECT) ||
                    (kind == HTML.Tag.TEXTAREA))
		    return new lapis.swing.FormView(elem);
                if ((kind == HTML.Tag.COMMENT) ||
                         (kind == HTML.Tag.HEAD) ||
                         (kind == HTML.Tag.TITLE) ||
                         (kind == HTML.Tag.META) ||
                         (kind == HTML.Tag.LINK) ||
                         (kind == HTML.Tag.STYLE) ||
                         (kind == HTML.Tag.SCRIPT) ||
                         (kind == HTML.Tag.AREA) ||
                         (kind == HTML.Tag.MAP) ||
                         (kind == HTML.Tag.PARAM) ||
                         (kind == HTML.Tag.APPLET))
                    return new lapis.swing.HiddenView(elem);
		if (kind==HTML.Tag.IMG)
		    return new lapis.swing.ImageView(elem, super.create (elem));
            }
             
            return super.create (elem);
        }
    }

    public javax.swing.text.Document createDefaultDocument() {
        return new lapis.swing.HTMLDocument (SwingDocument.defaultFactory,
                                             (Reader) null);
    }

    public void install (JEditorPane editorPane) {
        editorPane.addMouseListener (linkHandler);
        if (linkHandler instanceof MouseMotionListener)
            editorPane.addMouseMotionListener ((MouseMotionListener) linkHandler);

        if (map == null)
            map = new EditAction.ActionMap (super.getActions ());
        map.fixKeymap (editorPane.getKeymap ());

        // Don't want to call HTMLEditorKit.install(), because it
        // installs a mouseListener that triggers hyperlinks on any mouse
        // click (we only want hyperlinks to be triggered by left mouse clicks).
        // However, someday we may need to call the install() method of HTMLEditorKit's
        // parent class (StyledEditorKit).  StyledEditorKit.install() adds a listener
        // used only for editing, so we don't need it right now.
        //super.install (editorPane);
    }

    public void deinstall (JEditorPane editorPane) {
        // See note in install(), above.
        //super.deinstall (editorPane);
        editorPane.removeMouseListener (linkHandler);
        if (linkHandler instanceof MouseMotionListener)
            editorPane.removeMouseMotionListener ((MouseMotionListener) linkHandler);
    }

    MyLinkController linkHandler = new MyLinkController ();

    /**
     * Class to watch the associated component and fire
     * hyperlink events on it when appropriate.
	 * Follows hyperlinks when the left mouse button is pressed and
	 * released.
     */
    public static class MyLinkController extends LinkController {
        int posPressed;      // position in document where mouse was pressed

        public void mouseMoved(MouseEvent e) {
            BrowserPane bp = (BrowserPane) e.getSource();
            bp.pretendToBeReadOnly = true;
            try {
                super.mouseMoved (e);
            } finally {
                bp.pretendToBeReadOnly = false;
            }
        }

    	/**
         * Called for a mouse click event.
    	 * @param e the mouse event
    	 */
        public void mousePressed (MouseEvent e) {
            if (SwingUtilities.isLeftMouseButton (e)
                && !e.isShiftDown ()
                && !e.isControlDown ()
                && !e.isAltDown ()) {
                JEditorPane editor = (JEditorPane) e.getSource();
                Point pt = new Point(e.getX(), e.getY());
                posPressed = editor.viewToModel(pt);
            } else
                posPressed = -1;
        }

    	/**
         * Called for a mouse release event.
    	 *
    	 * @param e the mouse event
    	 */
        public void mouseReleased (MouseEvent e) {
            if (posPressed >= 0) {
                JEditorPane editor = (JEditorPane) e.getSource();
                Point pt = new Point(e.getX(), e.getY());
                int posReleased = editor.viewToModel(pt);
                if (posPressed == posReleased)
                    activateLink(posPressed, editor);
                posPressed = -1;
            }
        }

    	/**
         * Called for a mouse release event.
    	 *
    	 * @param e the mouse event
    	 */
        public void mouseClicked (MouseEvent e) {
            return;
        }
    }
}
