/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import lapis.maps.*;
import lapis.parsers.HTMLElement;
import lapis.parsers.HTMLTag;
import lapis.parsers.HTMLFakeElement;
import java.io.*;
import java.util.*;
import lapisx.util.Str;

public class HTMLParser extends lapis.parsers.HTMLParser {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    PiecewiseLinearMonotonicMap map;      // coordinate map of doc (or null if not needed)
    lapis.swing.HTMLDocument.MyHTMLReader callback;    // Swing parser callback

    int state;
    static final int START_STATE = 0;
    static final int HTML_STATE = 1;
    static final int HEAD_STATE = 2;
    static final int BODY_STATE = 3;

    Hashtable skippedElements = new Hashtable ();

    /**
     * Make an HTMLParser.
     */
    public HTMLParser () {
        setAlwaysParse (true);
    }

    public javax.swing.text.html.HTMLEditorKit.Parser getSwingParser () {
        return new SwingParser ();
    }

    class SwingParser extends javax.swing.text.html.HTMLEditorKit.Parser {
        /*
         * HTMLEditorKit.Parser interface
         */
        public void parse(Reader r,
                          javax.swing.text.html.HTMLEditorKit.ParserCallback cb,
                          boolean ignoreCharSet) throws IOException {
            callback = (lapis.swing.HTMLDocument.MyHTMLReader)cb;
            Document doc = (Document) callback.getDocument ();
            if (doc instanceof lapis.swing.HTMLDocument) {
                map = new PiecewiseLinearMonotonicMap ();
                map.add (0, 0);
            }
            else
                map = null;

            StringBuffer content = new StringBuffer ();

            HTMLParser.this.parse (doc, r, content);

            try {
                callback.flush ();
            } catch (javax.swing.text.BadLocationException e) {
            }


            if (doc instanceof lapis.swing.HTMLDocument) {
                String contentText = content.toString ();
                ((lapis.swing.HTMLDocument)doc).rawView.rawText = contentText;
                ((lapis.swing.HTMLDocument)doc).setCookedMap (map);

                map.add (contentText.length (), callback.getCurrentLength ());

                if (debug.printEnabled ()) {
                    String raw = contentText;
                    String cooked = "";
                    try {
                        lapis.swing.HTMLDocument sdoc = 
                            (lapis.swing.HTMLDocument) doc;
                        cooked = sdoc.getText (0, sdoc.getLength ());
                    } catch (javax.swing.text.BadLocationException e) {
                        debug.report (e);
                    }

                    map.dump (raw, cooked);
                }
            }
        }
    }

    protected void handleInitParse () {
        debug.println ("--- starting parse ---");
        super.handleInitParse ();

        state = START_STATE;
        skippedElements.clear ();

        //handleStartElement (new HTMLFakeElement (HTMLTag.HTML));
        handleStartElement (new HTMLFakeElement (HTMLTag.HTML, doc));
    }

    protected void handleDoneParse (int pos) {
        super.handleDoneParse (pos);

        skippedElements.clear ();
    }

    protected void handleStartWord (int pos) {
        super.handleStartWord (pos);
        if (map != null)
            map.add (pos, callback.getCurrentLength ());
    }

    protected void handleEndWord (char[] word, int offset, int len, int pos) {
        if (state == HTML_STATE) { 
			//handleStartElement (new HTMLFakeElement (HTMLTag.BODY));
			handleStartElement (new HTMLFakeElement (HTMLTag.BODY, doc));
        }

        super.handleEndWord (word, offset, len, pos);

        char[] data = new char[len];
        System.arraycopy (word, offset, data, 0, len);

        debug.println ("text(" + new String (data) + "@" + pos + ")");
        callback.handleText (data, pos);
        if (map != null)
            map.add (pos, callback.getCurrentLength ());
    }

    static String imageHref = null;

    protected void handleEmptyElement (HTMLElement e) {
        boolean isFake = (e instanceof HTMLFakeElement);

        if (state == HTML_STATE) {
			//handleStartElement (new HTMLFakeElement (HTMLTag.BODY));
			handleStartElement (new HTMLFakeElement (HTMLTag.BODY, doc));
        }

        String tagName = e.getTagName ();

        super.handleEmptyElement (e);

        // Pass element to callback interface
        if (map != null && !isFake)
            map.add (e.getStart (), callback.getCurrentLength ());

        if (!skippedTags.contains (tagName)) {
            // fix "doc:" image links
            if (HTMLTag.IMG.equals (tagName)) {
                String href = e.getHTMLAttribute (HTMLTag.SRC_ATTRIB);
                if (href != null
                    && href.startsWith ("doc:/")) {

                    if (imageHref == null) {
                        imageHref = SwingConfig.makeURL ("images")
                            .toString ();
                    }
                    
                    href = Str.replace (href, 
                                        "doc:/lib/images/ftp",
                                        imageHref);
                    debug.println ("changed img src to " + href);
                    e.changeHTMLAttribute (HTMLTag.SRC_ATTRIB, href);
                }
            }

            debug.println ("simpleTag(" + tagName + "@" + e.getEnd () + ")");
            callback.handleSimpleTag (getSwingTag (e.startTag),
                                      getSwingAttributes (e.startTag),
                                      e.getStart ());
        }  

        if (map != null && !isFake)
            map.add (e.getEnd (), callback.getCurrentLength ());
    }

    protected void handleStartElement (HTMLElement e) {
        boolean isFake = (e instanceof HTMLFakeElement);

        // check that this element has the context that Swing requires
        // (e.g., all TRs require a TABLE, all TDs require a TR)
        String tagName = e.getTagName ();
        String required = (String)requiredSwingContext.get (tagName);
        
        if (required != null && findOpenElement (required) == -1) {
            // Create a fake element to supply missing context.
            // The fake element will be closed automatically (I hope).
            // handleStartElement (new HTMLFakeElement (required));
            handleStartElement (new HTMLFakeElement (required, doc));
        }

        boolean skip = false;

        if  (skippedTags.contains (tagName))
            skip = true;
        else if (tagName.equals (HTMLTag.HTML)) {
            if (state >= HTML_STATE) {
                skip = true;
                skippedElements.put (e, e);
            } else
                state = HTML_STATE;
        } else if (tagName.equals (HTMLTag.HEAD)) {
            if (state >= HEAD_STATE) {
                skip = true;
                skippedElements.put (e, e);
            } else
                state = HEAD_STATE;
        } else if (tagName.equals (HTMLTag.BODY)) {
            if (state >= BODY_STATE) {
                skip = true;
                skippedElements.put (e, e);
            } else
                state = BODY_STATE;
//         } else if (headTags.contains (tagName)) {
//             if (state > HEAD_STATE)
//                 skip = true;
        } else {
            if (state < HEAD_STATE) {
				//handleStartElement (new HTMLFakeElement (HTMLTag.BODY));
				handleStartElement (new HTMLFakeElement (HTMLTag.BODY, doc));
            }
        }

        super.handleStartElement (e);

        if (map != null && !isFake)
            map.add (e.startTag.getStart (), callback.getCurrentLength ());

        if (!skip) {
            debug.println ("startTag(" + tagName + "@" + e.startTag.getEnd () + ")");
            callback.handleStartTag (getSwingTag (e.startTag),
                                     getSwingAttributes (e.startTag),
                                     e.getStart ());
        }

        if (map != null && !isFake)
            map.add (e.startTag.getEnd (), callback.getCurrentLength ());
    }

    protected void handleEndElement (HTMLElement e) {
        boolean isFake = (e instanceof HTMLFakeElement);

        String tagName = e.getTagName ();

        super.handleEndElement (e);

        if (e.endTag != null && map != null && !isFake)
            map.add (e.endTag.getStart (), callback.getCurrentLength ());

        if (!skippedTags.contains (tagName)
            && !skippedElements.contains (e)) {
            debug.println ("endTag(" + tagName + "@" + e.getEnd () + ")");
            callback.handleEndTag (getSwingTag (e.startTag), 
                                   (e.endTag != null) ? e.endTag.getStart () : e.getEnd ());
        }

        if (map != null && !isFake)
            map.add (e.getEnd (), callback.getCurrentLength ());
    }

    private javax.swing.text.MutableAttributeSet attrs =
            new javax.swing.text.SimpleAttributeSet ();

    private javax.swing.text.html.HTML.Tag getSwingTag (HTMLTag lapisTag) {
        String tagName = lapisTag.getTagName ();
        javax.swing.text.html.HTML.Tag swingTag
            = javax.swing.text.html.HTML.getTag (tagName);
        if (swingTag == null) {
            debug.println ("no Swing tag for " + tagName 
                           + ", using UnknownTag");
            swingTag = new javax.swing.text.html.HTML.UnknownTag (tagName);
        }
        return swingTag;
    }

    private javax.swing.text.MutableAttributeSet getSwingAttributes (HTMLTag lapisTag) {
    	attrs.removeAttributes(attrs);
    	if (lapisTag.countHTMLAttributes() > 0) {
    	    String[] htmlAttributes = lapisTag.getHTMLAttributes ();
            for (int i=0; i < htmlAttributes.length; ) {
                String name = htmlAttributes[i++];
                String val = htmlAttributes[i++];
                javax.swing.text.html.HTML.Attribute key =
                    javax.swing.text.html.HTML.getAttributeKey (name);
                if (key != null)
                    attrs.addAttribute (key, val);
            }
        }
        return attrs;
    }

    // maps element names to the context element required by Swing 
    // for exception-free rendering
    private static Hashtable requiredSwingContext = new Hashtable ();
    static {
          requiredSwingContext.put (HTMLTag.TR, HTMLTag.TABLE);
          requiredSwingContext.put (HTMLTag.TD, HTMLTag.TR);
          requiredSwingContext.put (HTMLTag.TH, HTMLTag.TR);
    }

    // element names that should not be passed to Swing
    private static Hashtable skippedTags = new Hashtable ();
    static {
          skippedTags.put (HTMLTag.COMMENT, HTMLTag.COMMENT);
    }
}
