/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import javax.swing.table.*;
import lapisx.swing.*;
import lapisx.util.*;
import lapisx.enum.RestartableEnumeration;
import java.text.*;

public class HypothesisDisplay extends JPanel {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    RestartableEnumeration hypothesisEnum = null;
    Vector hypotheses = new Vector ();
    boolean expanded = false;
    Set listeners = new SafeIteratorSet ();

    JTable table;
    JScrollPane scroller;
    JButton expandButton;

    public HypothesisDisplay () {
        table = new JTable ();
        table.setModel (new HypothesisModel ());
        table.setShowGrid (true);
        table.setSelectionMode (ListSelectionModel.SINGLE_SELECTION);
        table.getTableHeader().setReorderingAllowed (false);
        table.getSelectionModel ()
            .addListSelectionListener (selectionChanged);
        table.setSelectionBackground (BrowserPane.selectionWithFocus);

        TableColumnModel columns = table.getColumnModel  ();

        TableColumn pattern = columns.getColumn (0);
        PatternRenderer patternRenderer = new PatternRenderer ();
        pattern.setCellRenderer (patternRenderer);

        TableColumn matches = columns.getColumn (1);
        matches.setMaxWidth (65);
        matches.setCellRenderer (new NumberRenderer (NumberFormat.getInstance ()));

        // make a scroll pane whose maximum size depends on whether or not
        // we're expanded
        scroller = new JScrollPane (table) {
                public Dimension getMaximumSize () {
                    if (expanded)
                        return super.getMaximumSize ();
                    else
                        return new Dimension 
                            (Short.MAX_VALUE, getPreferredSize ().height);
                }
                
                public Dimension getPreferredSize () {
                    Dimension d = table.getPreferredSize ();
                    d.height = Math.max (d.height, table.getRowHeight () + 18);
                    debug.println ("preferred size = " + d);
                    return d;
                }
            };

        // button that toggles expansion of pattern list
        expandButton = new JButton ("Show More Guesses") {
                public Dimension getMaximumSize () {
                    return getPreferredSize ();
                }
            };
        expandButton.addActionListener (new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    boolean expandNow = !expanded;
                    setExpanded (expandNow);
                    expandButton.setText (expandNow
                                          ? "Show Only Best Guess"
                                          : "Show More Guesses");
                }
            });

        Layout.vertical (this)
            .add (scroller, Layout.EAST)
            .add (expandButton, Layout.EAST);
    }

    public void setExpanded (boolean f) {
        expanded = f;
        ((JComponent) getParent ()).revalidate ();

        if (expanded && hypothesisEnum != null) {
            // make sure we have all the hypotheses in the hypotheses vector
            hypothesisEnum.restart ();
            setHypotheses (hypothesisEnum);
        } else {
            ((HypothesisModel) table.getModel ()).fireTableDataChanged ();
            if (!hypotheses.isEmpty ())
                table.getSelectionModel ().setSelectionInterval (0, 0);
        }
    }

    public boolean isExpanded () {
        return expanded;
    }

    public void setHypotheses (RestartableEnumeration e) {
        hypothesisEnum = e;

        Vector h = new Vector ();
        Font font = new PatternRenderer ().getFont ();
        int fh;
        try {
            fh = 2 + FontUtil.getHeight (font, this);
        } catch (IllegalArgumentException ex) {
            fh = 18;
        }

        int rowHeight = table.getRowHeight ();
        if (e != null) {
             while (e.hasMoreElements ()) {
                NamedRegionSet f = (NamedRegionSet) e.nextElement ();
                rowHeight = Math.max (fh * countLines (f.toString ()),
                                      rowHeight);
                h.addElement (f);

                if (!isExpanded ())
                    break; // only want one hypothesis if we're not expanded
             }
        }

        hypotheses = h;

        table.setRowHeight (rowHeight);

        ((JComponent) getParent()).revalidate ();

        ((HypothesisModel) table.getModel ()).fireTableDataChanged ();

        table.getSelectionModel ().setSelectionInterval (0, 0);
    }

    static int countLines (String s) {
        int n = 1;
        int i = -1;
        while ((i = s.indexOf ('\n', i+1)) != -1)
            ++n;
        return n;
    }

    ListSelectionListener selectionChanged = new ListSelectionListener () {
            public void valueChanged (ListSelectionEvent event) {
                int i = table.getSelectedRow ();
                if (event.getValueIsAdjusting () 
                    || i == -1)
                    return;

                fireSelectedHypothesis ((NamedRegionSet) hypotheses.elementAt (i));
            }
        };

    public void addSelectionListener (SelectionListener l) {
        listeners.add (l);
    }

    public void removeSelectionListener (SelectionListener l) {
        listeners.remove (l);
    }

    private void fireSelectedHypothesis (NamedRegionSet sel) {
        for (Iterator g = listeners.iterator (); g.hasNext (); )
            ((SelectionListener) g.next ()).selectedHypothesis (sel);
    }

    public interface SelectionListener {
        void selectedHypothesis (NamedRegionSet sel);
    }


    class HypothesisModel extends AbstractTableModel {
        public String getColumnName(int col) { 
            switch (col) {
            case 0:
                return "Pattern";
            case 1:
                return "Matches";
            default:
                return "";
            }
        }

        public int getRowCount() { 
            if (!expanded)
                return Math.min (1, hypotheses.size ());
            else
                return hypotheses.size ();
        }

        public int getColumnCount() { 
            return 2;
        }

        public Class getColumnClass (int col) {
            switch (col) {
            case 0:
                return Object.class;
            case 1:
                return Integer.class;
            default:
                return Object.class;
            }
        }

        public Object getValueAt(int row, int col) { 
            NamedRegionSet f = (NamedRegionSet) hypotheses.elementAt (row);
            switch (col) {
            case 0:
                return f;
            case 1:
                return new Integer (f.getSize ());
            default:
                return null;
            }
            
        }
    }

    class NumberRenderer extends DefaultTableCellRenderer {
        NumberFormat fmt;

        public NumberRenderer (NumberFormat fmt) {
            this.fmt = fmt;
            setHorizontalAlignment (SwingConstants.RIGHT);
            setVerticalAlignment (SwingConstants.TOP);
        }

        public Component getTableCellRendererComponent(JTable table,
                                                       Object value,
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int row,
                                                       int column) {
            DefaultTableCellRenderer r = 
                (DefaultTableCellRenderer)
                super.getTableCellRendererComponent (table, value, isSelected,
                                                     hasFocus, row, column);

            String result = 
                (value == null)
                ? ""
                : fmt.format ((Number) value);

            r.setText (result);
            return r;
        }
    }


    static Border noFocusBorder = new EmptyBorder(1, 1, 1, 1); 

    class PatternRenderer extends TCEditor implements TableCellRenderer {
        public PatternRenderer () {
            super (true, false);
            setEditable (false);
        }


        public Component getTableCellRendererComponent(JTable table,
                                                       Object value,
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int row,
                                                       int column) {
            setText (value.toString ());

            if (isSelected) {
                super.setForeground(table.getSelectionForeground());
                super.setBackground(table.getSelectionBackground());
            } else {
                super.setForeground(table.getForeground());
                super.setBackground(table.getBackground());
            }
	
            if (hasFocus) {
                PatternRenderer.this.setBorder( UIManager.getBorder("Table.focusCellHighlightBorder") );
            } else {
                PatternRenderer.this.setBorder(noFocusBorder);
            }
                
            // ---- begin optimization to avoid painting background ----
            Color back = PatternRenderer.this.getBackground();
            boolean colorMatch = (back != null) && 
                ( back.equals(table.getBackground()) ) && table.isOpaque();
            PatternRenderer.this.setOpaque(!colorMatch);
            // ---- end optimization to aviod painting background ----
            
            return this;
        }

    }


}
