/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */
package lapis.swing;
import lapis.*;
import lapis.tc.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.awt.dnd.*;
import java.io.*;
import java.util.*;
import java.text.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import lapisx.sort.*;
import lapisx.util.SafeIteratorSet;
import lapisx.util.Str;
import java.util.Set;
import java.util.HashSet;


public class LibraryTree
    extends JTree
    implements BrowserListener, TreeSelectionListener, DragGestureListener, DropTargetListener, DragSourceListener {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    private Browser browser;
    private PatternLibrary lib;
    
    /** Variables needed for DnD */
    private DragSource dragSource = null;
    private DragSourceContext dragSourceContext = null;
    /** Stores the selected node info */
    private TreePath SelectedTreePath = null;
    private Node SelectedNode = null;
    private TreePath LastTreePath = null;
    private boolean dragOccured = false;
    private DropTarget dropTarget;   
    private final Set libTreeListeners = new SafeIteratorSet();
    public LibraryTree (Browser browser) {
        this.browser = browser;
        this.lib = PatternLibrary.getDocumentLibrary (browser.getDocument ());

        setModel (new Model ());

	setRootVisible(false);
	setShowsRootHandles(true);
	setToggleClickCount(1000);
	// Only allow library names to be selected (not categories)
	setSelectionModel(new LibraryTreeSelectionModel());
	// Draw lines between parents and children
	putClientProperty("JTree.lineStyle", "Angled");
	// Clear all icons
	setCellRenderer(new Renderer());
	// Show the root
	expandPath(new TreePath(getModel().getRoot()));
	// Handle clicks
	addMouseListener(new MouseAdapter() {
		TreePath pathPressed = null;
		public void mousePressed(MouseEvent e) {
		    pathPressed = getPathForLocation(e.getX(), e.getY());
		    if (pathPressed == null) {
			SelectedNode = null;
		    }
		}
		public void mouseReleased(MouseEvent e) {
		    if (dragOccured) {
			dragOccured = false;
			return;
		    }
		    TreePath pathReleased = getPathForLocation(e.getX(), e.getY());
		    if (pathPressed == null || pathReleased == null || !pathReleased.equals(pathPressed))
			return;
		    debug.println("expanding");
		    expandPath(pathPressed);
		    //check to see if the selected node is grayed appropriately
		    grayOut(getLastNode(pathPressed));
		    runPatterns();
		}
	    });
	// Handle Enter key
	addKeyListener(new KeyAdapter() {
		public void keyPressed(KeyEvent event) {
		    if (event.getKeyCode() == KeyEvent.VK_ENTER)
			runPattern(getSelectedName());
		}
	    });
	addTreeSelectionListener(this);
	addTreeExpansionListener(new LibraryTreeExpansionListener());
	
	/* drag and drop */
	dragSource = DragSource.getDefaultDragSource();
	DragGestureRecognizer dgr = dragSource.createDefaultDragGestureRecognizer(this, //DragSource
										  DnDConstants.ACTION_COPY_OR_MOVE, //specifies valid actions
										  this //DragGestureListener
										  );
	/* Eliminates right mouse clicks as valid actions - useful especially
	 * if you implement a JPopupMenu for the JTree
	 */
	dgr.setSourceActions(dgr.getSourceActions() & ~InputEvent.BUTTON3_MASK);
	/* First argument:  Component to associate the target with
	 * Second argument: DropTargetListener 
	 */
	dropTarget = new DropTarget(this, this);

	/*Eliminates ellipsis when tree label dynamically becomes larger
	 *than the bounds originally allotted
	 */
	
	ToolTipManager.sharedInstance().registerComponent(this);

	setLargeModel(true);
 	setRowHeight(18);
    }
    public void loadingDocument(Browser browser, SwingDocument doc) {
	//do nothing
    }
    public void loadedDocument(Browser browser, SwingDocument doc) {
        PatternLibrary newLib = PatternLibrary.getDocumentLibrary (doc);
	((Model) getModel ()).changedLibrary (lib, newLib);
	grayOut();
    }

    public int getTreeWidth() {
	return getWidth();
    }

    void runPattern(String name) {
	if (name == null)
	    return;
	Browser.debug.println("clicked on " + name + " in label tree");
	debug.println("clicked on " + name + " in label tree");
	// Display minimal name in pattern box
	String minimalName = lib.getShortestName(name);
	Browser.debug.println("minimal name = " + minimalName);
	debug.println("minimal name = " + minimalName);
	browser.setPattern(minimalName, true);
	// Search for it
	browser.find(TC.make(name));
    }
    /**
     * runs the Patterns of all currently selected nodes. 
     */
    void runPatterns() {
	String pattern = "";
	TreePath[] paths = getSelectionPaths();
	if (paths == null) {
	    browser.clearSelection();
	    return;
	}
	for (int i = 0; i < paths.length - 1; i++) {
	    Node node = (Node)paths[i].getPathComponent(paths[i].getPathCount() - 1);
	    if (node.isDefined) {
		debug.println(node.fullName);
		String minimalName = lib.getShortestName(node.fullName);
		pattern = pattern + minimalName + " or ";
	    }
	}
	Node node = (Node)paths[paths.length - 1].getPathComponent(paths[paths.length - 1].getPathCount() - 1);
	if (node.isDefined) {
	    debug.println(node.fullName);
	    String minimalName = lib.getShortestName(node.fullName);
	    pattern = pattern + minimalName;
	}
	else {
	    pattern = pattern.substring(0, pattern.length() - 4);
	}
	runPattern(pattern);
    }
    public void setSelectedName(String name) {
	if (name == null)
	    clearSelection();
	else {
	    TreePath path = getPathForName(name);
	    debug.println("name to be selected: " + name);
	    if (path == null)
		clearSelection();
	    else {
		setSelectionPath(path);
		scrollPathToVisible(path);
	    }
	}
    }
    public void setSelectedNames(Set names) {
	if (names == null)
	    clearSelection();
	else {
	    Iterator iter = names.iterator();
	    TreePath[] paths = new TreePath[1];
	    HashSet pathSet = new HashSet();
	    for (int i = 0; iter.hasNext(); i++) {
		String name = (String)iter.next();
		TreePath path = getPathForName(name);
		debug.println("name to be selected: " + name);
		if (path == null) {
		    clearSelection();
		}
		else {
		    pathSet.add(path);
		}
	    }
	    if (pathSet.size() > 0) {
		paths = (TreePath[])pathSet.toArray(paths);
		setSelectionPaths(paths);
		scrollPathToVisible(paths[0]);
	    }
	}
    }
    /** Returns The currently selected node */
    Node getSelectedNode() {
	return SelectedNode;
    }
    TreePath getSelectedPath() {
	return SelectedTreePath;
    }
    TreePath getPathForName(String name) {
	name += ".";
	// Count the number of path components
	int pathLength = 1; // root
	int i = 0;
	while ((i = name.indexOf('.', i)) > -1) {
	    ++pathLength;
	    ++i;
	}
	// Allocate the path and start with the root node
	Node[] path = new Node[pathLength];
	int iPath = 0;
	Node node = (Node)getModel().getRoot();
	path[iPath++] = node;
	// Consume each component of the name, searching for it in 
	// the current node.
	pathsearch : while ((i = name.indexOf('.')) > -1) {
	    String head = name.substring(0, i);
	    name = name.substring(i + 1);
	    for (int j = 0, n = node.children.size(); j < n; ++j) {
		Node child = (Node)node.children.elementAt(j);
		if (child.baseName.equalsIgnoreCase(head)) {
		    node = child;
		    path[iPath++] = child;
		    continue pathsearch;
		}
	    }
	    // Library component not found among children of current node.
	    // Fail out.
	    debug.println("can't find " + head + " in " + node.fullName);
	    return null;
	}
	return new TreePath(path);
    }
    public String getSelectedName() {
	TreePath path = getSelectionPath();
	if (path == null)
	    return null;
	return getNameForTreePath(path);
    }
    Node getLastNode(TreePath path) {
	return (Node)path.getPathComponent(path.getPathCount() - 1);
    }
    boolean isDefinedPattern(TreePath path) {
	return getLastNode(path).isDefined;
    }
    String getNameForTreePath(TreePath path) {
	return getLastNode(path).fullName;
    }
    String getNameAtLocation(int x, int y) {
	TreePath path = getPathForLocation(x, y);
	if (path == null)
	    return null;
	return getNameForTreePath(path);
    }

    /**grays the visible libraries from the root and repaints*/
    void grayOut() {
	debug.println("Graying inapplicable patterns out");
	Node root = (Node) getModel().getRoot();
	grayOut(new TreePath(root));
	treeDidChange();
        repaint ();
    }

    /**
     *grays out an individual node by checking to see
     *if it has zero matches to the document
     */
    void grayOut(Node node) {
	Document doc = browser.getDocument();
	boolean grayedOut = false;
	int matchCount = 0;
	if (doc != null && node.isDefined) 
	    try {
		Pattern pat = lib.get (node.fullName);
		matchCount = pat.match(doc).getSize();
		grayedOut = (matchCount==0);
	    } catch(Exception ex) {
		debug.println("error testing pattern");
		debug.report(ex);
	    }
	debug.println (node.fullName + " is " + (grayedOut ? "disabled" : "enabled"));
	node.setMatchCount(matchCount);
	node.setGrayValue(grayedOut);
    }

    /**
     *grays out the visible nodes on a certain path
     *including nodes on expanded subpaths
     */
    void grayOut(TreePath path) {
	java.util.List visibleNodes = new ArrayList();
	Node tempNode;

	//find the visible nodes directly under this path
	tempNode = getLastNode(path);
	visibleNodes.addAll(tempNode.children);

	//check all the visible nodes
	for(int i = 0; i< visibleNodes.size(); i++) {
	    tempNode = (Node) visibleNodes.get(i);
	    grayOut(tempNode);
	}

	//find the visible nodes of the children
	Enumeration e = getExpandedDescendants(path);
	while(e.hasMoreElements())
	    grayOut((TreePath) e.nextElement());
    }
    public void addLibraryTreeListener(LibraryTreeListener ltl) {
    	libTreeListeners.add(ltl);
    }
    
    class LibraryTreeExpansionListener implements TreeExpansionListener {
	public void treeExpanded(TreeExpansionEvent evt) {
	    LibraryTree tree = (LibraryTree) evt.getSource();
	    tree.grayOut(evt.getPath());
	}
	public void treeCollapsed(TreeExpansionEvent evt) {}
    }

    class LibraryTreeModelListener implements TreeModelListener {
	public void treeNodesChanged(TreeModelEvent evt) {}
	public void treeNodesInserted(TreeModelEvent evt) {
	    grayOut(getLastNode(evt.getTreePath()));
	    treeDidChange();
	}
	public void treeNodesRemoved(TreeModelEvent evt) {}
	public void treeStructureChanged(TreeModelEvent evt) {}
    }

    class Model implements javax.swing.tree.TreeModel, PatternLibrary.LibraryListener {
	Collator collator;
	Node root;
	EventListenerList listeners = new EventListenerList();
	public Model () {
	    collator = Collator.getInstance();
	    collator.setStrength(Collator.PRIMARY);
	    root = new Node("*root*", collator, false);            
            changedLibrary (null, lib);
	}
	public void changedLibrary (PatternLibrary oldLib, PatternLibrary newLib) {
            if (oldLib != null)
                oldLib.removeLibraryListener (this);
            newLib.addLibraryListener (this);

	    lib = newLib;
	    rebuild();
	}
	public void patternAdded(PatternLibrary lib, String name) {
	    debug.println("inserting " + name);
	    insert(new Node[] { root }, new Node(name, collator, true, lib));
	}
	public void patternRemoved(PatternLibrary lib, String name) {
	    debug.println("removing " + name);
	    remove(new Node[] { root }, new Node(name, collator, true, lib));
	}
        public void parserAdded (PatternLibrary lib, Parser parser) {
            // refetch our library, so that parser's document-specific patterns
            // will be added to it.
            PatternLibrary.getDocumentLibrary (browser.getDocument ());            
        }
        
	void rebuild() {
	    debug.println("Rebuilding label tree");
	    // Extract all (non-system) names from labeling
	    Vector nodes = new Vector();
	    if (lib != null) {
            
		Document doc = browser.getDocument();
    		for (Enumeration e = lib.visibleNames(); e.hasMoreElements();) {
    		    String name = e.nextElement().toString();
/*                
                    boolean grayedOut = false;
                    if (doc != null)                  
        		    try {
        			Pattern pat = lib.get (name);
        			grayedOut = (pat.match(doc).isEmpty());
        		    } catch(Exception ex) {
                                debug.report (ex);
        		    }
                    
                    //grayedOut = name.endsWith ("d");
                    debug.println (name + " is " + (grayedOut ? "disabled" : "enabled"));
		    */
                    nodes.addElement(new Node(name, collator, true, lib));
		}
	    }
	    // Sort the nodes to simplify tree construction
	    Quicksort.sort(new SortableVector(nodes) {
		    public int compare(Object o1, Object o2) {
			return ((Node)o1).compareTo((Node)o2);
		    }
		});
	    // Reconstruct the tree from the sorted listing.
	    rebuild(new Node[] { root }, nodes, 0, true);
	}
		    
	    
 	int rebuild(Node[] path, Vector nodes, int i, boolean parentAlreadyExisted) {
	    Node parent = path[path.length - 1];
	    String myprefix = (parent == root) ? "" : parent.fullName + ".";
	    int j = 0; // current index in children
	    int m = parent.children.size();
	    int n = nodes.size();
	    while (i < n) {
		Node newChild = (Node)nodes.elementAt(i);
		if (!Str.startsWithIgnoreCase(newChild.fullName, myprefix))					// newChild isn't a descendent of this node; since
		    // nodes is a sorted array, there are no more
		    // descendents of this node. 
		    break;
		// Check whether newChild is a child or a deeper descendent
		int nextdot = newChild.fullName.indexOf('.', myprefix.length());
		if (nextdot == -1)					// Child.  Consume the element from nodes, since we'll
		    // add it to this node's child list.
		    ++i;
		else					// Deeper descendent, so the child that leads to
		    // newChild is missing.  Create the missing child, but
		    // don't consume the element from nodes yet -- the child
		    // or one of its descendents will.
		    newChild = new Node(newChild.fullName.substring(0, nextdot), collator, false, lib);
		// newChild is this node's next new child.  
		// Check whether it already existed in the previous tree.
		boolean childAlreadyExisted = false;
		while (j < m) {
		    Node oldChild = (Node)parent.children.elementAt(j);
		    int diff = newChild.key.compareTo(oldChild.key);
		    if (diff == 0) {
			    // newChild == oldChild
			    oldChild.isDefined = newChild.isDefined;
                oldChild.isGray = newChild.isGray;
                oldChild.toolTip = newChild.toolTip;
			    newChild = oldChild;
			    debug.println("same: " + newChild.fullName);
			    childAlreadyExisted = true;
			    ++j;
			    break;
		    }
		    else
			if (diff > 0) {
			    // newChild > oldChild
			    parent.children.removeElementAt(j);
			    debug.println("removed: " + oldChild.fullName);
			    if (parentAlreadyExisted)								//changed FROM fileTreeNodesChanged - MN
				fireTreeNodesRemoved(path, new int[] { j }, new Node[] { oldChild });
			    --m;
			}
			else							// newChild < oldChild
			    break;
		}
		if (!childAlreadyExisted) {
		    // Add the child
		    parent.children.insertElementAt(newChild, j);
		    debug.println("added: " + newChild.fullName);
		    ++j;
		    ++m;
		}
		// Give newChild a chance to consume its descendents.
		i = rebuild(extendPath(path, newChild), nodes, i, childAlreadyExisted);
		if (parentAlreadyExisted && !childAlreadyExisted)
		    fireTreeNodesInserted(path, new int[] { j - 1 }, new Node[] { newChild });
	    }
	    // Remove all remaining original children
	    while (j < m) {
		Node oldChild = (Node)parent.children.elementAt(j);
		parent.children.removeElementAt(j);
		debug.println("removed: " + oldChild.fullName);
		if (parentAlreadyExisted)
		    fireTreeNodesRemoved(path, new int[] { j }, new Node[] { oldChild });
		--m;
	    }
	    return i;
	}
	Node[] extendPath(Node[] path, Node node) {
	    Node[] newPath = new Node[path.length + 1];
	    System.arraycopy(path, 0, newPath, 0, path.length);
	    newPath[path.length] = node;
	    return newPath;
	}
	void insert(Node[] path, Node node) {
	    debug.println("inserting " + node.toString() + " to " + path.toString());
	    Node parent = path[path.length - 1];
	    String myprefix = (parent == root) ? "" : parent.fullName + ".";
	    // Check whether node is a child or a deeper descendent
	    int nextdot = node.fullName.indexOf('.', myprefix.length());
	    Node newChild;
	    if (nextdot == -1)				// Child
		newChild = node;
	    else				// Deeper descendent, so the child that leads to
		// newChild is missing.
		newChild = new Node(node.fullName.substring(0, nextdot), collator, false, lib);
	    // newChild is this node's next new child.  
	    // Check whether it already exists in the tree.
	    boolean childAlreadyExisted = false;
	    int j, m;
	    for (j = 0, m = parent.children.size(); j < m; ++j) {
		Node oldChild = (Node)parent.children.elementAt(j);
		int diff = newChild.compareTo(oldChild);
		if (diff == 0) {
		    // newChild == oldChild
		    if (newChild.isDefined)
			oldChild.isDefined = true;
		    newChild = oldChild;
		    debug.println("same: " + newChild.fullName);
		    fireTreeNodesChanged(path, new int[] { j }, new Node[] { oldChild });
		    childAlreadyExisted = true;
		    break;
		}
		else
		    if (diff < 0)						// newChild < oldChild, so insert newChild here
			break;
	    }
	    if (!childAlreadyExisted) {
		// insert newChild here
		parent.children.insertElementAt(newChild, j);
		debug.println("different: " + newChild.fullName);
		debug.println("added: " + newChild.fullName);
		grayOut(newChild);
		fireTreeNodesInserted(path, new int[] { j }, new Node[] { newChild });
	    }
	    if (nextdot != -1)
		insert(extendPath(path, newChild), node);
	}
	boolean remove(Node[] path, Node node) {
	    Node parent = path[path.length - 1];
	    String myprefix = (parent == root) ? "" : parent.fullName + ".";
	    // Check whether node is a child or a deeper descendent
	    int nextdot = node.fullName.indexOf('.', myprefix.length());
	    Node newChild;
	    if (nextdot == -1)				// Child
		newChild = node;
	    else				// Deeper descendent, so the child that leads to
		// newChild is missing.
		newChild = new Node(node.fullName.substring(0, nextdot), collator, false, lib);
	    int j, m;
	    for (j = 0, m = parent.children.size(); j < m; ++j) {
		Node oldChild = (Node)parent.children.elementAt(j);
		int diff = newChild.compareTo(oldChild);
		debug.println(newChild.fullName + " vs. " + oldChild.fullName);
		if (diff == 0) {
		    // newChild == oldChild
		    if (newChild == node) {
			if (oldChild.children.size() > 0) {
			    debug.println("changed: " + oldChild.fullName);
			    oldChild.isDefined = false;
			    fireTreeNodesChanged(path, new int[] { j }, new Node[] { oldChild });
			}
			else {
			    debug.println("removed: " + oldChild.fullName);
			    parent.children.removeElementAt(j);
			    fireTreeNodesRemoved(path, new int[] { j }, new Node[] { oldChild });
			}
		    }
		    else
			if (remove(extendPath(path, oldChild), node)) {
			    debug.println("removed: " + oldChild.fullName);
			    parent.children.removeElementAt(j);
			    fireTreeNodesRemoved(path, new int[] { j }, new Node[] { oldChild });
			}
		    break;
		}
	    }
	    return !parent.isDefined && parent.children.size() == 0;
	}
	////////////////////
	//////////////////// TreeModel interface
	////////////////////
	public Object getRoot() {
	    return root;
	}
	public boolean isLeaf(Object node) {
	    return getChildCount(node) == 0;
	}
	public Object getChild(Object parent, int index) {
	    return ((Node)parent).children.elementAt(index);
	}
	public int getChildCount(Object parent) {
	    return ((Node)parent).children.size();
	}
	public int getIndexOfChild(Object parent, Object child) {
	    return ((Node)parent).children.indexOf(child);
	}
	public void addTreeModelListener(TreeModelListener listener) {
	    listeners.add(TreeModelListener.class, listener);
	}
	public void removeTreeModelListener(TreeModelListener listener) {
	    listeners.remove(TreeModelListener.class, listener);
	}
	public void valueForPathChanged(javax.swing.tree.TreePath path, Object newValue) {
	    // model is read-only
	}
	//
	// Firing events
	//
	protected void fireTreeNodesChanged(Object[] path, int[] childIndices, Object[] children) {
	    // Guaranteed to return a non-null array
	    Object[] lst = listeners.getListenerList();
	    TreeModelEvent e = null;
	    // Process the listeners last to first, notifying
	    // those that are interested in this event
	    for (int i = lst.length - 2; i >= 0; i -= 2) {
		if (lst[i] == TreeModelListener.class) {
		    // Lazily create the event:
		    if (e == null)
			e = new TreeModelEvent(this, path, childIndices, children);
		    ((TreeModelListener)lst[i + 1]).treeNodesChanged(e);
		}
	    }
	}
	/*
	 * Notify all listeners that have registered interest for
	 * notification on this event type.  The event instance 
	 * is lazily created using the parameters passed into 
	 * the fire method.
	 * @see EventListenerList
	 */
	protected void fireTreeNodesInserted(Object[] path, int[] childIndices, Object[] children) {
	    // Guaranteed to return a non-null array
	    Object[] lst = listeners.getListenerList();
	    TreeModelEvent e = null;
	    // Process the listeners last to first, notifying
	    // those that are interested in this event
	    for (int i = lst.length - 2; i >= 0; i -= 2) {
		if (lst[i] == TreeModelListener.class) {
		    // Lazily create the event:
		    if (e == null)
			e = new TreeModelEvent(this, path, childIndices, children);
		    ((TreeModelListener)lst[i + 1]).treeNodesInserted(e);
		}
	    }
	}
	/**
	 * Notify all listeners that have registered interest for
	 * notification on this event type.  The event instance 
	 * is lazily created using the parameters passed into 
	 * the fire method.
	 * @see EventListenerList
	 */
	protected void fireTreeNodesRemoved(Object[] path, int[] childIndices, Object[] children) {
	    // Guaranteed to return a non-null array
	    Object[] lst = listeners.getListenerList();
	    TreeModelEvent e = null;
	    // Process the listeners last to first, notifying
	    // those that are interested in this event
	    for (int i = lst.length - 2; i >= 0; i -= 2) {
		if (lst[i] == TreeModelListener.class) {
		    // Lazily create the event:
		    if (e == null)
			e = new TreeModelEvent(this, path, childIndices, children);
		    ((TreeModelListener)lst[i + 1]).treeNodesRemoved(e);
		}
	    }
	}
	protected void fireTreeStructureChanged(Object[] path) {
	    // Guaranteed to return a non-null array
	    Object[] lst = listeners.getListenerList();
	    TreeModelEvent e = null;
	    // Process the listeners last to first, notifying
	    // those that are interested in this event
	    for (int i = lst.length - 2; i >= 0; i -= 2) {
		if (lst[i] == TreeModelListener.class) {
		    // Lazily create the event:
		    if (e == null)
			e = new TreeModelEvent(this, path, null, null);
		    ((TreeModelListener)lst[i + 1]).treeStructureChanged(e);
		}
	    }
	}
    }
    //
    // Tree node
    //
    static class Node extends Object implements Transferable {
	boolean isDefined; // true iff corresponds to a label in the labeling
	boolean isTopLevel; // true iff child of root
	boolean isGray;
	int matchCount;
	String fullName;
	String baseName;
	String label; // label shown in tree, usually same as baseName
	String toolTip;
	Vector children;
	CollationKey key;
	final public static DataFlavor NODE_FLAVOR = new DataFlavor(Node.class, "Library Tree Node");
	DataFlavor[] dataFlavorArray = { NODE_FLAVOR };
	public Node(String fullName, Collator collator, boolean isDefined) {
	    this.fullName = fullName;
	    this.baseName = PatternLibrary.getBasename(fullName);
	    // Workaround for a silly bug:
	    // * nodes are rendered using Renderer, which extends JLabel.
	    // * JLabel interprets any label that starts with <html> as HTML
	    // * so if the node's label starts with <html>, it is misdisplayed.
	    // Workaround simply prepends a space.
	    this.label = (baseName.toLowerCase().startsWith("<html>")) ? " " + baseName : baseName;
	    this.children = new Vector();
	    this.key = collator.getCollationKey(fullName);
	    this.isDefined = isDefined;
	    this.isGray = false;
	    this.matchCount = 0;
	    this.toolTip = null;
	    int p = fullName.indexOf('.');
	    if (p == -1)
		isTopLevel = true;
	    else
		if (p > 0)
		    isTopLevel = false;
		else
		    isTopLevel = fullName.indexOf('.', 1) == -1;
	}

	public Node(String fullName, Collator collator, boolean isDefined, 
	            PatternLibrary patLib) {
        this.fullName = fullName;
        this.baseName = PatternLibrary.getBasename(fullName);
        // Workaround for a silly bug:
        // * nodes are rendered using Renderer, which extends JLabel.
        // * JLabel interprets any label that starts with <html> as HTML
        // * so if the node's label starts with <html>, it is misdisplayed.
        // Workaround simply prepends a space.
        this.label = (baseName.toLowerCase().startsWith("<html>")) ? " " + baseName : baseName;
        this.children = new Vector();
        this.key = collator.getCollationKey(fullName);
        this.isDefined = isDefined;
        this.isGray = false;
        this.matchCount = 0;
        /*Pattern tempPat = patLib.get(fullName);
        if(tempPat instanceof CachedPattern)
            tempPat = ((CachedPattern) tempPat).getPattern();
        if(tempPat instanceof TC) {
            this.toolTip = ((TC) tempPat).getExpression();
            //this.toolTip = ((TC) tempPat).parseSyntax();
        }
        */
        this.toolTip = null;
        int p = fullName.indexOf('.');
        if (p == -1)
            isTopLevel = true;
        else if (p > 0)
            isTopLevel = false;
        else
            isTopLevel = fullName.indexOf('.', 1) == -1;
    }

	public void setGrayValue(boolean isGray) {
	    this.isGray = isGray;
	}

	public void setMatchCount(int mc) {
	    this.matchCount = mc;
	}
	
	public boolean equals(Object obj) {
	    return (obj instanceof Node) && fullName.equals(((Node)obj).fullName);
	}
	
	public String toString() {
	    if(isDefined) {
		return label+" ("+matchCount+")";
	    }
	    else return label;
	}
	
	// Sort: primary key puts toplevel childless nodes after all others, 
	// secondary key is just alphabetical (ignoring case).
	public int compareTo(Node n) {
	    int diff = ((isTopLevel && isDefined) ? 1 : 0) - ((n.isTopLevel && n.isDefined) ? 1 : 0);
	    if (diff == 0)
		diff = key.compareTo(n.key);
	    return diff;
	}
	/* Transferable interface */
	/*  Returns an object which represents the data to be transferred. */
	public Object getTransferData(DataFlavor flavor) {
	    //	    throws UnsupportedFlavorException, IOException {
	    //	    if (flavor.equals(NODE_FLAVOR)) {
	    return this;
	    //	    }
	    //	    else throw new UnsupportedFlavorException(flavor);
	}
	/* Returns an array of DataFlavor objects indicating the flavors the data can be provided in. */
	public DataFlavor[] getTransferDataFlavors() {
	    return dataFlavorArray;
	}
	/*Returns whether or not the specified data flavor is supported for this object. */
	public boolean isDataFlavorSupported(DataFlavor flavor) {
	    return flavor.equals(NODE_FLAVOR);
	}
    }
    final static Color lightBlue = new Color(100, 100, 255);
    class Renderer extends DefaultTreeCellRenderer {
	Icon openCategoryIcon;
	Icon closedCategoryIcon;
	Icon regionsIcon;
	Icon docSpecificIcon;
	Icon blankIcon;
	
	public Renderer() {
	    openCategoryIcon = getOpenIcon();
	    closedCategoryIcon = getClosedIcon();
	    regionsIcon = SwingConfig.makeIcon("region-set-icon.gif");
	    docSpecificIcon = SwingConfig.makeIcon("red.gif");
	    blankIcon = SwingConfig.makeIcon("blank.gif");
	}

	public Component getTreeCellRendererComponent(
						      JTree tree,
						      Object value,
						      boolean sel,
						      boolean expanded,
						      boolean leaf,
						      int row,
						      boolean hasFocus) {

	    Node n = (Node)value;
	    Font f = LibraryTree.this.getFont();
// 	    setFont(f.deriveFont(f.getStyle() ^ Font.BOLD));
	    setTextNonSelectionColor(Color.BLACK);
	    setTextSelectionColor(Color.BLACK);

	    if (!n.isDefined) {
		setOpenIcon(openCategoryIcon);
		setClosedIcon(closedCategoryIcon);
	    }
	    else {
		setClosedIcon(regionsIcon);
		setOpenIcon(regionsIcon);
		setLeafIcon(regionsIcon);
		
		try{
		    PatternLibrary.getGlobalLibrary().lookup(n.fullName);
		}catch(PatternLibrary.LabelNotFoundException lnfe) {
		    setClosedIcon(docSpecificIcon);
		    setOpenIcon(docSpecificIcon);
		    setLeafIcon(docSpecificIcon);
		}catch(PatternLibrary.LabelAmbiguousException lae) {}

		if(n.isGray) {
		    setClosedIcon(blankIcon);
		    setOpenIcon(blankIcon);
		    setLeafIcon(blankIcon);
		    setTextNonSelectionColor(Color.GRAY);
		    setTextSelectionColor(Color.GRAY);
		    setFont(f);
		}
		    
	    }
	    Component c = super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf, row, hasFocus);
    	JComponent labelWithToolTip = (JComponent) c;
    	labelWithToolTip.setToolTipText(n.toolTip); 
		ToolTipManager.sharedInstance().registerComponent(labelWithToolTip); 
    	return c;
	}
    }
    static class LibraryTreeSelectionModel extends DefaultTreeSelectionModel {
	public LibraryTreeSelectionModel() {
	    //            setSelectionMode (SINGLE_TREE_SELECTION);
	    setSelectionMode(DISCONTIGUOUS_TREE_SELECTION);
	}
	TreePath[] pruneUnselectablePaths(TreePath[] paths) {
	    for (int i = paths.length - 1; i >= 0; --i) {
		Node n = (Node)paths[i].getLastPathComponent();
		if (!n.isDefined) {
		    TreePath[] newpaths = new TreePath[paths.length - 1];
		    System.arraycopy(paths, 0, newpaths, 0, i);
		    if (i < paths.length - 1)
			System.arraycopy(paths, i + 1, newpaths, i, paths.length - i);
		    paths = newpaths;
		}
	    }
	    return paths;
	}
	public void addSelectionPaths(TreePath[] paths) {
	    paths = pruneUnselectablePaths(paths);
	    if (paths.length > 0)
		super.addSelectionPaths(paths);
	}
	public void setSelectionPaths(TreePath[] paths) {
	    paths = pruneUnselectablePaths(paths);
	    if (paths.length > 0)
		super.setSelectionPaths(paths);
	}
    }
    /** DragGestureListener interface method */
    public void dragGestureRecognized(DragGestureEvent e) {
	//Get the selected node
	dragOccured = true;
	Node dragNode = getSelectedNode();
	LastTreePath = getSelectedPath();
	if (dragNode != null) {
	    debug.println("trying to drag " + dragNode.toString());
	    //Get the Transferable Object
	    Transferable transferable = (Transferable)dragNode;
	    //Select the appropriate cursor;
	    // This part still doesn't work. Don't realy know why. Somewhere it's getting changed 
	    Cursor cursor = DragSource.DefaultCopyNoDrop;
	    int action = e.getDragAction();
	    if (action == DnDConstants.ACTION_MOVE)
		cursor = DragSource.DefaultMoveNoDrop;
	    //begin the drag
	    dragSource.startDrag(e, cursor, transferable, this);
	    //   }
	}
    }
    /** DragSourceListener interface method */
    public void dragDropEnd(DragSourceDropEvent dsde) {
	this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	this.setSelectionModel(new LibraryTreeSelectionModel());
    }
    /** DragSourceListener interface method */
    public void dragEnter(DragSourceDragEvent dsde) {
	//	this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
    }
    /** DragSourceListener interface method */
    public void dragOver(DragSourceDragEvent dsde) {
	debug.println("dragged over PathName ");
	//+ getNameAtLocation ((int)(dsde.getX()), 
	// 									 (int)(dsde.getY())));
	// 	String pathName = getNameAtLocation ((int)(dsde.getX()),	
	// 						 (int)(dsde.getY()));
	// 	if (pathName != null) {
	// 	    this.setCursor(DragSource.DefaultMoveDrop);
	// 	}
    }
    /** DragSourceListener interface method */
    public void dropActionChanged(DragSourceDragEvent dsde) {
	int action = dsde.getDropAction();
	if (action == DnDConstants.ACTION_MOVE)
	    this.setCursor(DragSource.DefaultMoveDrop);
	if (action == DnDConstants.ACTION_NONE)
	    this.setCursor(DragSource.DefaultMoveNoDrop);
	debug.println("changed drop action");
    }
    /** DragSourceListener interface method */
    public void dragExit(DragSourceEvent dsde) {
	//	this.setCursor(DragSource.DefaultMoveNoDrop);
    }
    /** DropTargetListener interface method - What we do when drag is released */
    public void drop(DropTargetDropEvent e) {
	try {
	    Transferable tr = e.getTransferable();
	    //flavor not supported, reject drop
	    if (!tr.isDataFlavorSupported(Node.NODE_FLAVOR)) {
		e.rejectDrop();
		this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		this.setSelectionModel(new LibraryTreeSelectionModel());
	    }
	    Node newChild = (Node)tr.getTransferData(Node.NODE_FLAVOR);
	    String destinationPath = getNameAtLocation((int) (e.getLocation().getX()), (int) (e.getLocation().getY()));
	    debug.println("destPath: " + destinationPath);
	    debug.println("formerPath: " + SelectedNode.fullName);
	    if ((destinationPath != null) && (!destinationPath.equals(SelectedNode.fullName))) {
		String oldName = SelectedNode.fullName;
		String newName = new String(destinationPath + "." + newChild.baseName);
		debug.println("named " + newName);              
                lib.rename (oldName, newName);
		runPattern(newName);
	    }
	    this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	    this.setSelectionModel(new LibraryTreeSelectionModel());
	}
	catch (IOException io) {
	    e.rejectDrop();
	    this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	    this.setSelectionModel(new LibraryTreeSelectionModel());
	}
	catch (UnsupportedFlavorException ufe) {
	    e.rejectDrop();
	    this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	    this.setSelectionModel(new LibraryTreeSelectionModel());
	}
	this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	this.setSelectionModel(new LibraryTreeSelectionModel());
    }
    /** DropTaregetListener interface method */
    public void dragEnter(DropTargetDragEvent e) {
	//	this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	this.setCursor(DragSource.DefaultMoveDrop);
	debug.println("entered drop target");
    }
    /** DropTaregetListener interface method */
    public void dragExit(DropTargetEvent e) {
	this.setCursor(DragSource.DefaultMoveNoDrop);
	debug.println("exited drop target");
    }
    /** DropTaregetListener interface method */
    public void dragOver(DropTargetDragEvent e) {
	String pathName = getNameAtLocation((int) (e.getLocation().getX()), (int) (e.getLocation().getY()));
	this.setSelectionModel(new DefaultTreeSelectionModel());
	DefaultTreeSelectionModel model = (DefaultTreeSelectionModel)this.getSelectionModel();
	debug.println("cursor type: " + this.getCursor().getName());
	//	debug.println("dragged over PathName " + pathName)
	if (pathName != null) {
	    dropTarget.setDefaultActions(DnDConstants.ACTION_MOVE);
	    this.setCursor(DragSource.DefaultMoveDrop);
	    model.removeSelectionPath(LastTreePath);
	    LastTreePath = getPathForLocation((int) (e.getLocation().getX()), (int) (e.getLocation().getY()));
	    debug.println("LastPath: " + LastTreePath.toString());
	    TreePath[] treePaths = { LastTreePath, getSelectedPath()};
	    model.setSelectionPath(LastTreePath);
	    model.setSelectionPaths(treePaths);
	    getDropTarget().setDefaultActions(DnDConstants.ACTION_MOVE);
	    debug.println("dragged over PathName2 " + pathName);
	    debug.println("cursor type: " + this.getCursor().getName());
	}
	else {
	    this.setCursor(DragSource.DefaultMoveNoDrop);
	    model.removeSelectionPath(LastTreePath);
	    model.setSelectionPath(getSelectedPath());
	    getDropTarget().setDefaultActions(DnDConstants.ACTION_NONE);
	    debug.println("cursor type: " + this.getCursor().getName());
	}
	debug.println("drop action" + e.getDropAction());
	debug.println("cursor type: " + this.getCursor().getName());
    }
    /** DropTargetListener interface method */
    public void dropActionChanged(DropTargetDragEvent e) {
	int action = e.getDropAction();
	if (action == DnDConstants.ACTION_MOVE)
	    this.setCursor(DragSource.DefaultMoveDrop);
	if (action == DnDConstants.ACTION_NONE)
	    this.setCursor(DragSource.DefaultMoveNoDrop);
	debug.println("changed drop action");
    }
    /** TreeSelectionListener - sets selected node */
    public void valueChanged(TreeSelectionEvent evt) {
	SelectedTreePath = evt.getNewLeadSelectionPath();
	if (SelectedTreePath == null) {
	    SelectedNode = null;
		for (Iterator g = libTreeListeners.iterator(); g.hasNext (); )
			((LibraryTreeListener) g.next ()).treeNodeNotSelected();
	    return;
	}
	SelectedNode = (Node)SelectedTreePath.getLastPathComponent();
	debug.println("Node selected = " + SelectedNode.toString());
	for (Iterator g = libTreeListeners.iterator(); g.hasNext (); )
	    ((LibraryTreeListener) g.next ()).treeNodeSelected();
    }
    /** Convenience method to test whether drop location is valid
	@param destination The destination path 
	@param dropper The path for the node to be dropped
	@return null if no problems, otherwise an explanation
    */
    private String testDropTarget(TreePath destination, TreePath dropper) {
	return "";
    }
    public void setCursor(Cursor cursor) {
	debug.println("setCursor called");
	debug.printStackTrace();
	super.setCursor(cursor);
	debug.println("cursor type: " + cursor.getName());
    }
}
