/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import lapis.tc.*;
import lapis.ml.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.table.*;
import lapisx.swing.*;
import java.text.*;

public class OutlierFinder extends JFrame implements BrowserListener, SelectionListener {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    Browser browser;
    
    RegionSet allExamples;

    boolean selectionIsPositive;

    JSplitPane splitPane;
//     JButton fixButton;
//     JButton prevButton, nextButton;
//     JButton closeButton;
    FinderHistogram histogram;
    JLabel message;

    ExplanationModel explModel;
    JTable explTable;
    Similarity.Explanation[] explanation;

    public OutlierFinder (Browser browser) {
        super ("Unusual Matches");
        this.browser = browser;

        // Make GUI components
        Layout.vertical (getContentPane ())
            .setContainerInsets (5, 5, 5, 5)
            .add (splitPane = new JSplitPane 
                  (JSplitPane.VERTICAL_SPLIT,
                   Layout.vertical ()
                   .add (histogram = new FinderHistogram (browser))
                   .done (),
                   Layout.vertical ()
                   .add (message = new JLabel 
                         ("Searching for unusual matches.  Please wait..."))
                   .add (makeExplanation ())
//                    .add (Layout.horizontal ()
//                          .setContainerInsets (2, 2, 2, 2)
//                          .add (fixButton = ActionAdapters.toButton (fixAction))
//                          .add (prevButton = ActionAdapters.toButton (prevAction))
//                          .add (nextButton = ActionAdapters.toButton (nextAction))
//                          .add (Box.createHorizontalStrut (30))
//                          .add (closeButton = ActionAdapters.toButton (closeAction))
//                          .done ())
                   .done ()))
            ;

        splitPane.setResizeWeight (1);

//         fixButton.setEnabled (false);
//         prevButton.setEnabled (false);
//         nextButton.setEnabled (false);

        // How to handle window-close
        setDefaultCloseOperation (HIDE_ON_CLOSE);

        // Set default window size and position
        pack ();
        Win.position (this, browser, 2.0, 1.25);

        // Restore saved window size and position, if any 
        restoreWindowConfig ();
    }

    void saveWindowConfig () {
        SwingConfig.saveLocation ("outlier.location", this);
        SwingConfig.saveSize ("outlier.size", this);
        SwingConfig.saveSplitter ("outlier.splitter", splitPane);
        Main.getConfig ().save ();
    }

    void restoreWindowConfig () {
        SwingConfig.restoreLocation ("outlier.location", this);
        SwingConfig.restoreSize ("outlier.size", this);
        SwingConfig.restoreSplitter ("outlier.splitter", splitPane);
    }

    JComponent makeExplanation () { 
        explModel = new ExplanationModel ();
        explTable = new JTable (explModel);
        explTable.setShowGrid (false);
        explTable.setSelectionMode (ListSelectionModel.SINGLE_SELECTION);
        explTable.getTableHeader().setReorderingAllowed (false);

        // Make explanation scrollable
        JScrollPane scrollPane = new JScrollPane (explTable);
        scrollPane.setPreferredSize (new Dimension (450, 100));

        return scrollPane;
    }

    class ExplanationModel extends AbstractTableModel {
        public String getColumnName(int col) { 
            return "";
//             switch (col) {
//             case 0:
//                 if (clusters == null || curr < 0 || curr >= clusters.length)
//                     return "";
//                 else if (clusters[curr].isPosExample)
//                     return "Differences From Typical Matches";
//                 else
//                     return "Similarities To Typical Matches";
//             case 1:
//                 return "";
//             default:
//                 return "";
//             }
        }

        public int getRowCount() { 
            return explanation != null ? explanation.length : 0;
        }

        public int getColumnCount() { 
            return 2;
        }

        public Object getValueAt(int row, int col) { 
            switch (col) {
            case 0: 
                return explanation[row].feature;
            case 1: {
                NumberFormat fmt = NumberFormat.getPercentInstance ();
                double mean = explanation[row].posMean;
                return (selectionIsPositive
                          ? ("unlike " + fmt.format (1 - mean))
                          : ("like " + fmt.format (mean)))
                    + " of matches";
            }
            default:
                return "";
            }
        }
    }

    public void refresh () {
        NamedRegionSet sel = browser.getSelection ();
        Document doc = (lapis.Document)browser.getDocument();
        
        browser.getActivePane ().getFinderHighlighter().setSelection (null);
        
        if (doc == null || sel == null) {
            histogram.setExamples (null, null);
            return;
        }
        
        Inference inf;
        SelectionMode mode = browser.getSelectionMode ();
        TC pattern = browser.getSelection ().getPattern ();
            
        if (pattern != null) {
            allExamples = pattern.getUniversePattern ().match (doc);
        } else 
            allExamples = sel;
        
        if (allExamples.getSize () <= 1) {
            histogram.setExamples (null, null);
            return;
        }
                 
        RegionSet negExamples = allExamples.butnot (sel);
        
        message.setText ("Searching for unusual matches.  Please wait...");
		histogram.setExamples (sel, new NamedRegionSet (negExamples));
    }


    //
    // BrowserListener interface
    //

    public void loadingDocument (Browser browser, SwingDocument doc) {
	debug.println("Outlier Finder detects loading Document");
        if (isVisible ())
            refresh ();
    }

    public void loadedDocument (Browser browser, SwingDocument doc) {
    }

    //
    // SelectionListener interface
    //

    public void changedSelection (Browser browser, NamedRegionSet sel) {
        if (isVisible ())
            refresh ();
    }


    class FinderHistogram extends OutlierHistogram {
        
        public FinderHistogram (Browser browser) {
            super (browser);
        }

        protected synchronized void clustersChanged (Similarity.Cluster[] clusters) {
            super.clustersChanged (clusters);

            clearSelection ();
            for (int i = clusters.length-1; i >= 0; --i) {
                if (clusters[i].isPositive () || i == 0) {
                    select (i);
                    
                    final int j = i;
                    SwingUtilities.invokeLater (new Runnable () {
                            public void run () {
                                centerCluster (j);
                            }
                        });
                    break;
                }
            }
            endSelection ();
        }

        protected void selectionChanged (boolean[] selected) {
            Similarity.Cluster[] clusters = getClusters (); 
            
            RegionSet regions = RegionSet.EMPTY;
            int nPos = 0;
            int nNeg = 0;
            int nClusters = 0;
            int firstCluster = -1;
            
            selectionIsPositive = false;
            
            for (int i = 0; i < clusters.length; ++i) 
                if (selected[i]) {
                    ++nClusters;
                    
                    Similarity.Cluster ex = clusters[i];
                    //RegionSet r = ex.getRegions ();
                    
                    if (ex.isPositive ()) {
                        selectionIsPositive = true;
                        nPos += ex.getSize ();
                    } else
                        nNeg += ex.getSize ();
                    
                    regions = regions.or (ex);
                    
                    if (firstCluster == -1) 
                        firstCluster = i;
                }
            
            // update browser selection
            Browser b = OutlierFinder.this.browser;

			b.getActivePane ().getFinderHighlighter().setSelection(new NamedRegionSet (regions));
            b.getActivePane ().scrollRegionsToVisible (regions);
            
            // update explanation
            if (nClusters != 1)
                explanation = null;
            else
                explanation = clusters[firstCluster].getExplanation ();

            if (isVisible ()) {
                Font font = explTable.getFont ();
                explTable.setRowHeight (FontUtil.getHeight (font, this)
                                        + explTable.getRowMargin ());
            }

            explModel.fireTableDataChanged ();
            
            // set message label
            String label = "";
            if (clusters.length == 0)
                label = "Make a multiple selection to view unusual matches.";
            else if (nClusters > 0) {
                if (nPos > 0)
                    label += toCountLabel (nPos, true);
                if (nNeg > 0) {
                    if (label.length () > 0)
                        label += " and ";
                    label += toCountLabel (nNeg, false);
                }
                
                int n = regions.getSize ();
                label +=
                    ((nPos + nNeg == 1) ? " is" : " are")
                    + " highlighted in red.";
            }
            
            if (explanation != null && explanation.length > 0)
                label += " Reasons why " 
                    + ((nPos + nNeg == 1) ? "it" : "they") 
                    + " might be wrong:";
            
            message.setText (label);

            FinderHistogram.this.repaint ();
        }

        String toCountLabel (int n, boolean isPositive) {
            String isa = isPositive ? "match" : "mismatch";
            return
                (n == 1) 
                ? "1 " + isa
                : n + " " + isa + "es";
        }
    }


//     public Action closeAction = new AcceleratedAction ("&Close") {
//         public void actionPerformed (ActionEvent evt) {
//             dispose ();
//         }
//     };

    public void setVisible (boolean visible) {
        super.setVisible (visible);

        if (visible)
            refresh ();
        else
            browser.getActivePane ().getFinderHighlighter().setSelection (null);
    }

    public void dispose () {
        saveWindowConfig ();
        super.dispose ();
    }        
    
    public void jumpTo (Region r) {
        Similarity.Cluster[] clusters = histogram.getClusters ();

        histogram.clearSelection ();
        for (int i = 0; i < clusters.length; ++i) {
            if (clusters[i].intersects (r)) {
                histogram.select (i);
                break;
            }
        }
        histogram.endSelection ();
        debug.println ("region " + r + " is not an example");
    }
}
