/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import lapis.ml.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import lapisx.swing.*;
import java.text.*;
import javax.swing.plaf.metal.MetalLookAndFeel;

public class OutlierHistogram extends JPanel {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    static int BLOCK_WIDTH = 15;
    static int BLOCK_HEIGHT = 9;
    static {
        try {
            String s = Main.getConfig ().getProperty ("browser.fontsize");
            if (s != null && s.length () > 0) {
                int size = Integer.parseInt (s);
                double expansion = size / 12.0;
                if (expansion > 1.0) {
                    BLOCK_WIDTH = (int) (BLOCK_WIDTH * expansion);
                    BLOCK_HEIGHT = (int) (BLOCK_HEIGHT * expansion);
                }
            }
        } catch (NumberFormatException e) {
            debug.report (e);
        }
    }

    static int DEFAULT_WIDTH = 6 * BLOCK_WIDTH;
    static int DEFAULT_HEIGHT = 15 * BLOCK_HEIGHT;

    static Color selectedColor = Color.red;
    static Color positiveColor = new Color (180, 180, 255);
    static Color negativeColor = Color.white;

    static final int SEPARATION = 300; // pixels to allocate to separation

    static final double[] PERCENTILE = { 0.0, 0.50, 0.75, 0.90, 0.95, 1.00 };

    Browser browser;

    int nPositiveExamples;
    Similarity sim;
    Similarity.Cluster[] clusters;
    boolean[] selected;

    Histogram histogram;
    Container matchLegend, mismatchLegend;

    boolean hasNegativeExamples;
    double scale;
    double stddev;  // standard deviation of positive examples' distances

    Point click;
    Rectangle rubberBand;
    boolean incremental;


    public OutlierHistogram (Browser browser) { 
        this.browser = browser;
        histogram = new Histogram ();

        clusters = new Similarity.Cluster[0];
        selected = new boolean[0];

        JScrollPane scroller = new JScrollPane (histogram);
        scroller.setPreferredSize (histogram.getPreferredSize ());

        JComponent xaxis = (JComponent) Layout.horizontal ()
            .add (new JLabel (SwingConfig.makeIcon ("left-arrow.gif")))
            .add (new JLabel ("more typical"))
            .addGlue ()
            .add (Box.createHorizontalStrut (30))
            .add (new JLabel ("more unusual"))
            .add (new JLabel (SwingConfig.makeIcon ("right-arrow.gif")))
            .done ();

        JComponent legend = (JComponent) Layout.vertical ()
            .setInsets (2, 2, 2, 2)
            .setSize (Layout.FIXED_SIZE)
            .setAlignment (Layout.WEST)
            .add (matchLegend = Layout.horizontal ()
                  .setInsets (2, 2, 2, 2)
                  .add (new Block (true))
                  .add (new JLabel ("Match"))
                  .done ())
            .add (mismatchLegend = Layout.horizontal ()
                  .setInsets (2, 2, 2, 2)
                  .add (new Block (false))
                  .add (new JLabel ("Mismatch"))
                  .done ())
            .done ();

        matchLegend.setVisible (true);
        mismatchLegend.setVisible (false);

        Layout.horizontal (this)
            .setAlignment (Layout.NORTH)
            .add (Layout.vertical ()
                  .add (scroller)
                  .add (xaxis, Layout.FIXED_HEIGHT, new Insets (2, 5, 5, 5))
                  .done ())
            .add (legend, Layout.FIXED_WIDTH, Layout.NORTH, 
                  new Insets (20, 5, 0, 5));
    }

    public void setExamples (NamedRegionSet positive, NamedRegionSet negative) {
        if (positive == null && negative == null) {
            sim = null;
            clusters = new Similarity.Cluster[0];
            nPositiveExamples = 0;
            selected = new boolean[0];
            clearSelection ();
            histogram.repaint ();
            return;
        }

        nPositiveExamples = positive.getSize ();

        sim = new Similarity (positive, negative);

        browser.run (new RunnableWithExceptions () {
                public void run () {
                    try {
                        clusters = sim.getClusters ();
                        
                        SwingUtilities.invokeLater (new Runnable () {
                                public void run () {
                                    clustersChanged (clusters);
                                }
                            });
                    } catch (Exception e) {
                        debug.report (e);
                    }
                }
            });
    }

    protected void clustersChanged (Similarity.Cluster[] clusters) {
        selected = new boolean[clusters.length];

        boolean hasMismatches = false;
        for (int i = 0; i < clusters.length; ++i)
            if (!clusters[i].isPositive ()) {
                hasMismatches = true;
                break;
            }
        mismatchLegend.setVisible (hasMismatches);

        histogram.recalc ();
    }

    public Similarity getSimilarity () {
        return sim;
    }

    public Similarity.Cluster[] getClusters () {
        return clusters;
    }

    public boolean[] getSelection () {
        return selected;
    }

    public void centerCluster (int i) {
        histogram.centerCluster (i);
    }

    public void scrollClusterToVisible (int i) {
        histogram.scrollClusterToVisible (i);
    }


    void clearSelection () {
        for (int i = 0; i < selected.length; ++i)
            selected[i] = false;
        endSelection ();
    }

    void select (int index) {
        if (index >= 0 && index < selected.length)
            selected[index] = true;
    }

    void endSelection () {
        selectionChanged (selected);
    }

    protected void selectionChanged (boolean[] selection) {
        for (int i = 0; i < selected.length; ++i) {
            if (selected[i]) {
                scrollClusterToVisible (i);
                return;
            }
        }
        histogram.repaint ();
    }


    class Histogram extends JComponent {

        public Histogram () {
            Histogram.this.setAutoscrolls (true);
            Histogram.this.setForeground (MetalLookAndFeel.getSystemTextColor ());
            Histogram.this.setFont (MetalLookAndFeel.getControlTextFont ());

            recalc ();
            Histogram.this.addMouseListener (new MouseAdapter () {
                    public void mousePressed (MouseEvent e) {
                        click = e.getPoint ();
                        ComponentToBlock (click);
                        rubberBand = new Rectangle (click.x, click.y, 1, 1);
                        Histogram.this.repaint ();

                        incremental = e.isControlDown ();
                        if (!incremental)
                            clearSelection ();
                    }
                    public void mouseReleased (MouseEvent e) {
                        if (clusters != null) {
                            for (int i = 0; i < clusters.length; ++i) {
                                Rectangle b = getBlock (i);
                                if (incremental && b.contains (rubberBand))
                                    selected[i] = !selected[i];
                                else if (b.intersects (rubberBand))
                                    select (i);
                            }
                            endSelection ();
                        }
                        rubberBand = null;
                        Histogram.this.repaint ();
                    }

                });
            Histogram.this.addMouseMotionListener (new MouseMotionAdapter () {
                    public void mouseDragged (MouseEvent e) {
                        Point drag = e.getPoint ();
                        ComponentToBlock (drag);
                        rubberBand = new Rectangle (Math.min (click.x, drag.x),
                                                    Math.min (click.y, drag.y),
                                                    Math.abs (click.x - drag.x) + 1,
                                                    Math.abs (click.y - drag.y) + 1);
                        Histogram.this.repaint ();
                    }
                });
        }

        public void recalc () {
            hasNegativeExamples = false;

            if (clusters == null || clusters.length == 0) {
                Dimension d = new Dimension (DEFAULT_WIDTH, DEFAULT_HEIGHT);
                Histogram.this.setSize (d);
                Histogram.this.setPreferredSize (d);
                stddev = 0;
                return;
            }

            int maxY = 0;
            int minY = 0;
            int maxX = 0;
            double sumSquares = 0;
            int nPositive = 0;

            scale = SEPARATION / clusters[clusters.length-1].getDistance ();
            for (int i = 0; i < clusters.length; ++i) {
                Similarity.Cluster ex = clusters[i];
                Rectangle b = getBlock (i);
                maxX = Math.max (maxX, b.x + b.width + (clusters[i].isPositive () ? 120 : 0));
                minY = Math.min (minY, b.y);
                maxY = Math.max (maxY, b.y + b.height);

                if (clusters[i].isPositive ()) {
                    double d = ex.getDistance ();
                    sumSquares += ex.getSize () * d * d;
                    ++nPositive;
                } else
                    hasNegativeExamples = true;
            }

            stddev = Math.sqrt (sumSquares / nPositive);
            
            Insets insets = Histogram.this.getInsets ();
            int width = Math.max (maxX + insets.left + insets.right, DEFAULT_WIDTH);
            Dimension d = new Dimension (width, DEFAULT_HEIGHT);
            Histogram.this.setSize (d);
            Histogram.this.setPreferredSize (d);

            Histogram.this.repaint ();
        }
        
        Rectangle getBlock (int i) {
            int x = i * BLOCK_WIDTH + ((int) (scale * clusters[i].getDistance ()));
            int w = BLOCK_WIDTH;
            int h = clusters[i].getSize () * (BLOCK_HEIGHT-1) + 1;
            int y = (clusters[i].isPositive ()) ? -h : 1;
            return new Rectangle (x, y, w, h);
        }

        Rectangle getBlockInComponentCoords (int i) {
            Rectangle b = getBlock (i);
            Point pt = new Point (b.x, b.y);
            BlockToComponent (pt);
            b.x = pt.x;
            b.y = pt.y;
            return b;
        }

        int getMidline () {
            Insets insets = Histogram.this.getInsets ();
            int height = Histogram.this.getSize ().height;
            return hasNegativeExamples
                ? insets.top + (height - insets.top - insets.bottom)/2
                : height - insets.bottom;
            
        }

        void ComponentToBlock (Point pt) {
            Insets insets = Histogram.this.getInsets ();
            pt.x -= insets.left;
            pt.y -= getMidline ();
        }

        void BlockToComponent (Point pt) {
            Insets insets = Histogram.this.getInsets ();
            pt.x += insets.left;
            pt.y += getMidline ();
        }

        public void paintComponent (Graphics g) {
            super.paintComponent (g);

            // get font metrics, insets, and visible rectangle
            FontMetrics fm = Histogram.this.getFontMetrics (Histogram.this.getFont ());
            int th = fm.getHeight ();
            Insets insets = Histogram.this.getInsets ();
            int width = Histogram.this.getSize ().width
                - insets.left - insets.right;
            Rectangle visible = Histogram.this.getVisibleRect ();

            // set origin to left end of midline
            Point origin = new Point (0,0);
            BlockToComponent (origin);
            g.translate (origin.x, origin.y);
            visible.translate (-origin.x, -origin.y);
            
            // draw lines for each percentile
            if (clusters != null) {
                int nDrawn = 0;
                int p = PERCENTILE.length - 1;
                int nPercentile = (int) (nPositiveExamples * (1 - PERCENTILE[p]));
                debug.println (p + " --> " + nPercentile);
                int percentileGap = fm.stringWidth ("99%") * 2;
                int lastPercentileX = width;
                NumberFormat percentFmt = NumberFormat.getPercentInstance ();
                NumberFormat distFmt = NumberFormat.getInstance ();
                distFmt.setMaximumFractionDigits (1);

                g.setColor (Histogram.this.getForeground ());
                percentileLoop:
                for (int i = clusters.length - 1; i >= 0; --i) {
                    if (clusters[i].isPositive ()) {
                        if (nPercentile <= nDrawn) {
                            debug.println ("cluster " + i + " is at " + PERCENTILE[p]); 
                            Rectangle r = getBlock (i);
                            int x = r.x + r.width/2;
                            if (i > 0 && lastPercentileX - x > percentileGap) {
                                g.drawLine (x, visible.y, x, visible.y + visible.height);
                                drawStringRightJustified (g, 
                                                          percentFmt.format (PERCENTILE[p]),
                                                          x-1, 
                                                          visible.y + th,
                                                          fm);
                                
                                //                                 double d = clusters[i].getDistance ();
                                //                                 g.drawString (distFmt.format (d),
                                //                                               x+1, heightBelow);
                                
                                lastPercentileX = x;
                            }
                            do {
                                if (--p < 0)
                                    break percentileLoop;
                                else 
                                    nPercentile = (int) (nPositiveExamples * (1 - PERCENTILE[p]));
                                debug.println ("a" + nDrawn + " drawn," + p + "(" + nPercentile +")");
                            } while (nPercentile <= nDrawn);
                        }
                        
                        nDrawn += clusters[i].getSize ();
                    }
                }
            }

            //             // draw lines at integral standard deviations
            //             if (stddev > 0) {
            //                 g.setColor (Color.lightGray);
            //                 double sigma = scale * stddev;
            //                 for (double d = sigma; d < width; d += sigma)
            //                     g.drawLine ((int) d, -heightAbove, (int) d, heightBelow);
            //             }

            // draw midline across box
            g.setColor (Color.black);
            g.drawLine (0, 0, width, 0);

            // draw blocks
            if (clusters != null) {
                for (int i = 0; i < clusters.length; ++i) {
                    Rectangle b = getBlock (i);

                    boolean paintSelected = selected[i];
                    if (rubberBand != null) {
                        if (incremental && b.contains (rubberBand))
                            paintSelected = !paintSelected;
                        else if (rubberBand.intersects (b))
                            paintSelected = true;
                    }
                    paintBlock (g, 
                                b,
                                clusters[i].isPositive (),
                                paintSelected);
                }
            }

            // draw rubber-band rectangle
            g.setColor (Color.black);
            if (rubberBand != null)
                g.drawRect (rubberBand.x, rubberBand.y,
                            rubberBand.width, rubberBand.height);
        }

        public void centerCluster (int i) {
            if (i < 0 || i >= clusters.length)
                return;

            Rectangle b = getBlockInComponentCoords (i);
            Dimension d = Histogram.this.getPreferredSize ();
            Rectangle vr = Histogram.this.getVisibleRect ();
            debug.println ("b=" + b + ", d=" + d + ", vr=" + vr);
            vr.x = b.x - vr.width/2;
            vr.y = (d.height - vr.height)/2;
            Histogram.this.scrollRectToVisible (vr);
        }

        public void scrollClusterToVisible (int i) {
            if (i < 0 || i >= clusters.length)
                return;

            Rectangle b = getBlockInComponentCoords (i);
            Histogram.this.scrollRectToVisible (b);
        }
    }

    class Block extends JComponent {
        boolean isPositive;

        public Block (boolean isPositive) {
            this.isPositive = isPositive;
        }

        public Dimension getPreferredSize () {
            Insets insets = Block.this.getInsets ();
            int dx = insets.left + insets.right;
            int dy = insets.top + insets.bottom;

            return new Dimension (dx + BLOCK_WIDTH, dy + BLOCK_HEIGHT);
        }

        public Dimension getMinimumSize () {
            return getPreferredSize ();
        }

        public void paintComponent (Graphics g) {
            super.paintComponent (g);

            Insets insets = Block.this.getInsets ();
            int x = insets.left;
            int y = insets.top;
            paintBlock (g, new Rectangle (x, y, BLOCK_WIDTH, BLOCK_HEIGHT),
                          isPositive, false);
        }
    }

    // drawing routines

    static void drawStringRightJustified (Graphics g, String s, int w, int y,
                                          FontMetrics fm) {
        g.drawString (s, w - fm.stringWidth (s), y);
    }
    
    static void drawStringCentered (Graphics g, String s, int x, int y,
                                    FontMetrics fm) {
        g.drawString (s, x - fm.stringWidth (s)/2, y);
    }
    
    static void paintBlock (Graphics g, Rectangle b, 
                              boolean isPos, boolean isSelected) {
            if (isSelected)
                g.setColor (selectedColor);
            else if (isPos)
                g.setColor (positiveColor);
            else
                g.setColor (negativeColor);
            
            g.fillRect (b.x, b.y, b.width, b.height);
            
            g.setColor (Color.black);
            for (int d = 0, e = b.height - 1; d < e; d += BLOCK_HEIGHT-1)
                g.drawRect (b.x, b.y + d, b.width-1, BLOCK_HEIGHT-1);

            if (isSelected) {
                // darken outline a little more
                g.drawRect (b.x-1, b.y-1, b.width+1, b.height+1);
                g.drawRect (b.x+1, b.y+1, b.width-3, b.height-3);
            }
    }

}
