/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Enumeration;

import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JList;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.text.JTextComponent;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import lapis.*;
import lapis.Main;
import lapis.Pattern;
import lapis.PatternLibrary;
import lapis.RegionSet;
import lapis.NamedRegionSet;
import lapis.tc.Evaluator;
import lapis.tc.TC;
import lapis.tc.TCParseException;
import lapis.tc.TCToken;
import lapisx.swing.AcceleratedAction;
import lapisx.swing.ActionAdapters;
import lapisx.swing.FlashingLabel;
import lapisx.swing.HistoryComboBoxModel;
import lapisx.swing.InputThreeButtonDialog;
import lapisx.swing.Layout;

public class PatternPane extends JPanel implements ItemListener, LibraryTreeListener {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    Browser browser;

    HistoryComboBoxModel patternHistory; // history of patterns
    TCComboBox patternComboBox;
    // combo box for selecting from pattern history
    TCEditor pattern; // TC pattern to select
    FlashingLabel patternMatchCount; // displays # of matches
	JButton nameButton;
	JButton deleteNameButton;
	
    LibraryTree tree;
    JScrollPane treeScrollPane;

    // grouping components
    JPanel patternPanel;
    JPanel treePanel;
    JSplitPane splitter;

    // actions
    Find findAction;
    Action clearPatternAction;

    ////////////////////////////////////////
    // constraints
    static JCheckBox caseCheckBox;
    static JCheckBox wordCheckBox;
    static boolean caseSensitive;
    static boolean wholeWord;
    ////////////////////////////////////////

    public PatternPane(Browser browser) {
	this.browser = browser;

	////////////////////////////////////////
	caseCheckBox = new JCheckBox("Match Case");
	caseCheckBox.setSelected(false);
	caseCheckBox.addItemListener(this);
	wordCheckBox = new JCheckBox("Match Whole Word(s)");
	wordCheckBox.setSelected(false);
	wordCheckBox.addItemListener(this);
	caseSensitive = false;
	wholeWord = false;
	////////////////////////////////////////

	// Create the pattern box
	int patHistorySize =
	    Integer.parseInt(
			     Main.getConfig().getProperty("browser.pattern-history-size"));
	patternHistory = new HistoryComboBoxModel(patHistorySize);
	patternComboBox = new TCComboBox(patternHistory, true, true);
	pattern = (TCEditor) patternComboBox.getEditor();

	findAction = new Find(browser, pattern);
	clearPatternAction = new ClearPattern(pattern);
	pattern.addKeyListener(findAction);

	patternPanel =
	    (JPanel) Layout
	    .vertical()
	    .setAlignment(Layout.WEST)
	    .setContainerInsets(5, 5, 5, 5)
	    .add(
		 Layout
		 .horizontal()
		 .setAlignment(Layout.SOUTH)
		 .add(new JLabel("Pattern: "))
		 .addGlue()
		 .add(ActionAdapters.toTinyButton(findAction))
		 .add(ActionAdapters.toTinyButton(clearPatternAction))
		 .done(),
		 Layout.FIXED_HEIGHT)
	    .add(patternComboBox)
	    .add(Layout.horizontal().setAlignment(Layout.SOUTH)
		 //.add (caseCheckBox = ActionAdapters.toCheckBox (caseAction))
		 .add(caseCheckBox).addGlue()
		 //.add (wordCheckBox = ActionAdapters.toCheckBox (wordAction))
		 .add(wordCheckBox).done(),
		 Layout
		 .FIXED_HEIGHT)
	    .add(
		 patternMatchCount = new FlashingLabel("0 regions selected"),
		 Layout.FIXED_HEIGHT)
	    .done();
	patternPanel.setMinimumSize(new Dimension(0, 0));
	//         patternPanel.setPreferredSize (new Dimension (300, 160));

	// Create the label tree and its decorations
	tree = new LibraryTree(browser);
 	tree.addLibraryTreeListener(this);
	browser.addBrowserListener(tree);
	treeScrollPane = new JScrollPane(tree);
	treeScrollPane.setMinimumSize(new Dimension(0, 0));

	treePanel =
	    (JPanel) Layout
	    .vertical()
	    .setAlignment(Layout.WEST)
	    .setContainerInsets(5, 5, 5, 5)
	    .add(
		 Layout
		 .horizontal()
		 .setAlignment(Layout.SOUTH)
		 .add(new JLabel("Named Patterns:"))
		 .addGlue()
		 .add(nameButton = ActionAdapters.toTinyButton(nameAction))
		 .add(
		      deleteNameButton =
		      ActionAdapters.toTinyButton(deleteNameAction))
		 .done(),
		 Layout.FIXED_HEIGHT)
	    .add(treeScrollPane)
	    .done();
    
    deleteNameButton.setEnabled(false);
	if (SwingConfig.inDemoMode())
	    deleteNameButton.setEnabled(false);

	treePanel.setMinimumSize(new Dimension(0, 0));

	splitter =
	    new JSplitPane(JSplitPane.VERTICAL_SPLIT, patternPanel, treePanel);
	splitter.setOneTouchExpandable(true);
	splitter.setContinuousLayout(false);
	splitter.setResizeWeight(0.5);

	setLayout(new BorderLayout());
	add(splitter, BorderLayout.CENTER);
    }

    public void saveWindowPositions() {
	SwingConfig.saveSplitter("browser.treeSplitter", splitter);
    }

    public void restoreWindowPositions() {
	SwingConfig.restoreSplitter("browser.treeSplitter", splitter);
    }

    /**
     * Set number of matches displayed in the pattern pane.
     * @param n number of matches
     * @effects displays "n regions selected" in the pattern pane
     */
    public void setMatchCount(int n) {
	patternMatchCount.setText(
				  n + " region" + ((n == 1) ? "" : "s") + " selected");
    }

    /**
     * Get last pattern entered in pattern box.
     * @return contents of pattern box as a string
     */
    public String getPattern() {
	return pattern.getText();
    }

    /**
     * Action that searches for a text constraint pattern and
     * selects the matching regions.
     */
    class Find extends AcceleratedAction implements KeyListener {
	// keep this action from firing recursively
	boolean blockFindAction = false;
	Browser browser;
	TCEditor pattern;

	public Find(Browser browser, TCEditor pattern) {
	    super("Go", "Select all matches to the pattern");
	    this.browser = browser;
	    this.pattern = pattern;
	}

	public void actionPerformed(ActionEvent evt) {
	    try { // finally
		if(blockFindAction)
		    return;
		blockFindAction = true;
		attemptToFind();
	    } catch (TCParseException e) {
		debug.println("parse errors: " + e.getMessage());

		FeedbackForm.showMessageDialog(
					       browser,
					       e.errors().nextElement().toString(),
					       "Parse Error",
					       JOptionPane.ERROR_MESSAGE);
	    } finally {
		blockFindAction = false;
	    }
	}


	public void keyPressed(KeyEvent e) {
	    if (e.getKeyCode() == KeyEvent.VK_ENTER
		&& !e.isShiftDown()
		&& !e.isControlDown()
		&& !e.isMetaDown()
		&& !e.isAltDown()) {
		    
		actionPerformed(new ActionEvent(e.getSource(), 0, null));
		e.consume();
	    }
	}

	public void keyReleased(KeyEvent e) {
	}
	    
	public void keyTyped(KeyEvent e) {
	}

    }

    private void attemptToFind() throws TCParseException {
		
	//                 // get pattern expression
	//                 ComboBoxEditor ed = pattern.getEditor ();
	//                 String expr = ed.getItem ().toString ();
	//                 if (expr.length () == 0)
	//                     return;
		
	//                 // add it to pattern history
	//                 History history = (History) pattern.getModel ();
	//                 if (!history.contains (expr))
	//                     history.add (expr);
		
	debug.println("find pattern:\n" + pattern.getText().trim());
		
	// compile pattern into a TC structure
		
	TC system = makeTCFromTextComponent(pattern, true);
		
	if ( caseSensitive ) {
	    system = new TC(new TC.CaseSensitivity(caseSensitive, system.getRoot()));
	}
	if ( wholeWord ) {
	    system = new TC(new TC.And(new TC.Ending(new TC.Id("word")),
				       new TC.And(new TC.Starting(new TC.Id("word")),
						  system.getRoot())));
	}
		
	/*will want to add a try catch here that catches if pattern already exists and overwrite if needed
	  probably catch a thrown exception and then force overwrite
	  or can do error check here (does pattern already exist) except we don't know it's name
	  E2E suggests that we check if second word is "is" and do everything at this level. 
	  ^ won't work, ie: pronoun and you is you
	  try catch seems most reasonable either that or a parsing check like will make binding occur? 
	*/
		
	// pattern compiled OK, so add it to pattern history
	patternHistory.add(pattern.getText());

	LibraryLoader.checkForLog(system);
		
	// evaluate pattern in background
	find(system);
    }

    /**
     * Set selection to the result of a pattern match.
     * @param tc TC pattern to search for
     * @effects searches for TC pattern in active pane and sets selection
     * to the matching region set
     */
    public void find(final TC pat) {
	browser.run(new RunnableWithExceptions() {
		public void run() {
		    SwingDocument doc = browser.getDocument();
		    if (doc == null)
			return;

		    try {
			RegionSet result = ((TC) pat).matchWithWarnings(doc);
			//final NamedRegionSet sel = new NamedRegionSet(doc, result, pat);
			final NamedRegionSet sel = new NamedRegionSet(result, pat);
			debug.println(
				      "TC pattern of Pattern Pane: " + pat.toString());
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
				    browser.clearSelection();
				    browser.setSelection(sel);
				    browser.getActivePane().scrollRegionsToVisible(sel);

				    if (sel.isEmpty())
					patternMatchCount.startFlashing ();
				}
			    });
		    } catch (final TCParseException warnings) {
			
			final List ambiguities = new LinkedList();  //holds generated AmbiguousErrors
			Enumeration errors = warnings.errors();

			//find all AmbiguousErrors and put them in ambiguities
			while(errors.hasMoreElements()) {    
			    TCParseException.ParseError tempError = (TCParseException.ParseError) errors.nextElement();
			    if(tempError instanceof Evaluator.AmbiguousError) 
				ambiguities.add(tempError);
			}

			//No AmbiguousErrors
			if(ambiguities.size()==0) {
			    SwingUtilities.invokeLater(new Runnable() {
				    public void run() {
					browser.clearSelection();
					FeedbackForm.showMessageDialog(
								       browser,
								       warnings.getMessage(),
								       "Warning",
								       JOptionPane.ERROR_MESSAGE);
				    }
				});
			}

			//if there are AmbiguousErrors, bring up ambiguity dialog
			else {
			    Collections.sort(ambiguities, new AmbiguityComparator());
			    showAmbiguityDialog(pat, ambiguities);
			}	
		    }
		}
	    });
    }
	private static class AmbiguityComparator implements java.util.Comparator {
	/**
	 *@requires o1, o2 are CollisionEvents
	 *@return -1 if time1<time2, 0 if time1==time2, 1 if time1>time2
	 */
	public int compare(Object o1, Object o2) {
	    Evaluator.AmbiguousError c1 = (Evaluator.AmbiguousError)o1;
	    Evaluator.AmbiguousError c2 = (Evaluator.AmbiguousError)o2;

	    if (c1.getToken().start < c2.getToken().start) {
		return -1;
	    }
	    else if (c1.getToken().start > c2.getToken().start) {
		return +1;
	    }
	    else {
		return 0;
	    }
	}
	}

    private void showAmbiguityDialog(final TC pat, final List ambiguities) {
	int length = ambiguities.size();
	final Object[] ambiguousOptions = new Object[length+1];  //Array of option menus; one for each ambiguity plus error message
	final String patternText = pat.getExpression();
	TCToken errorToken;
	String errorMessage;
	Evaluator.AmbiguousError aError;

	//construct errorMessage (stored in ambiguousOptions[0])
	switch(length) {
	case 0:
	    break;
	case 1:
	    aError=(Evaluator.AmbiguousError) ambiguities.get(0);
	    errorToken=aError.getToken();
	    ambiguousOptions[0]="pattern name "+patternText.substring(errorToken.start, errorToken.end).toUpperCase()+" is ambiguous.\nPlease make a selection:";
	    ambiguousOptions[1]=new JList(aError.getChoices());
	    break;
        case 2:
	    errorMessage="pattern names ";
	    aError=(Evaluator.AmbiguousError) ambiguities.get(0);
	    errorToken=aError.getToken();
	    errorMessage+=patternText.substring(errorToken.start, errorToken.end).toUpperCase()+" and ";
	    ambiguousOptions[1]=new JList(aError.getChoices());
	    aError=(Evaluator.AmbiguousError) ambiguities.get(1);
	    errorToken=aError.getToken();
	    errorMessage+=patternText.substring(errorToken.start, errorToken.end).toUpperCase()+" are ambiguous.\nPlease make clarifying selections:";
	    ambiguousOptions[2]=new JList(aError.getChoices());
	    ambiguousOptions[0]=errorMessage;
	    break;
	default:
	    errorMessage="pattern names ";
	    for(int i=0; i<length-1; i++) {
		aError=(Evaluator.AmbiguousError) ambiguities.get(i);
		errorToken=aError.getToken();
		errorMessage+=patternText.substring(errorToken.start, errorToken.end).toUpperCase()+", ";
		ambiguousOptions[i+1]=new JList(aError.getChoices());
	    }
	    aError=(Evaluator.AmbiguousError) ambiguities.get(length-1);
	    errorToken=aError.getToken();
	    errorMessage+="and "+patternText.substring(errorToken.start, errorToken.end).toUpperCase()+" are ambiguous.\nPlease make clarifying selections:";
	    ambiguousOptions[0]=errorMessage;
	    ambiguousOptions[length]=new JList(aError.getChoices());
	}

	//create buttons: important for disable and enable
	final JButton okButton = new JButton("OK");
	final JButton cancelButton = new JButton("Cancel");
	final JButton feedbackButton = new JButton("Send Feedback");
	final Object[] buttons = new Object[3];
	buttons[0]=okButton;
	buttons[1]=cancelButton;
	buttons[2]=feedbackButton;
	okButton.setEnabled(false);

	//add listSelectionListener to all lists for changes in selections
	ListSelectionListener listener = new ListenerForLists(okButton, ambiguousOptions);
	for(int i=1;i<length+1;i++) {
	    JList tempList = (JList) ambiguousOptions[i];
	    tempList.addListSelectionListener(listener);
	    tempList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	}

	//run dialog
	SwingUtilities.invokeLater(new Runnable() {
		public void run() {
		    browser.clearSelection();
		    String[] options = {"OK", "Cancel", "Send Feedback"};
		    JOptionPane pane = new JOptionPane (ambiguousOptions,
							JOptionPane.ERROR_MESSAGE,
							JOptionPane.YES_NO_CANCEL_OPTION,
							null,
							buttons,
							buttons[0]);
		    pane.setInitialValue(buttons[0]);
		    
		    JDialog dialog = pane.createDialog(browser, "Warning");
		    pane.selectInitialValue();
		    okButton.addActionListener(new OKClick(dialog, ambiguities, ambiguousOptions, patternText));
		    cancelButton.addActionListener(new CancelClick(dialog));
		    feedbackButton.addActionListener(new FeedbackClick(dialog));
		    if(SwingConfig.inDemoMode())
			new FeedbackForm.DialogTimeout(dialog,60).start();
		    dialog.show();
		}});}

    //user clicks feedback button
    private class FeedbackClick implements ActionListener {
	JDialog dialog;

	public FeedbackClick(JDialog dialog) {
	    this.dialog=dialog;
	}
	public void actionPerformed(ActionEvent e) {
	    dialog.dispose();
	    FeedbackForm.showFeedbackForm(browser);
	}
    }

    //user clicks cancel button
    private class CancelClick implements ActionListener {
	JDialog dialog;

	public CancelClick(JDialog dialog) {
	    this.dialog=dialog;
	}
	public void actionPerformed(ActionEvent e) {
	    dialog.dispose();
	}
    }

    //user clicks OK button
    private class OKClick implements ActionListener {
	JDialog dialog;
	List ambiguities;
	Object[] ambiguousOptions;
	String patternText;

	public OKClick(JDialog dialog, List ambiguities, Object[] ambiguousOptions, String patternText) {
	    this.dialog=dialog;
	    this.ambiguities=ambiguities;
	    this.ambiguousOptions=ambiguousOptions;
	    this.patternText=patternText;
	}
	
	public void actionPerformed(ActionEvent e) {
	    dialog.dispose();
	    String replacementMessage = "";
	    int startIndex = 0;
	    for(int i=0; i<ambiguities.size(); i++) {
		TCToken token=((Evaluator.AmbiguousError) ambiguities.get(i)).tok;
		replacementMessage+=patternText.substring(startIndex, token.start);
		replacementMessage+=(String) ((JList) ambiguousOptions[i+1]).getSelectedValue(); 
		startIndex=token.end;
	    }
	    replacementMessage+=patternText.substring(startIndex, patternText.length());
	    
	    pattern.setText(replacementMessage);
	    try {
		attemptToFind();
	    } catch(TCParseException err) {
		debug.println("parse errors: " + err.getMessage());
		
		FeedbackForm.showMessageDialog(browser,
					       err.errors().nextElement().toString(),
					       "Parse Error",
					       JOptionPane.ERROR_MESSAGE);
	    }
	}
    }

    //enables button when each JList in options has valid selection
    private static class ListenerForLists implements ListSelectionListener{
	JButton button;
	Object[] options;
	
	public ListenerForLists(JButton button, Object[] options) {
	    this.button=button;
	    this.options=options;
	}
	
	public void valueChanged(ListSelectionEvent e) {
	    int i = 1;
	    boolean valid=true;
	    while(i<options.length) {
		if (((JList) options[i]).getSelectedIndex()==-1) {
		    valid=false;
		    break;
		}
		i++;
	    }
	    if(!valid)
		button.setEnabled(false);
	    else button.setEnabled(true);
	}
    }

    
			    
    /**
     * Action invoked when user clicks on Clear button.
     * Clears the pattern box.
     */

    static class ClearPattern extends AcceleratedAction {
	TCEditor pattern;

	public ClearPattern(TCEditor pattern) {
	    super("Clear", "Clear the Pattern box");
	    this.pattern = pattern;
	    /*
	      if ( caseSensitive ) {
	      caseCheckBox.doClick();
	      }
	      if ( wholeWord ) {
	      wordCheckBox.doClick();
	      }
	    */
	    /*
	      caseSensitive = false;
	      wholeWord = false;
	      caseCheckBox.setSelected( false );
	      wordCheckBox.setSelected( false );
	    */
	}

	public void actionPerformed(ActionEvent evt) {
	    //             setPattern ("", false);
	    pattern.setText("");
	    caseSensitive = false;
	    wholeWord = false;
	    caseCheckBox.setSelected(false);
	    wordCheckBox.setSelected(false);
	    pattern.requestFocus();
	}
    }

    /**
     * Set pattern in pattern box.
     * @param pat String that should appear in pattern box.
     * @param addToHistory true if pattern should be added to the
     * end of the pattern history too.
     * @effects sets content of pattern box to <em>pat</em>, and
     * if <em>addToHistory</em> is true, also adds it to the drop-down
     * history.
     */
    public void setPattern(String pat, boolean addToHistory) {
	pattern.setText(pat);
	if (addToHistory)
	    patternHistory.add(pat);
    }

    /**
     * Action invoked when user selects an earlier pattern from the Find drop-down list.
     */
    //     public ItemListener findHistoryListener = new ItemListener () {
    //         public void itemStateChanged(ItemEvent e) {
    //             Object item = pattern.getSelectedItem ();
    //             if (item == null)
    //                 item = "";                
    //             pattern.getEditor ().setItem (item);
    //         }
    //     };

    public static TC makeTCFromTextComponent(
					     JTextComponent textbox,
					     boolean justOneExpression)
	throws TCParseException {
	String text = textbox.getText();
	try {
	    return new TC(text, justOneExpression);
	} catch (TCParseException e) {
	    // Select the error, then re-throw the exception
	    TCEditor.selectParseError(textbox, e);
	    e.setStringContext(text);
	    throw e;
	}
    }

    /**
     * Append a name to the pattern box (usually because it was selected from
     * a popup menu or the label tree).
     */
    /*
      protected void putNameInPattern (Label name) {
      ComboBoxEditor editor = pattern.getEditor ();
      String item = (String)editor.getItem ();
      if (item.length () != 0)
      item += ' ';
      item += name;
      editor.setItem (item);
      }
    */

    /**
     * Action invoked when user clicks on Name... button.
     * Pops up a dialog to assign a label to current selection.
     */
    Action nameAction =
	new AcceleratedAction(
			      "Add...",
			      "Give a name to the current selection") {
	    public void actionPerformed(ActionEvent evt) {
		String name = promptForName();
		if (name == null)
		    return;

		debug.println("named " + name);

		PatternLibrary lib = PatternLibrary.getDocumentLibrary (browser.getDocument ());
		NamedRegionSet sel = browser.getSelection();
		Pattern pat = sel.getPattern();

		//check to make sure name doesn't already exist before adding
		if (!lib.contains(name)) {
		    //	lib.put (name, pat);
		}
		/* else prompt user to change or overwrite
		// 2= confirm overwrite
		// 3 = cancel
		*/
		else {
		    OverwriteOptions overwriteOption = new OverwriteOptions(name);
		    while (lib.contains(name)) {
			overwriteOption = confirmOverwrite(name);
			//user chose to overwrite
			if (overwriteOption.option == 2)
			    break;
			//user chose to cancel or didn't type in anything
			if (overwriteOption.option == 3
			    || overwriteOption.text == null)
			    return;
			else
			    name = overwriteOption.text;

		    }

		}

		if (pat == null) {
		    pat = new TC(sel);
		    PatternLibrary.getDocumentLibrary(browser.getDocument()).put(name, pat);
		}
		else {
		    if (pat instanceof TC) 
			pat = ((TC) pat).toBoundExpression(lib);
		    PatternLibrary.getGlobalLibrary ().put(name, pat);
		    LibraryLoader.addToLog(name, pat);
		}
		browser.setSelection(new NamedRegionSet(sel, TC.make(name)));
	    }
	};

    String promptForName() {
/*		JOptionPane pane = new JOptionPane (ambiguousOptions,
								  JOptionPane.ERROR_MESSAGE,
								  JOptionPane.YES_NO_CANCEL_OPTION,
								  null,
								  buttons,
								  buttons[0]);
  */
        JTextField name = new JTextField();
        final JTextArea warning = new JTextArea();
        warning.setOpaque(false);
        name.addKeyListener(new KeyAdapter() {
			public void keyPressed(KeyEvent event) {
			    if (event.getKeyCode() == KeyEvent.VK_SPACE)
			        warning.setText("Spaces are not allowed in pattern names");
			    else if (event.getKeyCode() == KeyEvent.VK_TAB)
				    warning.setText("Spaces are not allowed in pattern names");
				else warning.setText("");
			}
        });
        Object[] objects = new Object[3];
        objects[0]="Name for current selection: ";
        objects[1]=name;
        objects[2]=warning;
        JOptionPane pane = new JOptionPane(objects,
                                    JOptionPane.PLAIN_MESSAGE,
                                    JOptionPane.OK_CANCEL_OPTION,
                                    null);
        JDialog dialog = pane.createDialog(browser, "Name This Selection");
        dialog.show();
        Object blah = pane.getOptions();
        Object selectedValue = pane.getValue();
        
        if (selectedValue == null)
            return null;

        if (selectedValue instanceof Integer)
            if(((Integer)selectedValue).intValue() == 0) {
                String answer = name.getText();
                if(answer == null)
                    return null;
                answer = answer.trim();
                if(answer.length()==0)
                    return null;
                return answer;
            }
            
        return null;
    
	//         try {
	//             return getDocument ().getLabeling ().lookup (answer);
	//         } catch (Labeling.LabelNotFoundException e) {
	//             return answer;
	//         } catch (Labeling.LabelAmbiguousException e) {
	//             return (String)
	//                 JOptionPane.showInputDialog (null,
	//                                              "Several labels match that name.\n"
	//                                              + "Please select the one you want.",
	//                                              "Choose Label",
	//                                              JOptionPane.PLAIN_MESSAGE,
	//                                              null,
	//                                              e.getChoices (),
	//                                              null);
	//         }
    }

    /* Wrapper class of int and string
     */
    static class OverwriteOptions {
	int option;
	String text;

	public OverwriteOptions(String inText) {
	    option = 0;
	    text = inText;
	}
	public OverwriteOptions(int in, String inText) {
	    option = in;
	    text = inText;
	}
    }

    /* invoked when durring an Add Name, name chosen has already been input. 
       pops up a box asking user to confirm name overwrite or chose a new name
    */

    OverwriteOptions confirmOverwrite(String overwrite) {
	final String message =
	    new String(
		       "The name you have chosen: \n         "
		       + overwrite
		       + "\nis already in use. \nPlease chose a new name or confirm overwrite.");
	final String btnString1 = "OK";
	final String btnString2 = "Overwrite";
	final String btnString3 = "Cancel";
	Object[] options = { btnString1, btnString2, btnString3 };
	InputThreeButtonDialog dialog =
	    new InputThreeButtonDialog(
				       browser,
				       "Warning",
				       message,
				       overwrite,
				       options);

	dialog.pack();
	dialog.setVisible(true);
	int selection = dialog.getSelection();
	/* Ok = 1
	   Overwrite = 2
	   Cancel = 3 */

	if (selection == JOptionPane.YES_OPTION) {
	    return new OverwriteOptions(1, dialog.getText());
	} else if (selection == JOptionPane.NO_OPTION)
	    return new OverwriteOptions(2, overwrite);
	return new OverwriteOptions(3, null);

    }

    ////////////////////
    //////////////////// Deleting labels from pattern library.
    ////////////////////

    /**
     * Action invoked by Delete Label command.  Deletes selected label(s)
     * from pattern library.
     */
    Action deleteNameAction = new AcceleratedAction("Delete") {
	    public void actionPerformed(ActionEvent evt) {

		//String name = browser.getSelection().getName();
		//if (name != null) {
		//    browser.getPatternLibrary().remove(name);
		//    browser.clearSelection();
		//    LibraryLoader.unbindToLog(name);
		//}

		Set names = browser.getSelectionNames();
		//Set names = browser.getSelection().getNames();
		if (names != null) {
		    Iterator iter = names.iterator();
		    while (iter.hasNext()) {
			String tempName = (String) iter.next();
			browser.getPatternLibrary().remove(tempName);
			LibraryLoader.unbindToLog(tempName);
		    }
		    browser.clearSelection();
		}

	    }
	};

    ////////////////////////////////////////

    public void itemStateChanged(ItemEvent e) {
	Object source = e.getItemSelectable();

	if (source == caseCheckBox) {
	    if (e.getStateChange() == ItemEvent.DESELECTED) {
		caseSensitive = false;
		debug.println("case sensitive is now off");
		/*
		  JOptionPane.showMessageDialog (null, 
		  "alert", "case sensitive = off", 
		  JOptionPane.PLAIN_MESSAGE);
		*/
	    } else if (e.getStateChange() == ItemEvent.SELECTED) {
		caseSensitive = true;
		debug.println("case sensitive is now on");
		/*
		  JOptionPane.showMessageDialog (null,
		  "alert", "case sensitive = on",
		  JOptionPane.PLAIN_MESSAGE);
		*/
	    }
	} else if (source == wordCheckBox) {
	    if (e.getStateChange() == ItemEvent.DESELECTED) {
		wholeWord = false;
		debug.println("whole word is now off");
		/*
		  JOptionPane.showMessageDialog (null,
		  "alert", "whole word = off",
		  JOptionPane.PLAIN_MESSAGE);
		*/
	    } else if (e.getStateChange() == ItemEvent.SELECTED) {
		wholeWord = true;
		debug.println("whole word is now on");
		/*
		  JOptionPane.showMessageDialog (null,
		  "alert", "whole word = on",
		  JOptionPane.PLAIN_MESSAGE);
		*/
	    }
	}
    }
	public void treeNodeSelected() {
		deleteNameButton.setEnabled(true);
	}

	
	public void treeNodeNotSelected() {
		deleteNameButton.setEnabled(false);
	}

}
