/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import lapis.maps.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.net.*;
import java.io.*;
import java.util.*;
import lapisx.util.Str;

public class PlainDocument extends javax.swing.text.PlainDocument 
						   implements SwingDocument, DocumentView {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    Reader in;              
    // stream from which content should be read.  Null if content has already
    // been read.

    // loading state
    int state = NOT_LOADED;
    static final int NOT_LOADED = 0;
    static final int LOADING = 1;
    static final int LOADED = 2;

    PatternCache patternCache;

    DocumentFactory factory;

    DocumentVersion version;
    // current document version

    CoordinateMap changes;
    // changes since last call to getVersion ()
    
    // EditableView view;

    public PlainDocument (DocumentFactory factory, Reader in) {
        putProperty (MIMEProperty, "text/plain");
        this.factory = factory;
        this.in = in;
        //this.view = new EditableView(this);
    }

    public PlainDocument (DocumentFactory factory, lapis.Document doc,
                          String contentType) {
        this (factory, doc.getReader ());

        BasicDocument.copyMissingProperties (this, doc);
        if (contentType != null)
            putProperty (MIMEProperty, contentType);
    }

    private static PlainEditorKit defaultKit = new PlainEditorKit ();

    public String toString () {
        String cmd = (String) getProperty (CommandProperty);
        if (cmd != null && cmd.length () > 0)
            return Str.abbreviate (cmd, 200);

        URL url = (URL) getProperty (URLProperty);
        if (url != null)
            return url.toString ();

        return super.toString ();
    }

    public String getText () {
        if (!load (defaultKit))
            return "";

        try {
            return getText (0, getLength ());
        } catch (BadLocationException e) {
            debug.report (e);
            return "";
        }
    }

    public String getText (Region r) {
        if (!load (defaultKit))
            return "";

        try {
            int l = getLength ();
            int s = Math.min (l, Math.max (0, r.getStart ()));
            int e = Math.min (l, Math.max (0, r.getEnd ()));
            return getText (s, e - s);
        } catch (BadLocationException e) {
            debug.report (e);
            return "";
        }
    }

    public Region toRegion () {
        load (defaultKit);
        return new Region (0, getLength (), getVersion());
    }

    public Reader getReader () {
        return new StringReader (getText ());
    }

    public Dictionary getProperties () {
        return getDocumentProperties ();
    }

    public PatternCache getPatternCache () {
        if (patternCache == null)
            patternCache = new PatternCache ();
        return patternCache;
    }

    public DocumentFactory getFactory () {
        return factory;
    }

    public DocumentVersion getVersion () {
        if (version == null) {
            // set first version 
            version = DocumentVersion.getOriginal((DocumentView)this); //DocumentVersion.ORIGINAL;
            changes = IdentityMap.IDENTITY;
        } else if (changes != IdentityMap.IDENTITY) {
            // changes made since last version returned
            version = new DocumentVersion (changes, version, this);
            changes = IdentityMap.IDENTITY;
        }
        return version;
    }

    void setVersion (DocumentVersion ver) {
        this.version = ver;
        changes = IdentityMap.IDENTITY;
    }

    public boolean isLoaded () {
        return state == LOADED;
    }

    public boolean load (EditorKit kit) {
        if (in == null) {
            state = LOADED;
            return true;
        }

        synchronized (this) {
            if (state == LOADING)
                return false;
            state = LOADING;
        }

        // Fetch the document content, simultaneously
        // constructing the view
        try {
            try {
                kit.read(in, this, 0);
            } catch (BadLocationException e) {
                throw new IOException (e.toString ());
            }
            in.close ();

            addDocumentListener (new DocumentListener () {
                    public void changedUpdate (DocumentEvent e) {
			debug.println ("changedUpdate " + e);
                        documentChanged (IdentityMap.IDENTITY);
                    }
                    public void insertUpdate (DocumentEvent e) {
			debug.println ("insertUpdate " + e);
                        int start = e.getOffset ();
                        int end = start + e.getLength ();
                        documentChanged (new EditCoordinateMap.Insertion (start, start, end));
                    }
                    public void removeUpdate (DocumentEvent e) {
			debug.println ("removeUpdate " + e);
                        int start = e.getOffset ();
                        int end = start + e.getLength ();
                        documentChanged (new EditCoordinateMap.Deletion (start, end, start));
                    }
                });
	    
        } catch (IOException e) {
            debug.report (e);
        } finally {
            synchronized (this) {
                state = LOADED;
                in = null;
            }
        }

        return true;
    }

    protected void documentChanged (CoordinateMap change) {
        if (patternCache != null)
            patternCache.clear ();

        if (this.changes != null)
            // somebody is interested in versions, so keep track
            this.changes = new CompositeCoordinateMap (change, this.changes); 
    }

 	/*
	public class EditableView extends DocumentView {
		CoordinateMap changes = IdentityMap.IDENTITY;
		// changes since last call to getVersion ()

		public EditableView (lapis.Document doc){
			super (doc);
		}
		
		public String getText () {
			if (!load (defaultKit))
				return "";

			try {
				return ((javax.swing.text.PlainDocument)getDocument())
						.getText (0, getLength ());
			} catch (BadLocationException e) {
				debug.report (e);
				return "";
			}
		}

		public String getText (Region r) {
			if (!load (defaultKit))
				return "";

			try {
				int l = getLength ();
				int s = Math.min (l, Math.max (0, r.getStart ()));
				int e = Math.min (l, Math.max (0, r.getEnd ()));
				return ((javax.swing.text.PlainDocument)getDocument())
						.getText (s, e - s);
			} catch (BadLocationException e) {
				debug.report (e);
				return "";
			}
		}

		public Region toRegion () {
			load (defaultKit);
			return new Region (0, getLength (), this.getVersion());
		}
		
		public DocumentVersion getVersion () {
			//if (this.version == null) {
				// set first version 
			//	this.version = DocumentVersion.getOriginal(this); //DocumentVersion.ORIGINAL;
			//	changes = IdentityMap.IDENTITY;
			//} else if (changes != IdentityMap.IDENTITY) {
				// changes made since last version returned
			//	this.version = new DocumentVersion (changes, version, this);
			//	changes = IdentityMap.IDENTITY;
			//}
			//return version;
			
			if (changes != IdentityMap.IDENTITY) {
				// changes made since last version returned
				this.version = new DocumentVersion (changes, version, this);
				changes = IdentityMap.IDENTITY;
			}
			return this.version;
		}

		void setVersion (DocumentVersion ver) {
			this.version = ver;
			changes = IdentityMap.IDENTITY;
		}
	}
	*/
		
	public DocumentView getDefaultView () {
		return this;
	}
	
	public DocumentView getCookedView () {
		return this;
	}
	
	public lapis.Document getDocument () {
		return this;
	}
}
