/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import javax.swing.*;
import javax.swing.text.*;
import java.io.*;

public class PlainEditorKit extends DefaultEditorKit {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;
  
    private ViewFactory viewFactory = new ColoringContext();

    private static EditAction.ActionMap map;

    public Action[] getActions () {
        if (map == null)
            map = new EditAction.ActionMap (super.getActions ());
        return map.getActions ();
    }


    /**
     * Creates an uninitialized text storage model
     * that is appropriate for this type of editor.
     *
     * @return the model
     */
    public javax.swing.text.Document createDefaultDocument() {
        return new lapis.swing.PlainDocument (SwingDocument.defaultFactory,
                                              null);
    }

    /**
     * Clone the editor kit.  Must always implement this method,
     * because Swing 1.1's DefaultEditorKit.clone() uses new to
     * create the cloned editor kit, rather than calling Object.clone()
     */
    public Object clone() {
    	return new PlainEditorKit ();
    }

	/**
	 * Fetches a factory that is suitable for producing
	 * views of any models that are produced by this
	 * kit.  The default is to have the UI produce the
	 * factory, so this method has no implementation.
	 *
	 * @return the view factory
	 */
    public ViewFactory getViewFactory() {
        return viewFactory;
    }


  //
  //   /**
  //	 * Creates a view from the given structural element of a
  //	 * document.
  //	 *
  //	 * @param elem  the piece of the document to build a view of
  //	 * @return the view
  //	 * @see View
  //	 */
  // public View create(Element elem) {
  //     return new WrappedPlainView(elem);
  // }
  
    public void install (JEditorPane c) {
        super.install (c);
        if (map == null)
            map = new EditAction.ActionMap (super.getActions ());
        map.fixKeymap (c.getKeymap ());
    }

    /**
     * Inserts content from the given stream, which will be 
     * treated as plain text.
     * 
     * @param in  The stream to read from
     * @param doc The destination for the insertion.
     * @param pos The location in the document to place the
     *   content >= 0.
     * @exception IOException on any I/O error
     * @exception BadLocationException if pos represents an invalid
     *   location within the document.
     */
    public void read(Reader in, javax.swing.text.Document doc, int pos) 
        throws IOException, BadLocationException {

        char[] raw = new char[4096]; // data read from Reader 
        char[] cooked = new char[4096]; // after \r\n --> \n conversion
        int n;
	boolean lastWasCR = false;
	boolean isCRLF = false;

        while ((n = in.read (raw, 0, raw.length)) != -1) {
            int j = 0;
	    for(int i = 0; i < n; i++) {
		switch(raw[i]) {
		case '\r':
		    if (lastWasCR)
                        cooked[j++] = '\r';
                    lastWasCR = true;
		    break;
		case '\n':
                    cooked[j++] = '\n';
		    if (lastWasCR) {
			lastWasCR = false;
			isCRLF = true;
		    }
		    break;
		default:
		    if (lastWasCR) {
                        cooked[j++] = '\r';
                        lastWasCR = false;
                    }
                    cooked[j++] = raw[i];
		    break;
		}
	    }

            if (j > 0) {
                String s = new String (cooked, 0, j);
//                 debug.println ("read: " + s);
                doc.insertString (pos, s, null);
                pos += j;
            }
        }

	if (lastWasCR)
	    doc.insertString (pos, "\r", null);

        doc.putProperty(EndOfLineStringProperty, isCRLF ? "\r\n" : null);
    }

}
