/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import lapisx.util.*;
import lapisx.swing.*;
import lapisx.enum.*;

public class PreferencesDialog extends OkCancelDialog {
    Browser browser;
    JTabbedPane tabs;

    // Startup tab
    JTextField homepage;
    JTextField initscript;

    // Network tab
    JRadioButton noProxyButton, proxyButton;
    JTextField proxyhost;
    JTextField proxyport;
    JLabel proxyhostLabel;
    JLabel proxyportLabel;

    public PreferencesDialog (Browser browser) {
        super (browser, "Preferences", true);
        this.browser = browser;

        Layout.vertical (getControlsPane ())
            .add (tabs = new JTabbedPane ());

        //
        // Startup tab
        //

        tabs.addTab ("Startup",
                     Layout.gridbag ()
                     .setContainerInsets (10, 10, 10, 10)
                     .setAlignment (Layout.WEST)
                     .setSize (Layout.FIXED_SIZE)
                     .add (new JLabel ("Home page: "))
                     .add (homepage = new JTextField (get ("browser.homepage"),
                                                      30), 
                           Layout.FIXED_HEIGHT)
                     .nextRow ()
                     .add (new JLabel ("Startup script: "))
                     .add (initscript = new JTextField (get ("tcl.personal-init"),
                                                        30), 
                           Layout.FIXED_HEIGHT)
                     .nextRow ()
                     .done ());

        //
        // Network tab
        //

        tabs.addTab ("Network",
                     Layout.gridbag ()
                     .setContainerInsets (10, 10, 10, 10)
                     .setAlignment (Layout.WEST) 
                     .setSize (Layout.FIXED_SIZE)
                     .add (noProxyButton = new JRadioButton ("Direct connection to Internet"))
                     .nextRow ()
                     .add (proxyButton = new JRadioButton ("HTTP proxy"))
                     .nextRow ()
                     .add (Layout.gridbag ()
                           .setAlignment (Layout.WEST)
                           .setSize (Layout.FIXED_SIZE)
                           .add (proxyhostLabel = new JLabel ("Host: "))
                           .add (proxyhost = new JTextField (), Layout.FIXED_HEIGHT)
                           .nextRow ()
                           .add (proxyportLabel = new JLabel ("Port: "))
                           .add (proxyport = new JTextField (5))
                           .nextRow ()
                           .done (),
                           Layout.ANY_SIZE,
                           new Insets (0, 30, 0, 0))
                     .nextRow ()
                     .done ());

        ButtonGroup proxyGroup = new ButtonGroup ();
        proxyGroup.add (noProxyButton);
        proxyGroup.add (proxyButton);
        ChangeListener disabler = new ChangeListener () {
                public void stateChanged (ChangeEvent e) {
                    boolean enabled = proxyButton.isSelected ();
                    proxyhost.setEnabled (enabled);
                    proxyport.setEnabled (enabled);
                    proxyhostLabel.setEnabled (enabled);
                    proxyportLabel.setEnabled (enabled);
                }
            };
        noProxyButton.addChangeListener (disabler);
        proxyButton.addChangeListener (disabler);

        String proxy = get ("http.proxy");
        if (proxy.length () == 0)
            noProxyButton.setSelected (true);
        else {
            proxyButton.setSelected (true);
            proxyhost.setText (Str.before (proxy, ":"));
            proxyport.setText (Str.after (proxy, ":"));
        }

        pack ();
    }

    public void ok () {
        // Startup tab
        put ("browser.homepage", homepage.getText ().trim ());
        put ("tcl.personal-init", initscript.getText ().trim ());

        // Network tab
        String proxy;
        if (noProxyButton.isSelected ())
            proxy = "";
        else {
            String host = proxyhost.getText ().trim ();
            String port = proxyport.getText ().trim ();
            if (host.length () == 0) {
                FeedbackForm.showMessageDialog (browser,
                                                "No hostname specified for HTTP proxy",
                                                "No proxy hostname",
                                                JOptionPane.ERROR_MESSAGE);
                proxyhost.requestFocus ();
                return;
            }
            if (port.length () == 0) {
                FeedbackForm.showMessageDialog (browser,
                                                "No port number specified for HTTP proxy",
                                                "No proxy port",
                                                JOptionPane.ERROR_MESSAGE);
                proxyport.requestFocus ();
                return;
            }

            try {
                Integer.parseInt (port);
            } catch (NumberFormatException e) {
                FeedbackForm.showMessageDialog (browser,
                                                "Port must be a number",
                                                "Proxy port is not a number",
                                                JOptionPane.ERROR_MESSAGE);
                proxyport.requestFocus ();
                return;
            }

            proxy = host + ":" + port;
        }
        if (put ("http.proxy", proxy))
            Main.setProxy (proxy);

        // Store changed configuration
        Config cfg = Main.getConfig ();
        cfg.save ();
        if (cfg.getLastException () != null) {
            FeedbackForm.showExceptionDialog (browser, cfg.getLastException ());
            return;
        }

        super.ok ();
    }

    static String get (String key) {
        String s = Main.getConfig ().getProperty (key);
        return (s != null) ? s : "";
    }

    static boolean put (String key, String value) {
        value = value.trim ();

        Config cfg = Main.getConfig ();
        String oldValue = cfg.getProperty (key);

        if (oldValue != null && oldValue.equals (value))
            return false;

        cfg.put (key, value);
        return true;
    }
    
    static Enumeration browsers () {
        return new FilteredEnumeration (Main.windows ()) {
                public void transform (Object obj) {
                    if (obj instanceof Browser)
                        yield (obj);
                }
            };
    }


    /**
     * Action that pops up preferences dialog.
     */
    public static class Preferences extends AcceleratedAction {
        Browser browser;

        public Preferences (Browser browser) {
            super ("Pr&eferences");
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent evt) {
            new PreferencesDialog (browser).show ();
        }
    }

    
}
