/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import javax.swing.plaf.TextUI;
import lapisx.swing.Win;
import lapisx.util.MergingHashtable;
import lapisx.swing.Layout;
import lapisx.swing.ScrollBarHighlighter;

public class RegionHighlighter
	extends DefaultHighlighter
	implements ScrollBarHighlighter {
	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

	NamedRegionSet selection; // selection to be highlighted

	JTextComponent textComponent;
	// text component in which this highlighter is installed
	Rectangle componentBounds; // bounds of textComponent

	RegionSet faint; // region sets preprocessed from selection
	RegionSet dark;
	RegionSet points;

	boolean rightFill = true;

	final static Color DEFAULT_GRADIENT_COLOR = new Color(61, 255, 108);
	int gradientLevels = 13;
	int maxGradientWidth = 26;
	float gradientFactor = 1.0f;
	Color[] gradientColors = null;

	int shrinkX = 1;
	int shrinkY = 1;
	Color shadowColor = new Color(90, 90, 90); // dark gray
	int shadowX = 0;
	int shadowY = 0;

	boolean outlining = false;
	boolean showScrollBarHighlights = false;
	boolean showHandles = true;
	boolean showCaret = true;
	boolean showCaretGradient = false;

	Color scrollBarHighlightColor = Color.red;

	final static Color DEFAULT_HANDLE_COLOR = new Color(244, 0, 0); // red    
	Color handleColor;
	Image startHandle, endHandle;

	Cursor startHandleCursor =
		Cursor.getPredefinedCursor(Cursor.W_RESIZE_CURSOR);
	Cursor endHandleCursor = Cursor.getPredefinedCursor(Cursor.E_RESIZE_CURSOR);

	final static Color DEFAULT_CARET_COLOR = DEFAULT_HANDLE_COLOR;
	Color caretColor;

	Highlighter nextHighlighter;
	ScrollBarHighlighter nextScrollBarHighlighter;

	public RegionHighlighter() {
		super();
		setGradientColor(DEFAULT_GRADIENT_COLOR);
		setHandleColor(DEFAULT_HANDLE_COLOR);
		setCaretColor(DEFAULT_CARET_COLOR);
	}

	public RegionHighlighter(Highlighter h) {
		this();
		nextHighlighter = h;
	}

	// 
	// chain of highlighters
	// 

	public void setNextHighlighter(Highlighter h) {
		nextHighlighter = h;
	}

	public Highlighter getNextHighlighter() {
		return nextHighlighter;
	}

	public void setNextScrollBarHighlighter(ScrollBarHighlighter h) {
		nextScrollBarHighlighter = h;
	}

	public ScrollBarHighlighter getNextScrollBarHighlighter() {
		return nextScrollBarHighlighter;
	}

	// 
	// Getting and setting the highlighted selection
	// 

	public void setSelection(NamedRegionSet selection) {
		NamedRegionSet oldSelection = this.selection;

		this.selection = selection;
		updateHighlight();

		// repaint scrollbar, if necessary
		if (showScrollBarHighlights) {
			JScrollBar sb = getScrollBar();
			if (sb != null)
				sb.repaint();
		}

		java.util.Hashtable updated = new java.util.Hashtable();
		if (oldSelection != null) {
			RegionSet u = oldSelection.convertToLatest();
			SubsetEnumeration se = u.singleDocize();
			for (RegionSet r = se.firstFast(); r != null; r = se.nextFast()) {
				if (!r.isEmpty()) {
					int s =
						Math.min(
							RegionSet.MAX_INDEX,
							((SingleDocRegionSet) r).getStartLow());
					int e =
						Math.max(
							RegionSet.MIN_INDEX,
							((SingleDocRegionSet) r).getEndHigh());
					updated.put(
						((SingleDocRegionSet) r).getVersion(),
						new int[] { s, e });
				}
			}
		}

		if (selection != null) {
			RegionSet u = selection.convertToLatest();
			SubsetEnumeration se = u.singleDocize();
			for (RegionSet r = se.firstFast(); r != null; r = se.nextFast()) {
				DocumentVersion ver = ((SingleDocRegionSet) r).getVersion();
				if (ver != null) {
					int[] endpoints = (int[]) updated.get(ver);
					int s, e;
					if (endpoints != null) {
						s =
							Math.min(
								endpoints[0],
								((SingleDocRegionSet) r).getStartLow());
						e =
							Math.max(
								endpoints[1],
								((SingleDocRegionSet) r).getEndHigh());
					} else {
						s =
							Math.min(
								RegionSet.MAX_INDEX,
								((SingleDocRegionSet) r).getStartLow());
						e =
							Math.max(
								RegionSet.MIN_INDEX,
								((SingleDocRegionSet) r).getEndHigh());
					}
					updated.put(ver, new int[] { s, e });
				}
			}
		}
		java.util.Enumeration enum = updated.keys();
		while (enum.hasMoreElements()) {
			DocumentVersion ver = (DocumentVersion) enum.nextElement();
			int[] endpoints = (int[]) updated.get(ver);
			int s = endpoints[0];
			int e = endpoints[1];
			if (s <= e)
				damage(s, e, ver);
		}

		/*
		// repaint the union of the two selections' bounding boxes
		int s = RegionSet.MAX_INDEX;
		int e = RegionSet.MIN_INDEX;
		
		if (oldSelection != null) {
		    RegionSet u = oldSelection.getLatestRegions ();
		    s = Math.min (s, u.singleDocize ().getStartLow ());
		    e = Math.max (e, u.singleDocize ().getEndHigh ());
		}
		
		if (selection != null) {
		    RegionSet u = selection.getLatestRegions ();
		    s = Math.min (s, u.singleDocize ().getStartLow ());
		    e = Math.max (e, u.singleDocize ().getEndHigh ());
		}
		
		if (s <= e)
		    damage (s, e);
		*/
	}

	public NamedRegionSet getSelection() {
		return selection;
	}

	//
	// Preparing the selection for painting
	//

	void updateHighlight() {
		faint = RegionSet.EMPTY;
		dark = RegionSet.EMPTY;
		points = RegionSet.EMPTY;
		clearFlags();
		clearException();

		if (selection == null || selection.isEmpty())
			return;

		MutableRegionSet faint = new MutableRegionSet();
		MutableRegionSet dark = new MutableRegionSet();
		MutableRegionSet points = new MutableRegionSet();

		// TODO: support multiple documents
		SwingDocument doc = (SwingDocument) textComponent.getDocument();
		
        // minwu
        //HashSet vers = new HashSet();
		//vers.add(doc.getVersion());
        
        Versions vers = new SingleDocVersions(doc.getVersion());

		DocumentVersion selVer =
			selection.getSingleDocRegionSet(doc).getVersion();
		if (doc.getVersion() != selVer)
			selection = (NamedRegionSet) selection.convert(vers);

		int length = doc.getCookedView().toRegion().getLength();
		CoordinateMap rawToCooked =
			doc.getCookedView().getVersion().diff(doc.getVersion());

		RegionEnumeration re = selection.regions();
		for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
			RegionSet s = r.convert(rawToCooked);
			if (s.isEmpty())
				continue;
			else
				r = (Region) s;

			// Make sure all MAX_INDEX coordinates are changed to
			// the true document length
			int sl = Math.min(length, r.getStartLow());
			int sh = Math.min(length, r.getStartHigh());
			int el = Math.min(length, r.getEndLow());
			int eh = Math.min(length, r.getEndHigh());

			int startDrawing = 0;
			int endDrawing = 0;

			// draw handles
			if (sl == eh) {
				if (showCaret)
					startDrawing |= CARET;
			} else if (showHandles) {
				if (sl == sh)
					startDrawing |= START_HANDLE;
				if (el == eh)
					endDrawing |= END_HANDLE;
			}

			// draw faint highlighting across full range of region [sl, eh]
			if (sl < sh || el < eh)
				faint.insert(sl, eh, doc);

			// draw dark highlighting across kernel of region [sh, el]
			if (sh < el)
				dark.insert(sh, el, doc);

			// draw gradients at boundaries between faint highlighting
			// and dark highlighting
			if (sl < sh)
				startDrawing |= ((sh == eh) ? FADEIN_BACKWARD : FADEIN_FORWARD);

			if (el < eh)
				endDrawing |= ((el == sl) ? FADEOUT_FORWARD : FADEOUT_BACKWARD);

			if (startDrawing != 0) {
				Region pt = new Region(sh, sh, doc);
				if (addFlags(pt, startDrawing))
					points.insert(pt);
			}

			if (endDrawing != 0) {
				Region pt = new Region(el, el, doc);
				if (addFlags(pt, endDrawing))
					points.insert(pt);
			}
		}

		this.faint = faint.flatten();
		this.dark = dark.flatten();
		this.points = points.flatten();
	}

	static final int START_HANDLE = 0x01;
	static final int END_HANDLE = 0x02;
	static final int HANDLES = START_HANDLE | END_HANDLE;
	static final int FADEIN_FORWARD = 0x04;
	static final int FADEOUT_FORWARD = 0x08;
	static final int FADEIN_BACKWARD = 0x10;
	static final int FADEOUT_BACKWARD = 0x20;
	static final int FADEIN = FADEIN_FORWARD | FADEIN_BACKWARD;
	static final int FADEOUT = FADEOUT_FORWARD | FADEOUT_BACKWARD;
	static final int FADE = FADEIN | FADEOUT;
	static final int CARET = 0x40;
	static final int MAX = 0x80;

	static final Integer[] INTEGER = new Integer[MAX];
	static {
		for (int i = 0; i < MAX; ++i)
			INTEGER[i] = new Integer(i);
	}

	MergingHashtable drawingFlags = new MergingHashtable() {
		protected Object merge(Object curr, Object added) {
			return INTEGER[((Integer) curr).intValue()
				| ((Integer) added).intValue()];
		}
	};

	void clearFlags() {
		drawingFlags.clear();
	}

	boolean addFlags(Region r, int flags) {
		return drawingFlags.add(r, INTEGER[flags]) == null;
	}

	int getFlags(Region r) {
		Integer f = (Integer) drawingFlags.get(r);
		return (f != null) ? f.intValue() : 0;
	}

	//
	// Finds textComponent's vertical scrollbar
	//

	JScrollBar getScrollBar() {
		// find ScrollPane ancestor, if any
		Component parent;
		for (parent = textComponent.getParent();
			parent != null && !(parent instanceof JScrollPane);
			parent = parent.getParent());

		if (parent == null)
			return null;
		else
			return ((JScrollPane) parent).getVerticalScrollBar();
	}

	// 
	// Painting
	// 

	// paint is called only once for each repaint
	public void paint(Graphics g) {
		// fetch latest size of text component
		componentBounds = getComponentBounds();

		if (selection != null) {
			SwingDocument doc = (SwingDocument) textComponent.getDocument();
			if (doc != null) {
				// TODO: support multiple documents
                
                // minwu
				//HashSet vers = new HashSet();
				//vers.add(doc.getVersion());
                
                Versions vers = new SingleDocVersions(doc.getVersion());
				// check that selection is still up-to-date
				DocumentVersion selVer =
					selection.getSingleDocRegionSet(doc).getVersion();
				if (doc.getVersion() != selVer)
					setSelection((NamedRegionSet) selection.convert(vers));
			}
		}

		if (nextHighlighter != null)
			nextHighlighter.paint(g);
	}

	// paintLayeredHighlights is called once for every View leaf (e.g.,
	// a line)
	public void paintLayeredHighlights(
		Graphics g,
		int p0,
		int p1,
		Shape viewBounds,
		JTextComponent editor,
		View view) {
		if (selection != null) {
			boolean plaintext =
				(textComponent.getDocument()
					instanceof lapis.swing.PlainDocument);
			Shape oldClip = g.getClip();
			if (!plaintext) {
				// set clipping region
				Shape viewClip = getViewClip(view, p0, p1, viewBounds);
				Rectangle viewClipBounds = viewClip.getBounds();
				// expand view clipping region by a pixel on all sides,
				// so that handles are more visible
				viewClipBounds.x -= 1;
				viewClipBounds.y -= 1;
				viewClipBounds.width += 2;
				viewClipBounds.height += 2;
				g.setClip(getShapeBounds(oldClip).intersection(viewClipBounds));
			}

			// compute visible region and view bounding box
			int realP0 = p0;
			int realP1 = p1;
			if (plaintext) {
				try {
					if (textComponent.getText(realP1, 1).equals("\n"))
						++realP1;
				} catch (BadLocationException e) {
				}
			}

			// this Region is only for drawing purpose 
			Region visible =
				new Region(
					realP0,
					realP1,
					(lapis.Document) textComponent.getDocument());
			Rectangle bounds = getShapeBounds(viewBounds);

			// draw region set
			drawHighlight(g, visible, view, bounds);

			// restore clipping region
			g.setClip(oldClip);
		}

		if (nextHighlighter instanceof LayeredHighlighter)
			((LayeredHighlighter) nextHighlighter).paintLayeredHighlights(
				g,
				p0,
				p1,
				viewBounds,
				editor,
				view);

		//         if (!plaintext && isImage (view))
		//             g.setXORMode (Color.white);
	}

	public void paintScrollBarHighlights(
		Graphics g,
		JScrollBar scrollbar,
		Rectangle trackBounds,
		Rectangle incrButtonBounds,
		Rectangle decrButtonBounds) {
		drawHighlight : if (showScrollBarHighlights) {
			TextUI ui = textComponent.getUI();
			int viewMin = scrollbar.getMinimum();
			int viewExtent = scrollbar.getMaximum() - viewMin;

			int trackExtent = trackBounds.height;

			int w = 4;
			int x = trackBounds.x + 1 + (trackBounds.width - w) / 2;

			double trackInflection = 0; //decrButtonBounds.height;
			double viewInflection = trackInflection + decrButtonBounds.height;
			double slope1 = trackInflection / viewInflection;
			double slope2 =
				(trackExtent - 2 * trackInflection)
					/ (viewExtent - 2 * viewInflection);

			g.setColor(scrollBarHighlightColor);

			if (dark == null || points == null)
				break drawHighlight;
			RegionEnumeration re =
				(!dark.isEmpty()) ? dark.regions() : points.regions();
			for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
				try {
					Rectangle p;

					p =
						ui.modelToView(
							textComponent,
							r.getStart(),
							Position.Bias.Forward);
					long v1 = p.y - viewMin;
					p =
						ui.modelToView(
							textComponent,
							r.getEnd(),
							Position.Bias.Backward);
					long v2 = p.y + p.height - viewMin;

					double w1, w2;
					if (v1 < viewInflection)
						w1 = v1 * slope1;
					else if (v1 > viewExtent - viewInflection)
						w1 = trackExtent - (viewExtent - v1) * slope1;
					else
						w1 = trackInflection + (v1 - viewInflection) * slope2;

					if (v2 < viewInflection)
						w2 = v2 * slope1;
					else if (v2 > viewExtent - viewInflection)
						w2 = trackExtent - (viewExtent - v2) * slope1;
					else
						w2 = trackInflection + (v2 - viewInflection) * slope2;

					int d1 = (int) Math.ceil(w1);
					int d2 = (int) Math.floor(w2 - 1);

					int y = trackBounds.y + d1;
					int h = Math.max(d2 - d1, 2);

					g.fillRect(x, y, w, h);

				} catch (BadLocationException e) {
				}
			}
		}

		if (nextScrollBarHighlighter != null)
			(
				(
					ScrollBarHighlighter) nextScrollBarHighlighter)
						.paintScrollBarHighlights(
				g,
				scrollbar,
				trackBounds,
				incrButtonBounds,
				decrButtonBounds);
	}

	static boolean isImage(View view) {
		Element element = view.getElement();
		AttributeSet attr = element.getAttributes();
		HTML.Tag t = (HTML.Tag) attr.getAttribute(StyleConstants.NameAttribute);
		return (t == HTML.Tag.IMG);
	}

	static Rectangle getShapeBounds(Shape s) {
		return (s instanceof Rectangle) ? (Rectangle) s : s.getBounds();
	}

	Rectangle getComponentBounds() {
		Rectangle bounds = new Rectangle(textComponent.getSize());
		Insets insets = textComponent.getInsets();
		bounds.x += insets.left;
		bounds.y += insets.top;
		bounds.width -= insets.left + insets.right;
		bounds.height -= insets.top + insets.bottom;
		return bounds;
	}

	Shape getViewClip(View view, int p0, int p1, Shape viewBounds) {
		if (p0 == view.getStartOffset() && p1 == view.getEndOffset())
			// Precisely contained in view
			return viewBounds;
		else {
			// Only part of view -- determine its bounding box
			try {
				Shape partialViewBounds =
					view.modelToView(
						p0,
						Position.Bias.Forward,
						p1,
						Position.Bias.Backward,
						viewBounds);

				if (!rightFill)
					// just use partial view
					return partialViewBounds;

				// if we're right-filling, then the clipping region we 
				// want is partialViewBounds, but expanded left and right 
				// to the full width of viewBounds
				Rectangle rectView = getShapeBounds(viewBounds);
				Rectangle rectPartialView = getShapeBounds(partialViewBounds);

				rectPartialView.x = rectView.x;
				rectPartialView.width = rectView.width;
				return rectPartialView;

			} catch (BadLocationException e) {
				// fall back to full view
				return viewBounds;
			}
		}
	}

	protected void drawHighlight(
		Graphics g,
		Region visible,
		View view,
		Rectangle bounds) {

		Region overlapsVisible = (Region) visible.overlaps();
		RegionEnumeration re;

		// Draw faint highlighting
		g.setColor(getGradientColor(0.0));
		re = faint.regions(overlapsVisible);
		for (Region r = re.firstFast(); r != null; r = re.nextFast())
			drawRegion(g, r, visible, view, bounds);

		// Draw dark highlighting
		g.setColor(getGradientColor(1.0));
		re = dark.regions(overlapsVisible);
		for (Region r = re.firstFast(); r != null; r = re.nextFast())
			drawRegion(g, r, visible, view, bounds);

		// draw gradients, handles and carets
		re = points.regions(overlapsVisible);
		for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
			int flags = getFlags(r);
			if ((flags & FADE) != 0)
				drawGradient(g, r.getStart(), flags, visible, view, bounds);
			if ((flags & HANDLES) != 0)
				drawHandle(g, r.getStart(), flags, visible, view, bounds);
			if ((flags & CARET) != 0)
				drawCaret(g, r.getStart(), flags, visible, view, bounds);
		}
	}

	protected void drawRegion(
		Graphics g,
		Region r,
		Region visible,
		View view,
		Rectangle bounds) {
		try {
			// Clip r to visible
			boolean startsBefore = (r.getStart() < visible.getStart());
			boolean endsAfter = (r.getEnd() > visible.getEnd());
			int start = (startsBefore) ? visible.getStart() : r.getStart();
			int end = (endsAfter) ? visible.getEnd() : r.getEnd();

			if (start == end)
				return;

			//             debug.println ("highlighting <" 
			//                            + textComponent.getText ().substring (start, end) 
			//                            + "> inside <"
			//                            + textComponent.getText ().substring (r.getStart (), r.getEnd ())
			//                            + ">");

			// Fetch view positions.
			Rectangle p0 =
				getViewPosition(start, Position.Bias.Forward, view, bounds);
			Rectangle p1 =
				getViewPosition(end, Position.Bias.Backward, view, bounds);

			// Undo shrinkage if total highlight starts before or
			// ends after the little part we're drawing
			if (!outlining) {
				if (startsBefore) {
					p0.x -= shrinkX;
					p0.width += shrinkX;

					// undo y shrinkage only if r.getStart () is on a previous line 
					// (i.e. pStart's y interval doesn't overlap p0's y interval)
					Rectangle pStart =
						getViewPosition(
							r.getStart(),
							Position.Bias.Forward,
							null,
							bounds);
					if (!sameLine(p0, pStart)) {
						p0.y -= shrinkY;
						p0.height += shrinkY;
					}
				}

				if (endsAfter) {
					p1.width += shrinkX;

					// undo y shrinkage only if r.getEnd () is on a later line
					Rectangle pEnd =
						getViewPosition(
							r.getEnd(),
							Position.Bias.Backward,
							null,
							bounds);
					if (!sameLine(p1, pEnd))
						p1.height += shrinkY;
				}
			}

			// Highlight from p0 to p1
			Rectangle p = p0.union(p1);
			drawRegionRect(g, p.x, p.y, p.width, p.height);
		} catch (BadLocationException e) {
			// can't render
			reportException(e);
		}
	}

	protected void drawRegionRect(Graphics g, int x, int y, int w, int h) {
		g.fillRect(x, y, w, h);
	}

	boolean sameLine(Rectangle r1, Rectangle r2) {
		return Math.max(r1.y, r2.y)
			< Math.min(r1.y + r1.height, r2.y + r2.height);
	}

	protected void drawGradient(
		Graphics g,
		int pos,
		int flags,
		Region visible,
		View view,
		Rectangle bounds) {
		if ((flags & FADEIN) != 0
			&& ((flags & FADEIN_FORWARD) != 0
				? pos < visible.getEnd()
				: pos > visible.getStart())) {
			try {
				// Fetch view position
				Position.Bias bias =
					((flags & FADEIN_FORWARD) != 0)
						? Position.Bias.Forward
						: Position.Bias.Backward;
				Rectangle p = getViewPosition(pos, bias, view, bounds);

				// draw it
				drawGradientRect(
					g,
					p.x - maxGradientWidth,
					p.y,
					maxGradientWidth,
					p.height,
					0.0,
					1.0);
			} catch (BadLocationException e) {
				// can't render
				reportException(e);
			}
		}

		if ((flags & FADEOUT) != 0
			&& ((flags & FADEOUT_FORWARD) != 0
				? pos < visible.getEnd()
				: pos > visible.getStart())) {
			try {
				// Fetch view position
				Position.Bias bias =
					((flags & FADEOUT_FORWARD) != 0)
						? Position.Bias.Forward
						: Position.Bias.Backward;
				Rectangle p = getViewPosition(pos, bias, view, bounds);

				// draw it
				drawGradientRect(
					g,
					p.x,
					p.y,
					maxGradientWidth,
					p.height,
					1.0,
					0.0);
			} catch (BadLocationException e) {
				// can't render
				reportException(e);
			}
		}
	}

	protected void drawHandle(
		Graphics g,
		int pos,
		int flags,
		Region visible,
		View view,
		Rectangle bounds) {
		if ((flags & START_HANDLE) != 0 && pos < visible.getEnd()) {
			try {
				// Fetch view position
				Rectangle p =
					getViewPosition(pos, Position.Bias.Forward, view, bounds);

				// draw the handle
				g.drawImage(startHandle, p.x, p.y, null);
			} catch (BadLocationException e) {
				// can't render
				reportException(e);
			}
		}

		if ((flags & END_HANDLE) != 0 && pos > visible.getStart()) {
			try {
				// Fetch view position
				Rectangle p =
					getViewPosition(pos, Position.Bias.Backward, view, bounds);

				// draw the handle
				int x = p.x + p.width - endHandleWidth;
				int y = p.y + p.height - endHandleHeight;
				g.drawImage(endHandle, x, y, null);

			} catch (BadLocationException e) {
				// can't render
				reportException(e);
			}
		}

	}

	protected void drawCaret(
		Graphics g,
		int pos,
		int flags,
		Region visible,
		View view,
		Rectangle bounds) {
		if ((flags & CARET) != 0 && pos < visible.getEnd()) {
			int oldShrinkX = shrinkX;
			shrinkX = 0;
			try {
				if (showCaretGradient) {
					drawGradient(
						g,
						pos,
						FADEIN_BACKWARD,
						visible,
						view,
						bounds);
					drawGradient(
						g,
						pos,
						FADEOUT_FORWARD,
						visible,
						view,
						bounds);
				}

				// Fetch view position
				Rectangle p =
					getViewPosition(pos, Position.Bias.Forward, view, bounds);

				// draw the caret
				g.setColor(caretColor);
				g.fillRect(p.x - 2, p.y + 1, 3, p.height - 2);
				g.drawLine(p.x - 3, p.y + 1, p.x + 1, p.y + 1);
				g.drawLine(
					p.x - 3,
					p.y + p.height - 2,
					p.x + 1,
					p.y + p.height - 2);

			} catch (BadLocationException e) {
				// can't render
				reportException(e);
			} finally {
				shrinkX = oldShrinkX;
			}

		}
	}

	//     int modelToViews = 0;
	//     boolean reportQueued = false;
	//     long started;

	Rectangle getViewPosition(
		int pos,
		Position.Bias bias,
		View view,
		Rectangle bounds)
		throws BadLocationException {
		Rectangle r;

		if (view != null) {
			r = getShapeBounds(view.modelToView(pos, bounds, bias));
		} else {
			TextUI mapper = textComponent.getUI();
			r = mapper.modelToView(textComponent, pos, bias);
		}
		if (r == null)
			return new Rectangle(0, 0, 0, 0);

		SwingDocument doc = (SwingDocument) textComponent.getDocument();
		if (doc instanceof PlainDocument
			&& rightFill
			&& bias == Position.Bias.Backward
			&& doc.isLoaded()
			&& "\n".equals(doc.getText(new Region(pos - 1, pos, doc)))) {
			r.x = componentBounds.x + componentBounds.width - r.width;
		}

		// Shrink the view position to leave a little whitespace
		// between adjacent regions.

		r.y += shrinkY;
		r.height -= 2 * shrinkY;
		if (bias == Position.Bias.Forward)
			r.x += shrinkX;
		else
			r.x -= shrinkX;

		return r;
	}

	void drawGradientRect(
		Graphics g,
		int x,
		int y,
		int w,
		int h,
		double density0,
		double density1) {
		// Draw a horizontal sequence of rectangles with grades of color
		double deltaDensity = density1 - density0;
		int n = (int) Math.round(Math.abs(deltaDensity * gradientLevels)) + 1;
		double dx = ((double) w) / n;
		double dden = deltaDensity / n;
		double x1 = x;
		int dxi = (int) Math.ceil(dx);
		double den = density0;
		for (int i = 0; i < n; ++i) {
			g.setColor(getGradientColor(den));
			double x2 = x1 + dx;
			g.fillRect((int) x1, y, dxi, h);
			x1 = x2;
			den += dden;
		}
	}

	public boolean isScrollBarHighlighted() {
		return showScrollBarHighlights;
	}

	public void setScrollBarHighlighted(boolean f) {
		showScrollBarHighlights = f;
	}

	public Color getScrollBarHighlightColor() {
		return scrollBarHighlightColor;
	}

	public void setScrollBarHighlightColor(Color c) {
		scrollBarHighlightColor = c;
	}

	public boolean isHandlesVisible() {
		return showHandles;
	}

	public void setHandlesVisible(boolean f) {
		showHandles = f;
	}

	public boolean isCaretVisible() {
		return showCaret;
	}

	public void setCaretVisible(boolean f) {
		showCaret = f;
	}

	public boolean isCaretGradientVisible() {
		return showCaretGradient;
	}

	public void setCaretGradientVisible(boolean f) {
		showCaretGradient = f;
	}

	public Color getGradientColor(double gradient) {
		return gradientColors[(int) Math.round(gradient * gradientLevels)];
	}

	public void setGradientColor(Color color) {
		gradientColors = new Color[gradientLevels + 1];

		float[] hsb = new float[3];
		Color.RGBtoHSB(color.getRed(), color.getGreen(), color.getBlue(), hsb);
		float hue = hsb[0];
		float maxSaturation = hsb[1];
		float minSaturation = maxSaturation / 3.0f;
		float brightness = hsb[2];

		// as gradient x goes from 0.0 to 1.0, 
		// saturation(x) = c0 exp(gradientFactor *x) + c1.
		// We need to compute c0 and c1 so that saturation(0.0) = minSaturation
		// and saturation(1.0) = maxSaturation.  Solving gives: 
		double c0 =
			(maxSaturation - minSaturation) / (Math.exp(gradientFactor) - 1);
		double c1 = minSaturation - c0;

		// precompute colors for every gradient quantum from 0.0 to 1.0
		for (int i = 0; i <= gradientLevels; ++i) {
			double x = i / (double) gradientLevels;
			double sat = c0 * Math.exp(gradientFactor * x) + c1;
			gradientColors[i] = Color.getHSBColor(hue, (float) sat, brightness);
		}
	}

	static int originalHandleColor = 0xfff40000;

	static int startHandleWidth = 5;
	static int startHandleHeight = 5;
	static int[] startHandleData =
		{
			0xfff40000,
			0xfff40000,
			0xfff40000,
			0xfff40000,
			0xfff40000,
			0xfff40000,
			0xfff40000,
			0xffffff,
			0xffffff,
			0xffffff,
			0xfff40000,
			0xffffff,
			0xffffff,
			0xffffff,
			0xffffff,
			0xfff40000,
			0xffffff,
			0xffffff,
			0xffffff,
			0xffffff,
			0xfff40000,
			0xffffff,
			0xffffff,
			0xffffff,
			0xffffff };

	static int endHandleWidth = 5;
	static int endHandleHeight = 5;
	static int[] endHandleData =
		{
			0xffffff,
			0xffffff,
			0xffffff,
			0xffffff,
			0xfff40000,
			0xffffff,
			0xffffff,
			0xffffff,
			0xffffff,
			0xfff40000,
			0xffffff,
			0xffffff,
			0xffffff,
			0xffffff,
			0xfff40000,
			0xffffff,
			0xffffff,
			0xffffff,
			0xfff40000,
			0xfff40000,
			0xfff40000,
			0xfff40000,
			0xfff40000,
			0xfff40000,
			0xfff40000 };

	public void setHandleColor(Color color) {
		handleColor = color;

		int argb = 0xff000000 | color.getRGB();
		int[] recoloredStartHandleData =
			replace(startHandleData, originalHandleColor, argb);
		int[] recoloredEndHandleData =
			replace(endHandleData, originalHandleColor, argb);

		Toolkit tkit = Toolkit.getDefaultToolkit();
		startHandle =
			tkit.createImage(
				new MemoryImageSource(
					startHandleWidth,
					startHandleHeight,
					recoloredStartHandleData,
					0,
					startHandleWidth));
		endHandle =
			tkit.createImage(
				new MemoryImageSource(
					endHandleWidth,
					endHandleHeight,
					recoloredEndHandleData,
					0,
					endHandleWidth));
	}

	protected int[] replace(int[] data, int oldpixel, int newpixel) {
		int[] out = new int[data.length];
		for (int i = 0; i < data.length; ++i)
			out[i] = (data[i] == oldpixel) ? newpixel : data[i];
		return out;
	}

	public void setCaretColor(Color color) {
		caretColor = color;
	}

	public void setOutlining(boolean outlining) {
		this.outlining = outlining;
	}

	public boolean getOutlining() {
		return outlining;
	}

	public void setShrinkage(int shrink) {
		shrinkX = shrinkY = shrink;
	}

	public int getShrinkage() {
		return shrinkX;
	}

	//
	// exception handling during paint -- avoids printing lots of redundant stack
	// traces by only printing the first exception
	//
	boolean reported = false;
	void reportException(Throwable t) {
		if (reported)
			return;
		reported = true;
		debug.report(t);
	}
	void clearException() {
		reported = false;
	}

	//
	// attaching to a JTextComponent
	//

	public void install(JTextComponent c) {
		textComponent = c;
		super.install(c);
		if (nextHighlighter != null)
			nextHighlighter.install(c);
	}

	public void deinstall(JTextComponent c) {
		textComponent = null;
		super.deinstall(c);
		if (nextHighlighter != null)
			nextHighlighter.deinstall(c);
	}

	//
	// repainting
	//

	public void damageAll() {
		damage(
			RegionSet.MIN_INDEX,
			RegionSet.MAX_INDEX,
			((SwingDocument) textComponent.getDocument()).getVersion());
	}

	public void damage(int start, int end, DocumentVersion ver) {
		debug.assertion(
			(SwingDocument) textComponent.getDocument() == ver.getDocument());
		debug.println("damaging [" + start + "," + end + "]");
		SwingDocument doc = (SwingDocument) ver.getDocument();
		CoordinateMap rawToCooked = doc.getCookedView().getVersion().diff(ver);

		start = rawToCooked.mapStart(start);
		end = rawToCooked.mapEnd(end);

		// clip start and end to into [0,length]
		int len = textComponent.getDocument().getLength();
		start = Math.max(0, Math.min(start, len));
		end = Math.max(0, Math.min(end, len));

		try {
			// TODO: getComponentBounds should be based on Document
			Rectangle bounds = getComponentBounds();
			Rectangle p0 =
				getViewPosition(start, Position.Bias.Backward, null, bounds);
			Rectangle p1 =
				getViewPosition(end, Position.Bias.Forward, null, bounds);
			debug.println("damaging [" + p0 + "," + p1 + "]");
			textComponent.repaint(
				0,
				p0.y,
				bounds.x + bounds.width,
				(p1.y + p1.height) - p0.y);
		} catch (BadLocationException e) {
			reportException(e);
		} catch (NullPointerException e) {
			// sometimes happens when changing documents
			reportException(e);
		}

	}

	//
	// region handles
	//

	public Handle handleAtPoint(Point pt) {
		if (selection == null)
			return null;

		debug.println("clicked on " + pt);

		TextUI ui = textComponent.getUI();
		SwingDocument doc = (SwingDocument) textComponent.getDocument();
		int pos = ui.viewToModel(textComponent, pt);

		RegionEnumeration re =
			points.regions(
				new FuzzyRegion(pos - 1, pos + 1, pos - 1, pos + 1, doc));
		for (Region h = re.firstFast(); h != null; h = re.nextFast()) {
			pos = h.getStartLow();

			int flags = getFlags(h);
			int handleClicked = 0; // 0, START_HANDLE, or END_HANDLE
			if ((flags & START_HANDLE) != 0) {
				try {
					Rectangle p =
						getViewPosition(
							pos,
							Position.Bias.Forward,
							null,
							getComponentBounds());
					p.x -= 2;
					p.y -= 2;
					p.width = endHandleWidth + 4;
					p.height = endHandleHeight + 4;
					debug.println("hit-testing start handle " + p);
					if (p.contains(pt))
						handleClicked = START_HANDLE;
				} catch (BadLocationException e) {
				}
			}
			if ((flags & END_HANDLE) != 0) {
				try {
					Rectangle p =
						getViewPosition(
							pos,
							Position.Bias.Backward,
							null,
							getComponentBounds());
					p.x += p.width - endHandleWidth - 2;
					p.y += p.height - endHandleHeight - 2;
					p.width = endHandleWidth + 4;
					p.height = endHandleHeight + 4;
					debug.println("hit-testing end handle " + p);
					if (p.contains(pt))
						handleClicked = END_HANDLE;
				} catch (BadLocationException e) {
				}
			}

			if (handleClicked == 0)
				continue;

			// find region for this handle
			Region handleSpot;
			CoordinateMap cookedToRaw =
				doc.getVersion().diff(doc.getCookedView().getVersion());
			int p0 = cookedToRaw.mapStart(pos);
			int p1 = cookedToRaw.mapEnd(pos);
			if (p0 == p1)
				handleSpot = new Region(p0, p1, doc);
			else
				handleSpot = new FuzzyRegion(p0, p1, p0, p1, doc);

			debug.println("clicked on handle at " + handleSpot);

			Region where =
				(Region) ((handleClicked == START_HANDLE)
					? handleSpot.starting(RegionSet.EMPTY)
					: handleSpot.ending(RegionSet.EMPTY));
			Region r = selection.regions(where).first();

			if (r == null) {
				debug.println("handle doesn't have a corresponding region!");
				continue;
			}

			return new Handle(handleClicked == START_HANDLE, r);
		}

		return null;
	}

	public static class Handle {
		boolean start;
		Region r;
		public Handle(boolean start, Region r) {
			this.start = start;
			this.r = r;
		}
		public boolean isStart() {
			return start;
		}
		public Region getRegion() {
			return r;
		}
	}

	//
	// Display options UI
	//

	public void showOptionDialog(Component comp) {
		new OptionDialog(comp).show();
	}

	public class OptionDialog extends JDialog {
		JButton colorButton;

		JSlider xIndentSlider;
		JLabel xIndentLabel;

		JSlider yIndentSlider;
		JLabel yIndentLabel;

		JButton shadowColorButton;

		JSlider xShadowSlider;
		JLabel xShadowLabel;

		JSlider yShadowSlider;
		JLabel yShadowLabel;

		JButton handleColorButton;
		JButton caretColorButton;

		JSlider gradientLevelsSlider;
		JLabel gradientLevelsLabel;

		JSlider maxGradientWidthSlider;
		JLabel maxGradientWidthLabel;

		JTextField gradientFactorTextField;

		JCheckBox paintLayersCheckBox;
		JCheckBox rightFillCheckBox;

		public OptionDialog(Component comp) {
			super(Win.findFrameOrMakeFrame(comp), "Highlight Options");

			Layout layout = Layout.gridbag(getContentPane());

			layout
				.add(new JLabel("Highlight Color: "))
				.add(colorButton = new JButton(" "))
				.nextRow();
			colorButton.setBackground(
				RegionHighlighter.this.getGradientColor(1.0));
			colorButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent event) {
					Color c =
						JColorChooser.showDialog(
							OptionDialog.this,
							"Pick Highlight Color",
							RegionHighlighter.this.getGradientColor(1.0));
					if (c != null) {
						setGradientColor(c);
						colorButton.setBackground(
							RegionHighlighter.this.getGradientColor(1.0));
						damageAll();
					}
				}
			});

			layout
				.add(new JLabel("X Indent: "))
				.add(xIndentSlider = new JSlider(0, 10, shrinkX))
				.add(xIndentLabel = new JLabel(String.valueOf(shrinkX)))
				.nextRow();
			xIndentSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					shrinkX = xIndentSlider.getValue();
					xIndentLabel.setText(String.valueOf(shrinkX));
					damageAll();
				}
			});

			layout
				.add(new JLabel("Y Indent: "))
				.add(yIndentSlider = new JSlider(0, 10, shrinkY))
				.add(yIndentLabel = new JLabel(String.valueOf(shrinkY)))
				.nextRow();
			yIndentSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					shrinkY = yIndentSlider.getValue();
					yIndentLabel.setText(String.valueOf(shrinkY));
					damageAll();
				}
			});

			layout
				.add(new JLabel("Shadow Color: "))
				.add(shadowColorButton = new JButton(" "))
				.nextRow();
			shadowColorButton.setBackground(shadowColor);
			shadowColorButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent event) {
					Color c =
						JColorChooser.showDialog(
							OptionDialog.this,
							"Pick Shadow Color",
							shadowColor);
					if (c != null) {
						shadowColor = c;
						shadowColorButton.setBackground(shadowColor);
						damageAll();
					}
				}
			});

			layout
				.add(new JLabel("X Shadow: "))
				.add(xShadowSlider = new JSlider(0, 10, shadowX))
				.add(xShadowLabel = new JLabel(String.valueOf(shadowX)))
				.nextRow();
			xShadowSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					shadowX = xShadowSlider.getValue();
					xShadowLabel.setText(String.valueOf(shadowX));
					damageAll();
				}
			});

			layout
				.add(new JLabel("Y Shadow: "))
				.add(yShadowSlider = new JSlider(0, 10, shadowY))
				.add(yShadowLabel = new JLabel(String.valueOf(shadowY)))
				.nextRow();
			yShadowSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					shadowY = yShadowSlider.getValue();
					yShadowLabel.setText(String.valueOf(shadowY));
					damageAll();
				}
			});

			layout
				.add(new JLabel("Handle Color: "))
				.add(handleColorButton = new JButton(" "))
				.nextRow();
			handleColorButton.setBackground(handleColor);
			handleColorButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent event) {
					Color c =
						JColorChooser.showDialog(
							OptionDialog.this,
							"Pick Handle Color",
							handleColor);
					if (c != null) {
						setHandleColor(c);
						handleColorButton.setBackground(handleColor);
						damageAll();
					}
				}
			});

			layout
				.add(new JLabel("Caret Color: "))
				.add(caretColorButton = new JButton(" "))
				.nextRow();
			caretColorButton.setBackground(caretColor);
			caretColorButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent event) {
					Color c =
						JColorChooser.showDialog(
							OptionDialog.this,
							"Pick Caret Color",
							caretColor);
					if (c != null) {
						setCaretColor(c);
						caretColorButton.setBackground(caretColor);
						damageAll();
					}
				}
			});

			layout
				.add(new JLabel("Gradient Levels: "))
				.add(gradientLevelsSlider = new JSlider(0, 20, gradientLevels))
				.add(
					gradientLevelsLabel =
						new JLabel(String.valueOf(gradientLevels)))
				.nextRow();
			gradientLevelsSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					Color color = getGradientColor(1.0);
					gradientLevels = gradientLevelsSlider.getValue();
					gradientLevelsLabel.setText(String.valueOf(gradientLevels));
					setGradientColor(color);
					damageAll();
				}
			});

			layout
				.add(new JLabel("Gradient Width: "))
				.add(
					maxGradientWidthSlider =
						new JSlider(0, 100, maxGradientWidth))
				.add(
					maxGradientWidthLabel =
						new JLabel(String.valueOf(maxGradientWidth)))
				.nextRow();
			maxGradientWidthSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					maxGradientWidth = maxGradientWidthSlider.getValue();
					maxGradientWidthLabel.setText(
						String.valueOf(maxGradientWidth));
					damageAll();
				}
			});

			layout
				.add(new JLabel("Gradient Factor: "))
				.add(
					gradientFactorTextField =
						new JTextField(String.valueOf(gradientFactor)))
				.nextRow();
			gradientFactorTextField.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent event) {
					gradientFactor =
						Float
							.valueOf(gradientFactorTextField.getText())
							.floatValue();
					setGradientColor(getGradientColor(1.0));
					damageAll();
				}
			});

			layout
				.nextColumn()
				.add(rightFillCheckBox = new JCheckBox("Right fill", rightFill))
				.nextRow();
			rightFillCheckBox.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent event) {
					rightFill = rightFillCheckBox.isSelected();
					damageAll();
				}
			});

			setDefaultCloseOperation(DISPOSE_ON_CLOSE);

			pack();
			Win.center(this, getParent());
		}
	}

}
