/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;
import lapis.*;
import java.awt.*;
import java.util.*;
import javax.swing.*;
import lapisx.swing.Layout;
import lapisx.util.Str;

public class RegionSetDisplay extends JFrame implements SelectionListener {

	RegionSpace display;
	JTextArea matches;

	NamedRegionSet sel;
	int len;

	static final Color regionColor = Color.red;
	static final Color treeColor = Color.blue;
	static final Color guideColor = Color.black;

	public RegionSetDisplay() {
		super("Region Set Display");

		Layout.vertical(getContentPane()).add(display = new RegionSpace()).add(
			new JScrollPane(matches = new JTextArea(8, 40)));

		setDefaultCloseOperation(DISPOSE_ON_CLOSE);

		pack();
	}

	public void setLength(int len) {
		this.len = len;
	}

	public void setSelection(NamedRegionSet sel) {
		this.sel = sel;

		if (sel != null) {
			//RegionSet r = sel.getRegions ();
			int n = sel.getSize();
			boolean flat = sel.isFlat();
			boolean allPoints = sel.isAllPoints();
			boolean crisp = sel.isCrisp();
			String msg =
				n
					+ " match"
					+ (n == 1 ? "" : "es")
					+ " found in "
					+ (RegionSet) sel
					+ (crisp ? " crisp" : "")
					+ (flat ? " flat" : "")
					+ (allPoints ? " all-points" : "")
					+ "\n"
					+ Str.abbreviateLines(
						sel.toStringWithContent(),
						20,
						"... %d regions omitted ...\n");
			matches.setText(msg);
		} else
			matches.setText("");

		matches.select(0, 0);
		display.repaint();
	}

	//
	// SelectionListener interface
	//

	public void changedSelection(Browser browser, NamedRegionSet sel) {
		if (isVisible())
			setSelection(sel);
	}

	public class RegionSpace extends JComponent {
		double scaleX, scaleY;

		public Dimension getPreferredSize() {
			return new Dimension(500, 500);
		}

		public void paintComponent(Graphics g) {
			super.paintComponent(g);

			Dimension d = RegionSpace.this.getSize();
			g.setColor(RegionSpace.this.getBackground());
			g.fillRect(0, 0, d.width, d.height);

			g.setColor(guideColor);
			g.drawLine(0, d.height, d.width, 0);

			if (sel != null) {
				len = 0;
                Iterator idoc = sel.getDocuments();
				while (idoc.hasNext())
					len += ((Document) idoc.next()).toRegion().getLength();

				if (len > 0) {
					scaleX = (d.width - 1) / (double) len;
					scaleY = (d.height - 1) / (double) len;

					drawTree(g, sel);
				}
			}
		}

		void drawTree(Graphics g, RegionSet v) {
			if (!(v instanceof SingleDocRegionSet)) {
				SubsetEnumeration se = v.singleDocize();
				for (RegionSet s = se.firstFast();
					s != null;
					s = se.nextFast())
					drawTree(g, s);
			} else {
				drawBoundingBox(g, (SingleDocRegionSet) v);

				if (!(v instanceof Region)) {
					SubsetEnumeration e = v.subsets();
					for (RegionSet s = e.firstFast();
						s != null;
						s = e.nextFast())
						drawTree(g, s);
				}
			}
		}

		void drawBoundingBox(Graphics g, SingleDocRegionSet s) {
			int sh = Math.min(len, s.getStartHigh());
			int eh = Math.min(len, s.getEndHigh());
			int x = (int) Math.floor(s.getStartLow() * scaleX);
			int y = (int) Math.floor((len - eh) * scaleY);
			int w = (int) Math.ceil((sh - s.getStartLow()) * scaleX);
			int h = (int) Math.ceil((eh - s.getEndLow()) * scaleY);
			if (w < 3)
				w = 3;
			if (h < 3)
				h = 3;
			g.setColor((s instanceof Region) ? regionColor : treeColor);
			if (s instanceof Region)
				g.fillRect(x, y, w + 1, h + 1);
			else
				g.drawRect(x, y, w, h);
		}
	}

}
