/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.net.*;
import javax.swing.*;
import lapisx.swing.*;
import lapisx.util.*;
import lapisx.enum.RestartableEnumeration;
import lapis.ml.*;

public class SimulEditMode extends JPanel 
    implements SelectionMode, BrowserListener, SelectionListener
{
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    JButton closeButton;
    JScrollPane helpScroll;
    HypothesisDisplay patterns;
    JCheckBox showOutliersCheckBox;
    int maxOutliers; // loaded from config

    Browser browser;

    NamedRegionSet range;
    InferenceAlgorithm inference;

    boolean mouseSelection = false;
    NamedRegionSet positiveExamples;
    NamedRegionSet negativeExamples;
    Region firstPositiveExample;
    Thread outlierThread;

    public SimulEditMode (Browser browser) {
        this.browser = browser;

        range = browser.getSelection ();
        if (!isOKToSimulEdit (range)) {
            SwingDocument doc = browser.getDocument ();
			range = new NamedRegionSet (doc.toRegion ());
        } else
            range = prepareForSimulEdit (range);

        // make inference algorithm
        inference = new HybridInference (range);
                    
        // initialize positive and negative examples
        clearExamples ();


        // close button
        closeButton = new JButton (new AbstractAction ("Close") {
                public void actionPerformed (ActionEvent event) {
                    SimulEditMode.this.browser.setSelectionMode 
                        (SimulEditMode.this.browser.getManualSelectionMode ());
                }
            });

        // make help box
        URL helpURL = SwingConfig.makeURL ("doc/messages/simuledit.html");
        JEditorPane helpMsg;
        try {
            helpMsg = new JEditorPane (helpURL);
        } catch (IOException e) {
            FeedbackForm.showExceptionDialog (browser, e);
            helpMsg = new JEditorPane ();
        }
        helpScroll = new JScrollPane (helpMsg);
        helpMsg.setOpaque (false);
        helpMsg.setEditable (false);

        // make pattern feedback
        patterns = new HypothesisDisplay ();
        patterns.addSelectionListener (selectionListener);

        // make check box
        showOutliersCheckBox = new JCheckBox 
            ("Highlight unusual selections in red",
             "true".equals (Main.getConfig ().getProperty ("browser.show-outliers")));
        showOutliersCheckBox.addActionListener (new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    Main.getConfig ().put ("browser.show-outliers",
                                           showOutliersCheckBox.isSelected ()
                                           ? "true" : "false");
                    findOutlierRecords ();
                }
            });

        maxOutliers = Integer.parseInt (Main.getConfig ()
                                        .getProperty ("browser.max-outliers"));


        // assemble layout
        Layout.gridbag (this)
            .setContainerInsets (5, 5, 5, 5)
            .add (helpScroll)
            .nextRow ()
            .add (patterns, new Insets (10, 0, 0, 0))
            .nextRow ()
            .add (Layout.horizontal ()
                  .add (showOutliersCheckBox, Layout.FIXED_HEIGHT)
                  .add (closeButton, Layout.FIXED_SIZE)
                  .done (), Layout.FIXED_HEIGHT)
            .nextRow ()
            ;

        setMinimumSize (new Dimension (0, 0));
    }

    private boolean isOKToSimulEdit (NamedRegionSet sel) {
        int nNonZeroRecords = 0;
        RegionEnumeration e = sel.regions ();
        for (Region r = e.firstFast (); r != null; r = e.nextFast ()) {
            if (r.getStartLow () < r.getEndHigh ())
                ++nNonZeroRecords;
        }
            
        return nNonZeroRecords >= 1;
    }

    private NamedRegionSet prepareForSimulEdit (NamedRegionSet sel) {
        if (!sel.isFlat ())
			sel = new NamedRegionSet (sel.flatten ());

    // Remove blank records (zero-length or containing only whitespace).
    // FIX: ask user?  provide an option?
//         MutableRegionSet newRecords = null;
//         RegionEnumeration e = records.regions ();
//         for (Region r = e.firstFast (); r != null; r = e.nextFast ()) {
//             if (Str.isWhitespace (doc.getText (r))) {
//                 if (newRecords == null)
//                     newRecords = records.modify ();
//                 newRecords = newRecords.delete (r);
//             }
//         }

        return sel;
    }

    //
    // SelectionMode interface
    //

    public void enteringMode (SelectionMode oldMode) {
        browser.addSidebarTab ("Simultaneous Editing", this);
        browser.getActivePane ().recordHighlighter.setSelection (range);
        browser.clearSelection ();
    }

    public void leavingMode (SelectionMode newMode) {
        clearExamples ();
        browser.getActivePane ().recordHighlighter.setSelection (null);
        browser.getActivePane ().outlierRecordHighlighter.setSelection (null);

        browser.removeSidebarTab (this);
    }

	public void selectRegion (RegionSet rs) {
		Document doc = browser.getDocument ();
		Region r = rs.getSingleDocRegionSet(doc).getBoundingBox();

		browser.getManualSelectionMode ().selectRegion (r);

		stopFindingOutlierRecords ();

		if (range.convertToLatest ()
			.intersects (r.contains ())) {
			if (firstPositiveExample == null)
				firstPositiveExample = r;
			setPositiveExamples (browser.getManualSelectionMode ().mergeSelection (positiveExamples, r));

			updateInference ();
		}

	}

	public void unselectRegion (RegionSet rs) {
		Document doc = browser.getDocument ();
		Region r = rs.getSingleDocRegionSet(doc).getBoundingBox();
		
		browser.getManualSelectionMode ().unselectRegion (r);

		if (range.convertToLatest ()
			.intersects (r.contains ())) {
			setNegativeExamples(new NamedRegionSet (negativeExamples.or (r)));
			if (positiveExamples.intersects (r))
				setPositiveExamples(new NamedRegionSet(positiveExamples.butnot (r)));
			updateInference ();
		}
	}

    //
    // BrowserListener interface
    //

    public void loadingDocument (Browser browser, SwingDocument doc) {
        browser.setSelectionMode (browser.getManualSelectionMode ());
    }

    public void loadedDocument (Browser browser, SwingDocument doc) {
    }

    //
    // SelectionListener interface
    //

    public void changedSelection (Browser browser, NamedRegionSet sel) {
        if (sel.isEmpty ())
            clearExamples ();
    }


    //
    // Inference
    //

    void updateInference () {
        if (positiveExamples.isEmpty ())
            return;

        browser.setHourglass (true);
                
        boolean success = false;
                
        try {
            RestartableEnumeration hypotheses = 
                inference.generalize (positiveExamples,
                                      negativeExamples);
            
            if (hypotheses.hasMoreElements ()) {
                patterns.setHypotheses (hypotheses);
                success = true;
            }
        } catch (Exception e) {
            debug.report (e);
        } finally {
            browser.setHourglass (false);
        }

        if (!success) {
            if (mouseSelection
                && browser.getSelection ().isEmpty ())
                getToolkit ().beep ();
                    
            browser.setPattern ("unknown selection", false);
        }

        findOutlierRecords ();
    }

    private HypothesisDisplay.SelectionListener selectionListener =
        new HypothesisDisplay.SelectionListener () {
            public void selectedHypothesis (NamedRegionSet hyp) {
                browser.setPattern (hyp.toString (), false);
                browser.setSelection (hyp);
            }
        };

    public void replaced (NamedRegionSet selection, NamedRegionSet newSelection,
                          TransferableSelection mss, String content) {
        SwingDocument doc = browser.getDocument ();

        Vector v = (mss != null) ? mss.getStrings () : null;
        int len = (content != null) ? content.length () : 0;

        if (mss != null || content != null) {
            // make a special coordinate map for
            // updating record highlight, because
            // inserts at the start of a record must be covered by
            // the record's highlight
            
            // TODO: support multiple documents
            Versions oldvers = selection.getVersions();
			RegionSet records = range.convert(oldvers);
                
            RegionEnumeration sel = selection.regions ();
            RegionEnumeration rec = records.regions ();
            lapis.maps.PiecewiseLinearMonotonicMap map = 
                new lapis.maps.PiecewiseLinearMonotonicMap (selection.getSize ()*2);
            map.add (0, 0);
            int adjust = 0;
            int i = 0;
            for (Region s = sel.firstFast (), r = rec.firstFast ();
                 s != null;
                 s = sel.nextFast (), r = rec.nextFast ()) {
                if (mss != null) {
                    content = (String) v.elementAt (i);
                    ++i;
                    if (i >= v.size ())
                        i = 0;
                    len = content.length ();
                }
                    
                boolean isPoint = 
                    (s.getStartLow () == s.getEndHigh ());
                boolean startsRecord = 
                    (r != null && s.getStartLow () == r.getStartLow ());
                    
                if (isPoint) {
                    if (startsRecord) {
                        map.add (s.getStartLow (), s.getStartLow () + adjust);
                        adjust += len;
                        map.add (s.getEndHigh ()+1, s.getEndHigh () + adjust + 1);
                    } else {
                        map.add (s.getStartLow ()-1, s.getStartLow () + adjust - 1);
                        adjust += len;
                        map.add (s.getEndHigh (), s.getEndHigh () + adjust);
                    }
                } else {
                    map.add (s.getStartLow (), s.getStartLow () + adjust);
                    adjust += len - (s.getEndHigh () - s.getStartLow ());
                    map.add (s.getEndHigh (), s.getEndHigh () + adjust);
                }
            }
                
            records = records.getSingleDocRegionSet(doc).convert(map, doc);
			range = new NamedRegionSet (records);
            browser.getActivePane ().recordHighlighter.setSelection (range);
                
            // convert outlierRecords highlight the same way
            NamedRegionSet outlierSel = browser.getActivePane ().outlierRecordHighlighter.getSelection ();
            if (outlierSel != null)
                browser.getActivePane ().outlierRecordHighlighter.setSelection
					(new NamedRegionSet (outlierSel.convert(oldvers).getSingleDocRegionSet(doc).convert (map, doc)));
        }

        if (mss != null)
            inference.addPasteLocation (newSelection, mss);
        else
            inference.addEditLocation (newSelection, len);
            
    }

    public void copied (NamedRegionSet selection) {
        inference.addCopyLocation (selection);
    }

    public void undidTo (DocumentVersion oldVersion) {
        inference.undoTo (oldVersion);
    }

    public void startMouseSelection () {
        mouseSelection = true;
    }

    public void stopMouseSelection () {
        mouseSelection = false;
    }

    public boolean showExampleHighlight () {
        return true;
    }

    public boolean clearSelectionOnClick () {
        return true;
    }

    // FIX: This code used to restart simulediting mode using the same record 
    // set but reparsing the document.  No longer connected to the UI, and
    // harder to implement now that we're dealing with a list of simuledit
    // objects.

//     void restart () {
//         browser.run (new RunnableWithExceptions () {
//                 public void run () throws Exception {
//                     browser.getDocument ().getPatternCache ().clear ();
//                     NamedRegionSet records = simulEdit.getRecords ();
//                     records = records.convert (browser.getDocument ().getVersion ());
//                     simulEdit = new SimulEdit (records);
//                 }
//             });
//     }

    void setPositiveExamples (NamedRegionSet sel) {
        positiveExamples = sel;
    }

    void setNegativeExamples (NamedRegionSet sel) {
        negativeExamples = sel;
    }

    void clearExamples () {
		setPositiveExamples (new NamedRegionSet (RegionSet.EMPTY));
		setNegativeExamples (new NamedRegionSet (RegionSet.EMPTY));
        firstPositiveExample = null;
    }

    void findOutlierRecords () {
        browser.getActivePane ().outlierRecordHighlighter.setSelection (null);

        if (!showOutliersCheckBox.isSelected ()
            || maxOutliers == 0)
            return;

        if (browser.getSelection ().getSize () < 2)
            return;
	    
        outlierThread = new Thread () {
                public void run () {
                    try {
                        SwingDocument doc = browser.getDocument ();
                        NamedRegionSet sel = browser.getSelection ();
                        Similarity sim = new Similarity (sel, new NamedRegionSet (RegionSet.EMPTY));

                        Thr.check ();

                        Similarity.Cluster[] clusters = sim.getClusters ();

                        Thr.check ();

                        
                        // take outliers down to half the distance of 
                        // outermost outlier
                        double cutoff = clusters[clusters.length - 1].getDistance() / 2;
                        MutableRegionSet outliers = new MutableRegionSet ();
                        int currMaxOutliers = Math.min (sel.getSize () / 2, maxOutliers);
                        for (int i = clusters.length - 1; 
                             i >= 0 && clusters[i].getDistance () > cutoff; 
                             --i) {
                            //RegionSet s = clusters[i].getRegions ();

                            // don't treat first positive example (or any of its
                            // similar outliers) as outliers
                            if (firstPositiveExample == null
                                || !clusters[i].intersects (firstPositiveExample))
                                outliers.insert (clusters[i]);
                            
                            if (outliers.getSize () > currMaxOutliers)
                                return;
                        }
                        
                        if (outliers.isEmpty ())
                            return;
                        
                        // get the record set and make sure it's up-to-date
                        RegionSet recs = inference.getRecords ().convertToLatest ();
                        
                        // find the record containing each outlier
                        final RegionSet outlierRecords = recs.and (outliers.contains ());
	    
                        Thr.check ();

                        final Thread me = Thread.currentThread ();

                        // highlight the outlier records
                        SwingUtilities.invokeLater (new Runnable () {
                                public void run () {
                                    if (outlierThread != me)
                                        return;

                                    SwingDocument odoc = browser.getDocument ();
                                    NamedRegionSet osel = new NamedRegionSet (outlierRecords);
                                    browser.getActivePane ().outlierRecordHighlighter
                                        .setSelection (osel);
                                    outlierThread = null;
                                }
                            });
                    } catch (Exception e) {
                        debug.report (e);
                    }
                }
            };
        outlierThread.setPriority (4);
        outlierThread.start ();
    }

    void stopFindingOutlierRecords () {
        if (outlierThread != null) {
            outlierThread.interrupt ();
            outlierThread = null;
        }
    }


    /**
     * Action that enters Simultaneous Editing mode.
     */
    public static class SimulEditAction extends AcceleratedAction {
        Browser browser;

        public SimulEditAction (Browser browser) {
            super ("&Simultaneous Editing",
                   SwingConfig.makeIcon ("simuledit.gif"),
                   'E', KeyEvent.CTRL_MASK,
                   "Simultaneous editing mode");
            this.browser = browser;
        }

        public void actionPerformed (ActionEvent event) {
            // since it takes some time to enter SimulEdit mode,
            // we'll revert the UI back to manual selection mode
            // just in case the user hits Stop
            browser.setSelectionMode (browser.getManualSelectionMode ());

            browser.run (new RunnableWithExceptions () {
                    public void run () throws Exception {
                        final SelectionMode sm = 
                            new SimulEditMode (browser);
                        Thr.check ();
                        SwingUtilities.invokeLater (new Runnable () {
                                public void run () {
                                    browser.setSelectionMode (sm);
                                }
                            });
                    }
                });
        }
    }


    /**
     * Action that clears the outlier flags on a record in
     * Simultaneous Editing mode.
     */
    public static class IgnoreOutlierAction extends AcceleratedAction {
        private Browser browser;

        /** Record that was right-clicked on */
        private Region clickedRecord = null;

        public IgnoreOutlierAction (Browser browser) {
            super ("This Record's NamedRegionSet Is OK");
            this.browser = browser;
            setEnabled (false);
        }

        public void setRecord (Region clickedRecord) {
            this.clickedRecord = clickedRecord;
            setEnabled (clickedRecord != null);
        }

        public void actionPerformed (ActionEvent evt) {
            if (clickedRecord == null)
                return;
            
            RegionHighlighter orh = 
                browser.getActivePane ().outlierRecordHighlighter;

            NamedRegionSet sel = orh.getSelection ();
            if (sel == null)
                return;
            
			sel = new NamedRegionSet (sel.butnot (clickedRecord));
            orh.setSelection (sel);
        }
    }


}
