/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapisx.util.Reflect;
import javax.swing.*;

/**
 * Utility methods for compatibility with older versions of Swing.
 * This code uses reflection to call Swing methods that are not available
 * in all JDK versions, or that are deprecated in newer versions
 * but still need to be called on older versions.
 */
public abstract class SwingCompatibility {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    /**
     * Disables buggy Swing focus manager on certain versions.
     * @effects On JDK 1.2/1.3, calls {@link
     * javax.swing.FocusManager#disableSwingFocusManager}
     * to prevent a bug that sends focus to the toolbar after using
     * menu commands.  On later versions of Java, does nothing.
     */
    public static void disableBuggyFocusManager () {
        String javaVersion = System.getProperty ("java.version");
        if (javaVersion.startsWith ("1.2")
            || javaVersion.startsWith ("1.3")) {
            try {
                // Call by reflection to avoid deprecation warning
                // on JDK 1.4+.
                Reflect.callStaticMethod ("javax.swing.FocusManager",
                                          "disableSwingFocusManager",
                                          new Class[] {},
                                          new Object[] {});
            } catch (Exception e) {}
        }
    }

    /**
     * Calls {@link javax.swing.JTabbedPane#setTabLayoutPolicy} if it's
     * available.
     * @param tabpane tabbed pane to call method on
     * @effects If tabpane provides the setTabLayoutPolicy
     * method (Java 1.4+), then tabpane.setTabLayoutPolicy
     * (SCROLL_TAB_LAYOUT) is called; see its docs for effects.
     * Otherwise, no effect.
     */
    public static void setScrollTabLayoutPolicy (JTabbedPane tabpane) {
        try {
            // call setTabLayoutPolicy() by reflection, since it's
            // only available in Java 1.4 and later
            Object SCROLL_TAB_LAYOUT =
                Reflect.getStaticField ("javax.swing.JTabbedPane",
                                        "SCROLL_TAB_LAYOUT");
            Reflect.callMethod (tabpane,
                                "setTabLayoutPolicy",
                                new Class[] { int.class },
                                new Object[] { SCROLL_TAB_LAYOUT });
        } catch (Exception e) {
            debug.report (e);
        }
    }


}
