/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.plaf.metal.*;
import java.net.*;
import lapisx.util.Str;
import lapisx.net.URLUtil;
import lapisx.swing.FontUtil;

/**
 * Utility methods that load and save Swing state
 * from the LAPIS configuration file ({@link lapis.Main#getConfig}).
 */
public abstract class SwingConfig {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    /**
     * Get the font size (in points) from the LAPIS configuration.
     * @throws NumberFormatException if configuration doesn't specify
     * a font size.
     */
    public static int getFontSize () throws NumberFormatException {
        String s = Main.getConfig ().getProperty ("browser.fontsize");
        if (s == null || s.length () == 0)
            throw new NumberFormatException ();
        return Integer.parseInt (s);
    }

    /**
     * Sets the look and feel specified in the LAPIS configuration file.
     * For best effect, call this method before any Swing 
     * components are made visible.
     */
    public static void restoreLookAndFeel () {
        String lookAndFeel = Main.getConfig ()
            .getProperty ("browser.look-and-feel", 
            UIManager.getSystemLookAndFeelClassName());
        try {
            UIManager.setLookAndFeel (lookAndFeel);
        } catch (Exception e) { 
            debug.report (e);
        }

        try {
            int size = getFontSize ();
            MetalTheme theme = FontUtil.makeResizedMetalTheme (size);
            MetalLookAndFeel.setCurrentTheme (theme);
        } catch (NumberFormatException e) {
        }
    }

    /**
     * Stores a component's location in the LAPIS configuration file.
     * @param name property name in config file
     * @param c component
     * @effects Sets property <i>name</i> to <i>x,y</i>, where
     * <i>x,y</i> is the point returned by c.getLocation().
     */
    public static void saveLocation (String name, Component c) {
        Point pt = c.getLocation ();
        Main.getConfig ().put (name, pt.x + "," + pt.y);
    }

    /**
     * Moves a component to a location specified in the LAPIS config file.
     * @param name property name in config file
     * @param c component to move
     * @effects If config file contains a property <i>name</i> of the
     * form <i>x,y</i>, then calls c.setLocation() to move c to that
     * position. Otherwise has no effect.
     */
    public static void restoreLocation (String name, Component c) {
        try {
            String s = Main.getConfig ().getProperty (name);
            c.setLocation (Integer.parseInt (Str.before (s, ",")),
                           Integer.parseInt (Str.after (s, ",")));
        } catch (Exception e) {
        }
    }
    
    /**
     * Stores a splitter's divider location in the LAPIS configuration file.
     * @param name property name in config file
     * @param splitter splitter
     * @effects sets property <i>name</i> to the value of 
     * splitter.getDividerLocation()
     */
    public static void saveSplitter (String name, JSplitPane splitter) {
        Main.getConfig ().put (name, splitter.getDividerLocation () + "");
    }

    /**
     * Moves a splitter's divider to a location specified in the
     * LAPIS config file.
     * @param name property name in config file
     * @param splitter splitter to adjust
     * @effects If config file contains a property <i>name</i> with an
     * integer value <i>x</i>, then splitter.setDividerLocation(x) is
     * called.  Otherwise has no effect.
     */
    public static void restoreSplitter (String name, JSplitPane splitter) {
        try {
            String s = Main.getConfig ().getProperty (name);
            splitter.setDividerLocation (Integer.parseInt (s));
        } catch (Exception e) {
        }
    }

    /**
     * Stores a window's size in the LAPIS configuration file.  The
     * stored size is only the window's client area, not including 
     * title and border.
     * @param name property name in config file
     * @param c window
     * @effects sets property <i>name</i> to <i>WxH</i> in
     * Main.getConfig.
     */
    public static void saveSize (String name, Window c) {
        Dimension d = c.getSize ();
        Insets border = c.getInsets ();

        d.width -= border.left + border.right;
        d.height -= border.top + border.bottom;

        Main.getConfig ().put (name, d.width + "x" + d.height);
    }

    /**
     * Resizes a window to a size specified in the LAPIS config file.
     * @param name property name in config file
     * @param c window to resize
     * @effects If config file contains a property <i>name</i> of the
     * form <i>WxH</i>, then calls c.setSize() to resize c to those
     * dimensions.  Otherwise has no effect.
     */
    public static void restoreSize (String name, Window c) {
        Insets border = c.getInsets ();
        try {
            String s = Main.getConfig ().getProperty (name);
            int w = Integer.parseInt (Str.before (s, "x"))
                + border.left + border.right;
            int h = Integer.parseInt (Str.after (s, "x")) 
                + border.top + border.bottom;
            c.setSize (w, h);
        } catch (Exception e) {
        }
    }

    /**
     * Test for demo mode.  Demo mode blocks some features of Browser,
     * such as saving files, saving settings, or deleting patterns
     * from the library.
     * @return true if LAPIS is in self-service demo mode.
     */
    public static boolean inDemoMode () {
        return "true".equals (Main.getConfig ()
                              .getProperty ("browser.demo-mode"));
    }

    /**
     * Loads an icon from the LAPIS images directory.
     * @param filename name of image file, relative to lapis/images/.
     * @return icon loaded from file
     */ 
    public static Icon makeIcon (String filename) {
        return new ImageIcon (Main.getLapisDir () + "/images/" + filename);
    }

    /**
     * Loads an image from the LAPIS images directory.
     * @param filename name of image file, relative to lapis/images/.
     * @return image loaded from file
     */ 
    public static Image makeImage (String filename) {
        return Toolkit.getDefaultToolkit ()
            .getImage (Main.getLapisDir () + "/images/" + filename);
    }

    /**
     * Makes a URL referring to a file in the LAPIS directory (if relative)
     * or anywhere (if absolute).
     * @param filename pathname (may be relative or absolute)
     * @return absolute file: URL.  Relative filenames are interpreted
     * with respect to the LAPIS directory.
     */
    public static URL makeURL (String filename) {
        return URLUtil.HrefToURL (filename, Main.getLapisDir ());
    }

    /**
     * Makes a URL referring to a file whose name is given by
     * a config property.
     * @param propname property name in config
     * @return URL for file given by the value of <i>name</i> in the
     * config file.
     */
    public static URL makeURLFromProperty (String propname) {
        return makeURL (Main.getConfig ().getProperty (propname));
    }

}
