/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;
import lapis.*;
import lapis.tc.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import lapisx.util.*;
import javax.swing.plaf.TextUI;
import lapisx.swing.*;

public class TCEditor extends JTextArea implements TCSymbols {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    boolean justOneExpression;
    TCToken rootToken = null;
    Hashtable tokenPositions = new Hashtable (); // maps token -> Position
    Hashtable firstTokens = new Hashtable (); // maps line Element -> TCToken

    BackgroundTokenizer backgroundTokenizer = new BackgroundTokenizer ();
    boolean tokenizeInBackground;

    int indenting = 0;
    Position exprStart = null;
    Position exprEnd = null;

    public TCEditor (boolean justOneExpression) {
        this (justOneExpression, true);
    }

    public TCEditor (boolean justOneExpression, boolean tokenizeInBackground) {
        this.justOneExpression = justOneExpression;
        this. tokenizeInBackground = tokenizeInBackground;

        setFont (new Font ("Monospaced", Font.PLAIN, 12));
        try {
            FontUtil.resizeFont (this, SwingConfig.getFontSize ());
        } catch (NumberFormatException e) {}

        getDocument ().addDocumentListener 
            (new javax.swing.event.DocumentListener () {
                public void changedUpdate (DocumentEvent event) {
                }
                public void insertUpdate (DocumentEvent event) {
                    if (indenting == 0
                           && event.getChange (getDocument ()
                                               .getDefaultRootElement ())
                        == null) {
                        int o = event.getOffset ();
                        int l = event.getLength ();
                        indentExpression (o, o+l);
                    }
                    retokenize ();
                }
                public void removeUpdate (DocumentEvent event) {
                    if (indenting == 0
                           && event.getChange (getDocument ()
                                               .getDefaultRootElement ())
                        == null) {
                        int o = event.getOffset ();
                        int l = event.getLength ();
                        indentExpression (o+l, o);
                    }

                    retokenize ();
                }
            });

        setHighlighter (new TCHighlighter ());

        // Bind editor keys
        Keymap keymap = JTextComponent.addKeymap (null, getKeymap ());
        keymap.addActionForKeyStroke
            (KeyStroke.getKeyStroke (KeyEvent.VK_ENTER, 0),
             newlineAction);
        keymap.addActionForKeyStroke
            (KeyStroke.getKeyStroke (KeyEvent.VK_ENTER, KeyEvent.SHIFT_MASK),
             newlineAction);
        keymap.addActionForKeyStroke
            (KeyStroke.getKeyStroke (KeyEvent.VK_ENTER, KeyEvent.CTRL_MASK),
             newlineAction);
        keymap.addActionForKeyStroke
            (KeyStroke.getKeyStroke (KeyEvent.VK_TAB, 0),
             tabAction);
        keymap.addActionForKeyStroke
            (KeyStroke.getKeyStroke (KeyEvent.VK_TAB, KeyEvent.SHIFT_MASK),
             backTabAction);
        keymap.addActionForKeyStroke
            (KeyStroke.getKeyStroke (KeyEvent.VK_BACK_SPACE, KeyEvent.SHIFT_MASK),
             backTabAction);
        setKeymap (keymap);

        setMargin (new Insets (2, 2, 2, 2));

        //setEditorKit (new TCEditorKit ());
        //setEditable (true);

        if (tokenizeInBackground)
            backgroundTokenizer.start ();
    }


    //
    // Tokenization
    //

    /*
     * Updates the tokenization.
     */
    void tokenize () {
        AbstractDocument doc = (AbstractDocument)getDocument ();
        doc.render (new Runnable () {
            public void run () {
                doTokenize ();
            }
        });
    }

    void doTokenize () {
//         debug.println ("tokenizing");

        TCToken rootToken;
        Hashtable tokenPositions = new Hashtable ();
        Hashtable firstTokens = new Hashtable ();

        // Tokenize the editor contents
        try {
            rootToken = TCParser.parseSyntax (new StringReader (getText ()),
                                              new TCParseException ());
        } catch (Exception e) {
            debug.report (e);
            rootToken = null;
        }

        javax.swing.text.Document doc = getDocument ();
        Element map = doc.getDefaultRootElement ();

        // Now scan the token tree to fill in the mappings.
        try {
            scanTokens (doc, map, rootToken, tokenPositions, firstTokens);
        } catch (BadLocationException e) {
            debug.report (e);
        }

        // Update the tokenization state
        synchronized (backgroundTokenizer) {
            this.rootToken = rootToken;
            this.tokenPositions = tokenPositions;
            this.firstTokens = firstTokens;
        }

        // Force all tree edges to repaint
        try {
            getUI().damageRange (TCEditor.this, 0, getDocument ().getLength ());
        } catch (NullPointerException e) {
            // debug.report (e);
        }

    }

    static void scanTokens (javax.swing.text.Document doc, 
                     Element map, TCToken tok,
                     Hashtable tokenPositions, Hashtable firstTokens) 
        throws BadLocationException {
        for (; tok != null; tok = tok.sibling) {
            tokenPositions.put (tok, doc.createPosition (tok.start));

            Element line = map.getElement (tok.line);
            if (firstTokens.get (line) == null)
                firstTokens.put (line, tok);

            scanTokens (doc, map, tok.child, tokenPositions, firstTokens);
        }
    }

    //
    // Background tokenization
    //

    void retokenize () {
        if (tokenizeInBackground) {
            synchronized (backgroundTokenizer) {
                backgroundTokenizer.retokenize = true;
                backgroundTokenizer.notify ();
            }
        } else
            doTokenize ();
    }

    class BackgroundTokenizer extends Thread {
        public boolean retokenize = true;

        public BackgroundTokenizer () {
            setPriority (2);
        }

        public void run () {
            while (true) {
                synchronized (this) {
                    if (!retokenize) {
                        try {
                            BackgroundTokenizer.this.wait ();
                        } catch (InterruptedException e) {
                        }
                    }
                    retokenize = false;
                }
                tokenize ();
            }
        }
    }


    //
    // Utility functions: translating between tokens, lines, columns, offsets
    //

    int getStartOffsetOfToken (TCToken tok) {
        synchronized (backgroundTokenizer) {
            Position p = (Position)tokenPositions.get (tok);
            return p.getOffset ();
        }
    }

    int getEndOffsetOfToken (TCToken tok) {
        return getStartOffsetOfToken (tok) + (tok.end - tok.start);
    }

    int getStartColumnOfToken (TCToken tok) {
        try {
            return getStartOffsetOfToken (tok) 
                    - getLineStartOffset (getLineOfToken (tok));
        } catch (BadLocationException e) {
            return 0;
        }
    }

    int getEndColumnOfToken (TCToken tok) {
        return getStartColumnOfToken (tok) + (tok.end - tok.start);
    }

    TCToken getFirstTokenOnLine (int line) {
        synchronized (backgroundTokenizer) {
            try {
                javax.swing.text.Document doc = getDocument ();
                Element rootElem = doc.getDefaultRootElement ();
                Element lineElem = rootElem.getElement (line);
                return (TCToken) firstTokens.get (lineElem);
            } catch (ArrayIndexOutOfBoundsException e) {
                return null;
            }
        }
    }

    int getLineOfToken (TCToken tok) throws BadLocationException {
        return getLineOfOffset (getStartOffsetOfToken (tok));
    }

    TCToken getTokenOfOffset (int offset) {
//         debug.println ("searching for " + offset);
        try {
            int line = getLineOfOffset (offset);
            
            for (TCToken tok = getFirstTokenOnLine (line);
                 tok != null && tok.line == line;
                 tok = tok.child) {
                int s = getStartOffsetOfToken (tok);
                int e = getEndOffsetOfToken (tok);
//                 debug.println (tok + " [" + s + "," + e + "]");
                if (s <= offset && offset <= e)
                    return tok;
            }
        } catch (BadLocationException e) {
        }

        return null;            
    }

    TCToken getNextTokenOnLine (int offset) {
//         debug.println ("searching for " + offset);
        try {
            int line = getLineOfOffset (offset);
            
            for (TCToken tok = getFirstTokenOnLine (line);
                 tok != null && tok.line == line;
                 tok = tok.child) {
                int s = getStartOffsetOfToken (tok);
                if (s > offset)
                    return tok;
            }
        } catch (BadLocationException e) {
        }

        return null;            
    }

    int getIndentLevel (int line) {
        try {
            int start = getLineStartOffset (line);
            int end = getLineEndOffset (line);
            String s = getText (start, end-start);
            for (int i = 0, n = s.length (); i < n; ++i)
                if (!Character.isWhitespace (s.charAt (i)))
                    return i;
        } catch (BadLocationException e) {
        }
        return Integer.MAX_VALUE;
    }


    //
    // Indentation
    //

    Action newlineAction = new AbstractAction ("Newline") {
        public void actionPerformed (ActionEvent evt) {
            insertNewline ();
        }
    };

    Action tabAction = new AbstractAction ("Tab") {
        public void actionPerformed (ActionEvent evt) {
            if (getSelectionStart () == getSelectionEnd ()) {
                // tab forward at cursor
                int amount = getTabAmount ();
                replaceSelection (Str.repeat (" ", amount));
            } else {
                // indent one or more full lines
                selectFullLines ();
                indentSelection (getTabAmount ());
            }
        }
    };

    Action backTabAction = new AbstractAction ("BackTab") {
        public void actionPerformed (ActionEvent evt) {
            if (getSelectionStart () == getSelectionEnd ()) {
                // unindent current line
                int amount = getBackTabAmount ();
                int pos = getSelectionStart ();
                replaceRange ("", pos - amount, pos);
            } else {
                // unindent one or more full lines
                selectFullLines ();
                indentSelection (-getBackTabAmount ());
            }
        }
    };

    void insertNewline () {
        int col = 0;
        boolean autoindent = true;
        try {
            int pos = getSelectionEnd ();

            for (int line = getLineOfOffset (pos);
                 line >= 0;
                 --line) {
                col = getIndentLevel (line);
                if (col < Integer.MAX_VALUE)
                    break;
            }
            if (col == Integer.MAX_VALUE) {
                autoindent = false;
                col = 0;
            }
        } catch (BadLocationException e) {
            debug.report (e);
        }
        
        ++indenting;
        try {
            // break the line and indent up to indent level of previous 
            // non-blank line
            replaceSelection ("\n" + Str.repeat (" ", col));

            if (autoindent) {
                // now simulate a Tab press
                int amount = getTabAmount ();
                replaceSelection (Str.repeat (" ", amount));
            }
        } finally {
            --indenting;
        }

    }

    void selectFullLines () {
        try {
            int start = getSelectionStart ();
            int end = getSelectionEnd ();
        
            int firstLine = getLineOfOffset (start);
            int lastLine = getLineOfOffset (end);

            int newStart = getLineStartOffset (firstLine);
            int newEnd =
                (end == getLineStartOffset (lastLine))
                ? getLineEndOffset (lastLine - 1) - 1
                : getLineEndOffset (lastLine) - 1;

            if (start != newStart || end != newEnd)
                select (newStart, newEnd);
        } catch (BadLocationException e) {
            debug.report (e);
        }
    }

    static final int TAB_SIZE = 4;

    int getTabAmount () {
        try {
            int start = getSelectionStart ();
            int end = getSelectionEnd ();
            int line = getLineOfOffset (start);
            int lineStart = getLineStartOffset (line);
            int col = start - lineStart;
            int indentLevel = getIndentLevel (line);

            if (col <= indentLevel) {
                // at or before first token on line.
                // jump cursor ahead to first token, and consider
                // adaptive tab stops.
                if (indentLevel < Integer.MAX_VALUE) {
                    col = indentLevel;
                    if (start == end)
                        select (lineStart + col, lineStart + col);
                }

                // Fetch the adaptive tab stops for this line
                // (computed from the tokens on previous lines
                // that might line up with it)
                int[] tabStops = getTabStops (line);
                
                // Look for the next tab stop to the right
                for (int i = 0; i < tabStops.length; ++i) {
                    if (tabStops[i] != -1 && tabStops[i] > col)
                        return tabStops[i] - col;
                }
            }

            // Couldn't find an adaptive tab stop,
            // just advance to the computed TAB_SIZE tab stop
            return (col/TAB_SIZE + 1) * TAB_SIZE
                - col;
        } catch (BadLocationException e) {
            debug.report (e);
            return 0;
        }
    }

    int getBackTabAmount () {
        try {
            int start = getSelectionStart ();
            int end = getSelectionEnd ();
            int line = getLineOfOffset (start);
            int lastLine = getLineOfOffset (end);
            int col = getIndentLevel (line);

            if (start == end) {
                // no selection
                int lineStart = getLineStartOffset (line);
                if (col == Integer.MAX_VALUE)
                    // line is blank -- just leave cursor where it is
                    col = start - lineStart;
                else
                    // line is nonblank: jump cursor to first token on line
                    select (lineStart + col, lineStart + col);
            } else {
                // selection: advance to first non-blank line in selection
                while (col == Integer.MAX_VALUE) {
                    if (++line > lastLine)
                        return 0; // selection is all blank lines
                    col = getIndentLevel (line);
                }
            }

            // Fetch the adaptive tab stops for this line
            // (computed from the tokens on previous lines
            // that might line up with it)
            int[] tabStops = getTabStops (line);
            
            // Look for the next tab stop to the left of col
            int amount = 0;
            for (int i = tabStops.length-1; i >= 0; --i) {
                if (tabStops[i] != -1 && tabStops[i] < col) {
                    amount = col - tabStops[i];
                    break;
                } 
            }

            // Clip amount by the indent level of all lines in the selection
            while (amount > 0 && ++line <= lastLine)
                amount = Math.min (amount, getIndentLevel (line));

            return amount;

        } catch (BadLocationException e) {
            debug.report (e);
            return 0;
        }
    }


    // Compute tab stops by reconstructing the token stack at this line.
    int[] getTabStops (int line) {
        synchronized (backgroundTokenizer) {
            Stack stack = TCParser.reconstructStackAtLine (rootToken, line);
            int n = stack.size ();
            int[] tabStops = new int[2*n + 1];
            int j = 0;
            for (int i = 0; i < n; ++i) {
                TCToken tok = (TCToken)stack.elementAt (i);
                int col = getStartColumnOfToken (tok);
                switch (tok.sym) {
                case PREFIX:
                    tabStops[j++] = col;
                    tabStops[j++] = col + 4; // 4 is a typical indent
                    break;
                case EITHER:
                    tabStops[j++] = col;
                    tabStops[j++] = col + 4; // strlen("either") - strlen ("or")
                    break;
                case FROM:
                    tabStops[j++] = col;
                    tabStops[j++] = col + 2;// strlen("from") - strlen ("to")
                    break;
                case BALANCES:
                    tabStops[j++] = col;
                    tabStops[j++] = col + 4; // strlen("balances") - strlen ("with")
                    break;
                case IS:
                    break;
                default:
                    tabStops[j++] = col;
                    break;
                }
            }

            try {
                TCToken tos = (TCToken)stack.peek ();
                tabStops[j++] = getEndColumnOfToken (tos) + 1;
            } catch (EmptyStackException e) {
            }

            while (j < tabStops.length)
                tabStops[j++] = -1;

//             for (j = 0; j < tabStops.length; ++j)
//                 debug.println ("tab[" + j + "]=" + tabStops[j]);

            return tabStops;
        }
    }


    void indentExpression (int offsetBeforeChange, int offsetAfterChange) {
        try {
            // compute the range of lines spanned by the expression
            // starting at offset
            int firstLine = getLineOfOffset (offsetAfterChange);
            int lastLine;

            if (exprStart != null 
                && exprStart.getOffset () == offsetAfterChange) {
                // use stored firstLine..endLine
                debug.println ("using stored exprStart/exprEnd");
                lastLine = getLineOfOffset (exprEnd.getOffset ());
            } else {
                // must compute firstLine..endLine
                debug.println ("finding end of expression");

                TCToken firstTok = getNextTokenOnLine (offsetAfterChange);
                if (firstTok == null)
                    lastLine = firstLine;
                else
                    lastLine = getLineOfToken (firstTok.getLastLeaf ());

                // store firstLine..endLine as Positions
                javax.swing.text.Document doc = getDocument ();
                exprStart = doc.createPosition (offsetAfterChange);
                exprEnd = doc.createPosition (getLineEndOffset (lastLine) - 1);
            }

            debug.println ("firstLine=" + firstLine + ", lastLine=" + lastLine);
            if (firstLine < lastLine) {
                final int firstIndentedLine = firstLine + 1;
                final int lastIndentedLine = lastLine;
                final int amount = offsetAfterChange - offsetBeforeChange;

                SwingUtilities.invokeLater (new Runnable () {
                    public void run () {
                        ++indenting;
                        try {
                            indentLines (firstIndentedLine, 
                                         lastIndentedLine, 
                                         amount);
                        } finally {
                            --indenting;
                        }
                    }
                });
            }
        } catch (BadLocationException e) {
            debug.report (e);
        }
    }

    void indentSelection (int amount) {
        if (amount == 0)
            return;

        try {
            int pos = getSelectionStart ();
            int lastPos = getSelectionEnd ();
            int firstLine = getLineOfOffset (pos);
            int lastLine = getLineOfOffset (lastPos);
            
            ++indenting;
            try {
                indentLines (firstLine, lastLine, amount);
            } finally {
                --indenting;
            }

            selectFullLines ();

        } catch (BadLocationException e) {
            debug.report (e);
        }
    }

    void indentLines (int firstLine, int lastLine, int amount) {
        try {
            for (int line = firstLine; line <= lastLine; ++line) {
                int pos = getLineStartOffset (line);
                if (amount > 0)
                    insert (Str.repeat (" ", amount), pos);
                else
                    replaceRange ("", pos, pos - amount);
            }
        } catch (BadLocationException e) {
            debug.report (e);
        }
    }


    //
    // Drawing tree edges
    //

    class TCHighlighter extends DefaultHighlighter {

        Color TREE_COLOR = new Color (160, 160, 255);
        int HALF_SPACE_WIDTH = -1;

        public TCHighlighter () {
            super ();
        }

        public void paint (Graphics g) {
            super.paint (g);

            if (HALF_SPACE_WIDTH == -1)
                HALF_SPACE_WIDTH = 
                    getFontMetrics (getFont ()).charWidth (' ') 
                    / 2;

            synchronized (backgroundTokenizer) {
                try {
                    g.setColor (TREE_COLOR);
                    paintTreeEdges (g, rootToken, getVisibleRegion ());
                } catch (BadLocationException e) {
                }
            }
        }

        protected void paintTreeEdges (Graphics g, 
                                       TCToken parent,
                                       Region visible) throws BadLocationException {
            for (; parent != null; parent = parent.sibling) {
                int parentStart = getStartOffsetOfToken (parent);
                int parentEnd = getEndOffsetOfToken (parent);
                
                if (parentStart > visible.getEnd ())
                    return;

                paintTreeEdges (g, parent.child, visible);

//                 if (parent.operand != null && parent.operand.sibling != null) {
//                     // Draw the operand tree, like this:
//                     // [PARENT] -+- [OP1]
//                     //           +- [OP2]
//                     // The vertical line is the "trunk"; the
//                     // horizontal lines are "branches"
//                     Rectangle rParentEnd = getViewPosition 
//                         (parentEnd, Position.Bias.Backward);
//                     int xTrunk = rParentEnd.x + HALF_SPACE_WIDTH;
//                     int y = rParentEnd.y + rParentEnd.height/2 - 1;
//                     int yMin = y;
//                     int yMax = y;

//                     // Draw branch from parent to trunk
//                     g.drawLine (rParentEnd.x, y, xTrunk, y);

//                     // Draw branches from trunk to each operand
//                     for (TCToken operand = parent.operand;
//                          operand != null;
//                          operand = operand.sibling) {
//                         Rectangle rOperandStart = getViewPosition 
//                             (getStartOffsetOfToken (operand),
//                              Position.Bias.Forward);
//                         y = rOperandStart.y + rOperandStart.height/2 - 1;
//                         if (y >= yMax) {
//                             yMax = y;
//                             g.drawLine (xTrunk, y, rOperandStart.x - 1, y);
//                         }
//                     }

//                     // Draw trunk
//                     g.drawLine (xTrunk, yMin, xTrunk, yMax);
//                 }

                if (parent.child != null) {
                    // Draw the children tree, like this:
                    // [PARENT]
                    //  +- [CHILD1]
                    //  +- [CHILD2]
                    // As before, the vertical line is the trunk,
                    // and the horizontal lines are branches.

                    Rectangle rParentStart = getViewPosition 
                        (parentStart, Position.Bias.Forward);
                    int xTrunk = rParentStart.x + HALF_SPACE_WIDTH;
                    int y = rParentStart.y + rParentStart.height - 1;
                    int yMin = y;
                    int yMax = y;
                    
                    TCToken child = parent.child;

                    // Skip children on the same line as parent
                    while (child != null && child.line == parent.line)
                        child = child.sibling;

                    // Skip children inside parentheses
                    if (parent.sym == LEFTPAREN
                        || parent.sym == LEFTCURLY
                        || parent.sym == CLOSED_LEFT) {
                        while (child != null) {
                            if (child.sym == CLOSED_RIGHT) {
                                child = child.sibling;
                                break;
                            } else
                                child = child.sibling;
                        }
                    }

                    // Draw branches from trunk to each child
                    for (; child != null; child = child.sibling) {
                        Rectangle rChildStart = getViewPosition 
                            (getStartOffsetOfToken (child),
                             Position.Bias.Forward);
                        y = rChildStart.y + rChildStart.height/2 - 1;
                        yMax = y;
                        g.drawLine (xTrunk, y, rChildStart.x - 1, y);
                    }

                    // Draw trunk
                    if (yMin < yMax)
                        g.drawLine (xTrunk, yMin, xTrunk, yMax);
                }
            }
        }

        // returns an interval representing the visible portion of the TC
        // code
        Region getVisibleRegion () {
            TextUI ui = getUI ();

            // visible rectangle in screen coordinates
            Rectangle visibleRect = getVisibleRect ();

            // start and end Swing document positions corresponding to 
            // upper-left and lower-right corners of visibleRect
            int startVis = ui.viewToModel 
                (TCEditor.this, new Point (visibleRect.x, visibleRect.y));
            int endVis = ui.viewToModel 
                (TCEditor.this, new Point (visibleRect.x + visibleRect.width,
                                           visibleRect.y + visibleRect.height));

            return new Region (startVis, endVis, lapis.Document.DEFAULT); 
        }

        protected Rectangle getViewPosition (int pos, Position.Bias bias) 
            throws BadLocationException {
            TextUI mapper = getUI();
            return mapper.modelToView (TCEditor.this, pos, bias);
        }
    }

    public static void selectParseError (JTextComponent textbox, 
                                         TCParseException e) {
        int len = textbox.getText ().length ();
        int start = Math.max (0, Math.min (len, e.getStart ()));
        int end = Math.max (0, Math.min (len, e.getEnd ()));
        textbox.select (start, end);
        if (start == end) {
            // insert a marker to make error position visible
            textbox.replaceSelection ("???");
            textbox.select (start, end+3);
        }
        textbox.requestFocus ();        
    }
}
