/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;
import lapis.tc.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import lapisx.swing.*;

public abstract class ToolAction extends AcceleratedAction {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    Browser browser;
    Tool tool;
    BrowserArguments args;
    boolean usesSelection;

    public ToolAction (Browser browser, Tool tool, String label, boolean usesSelection) {
        super (label);
        this.browser = browser;
        this.tool = tool;
        this.usesSelection = usesSelection;
    }

    public class BrowserArguments extends Arguments {
        String command;

        public BrowserArguments () {
            super (browser.getInterpreter ());
        }
    }

    public abstract BrowserArguments makeArguments ();

    public NamedRegionSet getSelection () {
        return browser.getSelection ();
    }

    public String quote (TC pat) {
        return pat != null
            ? quote (pat.getExpression ())
            : "selection";
    }

    public String quote (String s) {
        if (s == null)
            return "selection";

        String t = browser.getInterpreter ().quote (s);
        debug.println ("quoted '" + s + "' to get '" + t + "'");
        return t;
    }


    public void actionPerformed (ActionEvent event) {
        browser.stop ();

        if (usesSelection && getSelection ().isEmpty ()) {
            browser.getToolkit ().beep ();
            return;
        }

        args = makeArguments ();
        if (args == null)
            return;

        // Set the interpreter's input document
        browser.getInterpreter ().setVar (Interpreter.DOCUMENT_VAR, 
                                          browser.getDocument ());

        browser.run (new RunnableWithExceptions () {
                public void run () throws Exception {
                    final Document result = tool.invoke (args);
                    if (result instanceof StatusWriter) {
                        ((StatusWriter) result).close ();
                    } else if (result != null) {
                        if (args.command != null) {
                            if (result.getProperty (SwingDocument.CommandProperty) == null)
                                result.putProperty (SwingDocument.CommandProperty, 
                                                    args.command);
                        }

                        SwingUtilities.invokeLater (new Runnable () {
                                public void run () {
                                    if (args.command != null)
                                        browser.setCommand (args.command, true);
                                    browser.setDocument (result);
                                    browser.getActivePane ().requestFocus ();
                                }
                            });
                    }
                }
            });
    }

    static class ToolboxAction extends ToolAction {
        public ToolboxAction (Browser browser, Toolbox.Info info) {
            super (browser, info.getTool (), info.getName (), false);

            String tooltip = info.getDescription ();
            if (tooltip != null)
                putValue (SHORT_DESCRIPTION, tooltip);

            File iconFile = info.getIcon ();
            if (iconFile != null)
                putValue (SMALL_ICON, new ImageIcon (iconFile.toString ()));
        }
        
        public BrowserArguments makeArguments () {
            return new BrowserArguments ();
        }
    }


    static class ToolboxChange implements Toolbox.ToolBoxListener { 
	Browser browser;

        public ToolboxChange (Browser browser) {
            this.browser = browser;
        }

	public void toolAdded(Toolbox toolBox, Toolbox.Info info){
	    debug.println("Calling " + info.getName() + " to toolbar.");
	    browser.addToToolBar(info);
	    
	}
	
	public void toolChanged(Toolbox toolBox, Toolbox.Info info){
	    //	    browser.removeFromToolBar(info);
	    //	    browser.addToToolBar(info);
	    browser.changeInTool(info);
	    
	}
	public void toolRemoved(Toolbox toolBox, Toolbox.Info info){
	    debug.println("Calling " + info.getName() + " to toolbar.");
	    browser.removeFromTools(info);
	}
    }



    /**
     * Action for Keep tool.
     */
    public static class Keep extends ToolAction {
        public Keep (Browser browser) {
            super (browser, new lapis.tools.Keep (), "&Keep", true);
        }

        public BrowserArguments makeArguments () {
            NamedRegionSet sel = getSelection ();
            BrowserArguments args = (BrowserArguments) new BrowserArguments ()
                .add ("query", sel)
                .add ("outof", sel.getPattern ().getUniversePattern ())
                ;
            args.command = "keep " + quote (sel.getPattern ());
            return args;
        }
    }


    /**
     * Action for Sort tool.  Sorts selection
     * to create a new document, and displays that.
     */
    public static class Sort extends ToolAction {
        public Sort (Browser browser) {
            super (browser, new lapis.tools.Sort (), "&Sort...", true);
        }

        public BrowserArguments makeArguments () {
            SortDialog dialog = new SortDialog ();
            dialog.show ();
            return dialog.args;
        }

        class SortDialog extends OkCancelDialog {
            ButtonGroup keyGroup;
            JRadioButton useEntireRecord;  // use entire record as sort key
            JRadioButton usePattern;       // use specified pattern as sort key
            TCEditor keyPattern;
        
            JComboBox order;            // alphabetic, numeric, etc.        
            BrowserArguments args;             // null until created by ok()
        
            public SortDialog () {
                super (browser, "Sort");
            
                Layout layout = Layout.gridbag (getControlsPane ())
                    .setAlignment (Layout.NORTHWEST)
                    .setSize (Layout.FIXED_SIZE);

                layout
                    .add (new JLabel ("Sort Key: "))
                    .add (useEntireRecord = new JRadioButton ("Full record"))
                    .nextRow ()
                    .nextColumn ()
                    .add (usePattern = new JRadioButton ("Part of record matching: "))
                    .nextRow ()
                    .nextColumn ()
                    .add (keyPattern = new TCEditor (true), Layout.ANY_SIZE)
                    .nextRow ();

                keyGroup = new ButtonGroup ();
                keyGroup.add (useEntireRecord);
                keyGroup.add (usePattern);
                useEntireRecord.setSelected (true);
                keyPattern.setEnabled (false);

                keyPattern.addMouseListener (new MouseAdapter () {
                        public void mousePressed (MouseEvent event) {
                            usePattern.setSelected (true);
                            keyPattern.setEnabled (true);
                            keyPattern.requestFocus ();
                        }
                    });
                ActionListener enabler = new ActionListener () {
                        public void actionPerformed (ActionEvent event) {
                            boolean usingPattern = usePattern.isSelected ();
                            keyPattern.setEnabled (usingPattern);
                            if (usingPattern)
                                keyPattern.requestFocus ();
                        }
                    };
                useEntireRecord.addActionListener (enabler);
                usePattern.addActionListener (enabler);

                String yourLanguage = 
                    java.util.Locale.getDefault ().getDisplayLanguage ();
                if (yourLanguage.equals (""))
                    yourLanguage = "Dictionary";
                layout
                    .add (new JLabel ("Order: "))
                    .add (order = new JComboBox (new String[] {
                        yourLanguage + " (a b c ...)",
                        "Reverse " + yourLanguage + " (z y x ...)", 
                        "Numeric (1 2 3 ...)", 
                        "Reverse Numeric (9 8 7 ...)",
                        "Unicode (A B C ... a b c ...)", 
                        "Reverse Unicode (z y x ... Z Y X ...)",
                        "Random" }))
                    .nextRow ();
                order.setEditable (false);

                SortDialog.this.pack ();
            }

            public void ok () {
                try {
                    // Create arguments
                    NamedRegionSet sel = getSelection ();
                    args = (BrowserArguments)new BrowserArguments ()
                        .add ("query", sel)
                        ;
                    args.command = "sort " + quote (sel.getPattern ());

                    if (usePattern.isSelected ()) {
                        TC keys = PatternPane.makeTCFromTextComponent (keyPattern,
                                                                   true);
                        args.add ("by", keys);
                        args.command += " -by " + quote (keyPattern.getText ());
                    }

                    boolean reversed;
                    String orderCode;
                    args.addName ("order");
                    switch (order.getSelectedIndex ()) {
                    case 0:
                        reversed = false;
                        orderCode = "dictionary";
                        break;
                    case 1:
                        reversed = true;
                        orderCode = "dictionary";
                        break;
                    case 2:
                        reversed = false;
                        orderCode = "numeric";
                        break;
                    case 3:
                        reversed = true;
                        orderCode = "numeric";
                        break;
                    case 4:
                        reversed = false;
                        orderCode = "unicode";
                        break;
                    case 5:
                        reversed = true;
                        orderCode = "unicode";
                        break;
                    case 6:
                        reversed = false;
                        orderCode = "random";
                        break;
                    default:
                        throw new RuntimeException ("illegal sort order: " + order.getSelectedIndex ());
                    }

                    if (reversed)
                        args.addValue ("reverse");
                    args.addValue (orderCode);

                    if (! ("dictionary".equals (orderCode) && !reversed))
                        args.command += " -order " + (reversed ? "reverse " : "") + orderCode;
                

                    super.ok ();
                } catch (TCParseException e) {
                    FeedbackForm.showMessageDialog (browser, 
                                                    e.getMessage (), 
                                                    "Parse Error",
                                                    JOptionPane.ERROR_MESSAGE);
                    args = null;
                }
            }
        }
    }


    /**
     * Action for Replace tool.
     */
    public static class Replace extends ToolAction {
        public Replace (Browser browser) {
            super (browser, new lapis.tools.Replace (), "&Replace...", true);
        }

        public BrowserArguments makeArguments () {
            ReplaceDialog dialog = new ReplaceDialog ();
            dialog.show ();
            return dialog.args;
        }

        class ReplaceDialog extends OkCancelDialog {

            JTextArea template;
            JCheckBox isTemplate;
            BrowserArguments args;  // null until created by ok()
        
            public ReplaceDialog () {
                super (browser, "Replace");
            
                Layout.gridbag (getControlsPane ())
                    .setAlignment (Layout.WEST)
                    .add (new JLabel ("Replace Each Selection With: "), Layout.FIXED_SIZE)
                    .nextRow ()
                    .add (template = new JTextArea (5, 40))
                    .nextRow ()
                    .add (isTemplate = 
                          new JCheckBox ("Substitute curly braces", false))
                    .nextRow ()
                    ;
                ReplaceDialog.this.pack ();
            }
        
            public void ok () {
                try {
                    // Configure arguments
                    NamedRegionSet sel = getSelection ();
                    args = (BrowserArguments) new BrowserArguments ()
                        .add ("query", sel)
                        ;
                    String tclCommand = "replace " + quote (sel.getPattern ());

                    try {
                        String tmplt = template.getText ();
                        args.add ("replace",
                                  lapis.tools.Template.makeFromString (tmplt));
                        tclCommand += " " + quote (tmplt);
                    } catch (TCParseException e) {
                        // Select the error, then re-throw the exception
                        TCEditor.selectParseError (template, e);
                        throw e;
                    }

                    args.command = tclCommand;

                    super.ok ();
                } catch (TCParseException e) {
                    FeedbackForm.showMessageDialog (browser, 
                                                    e.getMessage (), 
                                                    "Parse Error",
                                                    JOptionPane.ERROR_MESSAGE);
                    args = null;
                }
            }
        }
    }



    /**
     * Action for Delete tool (called "Omit" in UI).
     */
    public static class Omit extends ToolAction {
        public Omit (Browser browser) {
            super (browser, new lapis.tools.Delete (), "&Omit", true);
        }

        public BrowserArguments makeArguments () {
            NamedRegionSet sel = getSelection ();
            BrowserArguments args = (BrowserArguments) new BrowserArguments ()
                .add ("query", sel)
                ;
            args.command = "omit " + quote (sel.getPattern ());
            return args;
        }
    }


    /**
     * Action for Extract tool.
     */
    public static class Extract extends ToolAction {
        public Extract (Browser browser) {
            super (browser, new lapis.tools.Extract (), "&Extract", true);
        }

        public BrowserArguments makeArguments () {
            NamedRegionSet sel = getSelection ();
            BrowserArguments args =
                (BrowserArguments) new BrowserArguments ()
                .add ("query", sel)
                ;
            args.command = "extract " + quote (sel.getPattern ());
            return args;
        }
    }


    /**
     * Action for Calc tool.
     */
    public static class Calc extends ToolAction {
        public Calc (Browser browser) {
            super (browser, new lapis.tools.Calc (), "&Calc", true);
        }

        public BrowserArguments makeArguments () {
            NamedRegionSet sel = getSelection ();
            BrowserArguments args = (BrowserArguments) new BrowserArguments ()
                .add ("query", sel)
                ;
            args.command = "calc " + quote (sel.getPattern ());
            return args;
        }
    }

}

