/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing;

import lapis.*;

import java.awt.event.*;
import lapisx.swing.*;

import java.util.*;

public abstract class TransformAction extends AcceleratedAction {
	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

	Browser browser;

	public TransformAction(Browser browser, String label) {
		super(label);
		this.browser = browser;
	}

	public void actionPerformed(ActionEvent evt) {
		NamedRegionSet selection = browser.getSelection();
		if (!selection.isAllPoints()) {
			TransferableSelection mss = new TransferableSelection(selection);
			Vector v = mss.getStrings();
			for (int i = 0, n = v.size(); i < n; ++i) {
				String s = (String)v.elementAt(i);
				String t = transform(s);
				v.setElementAt(t, i);
			}
			browser.getActivePane().replace(selection, mss, null, true);
		}
	}

	protected abstract String transform(String s);

	/**
	 * Convert to lowercase action.
	 */
	public static class Downcase extends TransformAction {
		public Downcase(Browser browser) {
			super(browser, "&lowercase");
		}

		protected String transform(String s) {
			return s.toLowerCase();
		}
	}

	/**
	 * Convert to uppercase action.
	 */
	public static class Upcase extends TransformAction {
		public Upcase(Browser browser) {
			super(browser, "&UPPERCASE");
		}

		protected String transform(String s) {
			return s.toUpperCase();
		}
	}

	/**
	 * Capitalize first word action.
	 */
	public static class SentenceCase extends TransformAction {
		public SentenceCase(Browser browser) {
			super(browser, "Capitalize &first word");
		}

		protected String transform(String s) {
			StringBuffer result = new StringBuffer();
			boolean capitalized = false;
			for (int j = 0, m = s.length(); j < m; ++j) {
				char c = s.charAt(j);
				if (!capitalized && (Character.isUpperCase(c) || Character.isLowerCase(c))) {
					result.append(Character.toUpperCase(c));
					capitalized = true;
				}
				else
					result.append(c);
			}
			return result.toString();
		}
	}

	/**
	 * Capitalize all words action.
	 */
	public static class TitleCase extends TransformAction {
		public TitleCase(Browser browser) {
			super(browser, "Capitalize &All Words");
		}

		protected String transform(String s) {
			StringBuffer result = new StringBuffer();
			boolean startOfWord = true;
			for (int j = 0, m = s.length(); j < m; ++j) {
				char c = s.charAt(j);
				if (Character.isUpperCase(c) || Character.isLowerCase(c)) {
					if (startOfWord)
						c = Character.toUpperCase(c);
					startOfWord = false;
				}
				else
					startOfWord = true;
				result.append(c);
			}
			return result.toString();
		}
	}

	/**
	 * Toggle case action.
	 */
	public static class ToggleCase extends TransformAction {
		public ToggleCase(Browser browser) {
			super(browser, "t&OGGLE cASE");
		}

		protected String transform(String s) {
			StringBuffer result = new StringBuffer();
			for (int j = 0, m = s.length(); j < m; ++j) {
				char c = s.charAt(j);
				if (Character.isUpperCase(c))
					result.append(Character.toLowerCase(c));
				else
					result.append(Character.toUpperCase(c));
			}
			return result.toString();
		}
	}

	public static class PreserveCase extends TransformAction {
		public PreserveCase(Browser browser) {
			super(browser, "PreserveCase");
		}
		protected String transform(String s) {
			return transform(s, false);
		}
		protected String transform(String s, boolean replaceCase) {
			String result = s;
			if (s.equals(s.toUpperCase())) {
				return s.toUpperCase();
			}
			else {

				// if removed word is only lowercased and no special case in replace string
				if (s.equals(s.toLowerCase()) && !replaceCase)
					return s.toLowerCase();
				// removed word has some special capitalization
				else {
					int j = 0;
					while ((Character.isWhitespace(s.charAt(j))) && j < s.length())
						j++;
					if (j == s.length())
						j = 0;
					//					// only pay attention if first word is capitalized
					if (Character.isUpperCase(s.charAt(j)) && !replaceCase) {
						TransformAction transform = new TransformAction.SentenceCase(browser);
						return transform.transform(s);
					}
					else
						if (!Character.isUpperCase(s.charAt(j)) && !replaceCase)
							return s.toLowerCase();
				}
			}
			return result;
		}
	}
}
