/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing.cluster;

import lapis.*;
import lapis.ml.*;
import lapis.swing.*;
import java.util.*;
import lapisx.sort.*;
import lapis.swing.cluster.MatchNodeCreator.Node;

public class ClusterNodeTableModel extends AbstractTreeTableModel implements TreeTableModel {
	Vector data;
	Vector column;
	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

	public ClusterNodeTableModel(Browser browser, RegionSet rs) {
		//this(browser, new NamedRegionSet(browser.getDocument(), rs));
		this(browser, new NamedRegionSet(rs));
	}

	public ClusterNodeTableModel(Browser browser, NamedRegionSet sel) {
		this(new ClusterTableCreator(browser, sel));
	}

	public ClusterNodeTableModel(ClusterTableCreator ctc) {
		super(null);
		root = ctc.getRoot();
		data = ctc.getData();
		column = ctc.getColumn();
	}
	public ClusterNodeTableModel(Node root) {
		this(new ClusterTableCreator(root));
	}

	//
	// The TreeModel interface
	//

	/**
	 * Returns the number of children of <code>node</code>.
	 */
	public int getChildCount(Object node) {
		return ((Node)node).getChildCount();
	}

	/**
	 * Returns the child of <code>node</code> at index <code>i</code>.
	 */
	public Object getChild(Object node, int i) {
		return ((Node)node).children.get(i);
	}

	/**
	 * Returns true if the passed in object represents a leaf, false
	 * otherwise.
	 */
	public boolean isLeaf(Object node) {
		return ((Node)node).isLeaf();
	}

	//
	//  The TreeTableNode interface. 
	//

	/**
	 * Returns the number of columns.
	 */
	public int getColumnCount() {
		return column.size();
	}

	/**
	 * Returns the name for a particular column.
	 */
	public String getColumnName(int column) {
		return (String)this.column.get(column);
	}

	/**
	 * Returns the class for the particular column.
	 */
	public Class getColumnClass(int column) {
		switch (column) {
			case 0 :
				return TreeTableModel.class;
			case 1 :
				return String.class;

		}
		return null;
	}

	/**
	 * Returns the value of the particular column.
	 */
	public Object getValueAt(Object node, int column) {
		Node n = (Node)node;

		switch (column) {
			case 0 :
				return n;
			case 1 :
				String str = "";
				if (n.getChildCount() > 0) {
					str = new String("" + n.getChildCount());
					if (n.getSelectedChildrenCount() > 0)
						str = str + "(" + n.getSelectedChildrenCount() + " selected)";
				}
				return str;
		}
		return null;
	}

	interface NodeTableCreator {

		public Vector getData();

		public Vector getColumn();

		public Node getRoot();
	}

	static class ClusterTableCreator implements NodeTableCreator {
		Vector data = new Vector();
		Vector column = new Vector();
		Node root;

		public ClusterTableCreator(Browser browser, RegionSet rs) {
			//this(browser, new NamedRegionSet(browser.getDocument(), rs));
			this(browser, new NamedRegionSet(rs));
		}

		public ClusterTableCreator(Browser browser, NamedRegionSet sel) {
			ClusterCreator cc = new ClusterCreator(browser, sel);
			this.root = cc.getRoot();
			for (Enumeration e = root.children(); e.hasMoreElements();) {
				Node child = (Node)e.nextElement();
				Vector v = new Vector();
				v.add(child);
				v.add(new Integer(child.getChildCount()));
				data.add(v);
			}
			column.add("description");
			column.add("size");

		}
		public ClusterTableCreator(Node root) {
			this.root = root;
			for (Enumeration e = root.children(); e.hasMoreElements();) {
				Node child = (Node)e.nextElement();
				Vector v = new Vector();
				v.add(child);
				v.add(new Integer(child.getChildCount()));
				data.add(v);
			}
			column.add("description");
			column.add("size");

		}

		public Vector getData() {
			return data;
		}

		public Vector getColumn() {
			return column;
		}
		public Node getRoot() {
			return root;
		}
	}
	interface NodeCreator {

		public Node getRoot();

	}

	static class ClusterCreator implements NodeCreator {
		Node root;
		Similarity sim;
		Similarity.Cluster[] clusters;
		Browser browser;
		NamedRegionSet sel;

		public ClusterCreator(Browser browser, RegionSet rs) {
			this(browser, new NamedRegionSet(rs));
		}

		public ClusterCreator(Browser browser, NamedRegionSet sel) {
			this.sel = sel;
			this.browser = browser;
			debug.println("getting Pos/Neg");
			int sucess = getPosNeg();
			debug.println("setting root");
			root = new Node(browser, "**root**", new NamedRegionSet(Region.EMPTY));
			if (sucess == 1) {
				Region r = sel.regions().first();
				NamedRegionSet rs = new NamedRegionSet(r);
				Node match = new Node(browser, "**root**." + "match #1 " + r.toString(), rs);
				root.add(match);
			}
			else {
				debug.println("" + clusters.length + " clusters found");
				int[] clusterSizeArray = new int[clusters.length];
				for (int j = 0; j < clusters.length; j++) {
					clusterSizeArray[j] = clusters[j].getSize();
				}
				PointerSort pointerArray = new PointerSort(clusterSizeArray);
				pointerArray.sort();
				int[] pArray = pointerArray.pointerArray();

				for (int j = 0; j < clusters.length; j++) {
					RegionSet matches = clusters[pArray[pArray.length - j - 1]];

					if (matches.getSize() < 2) {
						RegionEnumeration e = matches.regions();
						int k = 0;
						for (Region r = e.first(); r != null; r = e.next()) {
							Node match = null;
							NamedRegionSet rs = new NamedRegionSet(r);
							match = new Node(browser, "Unclusterable.match #" + k + " " + r.toString(), rs);
							match.setExplanation(clusters[pArray[pArray.length - j - 1]].getExplanation());
							root.add(match);
							k++;
						}
					}
					else {
						Node header = new Node(browser, "cluster #" + j, new NamedRegionSet(matches));
						header.setExplanation(clusters[pArray[pArray.length - j - 1]].getExplanation());
						root.add(header);
						int i = 0;
						RegionEnumeration e = matches.regions();
						for (Region r = e.first(); r != null; r = e.next()) {
							Node match = null;
							NamedRegionSet rs = new NamedRegionSet(r);
							match = new Node(browser, header.fullname + "." + "match #" + i + " " + r.toString(), rs);
							header.add(match);
							i++;
						}
					}
				}
			}
		}

		private int getPosNeg() {
			if (sel.getVersions().getSize() == 0 || sel == null) {
				setExamples(null, null);
				debug.println("doc null or posExamples null");
				return 0;
			}
			if (sel.getSize() <= 1) {
				setExamples(null, null);
				debug.println("too small posExamples");
				return 1;
			}
			debug.println("setting examples");
			setExamples(sel, new NamedRegionSet(RegionSet.EMPTY));
			return 100;

		}

		private void setExamples(NamedRegionSet positive, NamedRegionSet negative) {
			if (positive == null && negative == null) {
				sim = null;
				clusters = new Similarity.Cluster[0];
				debug.println("pos and neg examples null");
				return;
			}

			sim = new Similarity(positive, negative);
			debug.println("getting clusters");
			clusters = sim.getClusters();
		}

		public Node getRoot() {
			return root;
		}

	}

}
