/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */
package lapis.swing.cluster;
import lapis.Parser;
import lapis.PatternLibrary;
import java.util.*;
import java.io.*;
import lapis.*;
import lapis.Document;
import lapis.maps.*;
import lapis.maps.PiecewiseLinearMonotonicMap;
/**
 * @author patrick winston 
 * @modified by alisa_m
 *
 */
public class LexicalAttractionParser implements Parser {
	private String name = null;
	private HashMap sets = null; //map of MutableRegionSet objects
	private LexMap lexicalMap = null;
	
	long wordsProcessed = 0;
	Map wordTable = new HashMap();
	Map connectionTable = new HashMap();
	// As prescribed by Yuret; need some non-zero conditional probability to start:
	double defaultConditionalProbability = 0.0001;
	// Debugging switches, normally false:
	boolean graftSwitch;
	boolean memoryReferenceSwitch;
	boolean recursiveBaseSwitch;
	boolean splittingSwitch;
	boolean bestResultSwitch;
	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;
	/**
	 * Analyzes a sentence or group of sentences, delivered as a single string.
	 * Hands off to a parse method that parses an array of words.
	 * Returns a parse object, which describes final sentence parsed (possibly
	 * null);
	 */
	public Parse parseSentence(String sentence) {
		// If the "sentence" has a period, question mark, colon, or semicolon inside, split and recurse:
		int period = sentence.indexOf('.');
		int semicolon = sentence.indexOf(';');
		int colon = sentence.indexOf(':');
		int question = sentence.indexOf('?');
		int splitAt = Math.max(Math.max(Math.max(period, semicolon), colon), question);
		if (splitAt >= 0) {
			Parse first = parseSentence(sentence.substring(0, splitAt));
			Parse second = parseSentence(sentence.substring(splitAt + 1));
			if (second != null) {
				return second;
			}
			else {
				return first;
			}
		}
		// Otherwise, break sentence up using a tokenizer, and stick words in a list:
		LinkedList list = new LinkedList();
		try {
			StreamTokenizer st = new StreamTokenizer(new StringReader(sentence));
			st.slashStarComments(true);
			st.slashSlashComments(true);
			st.lowerCaseMode(true);
			while (st.nextToken() != StreamTokenizer.TT_EOF) {
				// If a number, can't handle it; forget the sentence entirely:
				if (st.ttype == StreamTokenizer.TT_NUMBER) {
					return null;
				}
				// Otherwise, add word to the list:
				list.add(st.sval);
			}
		}
		catch (Exception e) {
		}
		// Now, if there are no words, return:
		if (list.size() == 0) {
			return null;
		}
		// Otherwize, covert the list into a properly sized array, and parse:
		String[] array = new String[list.size()];
		for (int i = 0; i < list.size(); ++i) {
			array[i] = (String) (list.get(i));
		}
		return parse(array, true);
	}
	/**
	* does an initial pairwise parse
	 */
	private void parsePairwise(String[] sentence) {
		Parse parse = new Parse(sentence);
		int[] pairwise = new int[sentence.length];
		for (int i = 0; i < pairwise.length; i++)
			pairwise[i] = i - 1;
		parse.setLinks(pairwise);
		incrementCounts(parse.getWords(), parse.getLinks());
	}
	/**
	 * Parses a sentence, delivered as an array of words.  This method hands off
	 * the array to a method that takes more arguments and recurses.
	 */
	private Parse parse(String[] sentence, boolean increment) {
		// Covert all words lower case:
		for (int i = 0; i < sentence.length; ++i) {
			sentence[i] = sentence[i].toLowerCase();
		}
		// Create array for remembering partial results; need one for each way
		// tree can root, to left or to right:
		Parse[][] leftRootMemorandum = new Parse[sentence.length][sentence.length];
		Parse[][] rightRootMemorandum = new Parse[sentence.length][sentence.length];
		// Find best parse recursively, using a table of accumulating partial results
		Parse result = parse(sentence, 0, sentence.length - 1, leftRootMemorandum, rightRootMemorandum);
		// Adjust counts in word frequency and conditional word frequency tables:
		if (increment)
			incrementCounts(result.getWords(), result.getLinks());
		// Return parse structure
		return result;
	}
	/**
	 * Adds a depth-of-parse variable, used only in debugging.
	 */
	private Parse parse(
		String[] sentence,
		int start,
		int end,
		Parse[][] leftRootMemorandum,
		Parse[][] rightRootMemorandum) {
		return parseWithLeftRoot(sentence, start, end, 0, leftRootMemorandum, rightRootMemorandum);
	}
	private LexMap parse(String content, DocumentView doc) {
		debug.println("*************starting parse**************");
		// make document map
		// after every parse set locations
		// add to document map 
		StringReader sr = new StringReader(content);
		BufferedReader br = new BufferedReader(sr);
		LexMap lexMap = new LexMap(doc);
		int location = 0;
		try {
			String sentence = br.readLine();
			Parse bestParse;
			while (sentence != null) {
                            // this only worked in JDK 1.4
                            // String[] words = sentence.split(" ");
                            StringTokenizer tok = new StringTokenizer (sentence, " ");
                            String[] words = new String[tok.countTokens ()];
                            for (int i = 0; i < words.length; ++i)
                                words[i] = tok.nextToken ();

				//				for (int i = 0; i < words.length; i++)
				//					debug.println(words[i]);
				bestParse = parse(words, true);
				bestParse.setLocations(location);
				int[] startLocations = bestParse.getStartLocations();
				int[] endLocations = bestParse.getEndLocations();
				int[] links = bestParse.getLinks();
				String[] parseWords = bestParse.getWords();
				for (int i = 0; i < parseWords.length; i++) {
					String word = parseWords[i];
					int start = startLocations[i];
					int end = endLocations[i];
					int index = links[i];
					String lexLink = "root";
					if (index >= 0) {
						lexLink = words[index];
					}
					lexMap.add(word, start, end, lexLink);
				}
				for (int i = 0; i < startLocations.length; i++)
					debug.println(content.substring(startLocations[i], endLocations[i]));
				String testsentance = content.substring(location, location + sentence.length());
				location = location + sentence.length() + 1;
				debug.println(bestParse.toString() + "\n" + sentence + "\n" + testsentance);
				sentence = br.readLine();
			}
		}
		catch (IOException ioe) {
			debug.println("can't read line");
		}
		return lexMap;
	}
	/**
	 * Here is where the real work is done, by finding the best way to
	 * recursively divide the sentence.  Let's represent the sentence as a..z.
	 * Further, we know from Yuret that any word can be taken as the root of the
	 * parse tree, so for the moment, we might as well use the leftmost word, a.
	 * Hence, the procedure is parseWithLeftRoot.
	 * <p> We also know from Yuret that the probability of a parse is given by
	 * processing each word as follows in the parse tree: form a product in
	 * which each word contributes a term.  The term for the root word is the
	 * probability of the root word, which is the same for all parses, so we can
	 * take its probability to be one, because we are only interested in finding
	 * the highest-probability parse releative to the others.  The term for all
	 * other words is the conditional probability of the word, given its parent
	 * in the parse tree.
	 * <p> Now, the word a, in the best possible parse, may be linked to several
	 * other words in the sentence, but there must be some word connected to
	 * word a that is the furthest-to-the-right of all such words.  Call that
	 * word i.
	 * <p> For the moment, suppose i is not z.  Then the words after a and before
	 * i cannot link to the words after i, because that would break the
	 * no-link-crossing principle.  Word a also cannot link to a word after i
	 * because i was presumed to be the final word that is linked to a.
	 * <p> So, now we have divided the sentence into two parts that overlap by
	 * one word, i; words up to and including i, a..i, form a tree rooted in a;
	 * words after i have to link into the final parse through i, thus forming a
	 * tree rooted in i, i..z.  Plainly we have broken up the sentence into two
	 * parts, each of which has fewer words than the whole sentence, so the
	 * recursion is gauranteed to stop.
	 * <p> Note that the probability of the whole parse is given by the product
	 * of the probabilities of the two parts.  This is because the second tree
	 * is rooted in i, and contributes a term of one in that trees probability
	 * calculation.  The probability of the first part includes a term that is
	 * the conditional probability of i given its parent in the first part's
	 * tree.  Thus, word i and all other words are properly accounted for in the
	 * probability calculation by multiplying the probabilities of the two parts.
	 * <p> So, at this point, it would be just a matter of trying all i, and
	 * picking the best, except there is the possibility that a actually does
	 * link to z.  Too bad; we cannot use the method just described in that
	 * case, because the number of words in the next level of recursion would be
	 * the same as at the current level, hence the recursion would not stop.
	 * <p> But wait!  Considering the tree as a whole, there must be some words,
	 * other than z, that link ultimately to a, and others that link ultimately
	 * to z.  Suppose i is the rightmost word that links ultimately to a.  None
	 * of the words before i can link to z, because that would violate the
	 * no-link-crossing principle.  Similarly, none of the words after i can
	 * link to a because i was supposed to be the rightmost such word.  Thus, we
	 * have two independant trees, a..i and i+1..z.  a..i is constructed so as
	 * to be rooted in a; i+1..z is constructed so as to be rooted in z.
	 * <p> What then is the probability of this combination?  In the previous
	 * situation, there was a shared word, i, which contributed a probability of
	 * just the right sort, enabling the combined probability of the two trees
	 * to work out correctly.
	 * <p> Now, however, the right side part is assumed to be rooted in z, and
	 * z's parent is a.  Accordingly, the probability of the whole tree is the
	 * probability of the first part, multiplied by the probability of the
	 * second part, multiplied by the conditional probability of z, given a,
	 * which is the parent of z in this configuration.  Again, all words are
	 * properly accounted for in the probability calculation.
	 * <p> What is the base of the recursion?  You ultimately ground out a tree
	 * of one or two words.  A one word tree is the root, so its probability is
	 * taken to be one.  If, when combined, it is still the root, it still
	 * contributes one, as it should.  If not, its conditional probability
	 * enters into the combined probability combination, and again all works out.
	 * <p>What about a two-word tree.
	 * Such a tree is rooted in the first word or second word, so the probability
	 * of interest is the conditional probablility of the second word, given the
	 * first, or vice versa.
	 * <p> So, the best parse will be the best assuming that a links to an
	 * intermediate word in a sequence, i, or the final word, z.  Checking both
	 * possibilities iteratively ultimately yields the best parse.
	 * <p> Of course, as all this goes on, the sentence will be broken up a great
	 * deal, multiply requiring the same computation to be performed on a
	 * subpart of the sentence.  It is best to store these computations as they
	 * are performed in a two dimentional array, so that remembered values can
	 * be looked up rather than recalculated.  There are two dimensions because
	 * any subpart of the sentence will have a starting index and a concluding
	 * index.
	 * <p> Now, a note on data structure.  All words have a parent except one.
	 * Thus, you can capture the structure of the tree with an array of integers,
	 * in which the index identifies a word and the value the word's parent,
	 * except for the root, which has no parent, and therefore has a value of
	 * -1, by convention, in the representation used here.
	 * <p> When two trees are brought together, the structure-capturing arrays
	 * are merged as described in the documentaiton of the tree-grafting
	 * methods, which all keep merged trees correctly rooted.
	 */
	private Parse parseWithLeftRoot(
		String[] sentence,
		int start,
		int end,
		int depth,
		Parse[][] leftRootMemorandum,
		Parse[][] rightRootMemorandum) {
		// Create a string with spaces equal to recursion depth, for printing during debugging:
		String printPrefix = "";
		for (int i = 0; i < depth; ++i) {
			printPrefix += " ";
		}
		// Check to see if best parse has been recorded already; if so, return
		Parse result = leftRootMemorandum[start][end];
		if (result != null) {
			println(memoryReferenceSwitch, "Found remembered parsed phrase at " + start + ", " + end);
			return result;
		}
		// Create new parse, given that none was found; first extract words:
		String[] words = new String[end - start + 1];
		for (int i = 0; i < words.length; ++i) {
			words[i] = sentence[start + i];
		}
		// Then create object:
		result = new Parse(words);
		// Check to see if there are only one word, in which case, it must be a root
		if (end == start) {
			println(recursiveBaseSwitch, "Just one word in left root parse.");
			result.setProbability(1.0);
			result.rootOneWord();
			return result;
		}
		// Check to see if there are only two words, in which case, set probability and link up
		if (end == 1 + start) {
			println(recursiveBaseSwitch, "Just two words in left root parse.");
			// If just two words, probability is conditional probability of second given first
			result.setProbability(pFirstGivenSecond(sentence[end], sentence[start]));
			result.rootPairToLeft();
			return result;
		}
		// Record new parse in the parse storage array (probability and links will be set in the course of analysis):
		leftRootMemorandum[start][end] = result;
		// Set probability of parse to a negative number, so at least one parse triggers action, a hack:
		double bestProbabilitySoFar = -0.1;
		// Find best parse in which first node is linked to an intermediate node
		for (int i = start + 1; i < end; ++i) {
			// Calculate best parses of two pieces; both parts root in their leftmost words:
			Parse firstParse =
				parseWithLeftRoot(sentence, start, i, 1 + depth, leftRootMemorandum, rightRootMemorandum);
			Parse secondParse = parseWithLeftRoot(sentence, i, end, 1 + depth, leftRootMemorandum, rightRootMemorandum);
			// Calculate probability of this parse
			double product = firstParse.getProbability() * secondParse.getProbability();
			println(splittingSwitch, "Trying intermediate " + firstParse + ", " + secondParse);
			println(
				splittingSwitch,
				"...which yields "
					+ firstParse.getProbability()
					+ ", "
					+ secondParse.getProbability()
					+ ", "
					+ product);
			// See if this is better that best of previous parses:
			if (product > bestProbabilitySoFar) {
				// If so, note improvment...
				bestProbabilitySoFar = product;
				// ... and set probability and  merge trees
				result.setProbability(product);
				result.setLinks(graftLeftRootedTreeToLeftRootedTree(firstParse.getLinks(), secondParse.getLinks()));
				println(
					splittingSwitch,
					printPrefix + "Intermediate improvement, " + i + ", <" + result.getProbability() + ">");
			}
		}
		// Find best parse in which first node is linked to the final node
		for (int i = start; i < end; ++i) {
			// Calculate best parses of two pieces; left side roots left, right side roots right:
			Parse firstParse =
				parseWithLeftRoot(sentence, start, i, 1 + depth, leftRootMemorandum, rightRootMemorandum);
			Parse secondParse =
				parseWithRightRoot(sentence, i + 1, end, 1 + depth, leftRootMemorandum, rightRootMemorandum);
			// Calculate probability of rightmost word, given leftmost word.
			double p = pFirstGivenSecond(sentence[end], sentence[start]);
			// Calculate probability of this parse
			double product = p * firstParse.getProbability() * secondParse.getProbability();
			println(splittingSwitch, "Trying terminal " + firstParse + ", " + secondParse);
			println(
				splittingSwitch,
				"...which yields "
					+ firstParse.getProbability()
					+ ", "
					+ secondParse.getProbability()
					+ ", "
					+ p
					+ ", "
					+ product);
			// See if this is better that best of previous parses:
			if (product > bestProbabilitySoFar) {
				// If so, note improvment...
				bestProbabilitySoFar = product;
				// ... and set probability and  merge trees
				result.setProbability(product);
				result.setLinks(graftRightRootedTreeToLeftRootedTree(firstParse.getLinks(), secondParse.getLinks()));
				println(
					splittingSwitch,
					printPrefix + "Terminal improvement, " + i + ", <" + result.getProbability() + ">");
			}
		}
		println(bestResultSwitch, printPrefix + "Best observed: " + result + ", <" + result.getProbability() + ">");
		return result;
	}
	/**
	 * Parses; see documentation for parseWithLeftRoot.  This is the same
	 * method, except that left and right rooting are reversed, so this ends up
	 * parsing with the root at the rightmost word.  Probabably, the two version
	 * should be combined, with another argument specifying left or right
	 * rooting.
	 */
	private Parse parseWithRightRoot(
		String[] sentence,
		int start,
		int end,
		int depth,
		Parse[][] leftRootMemorandum,
		Parse[][] rightRootMemorandum) {
		// Create a string with spaces equal to recursion depth, for printing during debugging:
		String printPrefix = "";
		for (int i = 0; i < depth; ++i) {
			printPrefix += " ";
		}
		// Check to see if best parse has been recorded already; if so, return
		Parse result = rightRootMemorandum[start][end];
		if (result != null) {
			println(memoryReferenceSwitch, "Found remembered parsed phrase at " + start + ", " + end);
			return result;
		}
		// Create new parse, given that none was found; first extract words:
		String[] words = new String[end - start + 1];
		for (int i = 0; i < words.length; ++i) {
			words[i] = sentence[start + i];
		}
		// Then create object;
		result = new Parse(words);
		// Check to see if there are only one word, in which case, it must be a root
		if (end == start) {
			println(recursiveBaseSwitch, "Just one word in right root parse.");
			result.setProbability(1.0);
			result.rootOneWord();
			return result;
		}
		// Check to see if there are only two words, in which case, set probability and link up
		if (end == 1 + start) {
			println(recursiveBaseSwitch, "Just two words in right root parse.");
			// If just two words, probability is conditional probability of first given second
			result.setProbability(pFirstGivenSecond(sentence[start], sentence[end]));
			result.rootPairToRight();
			return result;
		}
		// Record new parse in the parse storage array (probability and links will be set in the course of analysis):
		// probability and links will be set in the course of analysis, of course:
		rightRootMemorandum[start][end] = result;
		// Set probability of parse to a negative number, so at least one parse triggers action, a hack:
		double bestProbabilitySoFar = -0.1;
		// Find best parse in which first node is linked to an intermediate node
		for (int i = start + 1; i < end; ++i) {
			// Calculate best parses of two pieces; both parts root in their rightMost words:
			Parse firstParse =
				parseWithRightRoot(sentence, start, i, 1 + depth, leftRootMemorandum, rightRootMemorandum);
			Parse secondParse =
				parseWithRightRoot(sentence, i, end, 1 + depth, leftRootMemorandum, rightRootMemorandum);
			// Calculate probability of this parse
			double product = firstParse.getProbability() * secondParse.getProbability();
			println(splittingSwitch, "Trying intermediate " + firstParse + ", " + secondParse);
			println(
				splittingSwitch,
				"...which yields "
					+ firstParse.getProbability()
					+ ", "
					+ secondParse.getProbability()
					+ ", "
					+ product);
			// See if this is better that best of previous parses:
			if (product > bestProbabilitySoFar) {
				// If so, note improvment...
				bestProbabilitySoFar = product;
				// ... and set probability and  merge trees
				result.setProbability(product);
				result.setLinks(graftRightRootedTreeToRightRootedTree(firstParse.getLinks(), secondParse.getLinks()));
				println(
					splittingSwitch,
					printPrefix + "Intermediate improvement, " + i + ", <" + result.getProbability() + ">");
			}
		}
		// Find best parse in which first node is linked to the final node
		for (int i = start; i < end; ++i) {
			// Calculate best parses of two pieces; left side roots left, right side roots right:
			Parse firstParse =
				parseWithLeftRoot(sentence, start, i, 1 + depth, leftRootMemorandum, rightRootMemorandum);
			Parse secondParse =
				parseWithRightRoot(sentence, i + 1, end, 1 + depth, leftRootMemorandum, rightRootMemorandum);
			// Calculate probability of leftmost word, given rightmost word.
			double p = pFirstGivenSecond(sentence[start], sentence[end]);
			// Calculate probability of this parse
			double product = p * firstParse.getProbability() * secondParse.getProbability();
			println(splittingSwitch, "Trying terminal " + firstParse + ", " + secondParse);
			println(
				splittingSwitch,
				"...which yields "
					+ firstParse.getProbability()
					+ ", "
					+ secondParse.getProbability()
					+ ", "
					+ p
					+ ", "
					+ product);
			// See if this is better that best of previous parses:
			if (product > bestProbabilitySoFar) {
				// If so, note improvment...
				bestProbabilitySoFar = product;
				// ... and set probability and  merge trees
				result.setProbability(product);
				result.setLinks(graftLeftRootedTreeToRightRootedTree(firstParse.getLinks(), secondParse.getLinks()));
				println(
					splittingSwitch,
					printPrefix + "Terminal improvement, " + i + ", <" + result.getProbability() + ">");
			}
		}
		println(bestResultSwitch, printPrefix + "Best observed: " + result + ", <" + result.getProbability() + ">");
		return result;
	}
	/**
	 * Grafts.  a..i i..z  i is root of left side, z is root of right side.  z is final root.
	 */
	private int[] graftRightRootedTreeToRightRootedTree(int[] left, int[] right) {
		// Size is one less than sum of sizes, because joined by common element:  
		int[] result = new int[left.length + right.length - 1];
		// First elements come from left tree:
		for (int i = 0; i < left.length - 1; ++i) {
			result[i] = left[i];
		}
		// Remaining elements come from right tree, except for first element:
		for (int i = left.length - 1; i < result.length - 1; ++i) {
			// Of course, element values have to be offset:
			result[i] = right[i - left.length + 1] + left.length - 1;
		}
		// Of course, right most is still root:
		result[result.length - 1] = -1;
		println(graftSwitch, "RightRight: " + stringify(left) + ", " + stringify(right) + " --> " + stringify(result));
		return result;
	}
	/**
	 * Grafts.  a..i i..z  a is root of left side, i is root of right side.  a is final root.
	 */
	private int[] graftLeftRootedTreeToLeftRootedTree(int[] left, int[] right) {
		// Size is one less than sum of sizes, because joined by common element:  
		int[] result = new int[left.length + right.length - 1];
		// First elements come from left tree:
		for (int i = 0; i < left.length; ++i) {
			result[i] = left[i];
		}
		// Remaining elements come from right tree, except for first element:
		for (int i = left.length; i < result.length; ++i) {
			// Of course, element values have to be offset:
			result[i] = right[i - left.length + 1] + left.length - 1;
		}
		println(graftSwitch, "LeftLeft: " + stringify(left) + ", " + stringify(right) + " --> " + stringify(result));
		return result;
	}
	/**
	 * Grafts.  a..i i+1..z  a is root of left side, z is root of right side.  a is final root.
	 */
	private int[] graftRightRootedTreeToLeftRootedTree(int[] left, int[] right) {
		// Size is sum of sizes:
		int[] result = new int[left.length + right.length];
		// First elements come from left tree:
		for (int i = 0; i < left.length; ++i) {
			result[i] = left[i];
		}
		// Remaining elements come from right tree
		for (int i = left.length; i < result.length; ++i) {
			// Of course, element values have to be offset:
			result[i] = right[i - left.length] + left.length;
		}
		// And for the finishing touch, rightmost word links to leftmost word
		result[result.length - 1] = 0;
		println(graftSwitch, "RightLeft: " + stringify(left) + ", " + stringify(right) + " --> " + stringify(result));
		return result;
	}
	/**
	 * Grafts.  a..i i+1..z  a is root of left side, z is root of right side.  z is final root.
	 */
	private int[] graftLeftRootedTreeToRightRootedTree(int[] left, int[] right) {
		// Size is sum of sizes:
		int[] result = new int[left.length + right.length];
		// First elements come from left tree:
		for (int i = 0; i < left.length; ++i) {
			result[i] = left[i];
		}
		// Remaining elements come from right tree
		for (int i = left.length; i < result.length; ++i) {
			result[i] = right[i - left.length] + left.length;
		}
		// And for the finishing touch, leftmost word links to rightmost word
		result[0] = result.length - 1;
		// Of course, right most is still root:
		result[result.length - 1] = -1;
		println(graftSwitch, "LeftRight: " + stringify(left) + ", " + stringify(right) + " --> " + stringify(result));
		return result;
	}
	/**
	 * Returns probability of a word by looking into count table.
	 */
	private double p(String word) {
		if (wordsProcessed == 0) {
			// No words process, so cannot have a probability; use a small number:
			return defaultConditionalProbability;
		}
		// Get count:
		Object o = wordTable.get(word);
		// If not observed, again return a small number:
		if (o == null) {
			return defaultConditionalProbability;
		}
		double result = ((Long) o).longValue();
		// Divide count by all words and return;
		return result / wordsProcessed;
	}
	/**
	 * Returns conditional probability of a word by looking into count tables.
	 */
	private double pFirstGivenSecond(String first, String second) {
		long secondCount = getCount(second);
		// Returns small number if no word:
		if (secondCount == 0) {
			return defaultConditionalProbability;
		}
		long firstWhenSecondCount = getConnectedCount(first, second);
		// Returns small number if no word:
		if (firstWhenSecondCount == 0) {
			return defaultConditionalProbability;
		}
		// Divide conditioned first count by second count and return:
		return (double) firstWhenSecondCount / secondCount;
	}
	/**
	 * Returns joint probability; not used.
	 */
	private double pJoint(String x, String y) {
		double result = pFirstGivenSecond(x, y) * p(y);
		if (result == 0.0) {
			return defaultConditionalProbability;
		}
		return result;
	}
	/**
	 * Gets number of times a word has appeared.
	 */
	private long getCount(String word) {
		return getCount(word, wordTable);
	}
	/**
	 * Gets number of times a word appears as determined by given table.
	 */
	private long getCount(String word, Map table) {
		Object o = table.get(word);
		if (o == null) {
			return 0;
		}
		else
			if (o instanceof Long) {
				Long l = (Long) o;
				return (l.longValue());
			}
			else {
				System.err.println("Found object of wrong type in word table.");
			}
		return 0;
	}
	/**
	 * Records number of times a word has appeared.
	 */
	private void putCount(String word, long l) {
		putCount(word, wordTable, l);
	}
	/**
	 * Records number of times a word has appeared into a given table.
	 */
	private void putCount(String word, Map table, long l) {
		table.put(word, new Long(l));
	}
	/**
	 * Gets table of conditioned appearances relative to given word.
	 */
	private Map getMap(String word) {
		Object o = connectionTable.get(word);
		if (o == null) {
			Map table = new HashMap();
			connectionTable.put(word, table);
			return table;
		}
		return (Map) o;
	}
	/**
	 * Increments table of word counts and tables of conditional word counts.
	 */
	private void incrementCounts(String[] words, int[] links) {
		for (int i = 0; i < links.length; ++i) {
			int target = links[i];
			incrementCount(words[i]);
			if (target >= 0) {
				// increment 1st degree
				incrementConnections(words[i], words[target]);
				// increment 2nd degree
				int secondTarget = links[target];
				if (secondTarget >= 0)
					incrementConnections(words[i], words[secondTarget]);
			}
		}
	}
	/**
	 * Increments count of word appearances.  Includes increment of total words
	 * processed.
	 */
	private void incrementCount(String word) {
		++wordsProcessed;
		putCount(word, getCount(word) + 1);
	}
	/**
	 * Increments conditional word counts.
	 */
	private void incrementConnections(String wordA, String wordB) {
		putConnectedCount(wordA, wordB, getConnectedCount(wordA, wordB) + 1);
		putConnectedCount(wordB, wordA, getConnectedCount(wordB, wordA) + 1);
	}
	/**
	 * Notes number of times connected word has been linked to given word.
	 */
	private void putConnectedCount(String connected, String given, long l) {
		Map table = getMap(given);
		putCount(connected, table, l);
	}
	/**
	 * Gets number of times connected word appears linked to given word.
	 */
	private long getConnectedCount(String connected, String given) {
		Map table = getMap(given);
		return getCount(connected, table);
	}
	/**
	 * Prints an integer array in readable form.
	 */
	private String stringify(int[] array) {
		String result = "[";
		for (int i = 0; i < array.length - 1; ++i) {
			result += array[i] + ", ";
		}
		result += array[array.length - 1] + "]";
		return result;
	}
	/**
	 * Prints for debugging, used in liu of non-functional logger.
	 */
	private void println(boolean x, Object o) {
		if (x) {
			System.out.println(o);
		}
	}
	
	/**
	 * Parses document
	 * @param doc
	 */
	public void parse(lapis.Document doc) {
		long start = System.currentTimeMillis();
		name = "lapis.swing.cluster.LexicalAttractionParser";
		sets = new HashMap(); //state => RegionSet
		try {
			PatternCache cache = doc.getPatternCache();
			//make plain text doc without any html

//			LexicalAttractionParser parser = new LexicalAttractionParser();
			lapis.DocumentView view = cook(doc);
			CoordinateMap map = view.getVersion().diff(doc.getVersion());
			String content = view.getText();
			//			Main.getProgress().fireStartProgress("External Parser", content.length());
			LexMap lexMap = parse(content, view);
			LexMap convertedLexMap = new LexMap(doc);
			Iterator i = lexMap.keySet().iterator();
			Region r = null;
			while (i.hasNext()) {
				r = (Region) i.next();
				LexMap.LexNode ln = (LexMap.LexNode) lexMap.get(r);
				RegionSet converted = r.convert(map);
				Region conR = converted.regions().first();
				convertedLexMap.add(conR, ln);
				debug.println(
				view.getText(r) + " : " + doc.getText(conR) + " : Lexically attracted to: " + ln.getLexLink());
				cache.put("lexically attracted to: " + ln.getLexLink(), converted);
			}
			lexicalMap = convertedLexMap;
			cache.setParsedBy(name); //"lapis.parsers.ListParser");
		}
		finally {
			//			Main.getProgress().fireDoneProgress();
			debug.println("ExternalParser ran in " + (System.currentTimeMillis() - start) + " ms");
		}
	} //end parse(Document)
	
	/*
	 * takes document and grabs all sentences, removes all punctuation
	 */
	public lapis.DocumentView cook(lapis.Document cookedOnce) {
		DocumentView htmlView = cookedOnce.getCookedView();

		CoordinateMap rawToHtml = htmlView.getVersion().diff(cookedOnce.getVersion());

		//go into pattern library
		//get ".English.Sentence" regions
		//for each region, r
		// remove all '\n and puncutation' from r
		// separate tokens "';.?...\n" from sentence
		// append '\n'
		PatternLibrary lib = PatternLibrary.getGlobalLibrary();
		Pattern sentence = lib.get(".English.Sentence");
		RegionSet regionSet = sentence.match(cookedOnce);
		regionSet = regionSet.getSingleDocRegionSet(cookedOnce).convert(rawToHtml);
		RegionEnumeration regions = regionSet.regions();
		PiecewiseLinearMonotonicMap map = new PiecewiseLinearMonotonicMap();
		map.add(0, 0);
		int loc = 0; //how far we've moved along original string
		String oldContent = htmlView.getText();
		debug.print("DOC TEXT: " + htmlView.getText());
		StringBuffer newContent = new StringBuffer();
		String text = null;
		String chars = ",:;'\"!?@#$%^&*()_+=-{}|][\\<>/.";
		String delims = chars + " \n\r";
		boolean previousSpace = false;
		for (Region r = regions.firstFast(); r != null; r = regions.nextFast()) {
			loc = r.getStart();
			map.add(loc, newContent.length());
			text = r.extract(oldContent);
			debug.println("TEXT: " + text);
			//int last = text.lastIndexOf('.');
			int last = text.length();
			String end = "";
			if (last != -1) {
				end = text.substring(last);
				text = text.substring(0, last);
			}
			StringTokenizer st = new StringTokenizer(text, delims, true);
			String tempToken = null;
			while (st.hasMoreTokens()) {
				tempToken = st.nextToken();
				if (tempToken.equals(" ")) {
					if (!previousSpace) {
						newContent.append(tempToken);
						previousSpace = true;
					}
					else {
					}
					loc += 1;
					map.add(loc, newContent.length());
				}
				//			else if (chars.indexOf(tempToken) != -1) {
				//				//INVARIANT: must be of length "1"
				//				if (previousSpace) {
				//				debug.println("previous space, therefore no need to append one");
				//				newContent.append(tempToken);
				//				loc += 1;
				//				map.add(loc, newContent.length());
				//				previousSpace = false;
				//				}
				//				else {
				//				debug.println("no space, therefore need to append one.");
				//				newContent.append(" ").append(tempToken);
				//				loc += 1;
				//				map.add(loc, newContent.length());
				//				}
				//			}
				else
					if ("\n".equals(tempToken) || "\r".equals(tempToken) || chars.indexOf(tempToken) != -1) {
						if (previousSpace) {
							debug.println("space before newline, therefore no need to add another");
							loc += 1;
							map.add(loc, newContent.length());
						}
						else {
							debug.println("no space, therefore need to add before removing newline");
							newContent.append(" ");
							loc += 1;
							map.add(loc, newContent.length());
							previousSpace = true;
						}
					}
					else {
						newContent.append(tempToken);
						loc += tempToken.length();
						map.add(loc, newContent.length());
						previousSpace = false;
					}
			} //end while
			if (newContent.charAt(newContent.length() - 1) == ' ') {
				newContent.append(end + "\n");
			}
			else {
				newContent.append(" " + end + "\n");
			}
			map.add(loc, newContent.length());
		} //end for
		debug.println("NEW CONTENT: " + newContent.toString());
		//	debug.println("DUMP:");
		//	map.dump(doc.getText(), newContent.toString());
		return new SentenceView(cookedOnce, cookedOnce.getVersion(), map, newContent.toString()); 
	}
	/**
	 * Tests.
	 */
	public static void run(Document document) {
		LexicalAttractionParser parser = new LexicalAttractionParser();
		parser.parse(document);
	}
	/*
	 * Parses a sentence string, if you want to increment the values in the frequency table, set the boolean to true.
	 */
	private void parseSentenceString(String sentence, boolean increment) throws IOException {
		LinkedList listWords = new LinkedList();
		StringReader sr = new StringReader(sentence);
		StreamTokenizer st = new StreamTokenizer(sr);
		st.ordinaryChar(39);
		//StringTokenizer st = new StringTokenizer("this, is. a' test");
		int type = st.ttype;
		while (type != StreamTokenizer.TT_EOF) {
			if (type == StreamTokenizer.TT_WORD) {
				//debug.println(st.sval.toLowerCase());
				listWords.add(st.sval.toLowerCase());
			}
			else {
				//debug.println("" + type);
			}
			st.nextToken();
			type = st.ttype;
			//debug.println("***" + type);
		}
		if (!(listWords.size() == 0)) {
			String[] array = new String[listWords.size()];
			for (int i = 0; i < listWords.size(); ++i) {
				array[i] = (String) (listWords.get(i));
			}
			Parse parse;
			if (increment) {
				parse = parse(array, true);
			}
			else {
				parse = parse(array, false);
			}
			debug.println(parse.toString() + "\n" + sentence);
		}
	}
	//test method
	private static void parseSentenceString(LexicalAttractionParser parser, String sentence, boolean increment)
		throws IOException {
		LinkedList listWords = new LinkedList();
		StringReader sr = new StringReader(sentence);
		StreamTokenizer st = new StreamTokenizer(sr);
		st.ordinaryChar(39);
		//StringTokenizer st = new StringTokenizer("this, is. a' test");
		int type = st.ttype;
		while (type != StreamTokenizer.TT_EOF) {
			if (type == StreamTokenizer.TT_WORD) {
				//debug.println(st.sval.toLowerCase());
				listWords.add(st.sval.toLowerCase());
			}
			else {
				//debug.println("" + type);
			}
			st.nextToken();
			type = st.ttype;
			//debug.println("***" + type);
		}
		if (!(listWords.size() == 0)) {
			String[] array = new String[listWords.size()];
			for (int i = 0; i < listWords.size(); ++i) {
				array[i] = (String) (listWords.get(i));
			}
			Parse parse;
			if (increment) {
				parse = parser.parse(array, true);
			}
			else {
				parse = parser.parse(array, false);
			}
			debug.println(parse.toString() + "\n" + sentence);
		}
	}
	/**
	 * A utility class for storing an array of words and an array of integer
	 * showing parse structure.  Keeps track of words parsed, tree structure,
	 * and probability of parse.
	 */
	class Parse {
		String[] words;
		int[] startLocations;
		int[] endLocations;
		int[] links;
		double probability = 0.0;
		/**
		 * Constructs.
		 */
		public Parse(String[] w) {
			words = w;
			links = new int[words.length];
		}
		/**
		 * Roots one word
		 */
		public void rootOneWord() {
			if (links.length != 1) {
				System.err.println("Tried to link one-word parse, but not one word!");
			}
			links[0] = -1;
		}
		/**
		 * Links two word parse to left.
		 */
		public void rootPairToLeft() {
			if (links.length != 2) {
				System.err.println("Tried to link two-word parse, but not two words!");
			}
			links[0] = -1;
			links[1] = 0;
		}
		/**
		 * Links two word parse to right.
		 */
		public void rootPairToRight() {
			if (links.length != 2) {
				System.err.println("Tried to link two-word parse, but not two words!");
			}
			links[0] = 1;
			links[1] = -1;
		}
		/**
		 * Set probability.
		 */
		public void setProbability(double p) {
			probability = p;
		}
		/**
		 * Get probability.
		 */
		public double getProbability() {
			return probability;
		}
		/**
		 * Set array recording tree structure.
		 */
		public void setLinks(int[] l) {
			links = l;
		}
		/**
		 * Get array recording tree structure.
		 */
		public int[] getLinks() {
			return links;
		}
		/**
		 * Get words parsed.
		 */
		public String[] getWords() {
			return words;
		}
		public void setLocations(int startLocation) {
			int offset = 0;
			int[] locs = new int[words.length];
			int[] endLocs = new int[words.length];
			for (int i = 0; i < words.length; ++i) {
				locs[i] = startLocation + offset;
				endLocs[i] = startLocation + offset + words[i].length();
				offset = offset + words[i].length() + 1;
			}
			this.startLocations = locs;
			this.endLocations = endLocs;
		}
		public int[] getStartLocations() {
			return this.startLocations;
		}
		public int[] getEndLocations() {
			return this.endLocations;
		}
//	private void putHelper(MutableRegionSet set, String regionsets) {
//			StringTokenizer st = new StringTokenizer(regionsets, "[], \n\t");
//			while (st.hasMoreTokens()) {
//				set.insert(Integer.parseInt(st.nextToken()), Integer.parseInt(st.nextToken()));
//			}
//		}
		/**
		 * Prints an array of words, with pointers to parents.  A utility method.
		 */
		public String toString() {
			String result = "";
			for (int i = 0; i < words.length; ++i) {
				int index = links[i];
				String reference = "root";
				if (index >= 0) {
					reference = words[index];
				}
				// result += (words[i] + "[" + reference + "] ");
				result += (words[i] + "[" + index + " " + reference + "] ");
			}
			result += " (p = " + getProbability() + ")";
			return result;
		}
	}
	/* (non-Javadoc)
	 * @see lapis.Parser#bind(lapis.PatternLibrary)
	 */
	public void bind(PatternLibrary lib) {
		// TODO Auto-generated method stub
	}

    public void bind (PatternLibrary lib, Document doc) {
        // no doc-specific patterns for this parser
    }

	public void setLexicalMap(LexMap lexicalMap) {
		this.lexicalMap = lexicalMap;
	}
/**
 * 
 * @return Lexical Mapping of document to features
 */
	public LexMap getLexicalMap() {
		return lexicalMap;
	}
	public class SentenceView implements DocumentView {
		Document doc;
		DocumentVersion version = DocumentVersion.getOriginal(this);
		String content = null;
		
		SentenceView (Document doc, DocumentVersion lastVersion,
					  CoordinateMap map, String content) {
			this.doc = doc;
			this.version = new DocumentVersion(map, lastVersion, this);
			this.content = content;
		}
		
		public String getText() {
			return content;
		}
		
		public String getText (Region r) {
			if (content == null)
				return null;
				
			int l = content.length ();
			int s = Math.min (l, Math.max(0, r.getStart ()));
			int e = Math.min (l, Math.max(0, r.getEnd ()));
			
			return (s < e) ? content.substring(s,e) : "";
		}
		
		public Region toRegion () {
			return new Region (0,
						content != null ? content.length () : 0,
						this.version);
		}
		
		public Reader getReader () {
			return new StringReader (getText ());
		}
		
		public DocumentVersion getVersion () {
			return this.version;
		}
		
		public Document getDocument () {
			return this.doc;
		}
	}
}
