/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing.cluster;

import lapis.*;
import lapis.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import javax.swing.undo.*;
import lapis.swing.cluster.MatchNodeCreator.Node;
import javax.swing.plaf.basic.BasicTreeUI;

public class MatchTree extends JTree implements StateEditable {
	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;
	Browser browser;
	ClusterFindReplacePane clusterPane;
	Node root;
	UndoManager undoManager;
	boolean cluster;
	int height = 0;
	boolean treeTable = false;
	MatchTreeTable matchTreeTable = null;
	Rectangle clusterRectangle = null;
	public MatchTree(ClusterFindReplacePane clusterPane, Browser browser, NamedRegionSet sel, boolean cluster) {

		TreeTableModel tm;
		if (cluster)
			tm = new ClusterNodeTableModel(browser, sel);
		else
			tm = new SingleNodeTableModel(browser, sel);

		setModel(tm);
		root = (Node)tm.getRoot();
		this.cluster = cluster;
		this.browser = browser;
		this.clusterPane = clusterPane;
		setProperties();
		setupListeners();
		undoManager = new UndoManager();
		//addSelectionRow(0);
		expandAll();
	}

	public MatchTree(ClusterFindReplacePane clusterPane, Browser browser, RegionSet matches, boolean cluster) {
		//this(clusterPane, browser, new NamedRegionSet(browser.getDocument(), matches), cluster);
		this(clusterPane, browser, new NamedRegionSet(matches), cluster);
	}

	public MatchTree(ClusterFindReplacePane clusterPane, Browser browser, NamedRegionSet sel, boolean cluster, TreeModel tm) {

		super(tm);
		setModel(tm);
		root = (Node)tm.getRoot();
		this.cluster = cluster;
		this.browser = browser;
		this.clusterPane = clusterPane;
		setProperties();
		setupListeners();
		undoManager = new UndoManager();
		//addSelectionRow(0);
		expandAll();
	}
public MatchTree() {
}
	protected void setProperties() {
		setUI(new MatchTreeUI());
		setRootVisible(false);
		setShowsRootHandles(true);
		setBackground(new Color(220, 220, 220));
		setCellRenderer(new Renderer(browser));
		//ToolTipManager.sharedInstance().registerComponent(this);
//		 ((BasicTreeUI)getUI()).setExpandedIcon(SwingConfig.makeIcon("TreeMinus.jpg"));
//		((BasicTreeUI)getUI()).setCollapsedIcon(SwingConfig.makeIcon("TreePlus.jpg"));

	}

	protected void expandAll() {
		for (int i = 0; i < getRowCount(); i++) {
			expandRow(i);
		}

	}

	protected void setupListeners() {

		treeModel.addTreeModelListener(new TreeModelListener() {
			public void treeNodesChanged(TreeModelEvent e) {
				debug.println("nodes changed");

			}

			public void treeNodesInserted(TreeModelEvent e) {
				debug.println("nodes inserted");
			}

			public void treeNodesRemoved(TreeModelEvent e) {
				debug.println("nodes removed");
			}

			public void treeStructureChanged(TreeModelEvent e) {
				debug.println("structure changed");
			}
		});

		addFocusListener(new FocusListener() {
			public void focusGained(FocusEvent e) {
				debug.println("Tree gained focus");
				// rehighlight selected nodes
				if (!treeTable)
					highlightSelectedNodes();
			}
			public void focusLost(FocusEvent e) {
				debug.println("Tree focus lost");
			}
		});

		browser.addSelectionListener(new SelectionListener() {
			public void changedSelection(Browser browser, NamedRegionSet sel) {
				if (MatchTree.this.hasFocus()
					|| clusterPane.componentHasFocus()
					|| (treeTable && matchTreeTable != null && matchTreeTable.hasFocus())) {}
				else {
					Node child = findNodeMatch(sel);
					if (child != null) {
						Node[] path = child.getAncestors();
						TreePath t = new TreePath(path);
						MatchTree.this.setSelectionPath(t);
						scrollPathToVisible(t);
						if (treeTable && matchTreeTable != null)
							matchTreeTable.scrollRectToVisible(matchTreeTable.getCellRect(getRowForPath(t), 0, false));
						child.setSelected(true);
					}
				} //else
			}
		});

		selectionModel.addTreeSelectionListener(new TreeSelectionListener() {
			protected boolean updating = false;
			public void valueChanged(TreeSelectionEvent e) {

				debug.println("****************************NamedRegionSet Changed*******************************");

				debug.printStackTrace();
				//debug.println("source: " + e.getSource());
				TreePath[] changedPaths = e.getPaths();
				TreePath[] paths = getSelectionPaths();
				Vector nodes = new Vector();

				if (paths != null) {

					for (int i = 0; i < paths.length; i++) {
						if (paths[i].getLastPathComponent() instanceof Node) {

							Node node = (Node)paths[i].getLastPathComponent();
							if (node.isTopLevel) {
								// select all paths below it
								debug.println("added or still selected " + node.toString());

								// if Node is expanded, action handled by mouse listener
								if (isExpanded(paths[i])) {
									node.setSelected(true);
									debug.println("expanded ");
								} // end expanded

								// else select all the children
								else {

									if (node.getSelectedChildrenCount() == 0) {
										debug.println("selecting all children");
										for (Enumeration f = node.children.elements(); f.hasMoreElements();) {
											Node child = (Node)f.nextElement();
											nodes.add(child);
											debug.println(child.toString());
											child.setSelected(true);
										}
									}
									else
										if (node.getSelectedChildrenCount() <= node.getChildCount()) {
											debug.println("selecting selected children");
											for (Enumeration f = node.selectedChildren.elements(); f.hasMoreElements();) {
												Node child = (Node)f.nextElement();
												nodes.add(child);
												debug.println(child.toString());
											}
										}

								} //end else !expanded
								debug.println(node.toString() + " sel " + node.selectedChildren.size());
							} // if toplevel
							else {
								debug.println("selecting: " + node.toString());
								node.setSelected(true);
								nodes.add(node);
							} // if not toplevel
						}
					}
				}
				if (changedPaths != null) {

					for (int i = 0; i < changedPaths.length; i++) {
						Node node = (Node)changedPaths[i].getLastPathComponent();
						if (!(e.isAddedPath(changedPaths[i]))) {
							debug.println("removed and no longer selected " + changedPaths[i].toString());

							if (node.isTopLevel) {
								node.setSelected(false);
								if (!isExpanded(changedPaths[i])) {
									for (Enumeration f = node.children.elements(); f.hasMoreElements();) {
										Node child = (Node)f.nextElement();
										child.setSelected(false);
									}
								}
								//					else{
								//						for (Enumeration f = node.children.elements();  f.hasMoreElements(); ){
								//							debug.println("test");
								//							Node child = (Node)f.nextElement();
								//							child.setSelected(false);
								//						}
								//					}
							} // if TopLevel
							else {

								if (node.getParent() != null) {
									if ((isExpanded(changedPaths[i].getParentPath()))) {
										node.setSelected(false);
										// if parent is expanded then check if parent needs to be deselected 
										try {
											if (e.isAddedPath(changedPaths[i].getParentPath())) {
												node.setSelected(true);
												debug.println("parent added, don't deselect");
											}
											else {
												debug.println("not all sibs selected, deselect parent");
												removeSelectionPath(changedPaths[i].getParentPath());
											}
										}
										catch (IllegalArgumentException iae) {
											debug.println("not all sibs selected, deselect parent");
											removeSelectionPath(changedPaths[i].getParentPath());
										}
										//}
										//}
									}
								}
							}
						} // if removed
						else {
							//node.setSelected(true);
							if (!node.isTopLevel) {
								if (isExpanded(changedPaths[i].getParentPath()))
									if (node.getParent() != null)
										if (((Node)node.getParent()).getSelectedChildrenCount() == ((Node)node.getParent()).getChildCount())
											if ((Node)node.getParent() != root) {
												addSelectionPath(changedPaths[i].getParentPath());

											}
							}
						}
					} //for
				}
				debug.println("node size: " + nodes.size());
				highlight(nodes);
				//clusterPane.setSelectedText(browser.getSelection().getSize());
				//highlightSelectedNodes();
			}
		});

		addTreeWillExpandListener(new TreeWillExpandListener() {

			public void treeWillCollapse(TreeExpansionEvent e) throws ExpandVetoException {
				//debug.println("tree collapsing " + e.getSource().toString());
				debug.println(e.getPath().toString() + " tree collapsing ");
				TreePath path = e.getPath();
				Node node = (Node)path.getLastPathComponent();
				Vector selectedChildren = new Vector();
				if (node.getChildCount() == node.getSelectedChildrenCount()) {
					// if all are selected, maintain this state
					for (Enumeration f = node.children.elements(); f.hasMoreElements();) {
						TreePath t = path.pathByAddingChild(f.nextElement());
						if (isPathSelected(t)) {
							Node selChild = (Node)t.getLastPathComponent();
							selectedChildren.add(t);
							selChild.setSelected(true);
							debug.println("adding selected child: " + t.toString());
						}
					}
				}
				else {
					removeSelectionPath(path);
					//if only partially selected, remove all selected children
					TreePath[] pathsToBeRemoved = new TreePath[node.getChildCount()];
					int i = 0;
					for (Enumeration f = node.children.elements(); f.hasMoreElements(); i++) {
						TreePath t = path.pathByAddingChild(f.nextElement());
						pathsToBeRemoved[i] = t;
						if (isPathSelected(t)) {
							Node selChild = (Node)t.getLastPathComponent();
							selectedChildren.add(t);
							selChild.setSelected(false);
							debug.println("removing selected child: " + t.toString());
						}
					}
					removeSelectionPaths(pathsToBeRemoved);

				}
				if (clusterRectangle != null) {
					clusterRectangle = null;
					repaint();
				}
			}

			public void treeWillExpand(TreeExpansionEvent e) {
				TreePath path = e.getPath();
				//debug.println(path.toString () + " expanding " + e.getSource().toString());
				debug.println(path.toString() + " expanding ");
				Node node = (Node)path.getLastPathComponent();
				Vector selectedChildren = node.getSelectedChildren();
				TreePath[] pathsToBeAdded = new TreePath[selectedChildren.size()];
				int i = 0;
				for (Enumeration f = selectedChildren.elements(); f.hasMoreElements(); i++) {
					TreePath t = path.pathByAddingChild(f.nextElement());
					pathsToBeAdded[i] = t;
					debug.println("setting previously selected child: " + t.toString());
				}
				if (selectedChildren.size() < node.children.size())
					removeSelectionPath(path);
				addSelectionPaths(pathsToBeAdded);
				debug.println("added selected paths");
				if (clusterRectangle != null) {
					clusterRectangle = null;
					repaint();
				}
			}
		});

		addMouseMotionListener(new MouseMotionAdapter() {

			public void mouseMoved(MouseEvent me) {
				//		    debug.println("mouse moved");
				TreePath treePath = getPathForLocation(me.getX(), me.getY());
				if (treePath != null) {
					//MatchTree.this.setCursor(new Cursor(Cursor.HAND_CURSOR));
					Node node = (Node)treePath.getLastPathComponent();
					if (node.isTopLevel)
						handleTopLevelSelection(treePath, node, me.getX(), me.getY());
					else {
						handleLowerLevelSelection(treePath);
					}

				}
				else {
					//MatchTree.this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
					treePath = getClosestPathForLocation(me.getX(), me.getY());
					if (treePath != null) {

						Node node = (Node)treePath.getLastPathComponent();
						if (!node.isTopLevel)
							treePath = treePath.getParentPath();
						handleTopLevelSelection(treePath, (Node)treePath.getLastPathComponent(), me.getX(), me.getY());

					}
				}
				//handleTopLevelSelection(treePath, (Node)treePath.getLastPathComponent());

			}

			private void handleLowerLevelSelection(TreePath treePath) {
				Rectangle treeBounds = getPathBounds(treePath);
				if (treeBounds != clusterRectangle) {
					clusterRectangle = treeBounds;
					repaint();
				}
			}

			private void handleTopLevelSelection(TreePath closestPath, Node node, int x, int y) {
				if (isExpanded(closestPath)) {
					int row = getRowForPath(closestPath);
					Rectangle top = getPathBounds(closestPath);
					int lastRow = node.getChildCount() + row;
					Rectangle bottom = getRowBounds(lastRow);
					Rectangle cluster = top.union(bottom);
					if (cluster.contains(new Point(x, y))) {
						if (clusterRectangle != cluster) {
							clusterRectangle = cluster;
							repaint();
						}
					}
					else
						if (clusterRectangle != null) {
							clusterRectangle = null;
							repaint();
						}

				}
				else {

					Rectangle top = getPathBounds(closestPath);
					if (top.contains(new Point(x, y))) {

						if (clusterRectangle != top) {
							clusterRectangle = top;
							repaint();
						}
					}
					else
						if (clusterRectangle != null) {
							clusterRectangle = null;
							repaint();
						}
				}
			}

		});

		addMouseListener(new MouseAdapter() {

			public void mouseExited(MouseEvent me) {
				clusterRectangle = null;
				repaint();
			}

			public void mousePressed(MouseEvent e) {
				debug.println("mouse click on tree at: x: " + e.getX() + " y: " + e.getY());
				debug.println("clicked on row: " + MatchTree.this.getRowForLocation(e.getX(), e.getY()));

				debug.println("pressed on " + MatchTree.this.getPathForLocation(e.getX(), e.getY()));
				TreePath path = MatchTree.this.getPathForLocation(e.getX(), e.getY());
				if (path == null) {
					debug.println("pressed closest to " + MatchTree.this.getClosestPathForLocation(e.getX(), e.getY()));
					TreePath closestPath = MatchTree.this.getClosestPathForLocation(e.getX(), e.getY());
					Node node = (Node)closestPath.getLastPathComponent();
					if (node.isTopLevel) {
						if (!((MatchTreeUI)MatchTree.this.getUI()).isLocationInExpandControl(closestPath, e.getX(), e.getY())
							&& (clusterRectangle != null)) {
							if (MatchTree.this.isPathSelected(closestPath)) {
								MatchTree.this.removeSelectionPath(closestPath);
							}
							else {
								MatchTree.this.addSelectionPath(closestPath);
							}
							handleTopLevelSelection(e, closestPath, node);
						}

					}
					else {
						if (clusterRectangle != null) {

							closestPath = new TreePath(((Node)node.getParent()).getAncestors());
							node = (Node)node.getParent();
							if (MatchTree.this.isPathSelected(closestPath)) {
								MatchTree.this.removeSelectionPath(closestPath);
							}
							else {
								MatchTree.this.addSelectionPath(closestPath);
							}

							handleTopLevelSelection(e, closestPath, node);
						}
					}
					return;
				}

				else {
					debug.println("sel?: " + ((Node)path.getLastPathComponent()).selected);
					//find out if it is over the ? and if it is display the tool tip 
					Node node = (Node)path.getLastPathComponent();
					if (!node.isTopLevel) {
						return;
					}
					//is it on the ?
					Rectangle nodeBounds = MatchTree.this.getPathBounds(path);
					debug.println(nodeBounds.toString());
					int xBound = nodeBounds.x + nodeBounds.width - 30;
					int yBound = nodeBounds.y + 20;
					//debug.println("xbound " + xBound + " ybound " + yBound + " x " + e.getX() + " y " + e.getY());
					if ((e.getX() > xBound) && (e.getY() < yBound)) {
						//debug.println("clicked on ?");
					}
				}

				//debug.println("clicked on " +  MatchTree.this.getPathForLocation (e.getX(), e.getY())); 
				path = MatchTree.this.getPathForLocation(e.getX(), e.getY());
				if (path == null)
					return;
				else {
					Node node = (Node)path.getLastPathComponent();
					if (node.isTopLevel) {
						handleTopLevelSelection(e, path, node);
					} // if top level
					else {}
				}
			}

			public void mouseClicked(MouseEvent e) {}

			public void mouseReleased(MouseEvent e) {
				debug.println("release on " + MatchTree.this.getPathForLocation(e.getX(), e.getY()));

			}

			private void handleTopLevelSelection(MouseEvent e, TreePath path, Node node) {

				//check if selected 
				if (MatchTree.this.isPathSelected(path)) {
					node.setSelected(true);
					debug.println(node.toString() + " selected, selecting all underneath");
					//select or deselect all underneath
					if (MatchTree.this.isExpanded(path)) {
						//debug.println("expanded ");

						// select all 
						TreePath[] pathsToBeAdded = new TreePath[node.children.size()];
						int j = 0;
						debug.println("selecting all ");
						for (Enumeration f = node.children.elements(); f.hasMoreElements(); j++) {
							Node child = (Node)f.nextElement();
							child.setSelected(true);
							TreePath t = path.pathByAddingChild(child);
							pathsToBeAdded[j] = t;
						}
						addSelectionPaths(pathsToBeAdded);
					}

				} // if selected
				else {
					if ((!MatchTree.this.isPathSelected(path)) || e.isControlDown()) {
						node.setSelected(false);
						//debug.println(node.toString() + " not selected, deselecting all underneath");
						if (isExpanded(path)) {
							//debug.println("expanded removal");
							TreePath[] pathsToBeRemoved = new TreePath[node.children.size()];
							int j = 0;
							//debug.println("deselecting all " );
							for (Enumeration f = node.children.elements(); f.hasMoreElements(); j++) {
								Node child = (Node)f.nextElement();
								child.setSelected(false);
								TreePath t = path.pathByAddingChild(child);
								pathsToBeRemoved[j] = t;
							}
							removeSelectionPaths(pathsToBeRemoved);
							//							node.setSelectedChildren(new Vector());
							//debug.println("deselected all, new selected children size : " + node.selectedChildren.size());
							removeSelectionPaths(pathsToBeRemoved);
						}
					}
				} // if not selected

			}

		});

	}
	protected void paintComponent(Graphics g) {

		super.paintComponent(g);
		if (clusterRectangle != null) {
			g.setColor(new Color(150, 150, 150, 50));
			g.fill3DRect(clusterRectangle.x, clusterRectangle.y, clusterRectangle.width, clusterRectangle.height, true);
			debug.println("want to paint: " + clusterRectangle.toString());
		}
	}

	public void setSelectionModel(TreeSelectionModel tsm) {
		super.setSelectionModel(tsm);
		setupListeners();
	}
	//undoable stuff
	public void restoreState(Hashtable state) {
		//debug.println("savedState: " + state.toString());
		TreeTableModel tm;
		Node root2 = (Node)state.get("root");
		if (root2 == null)
			return;
		if (cluster)
			tm = new ClusterNodeTableModel(root2);
		else
			tm = new SingleNodeTableModel(root2);

		setModel(tm);
		setupListeners();
		//debug.println("restoring: new model: " + tm.toString());
		this.root = (Node)tm.getRoot();
		if (treeTable) {
			matchTreeTable.setModel(tm);
		}

		this.setVisible(false);
		this.setVisible(true);
		this.repaint();
	}
	public void storeState(Hashtable state) {
		Node root2 = root.copy();
		state.put("root", root2);

	}
	public void undo() {
		try {
			undoManager.undo();
			expandAll();
		}
		catch (CannotUndoException cue) {}
	}
	public void highlightSelectedNodes() {
		debug.println("highlighting");

		browser.clearSelection();

		Vector nodes = root.getSelectedDecendents();
		highlight(nodes);
	}

	public void highlightAll() {
		setSelectionInterval(0, getRowCount());
	}

	public void setTreeTable(boolean table, MatchTreeTable mtt) {
		treeTable = table;
		if (table)
			matchTreeTable = mtt;
		else
			matchTreeTable = null;
	}
	public void highlight(Node node) {
		debug.println("highlighting " + node.toString() + " " + node.smatch.regions().first());
		browser.clearSelection();
		browser.setSelection(new NamedRegionSet(node.getMatch()));
		browser.getActivePane().scrollRegionsToVisible(node.getMatch());

	}

	public void highlight(Vector nodes) {
		debug.println("node sizen before: " + nodes.size());

		RegionSet rs = RegionSet.EMPTY;
		for (Enumeration e = nodes.elements(); e.hasMoreElements();) {
			Object o = e.nextElement();
			if (o instanceof Node) {
				rs = rs.or(((Node)o).getMatch());
				debug.println("selecting in browser: " + ((Node)o).toString());
			}
		}

		debug.println("node size after: " + nodes.size());
		browser.setSelection(new NamedRegionSet(rs));
		browser.getActivePane().scrollRegionsToVisible(rs);

	}
	public Node findNodeMatch(NamedRegionSet sel) {

		//debug.println("do something");
		if (sel.getSize() == 1) {
			// debug.println("one selection");
			// return corresponding node
			// if it exists, else do nothing
			//RegionSet rs = sel.getRegions();
			for (Enumeration e = root.children.elements(); e.hasMoreElements();) {
				Node n = (Node)e.nextElement();
				if (n.isTopLevel) {
					for (Enumeration f = n.children.elements(); f.hasMoreElements();) {
						Node child = (Node)f.nextElement();
						if (child.getMatch().equals((RegionSet)sel)) {
							return child;
						} // if
					} //for
				} //if
				else {
					//check singletons
					if (n.getMatch().equals((RegionSet)sel)) {
						return n;
					} //if
				} //else
			} //for
		} //if
		else {
			return null;
		}
		return null;

	}

	public void clearSelection() {
		super.clearSelection();
	}
	public int getRowHeight() {
		return height;
	}

	private void remove(TreePath[] paths) {

		// Create the edit during the "before" state of the object
		StateEdit newEdit = new StateEdit(this);

		// Modify the object

		if (paths != null) {
			for (int i = 0; i < paths.length; i++) {
				debug.println("plan remove path : " + paths[i].toString());
			}
			for (int i = 0; i < paths.length; i++) {
				debug.println("removing path : " + paths[i].toString());

				Object[] path = paths[i].getPath();
				Node parent = (Node)path[path.length - 2];
				Node child = (Node)paths[i].getLastPathComponent();
				debug.println("parent: " + parent.toString() + " child " + child.toString());

				if (child.isTopLevel) {
					debug.println("child is top level");
					if (!isExpanded(paths[i])) {
						debug.println("child is not expanded");
						Vector selectedChildren = child.getSelectedChildren();
						if (selectedChildren.size() > 0) {
							if (selectedChildren.size() == child.getChildCount()) {
								debug.println(
									"child " + child.toString() + " parent " + child.getParent().toString() + " index " + parent.getIndex(child));
								((TreeTableModel)treeModel).removeNodeFromParent(child);

							}
							else {
								debug.println("removing only selected children: " + selectedChildren.size());
								Enumeration f = selectedChildren.elements();
								int c = 0;
								for (; f.hasMoreElements(); c++) {
									Object selChild = f.nextElement();
									debug.println("removing match#: " + c);
									((TreeTableModel)treeModel).removeNodeFromParent((Node)selChild);
								}
								debug.println("finished removal");
								child.resetSelectedChildren();
								debug.println("reset selected children");
								removeSelectionPath(paths[i]);
								debug.println("removing selection paths");
							}
						}
					}
					else {
						Vector selectedChildren = child.getSelectedChildren();
						if (selectedChildren.size() > 0) {
							if (selectedChildren.size() == child.getChildCount()) {
								debug.println(
									"child " + child.toString() + " parent " + child.getParent().toString() + " index " + parent.getIndex(child));
								((TreeTableModel)treeModel).removeNodeFromParent(child);
							}
						}
					}
				}
				else {
					if (parent != null) {
						debug.println("removing singleton");
						((TreeTableModel)treeModel).removeNodeFromParent(child);
						parent.resetText();
					}
				}

				debug.println("parent info after: " + parent.toString() + " " + parent.children.size() + " " + +parent.getDescendantsCount());
				debug.println("child info after: " + child.toString() + " " + child.children.size());
				debug.println("removing upper node if 0");
				debug.println("parent.isLeaf: " + parent.isLeaf() + " parent.allowsChildren: " + parent.getAllowsChildren());
				// 		    if ( parent.isLeaf() && parent.getAllowsChildren())
				// 			((DefaultTreeModel)treeModel).removeNodeFromParent(parent);
				debug.println("child.isLeaf: " + child.isLeaf() + " child.allowsChildren: " + child.getAllowsChildren());
				if (child.isLeaf() && child.getAllowsChildren())
					 ((TreeTableModel)treeModel).removeNodeFromParent(child);
			}

		}
		// "end" the edit when you are done modifying the object
		newEdit.end();
		// debug.println("before " + newEdit.preState + " after: " + newEdit.postState);
		debug.println("can undo? " + newEdit.canUndo());
		undoManager.addEdit(newEdit);

	}

	public void selectAll() {
		setSelectionInterval(0, getRowCount());
	}

	public void deleteSelection() {
		//	expandAll();
		remove(getSelectionPaths());

	}

	class Renderer extends DefaultTreeCellRenderer {
		Browser browser;

		public Renderer(Browser browser) {
			super();
			this.browser = browser;

		}

		public Component getTreeCellRendererComponent(
			JTree tree,
			Object value,
			boolean sel,
			boolean expanded,
			boolean leaf,
			int row,
			boolean hasFocus) {
			Node n = (Node)value;
			//	    debug.println(n.toString());
			ClusterRenderer renderer = new ClusterRenderer();
			if (n.getAllowsChildren()) {
				JComponent m = renderer.getTopSingleClip(browser, n, expanded, sel, hasFocus);
				//tree.setRowHeight(m.getHeight());
				height = m.getPreferredSize().height;
				return m;
			}

			RegionSet rs = n.getMatch();
			Region r = rs.regions().first();
			JComponent c = renderer.getSingleClip(browser, n, r, sel, hasFocus);
			//c.setToolTipText(n.getExplanationText());
			height = c.getPreferredSize().height;
			return c;

		}

	}

}
