/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing.cluster;

import lapis.*;
import lapis.swing.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.awt.Dimension;
import javax.swing.tree.*;
import javax.swing.table.*;
import lapis.swing.cluster.MatchNodeCreator.Node;


public class MatchTreeTable extends JTable {
	/** A subclass of JTree. */
	protected TreeTableCellRenderer tree;
	private Browser browser;
	public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;
	protected Rectangle clusterRectangle = null;
	public MatchTreeTable(ClusterFindReplacePane clusterPane, Browser browser, NamedRegionSet sel, boolean cluster, TreeTableModel treeTableModel) {
		super();

		this.browser = browser;
		// Create the tree. It will be used as a renderer and editor. 
		tree = new TreeTableCellRenderer(clusterPane, browser, sel, cluster, treeTableModel);

		setModel(treeTableModel);

		setProperties();
		setupListeners();
	}

	public void setModel(TreeTableModel treeTableModel) {
		TreeTableModelAdapter treeTableModelAdapter = new TreeTableModelAdapter(treeTableModel, tree);
		setupModelListener(treeTableModelAdapter);
		// Install a tableModel representing the visible rows in the tree. 
		super.setModel(treeTableModelAdapter);
		getColumnModel().getColumn(0).setPreferredWidth(200);
		resetRowHeights();
	}

	private void setProperties() {
		// Force the JTable and JTree to share their row selection models. 
		//				ListToTreeSelectionModelWrapper selectionWrapper = new ListToTreeSelectionModelWrapper();
		//				tree.setSelectionModel(selectionWrapper);
		//				setSelectionModel(selectionWrapper.getListSelectionModel());

		// Install the tree editor renderer and editor. 
		setDefaultRenderer(TreeTableModel.class, tree);
		setDefaultEditor(TreeTableModel.class, new TreeTableCellEditor());

		setShowGrid(false);

		// No intercell spacing
		setIntercellSpacing(new Dimension(0, 0));

		tree.setCellRenderer(new Renderer(browser));
		setColumnSelectionAllowed(false);
		setRowSelectionAllowed(false);
		getColumnModel().getColumn(0).setPreferredWidth(200);
		getTableHeader().setReorderingAllowed(false);
		tree.setTreeTable(true, this);
		this.setBackground(Color.lightGray);
	}

	private void setupListeners() {

		this.addMouseListener(new MouseInputAdapter() {
			public void mouseExited(MouseEvent me) {
				clusterRectangle = null;
				repaint();
			}
		});

		addMouseMotionListener(new MouseMotionAdapter() {

			public void mouseMoved(MouseEvent me) {
				//		    debug.println("mouse moved");
				TreePath treePath = MatchTreeTable.this.getTree().getPathForLocation(me.getX(), me.getY());
				if (MatchTreeTable.this.getCellRect(MatchTreeTable.this.tree.getRowForPath(treePath), 0, false).width > me.getX()) {
					debug.println(
						"Mouse at: "
							+ me.getX()
							+ " width of column: "
							+ MatchTreeTable.this.getCellRect(MatchTreeTable.this.tree.getRowForPath(treePath), 0, false).width);
					if (treePath != null) {
						MatchTreeTable.this.setCursor(new Cursor(Cursor.HAND_CURSOR));
						Node node = (Node)treePath.getLastPathComponent();
						if (node.isTopLevel)
							handleTopLevelSelection(treePath, node);
						else {
							handleLowerLevelSelection(treePath);
						}

					}
					else {
						MatchTreeTable.this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
						treePath = MatchTreeTable.this.getTree().getClosestPathForLocation(me.getX(), me.getY());

						Node node = (Node)treePath.getLastPathComponent();
						if (!node.isTopLevel)
							treePath = treePath.getParentPath();
						handleTopLevelSelection(treePath, (Node)treePath.getLastPathComponent());

					}
					//handleTopLevelSelection(treePath, (Node)treePath.getLastPathComponent());
				}
				else {
					MatchTreeTable.this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
					if (clusterRectangle != null) {
						clusterRectangle = null;
						repaint();
					}
				}

			}

			private void handleLowerLevelSelection(TreePath treePath) {
				Rectangle treeBounds = MatchTreeTable.this.tree.getPathBounds(treePath);
				int row = MatchTreeTable.this.tree.getRowForPath(treePath);
				Rectangle tableBounds = MatchTreeTable.this.getCellRect(row, 0, false);
				Rectangle bounds =
					new Rectangle(treeBounds.x, treeBounds.y, Math.min(treeBounds.width, tableBounds.width - treeBounds.x), treeBounds.height);
				if (bounds != clusterRectangle) {
					clusterRectangle = bounds;
					repaint();
				}
			}

			private void handleTopLevelSelection(TreePath closestPath, Node node) {

				int row = MatchTreeTable.this.tree.getRowForPath(closestPath);
				Rectangle top = MatchTreeTable.this.getCellRect(row, 0, false);
				if (MatchTreeTable.this.tree.isExpanded(closestPath)) {
					int lastRow = node.getChildCount() + row;
					Rectangle bottom = MatchTreeTable.this.getCellRect(lastRow, 0, false);
					Rectangle cluster = top.union(bottom);
					if (clusterRectangle != cluster) {
						clusterRectangle = cluster;
						repaint();
					}
				}
				else {
					if (clusterRectangle != top) {
						clusterRectangle = top;
						repaint();
					}
				}
			}

		});

		addFocusListener(new FocusListener() {
			public void focusGained(FocusEvent e) {
				debug.println("table gained focus");
				tree.highlightSelectedNodes();
			}
			public void focusLost(FocusEvent e) {
				debug.println("table focus lost");
			}
		});

		//		selectionModel.addListSelectionListener(new ListSelectionListener() {
		//			public void valueChanged(ListSelectionEvent e) {
		//				debug.println("selectionChanged");
		//				clearSelection();
		//			}
		//		});

		getTree().addTreeSelectionListener(new TreeSelectionListener() {
			public void valueChanged(TreeSelectionEvent e) {
				debug.println("selectionChanged");
				//debug.printStackTrace();
				TreePath[] paths = getTree().getSelectionPaths();
				if (paths != null)
					for (int i = 0; i < paths.length; i++) {
						debug.println(paths[i].toString());
					}
				MatchTreeTable.this.repaint();
			}
		});
		addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent e) {
				debug.println("mouse click on table at: x: " + e.getX() + " y: " + e.getY());
				debug.println("clicked on row: " + MatchTreeTable.this.rowAtPoint(new Point(e.getX(), e.getY())));
				repaint();

			}
		});

	}

	private void setupModelListener(TreeTableModelAdapter ttma) {
		ttma.addTableModelListener(new TableModelListener() {
			public void tableChanged(TableModelEvent e) {
				debug.println("table model changed");
				resetRowHeights();
				if (clusterRectangle != null) {
					clusterRectangle = null;
					repaint();
				}
			}
		});
	}

	/**
	 * Overridden to message super and forward the method to the tree.
	 * Since the tree is not actually in the component hieachy it will
	 * never receive this unless we forward it in this manner.
	 */
	public void updateUI() {
		super.updateUI();
		if (tree != null) {
			tree.updateUI();
		}
		// Use the tree's default foreground and background colors in the
		// table. 
		LookAndFeel.installColorsAndFont(this, "Tree.background", "Tree.foreground", "Tree.font");
	}

	/* Workaround for BasicTableUI anomaly. Make sure the UI never tries to 
	 * paint the editor. The UI currently uses different techniques to 
	 * paint the renderers and editors and overriding setBounds() below 
	 * is not the right thing to do for an editor. Returning -1 for the 
	 * editing row in this case, ensures the editor is never painted. 
	 */
	public int getEditingRow() {
		return (getColumnClass(editingColumn) == TreeTableModel.class) ? -1 : editingRow;
	}
	protected void paintComponent(Graphics g) {

		super.paintComponent(g);
		if (clusterRectangle != null) {
			g.setColor(new Color(150, 150, 150, 50));
			g.fill3DRect(clusterRectangle.x, clusterRectangle.y, clusterRectangle.width, clusterRectangle.height, true);
			debug.println("want to paint: " + clusterRectangle.toString());
		}
	}

	/**
	 * Returns the tree that is being shared between the model.
	 */
	public MatchTree getTree() {
		return tree;
	}
	private void resetRowHeights() {
		if (tree != null) {
			for (int i = 0; i < getRowCount(); i++) {
				setRowHeight(i, tree.getRowBounds(i).height);
			}
			for (int i = 0; i < getRowCount(); i++) {
				debug.println("set table row " + i + " height to " + getRowHeight(i));
			}
		}
	}

	public void undo() {
		setVisible(false);
		((TreeTableModelAdapter)getModel()).fireTableDataChanged();
		setVisible(true);
	}
	/**
	 * A TreeCellRenderer that displays a JTree.
	 */
	public class TreeTableCellRenderer extends MatchTree implements TableCellRenderer {
		/** Last table/tree row asked to renderer. */
		protected int visibleRow;
		protected boolean initialized = false;

		public TreeTableCellRenderer(ClusterFindReplacePane clusterPane, Browser browser, NamedRegionSet sel, boolean cluster, TreeModel model) {
			super(clusterPane, browser, sel, cluster, model);
			this.debug.println("created renderer");
			initialized = true;
		}

		/**
		 * updateUI is overridden to set the colors of the Tree's renderer
		 * to match that of the table.
		 */
		public void updateUI() {
			super.updateUI();
			// Make the tree's cell renderer use the table's cell selection
			// colors. 
			TreeCellRenderer tcr = getCellRenderer();
			if (tcr instanceof DefaultTreeCellRenderer) {
				DefaultTreeCellRenderer dtcr = ((DefaultTreeCellRenderer)tcr);
				// For 1.1 uncomment this, 1.2 has a bug that will cause an
				// exception to be thrown if the border selection color is
				// null.
				// dtcr.setBorderSelectionColor(null);
				dtcr.setTextSelectionColor(UIManager.getColor("Table.selectionForeground"));
				dtcr.setBackgroundSelectionColor(UIManager.getColor("Table.selectionBackground"));
			}
		}

		/**
		 * Sets the row height of the tree, and forwards the row height to
		 * the table.
		 */
		//		public void setRowHeight(int rowHeight) {
		//			if (rowHeight > 0) {
		//				super.setRowHeight(rowHeight);
		//				if (MatchTreeTable.this != null && MatchTreeTable.this.getRowHeight() != rowHeight) {
		//					MatchTreeTable.this.setRowHeight(getRowHeight());
		//				}
		//			}
		//		}
		public void setRowHeight(int rowHeight) {
			MatchTreeTable.debug.println("*********************************setting row heights**********");
			if (!initialized) {
				if (rowHeight > 0) {
					super.setRowHeight(rowHeight);
					if (MatchTreeTable.this != null && MatchTreeTable.this.getRowHeight() != rowHeight) {
						MatchTreeTable.this.setRowHeight(getRowHeight());
					}
				}
			}
			else
				resetRowHeights();
		}
		/**
		 * This is overridden to set the height to match that of the JTable.
		 */
		public void setBounds(int x, int y, int w, int h) {
			super.setBounds(x, 0, w, MatchTreeTable.this.getHeight());
		}

		/**
		 * Sublcassed to translate the graphics such that the last visible
		 * row will be drawn at 0,0.
		 */
		public void paint(Graphics g) {
			//MatchTreeTable.this.debug.println("painting tree table cell");

			//			MatchTreeTable.this.debug.println("bounds: " + getBounds());
			g.translate(0, -getRowBounds(visibleRow).y);

			//Rectangle oldBounds = g.getClipBounds();
			//			g.setClip(oldBounds.x, oldBounds.y, oldBounds.width, getRowBounds(visibleRow).height);
			//			g.clearRect(oldBounds.x, oldBounds.y, oldBounds.width, oldBounds.height);
			//g.setClip(0, getRowBounds(visibleRow).y, oldBounds.width, Math.min(getRowBounds(visibleRow).height, oldBounds.height));
			super.paint(g);
			//g.setClip(oldBounds);
			//			g.setClip(oldBounds);
		}

		/**
		 * TreeCellRenderer method. Overridden to update the visible row.
		 */
		public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
			if (isSelected)
				setBackground(table.getSelectionBackground());
			else
				setBackground(table.getBackground());
			//			if (MatchTreeTable.this.getRowHeight(row) != getRowBounds(row).height)
			//				resetRowHeights();
			//MatchTreeTable.this.setRowHeight(row, getRowBounds(row).height);

			//MatchTreeTable.this.debug.println("row " + row + " has height " + MatchTreeTable.this.getRowHeight(row));
			visibleRow = row;

			return this;

		}
	}

	class Renderer extends DefaultTreeCellRenderer {
		Browser browser;

		public Renderer(Browser browser) {
			super();
			this.browser = browser;

		}

		public Component getTreeCellRendererComponent(
			JTree tree,
			Object value,
			boolean sel,
			boolean expanded,
			boolean leaf,
			int row,
			boolean hasFocus) {
			Node n = (Node)value;
			//	    debug.println(n.toString());
			ClusterRenderer renderer = new ClusterRenderer();
			if (n.getAllowsChildren()) {
				JComponent m = renderer.getTopSingleClip(browser, n, expanded, sel, hasFocus);
				return m;
			}

			RegionSet rs = n.getMatch();
			Region r = rs.regions().first();
			JComponent c = renderer.getSingleClip(browser, n, r, sel, hasFocus);
			//c.setToolTipText(n.getExplanationText());
			return c;

		}
	}

	/**
	 * ListToTreeSelectionModelWrapper extends DefaultTreeSelectionModel
	 * to listen for changes in the ListSelectionModel it maintains. Once
	 * a change in the ListSelectionModel happens, the paths are updated
	 * in the DefaultTreeSelectionModel.
	 */
	class ListToTreeSelectionModelWrapper extends DefaultTreeSelectionModel {
		/** Set to true when we are updating the ListSelectionModel. */
		protected boolean updatingListSelectionModel;

		public ListToTreeSelectionModelWrapper() {
			super();
			getListSelectionModel().addListSelectionListener(createListSelectionListener());
		}

		/**
		 * Returns the list selection model. ListToTreeSelectionModelWrapper
		 * listens for changes to this model and updates the selected paths
		 * accordingly.
		 */
		ListSelectionModel getListSelectionModel() {
			return listSelectionModel;
		}

		/**
		 * This is overridden to set <code>updatingListSelectionModel</code>
		 * and message super. This is the only place DefaultTreeSelectionModel
		 * alters the ListSelectionModel.
		 */
		public void resetRowSelection() {
			//			debug.println("reseting Row NamedRegionSet");
			if (!updatingListSelectionModel) {
				updatingListSelectionModel = true;
				try {
					super.resetRowSelection();
				}
				finally {
					updatingListSelectionModel = false;
				}
			}
			// Notice how we don't message super if
			// updatingListSelectionModel is true. If
			// updatingListSelectionModel is true, it implies the
			// ListSelectionModel has already been updated and the
			// paths are the only thing that needs to be updated.
		}

		/**
		 * Creates and returns an instance of ListSelectionHandler.
		 */
		protected ListSelectionListener createListSelectionListener() {
			return new ListSelectionHandler();
		}

		/**
		 * If <code>updatingListSelectionModel</code> is false, this will
		 * reset the selected paths from the selected rows in the list
		 * selection model.
		 */
		protected void updateSelectedPathsFromSelectedRows() {
			if (!updatingListSelectionModel) {
				updatingListSelectionModel = true;
				try {
					// This is way expensive, ListSelectionModel needs an
					// enumerator for iterating.
					int min = listSelectionModel.getMinSelectionIndex();
					int max = listSelectionModel.getMaxSelectionIndex();
					int min2 = getTree().getSelectionModel().getMinSelectionRow();
					int max2 = getTree().getSelectionModel().getMaxSelectionRow();

					debug.println("min: " + min + " max; " + max);
					debug.println("min2: " + min2 + " max2; " + max2);
					debug.println("clearing selection ");
					if (getModel() instanceof TreeTableModelAdapter) {

						if (((TreeTableModelAdapter)getModel()).getExpanding()) {}
					}
					else
						this.clearSelection();

					if (min != -1 && max != -1) {
						for (int counter = min; counter <= max; counter++) {
							if (listSelectionModel.isSelectedIndex(counter)) {
								TreePath selPath = tree.getPathForRow(counter);
								debug.println("selected Path: " + selPath);
								if (selPath != null) {
									addSelectionPath(selPath);
								}
							}
						}
					}
				}
				finally {
					updatingListSelectionModel = false;
				}
			}
		}

		/**
		 * Class responsible for calling updateSelectedPathsFromSelectedRows
		 * when the selection of the list changse.
		 */
		class ListSelectionHandler implements ListSelectionListener {
			public void valueChanged(ListSelectionEvent e) {
				updateSelectedPathsFromSelectedRows();
			}
		}

	}

	public class TreeTableCellEditor extends AbstractCellEditor implements TableCellEditor {
		public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int r, int c) {
			return tree;
		}
	}

}
