/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.swing.cluster;

import javax.swing.plaf.metal.MetalTreeUI;
import java.awt.event.MouseEvent;
import javax.swing.SwingUtilities;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import java.awt.Rectangle;
/**
 * @author alisa_m
 *
 * a different TreeUI  that allows click on click off capability
 */

public class MatchTreeUI extends MetalTreeUI {
	private boolean ignoreLAChange;
	private int leadRow;
	/**
	  * Returning true signifies a mouse event on the node should toggle
	  * the selection of only the row under mouse.
	  */
	protected boolean isToggleSelectionEvent(MouseEvent event) {
		return true;
	}

	/**
	  * Messaged to update the selection based on a MouseEvent over a
	  * particular row. If the event is a toggle selection event, the
	  * row is either selected, or deselected. If the event identifies
	  * a multi selection event, the selection is updated from the
	  * anchor point. Otherwise the row is selected, and if the event
	  * specified a toggle event the row is expanded/collapsed.
	  */
	protected void selectPathForEvent(TreePath path, MouseEvent event) {
		/* Adjust from the anchor point. */

		if (isMultiSelectEvent(event)) {
			TreePath anchor = getAnchorSelectionPath();
			int anchorRow = (anchor == null) ? -1 : getRowForPath(tree, anchor);

			if (anchorRow == -1 || tree.getSelectionModel().getSelectionMode() == TreeSelectionModel.SINGLE_TREE_SELECTION) {
				tree.setSelectionPath(path);
			}
			else {
				int row = getRowForPath(tree, path);
				TreePath lastAnchorPath = anchor;

				if (row < anchorRow)
					tree.addSelectionInterval(row, anchorRow);
				else
					tree.addSelectionInterval(anchorRow, row);
				lastSelectedRow = row;
				setAnchorSelectionPath(lastAnchorPath);
				setLeadSelectionPath(path);
			}
		}

		// Should this event toggle the selection of this row?
		/* Control toggles just this node. */
		else
			if (isToggleSelectionEvent(event)) {
				if (tree.isPathSelected(path))
					tree.removeSelectionPath(path);
				else
					tree.addSelectionPath(path);
				lastSelectedRow = getRowForPath(tree, path);
				setAnchorSelectionPath(path);
				setLeadSelectionPath(path);
			}
		/* Adjust from the anchor point. */
		else
			if (isMultiSelectEvent(event)) {
				TreePath anchor = getAnchorSelectionPath();
				int anchorRow = (anchor == null) ? -1 : getRowForPath(tree, anchor);

				if (anchorRow == -1 || tree.getSelectionModel().getSelectionMode() == TreeSelectionModel.SINGLE_TREE_SELECTION) {
					tree.setSelectionPath(path);
				}
				else {
					int row = getRowForPath(tree, path);
					TreePath lastAnchorPath = anchor;

					if (row < anchorRow)
						tree.setSelectionInterval(row, anchorRow);
					else
						tree.setSelectionInterval(anchorRow, row);
					lastSelectedRow = row;
					setAnchorSelectionPath(lastAnchorPath);
					setLeadSelectionPath(path);
				}
			}
		/* Otherwise set the selection to just this interval. */
		else
			if (SwingUtilities.isLeftMouseButton(event)) {
				tree.setSelectionPath(path);
				if (isToggleEvent(event)) {
					toggleExpandState(path);
				}
			}
	}
	//
	// The following selection methods (lead/anchor) are covers for the
	// methods in JTree.
	//
	private void setAnchorSelectionPath(TreePath newPath) {
		ignoreLAChange = true;
		try {
			tree.setAnchorSelectionPath(newPath);
		}
		finally {
			ignoreLAChange = false;
		}
	}

	private TreePath getAnchorSelectionPath() {
		return tree.getAnchorSelectionPath();
	}
	private void setLeadSelectionPath(TreePath newPath) {
		setLeadSelectionPath(newPath, false);
	}
	private void setLeadSelectionPath(TreePath newPath, boolean repaint) {
		Rectangle bounds = repaint ? getPathBounds(tree, getLeadSelectionPath()) : null;

		ignoreLAChange = true;
		try {
			tree.setLeadSelectionPath(newPath);
		}
		finally {
			ignoreLAChange = false;
		}
		leadRow = getRowForPath(tree, newPath);

		if (repaint) {
			if (bounds != null)
				tree.repaint(bounds);
			bounds = getPathBounds(tree, newPath);
			if (bounds != null)
				tree.repaint(bounds);
		}
	}

	private TreePath getLeadSelectionPath() {
		return tree.getLeadSelectionPath();
	}

//	private int getLeadSelectionRow() {
//		return leadRow;
//	}

	public boolean isLocationInExpandControl(TreePath path, int mouseX, int mouseY){
		return super.isLocationInExpandControl(path, mouseX, mouseY);
	}

}
