/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tc;

import lapis.*;
import lapisx.util.*;
import java.util.*;
import java.io.Serializable;

public class Evaluator implements Visitor, Cloneable, Serializable {
    public static Debug debug = Debug.QUIET;

    // used by lapis.tools.Test
    public static long totalOps;
    public static long totalLits;
    public static long totalRegexps;
    public static long totalRegions;

    // document we're matching the TC pattern against (constant during 
    // evalution)
    transient Document doc;
    
    // current namespace prefix (changed by Prefix node, used by
    // Id and Assign nodes)
    String prefix;

    // current match flags (changed by CaseSensitivity and View nodes,
    // used by Literal and Regexp nodes)
    int matchFlags;
    static final int CASE_SENSITIVE = 0x01;
    static final int RAW = 0x02;

    // current background (changed by IgnoreBinary nodes,
    // used by NextTo, JustBefore, JustAfter, Starting, Ending nodes)
    lapis.Pattern background;

    // container for warnings generated by evaluation 
    transient TCParseException warnings;

    // pattern library
    transient PatternLibrary lib;

    public Evaluator (Document doc) {
        this (doc, new TCParseException ());
    }

    public Evaluator (PatternLibrary lib) {
        this (null, new TCParseException ());
        this.lib = lib;
    }

    public Evaluator (Document doc, TCParseException warnings) {
        this.doc = doc;
        prefix = "";
        matchFlags = 0; 
        this.warnings = warnings;
        if (doc != null)
            lib = PatternLibrary.getDocumentLibrary (doc);
    }

    public Object clone () {
        try {
            return super.clone ();
        } catch (CloneNotSupportedException e) {
            // shouldn't happen
            throw new RuntimeException (e.toString ());
        }
    }

    RegionSet getBackground () {
        if (background == null) {
            PatternLibrary lib = PatternLibrary.getDocumentLibrary (doc);
            background = lib.get (lapis.parsers.SystemParser.BACKGROUND);
        }
        PatternCache cache = doc.getPatternCache ();
        RegionSet bkgd = cache.get (background);
        if (bkgd == null) {
            bkgd = background.match (doc);
            cache.put (background, bkgd);
        }
        return bkgd;
    }

    public Object forConstant (TC.Constant n) {
        NamedRegionSet sel = n.getSelection ();

        // check whether it's the same document
        //Document orig = sel.getDocument ();  // rcm: call getSingleDocRegionSet(doc)
        //if (orig != null && orig != doc)
        //    return RegionSet.EMPTY;

        //return sel.getRegions (doc.getVersion ());
        return sel.getSingleDocRegionSet(doc);
    }

    public Object forExternalPattern (TC.ExternalPattern n) {
        return n.getPattern ().match (doc);
    }

    static RegionSet timed (long start, 
                            String op,
                            RegionSet r1, RegionSet r2, RegionSet r3,
                            RegionSet result) {
        if ("literal".equals (op))
            ++totalLits;
        else if ("regexp".equals (op))
            ++totalRegexps;
        else 
            ++totalOps;
        if (r1 != null)
            totalRegions += r1.getSize ();
        if (r2 != null)
            totalRegions += r2.getSize ();
        if (r3 != null)
            totalRegions += r3.getSize ();
//         debug.print (op + "(");
//         dump ("", r1);
//         dump (",", r2);
//         dump (",", r3);
//         dump (")-->", result);
//         debug.println (" in "
//                        + (System.currentTimeMillis () - start) + " ms");
        return result;
    }

    static void dump (String name, RegionSet r) {
        if (r == null)
            return;
        debug.print (name + r + "#" + r.getSize ());
    }

    public Object forLiteral (TC.Literal n) {
        MutableRegionSet result = new MutableRegionSet ();

        if (doc == null)
            return result;

        long start = System.currentTimeMillis ();
        String string = n.getString ();
        int len = string.length();
        if (len == 0)
            return result;

        DocumentView view = doc.getDefaultView();
        CoordinateMap cookedToRaw = IdentityMap.IDENTITY;

        // Set searchedDoc to cooked document, if appropriate
        if ((matchFlags & RAW) == 0) {
            view = doc.getCookedView();
            cookedToRaw = doc.getVersion().diff (view.getVersion());
        }

        // Fetch content
        String content = view.getText ();

        if ((matchFlags & CASE_SENSITIVE) == 0) {
            // Handle case-insensitive search by comparing characters in lowercase
            string = string.toLowerCase ();
            int i = 0;
            int j = 0;
            int stop = content.length () - len + 1;
            if (stop > j) {
                while (true) {
                    // invariant: string[0..i-1] == content[j...j+i-1]
                    if (i == len) {
                        result.insert (new Region (j, j+len, doc)
                                           .convert (cookedToRaw));
                        i = 0;
                        //j += len;
                        ++j;
                        if (j >= stop)
                            break;
                    } else if (string.charAt(i) 
                               == Character.toLowerCase (content.charAt (j + i))) {
                        ++i;
                    } else {
                        i = 0;
                        ++j;
                        if (j >= stop)
                            break;
                    }       
                }
            }
        } else {
            // case-sensitive search can use indexOf(), which is hopefully faster 
            for (int p, i=0; (p = content.indexOf (string, i)) != -1; 
                 i = p + 1/*len*/) {
                 result.insert (new Region (p, p + len, doc)
                                            .convert (cookedToRaw));
            }
        }

        return timed (start, "literal", null, null, null, result);
    }

    public Object forRegexp (TC.Regexp n) {
        MutableRegionSet result = new MutableRegionSet ();

        if (doc == null)
            return result;

        long start = System.currentTimeMillis ();
        String expr = n.getString ();

        DocumentView view = doc.getDefaultView ();
        CoordinateMap cookedToRaw = IdentityMap.IDENTITY;

        // Set searchedDoc to cooked document, if appropriate
        if ((matchFlags & RAW) == 0) {
            view = doc.getCookedView();
            cookedToRaw = doc.getVersion().diff(view.getVersion());
        }

        String content = view.getText ();

        try {
            org.apache.regexp.RE rp = 
                new org.apache.regexp.RE 
                (expr,
                 org.apache.regexp.RE.MATCH_MULTILINE // ^ and $ match around \n
                 | org.apache.regexp.RE.MATCH_SINGLELINE // . may match newline
                 | ((matchFlags & CASE_SENSITIVE) != 0
                    ? 0
                    : org.apache.regexp.RE.MATCH_CASEINDEPENDENT));

            int i = 0;
            int end = content.length ();
            while (rp.match (content, i)) {
                int s = rp.getParenStart (0);
                int e = rp.getParenEnd (0);
                result.insert (new Region (s, e, doc)
                                   .convert (cookedToRaw));

                // return only nonoverlapping sets
                if (e == end)
                    break;
                else
                    i = Math.max (i+1, e);

                // return overlapping sets
//                 if (s == end)
//                     break;
//                 else
//                     i = s+1;
            }

        } catch (org.apache.regexp.RESyntaxException e) {
            throw new RuntimeException ("syntax error in regular expression: "
                                        + e.getMessage());
        }
        
        return timed (start, "regexp", null, null, null, result);
    }
    
    public static class AmbiguousError extends TCParseException.ParseError {
	private String[] choices;
	public AmbiguousError(TCToken tok, String message, String[] choices) {
	    super(tok, message);
	    this.choices=choices;
	}
	public String[] getChoices() {
	    for(int i=0;i<choices.length;i++) {
		for(int j=i+1;j<choices.length; j++) {
		    if(choices[i].compareToIgnoreCase(choices[j])>0) {
			String temp = choices[i];
			choices[i]=choices[j];
			choices[j]=temp;
		    }
		}
	    }
	    return choices;
	}
    }
    
    public Object forId (TC.Id n) {
        if (doc == null)
            return RegionSet.EMPTY;

        String id = n.getString ();

        try {
            // first check pattern library
            String name = lib.lookup (id, prefix);
            lapis.Pattern pat = lib.get (name);
            RegionSet matches = pat.match (doc);

            if (matches instanceof ErrorRegionSet) {
                warnings.addError (((ErrorRegionSet)matches).getMessage ());
                matches = RegionSet.EMPTY;
            }

            return matches;
        } catch (PatternLibrary.LabelAmbiguousException e) {
            String msg = e.getMessage () +":";
            String[] choices = e.getChoices ();
            for (int i = 0; i < choices.length; ++i)
                msg += "\n    " + choices[i];
            warnings.addError (new AmbiguousError(n.getToken (), msg, choices));
            return RegionSet.EMPTY;
        } catch (PatternLibrary.LabelNotFoundException e) {
            // then try literal
            debug.println (id + " not found in library, searching for literal");
            RegionSet r = (RegionSet) forLiteral (new TC.Literal (id));
            if (r.isEmpty ())
                warnings.addError (id + " is not a pattern name, and\n" +
                                   "\"" + id + "\" is not found in the document.");

            return r;
        }
    }

    public Object forAny (TC.Any n) {
        if (doc == null)
            return RegionSet.EMPTY;

        PatternLibrary lib = PatternLibrary.getDocumentLibrary (doc);
        return lib.get (lapis.parsers.SystemParser.ANY).match (doc);
    }

    public Object forNothing (TC.Nothing n) {
        return RegionSet.EMPTY;
    }

    public Object forEquals (TC.Equals n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "equalTo", r, null, null, r.equalTo (getBackground ()));
    }

    public Object forBefore (TC.Before n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
//         if (r.getSize () > 1)
//             warnings.addError ("the right side of 'anywhere before' matches more than one region; using the last region as reference point");
        long start = System.currentTimeMillis ();
        return timed (start, "before", r, null, null, r.before ());
    }

    public Object forAfter (TC.After n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
//         if (r.getSize () > 1)
//             warnings.addError ("the right side of 'anywhere after' matches more than one region; using the first region as the reference point");
        long start = System.currentTimeMillis ();
        return timed (start, "after", r, null, null, r.after ());
    }

    public Object forJustBefore (TC.JustBefore n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "just before", r, null, null, r.justBefore (getBackground ()));
    }


    public Object forJustAfter (TC.JustAfter n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "just after", r, null, null, r.justAfter (getBackground ()));
    }

    public Object forBeforeDelimited (TC.BeforeDelimited n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        return r.beforeDelimited ();
    }

    public Object forAfterDelimited (TC.AfterDelimited n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        return r.afterDelimited ();
    }

    public Object forJustBeforeDelimited (TC.JustBeforeDelimited n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        return r.justBeforeDelimited (getBackground ());
    }

    public Object forJustAfterDelimited (TC.JustAfterDelimited n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        return r.justAfterDelimited (getBackground ());
    }

    public Object forIn (TC.In n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "in", r, null, null, r.in ());
    }

    public Object forContains (TC.Contains n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "contains", r, null, null, r.contains ());
    }

    public Object forDescendant (TC.Descendant n) {
	RegionSet r = (RegionSet) n.getArg ().accept (this);
	long start = System.currentTimeMillis ();
	return timed (start, "descendant of", r, null, null, r.descendant ());
    }

    public Object forAncestor (TC.Ancestor n) {
	RegionSet r = (RegionSet) n.getArg ().accept (this);
	long start = System.currentTimeMillis ();
	return timed (start, "ancestor of", r, null, null, r.ancestor ());
    }

    public Object forStarting (TC.Starting n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "starting", r, null, null, r.starting (getBackground ()));
    }

    public Object forEnding (TC.Ending n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "ending", r, null, null, r.ending (getBackground ()));
    }

    public Object forStartingDelimited (TC.StartingDelimited n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "starts with delim", r, null, null, r.startingDelimited (getBackground ()));
    }

    public Object forEndingDelimited (TC.EndingDelimited n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "ends with delim", r, null, null, r.endingDelimited (getBackground ()));
    }

    public Object forOverlapsStart (TC.OverlapsStart n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "overlaps start", r, null, null, r.overlapsStart ());
    }

    public Object forOverlapsEnd (TC.OverlapsEnd n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "overlaps end", r, null, null, r.overlapsEnd ());
    }

    public Object forOverlaps (TC.Overlaps n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "overlaps", r, null, null, r.overlaps ());
    }

    public Object forSeparatedBy (TC.SeparatedBy n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "separated by", r, null, null, r.separatedBy ());
    }

    public Object forStartOf (TC.StartOf n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "start of", r, null, null, r.startOf ());
    }

    public Object forEndOf (TC.EndOf n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "end of", r, null, null, r.endOf ());
    }

    public Object forBalancedUnary (TC.BalancedUnary n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        return r.balancedWith (null);
    }

    public Object forFlatten (TC.Flatten n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "flatten", r, null, null, r.flatten ());
    }

    public Object forUnbind (TC.Unbind n) {
        TC.Id arg = (TC.Id) n.getArg();
        String id = arg.getString ();
        try {
            // first check pattern library
            String name = lib.lookup (id, prefix);
	    lib.remove(name);
	} catch (PatternLibrary.LabelNotFoundException lnfe) {
	    debug.println("label not found; could not unbind");
	} catch (PatternLibrary.LabelAmbiguousException lae) {
	    debug.println("label ambiguous; could not unbind");
	}
	return RegionSet.EMPTY;
    }
    
    public Object forMelt (TC.Melt n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "melt", r, null, null, r.melt ());
    }

    public Object forCrispen (TC.Crispen n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "crispen", r, null, null, r.crispen ());
    }

    public Object forNonempty (TC.Nonempty n) {
        RegionSet r = (RegionSet) n.getArg ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "flatten", r, null, null, r.stripEmptyRegions ());
    }

    public Object forAnd (TC.And n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "and", r1, r2, null, r1.and (r2));
    }

    public Object forOr (TC.Or n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "or", r1, r2, null, r1.or (r2));
    }

    public Object forButNot (TC.ButNot n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "but not", r1, r2, null, r1.butnot (r2));
    }

    public Object forNextTo (TC.NextTo n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "then", r1, r2, null, r1.nextTo (r2, getBackground ()));
    }

    public Object forTrim (TC.Trim n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        long start = System.currentTimeMillis ();
        r2 = r2.melt ();
        return timed (start, "trim", r1, r2, null, r1.trim (r2));
    }

    public Object forInclude (TC.Include n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        long start = System.currentTimeMillis ();
        r2 = r2.melt ();
        return timed (start, "include", r1, r2, null, r1.include (r2));
    }

    public Object forSeq (TC.Seq n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        return r2;
    }

    public Object forBalancedBinary (TC.BalancedBinary n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        return r1.balancedWith (r2);
    }

    public Object forFromTo (TC.FromTo n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        return r1.fromTo (r2);
    }

    public Object forIgnoreBinary (TC.IgnoreBinary n) {
        lapis.Pattern oldBackground = background;
        background = new TC (new TC.Melt (n.getArg2 ()), (Evaluator) clone ());
        try {
            return (RegionSet) n.getArg1 ().accept (this);
        } finally {
            background = oldBackground;
        }
    }

    static long subPatterns = 0;

    public Object forAssign (TC.Assign n) {
        // Form the identifier 
        String name = prefix + "." + n.getString ();

        // Copy the subexpression, binding its identifiers to 
        // the current patterns in the library
        TC.Node boundExpr = TC.bindExpression (n.getArg (), 
                                               lib,
                                               prefix,
                                               warnings);

        // Save the evaluator's state (background, matchFlags, etc.)
        Evaluator evaluator = (Evaluator) clone ();

        // Make a cacheable pattern out of the bound expression
        lapis.Pattern pat = new TC (boundExpr, evaluator);

        // Put a debugging wrapper around the pattern
        if (debug.printEnabled ())
            pat = new DebugPattern (name, pat);

        // Make pattern cacheable
        pat = new CachedPattern (pat);

        // Store the reference in the global pattern library
        PatternLibrary.getGlobalLibrary ().put (name, pat);

        // Return a delayed reference to the pattern
        return new DelayedRegionSet (pat, doc);
    }

    static Hashtable namedPatterns;
    static class DebugPattern implements Pattern {
        String name;
        Pattern pat;

        public DebugPattern (String name, Pattern pat) {
            if (namedPatterns == null)
                namedPatterns = new Hashtable ();

            Integer n = (Integer) namedPatterns.get (name);
            if (n == null)
                n = new Integer (0);
            else
                n = new Integer (n.intValue () + 1);
            namedPatterns.put (name, n);

            this.name = name + n;
            this.pat = pat;
        }

        public RegionSet match (Document doc) {
            long start = System.currentTimeMillis ();
            debug.println ("start " + name);
            try {
                return pat.match (doc);
            } finally {
                debug.println ("end " + name + " " + (System.currentTimeMillis () - start) + " ms");
            }
        }
    }

    public Object forOrButNot (TC.OrButNot n) {
        RegionSet r1 = (RegionSet) n.getArg1 ().accept (this);
        RegionSet r2 = (RegionSet) n.getArg2 ().accept (this);
        RegionSet r3 = (RegionSet) n.getArg3 ().accept (this);
        long start = System.currentTimeMillis ();
        return timed (start, "", r1, r2, r3, r1.and (r2.or (r1.butnot (r3))));
    }

    public Object forNth (TC.Nth n) {
        int count = n.getCount ();
        RegionSet counter = (RegionSet) n.getCounter ().accept (this);
        RegionSet context = (RegionSet) n.getContext ().accept (this);
        return counter.nth (count, context);
    }

    public Object forPrefix (TC.Prefix n) {
        String oldPrefix = prefix;

        String p = n.getPrefix ();
        if (!p.startsWith ("."))
            p = prefix + "." + p;
        prefix = p;

        try {
            return (RegionSet) n.getArg ().accept (this);
        } finally {
            prefix = oldPrefix;
        }
    }

    public Object forCaseSensitivity (TC.CaseSensitivity n) {
        int oldMatchFlags = matchFlags;

        if (n.getCaseSensitive ())
            matchFlags |= CASE_SENSITIVE;
        else
            matchFlags &= ~CASE_SENSITIVE;
        try {
            return (RegionSet) n.getArg ().accept (this);
        } finally {
            matchFlags = oldMatchFlags;
        }
    }

    public Object forView (TC.View n) {
        int oldMatchFlags = matchFlags;

        if ("text".equals (n.getView ()))
            matchFlags |= RAW;
        else
            matchFlags &= ~RAW;
        try {
            return (RegionSet) n.getArg ().accept (this);
        } finally {
            matchFlags = oldMatchFlags;
        }
    }

}
