/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tc;

import lapis.*;
import java.io.*;
import java.util.*;
import lapisx.util.Str;
import lapisx.util.Debug;

public class TC implements Pattern, Parser, Serializable {
    public static Debug debug = Debug.QUIET;

    Node root;    // root of Node expression tree; never null
    TCToken syntax; // root of TCToken syntax tree; may be null
    String href;    // URL or resource name from which tree was parsed 
                    // (or null)
    String expression; // expression from which tree was parsed (or null)

    Evaluator evaluator;  // evaluation context (or null)

    public TC (Node root) {
        this.root = (root != null) ? root : new Nothing ();
    }

    public TC (Node root, String expression) {
        this (root);
        this.expression = expression;
    }

    public TC (Node root, Evaluator evaluator) {
        this (root);
        this.evaluator = evaluator;
    }

    /**
     * Create a TC expression tree by parsing a Reader stream.  Doesn't
     * close the input stream.
     * @param in Stream to parse
     * @exception TCParseException if expression is not a valid TC pattern
     * @exception IOException if an I/O error occurs
     */
    public TC (Reader reader) throws TCParseException, IOException {
        parse (reader, false);
    }

    void parse (Reader reader, boolean justOneExpression) 
        throws TCParseException, IOException {
        TCParseException errors = new TCParseException ();

        syntax = TCParser.parseSyntax (reader, errors);
        reader.close ();

        if (justOneExpression && syntax != null && syntax.sibling != null) {
            errors.addError (syntax.sibling, "too many expressions");
            throw errors;
        }

        root = 
            justOneExpression
            ? TCParser.translateNoun (syntax, errors)
            : TCParser.translateSequence (syntax, errors);
        if (root == null)
            root = new Nothing ();

        if (!errors.isEmpty ())
            throw errors;
    }

    /**
     * Create a TC expression tree by parsing it from a string.
     * @param expression TC expression
     * @exception TCParseException if expression is not a valid TC pattern
     */
    public TC (String expression) throws TCParseException {
        this (expression, false);
    }

    /**
     * Create a TC expression tree by parsing it from a string.
     * @param expression TC expression
     * @param justOneExpression if true, allows expression to contain only
     * one TC expression (as opposed to a sequence of assignments)
     * @exception TCParseException if expression is not a valid TC pattern
     */
    public TC (String expression, boolean justOneExpression) throws TCParseException {
        try {
            parse (new StringReader (expression), justOneExpression);
        } catch (IOException e) {
            throw new TCParseException (e.toString ());
        }
        this.expression = expression;
    }

    /**
     * Create a TC expression tree by parsing it from a string.  
     * By using this method, the caller
     * is asserting that the TC expression is syntactically correct.
     * An unchecked RuntimeException will be thrown if a syntax error is 
     * found.
     * @param expression TC expression
     * @return parsed TC object 
     * @exception RuntimeException if expression is not a valid TC pattern
     */
    public static TC make (String expression) {
        try {
            return new TC (expression);
        } catch (TCParseException e) {
            // caller asserts that this shouldn't happen
            debug.report (e);
            throw new RuntimeException (e.toString ());
        }       
    }


    /**
     * Create a TC expression that matches a literal string.
     * Deals with newlines and quotes properly.
     */
    public static String makeLiteral (String lit) {
        boolean hasLinebreak = (lit.indexOf ('\n') != -1);
        boolean hasSingleQuote = (lit.indexOf ('\'') != -1);
        boolean hasDoubleQuote = (lit.indexOf ('"') != -1);

        if (hasLinebreak || (hasSingleQuote && hasDoubleQuote))
            return getTCRegexp (lit);
        else if (hasDoubleQuote)
            return "'" + lit + "'";
        else
            return "\"" + lit + "\"";
    }

    static String getTCRegexp (String lit) {
        return "/" + Str.escape (lit, "\n\\/().*+?{}", '\\', "n\\/().*+?{}") + "/";
    }


    /**
     * Create a TC expression tree by parsing it from a file.
     * @param f File to parse
     * @exception TCParseException if expression is not a valid TC pattern
     * @exception IOException if an I/O error occurs
     */
    public TC (File srcFile) throws TCParseException, IOException {
        href = srcFile.toString ();

        // first try to load preparsed object file (.tco file)
        File objFile = toObjectFile (srcFile);
        try {
            if (newerThan (objFile, srcFile)) {
                ObjectInputStream in = 
                    new ObjectInputStream (new FileInputStream (objFile));
                root = (TC.Node) in.readObject ();
                in.close ();
                return;
            }
        } catch (Exception e) {
        }

        // if we got here, the .tco file wasn't found or was corrupt.
        // load and parse the source (.tc file) instead.
        parse (new FileReader (srcFile), false);

        // store the parse tree out to a .tco file
        ObjectOutputStream out = 
            new ObjectOutputStream (new FileOutputStream (objFile));
        out.writeObject (root);
        out.close ();

    }

    static File toObjectFile (File f) {
        String pathname = f.getAbsolutePath ();
        return new File (Str.before (pathname, ".tc") + ".tco");
    }

    public static boolean newerThan (File f, File g) {
        long fTime = f.lastModified ();
        long gTime = g.lastModified ();
        return (fTime != 0 && gTime != 0 && fTime >= gTime);
    }

    static boolean isTC(Pattern pat) {
	if(pat instanceof CachedPattern)
	    pat=((CachedPattern) pat).getPattern();
	return (pat instanceof TC);
    }

    /**
     * Create a TC expression tree by parsing it from a 
     * class resource (i.e., a file found inside a
     * JAR archive or in a Java class hierarchy).
     * Returns the root of a TC tree representing the expression
     * @param cls Class to use for Class.getResourceAsStream() call
     * @param name Name to pass to Class.getResourceAsStream()
     * @exception TCParseException if expression is not a valid TC pattern
     * @exception IOException if an I/O error occurs
     */
    public TC (Class cls, String name) throws TCParseException, IOException {
        InputStream stream = cls.getResourceAsStream (name);
        if (stream == null)
            throw new IOException ("Resource " + name + " not found");
        Reader reader = new InputStreamReader (stream);
        parse (reader, false);
        href = name;
    }
    
    /**
     * Create a TC expression tree representing a constant region set.
     * Returns a TC expression that returns sel.regions when it is evaluated
     * on sel.doc, but empty set when evaluated on any other document.
     * @param sel a selection (a RegionSet/Document/DocumentVersion triple)
     */
    public TC (NamedRegionSet sel) {
        root = new Constant (sel);
    }

    public String getHref () {
        return href;
    }

    public String getExpression () {
        return expression;
    }

    public TCToken getSyntax () {
        return syntax;
    }

    public TC.Node getRoot () {
        return root;
    }

	/*
    public String getName () {
		Set set = getNames();
		if (set == null)
	    	return null;
		if (set.size() == 1) {
	    	Iterator i = set.iterator();
	    	return (String)i.next();
		} 
		else 
	    	return null;
    }
	*/
    /*
     * returns the set of Names if the pattern has a name
     */
    public Set getNames () {
		debug.println(root.accept (LABEL_FINDER).getClass().getName());
		return (Set) root.accept (LABEL_FINDER); 
    }

        
	public String parseSyntax() {
        return parseSyntax(syntax);
    }

    private String parseSyntax(TCToken tok) {
    	if(tok == null)
    		return "";
        else return tok.text+parseSyntax(tok.child)+parseSyntax(tok.sibling);
    }
    
    public String toString () {
        return expression != null ? expression : super.toString ();
    }

    static final Visitor LABEL_FINDER = new Visitor () {
        public Object forConstant (TC.Constant n) { return Collections.EMPTY_SET; }
        public Object forExternalPattern (TC.ExternalPattern n) { return Collections.EMPTY_SET; }
        public Object forLiteral (TC.Literal n) { return Collections.EMPTY_SET; }
        public Object forRegexp (TC.Regexp n) { return Collections.EMPTY_SET; }
        public Object forId (TC.Id n) { 
	    Set set =  new HashSet ();
	    debug.println("forID " + n.getString());
	    set.add(n.getString());
	    return set;
	}
	
        public Object forAny (TC.Any n) { return Collections.EMPTY_SET; }
        public Object forNothing (TC.Nothing n) { return Collections.EMPTY_SET; }
        public Object forEquals (TC.Equals n) { return Collections.EMPTY_SET; }
        public Object forBefore (TC.Before n) { return Collections.EMPTY_SET; }
        public Object forAfter (TC.After n) { return Collections.EMPTY_SET; }
        public Object forJustBefore (TC.JustBefore n) { return Collections.EMPTY_SET; }
        public Object forJustAfter (TC.JustAfter n) { return Collections.EMPTY_SET; }
        public Object forBeforeDelimited (TC.BeforeDelimited n) { return Collections.EMPTY_SET; }
        public Object forAfterDelimited (TC.AfterDelimited n) { return Collections.EMPTY_SET; }
        public Object forJustBeforeDelimited (TC.JustBeforeDelimited n) { return Collections.EMPTY_SET; }
        public Object forJustAfterDelimited (TC.JustAfterDelimited n) { return Collections.EMPTY_SET; }
        public Object forIn (TC.In n) { return Collections.EMPTY_SET; }
        public Object forContains (TC.Contains n) { return Collections.EMPTY_SET; }
		
	
	public Object forDescendant (TC.Descendant n) { return Collections.EMPTY_SET; }
	public Object forAncestor (TC.Ancestor n) { return Collections.EMPTY_SET; }
	
	public Object forStarting (TC.Starting n) { return Collections.EMPTY_SET; }
        public Object forEnding (TC.Ending n) { return Collections.EMPTY_SET; }
        public Object forStartingDelimited (TC.StartingDelimited n) { return Collections.EMPTY_SET; }
        public Object forEndingDelimited (TC.EndingDelimited n) { return Collections.EMPTY_SET; }
        public Object forOverlapsStart (TC.OverlapsStart n) { return Collections.EMPTY_SET; }
        public Object forOverlapsEnd (TC.OverlapsEnd n) { return Collections.EMPTY_SET; }
        public Object forOverlaps (TC.Overlaps n) { return Collections.EMPTY_SET; }
        public Object forSeparatedBy (TC.SeparatedBy n) { return Collections.EMPTY_SET; }
        public Object forStartOf (TC.StartOf n) { return Collections.EMPTY_SET; }
        public Object forEndOf (TC.EndOf n) { return Collections.EMPTY_SET; }
        public Object forBalancedUnary (TC.BalancedUnary n) { return Collections.EMPTY_SET; }
        public Object forFlatten (TC.Flatten n) { return Collections.EMPTY_SET; }
        public Object forMelt (TC.Melt n) { return Collections.EMPTY_SET; }
        public Object forCrispen (TC.Crispen n) { return Collections.EMPTY_SET; }
        public Object forNonempty (TC.Nonempty n) { return Collections.EMPTY_SET; }
        public Object forAnd (TC.And n) { return Collections.EMPTY_SET; }
        public Object forOr (TC.Or n) { 
	    Set args = new HashSet();
	    args.addAll((Set)n.getArg1 ().accept (this));
	    args.addAll((Set)n.getArg2 ().accept (this));
	    return args;
	}
        public Object forButNot (TC.ButNot n) { return Collections.EMPTY_SET; }
        public Object forNextTo (TC.NextTo n) { return Collections.EMPTY_SET; }
        public Object forTrim (TC.Trim n) { return Collections.EMPTY_SET; }
        public Object forInclude (TC.Include n) { return Collections.EMPTY_SET; }
        public Object forSeq (TC.Seq n) { return Collections.EMPTY_SET; }
        public Object forBalancedBinary (TC.BalancedBinary n) { return Collections.EMPTY_SET; }
        public Object forFromTo (TC.FromTo n) { return Collections.EMPTY_SET; }
        public Object forIgnoreBinary (TC.IgnoreBinary n) { return Collections.EMPTY_SET; }
        public Object forAssign (TC.Assign n) {
	    Set set =  new HashSet ();
	    set.add(n.getString());
	    return set;
	}
        public Object forOrButNot (TC.OrButNot n) { return Collections.EMPTY_SET; }
        public Object forNth (TC.Nth n) { return Collections.EMPTY_SET; }
        public Object forPrefix (TC.Prefix n) { return Collections.EMPTY_SET; }
	public Object forCaseSensitivity (TC.CaseSensitivity n) { return Collections.EMPTY_SET; }
	public Object forView (TC.View n) { return Collections.EMPTY_SET; }
	public Object forUnbind (TC.Unbind n) { return Collections.EMPTY_SET; }
    };
    //
    // Evaluation
    //

    public RegionSet match (Document doc) {
		long start = System.currentTimeMillis();
        try {
            return matchWithWarnings (doc);
        } catch (TCParseException e) {
            return RegionSet.EMPTY;
        }
	finally {
	    debug.println("TC parser ran in: " + 
			  (System.currentTimeMillis() - start) + 
			  " ms");
	}
    }

    public RegionSet matchWithWarnings (Document doc) throws TCParseException {
        TCParseException warnings = new TCParseException ();
        Evaluator e;
        if (evaluator != null) {
            e = (Evaluator) evaluator.clone();
            e.doc = doc;
            e.warnings = warnings;
	    e.lib = PatternLibrary.getDocumentLibrary(doc);
        } else
            e = new Evaluator (doc, warnings);
        
        RegionSet s = (RegionSet) root.accept (e);
        if (!warnings.isEmpty ())
            throw warnings;

        return s;
    }

    //
    // Binding a set of named patterns into PatternLibrary
    //

    public void bind (PatternLibrary lib) {
        root.accept (new Evaluator (lib));
    }

    public void bind (PatternLibrary lib, Document doc) {
        // no doc-specific patterns for this parser
    }

    public TC toBoundExpression (PatternLibrary lib) {
        return new TC (bindExpression (getRoot (),
                                       lib,
                                       "",
                                       new TCParseException ()),
                       getExpression ());
    }

    public static TC.Node bindExpression (TC.Node node, 
                                          PatternLibrary lib,
                                          String prefix,
                                          TCParseException warnings) {
        return (TC.Node) node.accept 
            (new BindingCloner (lib, prefix, warnings));
    }

    static class BindingCloner extends Cloner {
        PatternLibrary lib;
        String prefix;
        TCParseException warnings;
        public BindingCloner (PatternLibrary lib, 
                              String prefix,
                              TCParseException warnings) {
            this.lib = lib;
            this.prefix = prefix;
            this.warnings = warnings;
        }

        public Object forId (TC.Id n) {
            try {
                lapis.Pattern pat = lib.get (lib.lookup (n.getString (), prefix));
                if (isTC (pat)) {
                    debug.println ("binding " + n.getString ());
                    return new TC.ExternalPattern (pat);
                } else {
                    debug.println ("leaving unbound " + n.getString ());
                    return n;
                }
            } catch (PatternLibrary.LabelAmbiguousException e) {
                String msg = e.getMessage () +":";
                String[] choices = e.getChoices ();
                for (int i = 0; i < choices.length; ++i)
                    msg += "\n    " + choices[i];
                warnings.addError (msg);
                return new TC.Nothing ();
            } catch (PatternLibrary.LabelNotFoundException e) {
                return new TC.Literal (n.getString ());
            }
        }

        public Object forPrefix (TC.Prefix n) {
            String oldPrefix = prefix;
            String p = n.getPrefix ();
            if (!p.startsWith ("."))
                p = prefix + "." + p;
            prefix = p;
            try {
                return super.forPrefix (n);
            } finally {
                prefix = oldPrefix;
            }
        }
    }


    //
    // Printing debugging representation
    //

    public void print () {
        PrintWriter out = new PrintWriter (System.err);
        print (out, 0);
    }

    public void print (final PrintWriter out, final int originalIndent) {
        root.accept (new Traversal () {
            int indent = originalIndent;
            void printNode (TC.Node n) {
                out.println (Str.repeat (" ", indent) + n);
            }
            public Object forConstant (TC.Constant n) {
                printNode (n);
                indent += 2;
                super.forConstant (n);
                indent -= 2;
                return null;
            }
            public Object forExternalPattern (TC.ExternalPattern n) {
                printNode (n);
                indent += 2;
                super.forExternalPattern (n);
                indent -= 2;
                return null;
            }
            public Object forLiteral (TC.Literal n) {
                printNode (n);
                indent += 2;
                super.forLiteral (n);
                indent -= 2;
                return null;
            }
            public Object forRegexp (TC.Regexp n) {
                printNode (n);
                indent += 2;
                super.forRegexp (n);
                indent -= 2;
                return null;
            }
            public Object forId (TC.Id n) {
                printNode (n);
                indent += 2;
                super.forId (n);
                indent -= 2;
                return null;
            }
            public Object forAny (TC.Any n) {
                printNode (n);
                indent += 2;
                super.forAny (n);
                indent -= 2;
                return null;
            }
            public Object forNothing (TC.Nothing n) {
                printNode (n);
                indent += 2;
                super.forNothing (n);
                indent -= 2;
                return null;
            }
            public Object forEquals (TC.Equals n) {
                printNode (n);
                indent += 2;
                super.forEquals (n);
                indent -= 2;
                return null;
            }
            public Object forBefore (TC.Before n) {
                printNode (n);
                indent += 2;
                super.forBefore (n);
                indent -= 2;
                return null;
            }
            public Object forAfter (TC.After n) {
                printNode (n);
                indent += 2;
                super.forAfter (n);
                indent -= 2;
                return null;
            }
            public Object forJustBefore (TC.JustBefore n) {
                printNode (n);
                indent += 2;
                super.forJustBefore (n);
                indent -= 2;
                return null;
            }
            public Object forJustAfter (TC.JustAfter n) {
                printNode (n);
                indent += 2;
                super.forJustAfter (n);
                indent -= 2;
                return null;
            }
            public Object forBeforeDelimited (TC.BeforeDelimited n) {
                printNode (n);
                indent += 2;
                super.forBeforeDelimited (n);
                indent -= 2;
                return null;
            }
            public Object forAfterDelimited (TC.AfterDelimited n) {
                printNode (n);
                indent += 2;
                super.forAfterDelimited (n);
                indent -= 2;
                return null;
            }
            public Object forJustBeforeDelimited (TC.JustBeforeDelimited n) {
                printNode (n);
                indent += 2;
                super.forJustBeforeDelimited (n);
                indent -= 2;
                return null;
            }
            public Object forJustAfterDelimited (TC.JustAfterDelimited n) {
                printNode (n);
                indent += 2;
                super.forJustAfterDelimited (n);
                indent -= 2;
                return null;
            }
            public Object forIn (TC.In n) {
                printNode (n);
                indent += 2;
                super.forIn (n);
                indent -= 2;
                return null;
            }
            public Object forContains (TC.Contains n) {
                printNode (n);
                indent += 2;
                super.forContains (n);
                indent -= 2;
                return null;
            }

	    public Object forDescendant (TC.Descendant n) {
		printNode (n);
		indent += 2;
		super.forDescendant (n);
		indent -= 2;
		return null;
	    }
	    public Object forAncestor (TC.Ancestor n) {
		printNode (n);
		indent += 2;
		super.forAncestor (n);
		indent -= 2;
		return null;
	    }

            public Object forStarting (TC.Starting n) {
                printNode (n);
                indent += 2;
                super.forStarting (n);
                indent -= 2;
                return null;
            }
            public Object forEnding (TC.Ending n) {
                printNode (n);
                indent += 2;
                super.forEnding (n);
                indent -= 2;
                return null;
            }
            public Object forStartingDelimited (TC.StartingDelimited n) {
                printNode (n);
                indent += 2;
                super.forStartingDelimited (n);
                indent -= 2;
                return null;
            }
            public Object forEndingDelimited (TC.EndingDelimited n) {
                printNode (n);
                indent += 2;
                super.forEndingDelimited (n);
                indent -= 2;
                return null;
            }
            public Object forOverlapsStart (TC.OverlapsStart n) {
                printNode (n);
                indent += 2;
                super.forOverlapsStart (n);
                indent -= 2;
                return null;
            }
            public Object forOverlapsEnd (TC.OverlapsEnd n) {
                printNode (n);
                indent += 2;
                super.forOverlapsEnd (n);
                indent -= 2;
                return null;
            }
            public Object forOverlaps (TC.Overlaps n) {
                printNode (n);
                indent += 2;
                super.forOverlaps (n);
                indent -= 2;
                return null;
            }
            public Object forSeparatedBy (TC.SeparatedBy n) {
                printNode (n);
                indent += 2;
                super.forSeparatedBy (n);
                indent -= 2;
                return null;
            }
            public Object forStartOf (TC.StartOf n) {
                printNode (n);
                indent += 2;
                super.forStartOf (n);
                indent -= 2;
                return null;
            }
            public Object forEndOf (TC.EndOf n) {
                printNode (n);
                indent += 2;
                super.forEndOf (n);
                indent -= 2;
                return null;
            }
            public Object forBalancedUnary (TC.BalancedUnary n) {
                printNode (n);
                indent += 2;
                super.forBalancedUnary (n);
                indent -= 2;
                return null;
            }
            public Object forFlatten (TC.Flatten n) {
                printNode (n);
                indent += 2;
                super.forFlatten (n);
                indent -= 2;
                return null;
            }
            public Object forMelt (TC.Melt n) {
                printNode (n);
                indent += 2;
                super.forMelt (n);
                indent -= 2;
                return null;
            }
            public Object forCrispen (TC.Crispen n) {
                printNode (n);
                indent += 2;
                super.forCrispen (n);
                indent -= 2;
                return null;
            }
            public Object forNonempty (TC.Nonempty n) {
                printNode (n);
                indent += 2;
                super.forNonempty (n);
                indent -= 2;
                return null;
            }
            public Object forAnd (TC.And n) {
                printNode (n);
                indent += 2;
                super.forAnd (n);
                indent -= 2;
                return null;
            }
            public Object forOr (TC.Or n) {
                printNode (n);
                indent += 2;
                super.forOr (n);
                indent -= 2;
                return null;
            }
            public Object forButNot (TC.ButNot n) {
                printNode (n);
                indent += 2;
                super.forButNot (n);
                indent -= 2;
                return null;
            }
            public Object forNextTo (TC.NextTo n) {
                printNode (n);
                indent += 2;
                super.forNextTo (n);
                indent -= 2;
                return null;
            }
            public Object forTrim (TC.Trim n) {
                printNode (n);
                indent += 2;
                super.forTrim (n);
                indent -= 2;
                return null;
            }
            public Object forInclude (TC.Include n) {
                printNode (n);
                indent += 2;
                super.forInclude (n);
                indent -= 2;
                return null;
            }
            public Object forSeq (TC.Seq n) {
                printNode (n);
                indent += 2;
                super.forSeq (n);
                indent -= 2;
                return null;
            }
            public Object forFromTo (TC.FromTo n) {
                printNode (n);
                indent += 2;
                super.forFromTo (n);
                indent -= 2;
                return null;
            }
            public Object forBalancedBinary (TC.BalancedBinary n) {
                printNode (n);
                indent += 2;
                super.forBalancedBinary (n);
                indent -= 2;
                return null;
            }
            public Object forIgnoreBinary (TC.IgnoreBinary n) {
                printNode (n);
                indent += 2;
                super.forIgnoreBinary (n);
                indent -= 2;
                return null;
            }
            public Object forAssign (TC.Assign n) {
                printNode (n);
                indent += 2;
                super.forAssign (n);
                indent -= 2;
                return null;
            }
            public Object forOrButNot (TC.OrButNot n) {
                printNode (n);
                indent += 2;
                super.forOrButNot (n);
                indent -= 2;
                return null;
            }
            public Object forNth (TC.Nth n) {
                printNode (n);
                indent += 2;
                super.forNth (n);
                indent -= 2;
                return null;
            }
            public Object forPrefix (TC.Prefix n) {
                printNode (n);
                indent += 2;
                super.forPrefix (n);
                indent -= 2;
                return null;
            }
            public Object forCaseSensitivity (TC.CaseSensitivity n) {
                printNode (n);
                indent += 2;
                super.forCaseSensitivity (n);
                indent -= 2;
                return null;
            }
            public Object forView (TC.View n) {
                printNode (n);
                indent += 2;
                super.forView (n);
                indent -= 2;
                return null;
            }
            public Object forUnbind (TC.Unbind n) {
                printNode (n);
                indent += 2;
                super.forUnbind (n);
                indent -= 2;
                return null;
            }
		
        });
        out.flush ();
    }


    //
    // Node types
    //
    public abstract static class Node implements Serializable {
        public abstract Object accept (Visitor visitor);
        public String toString () {
            return Str.after (getClass ().getName (), "$");
        }
    }

    // 
    // Primitive terms
    //
    public static class Constant extends Node {
        NamedRegionSet sel;
        public Constant (NamedRegionSet sel) {
            this.sel = sel;
        }
        public Constant (Document doc, RegionSet regions) {
            //this.sel = new NamedRegionSet (doc, regions);
			this.sel = new NamedRegionSet (regions);
        }
        public NamedRegionSet getSelection () {
            return sel;
        }
        public Object accept (Visitor v) {
            return v.forConstant (this);
        }
    }
    public static class ExternalPattern extends Node {
        Pattern pat;
        public ExternalPattern (Pattern pat) {
            this.pat = pat;
        }
        public Pattern getPattern () {
            return pat;
        }
        public Object accept (Visitor v) {
            return v.forExternalPattern (this);
        }
    }
    public static class Literal extends Node {
        String literal;
        public Literal (String s) {
            literal = s;
        }
        public String getString () {
            return literal;
        }
        public Object accept (Visitor v) {
            return v.forLiteral (this);
        }
        public String toString () {
            return "\"" + literal + "\"";
        }
    }
    public static class Regexp extends Node {
        String regexp;
        public Regexp (String s) {
            regexp = s;
        }
        public String getString () {
            return regexp;
        }
        public Object accept (Visitor v) {
            return v.forRegexp (this);
        }
        public String toString () {
            return "/" + regexp + "/";
        }
    }
    public static class Id extends Node {
        String id;
        transient TCToken tok;
        public Id (String s) {
            id = s;
        }
        public Id (String s, TCToken tok) {
            id = s;
            this.tok = tok;
        }
        public String getString () {
            return id;
        }
        public TCToken getToken () {
            return tok;
        }
        public Object accept (Visitor v) {
            return v.forId (this);
        }
        public String toString () {
            return id;
        }
    }
    public static class Any extends Node {
        public Any () {
        }
        public Object accept (Visitor v) {
            return v.forAny (this);
        }
    }
    public static class Nothing extends Node {
        public Nothing () {
        }
        public Object accept (Visitor v) {
            return v.forNothing (this);
        }
    }

    //
    // Unary operators
    //
    abstract static class UnaryOp extends Node {
        Node arg;
        public UnaryOp (Node arg) {
            this.arg = arg;
        }
        public Node getArg () {
            return arg;
        }
    }
    public static class Equals extends UnaryOp {
        public Equals (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forEquals (this);
        }
    }
    public static class Before extends UnaryOp {
        public Before (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forBefore (this);
        }
    }
    public static class After extends UnaryOp {
        public After (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forAfter (this);
        }
    }
    public static class JustBefore extends UnaryOp {
        public JustBefore (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forJustBefore (this);
        }
    }
    public static class JustAfter extends UnaryOp {
        public JustAfter (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forJustAfter (this);
        }
    }
    public static class BeforeDelimited extends UnaryOp {
        public BeforeDelimited (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forBeforeDelimited (this);
        }
    }
    public static class AfterDelimited extends UnaryOp {
        public AfterDelimited (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forAfterDelimited (this);
        }
    }
    public static class JustBeforeDelimited extends UnaryOp {
        public JustBeforeDelimited (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forJustBeforeDelimited (this);
        }
    }
    public static class JustAfterDelimited extends UnaryOp {
        public JustAfterDelimited (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forJustAfterDelimited (this);
        }
    }
    public static class In extends UnaryOp {
        public In (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forIn (this);
        }
    }
    public static class Contains extends UnaryOp {
        public Contains (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forContains (this);
        }
    }

    public static class Descendant extends UnaryOp {
	public Descendant (Node arg) {
	    super (arg);
	}
	public Object accept (Visitor v) {
	    return v.forDescendant (this);
	}
    }
    public static class Ancestor extends UnaryOp {
	public Ancestor (Node arg) {
	    super (arg);
	}
	public Object accept (Visitor v) {
	    return v.forAncestor (this);
	}
    }

    public static class Starting extends UnaryOp {
        public Starting (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forStarting (this);
        }
    }
    public static class Ending extends UnaryOp {
        public Ending (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forEnding (this);
        }
    }
    public static class StartingDelimited extends UnaryOp {
        public StartingDelimited (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forStartingDelimited (this);
        }
    }
    public static class EndingDelimited extends UnaryOp {
        public EndingDelimited (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forEndingDelimited (this);
        }
    }
    public static class OverlapsStart extends UnaryOp {
        public OverlapsStart (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forOverlapsStart (this);
        }
    }
    public static class OverlapsEnd extends UnaryOp {
        public OverlapsEnd (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forOverlapsEnd (this);
        }
    }
    public static class Overlaps extends UnaryOp {
        public Overlaps (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forOverlaps (this);
        }
    }
    public static class SeparatedBy extends UnaryOp {
        public SeparatedBy (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forSeparatedBy (this);
        }
    }
    public static class StartOf extends UnaryOp {
        public StartOf (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forStartOf (this);
        }
    }
    public static class EndOf extends UnaryOp {
        public EndOf (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forEndOf (this);
        }
    }
    public static class BalancedUnary extends UnaryOp {
        public BalancedUnary (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forBalancedUnary (this);
        }
    }
    public static class Flatten extends UnaryOp {
        public Flatten (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forFlatten (this);
        }
    }
    public static class Unbind extends UnaryOp {
        public Unbind (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forUnbind (this);
        }
    }
    public static class Melt extends UnaryOp {
        public Melt (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forMelt (this);
        }
    }
    public static class Crispen extends UnaryOp {
        public Crispen (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forCrispen (this);
        }
    }
    public static class Nonempty extends UnaryOp {
        public Nonempty (Node arg) {
            super (arg);
        }
        public Object accept (Visitor v) {
            return v.forNonempty (this);
        }
    }

    //
    // Binary operators
    //
    abstract static class BinaryOp extends Node {
        Node arg1;
        Node arg2;
        public BinaryOp (Node arg1, Node arg2) {
            this.arg1 = arg1;
            this.arg2 = arg2;
        }
        public Node getArg1 () {
            return arg1;
        }
        public Node getArg2 () {
            return arg2;
        }
    }

    public static class And extends BinaryOp {
        public And (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forAnd (this);
        }
    }
    public static class Or extends BinaryOp {
        public Or (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forOr (this);
        }

    }
    public static class ButNot extends BinaryOp {
        public ButNot (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forButNot (this);
        }
    }
    public static class NextTo extends BinaryOp {
        public NextTo (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forNextTo (this);
        }
    }
    public static class Trim extends BinaryOp {
        public Trim (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forTrim (this);
        }
    }
    public static class Include extends BinaryOp {
        public Include (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forInclude (this);
        }
    }
    public static class Seq extends BinaryOp {
        public Seq (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forSeq (this);
        }
    }
    public static class FromTo extends BinaryOp {
        public FromTo (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forFromTo (this);
        }
    }
    public static class BalancedBinary extends BinaryOp {
        public BalancedBinary (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forBalancedBinary (this);
        }
    }
    public static class IgnoreBinary extends BinaryOp {
        public IgnoreBinary (Node arg1, Node arg2) {
            super (arg1, arg2);
        }
        public Object accept (Visitor v) {
            return v.forIgnoreBinary (this);
        }
    }


    // 
    // Other operators
    //
    public static class Assign extends Node {
        String id;
        Node arg;
        public Assign (String s, Node arg) {
            this.arg = arg;
            id = s;
        }
        public Node getArg () {
            return arg;
        }
        public String getString () {
            return id;
        }
        public Object accept (Visitor v) {
            return v.forAssign (this);
        }

        public String toString () {
            return id + " = ";
        }


    }
    public static class Nth extends Node {
        int n;
        Node counter;
        Node context;
        public Nth (int n, Node counter, Node context) {
            this.n = n;
            this.counter = counter;
            this.context = context;
        }
        public int getCount () {
            return n;
        }
        public Node getCounter () {
            return counter;
        }
        public Node getContext () {
            return context;
        }
        public Object accept (Visitor v) {
            return v.forNth (this);
        }

        public String toString () {
            return super.toString () + " " + n;
        }
    }
    public static class OrButNot extends Node {
        Node arg1;
        Node arg2;
        Node arg3;
        public OrButNot (Node arg1, Node arg2, Node arg3) {
            this.arg1 = arg1;
            this.arg2 = arg2;
            this.arg3 = arg3;
        }
        public Node getArg1 () {
            return arg1;
        }
        public Node getArg2 () {
            return arg2;
        }
        public Node getArg3 () {
            return arg3;
        }
        public Object accept (Visitor v) {
            return v.forOrButNot (this);
        }
    }
    public static class Prefix extends Node {
        String prefix;
        Node arg;
        public Prefix (String prefix, Node arg) {
            this.prefix = prefix;
            this.arg = arg;
        }
        public String getPrefix () {
            return prefix;
        }
        public Node getArg () {
            return arg;
        }
        public Object accept (Visitor v) {
            return v.forPrefix (this);
        }
        public String toString () {
            return super.toString () + " " + prefix;
        }
    }
    public static class CaseSensitivity extends Node {
        boolean enabled;
        Node arg;
        public CaseSensitivity (boolean enabled, Node arg) {
            this.enabled = enabled;
            this.arg = arg;
        }
        public boolean getCaseSensitive () {
            return enabled;
        }
        public Node getArg () {
            return arg;
        }
        public Object accept (Visitor v) {
            return v.forCaseSensitivity (this);
        }
        public String toString () {
            return super.toString () + " " + enabled;
        }
    }
    public static class View extends Node {
        String view;
        Node arg;
        public View (String view, Node arg) {
            this.view = view;
            this.arg = arg;
        }
        public String getView () {
            return view;
        }
        public Node getArg () {
            return arg;
        }
        public Object accept (Visitor v) {
            return v.forView (this);
        }
        public String toString () {
            return super.toString () + " " + view;
        }
    }


    //
    // Converting to related patterns
    //
    
    public TC getPointPattern () {
        return new TC ((Node) root.accept (POINT_CONVERTER));
    }
    
    static final Visitor POINT_CONVERTER = new Cloner () {
        public Object forBefore (TC.Before n) {
            return new TC.StartOf (n.getArg ());
        }
        public Object forAfter (TC.After n) {
            return new TC.EndOf (n.getArg ());
        }
        public Object forJustBefore (TC.JustBefore n) {
            return new TC.StartOf (n.getArg ());
        }
        public Object forJustAfter (TC.JustAfter n) {
            return new TC.EndOf (n.getArg ());
        }
        public Object forBeforeDelimited (TC.BeforeDelimited n) {
            return new TC.StartOf (n.getArg ());
        }
        public Object forAfterDelimited (TC.AfterDelimited n) {
            return new TC.EndOf (n.getArg ());
        }
        public Object forJustBeforeDelimited (TC.JustBeforeDelimited n) {
            return new TC.StartOf (n.getArg ());
        }
        public Object forJustAfterDelimited (TC.JustAfterDelimited n) {
            return new TC.EndOf (n.getArg ());
        }
        public Object forStarting (TC.Starting n) {
            return new TC.StartOf (n.getArg ());
        }
        public Object forEnding (TC.Ending n) {
            return new TC.EndOf (n.getArg ());
        }
        public Object forStartingDelimited (TC.StartingDelimited n) {
            return new TC.StartOf (n.getArg ());
        }
        public Object forEndingDelimited (TC.EndingDelimited n) {
            return new TC.EndOf (n.getArg ());
        }
    };


    public TC getUniversePattern () {
        return new TC ((Node) root.accept (UNIVERSE_CONVERTER));
    }

    static final Visitor UNIVERSE_CONVERTER = new Cloner () {
        public Object forAnd (TC.And n) {
            return n.getArg1 ().accept (this);
        }
        //public Object forOr (TC.Or n);
        public Object forButNot (TC.ButNot n) {
            return n.getArg1 ().accept (this);
        }
        public Object forOrButNot (TC.OrButNot n) {
            return n.getArg1 ().accept (this);
        }
        public Object forEquals (TC.Equals n) {
            return n.getArg ().accept (this);
        }
        public Object forBefore (TC.Before n) {
            return n;
        }
        public Object forAfter (TC.After n) {
            return n;
        }
        public Object forJustBefore (TC.JustBefore n) {
            return n;
        }
        public Object forJustAfter (TC.JustAfter n) {
            return n;
        }
        public Object forBeforeDelimited (TC.BeforeDelimited n) {
            return n;
        }
        public Object forAfterDelimited (TC.AfterDelimited n) {
            return n;
        }
        public Object forJustBeforeDelimited (TC.JustBeforeDelimited n) {
            return n;
        }
        public Object forJustAfterDelimited (TC.JustAfterDelimited n) {
            return n;
        }
        public Object forIn (TC.In n) {
            return n;
        }
        public Object forContains (TC.Contains n) {
            return n;
        }

	public Object forDescendant (TC.Descendant n) {
	    return n;
	}
	public Object forAncestor (TC.Ancestor n) {
	    return n;
	}

        public Object forStarting (TC.Starting n) {
            return n;
        }
        public Object forEnding (TC.Ending n) {
            return n;
        }
        public Object forStartingDelimited (TC.StartingDelimited n) {
            return n;
        }
        public Object forEndingDelimited (TC.EndingDelimited n) {
            return n;
        }
        public Object forOverlapsStart (TC.OverlapsStart n) {
            return n;
        }
        public Object forOverlapsEnd (TC.OverlapsEnd n) {
            return n;
        }
        public Object forOverlaps (TC.Overlaps n) {
            return n;
        }
        public Object forSeparatedBy (TC.SeparatedBy n) {
            return n;
        }
        public Object forStartOf (TC.StartOf n) {
            return n;
        }
        public Object forEndOf (TC.EndOf n) {
            return n;
        }
        public Object forBalancedUnary (TC.BalancedUnary n) {
            return n;
        }
        public Object forBalancedBinary (TC.BalancedBinary n){
            return n;
        }
        public Object forFromTo (TC.FromTo n){
            return n;
        }
        //public Object forConstant (TC.Constant n);
        //public Object forExternalPattern (TC.ExternalPattern n);
        //public Object forLiteral (TC.Literal n);
        //public Object forRegexp (TC.Regexp n);
        //public Object forId (TC.Id n);
        //public Object forAny (TC.Any n);
        //public Object forNothing (TC.Nothing n);
        public Object forNextTo (TC.NextTo n) {
            return n;
        }
        public Object forTrim (TC.Trim n) {
            return n;
        }
        public Object forInclude (TC.Include n) {
            return n;
        }
        public Object forAssign (TC.Assign n) {
            return n;
        }
        public Object forSeq (TC.Seq n) {
            return n.getArg2 ().accept (this);
        }
        //public Object forIgnoreBinary (TC.IgnoreBinary n);
        public Object forNth (TC.Nth n){
            return n.getCounter ().accept (this);
        }
        public Object forFlatten (TC.Flatten n) {
            return n;
        }
        public Object forUnbind (TC.Unbind n) {
            return n;
        }
        public Object forMelt (TC.Melt n) {
            return n;
        }
        public Object forCrispen (TC.Crispen n) {
            return n;
        }
        public Object forNonempty (TC.Nonempty n) {
            return n;
        }
    };



    public TC getObjectPattern () {
        return new TC ((Node) root.accept (OBJECT_CONVERTER));
    }

    static final Visitor OBJECT_CONVERTER = new Cloner () {
        public Object forAnd (TC.And n) {
            return n.getArg1 ().accept (this);
        }
        //public Object forOr (TC.Or n);
        public Object forButNot (TC.ButNot n) {
            return n.getArg1 ().accept (this);
        }
        public Object forOrButNot (TC.OrButNot n) {
            return n.getArg1 ().accept (this);
        }
        public Object forEquals (TC.Equals n) {
            return n.getArg ().accept (this);
        }
        public Object forBefore (TC.Before n) {
            return n.getArg ().accept (this);
        }
        public Object forAfter (TC.After n) {
            return n.getArg ().accept (this);
        }
        public Object forJustBefore (TC.JustBefore n) {
            return n.getArg ().accept (this);
        }
        public Object forJustAfter (TC.JustAfter n) {
            return n.getArg ().accept (this);
        }
        public Object forBeforeDelimited (TC.BeforeDelimited n) {
            return n.getArg ().accept (this);
        }
        public Object forAfterDelimited (TC.AfterDelimited n) {
            return n.getArg ().accept (this);
        }
        public Object forJustBeforeDelimited (TC.JustBeforeDelimited n) {
            return n.getArg ().accept (this);
        }
        public Object forJustAfterDelimited (TC.JustAfterDelimited n) {
            return n.getArg ().accept (this);
        }
        public Object forIn (TC.In n) {
            return n.getArg ().accept (this);
        }
        public Object forContains (TC.Contains n) {
            return n.getArg ().accept (this);
        }

	public Object forDescendant (TC.Descendant n) {
	    return n.getArg ().accept (this);
	}
	public Object forAncestor (TC.Ancestor n) {
	    return n.getArg ().accept (this);
	}

        public Object forStarting (TC.Starting n) {
            return n.getArg ().accept (this);
        }
        public Object forEnding (TC.Ending n) {
            return n.getArg ().accept (this);
        }
        public Object forStartingDelimited (TC.StartingDelimited n) {
            return n.getArg ().accept (this);
        }
        public Object forEndingDelimited (TC.EndingDelimited n) {
            return n.getArg ().accept (this);
        }
        public Object forOverlapsStart (TC.OverlapsStart n) {
            return n.getArg ().accept (this);
        }
        public Object forOverlapsEnd (TC.OverlapsEnd n) {
            return n.getArg ().accept (this);
        }
        public Object forOverlaps (TC.Overlaps n) {
            return n.getArg ().accept (this);
        }
        public Object forSeparatedBy (TC.SeparatedBy n) {
            return n.getArg ().accept (this);
        }
        public Object forStartOf (TC.StartOf n) {
            return n.getArg ().accept (this);
        }
        public Object forEndOf (TC.EndOf n) {
            return n.getArg ().accept (this);
        }
        public Object forBalancedUnary (TC.BalancedUnary n) {
            return n;
        }
        public Object forBalancedBinary (TC.BalancedBinary n){
            return n;
        }
        public Object forFromTo (TC.FromTo n){
            return n;
        }
        //public Object forConstant (TC.Constant n);
        //public Object forExternalPattern (TC.ExternalPattern n);
        //public Object forLiteral (TC.Literal n);
        //public Object forRegexp (TC.Regexp n);
        //public Object forId (TC.Id n);
        //public Object forAny (TC.Any n);
        //public Object forNothing (TC.Nothing n);
        public Object forNextTo (TC.NextTo n) {
            return n;
        }
        public Object forTrim (TC.Trim n) {
            return n;
        }
        public Object forInclude (TC.Include n) {
            return n;
        }
        public Object forAssign (TC.Assign n) {
            return n;
        }
        public Object forSeq (TC.Seq n) {
            return n.getArg2 ().accept (this);
        }
        //public Object forIgnoreBinary (TC.IgnoreBinary n);
        public Object forNth (TC.Nth n){
            return n.getCounter ().accept (this);
        }
        public Object forFlatten (TC.Flatten n) {
            return n;
        }
        public Object forUnbind (TC.Unbind n) {
            return n;
        }
        public Object forMelt (TC.Melt n) {
            return n;
        }
        public Object forCrispen (TC.Crispen n) {
            return n;
        }
        public Object forNonempty (TC.Nonempty n) {
            return n;
        }
    };


}
