/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tc;

import java.io.*;
import java.util.*;
import lapisx.net.URLUtil;
import lapisx.io.FileUtil;

public class TCParseException extends Exception {
    String href = null;
    Vector errorVector = new Vector ();
    int offset = 0;

    public static class ParseError {
        public TCToken tok;
        public String message;
        public ParseError (TCToken tok, String message) {
            this.tok = tok;
            this.message = message;
        }
	public TCToken getToken() {
	    return tok;
	}
	public String toString () {
            return message;
        }
    }

    public TCParseException () {
    }

    public TCParseException (String message) {
        addError (message);
    }

    public Enumeration errors () {
        return errorVector.elements ();
    }
    
    public int getErrorCount () {
        return errorVector.size ();
    }

    public boolean isEmpty () {
        return errorVector.isEmpty ();
    }

    public String getMessage () {
        if (errorVector.isEmpty ())
            return "no errors";

        StringBuffer result = new StringBuffer ();
        for (Enumeration e = errors (); e.hasMoreElements (); ) {
            result.append (e.nextElement ().toString ());
            result.append ('\n');
        }
        return result.toString ();
    }


    public int getStart () {
        if (errorVector.isEmpty ())
            return 0;
        ParseError err = (ParseError) errorVector.elementAt (0);
        if (err.tok == null)
            return 0;
        return err.tok.start;
    }

    public int getEnd () {
        if (errorVector.isEmpty ())
            return 0;
        ParseError err = (ParseError) errorVector.elementAt (0);
        if (err.tok == null)
            return 0;
        return err.tok.end;
    }

    public void printErrors (PrintWriter out) {
        for (Enumeration e = errors (); e.hasMoreElements (); ) {
            ParseError err = (ParseError)e.nextElement ();
            if (href != null)
                out.print (href + ":");
            if (err.tok != null)
                out.print ((err.tok.line+1) + ":");
            out.println (err.message);
        }
        out.flush ();
    }

    public void addError (String message) {
        addError (null, message);
    }

    public void addError (TCToken tok, String message) {
        addError (new ParseError (tok, message));
    }

    public void addError (ParseError err) {
        errorVector.addElement (err);
    }
    
    public void addOffset (int offset) {
        this.offset += offset;
    }

    public void setStringContext (String expression) {
        for (Enumeration e = errors (); e.hasMoreElements (); ) {
            ParseError err = (ParseError)e.nextElement ();
            if (err.tok != null)
                err.message += "\n    " + highlight (expression, err.tok);
        }
    }

    public void setHrefContext (String href) {
        this.href = href;
        try {
            setStringContext (FileUtil.load (URLUtil.HrefToURL (href)
                                             .openConnection ()
                                             .getInputStream ()));
        } catch (IOException e) {
            TC.debug.report (e);
        }
    }

    private String highlight (String s, TCToken tok) {
        int len = s.length ();
        int start = Math.max (0, Math.min (len, offset + tok.start));
        int end = Math.max (0, Math.min (len, offset + tok.end));

        // Find boundaries of line containing tok
        int lineStart = Math.max (s.lastIndexOf ('\n', start),
                                  s.lastIndexOf ('\r', start)) + 1;
        int nl = s.indexOf ('\n', end);
        if (nl == -1)
            nl = len;
        int cr = s.indexOf ('\r', end);
        if (cr == -1)
            cr = len;
        int lineEnd = Math.min (cr, nl);

        // Return just tok's line, with tok highlighted
        return s.substring (lineStart, start)
                + ">>"
                + s.substring (start, end)
                + "<<"
                + s.substring (end, lineEnd);
    }

}
