/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tc;

import java.io.*;
import java.util.*;
import lapisx.util.Str;

public abstract class TCParser implements TCSymbols {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    static int debugFlags = 0;
    static final int STACK_TRACES = 0x1;
    static final int RECONSTRUCT_STACK = 0x2;

    public static TCToken parseSyntax (Reader reader, 
                                       TCParseException errors) 
        throws IOException {
        TCLexer scanner = new TCLexer (reader);
        scanner.setErrorHolder (errors);

        Stack stack = new Stack ();

        // Push a fake root token on the stack to act as
        // an anchor for other tokens.  Put it at 
        // line -1, column -1, so that all other tokens
        // (which must have column >= 0) end up as
        // children of the root.
        TCToken root = new TCToken (NIL, "",
                                    -1, -1, -1,
                                    0, 0);
        stack.push (root);

        int maxline = -1;
        TCToken tok = root;
        try {
            while ((tok = scanner.nextToken ()) != null) {
                maxline = Math.max (tok.line, maxline);

                if ((debugFlags & RECONSTRUCT_STACK) != 0)
                    System.err.println (tok.line + ": " + stack);

                TCToken tos = (TCToken)stack.peek ();

                // pop impossible parents off stack -- tokens that start to the
                // right of tok, or tokens that already have their quota
                // of children.
                while ((tos.sym != LEFTPAREN
                        && tos.sym != LEFTCURLY)
                       /* can't pop these delimiters without closing them first */

                       && (tos.columnStart >= tok.columnStart 
                           || noMoreChildren (tos))
                       && !(tos.columnStart == tok.columnStart
                            && ((tok.sym == OR && tos.sym == EITHER)
                                || (tok.sym == TO && tos.sym == FROM)
                                || (tok.sym == WITH && tos.sym == BALANCES)))) {
                    stack.pop ();
                    tos = (TCToken)stack.peek ();
                }


                // scan up stack for preferred parent -- e.g., OR prefers 
                // EITHER.
                if (tok.sym == OR)
                    tos = popTo (stack, EITHER);
                else if (tok.sym == TO)
                    tos = popTo (stack, FROM);
                else if (tok.sym == WITH)
                    tos = popTo (stack, BALANCES);
                else if (tok.sym == RIGHTPAREN) {
                    tos = popTo (stack, LEFTPAREN);
                    if (tos.sym == LEFTPAREN) {
                        tos.sym = CLOSED_LEFT;
                        tok.sym = CLOSED_RIGHT;
                    }
                } else if (tok.sym == RIGHTCURLY) {
                    tos = popTo (stack, LEFTCURLY);
                    if (tos.sym == LEFTCURLY) {
                        tos.sym = CLOSED_LEFT;
                        tok.sym = CLOSED_RIGHT;
                    }
                }

                // Certain parents need an implicit ANY token for 
                // multiple constraints.
                // Create it when the second child appears.
                if (tos.child != null 
                    && tos.child.sibling == null
                    && (((tos.sym == EITHER || tos.sym == OR) && tok.sym != OR)
                        || tos.sym == NOT)) {
                    TCToken child = tos.child;
                    TCToken implicit = new TCToken (ANY, "", 
                                                    child.line, 
                                                    child.columnStart-1, child.columnStart,
                                                    child.start-1, child.start);
                    tos.child = implicit;
                    implicit.child = child;
                    stack.push (implicit);
                    tos = implicit;
                }

                // connect tok to tos
                tos.addChild (tok);

                // push it on stack
                stack.push (tok);
            }
        } catch (EmptyStackException e) {
            throw new RuntimeException ("stack underflow: " + tok);
        }

        // All tokens should have been added as descendents
        // of fake root token.  Return the first token.
        root = root.child;

        if ((debugFlags & RECONSTRUCT_STACK) != 0) {
            // Reconstruct the stack at every line
            for (int line = 0; line <= maxline; ++line)
                System.err.println (line + ": " + 
                                    reconstructStackAtLine (root, line));
        }

        return root;
    }

    // search up stack for sym
    static TCToken popTo (Stack stack, int sym) {
        TCToken tos = (TCToken)stack.peek ();
        for (int i = stack.size () - 1; i >= 0; --i) {
            TCToken t = (TCToken)stack.elementAt (i);
            if (t.sym == sym) {
                // pop up to t
                while (tos != t) {
                    stack.pop ();
                    tos = (TCToken)stack.peek ();
                }
                return tos;
            }
        }
        return tos;
    }

    static boolean noMoreChildren (TCToken tok) {
        switch (tok.sym) {
        case EQUALS:
        case BEFORE:
        case AFTER:
//         case BEFORE_DELIMITED:
//         case AFTER_DELIMITED:
        case JUST_BEFORE:
        case JUST_AFTER:
//         case JUST_BEFORE_DELIMITED:
//         case JUST_AFTER_DELIMITED:
        case CONTAINS:
//         case STARTING_DELIMITED:
//         case ENDING_DELIMITED:
        case IN:
	case DESCENDANT:
	case ANCESTOR:
        case OVERLAPS:
        case OVERLAPS_START:
        case OVERLAPS_END:
        case STARTING:
        case ENDING:
        case ANYWHERE_BEFORE:
        case ANYWHERE_AFTER:
        case IS:
        case OR:
        case THEN:
        case SEPARATED_BY:
        case IGNORE:
        case TRIM:
        case INCLUDE:
            return (tok.child != null);
        case BALANCES:
            return (tok.child != null && tok.child.sibling != null);
        default:
            return false;
        }
    }

    public static Stack reconstructStackAtLine (TCToken root, 
                                                int line) {
        Stack stack = new Stack ();
        TCToken tok = root;

        while (tok != null && tok.line < line) {
            // Find latest sibling which is still above line
            while (tok.sibling != null
                   && tok.sibling.line < line)
                tok = tok.sibling;
            
            stack.push (tok);
            tok = tok.child;
        }

        return stack;
    }

    static TC.Node translateSequence (TCToken tok,
                                      TCParseException errors) {
        TC.Node expr = null;
        for (; tok != null; tok = tok.sibling) {
            TC.Node node = translateAssignment (tok, errors);
            expr = (expr != null)
                ? new TC.Seq (expr, node)
                : node;
        }
        return (expr != null) ? expr : new TC.Nothing ();
    }

    static TC.Node translateAssignment (TCToken tok, TCParseException errors) {
        switch (tok.sym) {
        case CASE_SENSITIVE:
        case NOT_CASE_SENSITIVE:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.CaseSensitivity 
                    ((tok.sym == CASE_SENSITIVE), translateSequence (tok.child, errors));
            break;
                
        case AS_TEXT:
        case AS_HTML:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.View ((tok.sym == AS_TEXT) ? "text" : "html",
                                    translateSequence (tok.child, errors));
            break;

        case CLOSED_LEFT:
        case CLOSED_RIGHT:
            return translateSequence (tok.child, errors);

        default:
            return translateNoun (tok, errors);
        }

        return new TC.Nothing (); // by default
    }


    static TC.Node translateConstraints (TC.Node expr, 
                                         TCToken tok, 
                                         TCParseException errors) {
        TC.Node noun;

        for (; tok != null; tok = tok.sibling) {
            if (isNoun (tok))
                 recordError (tok, "must be constraint operator", errors);

            switch (tok.sym) {
            case EITHER:
                {
                    TCToken firstEither = tok;
                    TC.Node positive = null, negative = null;
                    // positive is an OR of the subexprs not preceded by NOT
                    // negative is an AND of the subexprs preceded by NOT
                    while (tok != null && (tok == firstEither
                                           || tok.sym == OR)) {
                        if (tok.child == null)
                            recordError (tok, "must have right-hand side", errors);
                        else if (tok.child.sym == NOT) {
                            if (tok.child.child == null)
                                recordError (tok.child, "must have right-hand side", errors);
                            else {
                                TC.Node node = translateNoun (tok.child.child, errors);
                                negative = negative != null
                                    ? new TC.And (negative, node)
                                    : node;
                            }
                        } else {
                            TC.Node node = translateNoun (tok.child, errors);
                            positive = positive != null
                                ? new TC.Or (positive, node)
                                : node;
                        }

                        tok = (tok == firstEither) 
                            ? tok.child.sibling
                            : tok.sibling;
                    }

                    if (negative != null)
                        recordError (firstEither, "syntax error", errors);
                    else if (positive != null && negative != null)
                        expr = new TC.OrButNot (expr != null ? expr : new TC.Nothing (), 
                                                positive, 
                                                negative);
                    else if (negative != null) {
                        if (expr == null)
                            recordError (tok, "must have left-hand side", errors);
                        else 
                            expr = new TC.ButNot (expr, negative);
                    } else {
                        // Translate remaining children as constraints
                        // on entire either expression.  FIX: make this
                        // work when negative!=null, too
                        positive = translateConstraints (positive, tok, errors);
                        expr = (expr != null) ? new TC.And (expr, positive) : positive;
                    }
                    
                    tok = firstEither;
                }
                break;

            case OR: // OR appearing by itself (with no EITHER)
                noun = translateNoun (tok.child, errors);
                expr = (expr != null) ? new TC.Or (expr, noun) : noun;
                break;

            case AND:
                noun = translateNoun (tok.child, errors);
                expr = (expr != null) ? new TC.And (expr, noun) : noun;
                break;

            case NOT:
                if (tok.child == null)
                    recordError (tok, "must have right-hand side", errors);
                else if (expr == null)
                    recordError (tok, "must have left-hand side", errors);
                else
                    expr = new TC.ButNot (expr, translateNoun (tok.child, errors));
                break;

            case THEN:
                if (tok.child == null)
                    recordError (tok, "must have right-hand side", errors);
                else if (expr == null)
                    recordError (tok, "must have left-hand side", errors);
                else
                    expr = new TC.NextTo (expr, translateNoun (tok.child, errors));
                break;
                
            case TRIM:
                if (tok.child == null)
                    recordError (tok, "must have right-hand side", errors);
                else if (expr == null)
                    recordError (tok, "must have left-hand side", errors);
                else
                    expr = new TC.Trim (expr, translateNoun (tok.child, errors));
                break;

            case INCLUDE:
                if (tok.child == null)
                    recordError (tok, "must have right-hand side", errors);
                else if (expr == null)
                    recordError (tok, "must have left-hand side", errors);
                else
                    expr = new TC.Include (expr, translateNoun (tok.child, errors));
                break;

            case IGNORE:
                if (tok.child == null)
                    recordError (tok, "must have right-hand side", errors);
                else if (expr == null)
                    recordError (tok, "must have left-hand side", errors);
                else
                    expr = new TC.IgnoreBinary (expr, translateNoun (tok.child, errors));
                break;

            case CLOSED_LEFT:
            case CLOSED_RIGHT:
                if (tok.child != null) {
                    if (expr != null)
                        expr = translateConstraints (expr, tok.child, errors);
                    else
                        expr = translateConstraints (translateNoun (tok.child, errors),
                                                     tok.child.sibling, errors);
                }
                break;

            default:
                noun = translateNoun (tok, errors);
                expr = (expr != null) ? new TC.And (expr, noun) : noun;
                break;
            }
        }

        return expr;
    }

    // Translate tok and its subtree to a TC.Node
    static TC.Node translateNoun (TCToken tok, TCParseException errors) {
        if (tok == null)
            return new TC.Nothing ();

            // Combine expr with translated node
        switch (tok.sym) {
            /* Nouns */
        case NOTHING:
            return translateConstraints (new TC.Nothing (), tok.child, errors);
        case ANY:
            return translateConstraints (new TC.Any (), tok.child, errors);
        case LIT:
            return translateConstraints (new TC.Literal (tok.text), tok.child, errors);
        case REGEXP:
            return translateConstraints (new TC.Regexp (tok.text), tok.child, errors);
        case ID:
            if (tok.child != null && tok.child.sym == IS) {
                if (tok.child.child == null)
                    recordError (tok.child, 
                                 "IS must have right-hand side", errors);
                else if (tok.child.sibling != null)
                    recordError (tok.child.sibling, 
                                 "syntax error", errors);
                else
                    return new TC.Assign (tok.text, 
                                          translateNoun (tok.child.child, errors));
            } else
                return translateConstraints (new TC.Id (tok.text, tok), tok.child, errors);
            break;
            
            /* Relops */
        case EQUALS:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.Equals (translateNoun (tok.child, errors));
            break;
        case ANYWHERE_BEFORE:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.Before (translateNoun (tok.child, errors));
            break;
        case ANYWHERE_AFTER:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.After (translateNoun (tok.child, errors));
            break;
        case JUST_BEFORE:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.JustBefore 
                    (translateNoun (tok.child, errors));
            break;
        case JUST_AFTER:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.JustAfter 
                    (translateNoun (tok.child, errors));
            break;
//         case BEFORE_DELIMITED:
//             if (tok.child == null)
//                 recordError (tok, "must have right-hand side", errors);
//             else
//                 return new TC.BeforeDelimited 
//                     (translateNoun (tok.child, errors));
//             break;
//         case AFTER_DELIMITED:
//             if (tok.child == null)
//                 recordError (tok, "must have right-hand side", errors);
//             else
//                 return new TC.AfterDelimited 
//                     (translateNoun (tok.child, errors));
//             break;
//         case JUST_BEFORE_DELIMITED:
//             if (tok.child == null)
//                 recordError (tok, "must have right-hand side", errors);
//             else
//                 return new TC.JustBeforeDelimited 
//                     (translateNoun (tok.child, errors));
//             break;
//         case JUST_AFTER_DELIMITED:
//             if (tok.child == null)
//                 recordError (tok, "must have right-hand side", errors);
//             else
//                 return new TC.JustAfterDelimited 
//                     (translateNoun (tok.child, errors));
//             break;
        case IN:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.In (translateNoun (tok.child, errors));
            break;
        case CONTAINS:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.Contains (translateNoun (tok.child, errors));
            break;
	case DESCENDANT:
	    if (tok.child == null)
		recordError (tok, "must have right-hand side", errors);
	    else
		return new TC.Descendant (translateNoun (tok.child, errors));
	    break;
	case ANCESTOR:
	    if (tok.child == null)
		recordError (tok, "must have right-hand side", errors);
	    else
		return new TC.Ancestor (translateNoun (tok.child, errors));
	    break;
        case STARTING:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.Starting (translateNoun (tok.child, errors));
            break;
        case ENDING:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.Ending (translateNoun (tok.child, errors));
            break;
        case STARTOF:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return translateConstraints (new TC.StartOf (translateNoun (tok.child, errors)),
                                             tok.child.sibling, errors);
            break;
        case ENDOF:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return translateConstraints (new TC.EndOf (translateNoun (tok.child, errors)),
                                             tok.child.sibling, errors);
            break;
//         case STARTING_DELIMITED:
//             if (tok.child == null)
//                 recordError (tok, "must have right-hand side", errors);
//             else
//                 return new TC.StartingDelimited (translateNoun (tok.child, errors));
//             break;
//         case ENDING_DELIMITED:
//             if (tok.child == null)
//                 recordError (tok, "must have right-hand side", errors);
//             else
//                 return new TC.EndingDelimited (translateNoun (tok.child, errors));
//             break;
        case OVERLAPS:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.Overlaps (translateNoun (tok.child, errors));
            break;
        case OVERLAPS_START:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.OverlapsStart (translateNoun (tok.child, errors));
            break;
        case OVERLAPS_END:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.OverlapsEnd (translateNoun (tok.child, errors));
            break;
        case SEPARATED_BY:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.SeparatedBy (translateNoun (tok.child, errors));
            break;
        case BALANCES:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else {
                TC.Node open = translateNoun (tok.child, errors);
                if (tok.child.sibling != null && tok.child.sibling.sym == WITH) {
                    TCToken with = tok.child.sibling;
                    TC.Node close = translateNoun (with.child, errors);
                    return new TC.BalancedBinary (open, close);
                } else
                    return new TC.BalancedUnary (open);
            }
            break;
                
        case FLATTEN:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.Flatten (translateNoun (tok.child, errors));
            break;

	case UNBIND:
	    if (tok.child == null)
		recordError (tok, "must have right-hand side", errors);
	    else if (tok.child.sym != ID) 
		recordError (tok, "UNBIND must be followed by a name", errors);
	    else return new TC.Unbind (translateNoun (tok.child, errors));
	    break;
	    
        case MELT:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.Melt (translateNoun (tok.child, errors));
            break;
                
        case NONEMPTY:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.Nonempty (translateNoun (tok.child, errors));
            break;

        case CASE_SENSITIVE:
        case NOT_CASE_SENSITIVE:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.CaseSensitivity 
                    ((tok.sym == CASE_SENSITIVE), translateNoun (tok.child, errors));
            break;
                
        case AS_TEXT:
        case AS_HTML:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else
                return new TC.View ((tok.sym == AS_TEXT) ? "text" : "html",
                                    translateNoun (tok.child, errors));
            break;

        case FROM:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else {
                TC.Node start = translateNoun (tok.child, errors);
                if (tok.child.sibling == null)
                    recordError (tok, "FROM be followed by TO", errors);
                else {
                    tok = tok.child.sibling;
                    TC.Node end = translateNoun (tok.child, errors);
                    tok = tok.sibling;
                    TC.Node expr = new TC.FromTo (start, end);

                    // Translate remaining children as constraints
                    // on entire expression.
                    return translateConstraints (expr, tok, errors);
                }
            }
            break;

        case EITHER:
            {
                TCToken firstEither = tok;
                TC.Node expr = null;
                TC.Node node = null;
                while (tok != null && (tok == firstEither || tok.sym == OR)) {
                    if (tok.child == null)
                        recordError (tok, "must have right-hand side", errors);
                    else {
                        node = translateNoun (tok.child, errors);
                        expr = (expr != null)
                            ? new TC.Or (expr, node)
                            : node;
                    }

                    tok = (tok == firstEither) 
                        ? tok.child.sibling
                        : tok.sibling;
                }

                // Translate remaining children as constraints
                // on entire expression.
                return translateConstraints (expr, tok, errors);
            }
                
        case PREFIX:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else if (tok.child.sym != ID)
                recordError (tok.child, "PREFIX must be followed by a name", errors);
            else
                return new TC.Prefix (tok.child.text,
                                      translateSequence (tok.child.sibling, errors));
            break;

        case NTH:
            if (tok.child == null)
                recordError (tok, "must have right-hand side", errors);
            else {
                // Search for context constraint (IN EACH, etc.)
                TC.Node context = null;
                TCToken contextToken = tok.child.child;
                if (contextToken == null)
                    context = new TC.Any ();
//                     recordError (tok, 
//                                  tok.text + " must be followed by IN, BEFORE, or AFTER:\n"
//                                  + "e.g., " + tok.text + " word in sentence", errors);
                else if (contextToken.sibling != null)
                    recordError (contextToken.sibling, 
                                 tok.text + " can have only one constraint", errors);
                else {
                    switch (contextToken.sym) {
                    case IN:
                        if (contextToken.child == null)
                            recordError (contextToken, "must have right-hand side", errors);
                        else
                                // FIX: IN is not quite right for nested sets
                            context = new TC.In 
                                (translateNoun (contextToken.child, errors));
                        break;
                    case BEFORE:
                        if (contextToken.child == null)
                            recordError (contextToken, "must have right-hand side", errors);
                        else
                            context = new TC.BeforeDelimited 
                                (translateNoun (contextToken.child, errors));
                        break;
                    case AFTER:
                        if (contextToken.child == null)
                            recordError (contextToken, "must have right-hand side", errors);
                        else
                            context = new TC.AfterDelimited
                                (translateNoun (contextToken.child, errors));
                        break;
                    default:
                        recordError (tok, 
                                     tok.text + " must be followed by IN, BEFORE, or AFTER:\n"
                                     + "e.g., " + tok.text + " Word in Sentence", errors);
                        break;
                    }
                }

                // Parse n
                int n;
                if (tok.text.equalsIgnoreCase ("last"))
                    n = -1;
                else {
                    n = Str.parseNumber (tok.text).intValue ();
                    if (Str.endsWithIgnoreCase (tok.text, "last"))
                        n = -n;
                }

                // Create NTH node.  If there was a context constraint,
                // temporarily clear it to keep translateNoun from translating it.
                tok.child.child = null;
                TC.Node node = new TC.Nth 
                    (n,
                     translateNoun (tok.child, errors),
                     context != null
                     ? context
                     : new TC.Any ());
                tok.child.child = contextToken;

                // Translate all other children as constraints
                // applied to the NTH node. 
                return translateConstraints (node, tok.child.sibling, errors);
            }

        case CLOSED_LEFT:
        case CLOSED_RIGHT:
            return (tok.child != null)
                ? translateConstraints (translateNoun (tok.child, errors),
                                        tok.child.sibling, errors)
                : new TC.Any ();

        case IS: // handled by ID
            recordError (tok, "IS should only be used in a name definition, e.g. name IS pattern", errors);
            break;

        case OR: // handled by EITHER
            recordError (tok, 
                         "OR should only be used after EITHER,\n"
                         + "e.g. EITHER \"A\" OR \"B\"", errors);
            break;

        case WITH: // handled by BALANCES
            recordError (tok, 
                         "WITH should only be used after BALANCES, e.g. BALANCES \"(\" with \")\"", errors);
            break;

        case BEFORE: // handled by NTH
        case AFTER: // handled by NTH
            recordError (tok, 
                         (tok.sym == BEFORE ? "BEFORE" : "AFTER")
                         + " should only be used after 1st,\n"
                         + "e.g. last Word " + tok.sym + " \".\"\n"
                         + "Do you mean JUST BEFORE or ANYWHERE BEFORE?", 
                         errors);
            break;

        default:
            recordError (tok, "syntax error", errors);
            break;
        }

        return new TC.Nothing (); // by default
    }

    static boolean isNoun (TCToken tok) {
        switch (tok.sym) {
        case NOTHING:
        case LIT:
        case REGEXP:
        case ID:
        case FLATTEN:
        case MELT:
        case NTH:
            return true;
        case EITHER:
        case CASE_SENSITIVE:
        case NOT_CASE_SENSITIVE:
        case AS_TEXT:
        case AS_HTML:
            return isNoun (tok.child);
        case PREFIX:
            return isNoun (tok.child.sibling);
        default:
            return false;
        }
    }
        
    static void recordError (TCToken tok, String message, TCParseException errors) {
        if ((debugFlags & STACK_TRACES) != 0)
            debug.printStackTrace ();
        errors.addError (tok, message);
    }


    //
    // Test main
    //
    public static void main (String[] argv) throws Exception {
        debugFlags = STACK_TRACES;
        
        for (int i = 0; i < argv.length; ++i)
            if (argv[i].equals ("-nostacktrace"))
                debugFlags &= ~STACK_TRACES;
            else if (argv[i].equals ("-reconstruct"))
                debugFlags |= RECONSTRUCT_STACK;

        TCParseException errors = new TCParseException ();
        
        Reader reader = new InputStreamReader (System.in);
        TCToken syntax = parseSyntax (reader, errors);
        if (syntax != null) {
            syntax.print ();
            System.err.println ();
        }
        
        TC.Node root = translateSequence (syntax, errors);
        if (!errors.isEmpty ()) {
            errors.printErrors (new PrintWriter (System.err));
            return;
        }

        TC pattern = new TC (root);
        pattern.print ();
    }


}
