/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tcl;

import java.io.*;
import java.util.*;
import lapisx.util.Str;

public abstract class Optimizer {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    /**
      * Optimize a Tcl script by commenting out any URL command
      * that is immediately followed by another URL command
      */
    public static String optimizeDeadURLs (String script) {
        StreamTokenizer in = new StreamTokenizer (new StringReader (script));
        in.resetSyntax ();
        in.wordChars (Character.MIN_VALUE, Character.MAX_VALUE);
        in.whitespaceChars ('\r', '\r');
        in.whitespaceChars ('\n', '\n');

        StringWriter sout = new StringWriter ();
        PrintWriter out = new PrintWriter (sout);

        String last = null;  // last command, including surrounding whitespace
        boolean lastWasURL = false; // last command was a URL

        try {
            while (in.nextToken () != StreamTokenizer.TT_EOF) {
                if (in.ttype == StreamTokenizer.TT_WORD) {
                    String cmd = trim (in.sval);

                    if (cmd.length () == 0) {
                        if (last != null)
                            last = last + in.sval + "\n";
                        else
                            last = in.sval + "\n";
                    } else {
                        if (isURL (cmd)) {
                            if (lastWasURL) 
                                // comment out last command
                                last = "# " + last;
                            lastWasURL = true;
                        } else {
                            lastWasURL = false;
                        }

                        // print previous command
                        if (last != null)
                            out.print (last);
                        last = in.sval + "\n";
                    }
                }
            }

            if (last != null)
                out.print (last);

            out.close ();
            return sout.toString ();
        } catch (IOException e) {
            debug.report (e);
            return script;
        }
    }


    /**
      * Optimize a Tcl script by converting a form submission
      * (URL, enter, enter, ..., submit) into a webquery command.
      */
    public static String optimizeFormSubmits (String script) {
        StreamTokenizer in = new StreamTokenizer (new StringReader (script));
        in.resetSyntax ();
        in.wordChars (Character.MIN_VALUE, Character.MAX_VALUE);
        in.whitespaceChars ('\r', '\r');
        in.whitespaceChars ('\n', '\n');

        StringWriter sout = new StringWriter ();
        PrintWriter out = new PrintWriter (sout);

        String formSubmission = null;  
        // current form submission, or null if not currently
        // reading a form submission

        String hrefForm = null;
        // URL of form
        Vector params = new Vector ();
        // parameters of Enter commands: name, value, name, value, ...

        try {
            while (in.nextToken () != StreamTokenizer.TT_EOF) {
                if (in.ttype == StreamTokenizer.TT_WORD) {
                    String cmd = trim (in.sval);
                    String line = in.sval + "\n";

                    if (formSubmission == null) {
                        if (isURL (cmd)) {
                            formSubmission = line;
                            hrefForm = cmd;
                            params.removeAllElements ();
                        }
                        else
                            out.print (line);
                    } else {
                        if (isURL (cmd)) {
                            // Jumped to a new URL before form submission 
                            // was complete. Print out partial form
                            // submission and reset for new form.
                            out.print (formSubmission);
                            formSubmission = line;
                            hrefForm = cmd;
                            params.removeAllElements ();
                        } else if (isCommand (cmd, "enter")) {
                            if (getEnterParameters (cmd, params)) 
                                // Continuing form submission
                                formSubmission += in.sval + "\n";
                            else {
                                // Can't handle this form submission
                                System.err.println ("Can't handle enter command: " + cmd);
                                out.print (formSubmission);
                                out.print (line);
                                formSubmission = null;
                            }
                        } else if (isCommand (cmd, "submit")) {
                                out.print (formSubmission);
                                out.print (line);
                                formSubmission = null;
                            
//                             try {
//                                 FormURL formURL = Submit.submit 
//                                     (new Arguments ().getFactory ().make (hrefForm),
                                     

//                                     new lapis.tools.Submit ()
//                                     .invoke (new Arguments ()
//                                              .add ("href", hrefForm));
                            
//                             out.print ("webquery");
//                             out.print (" -" + formURL.getMethod ().toLowerCase ());
//                             out.print (" " + formURL.getAction ());
                            
//                             Hashtable defaults = formURL.getQueryHash ();
                            
//                             // Print enter'ed params first
//                             for (int i = 0, n = params.size (); 
//                                  i < n; ++i) {
//                                 Object obj = params.elementAt (i);
//                                 out.print (" " + obj);
//                                 if (i % 2 == 0)
//                                     // it's a parameter name; delete it from
//                                     // list of defaults
//                                     defaults.remove (obj);
//                             }

//                             // Now print defaults 
//                             for (Enumeration e = defaults.keys (); 
//                                  e.hasMoreElements (); ) {
//                                 Object key = e.nextElement ();
//                                 out.print (" " + key + " " + defaults.get(key));
//                             }

//                             // finish command and return to start state
//                             out.println ();
//                             formSubmission = null;
//                             } catch (Exception e) {
//                                 // forget trying to optimize this submission
//                                 e.printStackTrace ();
//                                 out.print (formSubmission);
//                                 formSubmission = null;
//                             }
                        } else {
                            // Unfinished form submission, can't be optimized
                            out.print (formSubmission);
                            formSubmission = null;
                        }
                    }
                }
            }

            if (formSubmission != null)
                out.print (formSubmission);

            out.close ();
            return sout.toString ();
        } catch (IOException e) {
            debug.report (e);
            return script;
        }
    }



    /**
      * Deletes trailing comment (#...) and leading and trailing whitespace
      * from a string.
      */
    static String trim (String line) {
        int comment = line.lastIndexOf ('#');
        if (comment >= 0)
            line = line.substring (0, comment);
        return line.trim ();
    }

    /**
      * Tests whether command is a URL.
      */
    static boolean isURL (String cmd) {
        return cmd.startsWith ("http:")
            || cmd.startsWith ("ftp:")
            || cmd.startsWith ("file:")
            || cmd.startsWith ("systemresource:");
    }

    static boolean isCommand (String cmd, String cmdName) {
        return cmd.startsWith (cmdName)
            && (cmd.length() == cmdName.length() 
                || Character.isWhitespace (cmd.charAt (cmdName.length())));
    }

    static boolean getEnterParameters (String cmd, Vector params) {
        String args = Str.after (cmd, " ").trim ();

        String fieldArg = Str.before (args, " ").trim ();
        if (!fieldArg.toLowerCase ().startsWith ("Field.") 
            || fieldArg.indexOf (" ") != -1)
            // this enter command uses a pattern more complicated than
            // Field.foo -- we can't handle it
            return false;
        String fieldName = fieldArg.substring (6);
        
        String valueArg = Str.after (args, " ").trim ();

        params.addElement (fieldName);
        params.addElement (valueArg);
        return true;
    }

}

