/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tcl;

import lapis.*;
import tcl.lang.*;
import lapisx.util.*;
import lapisx.net.URLUtil;
import java.io.IOException;
import java.io.File;

public class TclInterpreter implements Interpreter {
    public static Debug debug = Debug.QUIET;

    Interp interp;

    public TclInterpreter () {
        this.interp = new Interp ();

        // create a fresh factory
        setVar (Interpreter.FACTORY_VAR, 
                new BasicDocument.Factory ());

        try {
            setVar (LAPIS_DIR_VAR, Main.getLapisDir ().toString ());

            interp.createCommand ("lapis::bind", new BindCommand ());

            // create document history
            int pageHistorySize = Integer.parseInt 
                (Main.getConfig ().getProperty ("shell.page-history-size"));
            History history = new History (pageHistorySize);
            interp.setVar (HISTORY_VAR, 
                           ReflectObject.newInstance (interp, 
                                                      history.getClass (), 
                                                      history), 
                           TCL.GLOBAL_ONLY);
        } catch (TclException e) {
            debug.report (e);
        }

        runInitFile ("tcl.system-init", Main.getLapisDir ());
        runInitFile ("tcl.personal-init", Config.getHomeDirectory ());
    }

    void runInitFile (String property, File dir) {
        String filename = Main.getConfig ().getProperty (property);
        if (filename == null
            || filename.trim ().length () == 0)
            return;

        try {
            filename = URLUtil.HrefToFile (filename, dir).toString ();
        } catch (IOException e) {
            debug.report (e);
            return;
        }

        debug.println (property + "=" + filename);

        try {
            interp.evalFile (filename);
        } catch (TclException e) {
            // FIX: print filename and line number
            switch (e.getCompletionCode ()) {
            case TCL.RETURN:
                break;
            case TCL.ERROR:
                System.err.println (e.getMessage ());
                System.err.println (getVar ("errorInfo").toString ());
                break;
                
            case TCL.BREAK:
            case TCL.CONTINUE:
                System.err.println ("TclException: break or continue without enclosing loop");
                break;
                
            default:
                System.err.println ("TclException: unknown error");
                break;
            }
        }
    }

    public TclInterpreter (Interp interp) {
        this.interp = interp;
    }

    static class BindCommand implements Command {
        public void cmdProc (Interp interp, TclObject argv[]) throws TclException {
            if (argv.length != 3)
                throw new TclException (interp, "lapis::bind <procedure name> <Java class>");

            String proc = argv[1].toString ();
            String className = argv[2].toString ();
            TclCommand cmd = new TclCommand (className);
            interp.createCommand (proc, cmd);
        }
    }
    
    /**
     * Evaluate a script.
     */
    public synchronized Object eval (String code) throws Exception {
        try {
            interp.eval (code);
        } catch (TclException e) {
            if (e.getCompletionCode () != TCL.RETURN) {
                throw new WrappedTclException (interp, e, 
                                               getVar ("errorInfo").toString ());
            }
        }

        return LapisUtil.tclObjectToJavaObject (interp,
                                                interp.getResult ());
    }

    /**
     * Evaluate a script.
     */
    public synchronized Object evalFile (String filename) throws Exception {
        try {
            interp.evalFile (filename);
        } catch (TclException e) {
            if (e.getCompletionCode () != TCL.RETURN)
                throw new TclException (interp, 
                                        getVar ("errorInfo").toString ());
            // otherwise everything's cool
        }

        return LapisUtil.tclObjectToJavaObject (interp,
                                                interp.getResult ());
    }

    public synchronized Object getVar (String name) {
        // handle some special-case variables first
        // current directory is stored internally in the interpreter
        if (CURRENT_DIRECTORY_VAR.equals (name))
            return LapisUtil.getWorkingDir (interp);
        else if (DOCUMENT_VAR.equals (name))
            return LapisUtil.getCurrentDocument (interp);
        else if (ARGUMENTS_VAR.equals (name)) {
            // arguments need to be converted to a string array
            try {
                TclObject list = interp.getVar (name, TCL.GLOBAL_ONLY);
                if (list == null)
                    return null;

                list.preserve ();
                try {
                    TclObject[] elements = TclList.getElements (interp, list);
                    String[] argv = new String[elements.length];
                    for (int i = 0; i < elements.length; ++i)
                        argv[i] = elements[i].toString ();
                    return argv;
                } finally {
                    list.release ();
                }
            } catch (TclException e) {
                return null;
            }
        } else {
            try {
                return LapisUtil.tclObjectToJavaObject (interp,
                                                        interp.getVar (name, TCL.GLOBAL_ONLY));
            } catch (TclException e) {
                return null;
            }
        }
    }

    public synchronized void setVar (String name, Object obj) {
        try {
            // handle some special-case variables first
            if (CURRENT_DIRECTORY_VAR.equals (name))
                // current directory is stored internally in the interpreter
                LapisUtil.setWorkingDir (interp, (java.io.File) obj);
            else if (DOCUMENT_VAR.equals (name))
                // current document is a local variable, not a global
                 LapisUtil.setCurrentDocument (interp, obj);
            else if (ARGUMENTS_VAR.equals (name)) {
                // arguments need to be converted to a list
                String[] args = (String[]) obj;
		if (args == null)
		    args = new String[0];
                TclObject argv = TclList.newInstance();
                argv.preserve();
                try {
                    int argc = args.length;
		    debug.println ("argc=" + argc);
                    for (int i = 0; i < args.length; ++i) {
			debug.println ("arg[" + i + "]=" + args[i]);
                        TclList.append(interp, argv, TclString.newInstance (args[i]));
		    }
                    interp.setVar("argv", argv, TCL.GLOBAL_ONLY);
                    interp.setVar("argc", TclInteger.newInstance (argc), 
				  TCL.GLOBAL_ONLY);
                } finally {
                    argv.release();
                }
                return;
            } else {
                // convert obj to appropriate Tcl type
                TclObject tobj;
                if (obj instanceof String)
                    tobj = TclString.newInstance ((String)obj);
                else if (obj != null)
                    tobj = ReflectObject.newInstance (interp, 
                                                      obj.getClass (), 
                                                      obj);
                else
                    tobj = ReflectObject.newInstance (interp, 
                                                      Object.class,
                                                      obj);
                
                // set it to a global variable
                interp.setVar (name, tobj, TCL.GLOBAL_ONLY);
            }
        } catch (TclException e) {
            debug.report (e);
        }
    }

    public synchronized void bindTool (String name, Tool tool) {
        interp.createCommand (name, new lapis.tcl.TclCommand (tool));
    }

    /** 
     * Quote a string to protect it from Tcl interpreter (but only if 
     * it contains Tcl metacharacters).
     */

    static final String metachars = " \r\n\t\f[]$;\"\\{}";
    static final String escapes =   " rntf[]$;\"\\{}";

    public String quote (String s) {
        // check if there are any metacharacters in the string
        int i = Str.indexOfAnyChar (s, metachars);
        if (i == -1)
            return s;

        // s contains metacharacters (first is at s[i]).  
        // determine whether we can quote it simply with {...}, or whether
        // we have to resort to escaping every metacharacter with \.
        int nestedBraces = 0;
        boolean mustEscape = false;

    loop:
        for (int n = s.length (); i < n; ++i) {
            char c = s.charAt (i);
            switch (c) {
            case '{':
                ++nestedBraces;
                break;
                
            case '}':
                --nestedBraces;
                if (nestedBraces < 0) {
                    // unmatched braces -- must escape
                    mustEscape = true;
                    break loop;
                }
                break;

            case '\\':
                if (i+1 == n || s.charAt(i+1) == '\n') {
                    mustEscape = true;
                    break loop;
                }
            }
        }

        if (nestedBraces != 0)
            // unmatched braces -- must escape
            mustEscape = true;

        if (mustEscape)
            return Str.escape (s, metachars, '\\', escapes);
        else
            return "{" + s + "}";
    }

}
