/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
import lapis.tc.TC;
import lapisx.util.Str;
import java.util.Vector;
import java.text.NumberFormat;

public class Calc implements Tool {
    /**
     * Invoke tool.
     */
    public Document invoke (Arguments args) throws Exception {
        TC tcRecords = null;
        Vector functions = new Vector ();
        boolean showCaptions = false;
        
        args.setUsage (
        "Usage:\n"
        +"   calc <pattern> [document]\n"
        +"       Returns a document that computes various functions of\n"
        +"       the regions matched by <pattern>.  If no function options\n"
        +"       are given, then all functions are computed and printed.\n"
        +"       NON-NUMERIC REGIONS ARE IGNORED.\n"
        +"\n"
        +"Options:\n"
        +"    -count       count\n"
        +"    -sum         sum\n"
        +"    -avg, -mean  average\n"
        +"    -min         minimum value\n"
        +"    -max         maximum value\n"
        +"    -stddev      standard deviation\n"
        +"\n"
        +"    -help        display this message\n"
        );

        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("query".equals (name)
                || (name == null && tcRecords == null))
                tcRecords = args.nextPattern ();
            else if ("count".equals (name))
                functions.addElement (new Count ());
            else if ("sum".equals (name))
                functions.addElement (new Sum ());
            else if ("average".equals (name)
                     || "avg".equals (name)
                     || "mean".equals (name))
                functions.addElement (new Average ());
            else if ("min".equals (name))
                functions.addElement (new Min ());
            else if ("max".equals (name))
                functions.addElement (new Max ());
            else if ("stddev".equals (name))
                functions.addElement (new StdDev ());
            else
                args.consume (name);
        }

        if (tcRecords == null)
            // error if no pattern
            args.nextValue ();

        if (functions.isEmpty ()) {
            functions.addElement (new Count ());
            functions.addElement (new Sum ());
            functions.addElement (new Average ());
            functions.addElement (new Min ());
            functions.addElement (new Max ());
            functions.addElement (new StdDev ());
            showCaptions = true;
        }

        Document doc = args.getInput ();
        DocumentWriter output = args.getOutput ();
        RegionSet records = tcRecords.matchWithWarnings (doc);

        records = records.flatten ();

        RegionEnumeration e = records.regions ();
        for (Region r = e.firstFast (); 
             r != null; 
             r = e.nextFast ()) {
            try {
                double num = Str.parseNumber (doc.getText (r)).doubleValue ();
                for (int i = 0, n = functions.size (); i < n; ++i) {
                    Function f = (Function) functions.elementAt (i);
                    f.accumulate (num);
                }
            } catch (NumberFormatException ex) {
                // ignore this region
            }
        }

        NumberFormat fmt = NumberFormat.getInstance ();
        fmt.setGroupingUsed (false);

        for (int i = 0, n = functions.size (); i < n; ++i) {
            Function f = (Function) functions.elementAt (i);
            String val;
            try {
                val = fmt.format (f.getAnswer ());
            } catch (NoAnswerException ex) {
                val = "NA";
            }
            if (showCaptions) {
                String name = 
                    Str.after (f.getClass ().getName (), "$")
                    .toLowerCase ();
                output.println (name + " = " + val);
            } else {
                if (i > 0)
                    output.print (" ");
                output.print (val);
            }
        }

        return output;
    }

    public static interface Function {
        public abstract void accumulate (double num);
        public abstract double getAnswer () throws NoAnswerException;
    }

    public static class NoAnswerException extends Exception {
    }
    
    public static class Count implements Function {
        protected int n = 0;
        public void accumulate (double num) {
            ++n;
        }
        public double getAnswer () throws NoAnswerException {
            return n;
        }
    }
        
    public static class Sum implements Function {
        protected double sum = 0.0;
        public void accumulate (double num) {
            sum += num;
        }
        public double getAnswer () throws NoAnswerException {
            return sum;
        }
    }
    
    public static class Average extends Sum {
        protected int n = 0;
        public void accumulate (double num) {
            super.accumulate (num);
            ++n;
        }
        public double getAnswer () throws NoAnswerException {
            if (n == 0)
                throw new NoAnswerException ();
            return sum / n;
        }
    }

    public static class Min implements Function {
        protected boolean noAnswer = true;
        protected double min = Double.MAX_VALUE;
        public void accumulate (double num) {
            min = Math.min (min, num);
            noAnswer = false;
        }
        public double getAnswer () throws NoAnswerException {
            if (noAnswer)
                throw new NoAnswerException ();
            return min;
        }
    }

    public static class Max implements Function {
        protected boolean noAnswer = true;
        protected double max = Double.MIN_VALUE;
        public void accumulate (double num) {
            max = Math.max (max, num);
            noAnswer = false;
        }
        public double getAnswer () throws NoAnswerException {
            if (noAnswer)
                throw new NoAnswerException ();
            return max;
        }
    }

    public static class StdDev extends Average {
        protected double squares = 0.0;
        public void accumulate (double num) {
            super.accumulate (num);
            squares += num * num;
        }
        public double getAnswer () throws NoAnswerException {
            double avg = super.getAnswer ();
            return Math.sqrt (squares/n - avg*avg);
        }
    }
}
