/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
import lapis.tc.*;
import java.net.*;
import lapisx.net.FormURL;

public class Click implements Tool {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    public Document invoke (Arguments args) throws Exception {
        TC control = null;
        boolean haveInput = false;

        args.setUsage (
        "Usage:\n"
        +"  click <pattern>    clicks on hyperlink/button matching <pattern>\n"
        +"\n"
        +"Options:\n"
        +"    -help                  Display this message\n"
        );
        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("control".equals (name)
                || (name == null && control == null))
                control = args.nextPattern ();
            else if (name == null) {
                if (haveInput)
                    args.usage ();
                haveInput = true;
                args.consume (name);
            }
            else
                args.consume (name);
        }

        if (control == null)
            args.nextValue ();
    
        Document doc = args.getInput ();

        TC tcLink = matchLink (control);
        TC tcButton = matchButton (control);

        RegionSet rLinks = tcLink.matchWithWarnings (doc);
        debug.println (rLinks.toStringWithContent ());
        RegionSet rButtons = tcButton.matchWithWarnings (doc);
        debug.println (rButtons.toStringWithContent ());

        int totalMatches = rLinks.getSize () + rButtons.getSize ();
        if (totalMatches == 0)
            throw new Arguments.UsageException ("No matches to pattern -- nothing to click on");
        else if (totalMatches > 1)
            throw new Arguments.UsageException ("Multiple matches to pattern -- don't know which to click on");

        else if (rLinks.getSize () > 0) {
            URL url = clickHyperlink (doc, rLinks.regions ().first ());
            return args.getFactory ().make (url);
        } else {
            FormURL formURL = Submit.submit (doc, null, rButtons.regions ().first ());
            if (formURL != null)
                return args.getFactory ().make (formURL);

            throw new Arguments.UsageException ("Can't click -- not a proper form button");
        }
    }

    TC matchLink (TC pat) {
        TC.Node n = pat.getRoot ();
        n = new TC.And (new TC.Id (".HTML.Element.[a]"), new TC.Equals (n));
        n = new TC.And (new TC.Id (".HTML.Tag.StartTag.<a>"), new TC.Starting (n));
        n = new TC.And (new TC.Id (".HTML.Attribute.@href"), new TC.In (n));
        n = new TC.And (new TC.Id (".HTML.Attribute.Value"), new TC.In (n));
        return new TC (n);
    }

    TC matchButton (TC pat) {
        TC.Node n = pat.getRoot ();
        n = new TC.And (new TC.Id (".Layout.Form.Control"), n);
        return new TC (n);
    }

    public static URL clickHyperlink (Document doc, Region link) throws Exception {
        String href = trimQuotes (doc.getText (link));
        return new URL (getBase (doc), href);
    }

    public static URL getBase (Document doc) {
        URL url = (URL) doc.getProperty (doc.BaseProperty);
        if (url != null)
            return url;

        return (URL) doc.getProperty (doc.URLProperty);
    }
    
    static String trimQuotes (String s) {
        if ((s.startsWith ("\"") && s.endsWith ("\""))
            || (s.startsWith ("'") && s.endsWith ("'")))
            return s.substring (1, s.length()-1);
        else
            return s;
    }
}
