/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
import lapis.tc.*;
import lapisx.util.Str;
//import java.io.*;
//import java.util.Enumeration;
//import java.util.Vector;

public class Extract implements Tool {
    static final String TEXT = "text/plain";
    static final String HTML = "text/html";
    static final String TCL_LIST = "application/x-tcl-list";

    /**
     * Invoke tool.
     */
    public Document invoke (Arguments args) throws Exception {
        TC tcRecords = null;
        String contentType = null;
        String startsWith = null;
        String endsWith = null;
        String separatedBy = null;
        Template template = null;
        boolean useDefaultTerminator;

        args.setUsage (
          "Usage:\n"
        + "   extract <pattern>      extracts regions matching <pattern>\n"
        + "\n"
        + "Options:\n"
        + "    -as [text|html|tcl]     Output type\n"
        + "                               (default: same type as input document)\n"
        + "    -startswith <string>    String to precede each extracted record\n"
        + "                               (default: none)\n"
        + "    -endswith <string>      String to follow each extracted record\n"
        + "                               (default: newline for text, <BR> for HTML, none for Tcl)\n"
        + "    -separatedby <string>   String to separate each pair of extracted records\n"
        + "                               (default: none for text/HTML, space for Tcl)\n"
        + "    -help                   Display this message\n"
          );

        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("query".equals (name)
                || (name == null && tcRecords == null)) {
                tcRecords = args.nextPattern ();
            }
            else if ("as".equals (name)) {
                String type = args.nextString ();
                if ("text".equalsIgnoreCase (type)
                    || TEXT.equalsIgnoreCase (type))
                    contentType = TEXT;
                else if ("html".equalsIgnoreCase (type)
                    || HTML.equalsIgnoreCase (type))
                    contentType = HTML;
                else if ("same".equalsIgnoreCase (type))
                    contentType = null;
                else if ("tcl".equalsIgnoreCase (type))
                    contentType = TCL_LIST;
                else
                    args.usage ();
            }
            else if ("startswith".equals (name))
                startsWith = Str.unescape (args.nextString (),
                                           '\\', 
                                           "btnfr\\", 
                                           "\b\t\n\f\r\\", 
                                           false);
            else if ("endswith".equals (name))
                endsWith = Str.unescape (args.nextString (),
                                           '\\', 
                                           "btnfr\\", 
                                           "\b\t\n\f\r\\", 
                                           false);
            else if ("separatedby".equals (name))
                separatedBy = Str.unescape (args.nextString (),
                                           '\\', 
                                           "btnfr\\", 
                                           "\b\t\n\f\r\\", 
                                           false);
            else if ("format".equals (name)) {
                Object obj = args.nextValue ();

                if (obj instanceof Template)
                    template = (Template)obj;
                else if (obj instanceof String) {
                    String strTemplate = (String)obj;
                    try {
                        template = Template.makeFromString (strTemplate);
                    } catch (TCParseException e) {
                        e.setStringContext (strTemplate);
                        throw e;
                    }
                }
                else
                    args.usage ();
            }
            else 
                args.consume (name);
        }

        // Error if no pattern
        if (tcRecords == null)
            args.usage ();

        useDefaultTerminator = (startsWith == null && endsWith == null && separatedBy == null);
        if (startsWith == null)
            startsWith = "";
        if (endsWith == null)
            endsWith = "";
        if (separatedBy == null)
            separatedBy = "";

        Interpreter interp = args.getInterpreter ();
        Document doc = args.getInput ();
        DocumentWriter output = args.getOutput ();
        boolean started = false;

        boolean tcl = TCL_LIST.equals (contentType);
        if (tcl) {
            useDefaultTerminator = false;
            startsWith = "";
            endsWith = "";
            separatedBy = " ";
        }

        String docContentType = (String) doc.getProperty (Document.MIMEProperty);

        if (contentType == null)
            contentType = docContentType;

        if (useDefaultTerminator && HTML.equals (contentType))
            endsWith = "<br>\n";

        BasicDocument.copyVitalProperties (output, doc);

        if (contentType != null)
            output.putProperty (Document.MIMEProperty, contentType);

        RegionSet records = tcRecords.matchWithWarnings (doc);
        records = records.flatten ();

        if (template != null) {
            template.bind (records, doc);
            template.print (records, doc, output);
        } else {
            boolean html2text = (TEXT.equals (contentType)
                                 && HTML.equals (docContentType));
            boolean text2html = (HTML.equals (contentType)
                                 && TEXT.equals (docContentType));

			DocumentView view = doc.getDefaultView();
            if (html2text) {
                view = doc.getCookedView();
                records = view.getVersion().convert(records.getSingleDocRegionSet(doc));
            }
                
            records = records.flatten ();

            RegionEnumeration e = records.regions (); 
            for (Region r = e.firstFast (); 
                 r != null; 
                 r = e.nextFast ()) {
                if (!started) 
                    started = true;
                else
                    output.print (separatedBy);
                    
                output.print (startsWith);
                    
                //String rec = doc.getText (r);
                String rec = view.getText(r);
                if (html2text)
                    rec = rec.trim ();
                else if (text2html) {
                    rec = Str.replace (rec, "&", "&amp;");
                    rec = Str.replace (rec, "<", "&lt;");
                    rec = Str.replace (rec, ">", "&gt;");
                }

                if (tcl)
                    rec = interp.quote (rec);
                    
                output.print (rec);
                    
                if (useDefaultTerminator && !rec.endsWith ("\n"))
                    output.println ();

                output.print (endsWith);
            }
        }

        return output;
    }
}
