/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
import lapis.tc.*;
import lapisx.util.Str;
//import java.io.*;
//import java.util.*;

public class Highlight implements Tool {
    public static final lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    static final String TEXT = "text/plain";
    static final String HTML = "text/html";
    static final String TCL_LIST = "application/x-tcl-list";

    /**
     * Invoke tool.
     */
    public Document invoke (Arguments args) throws Exception {
        TC tcItems = null;
        String contentType = HTML;
        TC tcUnits = null; 

        args.setUsage (
          "Usage:\n"
        + "   highlight <pattern>     highlights regions matching <pattern>.\n"
        + "                           In text, highlights >>like this<<\n"
        + "                           In HTML, highlights in bold\n"
        + "\n"
        + "Options:\n"
        + "    -as [text|html|same]   Output type (default is html)\n"
        + "    -units <pattern>       Units to round to\n"
        + "    -help                  Display this message\n"
          );

        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("query".equals (name)
                || (name == null && tcItems == null)) {
                tcItems = args.nextPattern ();
            }
            else if ("as".equals (name)) {
                String type = args.nextString ();
                if ("text".equalsIgnoreCase (type)
                    || TEXT.equalsIgnoreCase (type))
                    contentType = TEXT;
                else if ("html".equalsIgnoreCase (type)
                    || HTML.equalsIgnoreCase (type))
                    contentType = HTML;
                else if ("same".equalsIgnoreCase (type))
                    contentType = null;
                else
                    args.usage ();
            }
            else if ("units".equals (name) || "unit".equals (name)) {
                tcUnits = args.nextPattern ();
            }
            else 
                args.consume (name);
        }

        // Error if no pattern
        if (tcItems == null)
            args.usage ();

        if (tcUnits == null)
            tcUnits = new TC (new TC.Id (".Layout.Line"));

        String endsWith, highlightStart, highlightEnd;
        if (HTML.equals (contentType)) {
            endsWith = "<br>\n";
            highlightStart = "<b>";
            highlightEnd = "</b>";
        } else {
            endsWith = null;
            highlightStart = ">>";
            highlightEnd = "<<";
        }

        Document doc = args.getInput ();
        DocumentWriter output = args.getOutput ();
        String docContentType = (String) doc.getProperty (doc.MIMEProperty);

        if (contentType == null)
            contentType = docContentType;

        BasicDocument.copyVitalProperties (output, doc);
        if (contentType != null)
            output.putProperty (doc.MIMEProperty, contentType);

        RegionSet items = tcItems.matchWithWarnings (doc);
        RegionSet units = tcUnits.matchWithWarnings (doc);

        boolean html2text = (TEXT.equals (contentType)
                             && HTML.equals (docContentType));
        boolean text2html = (HTML.equals (contentType)
                             && TEXT.equals (docContentType));

		DocumentView view = doc.getDefaultView ();            
        if (html2text) {
            view = doc.getCookedView();
            items = view.getVersion().convert (items.getSingleDocRegionSet(doc));
            units = view.getVersion().convert (units.getSingleDocRegionSet(doc));
        }

        items = items.flatten ();
        units = units.flatten ();

        Relation recordItems = new Relation ();
        MutableRegionSet records = new MutableRegionSet ();
        RegionEnumeration e = items.regions ();
        for (Region item = e.firstFast (); item != null; item = e.nextFast ()) {
            Region record = (Region) item.include (units);
            if (!recordItems.add (record, item))
                records.insert (record);
        }

        e = records.regions (); 
        for (Region record = e.firstFast (); record != null; record = e.nextFast ()) {
            String rec = view.getText(record);
            if (html2text)
                rec = rec.trim ();
            else if (text2html) {
                rec = Str.replace (rec, "&", "&amp;");
                rec = Str.replace (rec, "<", "&lt;");
                rec = Str.replace (rec, ">", "&gt;");
            }
            
            int off = 0; // number of characters inserted in rec
            RegionEnumeration re = recordItems.get (record).regions ();
            for (Region item = re.firstFast ();
                 item != null;
                 item = re.nextFast ()) {
                debug.println ("highlighting " + item + " in " + record);
                int st = item.getStart () - record.getStart () + off;
                int en = item.getEnd () - record.getStart () + off;
                debug.println ("that's [" + st + "," + en + "] relative");
                rec = rec.substring (0, st)
                    + highlightStart
                    + rec.substring (st, en)
                    + highlightEnd
                    + rec.substring (en);
                off += highlightStart.length () + highlightEnd.length ();
            }
    
            output.print (rec);
            if (!rec.endsWith ("\n"))
                output.println ();
            output.print (endsWith);
        }

        return output;
    }
}
