/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
import lapis.tc.TC;
//import java.io.*;
//import java.util.Enumeration;
import java.net.URL;
import lapisx.util.Str;

public class PageRelocator implements Tool {
    /**
     * Invoke the tool.
     */
    public Document invoke (Arguments args) throws Exception {
        boolean override = false;
        String baseHref = null;

        args.setUsage (
        "Usage:\n"
        +"    relocate         adds a BASE element to current document\n"
        +"\n"
        +"Options:\n"
        + "    -override       Override existing BASE element\n"
        + "    -base <URL>     URL for BASE element (default is document URL)\n"
        + "    -help           Display this message\n"
        );
        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("override".equals (name))
                override = true;
            else if ("base".equals (name))
                baseHref = args.nextString ();
            else
                args.consume (name);
        }

        Document doc = args.getInput ();
        DocumentWriter output = args.getOutput ();

        BasicDocument.copyVitalProperties (output, doc);
        output.putProperty (doc.URLProperty, doc.getProperty (doc.URLProperty));
        
        if ("text/html".equals ((String) doc.getProperty (Document.MIMEProperty)))
            insertBaseElement (doc, override, baseHref, output);
    
        return output;
    }

    public static String getBase (Document doc) {
        URL url = (URL) doc.getProperty (doc.BaseProperty);
        if (url != null)
            return url.toString ();

        url = (URL) doc.getProperty (doc.URLProperty);
        if (url != null)
            return url.toString ();

        return null;
    }

    public static void insertBaseElement (Document doc, 
                                          boolean overrideExistingBase,
                                          String href,
                                          DocumentWriter output) {
        if (href == null)
            href = getBase (doc);
        if (href == null) {
            output.print (doc.getText ());
            return;
        }

        href = Str.replace (href, "\"", "%22");
        
        Region base = findBaseElement (doc);
        if (base != null && !overrideExistingBase) {
            output.print (doc.getText ());
            return;
        }

        if (base == null)
            base = findBaseInsertPoint (doc);

        RegionSet before = base.before ().flatten ();
        RegionSet after = base.after ().flatten ();

        if (!before.isEmpty ())
            output.print (doc.getText ((Region) before));
        output.print ("<base href=\"" + href + "\">");
        if (!after.isEmpty ())
            output.print (doc.getText ((Region) after));
    }

    public static Region findBaseElement (Document doc) {
        return 
            TC.make ("<base>")
            .match (doc)
            .regions ()
            .first ();
    }

    public static Region findBaseInsertPoint (Document doc) {
        Region r;

        // try a sequence of possible places
        r = TC.make ("point just after <head>")
            .match (doc)
            .regions ()
            .first ();
        if (r != null)
            return r;

        r = TC.make ("point just before <body>")
            .match (doc)
            .regions ()
            .first ();
        if (r != null)
            return r;

        r = TC.make ("point just after <html>")
            .match (doc)
            .regions ()
            .first ();
        if (r != null)
            return r;

        // Failing all those, just stick <base> at the top of the document
        return new Region (0, 0, doc);
    }
}
