/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
//import java.io.*;
import java.util.Enumeration;

public class Property implements Tool {

    final static int NONE = 0;
    final static int LIST = 1;
    final static int GET = 2;
    final static int SET = 3;
    final static int DELETE = 4;

    public Document invoke (Arguments args) throws Exception {
        int op = NONE;
        String property = null;
        Object value = null;

        args.setUsage (
        "Usage:\n"
        +"    property -list              Get list of property names for current doc\n"
        +"    property [-get] <name>      Get a property from curr doc\n"
        +"    property -set <name> <val>  Set a property on curr doc\n"
        +"    property -delete <name>     Delete a property from curr doc\n"
        +"\n"
        +"Options:\n"
        +"    -help                  Display this message\n"
        );
        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("list".equals (name))
                op = LIST;
            else if ("get".equals (name) 
                     || (name == null && op == NONE)) {
                op = GET;
                property = args.nextString ();
            } else if ("set".equals (name)) {
                op = SET;
                property = args.nextString ();
            } else if ("delete".equals (name)) {
                op = DELETE;
                property = args.nextString ();
            } else if (name == null && value == null && op == SET) {
                value = args.nextValue ();
            } else
                args.consume (name);
        }

        switch (op) {
        case NONE:
            args.usage ();
            break;
        case GET:
            break;
        case SET:
            break;
        }

        Document doc = args.getInput ();

        switch (op) {
        case LIST:
            {
                DocumentWriter output = args.getOutput ();
                for (Enumeration e = doc.getProperties ().keys ();
                     e.hasMoreElements (); ) {
                    output.print (e.nextElement ());
                    output.print (" ");
                }
                return output;
            }
            
        case GET:
            {
                if (property == null)
                    args.usage ();
                Object obj = doc.getProperty (property);
                if (obj instanceof Document)
                    return (Document) obj;
                else {
                    DocumentWriter output = args.getOutput ();
                    output.print (obj);
                    return output;
                }
            }

        case SET:
            {
                if (property == null || value == null)
                    args.usage ();

                doc.putProperty (property, value);
                return null;
            }

        case DELETE:
            {
                if (property == null)
                    args.usage ();

                doc.putProperty (property, null);
                return null;
            }

        default:
            args.usage ();
            return null;
        }                
    }
}
