/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
import lapis.tc.*;

public class Replace implements Tool {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET; 

    /**
     * Invoke tool.
     */
    public Document invoke (Arguments args) throws Exception {
        TC tcRecords = null;
        Template template = null;

        args.setUsage (
        "Usage:\n"
        +"    replace <pattern> <template>   replaces regions matching <pattern>\n"
        +"                                   with the replacement <template>.\n" 
        +"                                   <template> may contain embedded\n"
        +"                                   patterns in {...}\n"
        +"\n"
        +"Options:\n"
        + "    -help           Display this message\n"
        );
        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("query".equals (name)
                || (name == null && tcRecords == null))
                tcRecords = args.nextPattern ();
            else if ("replace".equals (name)
                     || (name == null && template == null)) {
                Object obj = args.nextValue ();

                if (obj instanceof Template)
                    template = (Template)obj;
                else if (obj instanceof String) {
                    String strTemplate = (String)obj;
                    try {
                        template = Template.makeFromString (strTemplate);
                    } catch (TCParseException e) {
                        e.setStringContext (strTemplate);
                        throw e;
                    }
                }
                else
                    args.usage ();
            }
            else
                args.consume (name);
        }

        // Error if no pattern or template
        if (tcRecords == null || template == null)
            args.usage ();

        Document doc = args.getInput ();
        DocumentWriter output = args.getOutput ();

        BasicDocument.copyVitalProperties (output, doc);
        output.putProperty (doc.URLProperty, doc.getProperty (doc.URLProperty));
        
        RegionSet records = tcRecords.matchWithWarnings (doc).flatten ();
        template.bind (records, doc);

        if (debug.printEnabled ())
            debug.println ("replacing: " + records.highlight (doc));

        int end = 0;
        RegionEnumeration e = records.regions (); 
        for (Region r = e.first (); r != null; r = e.next ()) {
            output.print (doc.getText (new Region (end, r.getStart (), doc)));
            end = r.getEnd ();
            template.print (r, doc, output);

            //                 System.err.print (r);
            //                 System.err.println (lapisx.util.Str.abbreviate (r.extract (doc), 40));
        }
        output.print (doc.getText (new Region (end, Region.MAX_INDEX, doc)));

        return output;
    }
}
