/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
import java.io.*;
import java.net.*;
import lapisx.net.URLUtil;
//import java.util.Enumeration;

public class Save implements Tool {
    
    public Document invoke (Arguments args) throws Exception {
        String filename = null;    // output filename
        String backupExtension = 
            (String) args.getInterpreter ().getVar (Interpreter.BACKUP_EXTENSION_VAR);

        args.setUsage (
        "Usage:\n"
        +"    save              saves document to its original location\n"
        +"                        (only works if originally loaded from disk)\n"
        +"    save <filename>   saves document to <filename>\n"
        +"\n"
        +"Options:\n"
        +"    -backup <extension>    If <filename> already exists, it is backed up\n"
        +"                           in <filename><extension>.  If <extension>\n"
        +"                           is an empty string, backup is disabled.\n"
        +"    -help                  Display this message\n"
        );
        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("filename".equals (name)
                || (name == null && filename == null))
                filename = args.nextString ();
            else
                args.consume (name);
        }

        Document doc = args.getInput ();

        // figure out what filename to save 
        File file = null;

        // if we had a filename argument, use it
        if (filename != null) {
            try {
                file = URLUtil.HrefToFile (filename);
            } catch (MalformedURLException e) {
                // filename was an http: URL
                throw e;
            }
        }
    
        // failing that, see if document's URL property
        // is a local file: URL
        if (file == null) {
            URL url = (URL) doc.getProperty (Document.URLProperty);
            if (url != null)
                try {
                    file = URLUtil.URLToFile (url);
                } catch (MalformedURLException e) {
                    // don't throw this one, because this filename
                    // is just a default
                }
        }
        
        if (file == null)
            // no filename could be found
            args.usage ();

        // make sure filename is absolute
        if (!file.isAbsolute ()) {
            File curdir = (File) args.getInterpreter ()
                .getVar (Interpreter.CURRENT_DIRECTORY_VAR);
            file = new File (curdir, file.toString ());
        }

        // Fetch text to save *before* opening the file.
        // That way, if getText() deadlocks for some reason,
        // we won't have changed anything on disk.
        String text = doc.getText ();

        // make backup of original file, if any
        if (backupExtension != null
            && backupExtension.length () > 0
            && file.exists ()) {
            File backup = new File (file + backupExtension);
            if (backup.exists ())
                backup.delete ();
            file.renameTo (backup);
        }

        // save the file
        PrintWriter output = new PrintWriter (new FileWriter (file));
        output.print (text);
        output.close ();

        // update document properties
        doc.putProperty (Document.URLProperty, URLUtil.FileToURL (file));
        doc.putProperty (Document.SavedVersionProperty, doc.getVersion ());
        return doc;
    }
}
