/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
import lapis.tc.*;
import lapisx.sort.*;
import lapisx.util.Str;
//import java.io.*;
//import java.util.Enumeration;
import java.text.*;

public class Sort implements Tool {

    /**
     * Invoke tool.
     */
    public Document invoke (Arguments args) throws Exception {
        TC tcRecords = null;    // records to sort
        TC tcKey = null;        // last -by argument
        Order order = null;

        boolean haveInput = false;    // have we read the input filename yet?

        args.setUsage (
         "Usage:\n"
         +"   sort <record pattern>     sorts regions matching <record pattern>\n"
        +"\n"
        +"Options:\n"
        + "    -by <keypattern>  Pattern describing sort key in each record\n"
        + "                      (if omitted, use entire record as sort key)\n"
        + "    -order [reverse] <dictionary|numeric|unicode|random>\n"
        + "                    Sort order: dictionary is your local language,\n"
        + "                    numeric is numbers, unicode is case- and whitespace-sensitive Unicode order,\n"
        + "                    random is randomized.  Reverse reverses the specified sort order.\n"
        + "    -help           Display this message\n"
        );

        while (args.hasMoreElements ()) {
            String name = args.nextName ();
            if ("query".equals (name)
                || (name == null && tcRecords == null))
                tcRecords = args.nextPattern ();
            else if ("by".equals (name)) {
                if (tcKey != null) {
                    Order newOrder = new DictionaryOrder (tcKey);
                    order = (order != null) ? new ProductOrder (order, newOrder) : newOrder;
                }
                tcKey = args.nextPattern ();
            } else if ("order".equals (name)) {
                String orderName = args.nextString ();

                boolean reversed;
                if ("reverse".equals (orderName)) {
                    reversed = true;
                    orderName = args.nextString ();
                } else
                    reversed = false;

                Order newOrder = null;
                if ("dictionary".equalsIgnoreCase (orderName))
                    newOrder = new DictionaryOrder (tcKey);
                else if ("numeric".equalsIgnoreCase (orderName))
                    newOrder = new NumericOrder (tcKey);
                else if ("unicode".equalsIgnoreCase (orderName))
                    newOrder = new UnicodeOrder (tcKey);
                else if ("random".equalsIgnoreCase (orderName))
                    newOrder = new RandomOrder (tcKey);
                else
                    args.usage ();

                if (reversed)
                    newOrder = newOrder.reverse ();

                order = (order != null) ? new ProductOrder (order, newOrder) : newOrder;
                tcKey = null;
            }
            else if (name == null) {
                if (haveInput)
                    args.usage ();
                haveInput = true;
                args.consume (name);
            }
            else
                args.consume (name);
        }

        if (order == null)
            order = new DictionaryOrder (tcKey);
            
        // Error if no patterns
        if (tcRecords == null)
            args.usage ();

        Document doc = args.getInput ();
        DocumentWriter output = args.getOutput ();

        BasicDocument.copyVitalProperties (output, doc);
        output.putProperty (doc.URLProperty, doc.getProperty (doc.URLProperty));

        // find and flatten the records
        RegionSet records = tcRecords.matchWithWarnings (doc);
        records = records.flatten ();
        int nRecords = records.getSize ();
        
        order.setDocument (doc);
        
        // Create array with sort key from each record.
        // Also gather up the stuff between the records (the "lint").
        Region lint[] = new Region[nRecords+1];
        SortableRegion header = order.makeSortableRegion (new Region (0,0, doc));
        // header is just a sentinel -- it'll be thrown away
        
        SortableRegion last = header;
        RegionEnumeration e1 = records.regions ();
        int i = 0;
        for (Region record = e1.first (); 
             record != null; 
             record = e1.next ()) {

            lint[i++] = new Region (last.getRegion().getEnd (), 
                                    record.getStart (), doc);
                
            SortableRegion sortable = order.makeSortableRegion (record);
            last.setNext (sortable);
            last = sortable;
        }
            
        lint[i++] = new Region (last.getRegion ().getEnd (), 
        						RegionSet.MAX_INDEX,
        						doc);
            
        SortableRegion recordList = (SortableRegion)header.next ();

        // Sort the records
        SortableRegion sortedList = (SortableRegion)MergeSort.sort (recordList);

        // Generate the output, consisting of records separated by lint
        i = 0;
        for (SortableRegion r = sortedList; r != null; r = (SortableRegion)r.next()) {
            output.print (doc.getText (lint[i++]));
            output.print (doc.getText (r.getRegion ()));
        }
            
        // output final piece of lint
        output.print (doc.getText (lint[i++]));

        return output;
    }

    public static interface Order {
        Order reverse ();
        void setDocument (Document doc) throws TCParseException;
        SortableRegion makeSortableRegion (Region r);
    }

    public static abstract class SortableRegion implements SortableList {
        Region region;
        SortableList nextPtr;

        public SortableRegion (Region r) {
            region = r;
        }

        public Object get () { 
            return this; 
        }

        public Region getRegion () {
            return region;
        }

        public SortableList next () { 
            return nextPtr;
        }

        public void setNext (SortableList l) { 
            nextPtr = l; 
        }
    }

    public abstract static class KeyOrder implements Order {
        TC pattern;
        Document doc;
        RegionSet matches;

        public KeyOrder (TC pattern) {
            this.pattern = pattern;
        }

        public void setDocument (Document doc) throws TCParseException {
            this.doc = doc;
            if (pattern != null)
                matches = pattern.matchWithWarnings (doc);
        }
        
        public String getKey (Region r) {
            Region key = (pattern != null)
                ? matches.regions ((Region)r.in ()).first ()
                : r;
            return (key != null)
                ? doc.getText (key)
                : "";
        }             
    }

    public static class ProductOrder implements Order {
        Document doc;
        Order primaryOrder;
        Order secondaryOrder;

        public ProductOrder (Order primary, Order secondary) {
            this.primaryOrder = primary;
            this.secondaryOrder = secondary;
        }

        public Order reverse () {
            return new ProductOrder (primaryOrder.reverse (), secondaryOrder.reverse ());
        }

        public void setDocument (Document doc) throws TCParseException {
            this.doc = doc;
            primaryOrder.setDocument (doc);
            secondaryOrder.setDocument (doc);
        }
        
        public SortableRegion makeSortableRegion (Region r) {
            return new ProductSortableRegion (r);
        }

        class ProductSortableRegion extends SortableRegion {
            SortableRegion primaryKey;
            SortableRegion secondaryKey;
            
            public ProductSortableRegion (Region r) {
                super (r);
                primaryKey = primaryOrder.makeSortableRegion (r);
                secondaryKey = null; // generated on-demand in compareTo ()
            }
            
            public int compareTo (SortableList l) {
                ProductSortableRegion psr2 = (ProductSortableRegion)l;

                int primaryResult = primaryKey.compareTo (psr2.primaryKey);
                if (primaryResult != 0)
                    return primaryResult;

                if (secondaryKey == null)
                    secondaryKey = secondaryOrder.makeSortableRegion (getRegion ());
                if (psr2.secondaryKey == null)
                    psr2.secondaryKey = secondaryOrder.makeSortableRegion (psr2.getRegion ());

                return secondaryKey.compareTo (psr2.secondaryKey);
            }
        }
    }

    public static class DictionaryOrder extends KeyOrder {
        Collator collator;

        public DictionaryOrder (TC pattern) {
            super (pattern);
            collator = Collator.getInstance ();
            collator.setStrength (Collator.PRIMARY);
        }

        public Order reverse () {
            return new ReverseDictionaryOrder (pattern);
        }

        public SortableRegion makeSortableRegion (Region r) {
            return new DictionarySortableRegion (r, collator.getCollationKey (getKey (r)));
        }        

        static class DictionarySortableRegion extends SortableRegion {
            CollationKey key;
            
            public DictionarySortableRegion (Region r, CollationKey key) {
                super (r);
                this.key = key;
            }
            
            public int compareTo (SortableList l) {
                return key.compareTo (((DictionarySortableRegion)l).key);
            }
        }
    }

    public static class ReverseDictionaryOrder extends DictionaryOrder {
        public ReverseDictionaryOrder (TC pattern) {
            super (pattern);
        }

        public Order reverse () {
            return new DictionaryOrder (pattern);
        }

        public SortableRegion makeSortableRegion (Region r) {
            return new ReverseDictionarySortableRegion (r, collator.getCollationKey (getKey (r)));
        }        

        static class ReverseDictionarySortableRegion extends DictionarySortableRegion {
            public ReverseDictionarySortableRegion (Region r, CollationKey key) {
                super (r, key);
            }
            
            public int compareTo (SortableList l) {
                return -key.compareTo (((ReverseDictionarySortableRegion)l).key);
            }
        }
    }

    public static class UnicodeOrder extends KeyOrder {
        public UnicodeOrder (TC pattern) {
            super (pattern);
        }

        public Order reverse () {
            return new ReverseUnicodeOrder (pattern);
        }

        public SortableRegion makeSortableRegion (Region r) {
            return new UnicodeSortableRegion (r, getKey (r));
        }        

        static class UnicodeSortableRegion extends SortableRegion {
            String key;
            
            public UnicodeSortableRegion (Region r, String key) {
                super (r);
                this.key = key;
            }
            
            public int compareTo (SortableList l) {
                return key.compareTo (((UnicodeSortableRegion)l).key);
            }
        }
    }

    public static class ReverseUnicodeOrder extends UnicodeOrder {
        public ReverseUnicodeOrder (TC pattern) {
            super (pattern);
        }

        public Order reverse () {
            return new UnicodeOrder (pattern);
        }

        public SortableRegion makeSortableRegion (Region r) {
            return new ReverseUnicodeSortableRegion (r,getKey (r));
        }        

        static class ReverseUnicodeSortableRegion extends UnicodeSortableRegion {
            public ReverseUnicodeSortableRegion (Region r, String key) {
                super (r, key);
            }
            
            public int compareTo (SortableList l) {
                return -key.compareTo (((ReverseUnicodeSortableRegion)l).key);
            }
        }
    }

    public static class NumericOrder extends KeyOrder {
        public NumericOrder (TC pattern) {
            super (pattern);
        }

        public Order reverse () {
            return new ReverseNumericOrder (pattern);
        }

        public SortableRegion makeSortableRegion (Region r) {
            double n;
            try {
                n = Str.parseNumber (getKey (r)).doubleValue ();
            } catch (NumberFormatException e) {
                n = 0;
            }
            return new NumericSortableRegion (r, n);
        }        

        static class NumericSortableRegion extends SortableRegion {
            double n;
            
            public NumericSortableRegion (Region r, double n) {
                super (r);
                this.n = n;
            }
            
            public int compareTo (SortableList l) {
                double n2 = ((NumericSortableRegion)l).n;
                if (n < n2)
                    return -1;
                else if (n > n2)
                    return 1;
                else
                    return 0;
            }
        }
    }

    public static class ReverseNumericOrder extends NumericOrder {
        public ReverseNumericOrder (TC pattern) {
            super (pattern);
        }

        public Order reverse () {
            return new NumericOrder (pattern);
        }

        public SortableRegion makeSortableRegion (Region r) {
            double n;
            try {
                n = Str.parseNumber (getKey (r)).doubleValue ();
            } catch (NumberFormatException e) {
                n = 0;
            }
            return new ReverseNumericSortableRegion (r, n);
        }        

        static class ReverseNumericSortableRegion extends NumericSortableRegion {
            public ReverseNumericSortableRegion (Region r, double n) {
                super (r, n);
            }
            
            public int compareTo (SortableList l) {
                double n2 = ((NumericSortableRegion)l).n;
                if (n < n2)
                    return 1;
                else if (n > n2)
                    return -1;
                else
                    return 0;
            }
        }
    }

    public static class RandomOrder extends NumericOrder {
        public RandomOrder (TC pattern) {
            super (pattern);
        }

        public Order reverse () {
            return this;
        }

        public SortableRegion makeSortableRegion (Region r) {
            return new NumericSortableRegion (r, Math.random ());
        }        
    }
}
