/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
import lapis.tc.*;
import lapisx.util.Str;
import java.io.*;
import java.util.Vector;

public abstract class Template {
    public static Template makeFromString (String str) 
        throws TCParseException {
        return SequenceTemplate.makeFromString (str);
    }

    public void bind (RegionSet list, Document doc) throws TCParseException {
    }

    public abstract void print (RegionSet s, Document doc, PrintWriter out);

    public String evaluate (RegionSet s, Document doc) {
	StringWriter out = new StringWriter ();
        PrintWriter pout = new PrintWriter (out);
	print (s, doc, pout);
        pout.close ();
	return out.toString ();
    }

    public static Template SELF = new SelfTemplate ();
    public static Template SIMPLE_LIST = new ListTemplate ("", SELF, "", "");
    public static Template HTML_LINES = new ListTemplate ("", SELF, 
                                                          "<br>\n", 
                                                          "\n");
    public static Template HTML_PARAGRAPHS = new ListTemplate ("<p>\n", 
                                                               SELF,
                                                               "</p>\n<p>\n",
                                                               "</p>\n");
    public static Template HTML_BULLETS = new ListTemplate ("<ul>\n<li>\n",
                                                            SELF,
                                                            "</li>\n<li>\n",
                                                            "</li></ul>\n");
    public static Template TEXT_LINES = new ListTemplate ("", 
                                                          SELF,
                                                          "\n", 
                                                          "\n"); 
    public static Template TEXT_PARAGRAPHS = new ListTemplate ("",
                                                               SELF,
                                                               "\n\n", 
                                                               "\n");
    public static Template TEXT_BULLETS = new ListTemplate ("* ", 
                                                            SELF,
                                                            "\n*", 
                                                            "\n");


    public static class LiteralTemplate extends Template {
        String literal;

        public LiteralTemplate (String literal) {
            this.literal = literal;
        }


        /** 
         * Translate escape sequences, discarding backslash if 
         * followed by an untranslatable character.
         */
        public static Template makeFromString (String literalWithEscapes) {
            return new LiteralTemplate (Str.unescape (literalWithEscapes,
                                                      '\\', 
                                                      "btnfr{}\\", 
                                                      "\b\t\n\f\r{}\\", 
                                                      false));
        }


        public void print (RegionSet list, Document doc, PrintWriter out) {
            out.print (literal);
        }
    }

    public static class SelfTemplate extends Template {
        public void print (RegionSet list, Document doc, PrintWriter out) {
            RegionEnumeration e = list.regions ();
            for (Region r =  e.firstFast (); 
                 r != null; 
                 r =  e.nextFast ())
                out.print (doc.getText (r));
        }
    }

    public static class SequenceTemplate extends Template {
        Template[] parts;

        protected SequenceTemplate () {
        }

        public SequenceTemplate (Template[] parts) {
            this.parts = parts;
        }

        public static Template makeFromString (String strTemplate) 
            throws TCParseException {
            Vector pieces = new Vector ();
            // vector of Templates
        
            // Split up strTemplate into literal strings and curly-bracketed
            // TC patterns.  Ignores nested and escaped curly braces.
            int startLiteral = 0;  // Start of current literal in strTemplate
            int startPattern = 0;  // Start of current pattern (just after {)
            int depth = 0;    // Curly brace nesting depth
            // 0 means we're outside of curly braces
            for (int i=0, n=strTemplate.length(); i < n; ++i) {
                switch (strTemplate.charAt (i)) {
                case '{':
                    if (depth == 0) {
                        if (startLiteral < i)
                            pieces.addElement (LiteralTemplate.makeFromString (strTemplate.substring (startLiteral, i)));
                        startPattern = i+1;
                    }
                    ++depth;
                    break;
                case '}':
                    --depth;
                    if (depth == 0) {
                        try {
                            String expr = strTemplate.substring (startPattern, i);
                            if (expr.equals ("match"))
                                pieces.addElement (SELF);
                            else {
                                TC tc = new TC (expr);
                                pieces.addElement (new FieldTemplate (tc, SELF));
                            }
                        } catch (TCParseException e) {
                            e.addOffset (startPattern);
                            throw e;
                        }
                        startLiteral = i+1;
                    }
                    break;
                case '\\':
                    // Just ignore the next character
                    ++i;
                    break;
                }
            }
            
            if (depth > 0)
                throw new TCParseException ("Unterminated {"); 
            
            // Grab last literal in strTemplate
            if (startLiteral < strTemplate.length())
                pieces.addElement (LiteralTemplate.makeFromString (strTemplate.substring (startLiteral)));
            
            // Convert to template
            SequenceTemplate t = new SequenceTemplate ();
            t.parts = new Template[pieces.size ()];
            pieces.copyInto (t.parts);
            return t;
        }

        public void bind (RegionSet list, Document doc) throws TCParseException {
            for (int i=0; i < parts.length; ++i)
                parts[i].bind (list, doc);
        }

        public void print (RegionSet list, Document doc, PrintWriter out) {
            RegionEnumeration e = list.regions ();
            for (Region r = e.first (); r != null; r = e.next ()) {
                for (int i=0; i < parts.length; ++i)
                    parts[i].print (r, doc, out);
            }
        }
    }

    public static class FieldTemplate extends Template {
        TC pattern;  // pattern that extracts field
        Template template; // template for field
        RegionSet binding; // matches to pattern (in entire document)

        public FieldTemplate (TC pattern, Template template) {
            this.pattern = pattern;
            this.template = template;
        }

        public void bind (RegionSet list, Document doc) throws TCParseException {
            binding = pattern.matchWithWarnings (doc);
            template.bind (binding, doc);
        }

        public void print (RegionSet list, Document doc, PrintWriter out) {
            RegionEnumeration e = list.regions ();
            for (Region r = e.firstFast (); r != null; r = e.nextFast ())
                template.print (binding.and (r.in ()).flatten (), doc, out);
        }
    }

    public static class ListTemplate extends Template {
        String header;    // precedes entire list (only if list is non-empty)
        Template template; // template for each list item
        String separater; // between each pair of list items
        String footer;    // follows entire list (only if list is non-empty)

        public ListTemplate (String header, 
                             Template template, 
                             String separater, 
                             String footer) {
            this.header = header;
            this.footer = footer;
            this.separater = separater;
            this.template = template;
        }

        public void bind (RegionSet list, Document doc) throws TCParseException {
            template.bind (list, doc);
        }

        public void print (RegionSet list, Document doc, PrintWriter out) {
            RegionEnumeration e = list.regions ();
            Region r = e.first ();
            if (r != null) {
                out.print (header);
                template.print (r, doc, out);

                for (r = e.next (); r != null; r = e.next ()) {
                    out.print (separater);
                    template.print (r, doc, out);
                }
                out.print (footer);
            }
        }
    }
}
