/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.tools;

import lapis.*;
//import lapis.tc.*;
import java.io.*;
import java.net.*;
//import java.util.Enumeration;
//import java.util.Vector;
import lapisx.net.URLUtil;
import lapisx.util.*;

/**
 * Unknown is called when the interpreter can't match the
 * command name with any known command.
 */
public class Unknown implements Tool {
    public static Debug debug = Debug.QUIET;

    public Document invoke (Arguments args) throws Exception {
        // Recover the original argument list
        Object[] originalArgs = args.getOriginalArguments ();
        String[] argv = new String[originalArgs.length];
        StringBuffer buf = new StringBuffer ();
        for (int i = 0; i < argv.length; ++i) {
            argv[i] = originalArgs[i].toString ();
            buf.append (argv[i]);
            buf.append (' ');
        }
        buf.setLength (buf.length () - 1); // delete last space 
        String cmd = buf.toString ();
        File curdir = 
            (File) args.getInterpreter ().getVar (Interpreter.CURRENT_DIRECTORY_VAR);

        debug.println ("handling unknown command: " + cmd);

        // Try interpreting as a URL
        try {
            URL url = new URL (cmd);

            if ("file".equals (url.getProtocol ())) {
                File file = URLUtil.URLToFile (url);
                debug.println ("URL corresponds to file " + file);
                if (!file.isAbsolute ()) {
                    String ref = url.getRef (); // save ref, if any
                    url = URLUtil.FileToURL (new File (curdir, file.toString ()));
                    if (ref != null && ref.length () > 0)
                        url = new URL (url, "#" + ref);
                    debug.println ("absolute URL is now " + url);
                }
            }
            return args.getFactory ().make (url);
        } catch (MalformedURLException e) {
            // not a URL -- try another interpretation
        }

        // Try invoking as external program
        boolean hasExecPrefix = cmd.startsWith ("exec:"); 
        if (hasExecPrefix)
            // delete exec: prefix
            argv[0] = argv[0].substring (5);
        Process process = null;
        try {
            process = execProcess (argv, curdir);
        } catch (IOException e) {
            if (hasExecPrefix)
                // can only interpret this as an external program,
                // so fail
                throw new Arguments.UsageException ("unknown command: " + argv[0]);
        }
        if (process != null) {
            try {
                return makeProcessDocument (process, args);
            } catch (InterruptedIOException e) {
                // somebody pushed Stop -- kill the process and die
                process.destroy ();
                throw new ThreadDeath ();
            }
        }

        // Otherwise try interpreting as filename
        File file = new File (cmd);
        if (!file.isAbsolute ())
            file = new File (curdir, cmd);
        if (file.exists ())
            return args.getFactory ().make (file);

        //             // Finally, try to interpret cmd as a domain name
        //             try {
        //                 URL url = new URL ("http://" + cmd);
        //                 interp.setResult (ReflectObject.newInstance (interp, url.getClass (), url));
        //                 return;
        //             } catch (MalformedURLException e) {
        //             }

        throw new Arguments.UsageException ("unknown command or filename: " + cmd);
    }

    static Process execProcess (String[] argv, File curdir) throws IOException {
        // Create a CGI-like environment
        String[] env = {
            "GATEWAY_INTERFACE=CGI/1.1",
            "SERVER_SOFTWARE=LAPIS",
            "SERVER_NAME=localhost",
            "SERVER_PORT=0",
            "REMOTE_ADDR=127.0.0.1",
            "REMOTE_HOST=localhost",
            "REQUEST_METHOD=GET",
            "QUERY_STRING=",
            "SERVER_PROTOCOL=HTTP/1.1",
//             "CONTENT_LENGTH=" + 
//             ((doc != null) ? doc.getContentLength () : 0),
//             "CONTENT_TYPE=" +
//             ((doc != null) ? doc.getContentType () : "text/plain"),
//             "SCRIPT_NAME=" + lapisx.util.Str.join (cmdarray, " ")
//             "PATH_INFO=",
//             "PATH_TRANSLATED=",
//             "REMOTE_USER=",
//             "AUTH_TYPE=",
        };

        // Start the process
        return Exec.exec (argv, env, curdir);
    }

    static Document makeProcessDocument (Process process, Arguments args) throws IOException {
        // Start pumping current document (if any) 
        // into child process's standard input
        Document doc = args.getInput ();
        if (doc != null)
            new ProcessWriter(doc, process.getOutputStream ())
                .start ();
        else
            process.getOutputStream ().close ();

        // Create a document for process's standard error, 
        // and start fetching it in a background thread.
        // This prevents deadlock -- e.g., where Lapis is 
        // blocked trying to read the process stdout and 
        // the process is blocked trying to write its stderr.
        // Make error always text/plain.
        final Document stderr = 
            args.getFactory ()
            .make (process.getErrorStream (), "text/plain");
        new Thread () {
            public void run () {
                stderr.getText ();
            }
        }.start ();
        
        // Create document for process stdout.
        Document stdout = args.getFactory ().make (process.getInputStream ());

        // Store the stderr document in a property of stdout, and vice versa
        stdout.putProperty (Document.ProcessProperty, process);
        stdout.putProperty (Document.StdoutProperty, stdout);
        stdout.putProperty (Document.StderrProperty, stderr);

        stderr.putProperty (Document.ProcessProperty, process);
        stderr.putProperty (Document.StdoutProperty, stdout);
        stderr.putProperty (Document.StderrProperty, stderr);

        return stdout;
    }

    /**
      * Thread to copy a document to a process's input.
      */
    static class ProcessWriter extends Thread {
        Document doc;
        PrintWriter writer;
        
        ProcessWriter(Document doc, OutputStream out) {
            setPriority(4);
            this.doc = doc;
            this.writer = new PrintWriter (out);
        }
        
        public void run() {
            String text = doc.getText ();
            writer.print (text);
            writer.close ();
        }
    }
}

