/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.trees;

import lapis.*;
import java.io.*;

abstract class LazyRelop
	extends SingleDocRegionSet
	implements Cloneable, Serializable {
	protected RegionSet kernel;
	protected RegionSet bkgd;
	protected RegionSet bbox;

	public LazyRelop(RegionSet kernel, RegionSet bkgd) {
		debug.assertion(
			bkgd == null
				|| (bkgd instanceof SingleDocRegionSet
					&& kernel instanceof SingleDocRegionSet));
		debug.assertion(
			bkgd == null
				|| ((SingleDocRegionSet) bkgd).getVersion() == null
				|| ((SingleDocRegionSet) bkgd).getVersion().equals(
					((SingleDocRegionSet) kernel).getVersion()));
		this.kernel = kernel;
		this.bkgd = bkgd;
		this.bbox =
			applyOp(((SingleDocRegionSet) kernel).getBoundingBox(), null);
	}

	//
	// Accessor methods
	//
	public int getStartLow() {
		return ((SingleDocRegionSet) bbox).getStartLow();
	}

	public int getStartHigh() {
		return ((SingleDocRegionSet) bbox).getStartHigh();
	}

	public int getEndLow() {
		return ((SingleDocRegionSet) bbox).getEndLow();
	}

	public int getEndHigh() {
		return ((SingleDocRegionSet) bbox).getEndHigh();
	}

	LazyRelop makeLazyRelop(RegionSet kernel, RegionSet last) {
		try {
			LazyRelop s = (LazyRelop) clone();
			s.kernel = kernel;
			s.bbox =
				applyOp(((SingleDocRegionSet) kernel).getBoundingBox(), last);
			return s;
		} catch (CloneNotSupportedException e) {
			throw new RuntimeException(e.toString());
		}
	}

	public abstract RegionSet applyOp(Region r, RegionSet last);

	public int getSize() {
		return kernel.getSize();
	}

	public RegionEnumeration regions(Region bbox) {
		return new LazyEnumeration(bbox);
	}

	public RegionEnumeration regions(RegionSet bboxs) {
		SingleDocRegionSet s = bboxs.getSingleDocRegionSet(this.getDocument());
		return new LazyEnumeration(s.getBoundingBox());
	}

	public SubsetEnumeration subsets() {
		return new LazySetEnumeration();
	}

	public RegionSet freeze() {
		if (kernel instanceof Mutable) {
			// kernel might change underneath us
			MutableRegionSet m = new MutableRegionSet(getSize());
			m.insert(this);
			return m;
		} else
			// kernel is immutable, so we're immutable too
			return this;
	}

	class LazyEnumeration implements RegionEnumeration {
		Region bbox;
		RegionEnumeration ke;
		RegionSet opR;

		public LazyEnumeration(Region bbox) {
			this.bbox = bbox;
			this.ke = kernel.regions();
			// FIX: transform bbox and pass it to regions(), instead of
			// storing it here and applying it below
			this.opR =
				new FuzzyRegion(
					MIN_INDEX,
					getStartLow() - 1,
					MIN_INDEX,
					MAX_INDEX,
					getVersion());

		}

		public Region first() {
			return firstFast();
		}

		public Region next() {
			return nextFast();
		}

		public Region firstFast() {
			Region r = ke.firstFast();
			if (r == null)
				return null;

			opR = applyOp(r, opR);
			if (!opR.isEmpty() && opR.mightIntersect(bbox))
				return (Region) opR;
			else
				return nextFast();
		}

		public Region nextFast() {
			while (true) {
				Region r = ke.nextFast();
				if (r == null)
					return null;

				opR = applyOp(r, opR);
				if (!opR.isEmpty() && opR.mightIntersect(bbox))
					return (Region) opR;
			}
		}
	}

	class LazySetEnumeration implements SubsetEnumeration {
		SubsetEnumeration ke;
		RegionSet opS;

		public LazySetEnumeration() {
			this.ke = kernel.subsets();
			this.opS =
				new FuzzyRegion(
					MIN_INDEX,
					getStartLow() - 1,
					MIN_INDEX,
					MAX_INDEX,
					getVersion());
		}

		public RegionSet firstFast() {
			RegionSet s = ke.firstFast();
			if (s == null)
				return null;
			else if (s instanceof Region)
				return opS = applyOp((Region) s, opS);
			else
				return opS = makeLazyRelop(s, opS);
		}

		public RegionSet nextFast() {
			RegionSet s = ke.nextFast();
			if (s == null)
				return null;
			else if (s instanceof Region)
				return opS = applyOp((Region) s, opS);
			else
				return opS = makeLazyRelop(s, opS);
		}
	}

	public SubsetEnumeration singleDocize() {
		return new SubsetEnumeration() {
			public RegionSet firstFast() {
				return LazyRelop.this;
			}
			public RegionSet nextFast() {
				return null;
			}
		};
	}

	public DocumentVersion getVersion() {
		return ((SingleDocRegionSet) this.kernel).getVersion();
	}

	public Document getDocument() {
		if (this.getVersion() == null)
			return null;
		else
			return this.getVersion().getDocument();
	}

	public SingleDocRegionSet getSingleDocRegionSet(Document doc) {
		DocumentVersion ver = getVersion();
		if (ver == null)
			return EmptyRegion.EMPTY;

		if (ver.getDocument() == doc)
			return this;
		else
			return EmptyRegion.EMPTY;
	}
    
    public Versions getVersions() {
        return new SingleDocVersions(this.getVersion());
    }
}
