/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.trees;

import lapis.*;
import lapisx.iterator.*;
import java.util.*;

/**
 * MultiDocRegionSet is RegionSet that contains regions from multiple documents.
 */
public class MultiDocRegionSet extends RegionSet implements Mutable {

	// different document mapping different SingleDocRegionSet 
	private Map regionsetByDocument = new HashMap();
    // different document mapping different version
    private Map versionByDocument = new HashMap();

	MultiDocRegionSet() {
	}

	MultiDocRegionSet(HashMap r, HashMap v) {
		this.regionsetByDocument = r;
        this.versionByDocument = v;
	}

	public MultiDocRegionSet(SingleDocRegionSet s) {
		this.regionsetByDocument.put(s.getDocument(), s);
        this.versionByDocument.put(s.getDocument(), s.getVersion());
	}

	/** 
	 * insert r to MultiDocRegionSet
	 */
	public void insert(Region r) {
		if (r.isEmpty())
			return;
		Document doc = r.getDocument();
		if (doc == null)
			return;
		SingleDocRegionSet s =
			(SingleDocRegionSet) regionsetByDocument.get(doc);

		try {
			if (s == null) {
				RegionArray ra = new RegionArray();
				ra.insert(r);
				this.regionsetByDocument.put(doc, ra);
                this.versionByDocument.put(doc, r.getVersion());
			} else {
				((Mutable) s).insert(r);
			}
		} catch (RepresentationChangeException e) {
            SingleDocRegionSet rep = (SingleDocRegionSet)e.getRepresentation();
			this.regionsetByDocument.put(doc, rep);
            this.versionByDocument.put(doc, rep.getVersion());
		}
	}

	public void insert(int start, int end, DocumentVersion ver) {
		insert(new Region(start, end, ver));
	}

	public void insert(RegionSet s) {
		if (s instanceof Region)
			insert((Region) s);
		else {
			RegionEnumeration e = s.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast())
				insert(r);
		}
	}

	/** 
	 * delete r from MultiDocRegionSet
	 */
	public void delete(Region r) {
		Document doc = r.getDocument();
		if (r.isEmpty() || doc == null)
			return;
		SingleDocRegionSet s =
			(SingleDocRegionSet) regionsetByDocument.get(doc);
		if (s == null || s.isEmpty())
			return;

		try {
			((Mutable) s).delete(r);
			if (s.isEmpty()) {
				this.regionsetByDocument.remove(doc);
                this.versionByDocument.remove(doc);
            }
		} catch (RepresentationChangeException e) {
            SingleDocRegionSet rep = (SingleDocRegionSet)e.getRepresentation();
			this.regionsetByDocument.put(doc, rep);
            this.versionByDocument.put(doc, rep.getVersion());
		}
	}

	public void delete(RegionSet s) {
		if (s instanceof Region)
			delete((Region) s);
		else {
			RegionEnumeration e = s.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast())
				delete(r);
		}
	}

	/** 
	 * return the number of regions that it contains
	 */
	public int getSize() {
		int num = 0;
		Iterator i = regionsetByDocument.values().iterator();
		while (i.hasNext())
			num += ((RegionSet) i.next()).getSize();
		return num;
	}

	public RegionEnumeration regions(Region bbox) {
		Document doc = bbox.getDocument();
		if (doc == null)
			return new EmptyRegionSet().regions();
		else {
			SingleDocRegionSet s =
				(SingleDocRegionSet) regionsetByDocument.get(doc);
			return s.regions(bbox);
		}
	}

	public RegionEnumeration regions(RegionSet bboxs) {
		return new MultiDocRegionSetEnumeration(bboxs);
	}

	class MultiDocRegionSetEnumeration implements RegionEnumeration {
		// TODO: enumerate RegionSet that contains less documents 
		//		 by checking the size of DocumentVersion list
		RegionSet bboxs;
		SubsetEnumeration se;
		RegionSet currSet;
		RegionEnumeration currRe;
		boolean renew = false;

		MultiDocRegionSetEnumeration(RegionSet bboxs) {
			this.bboxs = bboxs;
			se = MultiDocRegionSet.this.singleDocize();
			currSet = se.firstFast();
			if (currSet != null)
				currRe = currSet.regions(bboxs);
			else
				currRe = null;
		}

		public Region first() {
			Region r = firstFast();
			return (r != null) ? (Region) r.freeze() : null;
		}
		public Region next() {
			Region r = nextFast();
			return (r != null) ? (Region) r.freeze() : null;
		}
		public Region firstFast() {
			if (currRe != null) {
				return currRe.firstFast();
			} else
				return null;
		}
		public Region nextFast() {
			do {
				if (currRe != null) {
					Region next;
					if (renew) {
						next = currRe.firstFast();
						renew = false;
					} else
						next = currRe.nextFast();

					if (next != null) {
						return next;
					} else {
						currSet = se.nextFast();
						currRe =
							(currSet != null) ? currSet.regions(bboxs) : null;
						renew = true;
					}
				} else {
					return null;
				}
			} while (true);
		}
	}

	/** 
	 * return a copy of this regionset
	 */
	public RegionSet freeze() {
		return new MultiDocRegionSet(new HashMap(this.regionsetByDocument),
                                     new HashMap(this.versionByDocument));
	}

	public SubsetEnumeration singleDocize() {
		return new SubsetEnumeration() {
			Iterator i = regionsetByDocument.values().iterator();
			public RegionSet firstFast() {
				return i.hasNext() ? (RegionSet) i.next() : null;
			}
			public RegionSet nextFast() {
				return i.hasNext() ? (RegionSet) i.next() : null;
			}
		};
	}

	/** 
	 * return a SingleDocBased regionset that mapping to the given doc
	 */
	public SingleDocRegionSet getSingleDocRegionSet(Document doc) {
		if (doc == null)
			return Region.EMPTY;
		SingleDocRegionSet s =
			(SingleDocRegionSet) regionsetByDocument.get(doc);
		return (s == null) ? Region.EMPTY : s;
	}

	public void replace(RegionSet mr) {
		debug.assertion(mr instanceof SingleDocRegionSet);
		regionsetByDocument.put(((SingleDocRegionSet) mr).getDocument(), mr);
	}

    private Versions myVersions = new Versions () {
        public Iterator versions() {
            return MultiDocRegionSet.this.versionByDocument.values().iterator();
        }
        public Iterator documents() {
            return MultiDocRegionSet.this.versionByDocument.keySet().iterator();
        }
        public int getSize() {
            return MultiDocRegionSet.this.versionByDocument.size();
        }
        public boolean contains (DocumentVersion ver) {
            return MultiDocRegionSet.this.versionByDocument.containsValue(ver);
        }
        public boolean contains (Document doc) {
            return MultiDocRegionSet.this.versionByDocument.containsKey(doc);
        }
        public DocumentVersion getVersion(Document doc) {
            return (DocumentVersion)MultiDocRegionSet.this.versionByDocument.get(doc);
        }
        public boolean equals (Versions vs) {
            if (this.getSize() != vs.getSize())
                return false;
                        
            Iterator i = this.versions();
            while (i.hasNext()) {
                if (!vs.contains((DocumentVersion)i.next()))
                    return false;
            }
                    
            return true;
        }
        public boolean compatible(Versions vs) {
            // TODO: enumerate Versions that contains less documents by checking getSize
            Iterator i = this.versions();
            while (i.hasNext()) {
                DocumentVersion sourceVer = (DocumentVersion)i.next();
                Document sourceDoc = sourceVer.getDocument();
                DocumentVersion targetVer = vs.getVersion(sourceDoc);
                if (!sourceVer.equals(targetVer))
                    return false;
            }
                    
            return true;
        }
    };
     
    public Versions getVersions () {
        return myVersions;
    }
}
