/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.trees;
import lapis.*;
import lapisx.util.Debug;

/**
 * RB-tree
 *    B-tree of regions (ordered lexicographically by c,b,d,e)
 *    with bounding box stored at each node
 *
 */

public class RBTree extends SingleDocRegionSet implements Mutable {
	public static Debug debug = Debug.QUIET;

	// requires: MIN_CHILDREN >= 2
	//          and MIN_CHILDREN <= floor(MAX_CHILDREN/2)
	//     static final int MIN_CHILDREN = 2;
	//     static final int MAX_CHILDREN = 3;

	static final int MIN_CHILDREN = 4;
	static final int MAX_CHILDREN = 12;

	DocumentVersion version;

	// representation invariant: for all nodes v,
	//     v.nChildren >= MIN_CHILDREN
	//     v.nChildren <= MAX_CHILDREN
	//     (Region)v = minimum bounding box of v's children
	//     v.min = minimum region in subtree of v (minimum by lexicographic order lessThan())

	Node root;

	int depth = 1; // number of Node levels in the tree
	// always >= 1
	int nRegions = 0; // number of primitive Regions in tree

	public RBTree(DocumentVersion ver) {
		this.version = ver;
		this.root = new Leaf(ver);
	}

	protected RBTree(RBTree tree, DocumentVersion ver) {
		root = tree.root.copy();
		root.updateBoundingBoxDeeply();
		depth = tree.depth;
		nRegions = tree.nRegions;

		version = tree.version;
	}

	public Object clone() {
		return new RBTree(this, this.version);
	}

	//
	// Accessor methods
	//

	public int getStartLow() {
		return root.getStartLow();
	}

	public int getStartHigh() {
		return root.getStartHigh();
	}

	public int getEndLow() {
		return root.getEndLow();
	}

	public int getEndHigh() {
		return root.getEndHigh();
	}

	public void insert(Region r) {
		if (r.isEmpty())
			return;

		DocumentVersion ver = r.getVersion();

		if ((this.version != null) && !(this.version.equals(ver)))
			return;

		this.version = ver;

		try {
			// Merge tree with r to delete all regions enclosed by r
			merge(root, r, true);

			// Now insert r in the tree
			insertAtRoot(r.freeze());

			//             checkTree (root);
		} catch (DoneException e) {
			// merge() found that r was redundant with a region
			// already in the tree -- don't need to add it
		}
	}

	public void insert(int start, int end, DocumentVersion ver)
		throws RepresentationChangeException {
		insert(new Region(start, end, ver));
	}

	public void insert(RegionSet s) throws RepresentationChangeException {
		if (s instanceof Region)
			insert((Region) s);
		else {
			RegionEnumeration e = s.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast())
				insert(r);
		}
	}

	public void delete(Region r) {
		if (isEmpty() || r.isEmpty())
			return;

		if (!this.version.equals(r.getVersion()))
			return;

		try {
			// Merge tree with r to delete all regions enclosed by r
			merge(root, r, false);
		} catch (DoneException e) {
		}
	}

	public void delete(RegionSet s) throws RepresentationChangeException {
		if (s instanceof Region)
			delete((Region) s);
		else {
			RegionEnumeration e = s.regions();
			for (Region r = e.firstFast(); r != null; r = e.nextFast())
				delete(r);
		}
	}

	public int getSize() {
		return nRegions;
	}

	public RegionEnumeration regions(Region bbox) {
		return new RBTreeEnumeration(this, bbox);
	}

	public RegionEnumeration regions(RegionSet bboxs) {
		SingleDocRegionSet s = bboxs.getSingleDocRegionSet(this.getDocument());
		return new RBTreeEnumeration(this, s.getBoundingBox());
	}

	public SubsetEnumeration subsets() {
		return root.subsets();
	}

	public RegionSet freeze() {
		return new RBTree(this, this.version);
	}

	/*
	 *  Domination test
	 *
	 */

	void merge(Node t, Region r, boolean inserting) throws DoneException {
		//if (t == root)
		//    System.err.println ("redundant (" + r + ")");
		if (!r.mightIntersect(t))
			return;

		if (t instanceof Leaf) {
			for (int i = 0; i < t.nChildren; ++i) {
				Region s = (Region) t.children[i];
				if (!inserting) {
					// we're deleting r from the tree.
					// only need to check whether r encloses s, and if so,
					// delete s from the tree
					if (r.mightEnclose(s)) {
						i = delete(t, i);
						--i;
					}
				} else {
					// we're inserting r in the tree.
					// test both directions (r subset of s and s subset of
					// r) and merge appropriately
					if (s.mightEnclose(r)) {
						// s contains r, and s is already in the tree.
						// we don't need to add r, so jump out of the search
						throw new DoneException();
					} else if (r.mightEnclose(s)) {
						// s was contained in r, so remove it from the tree.
						// but keep looking for other merge possibilities
						i = delete(t, i) - 1;
					}
				}
			}
		} else {
			for (int i = 0; i < t.nChildren; ++i)
				merge((Node) t.children[i], r, inserting);
		}
	}

	static class DoneException extends Exception {
	}

	/*
	 *  Insertion algorithm
	 *
	 */

	void insertAtRoot(RegionSet r) {
		Node u = insert(root, r);

		if (u != null) {
			// split root
			Node t = root;

			root = new Node(version);
			root.children[0] = t;
			root.children[1] = u;
			root.nChildren = 2;

			t.parent = root;
			u.parent = root;

			root.updateBoundingBox();
			++depth;
		}

		nRegions += r.getSize();
	}

	static Node insert(Node t, RegionSet r) {
		// update t's bounding box and count
		t.enlargeBoundingBox(r);

		if (t instanceof Leaf) {
			// find position to insert r
			// scan from right to left to speed up common case
			// of receiving regions in sorted order)
			int pos;
			for (pos = t.nChildren - 1;
				pos >= 0 && ((SingleDocRegionSet) r).lessThan(t.children[pos]);
				--pos);

			if (pos == -1)
				t.min = r; // new minimum

			// insert r to the right of pos
			return insertInNode(t, pos + 1, r);
		} else {
			// find position to insert r
			// scan from right to left to speed up common case
			// of receiving regions in sorted order)
			int pos;
			for (pos = t.nChildren - 1;
				pos > 0 && lessThanSubtree(r, (Node) t.children[pos]);
				--pos);

			// recursively insert
			Node u = insert((Node) t.children[pos], r);

			if (pos == 0)
				t.min = ((Node) t.children[0]).min; // possibility of new min

			if (u != null) {
				// a split occured; insert new node u to the right of pos
				u.parent = t;
				return insertInNode(t, pos + 1, u);
			} else
				return null;
		}
	}

	// Inserts r (which may be a region or a tree node) as child pos
	// of t, moving other children over.
	// If the node must be split, the new node is created and returned;
	// if the node does not need splitting, null is returned.
	static Node insertInNode(Node t, int pos, RegionSet r) {
		if (t.nChildren < MAX_CHILDREN) {
			// no split necessary
			if (pos < t.nChildren)
				// make room in array
				System.arraycopy(
					t.children,
					pos,
					t.children,
					pos + 1,
					t.nChildren - pos);
			t.children[pos] = r;
			reparent(t, pos, pos + 1);
			++t.nChildren;
		} else {
			// try to overflow to adjacent sibling
			Node leftSibling = getSibling(t, -1);
			Node rightSibling = getSibling(t, 1);

			if (leftSibling != null && leftSibling.nChildren < MAX_CHILDREN)
				// overflow to leftSibling
				divideChildren(leftSibling, t, pos + leftSibling.nChildren, r);
			else if (
				rightSibling != null && rightSibling.nChildren < MAX_CHILDREN)
				// overflow to rightSibling
				divideChildren(t, rightSibling, pos, r);
			else {
				// split node and propagate split upwards
				Node u =
					(t instanceof Leaf)
						? new Leaf(t.getVersion())
						: new Node(t.getVersion());
				divideChildren(t, u, pos, r);
				return u;
			}
		}

		return null;
	}

	// Divides the children of t and u evenly among the two nodes, after inserting r.
	//    To be precise:  S <- t.children + u.children
	//                    S' <- S with r inserted at position pos
	//                    t.children <- first half of S'
	//                    u.children <- second half of S'
	static void divideChildren(Node t, Node u, int pos, RegionSet r) {
		int n = t.nChildren + u.nChildren + 1;
		int nT = n / 2;
		int nU = n - nT;
		int shift = t.nChildren - nT;
		// number of regions to move from T to U (possibly negative)
		if (pos < nT)
			++shift; // need to shift an extra region to make room for r in T

		if (shift > 0) {
			// push regions from end of t to start of u
			// first make room in u
			System.arraycopy(u.children, 0, u.children, shift, u.nChildren);
			// then copy them from t to u
			System.arraycopy(
				t.children,
				t.nChildren - shift,
				u.children,
				0,
				shift);
			// now delete them from t
			clear(t.children, t.nChildren - shift, t.nChildren);
			// reparent the moved regions
			reparent(u, 0, shift);
		} else if (shift < 0) {
			// push regions from start of u to end of t
			System.arraycopy(u.children, 0, t.children, t.nChildren, -shift);
			System.arraycopy(
				u.children,
				-shift,
				u.children,
				0,
				u.nChildren + shift);
			clear(u.children, u.nChildren + shift, u.nChildren);
			// reparent the moved regions
			reparent(t, t.nChildren, t.nChildren - shift);
		}
		t.nChildren -= shift;
		u.nChildren += shift;

		if (pos < nT) {
			// insert r in t
			// first make room in t
			System.arraycopy(
				t.children,
				pos,
				t.children,
				pos + 1,
				t.nChildren - pos);
			t.children[pos] = r;
			reparent(t, pos, pos + 1);
			++t.nChildren;
		} else {
			// insert r in u
			// first correct pos
			pos -= nT;
			System.arraycopy(
				u.children,
				pos,
				u.children,
				pos + 1,
				u.nChildren - pos);
			u.children[pos] = r;
			reparent(u, pos, pos + 1);
			++u.nChildren;
		}

		t.updateBoundingBox();
		u.updateBoundingBox();

		// now everything should be correct
		debug.assertion(t.nChildren == nT && u.nChildren == nU);
		//checkNode (t);
		//checkNode (u);
	}

	static Node getSibling(Node t, int offset) {
		Node parent = t.parent;
		if (parent == null)
			return null;

		int d = findIndex(parent, t) + offset;
		if (d < 0 || d >= parent.nChildren)
			return null;

		return (Node) parent.children[d];
	}

	static boolean lessThanSubtree(RegionSet r, Node t) {
		return ((SingleDocRegionSet) r).lessThan(t.min);
	}

	static void clear(Object[] array, int start, int end) {
		for (int i = start; i < end; ++i)
			array[i] = null;
	}

	static void reparent(Node t, int start, int end) {
		if (t instanceof Leaf)
			return;
		for (int i = start; i < end; ++i)
			 ((Node) t.children[i]).parent = t;
	}

	/*
	*  Deletion algorithm
	*
	*/

	// delete the cth child of node t
	// returns the new index of the c+1st child
	int delete(Node t, int c) {
		RegionSet r = t.children[c]; // the child we're deleting

		c = deleteFromNode(t, c);

		// propagate bounding box changes up to root.
		// stop propagating if we reach a node that doesn't change.
		for (Node u = t; u != null; u = u.parent) {
			int sl = u.getStartLow();
			int sh = u.getStartHigh();
			int el = u.getEndLow();
			int eh = u.getEndHigh();

			u.updateBoundingBox();

			if (sl == u.getStartLow()
				&& sh == u.getStartHigh()
				&& el == u.getEndLow()
				&& eh == u.getEndHigh())
				break;
		}

		//checkTree (root);

		nRegions -= r.getSize();
		return c;
	}

	int deleteFromNode(Node t, int c) {
		if (c < t.nChildren - 1)
			System.arraycopy(
				t.children,
				c + 1,
				t.children,
				c,
				t.nChildren - c - 1);
		--t.nChildren;
		t.children[t.nChildren] = null;

		if (t == root) {
			if (t.nChildren == 1 && depth > 1) {
				root = (Node) t.children[0];
				root.parent = null;
				--depth;
			}
		} else if (t.nChildren < MIN_CHILDREN) {
			// underflow: merge t with its right sibling (unless it
			// doesn't have one)
			Node parent = t.parent;
			int d = findIndex(parent, t);
			int e = (d < parent.nChildren - 1) ? d + 1 : d - 1;
			Node u = (Node) parent.children[e];

			if (u.nChildren + t.nChildren <= MAX_CHILDREN) {
				// move all of u into t, and delete u
				if (e > d) {
					// append u's children after t's children
					System.arraycopy(
						u.children,
						0,
						t.children,
						t.nChildren,
						u.nChildren);
					reparent(t, t.nChildren, t.nChildren + u.nChildren);
				} else {
					// insert u's children ahead of t's children
					System.arraycopy(
						t.children,
						0,
						t.children,
						u.nChildren,
						t.nChildren);
					System.arraycopy(u.children, 0, t.children, 0, u.nChildren);
					reparent(t, 0, u.nChildren);
					c += u.nChildren;
				}
				t.nChildren += u.nChildren;
				t.updateBoundingBox();

				deleteFromNode(t.parent, e);
			} else {
				// split the children evenly among t and u
				int nU = (u.nChildren + t.nChildren) / 2;
				int moved = u.nChildren - nU;

				if (e > d) {
					// append u's children after t's children, move u down
					System.arraycopy(
						u.children,
						0,
						t.children,
						t.nChildren,
						moved);
					System.arraycopy(u.children, moved, u.children, 0, nU);
					reparent(t, t.nChildren, t.nChildren + moved);
				} else {
					// insert u's children ahead of t's children
					System.arraycopy(
						t.children,
						0,
						t.children,
						moved,
						t.nChildren);
					System.arraycopy(u.children, nU, t.children, 0, moved);
					c += moved;
					reparent(t, 0, moved);
				}
				clear(u.children, nU, u.nChildren);

				t.nChildren += moved;
				u.nChildren = nU;

				t.updateBoundingBox();
				u.updateBoundingBox();
			}
		}

		//checkTree (root);
		return c;
	}

	static int findIndex(Node u, Node t) {
		for (int d = 0; d < u.nChildren; ++d)
			if (u.children[d] == t)
				return d;
		throw new java.util.NoSuchElementException();
	}

	/*
	 *  Managing the representation invariant, and debugging code
	 *
	 */

	public void validate() {
		if (debug.assertionEnabled())
			checkTree(root);
	}

	void checkTree(Node t) {
		checkNode(t);
		if (!(t instanceof Leaf))
			for (int i = 0; i < t.nChildren; ++i)
				checkTree((Node) t.children[i]);
	}

	void checkNode(Node t) {
		// is parent correct?
		if (t == root)
			debug.assertion(t.parent == null);

		// is nChildren legal?
		if (t == root)
			debug.assertion(t.nChildren >= 0 && t.nChildren <= MAX_CHILDREN);
		else
			debug.assertion(
				t.nChildren >= MIN_CHILDREN && t.nChildren <= MAX_CHILDREN);

		// are children correct?
		if (t instanceof Leaf) {
			for (int i = 0; i < t.nChildren; ++i) {
				RegionSet r = t.children[i];
				debug.assertion(r != null);
				debug.assertion(!(r instanceof Node));
				if (i > 0) {
					debug.assertion(r != t.children[i - 1]);
					debug.assertion(
						!((SingleDocRegionSet) r).lessThan(t.children[i - 1]));
				}
			}
		} else {
			for (int i = 0; i < t.nChildren; ++i) {
				Node u = (Node) t.children[i];
				debug.assertion(u != null);
				if (i > 0) {
					debug.assertion(u != t.children[i - 1]);
					debug.assertion(
						!((SingleDocRegionSet) u.min).lessThan(
							((Node) t.children[i - 1]).min));
				}
				debug.assertion(u.parent == t);
			}
		}

		// are remaining child slots empty?
		for (int i = t.nChildren; i < MAX_CHILDREN; ++i)
			debug.assertion(t.children[i] == null);

		// is t the minimum bounding box of its children?
		int l = MAX_INDEX;
		int m = MIN_INDEX;
		int r = MAX_INDEX;
		int s = MIN_INDEX;
		for (int i = 0; i < t.nChildren; ++i) {
			RegionSet x = t.children[i];
			l = Math.min(l, ((SingleDocRegionSet) x).getStartLow());
			m = Math.max(m, ((SingleDocRegionSet) x).getStartHigh());
			r = Math.min(r, ((SingleDocRegionSet) x).getEndLow());
			s = Math.max(s, ((SingleDocRegionSet) x).getEndHigh());
		}
		debug.assertion(
			t.getStartLow() == l
				&& t.getStartHigh() == m
				&& t.getEndLow() == r
				&& t.getEndHigh() == s);
	}

	//
	// Externalizable interface
	//
	/*
	    public void writeExternal (ObjectOutput out) throws IOException {
	        out.writeInt (depth);
	        out.writeInt (nRegions);
	        root.writeExternal (out);
	    }
	
	    public void readExternal(ObjectInput in) throws IOException {
	        depth = in.readInt ();
	        nRegions = in.readInt ();
	        root = readNode (in, null);
	        root.updateBoundingBoxDeeply ();
	    }
	
	    public static Node readNode (ObjectInput in, Node parent) throws IOException {
	        Node n;
	        char b = (char)in.readByte ();
	        switch (b) {
	        case 'R':
	            n = new Node ();
	            break;
	        case 'L':
	            n = new Leaf ();
	            break;
	        default:
	            throw new IOException ("bad node type " + b);
	        }
	        n.readExternal (in);
	        n.parent = parent;
	        return n;
	    }
	*/
	/**
	 * Node in the RB tree.
	 */
	static class Node extends SingleDocRegionSet {
		int startLow;
		int startHigh;
		int endLow;
		int endHigh;

		RegionSet[] children = new RegionSet[MAX_CHILDREN];
		int nChildren = 0; // number of actual children

		RegionSet min; // smallest leaf in this subtree 
		// (by lessThan() ordering)
		Node parent;

		DocumentVersion version;

		public Node(DocumentVersion ver) {
			startLow = (Region.EMPTY).getStartLow();
			startHigh = (Region.EMPTY).getStartHigh();
			endLow = (Region.EMPTY).getEndLow();
			endHigh = (Region.EMPTY).getEndHigh();
			version = ver;
		}

		public Node(Node node) {
			nChildren = node.nChildren;
			version = node.version;
			for (int i = 0; i < nChildren; ++i) {
				Node child = (Node) node.children[i];
				child = child.copy();
				child.parent = this;
				children[i] = child;
			}
			// don't initialize anything else -- let updateBoundingBoxDeeply() handle it
		}

		//
		// Accessor methods
		//

		public int getStartLow() {
			return startLow;
		}

		public int getStartHigh() {
			return startHigh;
		}

		public int getEndLow() {
			return endLow;
		}

		public int getEndHigh() {
			return endHigh;
		}

		protected void enlargeBoundingBox(RegionSet r) {
			startLow =
				Math.min(((SingleDocRegionSet) r).getStartLow(), startLow);
			startHigh =
				Math.max(((SingleDocRegionSet) r).getStartHigh(), startHigh);
			endLow = Math.min(((SingleDocRegionSet) r).getEndLow(), endLow);
			endHigh = Math.max(((SingleDocRegionSet) r).getEndHigh(), endHigh);
		}

		protected void setBoundingBox(RegionSet r) {
			startLow = ((SingleDocRegionSet) r).getStartLow();
			startHigh = ((SingleDocRegionSet) r).getStartHigh();
			endLow = ((SingleDocRegionSet) r).getEndLow();
			endHigh = ((SingleDocRegionSet) r).getEndHigh();
		}

		public Node copy() {
			return new Node(this);
		}

		public RegionSet freeze() {
			// this class shouldn't leak out of RBTree, so we don't
			// need to implement freeze ().  copy() does what freeze () would do, anyway.
			throw new RuntimeException("operation not supported");
		}

		public void updateBoundingBox() {
			startLow = (Region.EMPTY).getStartLow();
			startHigh = (Region.EMPTY).getStartHigh();
			endLow = (Region.EMPTY).getEndLow();
			endHigh = (Region.EMPTY).getEndHigh();

			for (int i = 0; i < nChildren; ++i) {
				RegionSet r = children[i];
				enlargeBoundingBox(r);
			}

			min = nChildren > 0 ? ((Node) children[0]).min : null;
		}

		public void updateBoundingBoxDeeply() {
			for (int i = 0; i < nChildren; ++i)
				 ((Node) children[i]).updateBoundingBoxDeeply();
			updateBoundingBox();
		}

		public int getSize() {
			int n = 0;
			for (int i = 0; i < nChildren; ++i)
				n += children[i].getSize();
			return n;
		}

		public RegionEnumeration regions(Region bbox) {
			return new RBTreeEnumeration(this, bbox);
		}

		public RegionEnumeration regions(RegionSet bboxs) {
			SingleDocRegionSet s =
				bboxs.getSingleDocRegionSet(this.getDocument());
			return new RBTreeEnumeration(this, s.getBoundingBox());
		}

		public SubsetEnumeration subsets() {
			return new SubsetEnumeration() {
				int i = 0; // index of next child to return
				public RegionSet firstFast() {
					i = 0;
					return nextFast();
				}
				public RegionSet nextFast() {
					if (i < nChildren)
						return children[i++];
					else
						return null;
				}
			};
		}

		/*
		        public void writeExternal (ObjectOutput out) throws IOException {
		            out.writeByte ('R');
		            out.writeInt (nChildren);
		            for (int i = 0; i < nChildren; ++i)
		                ((Node)children[i]).writeExternal (out);
		        }
		
		        public void readExternal(ObjectInput in) throws IOException {
		            nChildren = in.readInt ();
		            for (int i = 0; i < nChildren; ++i)
		                children[i] = readNode (in, this);
		        }
		*/
		public SubsetEnumeration singleDocize() {
			return new SubsetEnumeration() {
				public RegionSet firstFast() {
					return Node.this;
				}
				public RegionSet nextFast() {
					return null;
				}
			};
		}
		public DocumentVersion getVersion() {
			return this.version;
		}
		public Document getDocument() {
			if (this.version == null)
				return null;
			return this.version.getDocument();
		}
		public SingleDocRegionSet getSingleDocRegionSet(Document doc) {
			if (this.version.getDocument() == doc)
				return this;
			else
				return Region.EMPTY;
		}
        public Versions getVersions() {
            return new SingleDocVersions(this.version);
        }
	}

	/**
	 * Leaf node in the RB tree.
	 */
	static class Leaf extends Node {
		public Leaf(DocumentVersion ver) {
			super(ver);
		}

		public Leaf(Leaf node) {
			super(node.getVersion());
			nChildren = node.nChildren;
			System.arraycopy(node.children, 0, children, 0, nChildren);
		}

		public Node copy() {
			return new Leaf(this);
		}

		public void updateBoundingBox() {
			min = children[0];

			startLow = (Region.EMPTY).getStartLow();
			startHigh = (Region.EMPTY).getStartHigh();
			endLow = (Region.EMPTY).getEndLow();
			endHigh = (Region.EMPTY).getEndHigh();

			for (int i = 0; i < nChildren; ++i) {
				RegionSet r = children[i];
				enlargeBoundingBox(r);
			}
		}

		public void updateBoundingBoxDeeply() {
			updateBoundingBox();
		}
		/*
		        public void writeExternal (ObjectOutput out) throws IOException {
		            out.writeByte ('L');
		            out.writeInt (nChildren);
		            for (int i = 0; i < nChildren; ++i) {
		                RegionSet r = children[i];
		                out.writeInt (r.singleDocize ().getStartLow ());
		                out.writeInt (r.singleDocize ().getStartHigh ());
		                out.writeInt (r.singleDocize ().getEndLow ());
		                out.writeInt (r.singleDocize ().getEndHigh ());
		            }
		        }
		
		        public void readExternal(ObjectInput in) throws IOException {
		            nChildren = in.readInt ();
		            for (int i = 0; i < nChildren; ++i) {
		                int sl = in.readInt ();
		                int sh = in.readInt ();
		                int el = in.readInt ();
		                int eh = in.readInt ();
		                children[i] = new FuzzyRegion (sl, sh, el, eh);
		            }
		        }
		*/
	}

	/**
	 * RegionEnumeration for RBTrees.
	 */
	static class RBTreeEnumeration implements RegionEnumeration {
		Node root;
		Region range;

		Node curr; // current node in tree
		int[] stack; // path to next from root:
		// next = root.children[stack[0]]
		//         .children[stack[1]]
		//         ...
		//         .children[stack[depth-2]]
		//         .children[stack[depth-1]]

		public RBTreeEnumeration(RBTree rt, Region range) {
			this.range = range;
			stack = new int[rt.depth];
			curr = rt.root;
		}

		public RBTreeEnumeration(Node t, Region range) {
			this.range = range;
			stack = new int[countDepth(t)];
			curr = t;
		}

		public Region first() {
			stack[0] = -1;
			return findNext(0);
		}

		public Region next() {
			return findNext(stack.length - 1);
		}

		public Region firstFast() {
			return first();
		}

		public Region nextFast() {
			return next();
		}

		int countDepth(Node t) {
			if (t instanceof Leaf || t.nChildren == 0)
				return 1;
			else
				return 1 + countDepth((Node) t.children[0]);
		}

		Region findNext(int d) {
			int dLeaf = stack.length - 1;

			while (d >= 0) {
				// go right
				int i = ++stack[d];
				if (i < curr.nChildren) {
					RegionSet r = curr.children[i];

					if (r.mightIntersect(range)) {
						if (d == dLeaf)
							// found a new leaf, return it
							return (Region) r;
						else {
							// go down
							curr = (Node) r;
							++d;
							stack[d] = -1;
						}
					}
				} else {
					// go up
					curr = curr.parent;
					--d;
				}
			}

			return null; // end of enumeration
		}
	}

	public SubsetEnumeration singleDocize() {
		return new SubsetEnumeration() {
			public RegionSet firstFast() {
				return RBTree.this;
			}
			public RegionSet nextFast() {
				return null;
			}
		};
	}

	public DocumentVersion getVersion() {
		return this.version;
	}

	public Document getDocument() {
		if (this.version == null)
			return null;

		return this.version.getDocument();
	}

	public SingleDocRegionSet getSingleDocRegionSet(Document doc) {
		if (this.version.getDocument() == doc)
			return this;
		else
			return Region.EMPTY;
	}
    
    public Versions getVersions () {
        return new SingleDocVersions(this.version);
    }
}
