/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapis.trees;

import lapis.*;
import lapisx.util.BinarySearch;
import lapisx.sort.*;
import lapisx.util.Debug;

/**
 * Region array.  Sorted list of region rectangles.
 */

public class RegionArray extends SingleDocRegionSet implements Mutable {
	public static Debug debug = Debug.QUIET;

	int state = MONOTONIC;
	static final int UNSORTED = 0;
	static final int SORTED = 1;
	static final int MONOTONIC = 2;

	int[] points;
	int slOffset; // offset of startLow array in points
	int shOffset; // offset of startHigh array
	int elOffset; // offset of endLow array
	int ehOffset; // offset of endHigh array

	DocumentVersion version;

	int n = 0; // number of rectangles in points 
	int max = 0; // maximum rectangles that can fit in points
	int size = 0; // number of undeleted rectangles

	boolean[] deleted;

	static final int DEFAULT_SIZE = 1;

	/**
	 * Make a new, empty RegionArray.
	 */
	public RegionArray() {
		this(DEFAULT_SIZE);
	}

	public RegionArray(DocumentVersion ver) {
		this(DEFAULT_SIZE, ver);
	}
	/**
	 * Make a new, empty RegionArray.
	 * @param size Expected number of regions
	 */
	public RegionArray(int size) {
		max = (size > 0) ? size : DEFAULT_SIZE;
		points = new int[max * 2];
		slOffset = shOffset = 0;
		elOffset = ehOffset = max;
	}

	public RegionArray(int size, DocumentVersion ver) {
		max = (size > 0) ? size : DEFAULT_SIZE;
		points = new int[max * 2];
		slOffset = shOffset = 0;
		elOffset = ehOffset = max;
		version = ver;
	}

	/**
	 * Copy a RegionArray.
	 */
	protected RegionArray(RegionArray ra, boolean copy) {
		if (copy) {
			points = new int[ra.points.length];
			System.arraycopy(ra.points, 0, points, 0, ra.points.length);

			if (ra.deleted != null) {
				deleted = new boolean[ra.deleted.length];
				System.arraycopy(ra.deleted, 0, deleted, 0, ra.deleted.length);
			}
		} else {
			points = ra.points;
			deleted = ra.deleted;
		}

		state = ra.state;
		slOffset = ra.slOffset;
		elOffset = ra.elOffset;
		shOffset = ra.shOffset;
		ehOffset = ra.ehOffset;
		n = ra.n;
		max = ra.max;
		size = ra.size;
		state = ra.state;

		version = ra.version;
	}

	//
	// bounding box of region set
	//

	public int getStartLow() {
		validate();
		return (n > 0) ? points[slOffset] : (Region.EMPTY).getStartLow();
	}

	public int getStartHigh() {
		validate();
		return (n > 0)
			? points[shOffset + n - 1]
			: (Region.EMPTY).getStartHigh();
	}

	public int getEndLow() {
		validate();
		return (n > 0) ? points[elOffset] : (Region.EMPTY).getEndLow();
	}

	public int getEndHigh() {
		validate();
		return (n > 0) ? points[ehOffset + n - 1] : (Region.EMPTY).getEndHigh();
	}

	//
	// size of region set
	//

	public int getSize() {
		return size;
	}

	//
	// inserting regions
	//

	public void insert(Region s) throws RepresentationChangeException {
		if (!s.isEmpty())
			insert(
				s.getStartLow(),
				s.getStartHigh(),
				s.getEndLow(),
				s.getEndHigh(),
				s.getVersion());
	}

    public void insert (RegionSet s) throws RepresentationChangeException {
        if (s instanceof Region)
            insert ((Region) s);
        else {
            RegionEnumeration e = s.regions ();
            for (Region r = e.firstFast (); r != null; r = e.nextFast ())
                insert (r);
        }
    }

	public void insert(int start, int end, DocumentVersion ver)
		throws RepresentationChangeException {
		insert(start, start, end, end, ver);
	}

	void insert(int sl, int sh, int el, int eh, DocumentVersion ver)
		throws RepresentationChangeException {
		if ((this.version != null) && !(this.version.equals(ver)))
			return;

		this.version = ver;

		int lastSL;
		int lastSH;
		int lastEL;
		int lastEH;

		if (n > 0) {
			lastSL = points[slOffset + n - 1];
			lastEL = points[elOffset + n - 1];
			lastSH = points[shOffset + n - 1];
			lastEH = points[ehOffset + n - 1];
		} else {
			lastSL = lastEL = lastSH = lastEH = -1;
		}

		// if previous region encloses this new one,
		// don't insert
		if (lastSL <= sl && lastEL <= el && lastSH >= sh && lastEH >= eh)
			return;

		// if this region encloses previous region, remove the
		// previous region.  Repeat until can't.
		while (sl <= lastSL && el <= lastEL && sh >= lastSH && eh >= lastEH) {
			--n;
			--size;

			if (n > 0) {
				lastSL = points[slOffset + n - 1];
				lastEL = points[elOffset + n - 1];
				lastSH = points[shOffset + n - 1];
				lastEH = points[ehOffset + n - 1];
			} else {
				lastSL = lastEL = lastSH = lastEH = -1;
			}
		}

		// convert to fuzzy regions if necessary
		if (slOffset == shOffset && (sl != sh || el != eh))
			makeFuzzy();

		// make room for new region
		if (n >= max)
			makeRoom();

		// store the new region
		points[slOffset + n] = sl;
		points[elOffset + n] = el;
		points[shOffset + n] = sh;
		points[ehOffset + n] = eh;

		// update invariant state
		if (state != UNSORTED && n > 0) {
			if (sortable.compare(n - 1, n) > 0)
				state = UNSORTED;
			else if (
				state == MONOTONIC
					&& (lastSL > sl || lastEL > el || lastSH > sh || lastEH > eh))
				state = SORTED;
		}

		++n;
		++size;
	}

	void makeFuzzy() {
		int[] newpoints = new int[max * 4];
		int newSHOffset = max * 2;
		int newEHOffset = max * 3;
		System.arraycopy(points, slOffset, newpoints, slOffset, n);
		System.arraycopy(points, elOffset, newpoints, elOffset, n);
		System.arraycopy(points, shOffset, newpoints, newSHOffset, n);
		System.arraycopy(points, ehOffset, newpoints, newEHOffset, n);
		points = newpoints;
		shOffset = newSHOffset;
		ehOffset = newEHOffset;
	}

	void makeRoom() {
		int newmax = max * 2;
		int[] newpoints;
		if (slOffset == shOffset) {
			newpoints = new int[newmax * 2];
			System.arraycopy(points, slOffset, newpoints, slOffset * 2, n);
			System.arraycopy(points, elOffset, newpoints, elOffset * 2, n);
		} else {
			newpoints = new int[newmax * 4];
			System.arraycopy(points, slOffset, newpoints, slOffset * 2, n);
			System.arraycopy(points, elOffset, newpoints, elOffset * 2, n);
			System.arraycopy(points, shOffset, newpoints, shOffset * 2, n);
			System.arraycopy(points, ehOffset, newpoints, ehOffset * 2, n);
		}

		points = newpoints;
		max = newmax;
		slOffset *= 2;
		elOffset *= 2;
		shOffset *= 2;
		ehOffset *= 2;
	}

	//
	// finish insertion by sorting and trimming rectangles
	//
	void validateInsert() {
		if (state == MONOTONIC)
			return;

		if (state == UNSORTED) {
			Quicksort.sort(sortable);
			state = SORTED;
		}

		deleteRedundant();
		trim();

		if (isMonotonic()) {
			state = MONOTONIC;
			return;
		}

		throw toRBTree();
	}

	RepresentationChangeException toRBTree()
		throws RepresentationChangeException {
		//         debug.println ("converting nonmonotonic " + this + " to RBTree");
		//         dump ();
		RBTree t = new RBTree(this.version);

		if (slOffset == shOffset) {
			for (int i = 0; i < n; ++i)
				if (deleted == null || !deleted[i])
					t.insert(
						points[slOffset + i],
						points[ehOffset + i],
						this.version);
		} else {
			for (int i = 0; i < n; ++i)
				if (deleted == null || !deleted[i])
					t.insert(
						new FuzzyRegion(
							points[slOffset + i],
							points[shOffset + i],
							points[elOffset + i],
							points[ehOffset + i],
							version));
		}

		return new RepresentationChangeException(t);
	}

	transient Sortable2 sortable = new Sortable2() {
		public int size() {
			return n;
		}

		public void swap(int i, int j) {
			int sl = points[slOffset + i];
			int el = points[elOffset + i];
			int sh = points[shOffset + i];
			int eh = points[ehOffset + i];

			points[slOffset + i] = points[slOffset + j];
			points[elOffset + i] = points[elOffset + j];
			points[shOffset + i] = points[shOffset + j];
			points[ehOffset + i] = points[ehOffset + j];

			points[slOffset + j] = sl;
			points[elOffset + j] = el;
			points[shOffset + j] = sh;
			points[ehOffset + j] = eh;
		}

		public int compare(int i, int j) {
			int xi, xj;

			xi = points[slOffset + i];
			xj = points[slOffset + j];
			if (xi < xj)
				return -1;
			else if (xi > xj)
				return 1;

			xi = points[elOffset + i];
			xj = points[elOffset + j];
			if (xi < xj)
				return -1;
			else if (xi > xj)
				return 1;

			xi = points[shOffset + i];
			xj = points[shOffset + j];
			if (xi < xj)
				return -1;
			else if (xi > xj)
				return 1;

			xi = points[ehOffset + i];
			xj = points[ehOffset + j];
			if (xi < xj)
				return -1;
			else if (xi > xj)
				return 1;

			return 0;
		}
	};

	void deleteRedundant() {
		if (n < 2)
			return;

		int sl1 = points[slOffset];
		int el1 = points[elOffset];
		int sh1 = points[shOffset];
		int eh1 = points[ehOffset];

		for (int i = 1; i < n; ++i) {
			int sl2 = points[slOffset + i];
			int el2 = points[elOffset + i];
			int sh2 = points[shOffset + i];
			int eh2 = points[ehOffset + i];

			// if r1 encloses r2, delete r2
			if (sl1 <= sl2 && el1 <= el2 && sh1 >= sh2 && eh1 >= eh2) {
				debug.println(
					" ["
						+ points[slOffset
						+ i
						- 1]
						+ "-"
						+ points[shOffset
						+ i
						- 1]
						+ ","
						+ points[elOffset
						+ i
						- 1]
						+ "-"
						+ points[ehOffset
						+ i
						- 1]
						+ "] encloses ["
						+ points[slOffset
						+ i]
						+ "-"
						+ points[shOffset
						+ i]
						+ ","
						+ points[elOffset
						+ i]
						+ "-"
						+ points[ehOffset
						+ i]
						+ "]");
				markDeleted(i);
			}

			sl1 = sl2;
			el1 = el2;
			sh1 = sh2;
			eh1 = eh2;
		}

		compact();
	}

	void trim() {
		if (n < 2)
			return;

		int sl1 = points[slOffset];
		int el1 = points[elOffset];
		int sh1 = points[shOffset];
		int eh1 = points[ehOffset];

		for (int i = 1; i < n; ++i) {
			int sl2 = points[slOffset + i];
			int el2 = points[elOffset + i];
			int sh2 = points[shOffset + i];
			int eh2 = points[ehOffset + i];

			if (sl2 <= sh1 && el2 >= el1 && eh2 <= eh1)
				sl2 = points[slOffset + i] = sh1 + 1;

			if (sl2 > sh2)
				markDeleted(i);

			sl1 = sl2;
			el1 = el2;
			sh1 = sh2;
			eh1 = eh2;
		}

		compact();
	}

	// test whether rectangles are monotonic.
	// assumes rectangles are already sorted.
	boolean isMonotonic() {
		if (n < 2)
			return true;

		int sl1 = points[slOffset];
		int el1 = points[elOffset];
		int sh1 = points[shOffset];
		int eh1 = points[ehOffset];

		for (int i = 1; i < n; ++i) {
			int sl2 = points[slOffset + i];
			int el2 = points[elOffset + i];
			int sh2 = points[shOffset + i];
			int eh2 = points[ehOffset + i];

			if (sl1 > sl2 || el1 > el2 || sh1 > sh2 || eh1 > eh2) {
				debug.println(
					"not monotonic because "
						+ " ["
						+ points[slOffset
						+ i
						- 1]
						+ "-"
						+ points[shOffset
						+ i
						- 1]
						+ ","
						+ points[elOffset
						+ i
						- 1]
						+ "-"
						+ points[ehOffset
						+ i
						- 1]
						+ "] < ["
						+ points[slOffset
						+ i]
						+ "-"
						+ points[shOffset
						+ i]
						+ ","
						+ points[elOffset
						+ i]
						+ "-"
						+ points[ehOffset
						+ i]
						+ "]");
				return false;
			}

			sl1 = sl2;
			el1 = el2;
			sh1 = sh2;
			eh1 = eh2;
		}

		return true;
	}

	//
	// deleting regions
	//

	public void delete(Region s) throws RepresentationChangeException {
		if (isEmpty() || s.isEmpty())
			return;

		if (!this.version.equals(s.getVersion()))
			return;

		validateInsert();

		RegionArrayEnumeration re = new RegionArrayEnumeration(s);
		for (Region r = re.firstFast(); r != null; r = re.nextFast()) {
			if (s.mightEnclose(r))
				markDeleted(re.iCurr - 1);
			else {
				RepresentationChangeException e = toRBTree();
				((Mutable)e.getRepresentation()).delete(s);
				throw e;
			}
		}
	}

    public void delete (RegionSet s) throws RepresentationChangeException {
        if (s instanceof Region)
            delete ((Region)s);
        else {
            RegionEnumeration e = s.regions ();
            for (Region r = e.firstFast (); r != null; r = e.nextFast ())
                delete (r);
        }
    }

	void markDeleted(int i) {
		if (deleted == null)
			deleted = new boolean[max];
		if (!deleted[i])
			--size;
		deleted[i] = true;
	}

	void validateDelete() {
		if (deleted == null)
			return;

		compact();
		trim();
	}

	void compact() {
		if (deleted == null)
			return;

		int newmax;
		int[] newpoints;
		int newSLOffset;
		int newSHOffset;
		int newELOffset;
		int newEHOffset;

		// reallocate points array if it's shrunk below a quarter
		// of its current size
		if (size < max / 4) {
			newmax = Math.max(size * 2, DEFAULT_SIZE);
			if (slOffset == shOffset) {
				newpoints = new int[newmax * 2];
				newSLOffset = newSHOffset = 0;
				newELOffset = newEHOffset = newmax;
			} else {
				newpoints = new int[newmax * 4];
				newSLOffset = 0;
				newELOffset = newmax;
				newSHOffset = newmax * 2;
				newEHOffset = newmax * 3;
			}
		} else {
			newmax = max;
			newpoints = points;
			newSLOffset = slOffset;
			newELOffset = elOffset;
			newSHOffset = shOffset;
			newEHOffset = ehOffset;
		}

		// compact points into newpoints (may be the same array)
		int j = 0;
		if (slOffset == shOffset) {
			for (int i = 0; i < n; ++i) {
				if (!deleted[i]) {
					newpoints[j + newSLOffset] = points[i + slOffset];
					newpoints[j + newELOffset] = points[i + elOffset];
					++j;
				}
			}
		} else {
			for (int i = 0; i < n; ++i) {
				if (!deleted[i]) {
					newpoints[j + newSLOffset] = points[i + slOffset];
					newpoints[j + newELOffset] = points[i + elOffset];
					newpoints[j + newSHOffset] = points[i + shOffset];
					newpoints[j + newEHOffset] = points[i + ehOffset];
					++j;
				}
			}
		}
		if (j != size) {
			debug.println("j=" + j + ", size=" + size);
			debug.assertion(j != size);
		}

		// update fields
		points = newpoints;
		max = newmax;
		n = size;
		slOffset = newSLOffset;
		elOffset = newELOffset;
		shOffset = newSHOffset;
		ehOffset = newEHOffset;
		deleted = null;

		//         debug.println ("compacted " + this);
		//         dump ();
	}

	//
	// validating region array.  Sorts if necessary,
	// compacts if necessary, trims rectangles against each other. 
	//

	public void validate() {
		validateInsert();
		validateDelete();
	}

	//
	// converting to another coordinate system
	//

	public RegionSet convert(CoordinateMap coords, DocumentVersion ver) {
		if (coords == IdentityMap.IDENTITY && ver.equals(this.version))
			return this;

		try {
			validate();
			RegionArray ra = new RegionArray(this, true);

			if (shOffset == slOffset)
				for (int i = 0; i < n; ++i) {
					int s = coords.mapStart(points[slOffset + i]);
					int e = Math.max(s, coords.mapEnd(points[ehOffset + i]));
					ra.points[slOffset + i] = s;
					ra.points[ehOffset + i] = e;
				} else
				for (int i = 0; i < n; ++i) {
					ra.points[slOffset + i] =
						coords.mapStart(points[slOffset + i]);
					ra.points[shOffset + i] =
						coords.mapStart(points[shOffset + i]);
					ra.points[elOffset + i] =
						coords.mapEnd(points[elOffset + i]);
					ra.points[ehOffset + i] =
						coords.mapEnd(points[ehOffset + i]);
				}

			ra.version = ver;
			return ra;
		} catch (RepresentationChangeException e) {
			return ((RBTree) e.getRepresentation()).convert(coords, ver);
		}
	}

	public RegionSet convert(CoordinateMap coords) {
		return this.convert(coords, this.version);
	}

	public RegionSet convert(CoordinateMap coords, Document doc) {
		return this.convert(coords, doc.getVersion());
	}

	//
	// fuzzification
	//

	public int fuzzify(int x, int direction) {
		validate();

		int y = (direction == FUZZY_LOW) ? x - 1 : x;
		// search for a region containing [y,y+1]        

		if (n == 0 || y < points[slOffset])
			return x; // lies before all regions

		// find largest i such that sl[i] <= y < sl[i+1]
		int i = BinarySearch.search(points, slOffset, n, y + 1) - 1;

		// test assertion
		debug.assertion(points[slOffset + i] <= y);
		debug.assertion(i + 1 == n || y < points[slOffset + i + 1]);

		// check whether y lies in a region
		if (y >= points[elOffset + i])
			return x;

		// so [y,y+1] lies in [sl[i], el[i]]

		// return the start point or end point of the region
		return points[((direction == FUZZY_LOW) ? slOffset : elOffset) + i];
	}

	//
	// debugging
	//

	void dump() {
		for (int i = 0; i < n; ++i)
			debug.println(
				"\t["
					+ points[slOffset
					+ i]
					+ "-"
					+ points[shOffset
					+ i]
					+ ","
					+ points[elOffset
					+ i]
					+ "-"
					+ points[ehOffset
					+ i]
					+ "]");
	}

	//
	// switching between mutable and immutable reps
	//

	public RegionSet freeze() {
		validate();
		return new RegionArray(this, true);
	}

	//
	// subset enumeration
	//

	static final int NCHILDREN = 8;

	public SubsetEnumeration subsets() {
		validate();
		return n <= NCHILDREN
			? (SubsetEnumeration) super.subsets()
			: (SubsetEnumeration) new RegionArraySetEnumeration(this);
	}

	static class RegionArraySetEnumeration
		extends RegionArray
		implements SubsetEnumeration {
		RegionArray source;
		int i;

		public RegionArraySetEnumeration(RegionArray source) {
			super(source, false);
			this.source = source;
		}

		public RegionSet firstFast() {
			i = 0;
			n = (source.n + NCHILDREN - 1) / NCHILDREN;
			return this;
		}

		public RegionSet nextFast() {
			i += n;
			if (i >= source.n)
				return null;

			slOffset += n;
			elOffset += n;
			shOffset += n;
			ehOffset += n;
			n = Math.min(source.n - i, n);
			return this;
		}
	}

	//
	// region enumeration
	//

	public RegionEnumeration regions(Region bbox) {
		try {
			validate();
			return new RegionArrayEnumeration(bbox);
		} catch (RepresentationChangeException e) {
			return e.getRepresentation().regions(bbox);
		}
	}

	public RegionEnumeration regions(RegionSet bboxs) {
		try {
			validate();
			SingleDocRegionSet s =
				bboxs.getSingleDocRegionSet(this.getDocument());
			return new RegionArrayEnumeration(s.getBoundingBox());
		} catch (RepresentationChangeException e) {
			return e.getRepresentation().regions(bboxs);
		}
	}

	class RegionArrayEnumeration
		extends FuzzyRegion
		implements RegionEnumeration {
		Region bbox;
		int iCurr; // next region in the range
		int iEnd; // 1 more than the last region in the range

		public RegionArrayEnumeration(Region bbox) {
			this.bbox = bbox;
			this.version = RegionArray.this.version;
		}

		public Region first() {
			Region r = firstFast();
			return (r != null) ? (Region) r.freeze() : null;
		}

		public Region next() {
			Region r = nextFast();
			return (r != null) ? (Region) r.freeze() : null;
		}

		public Region firstFast() {
			// largest k such that sl[1..k] <= bbox.sh.
			// ==> sl[1..k] < bbox.sh+1 <= sl[k+1..n-1].
			// The BinarySearch.search() call returns k+1.
			int ksl =
				(BinarySearch
					.search(points, slOffset, n, bbox.getStartHigh() + 1)
					- 1)
					- slOffset;

			// largest k such that el[1..k] <= bbox.eh.
			// Analogous to ksl.
			int kel =
				(BinarySearch
					.search(points, elOffset, n, bbox.getEndHigh() + 1)
					- 1)
					- elOffset;

			// smallest k such that sh[k...n-1] >= bbox.sl.
			// ==> sh[1..k-1] < bbox.sl <= sh[k..n-1]
			int ksh =
				BinarySearch.search(points, shOffset, n, bbox.getStartLow())
					- shOffset;

			// smallest k such that eh[k...n-1] >= bbox.el.
			// analogous to ksh.
			int keh =
				BinarySearch.search(points, ehOffset, n, bbox.getEndLow())
					- ehOffset;

			// all regions intersecting bbox must lie in
			// [max(ksh, keh)..min(ksl, kel)].
			//             debug.println ("bbox=" + bbox
			//                            + " ==> [max(" + ksh + "," + keh + ")...min("
			//                            + ksl + "," + kel + ")]");
			iCurr = Math.max(ksh, keh);
			iEnd = Math.min(ksl, kel) + 1;

			return nextFast();
		}

		public Region nextFast() {
			if (iCurr >= iEnd)
				return null;

			RegionArrayEnumeration.this.start = points[iCurr + slOffset];
			RegionArrayEnumeration.this.end = points[iCurr + ehOffset];
			RegionArrayEnumeration.this.startHigh = points[iCurr + shOffset];
			RegionArrayEnumeration.this.endLow = points[iCurr + elOffset];
			++iCurr;

			if (RegionArrayEnumeration.this.isEmpty())
				throw new RuntimeException("empty fuzzy region in RegionArray");

			if (!bbox.mightIntersect(this))
				throw new RuntimeException(
					"bounding box " + bbox + " doesn't intersect " + this);

			return this;
		}

		public RegionSet freeze() {
			RegionSet s =
				make(
					RegionArrayEnumeration.this.getStartLow(),
					RegionArrayEnumeration.this.getStartHigh(),
					RegionArrayEnumeration.this.getEndLow(),
					RegionArrayEnumeration.this.getEndHigh(),
					RegionArrayEnumeration.this.getVersion());
			if (s.isEmpty())
				throw new RuntimeException(this +".freeze () returned EMPTY");
			return s;
		}
	}

	//
	// fast lazy implementations of a few relational operators
	//

	public RegionSet justBefore(RegionSet bgd) {
		validate();
		SingleDocRegionSet sbgd = bgd.getSingleDocRegionSet(this.getDocument());
		return new LazyRelop(this, (RegionSet) sbgd) {
			public RegionSet applyOp(Region r, RegionSet last) {
				return r.justBefore(bkgd);
			}
		};
	}

	public RegionSet justAfter(RegionSet bgd) {
		validate();
		SingleDocRegionSet sbgd = bgd.getSingleDocRegionSet(this.getDocument());
		return new LazyRelop(this, (RegionSet) sbgd) {
			public RegionSet applyOp(Region r, RegionSet last) {
				return r.justAfter(bkgd);
			}
		};
	}

	public RegionSet starting(RegionSet bgd) {
		validate();
		SingleDocRegionSet sbgd = bgd.getSingleDocRegionSet(this.getDocument());
		return new LazyRelop(this, (RegionSet) sbgd) {
			public RegionSet applyOp(Region r, RegionSet last) {
				return r.starting(bkgd);
			}
		};
	}

	public RegionSet ending(RegionSet bgd) {
		validate();
		SingleDocRegionSet sbgd = bgd.getSingleDocRegionSet(this.getDocument());
		return new LazyRelop(this, (RegionSet) sbgd) {
			public RegionSet applyOp(Region r, RegionSet last) {
				return r.ending(bkgd);
			}
		};
	}

	public RegionSet in() {
		validate();
		return new LazyRelop(this, null) {
			public RegionSet applyOp(Region r, RegionSet last) {
				return r.in();
			}
		};
	}

	public RegionSet contains() {
		validate();
		return new LazyRelop(this, null) {
			public RegionSet applyOp(Region r, RegionSet last) {
				Region s = (Region) (r.contains());

				if (last != null) {
					int newstart =
						Math.max(
							s.getStartLow(),
							((SingleDocRegionSet) last).getStartHigh() + 1);
					int newend = s.getStartHigh();
					if (newstart <= newend)
						s =
							new FuzzyRegion(
								newstart,
								newend,
								s.getEndLow(),
								s.getEndHigh(),
								s.getVersion());
				}

				return s;
			}
		};
	}

	/** debug for RegionArray manipulation
	 */
	void printPoints() {
		debug.println("points in RegionArray.java:");
		for (int i = 0; i < points.length; i++)
			System.out.print(points[i] + ", ");
		System.out.println();
	}

	public SubsetEnumeration singleDocize() {
		try {
			validate();
			return new SubsetEnumeration() {
				public RegionSet firstFast() {
					return RegionArray.this;
				}
				public RegionSet nextFast() {
					return null;
				}
			};
		} catch (RepresentationChangeException e) {
			return new RegionArraySubsetEnumeration(e.getRepresentation());
		}
	}

	private class RegionArraySubsetEnumeration implements SubsetEnumeration {
		RegionSet ms;
		RegionArraySubsetEnumeration(RegionSet ms) {
			this.ms = ms;
		}
		public RegionSet firstFast() {
			return this.ms;
		};
		public RegionSet nextFast() {
			return null;
		}
	}

	public DocumentVersion getVersion() {
		return this.version;
	}

	public Document getDocument() {
		if (this.version == null)
			return null;
		else
			return this.version.getDocument();
	}

	public SingleDocRegionSet getSingleDocRegionSet(Document doc) {

		if (this.version == null)
			return Region.EMPTY;

		if (this.version.getDocument() == doc) {
			try {
				validate();
				return this;
			} catch (RepresentationChangeException e) {
				return (SingleDocRegionSet) e.getRepresentation();
			}
		} else
			return Region.EMPTY;
	}
    
    public Versions getVersions() {
        return new SingleDocVersions (this.version);
    }
}
