/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.io;

import java.io.*;
import java.util.*;
import java.util.zip.*;

public class ZipArchive implements Serializable {

    private static final int BUFFER_SIZE = 512;

    byte[] zipdata;
        // ZIP archive

    Vector names = new Vector ();     
        // files contained in the ZIP archive

    // Used only for serialization
    static final long serialVersionUID = -9185404800237564786L;

    public ZipArchive (String[] filenames) throws IOException {
        byte[] buffer = new byte[BUFFER_SIZE];
        int n;

        ByteArrayOutputStream bout = new ByteArrayOutputStream ();
        ZipOutputStream out = new ZipOutputStream (bout);
        out.setMethod (ZipOutputStream.STORED);
        
        try {
            for (int i = 0; i < filenames.length; ++i) {
                String filename = filenames[i];
                if (filename.endsWith (File.separator)
                    && !filename.equals (File.separator))
                    filename = filename.substring (0, filename.length()-1);
                    
                addFile (out, new File (filename), "", buffer);
            }
            out.finish ();
        } finally {
            out.close ();
            zipdata = bout.toByteArray ();
        }
    }

    private void addFile (ZipOutputStream out, File f, String prefix, byte[] buffer) throws IOException {
        String name = prefix + f.getName ();
        
        if (f.isDirectory ()) {
            // Mark it as a directory for the ZipFile
            name += "/";
            
            // Create an entry for the directory
            ZipEntry entry = new ZipEntry (name);
            entry.setSize (0);
            entry.setCrc (0);
            
            // Add the entry to the ZIP archive
            out.putNextEntry (entry);            
            names.addElement (name);
            
            // Add each file in directory recursively
            String[] subfiles = f.list ();
            for (int i = 0; i < subfiles.length; ++i)
                addFile (out, new File (f, subfiles[i]), name, buffer);
        }
        else {
            // Create an entry for this file
            ZipEntry entry = new ZipEntry (name);
            entry.setSize (f.length ());

            // compute the file's CRC
            Checksum crc = new CRC32 ();
            CheckedInputStream checkedInput = new CheckedInputStream (new FileInputStream (f), crc);
            while (checkedInput.read (buffer) != -1)
                ;
            checkedInput.close ();
            entry.setCrc (crc.getValue ());

            // Add the entry to the ZIP archive
            out.putNextEntry (entry);
            names.addElement (name);
            
            // Copy the file's data into the ZIP archive
            InputStream in = new FileInputStream (f);
            int n;
            while ((n = in.read (buffer)) != -1)
                out.write (buffer, 0, n);
            in.close ();
        }
    }
    
    public ZipArchive (InputStream in) throws IOException {
        // Save the input stream to memory
        ByteArrayOutputStream out = new ByteArrayOutputStream ();
        byte[] buffer = new byte[BUFFER_SIZE];
        int n;
        while ((n = in.read (buffer)) != -1)
            out.write (buffer, 0, n);
        out.close ();
        zipdata = out.toByteArray ();

        afterCreatingFromStream ();
    }

    public ZipArchive (InputStream in, int length) throws IOException {
        // Save the input stream to memory
        ByteArrayOutputStream out = new ByteArrayOutputStream ();
        byte[] buffer = new byte[BUFFER_SIZE];
        int total = 0;
        int n;
        while (total < length && (n = in.read (buffer)) != -1) {
            out.write (buffer, 0, n);
            total += n;
        }
        out.close ();
        zipdata = out.toByteArray ();

        afterCreatingFromStream ();
    }

    private void afterCreatingFromStream () throws IOException {
        // Get the entry names out
        ZipInputStream zin = new ZipInputStream (getInputStream ());
        try {
            ZipEntry entry;
            while ((entry = zin.getNextEntry ()) != null) {
                names.addElement (entry.getName ());
                //System.err.println (entry.getName ());
                zin.closeEntry ();
            }
        } finally {
            zin.close ();
        }
    }
    
    public String[] getNames () {
        String[] fn = new String[names.size ()];
        names.copyInto (fn);
        return fn;
    }

    public String filenameAt (int i) {
        return names.elementAt (i).toString ();
    }

    public int getFilenameCount () {
        return names.size ();
    }

    public int length () {
        return zipdata.length;
    }

    public InputStream getInputStream () {
        return new ByteArrayInputStream (zipdata);
    }

    /**
     * Extract ZIP archive entries to the filesystem.
     * Each destination filename is computed by prefixing 
     * the ZIP entry name with directory.
     * @param directory Directory where files should be placed.
     * Null or empty string means current directory.  If directory
     * is non-empty and doesn't end with a file separator like / or \, 
     * then a file separator is added automatically.
     */
    public void extract (String directory) throws IOException {
        extract (directory, null, null);
    }

    public void extract (String directory, PrintStream nameStream) throws IOException {
        extract (directory, null, nameStream);
    }

    public void extract (String directory, String[] names) throws IOException {
        extract (directory, names, null);
    }

    public void extract (String directory, String[] names, PrintStream nameStream) throws IOException {
        byte[] buffer = new byte[BUFFER_SIZE];
        
        // Make sure dir has a terminating separator
        if (directory == null)
            directory = "";
        else if (directory.length () > 0
            && !(directory.endsWith ("/") ||
                 directory.endsWith (File.separator)))
            directory += File.separator;
          
        // Make a Hashtable out of the names
        Hashtable nameHash = null;
        if (names != null) {
            nameHash = new Hashtable (names.length);
            for (int i=0; i < names.length; ++i)
                nameHash.put (names[i], names[i]);        
        }

        ZipInputStream zin = new ZipInputStream (getInputStream ());
        try {
            ZipEntry entry;
            while ((entry = zin.getNextEntry ()) != null) {
                if (names == null || nameHash.contains (entry.getName ())) {
                    File f = new File (directory + entry.getName ());

                    if (nameStream != null)
                        nameStream.println (f.toString ());

                    if (entry.isDirectory ())
                        f.mkdir ();
                    else {
                        // Save the input stream to disk
                        OutputStream out = new FileOutputStream (f);
                        int length = (int)entry.getSize ();
                        int n;
                        while (length > buffer.length && (n = zin.read (buffer)) != -1) {
                            out.write (buffer, 0, n);
                            length -= n;
                        }
                        if (length > 0 && (n = zin.read (buffer, 0, length)) != -1) {
                            out.write (buffer, 0, n);
                            length -= n;
                        }

                        out.close ();
                    }
                }

                zin.closeEntry ();
            }
            
        } finally {
            zin.close ();
        }
    }
    
    public String[] extractToTempDir () throws IOException {
        return extractToTempDir (null);
    }

    public String[] extractToTempDir (String[] names) throws IOException {
        // Unpack to temporary directory
        String dir = TempFile.getTempDirectory ().toString ();
        extract (dir, names);

        if (names == null)
            names = getNames ();

        // Register all the temp files we just created (in reverse order
        // so that directories are freed when they're empty)
        for (int i = names.length - 1; i >= 0; --i)
            TempFile.registerTempFile (new File (dir + names[i]));
        
        // Return an array of just the root (top-level) names
        Vector v = new Vector ();
        for (int i = 0; i < names.length; ++i) {
            int slash = names[i].indexOf ("/");
            if (slash == -1 || slash == names[i].length()-1)
                v.addElement (dir + names[i]);
        }

        String[] extractedFilenames = new String[v.size ()];
        v.copyInto (extractedFilenames);
        return extractedFilenames;
    }
}
