/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.net;

import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

// FIX: support multipart/form-data as well
public class FormURL {
    String method;
    URL action;
    URL url;
    String queryString;
    Vector queryVector;

    /**
     * Create a query using a given method (get or post) and query data.
     * @param method "get" or "post" (case-insensitive).
     * @param url URL to send query to.  Must be http:
     * @param query Name/value pairs to simulate submitting a form.  Names and values may be arbitrary objects; this method uses toString() to convert them to strings, then URL-encodes them before sending to the web server.  Pass null if no form data is desired.  
     */
    public FormURL (String method, URL url, Vector query) {
        this (method, url, (query != null) ? encodeQuery (query) : null);
        queryVector = query;
    }

    /**
     * Open a URL using a given method (get or post) and query data.
     * @param method "get" or "post" (case-insensitive). 
     * @param url URL to open.  Must be http:
     * @param query URL-encoded name/value pairs, or null if no query desired
     */
    public FormURL (String method, URL url, String query) {
        this.method = method;
        this.action = url;
        this.url = url;
        this.queryString = (query != null) ? query : "";

        if (method == null || "get".equalsIgnoreCase (method)) {
            try {
                this.url = new URL (URLUtil.getServiceURL (url) + "?" + query);
            } catch (MalformedURLException e) {
                throw new RuntimeException (e.toString ());// shouldn't happen
            }
        }
    }

    // FIX: constructor from a String (assuming GET), or String + method

    public String getMethod () {
        return method;
    }

    public URL getAction () {
        return url;
    }

    public URL getURL () {
        return url;
    }

    public String getQueryString () {
        return queryString;
    }

    // FIX: write decodeQuery, so that this can work even when
    // FormURL is constructed from a string, not a hash
    public Vector getQueryVector () {
        return queryVector;
    }

    public String toString () {
        return url.toString ();
    }

    /**
     * Open a connection ready to submit query and retrieve its results.
     * @return URLConnection (set up for connection but not yet connected)
     */
    public URLConnection openConnection () throws IOException {
        if (!"http".equals (url.getProtocol ()))
            throw new IOException ("query URL must be http:"); 

        URLConnection conn = url.openConnection ();

        // handle POST method
        if ("post".equalsIgnoreCase (method)) {
            conn.setDoOutput (true);
            conn.setUseCaches (false);
            conn.setRequestProperty ("Content-type",
                                     "application/x-www-form-urlencoded");
            conn.setRequestProperty ("Content-length", String.valueOf (queryString.length()));

                // commence request
            PrintWriter out = null;
            try {
                out = new PrintWriter (new OutputStreamWriter
                                       (conn.getOutputStream ()));
//                 System.err.println ("Sending by POST:\n" + queryString + "\ndone");
                out.print (queryString);
                out.flush ();
            } finally {
                if (out != null)
                    out.close ();
            }
        }

        return conn;
    }

    /**
     * URL-encode a sequence of name-value pairs.
     * @param query Name/value pairs to encode; even indexes are names, odd indexes are values
     * @return String of URL-encoded name/value pairs
     */
    public static String encodeQuery (Vector query) {
        StringBuffer result = new StringBuffer ();

        for (int i = 0, n = query.size (); i < n; i+=2) {
            String name = (String)query.elementAt (i);
            String value = (String)query.elementAt (i+1);
            
            if (result.length () > 0)
                result.append ('&');
            
            result.append (ENCODER.encode (name));
            result.append ('=');
            result.append (ENCODER.encode (value));
        }
        return result.toString ();
    }

    //
    // The following craziness calls the deprecated URLEncoder.encode(string)
    // only when we have to, in Java 1.2 and 1.3.  In Java 1.4 and later,
    // we call the preferred URLEncoder.encode(string, character-encoding).
    //
    private static Encoder ENCODER;

    private interface Encoder {
        public String encode (String s);
    }

    /**
     * Wrapper for deprecated Java 1.2/1.3 URLEncoder.encode().
     */
    private static class Encoder2 implements Encoder {
        Class cls;
        Method method;
        public Encoder2 () throws Exception {
            cls = Class.forName ("java.net.URLEncoder");
            method = cls.getMethod ("encode", new Class[] {String.class});
        }            
        public String encode (String s) {
            try {
                return (String) method.invoke (cls, new Object[] {s});
            } catch (Exception e) {
                throw new RuntimeException (e.toString ());
            }
        }
    } 

    /**
     * Wrapper for preferred Java 1.4 & later URLEncoder.encode().
     */
    private static class Encoder4 implements Encoder {
        Class cls;
        Method method;
        public Encoder4 () throws Exception {
            cls = Class.forName ("java.net.URLEncoder");
            method = cls.getMethod ("encode", 
                                    new Class[] {String.class, String.class});
        }            
        public String encode (String s) {
            try {
                return (String) method.invoke (cls, new Object[] {s, "UTF-8"});
            } catch (Exception e) {
                throw new RuntimeException (e.toString ());
            }
        }
    } 

    /**
     * Determine which version of Java is running and
     * set ENCODER appropriately.
     */
    static {
        String javaVersion = System.getProperty ("java.version");
        try {
            if (javaVersion.startsWith ("1.2")
                || javaVersion.startsWith ("1.3"))
                ENCODER = new Encoder2 ();
            else
                ENCODER = new Encoder4 ();
        } catch (Exception e) {
            throw new RuntimeException (e.toString ());
        }
    }
}
