/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.net;

import java.net.*;
import java.io.*;
import java.util.*;
import lapisx.util.Exec;
import lapisx.util.Debug;

public class LocalFormURL extends FormURL {
    public static Debug debug = Debug.QUIET;

    /**
     * Create a query using a given method (get or post) and query data.
     * @param method "get" or "post" (case-insensitive).
     * @param url URL to send query to.  Must be file:
     * @param query Name/value pairs to simulate submitting a form.  Names and values may be arbitrary objects; this method uses toString() to convert them to strings, then URL-encodes them before sending to the web server.  Pass null if no form data is desired.  
     */
    public LocalFormURL (String method, URL url, Vector query) {
        super (method, url, query);
    }

    /**
     * Open a URL using a given method (get or post) and query data.
     * @param method "get" or "post" (case-insensitive). 
     * @param url URL to open.  Must be file:
     * @param query URL-encoded name/value pairs, or null if no query desired
     */
    public LocalFormURL (String method, URL url, String query) {
        super (method, url, query);
    }

    /**
     * Open a connection ready to submit query and retrieve its results.
     * @return URLConnection (set up for connection but not yet connected)
     */
    public URLConnection openConnection () throws IOException {
        String protocol = url.getProtocol ();
        if (!"file".equals (protocol))
            throw new IOException ("local form URL must be file:"); 

        CGIConnection conn = new CGIConnection (method, url);

        // handle POST method
        if ("post".equals (method)) {
            conn.setDoOutput (true);
            conn.setRequestProperty ("Content-type",
                                     "application/x-www-form-urlencoded");
            conn.setRequestProperty ("Content-length", String.valueOf (queryString.length()));
                
                // commence request
            PrintWriter out = null;
            try {
                out = new PrintWriter (new OutputStreamWriter
                                       (conn.getOutputStream ()));
                out.print (queryString);
                out.flush ();
            } finally {
                if (out != null)
                    out.close ();
            }
        }

        conn.connect ();
        return conn;
    }

    static class CGIConnection extends URLConnection {
        Process process;
        String method;
        String contentType = "text/plain";
        long contentLength = 0;

        public CGIConnection (String method, URL url) {
            super (url);
            this.method = (method != null) ? method.toUpperCase () : "GET";
        }

        public void setRequestProperty (String name, String value) {
            if ("Content-type".equalsIgnoreCase (name))
                contentType = value;
            else if ("Content-length".equalsIgnoreCase (name))
                contentLength = Long.valueOf (value).longValue ();
        }
 
        public void connect () throws IOException {
            if (connected) 
                return;

            URL serviceURL = URLUtil.getServiceURL (this.url);
            String command = serviceURL.getFile ();
            String query = URLUtil.getQuery (this.url);
            if (query.startsWith ("?"))
                query = query.substring (1);

            String[] env = {
                "GATEWAY_INTERFACE=CGI/1.1",
                "SERVER_SOFTWARE=LAPIS",
                "SERVER_NAME=localhost",
                "SERVER_PORT=0",
                "REMOTE_ADDR=127.0.0.1",
                "REMOTE_HOST=localhost",
                "REQUEST_METHOD=" + method,
                "CONTENT_LENGTH=" + contentLength,
                "CONTENT_TYPE=" + contentType,
                "SCRIPT_NAME=" + command,
                "QUERY_STRING=" + query,
                "SERVER_PROTOCOL=HTTP/1.1",
//                 "PATH_INFO=",
//                 "PATH_TRANSLATED=",
//                 "REMOTE_USER=",
//                 "AUTH_TYPE="
            };

            debug.println ("Connecting to " + this.url);
            debug.println ("Executing local CGI script '" + command + "'");
            debug.println (lapisx.util.Str.join (env, "\n"));

            process = Exec.exec (new String[] { command }, env);
            connected = true;

            // Close output stream if unneeded (to keep subprocess from reading from it)
            if (!doOutput)
                getOutputStream ().close ();
        }

        public InputStream getInputStream () throws IOException {
            connect ();
            return process.getInputStream ();
        }

        public OutputStream getOutputStream () throws IOException {
            connect ();
            return process.getOutputStream ();
        }

    }
}
