/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.net;

import java.net.*;
import java.io.*;
//import java.util.*;
import lapisx.util.Str;

public abstract class URLUtil {
    /**
     * Convert a local filename to a URL.
     * For example, if the filename is "C:\FOO\BAR\BAZ",
     * the resulting URL is "file:/C:/FOO/BAR/BAZ".
     * @param file File to convert
     * @return URL corresponding to file
     */
    public static URL FileToURL (File file) {
        try {
            String host = "";
            String path = file.getAbsolutePath ();
            path = path.replace (File.separatorChar, '/');
            path = Str.replace (path, "%", "%25");
            path = Str.replace (path, "#", "%23");
            if (!path.startsWith ("/"))
                path = "/" + path;
            else if (path.startsWith ("//"))
                host = "localhost";
            return new URL ("file", host, path);
        } catch (MalformedURLException e) {
            // shouldn't happen
            throw new RuntimeException (e.toString ());
        }
    }

    /**
     * Convert a file: URL to a filename appropriate to the
     * current system platform.  For example, on MS Windows,
     * if the URL is "file:/FOO/BAR/BAZ", the resulting
     * filename is "\FOO\BAR\BAZ".
     * @param url URL to convert
     * @return File corresponding to url
     * @exception MalformedURLException if url is not a
     * file: URL.
     */
    public static File URLToFile (URL url) throws MalformedURLException {
        if (!url.getProtocol().equals ("file"))
            throw new MalformedURLException ();

        String path = url.getFile ();
        path = path.replace ('/', File.separatorChar);
        path = Str.replace (path, "%23", "#");
        path = Str.replace (path, "%25", "%");
        // for MSWindows: change pathnames of the
        // form /X:/ to X:/
        if (path.length () > 3
            && path.charAt (0) == File.separatorChar
            && path.charAt(2) == ':'
            && path.charAt (3) == File.separatorChar)
            path = path.substring (1);

        return new File (path);
    }

    /**
     * Convert a string (representing either a filename or a local file: URL)
     * to a File. Tries to behave reasonably on common platforms
     * (especially Win32 and Unix).
     * <P>
     * Examples:
     * <UL><LI>HrefToFile ("file:/home/rcm") returns new File("/home/rcm");
     *     <LI>HrefToFile ("C:\FOO\BAR\BAZ") returns new File("C:\FOO\BAR\BAZ");
     *     <LI>HrefToFile ("http://www.yahoo.com/") throws MalformedURLException
     * @param href String to convert
     * @return File corresponding to href
     * @exception MalformedURLException if href uses a network protocol, like http: or ftp:.
     */
    public static File HrefToFile (String href) throws MalformedURLException {
        return HrefToFile (href, null);
    }

    public static File HrefToFile (String href, File curdir) throws MalformedURLException {
        File file = null;

        try {
            URL url = new URL (href);
            String protocol = url.getProtocol ();
            String host = url.getHost ();
            String path = url.getFile ();
            if (protocol.equals ("file")
                && (host == null 
                    || host.equals ("") 
                    || host.equals ("localhost"))) {
                path = path.replace ('/', File.separatorChar);
                // for MSWindows: change pathnames of the
                // form /X:/ to X:/
                if (path.length () > 3
                    && path.charAt (0) == File.separatorChar
                    && path.charAt(2) == ':'
                    && path.charAt (3) == File.separatorChar)
                    path = path.substring (1);
                file = new File (path);
            }
        } catch (MalformedURLException e) {
            file = new File (href);
            if (!file.isAbsolute () && curdir != null)
                file = new File (curdir, href);
        }

        if (file == null)
            throw new MalformedURLException ("not a filename or file: URL"); 
        return file;
    }

    /**
     * Convert a string (representing either a filename or a URL) to a URL.
     * <P>
     * Examples:
     * <UL><LI>HrefToURL ("http://www.yahoo.com/") returns new URL("http://www.yahoo.com/)";
     *     <LI>HrefToURL ("C:\FOO\BAR\BAZ") returns new URL("file:/C:/FOO/BAR/BAZ");
     * @param href String to convert
     * @return URL corresponding to href
     */
    public static URL HrefToURL (String href) {
        return HrefToURL (href, null);
    }

    /**
     * Convert a string (representing either a filename or a local URL) to a URL.
     * <P>
     * Examples:
     * <UL><LI>HrefToURL ("http://www.yahoo.com/") returns new URL("http://www.yahoo.com/)";
     *     <LI>HrefToURL ("C:\FOO\BAR\BAZ") returns new URL("file:/C:/FOO/BAR/BAZ");
     * @param href String to convert
     * @param curdir Current directory, in case href is a relative filename
     * @return URL corresponding to href
     */
    public static URL HrefToURL (String href, File curdir) {
        URL url;

        try {
            url = new URL (href);

            String protocol = url.getProtocol ();
            String host = url.getHost ();
            String path = url.getFile ();
            String ref = url.getRef ();
            if (protocol.equals ("file")
                && (host == null 
                    || host.equals ("") 
                    || host.equals ("localhost"))) {
                // add a trailing slash to local directories;
                // otherwise JDK 1.2 won't give them content-type text/html
                if (path == null || path.length () == 0)
                    path = "/";
                File file = new File (path);
                if (file.isDirectory () && !path.endsWith ("/"))
                    path += "/";
                url = new URL ("file:" + path);
                if (ref != null && ref.length () > 0)
                    url = new URL (url, "#" + ref);
            }
        }
        catch (MalformedURLException e) {
            File f = new File (href);
            if (!f.isAbsolute () && curdir != null)
                f = new File (curdir, href);
            url = FileToURL (f);
        }

        return url;
    }

    /**
     * Get the URL of a page, omitting any anchor reference (like #ref).
     * @param url URL to parse
     * @return url sans anchor reference
     */
    public static URL getPageURL (URL url) {
        String href = url.toExternalForm ();
        int i = href.indexOf ('#');
        try {
            return (i != -1) ? new URL(href.substring (0, i)) : url;
        } catch (MalformedURLException e) {
            return url;
        }
    }

    /**
     * Get the URL of a Web service, omitting any query or anchor reference.
     * @return the URL sans query and anchor reference
     */
    public static URL getServiceURL (URL url) {
        String href = url.toExternalForm ();
        int i = href.indexOf ('?');
        try {
            return (i != -1)
                ? new URL(href.substring (0, i))
                : getPageURL(url);
        } catch (MalformedURLException e) {
            return url;
        }
    }

    /**
     * Get the URL of a page's directory.
     * @param url URL to parse
     * @return url sans filename, query and anchor reference
     */
    public static URL getDirectoryURL (URL url) {
        String file = url.getFile();
        int qmark = file.indexOf ('?');
        if (qmark == -1)
            qmark = file.length();
        // find pivotal separator (between directory and filename)
        int pivot = file.lastIndexOf ('/', Math.max(qmark-1, 0));
        try {
            if (pivot == -1)
                return new URL (url, "/");
            else if (pivot == file.length()-1)
                return url;
            else
                return new URL (url, file.substring (0, pivot+1));
        } catch (MalformedURLException e) {
            return url;
        }
    }

    /**
     * Get the URL of a page's parent directory.
     * @param url URL to parse
     * @return url sans filename, query and anchor reference
     */
    public static URL getParentURL (URL url) {
        URL dirURL = getDirectoryURL (url);
        if (!dirURL.equals (url))
            return dirURL;

        String dir = dirURL.getFile ();
        int lastSlash = dir.length()-1;
        if (lastSlash == 0)
            return dirURL;

        int penultSlash = dir.lastIndexOf ('/', lastSlash-1);

        if (penultSlash == -1)
            return dirURL;

        try {
            return new URL (url, dir.substring (0, penultSlash+1));
        } catch (MalformedURLException e) {
            return dirURL;
        }
    }

    /**
     * Get the filename part of a URL, like "index.html".
     * Never contains '/'; may be the empty string.
     * @param url URL to parse
     * @return the filename portion of url
     */
    public static String getFilename (URL url) {
        String file = url.getFile ();
        String query = getQuery (url);
        int qmark = file.length() - query.length ();
        int slash = file.lastIndexOf ('/', qmark);
        return file.substring (slash+1, qmark);
    }

    /**
     * Get the query part of the link, like "?query".
     * Either starts with a '?', or is the empty string.
     * @param url URL to parse
     * @return the query portion
     */
    public static String getQuery (URL url) {
        String file = url.getFile();
        int qmark = file.indexOf ('?');
        if (qmark == -1)
            return "";
        else
            return file.substring (qmark);
    }

    /**
     * Get the basename part of a URL, which is the filename without extension:
     * e.g., "index". Never contains '/' or '.'; may be the empty string.
     * @param url URL to parse
     * @return the basename portion of url
     */
    public static String getBasename (URL url) {
        String filename = getFilename (url);
        int period = filename.indexOf ('.');
        return (period != -1) ? filename.substring (0, period) : filename;
    }

    /**
     * Get the extension part of a URL, e.g. ".html".  Either starts with
     * a period or is the empty string.
     * @param url URL to parse
     * @return the basename portion of url
     */
    public static String getExtension (URL url) {
        String filename = getFilename (url);
        int period = filename.indexOf ('.');
        return (period != -1) ? filename.substring (period) : "";
    }

    /**
     * Computes relative HREF for URL <I>there</I> when the current location
     * is URL <I>here</I>.
     */
    public static String relativeTo (URL here, URL there) {
        if (here == null)
            return there.toString();
        //System.err.println ("From: " + here);
        //System.err.println ("To:   " + there);
        if (here.getProtocol().equals (there.getProtocol())
            && here.getHost().equals (there.getHost ())
            && here.getPort() == there.getPort ()) {
            String fn = relativeTo (here.getFile (),
                                    there.getFile ());
            String ref = there.getRef ();
            return (ref != null) ? fn+ref : fn;
        }
        else {
          //System.err.println ("Use: " + there);
            return there.toString ();
        }
    }

    /**
     * Compute relative HREF for URL <i>there</i> when the current location
     * is URL <i>here</i>.
     */
    public static String relativeTo (URL here, String there) {
        if (here == null)
            return there;
      try {
        return relativeTo (here, new URL (here, there));
      } catch (MalformedURLException e) {
        return there;
      }
    }

    /**
     * Compute relative HREF for filename <there> when the current location
     * is filename <here.
     */
    public static String relativeTo (String here, String there) {
        StringBuffer result = new StringBuffer ();

        int lcp = 0;

        while (true) {
            int i = here.indexOf ('/', lcp);
            int j = there.indexOf ('/', lcp);

            if (i == -1 || i != j || !here.regionMatches (lcp, there, lcp, i-lcp))
                break;
            lcp = i+1;
        }

        // assert: first lcp characters of here and there are identical
        //         and (lcp==0 or here[lcp-1] == '/')

        // here[0..lcp-1] is the common ancestor directory of here and there

        // count hops up from here to the common ancestor
        for (int i = here.indexOf ('/', lcp);
             i != -1;
             i = here.indexOf ('/', i+1)) {
            result.append ("..");
            result.append ('/');
        }

        // append path down from common ancestor to there
        result.append (there.substring (lcp));

        //System.out.println ("Use:   " + result);
        //System.out.println ();

        return result.toString ();
    }

    /**
     * Open a file or URL.
     * @param href filename or URL to open
     * @return stream of data from open file or Web page
     */
    public static Reader open (String href) throws IOException {
        return open (HrefToURL (href));
    }

    /**
     * Open a URL.
     * @param url URL to open
     * @return stream of data from url
     */
    public static Reader open (URL url) throws IOException {
        URLConnection conn = url.openConnection ();
        InputStream in = conn.getInputStream ();
        return new InputStreamReader (in);
    }
}
