/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.progress;

import java.util.*;
//import lapisx.util.Thr;

public class BasicProgressGenerator extends ProgressEvent implements ProgressGenerator {
    boolean killThreadOnInterrupt = false;
    Vector listeners = new Vector ();
    double delta = 0.1;
    boolean inprogress = false;

    public BasicProgressGenerator (Object source) {
        this (source, false);
    }

    public BasicProgressGenerator (Object source, boolean killThreadOnInterrupt) {
        super (source);
        this.killThreadOnInterrupt = killThreadOnInterrupt;
    }

    public void addProgressListener (ProgressListener listener) {
        listeners.addElement (listener);
    }

    public void removeProgressListener (ProgressListener listener) {
        listeners.removeElement (listener);
    }

    public void fireStartProgress (String label) {
        fireStartProgress (label, 1000);
    }        

    public void fireStartProgress (String label, int max) {
        if (inprogress)
            push ();

        setLabel (label);
        setMaximum (Math.max (max, 1));
        setValue (0);

        for (Enumeration l = listeners.elements (); l.hasMoreElements (); )
            ((ProgressListener)l.nextElement ()).startProgress (this);
        inprogress = true;
    }
    
    public void fireProgress (int value) {
        if (killThreadOnInterrupt && Thread.currentThread ().isInterrupted ())
            throw new ThreadDeath ();

        if (!inprogress)
            return;

        setValue (value);
        for (Enumeration l = listeners.elements (); l.hasMoreElements (); )
            ((ProgressListener)l.nextElement ()).makeProgress (this);
    }

    public void fireAddProgress (int increment) {
        fireProgress (getValue () + increment);
    }

    public void fireProgress () {
        double delta = getDelta ();
        int value = getValue ();
        int remaining = getMaximum () - value;
        int newValue = value + (int)(delta*remaining); 
        fireProgress (newValue);
    }

    public double getDelta () {
        return delta;
    }

    public void setDelta (double delta) {
        this.delta = delta;
    }

    public void fireDoneProgress () {
        if (!inprogress)
            return;

        setValue (getMaximum ());

        for (Enumeration l = listeners.elements (); l.hasMoreElements (); )
            ((ProgressListener)l.nextElement ()).doneProgress (this);

        if (pop ())
            fireProgress (getValue ());
        else
            inprogress = false;
    }

    Stack stack;

    void push () {
        if (stack == null)
            stack = new Stack ();
        ProgressEvent pe = new ProgressEvent (getSource ());
        pe.setLabel (getLabel ());
        pe.setMinimum (getMinimum ());
        pe.setMaximum (getMaximum ());
        pe.setValue (getValue ());
        stack.push (pe);
    }

    boolean pop () {
        if (stack == null || stack.isEmpty ())
            return false;
        ProgressEvent pe = (ProgressEvent) stack.pop ();
        setLabel (pe.getLabel ());
        setMinimum (pe.getMinimum ());
        setMaximum (pe.getMaximum ());
        setValue (pe.getValue ());
        return true;
    }
}
