/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.sort;

import java.util.Vector;

public abstract class MergeSort {

    public static void sort (Sortable a) {
        Vector buffer = new Vector ();
        buffer.setSize (a.size ());
        sort (a, 0, a.size()-1, buffer);
    }

    public static void sort (Sortable a, int p, int r, Vector buffer) {
        if (p < r) {
            int q = (p + r)/2;
            sort (a, p, q, buffer);
            sort (a, q+1, r, buffer);
            merge (a, p, q, r, buffer);
        }
    }

    private static void merge (Sortable a, int p, int q, int r, Vector buffer) {
        int k = 0;
        int i = p;
        int j = q + 1;
        Object objI = a.get (i);
        Object objJ = a.get (j);
        while (true) {
            if (a.compare (objI, objJ) <= 0) {
                buffer.setElementAt (objI, k);
                ++k;
                ++i;
                if (i > q)
                    break;
                objI = a.get (i);
            } else {
                buffer.setElementAt (objJ, k);
                ++k;
                ++j;
                if (j > r)
                    break;
                objJ = a.get (j);
            }
        }
        
        while (i <= q) {
            buffer.setElementAt (a.get (i), k);
            ++k;
            ++i;
        }

        while (j <= r) {
            buffer.setElementAt (a.get (j), k);
            ++k;
            ++j;
        }

        for (i = 0; i < k; ++i)
            a.set (p + i, buffer.elementAt (i));
    }


    public static SortableList sort (SortableList l) {
        return sort (l, count (l));
    }

    public static SortableList sort (SortableList l, int n) {
        if (n < 2)
            return l;

        int half = n/2;
        SortableList l1 = l;
        SortableList l1End = traverse (l, half-1);
        SortableList l2 = l1End.next ();
        l1End.setNext (null);

        // l1 is now a list of half elements,
        // l2 is now a list of n-half elements.
        // Sort l1 and l2 separately and merge them.
        l1 = sort (l1, half);
        l2 = sort (l2, n-half);

        return merge (l1, l2);
    }

    private static SortableList merge (SortableList l1, SortableList l2) {
        SortableList header = new DummySortableList ();
        SortableList l = header;

        while (l1 != null && l2 != null) {
            if(l1.compareTo (l2) <= 0) {
                l.setNext (l1);
                l = l1;
                l1 = l1.next ();
            }
            else {
                l.setNext (l2);
                l = l2;
                l2 = l2.next ();
            }
        }

        if (l1 != null)
            l.setNext (l1);
        else if (l2 != null)
            l.setNext (l2);
        else
            l.setNext (null);

        return header.next ();
    }

    static class DummySortableList implements SortableList {
        SortableList nextPtr;
        public Object get () { return null; }
        public SortableList next () { return nextPtr; }
        public void setNext (SortableList l) { nextPtr = l; }
        public int compareTo (SortableList l) { return 0; }
    }

    private static SortableList traverse (SortableList l, int n) {
        while (n > 0) {
            l = l.next ();
            --n;
        }
        return l;
    }

    private static int count (SortableList l) {
        int n = 0;
        for (SortableList p = l; p != null; p = p.next ())
            ++n;
        return n;
    }

    public static void test (String[] args) {
        StringSortableList list = StringSortableList.make (args);
        list = (StringSortableList)MergeSort.sort (list);
        System.out.print ("list sort: ");
        if (list != null)
            list.dump ();
        System.out.println ();

        Vector v = new Vector ();
        for (int i = 0; i < args.length; ++i)
            v.addElement (args[i]);
        MergeSort.sort (new SortableVector (v));
        System.out.print ("array sort: ");
        for (int i = 0; i < v.size (); ++i)
            System.out.print (v.elementAt (i) + " ");
        System.out.println ();
        
        
    }

    public static void main (String[] args) {
        if (args.length > 0)
            test (args);
        else {
            test (new String[] {"b", "d", "a", "c"});
            test (new String[] {"a", "b", "c", "d"});
            test (new String[] {"d", "c", "b", "a"});

            test (new String[] {"a", "b", "c"});
            test (new String[] {"c", "b", "a"});

            test (new String[] {"a", "b"});
            test (new String[] {"b", "a"});

            test (new String[] {"a"});

            test (new String[] {});

            String[] random = new String[100];
            for (int i = 0; i < random.length; ++i)
                random[i] = String.valueOf ((char)('a' + 26 * Math.random ()));
            test (random);
        }
    }
}
