/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.swing;

import java.awt.event.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.undo.*;

/**
 * This class coalesces multiple typed keystrokes into a single undoable event.  
 * To use it, just replace your undo event listener attachment call
 * <tt>document.addUndoableEditListener (undoManager)</tt> with a call to
 * <tt>CoalescingUndo.addUndoableEditListener (document, undoManager)</tt>.
 */
public abstract class CoalescingUndo {

    /**
     * Attaches an undo listener to a document in such a way that
     * multiple typed keystrokes are coalesced into a single undoable edit.
     * <b>Only one undo listener may be attached to a document with this call.</b>
     * @param doc Document generating the edit events
     * @param undo Listener receiving the edit events, usually an instance of UndoManager
     */
    public static void setUndoableEditListener (Document doc, UndoableEditListener undo) {
        if (!installed)
            install ();
        doc.addUndoableEditListener (undo);
        doc.putProperty (UNDO_LISTENER, undo);
    }

    public static UndoableEditListener getUndoableEditListener (Document doc) {
        Object undoObject = doc.getProperty (UNDO_LISTENER);
        if (undoObject == null
            || !(undoObject instanceof UndoableEditListener))
            return null;
        return (UndoableEditListener) undoObject;
    }

    /*
     * Replace the default key-typed action in the default keymap with our own
     * KeyTypedAction.
     */
    protected static boolean installed = false;
    protected static void install () {
        Keymap keymap = JTextComponent.getKeymap (JTextComponent.DEFAULT_KEYMAP);
        keymap.setDefaultAction (new KeyTypedAction ());
        installed = true;
    }

    protected static final String UNDO_LISTENER = "lapisx.swing.CoalescingUndo.UndoableEditListener";

    /*
     * Action when an unmapped key is pressed.  If it's a character, we send a KeyTypedEdit edit 
     * to the undo listener which will absorb subsequent typed keystrokes.
     */
    protected static class KeyTypedAction extends DefaultEditorKit.DefaultKeyTypedAction {

        public void actionPerformed(ActionEvent e) {
            KeyTypedEdit edit = coalesce (e);
            super.actionPerformed (e);
            if (edit != null)
                edit.end ();
        }

        protected KeyTypedEdit coalesce (ActionEvent e) {
            if (e == null)
                return null;
            
            JTextComponent target = getTextComponent(e);
            if (target == null
                || !target.isEditable () 
                || !target.isEnabled ())
                return null;

            Document doc = target.getDocument ();
            if (doc == null)
                return null;

            UndoableEditListener undo = getUndoableEditListener (doc);
            if (undo == null)
                return null;

            int mod = e.getModifiers();
            if ((mod & ActionEvent.ALT_MASK) != 0
                || (mod & ActionEvent.CTRL_MASK) != 0)
                return null;
            
            String content = e.getActionCommand();
            if (content == null
                || content.length() == 0)
                return null;

            char c = content.charAt(0);
            if (c < 0x20 
                || c == 0x7F)
                return null;

            // if we're doing pending delete, take care of that first
            if (target.getSelectionStart () != target.getSelectionEnd ())
                target.replaceSelection ("");

            KeyTypedEdit edit = new KeyTypedEdit (target);
            UndoableEditEvent ue = new UndoableEditEvent (target, edit);
            undo.undoableEditHappened (ue);

            return edit;
        }
    }

    /*
     * Key typed undoable edit.  Absorbs multiple typed keystrokes until a non-KeyTypedEdit
     * edit occurs or until the caret moves.
     */
    protected static class KeyTypedEdit extends CompoundEdit implements CaretListener {
        JTextComponent editor;
        boolean caretMoved = false;

        public KeyTypedEdit (JTextComponent editor) {
            this.editor = editor;
        }

        public boolean replaceEdit (UndoableEdit edit) {
            if (edit instanceof KeyTypedEdit) {
                KeyTypedEdit te = (KeyTypedEdit) edit;
                if (!te.caretMoved) {
                    edits = te.edits;
                    editor.removeCaretListener (te);
                    return true;
                }
            }

            return super.replaceEdit (edit);
        }

        public void end () {
            editor.addCaretListener (this);
            super.end ();
        }

        public void caretUpdate (CaretEvent evt) {
            caretMoved = true;
            editor.removeCaretListener (this);
        }
    }
}

