/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.swing;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * A popup menu that appears only after the mouse is held down for 
 * a certain period of time.
 */
public class DelayedPopupMenu {
    Component frame;
    Component trigger;
    Timer timer;
    JPopupMenu menu;
    

    /**
     * Make a delayed popup menu.
     * @param frame Component in which the popup menu will appear (see
     * {@link javax.swing.JPopupMenu#show}.  This is often a JFrame.
     * @param trigger Component which should trigger the menu.
     * @param delay Delay before showing menu, in milliseconds.
     * @effects Creates a DelayedPopupMenu and attaches a mouse listener
     * to <i>trigger</i> so that when the mouse is pressed down <i>trigger</i>
     * and held for more than <i>delay</i> msec, the menu appears.
     */
    public DelayedPopupMenu (Component frame, Component trigger, int delay) {
        this.frame = frame;
        this.trigger = trigger;

        trigger.addMouseListener (new MouseAdapter () {
                public void mousePressed (MouseEvent event) {
                    timer.restart ();
                }
                public void mouseReleased (MouseEvent event) {
                    timer.stop ();

                    if (menu != null
                        && menu.isVisible ()
                        && !menu.contains 
                        (SwingUtilities.convertPoint ((Component) event.getSource (),
                                                      event.getPoint (),
                                                      menu)))
                        menu.setVisible (false);
                }
            });

        timer = new Timer (delay, new ActionListener () {
                public void actionPerformed (ActionEvent event) {
                    menu = makeMenu ();
                    showMenu (menu);
                }
            });
        timer.setRepeats (false);
    }

    /**
     * Test whether menu is popped up.
     * @return true iff menu is showing
     */
    public boolean isVisible () {
        return menu != null && menu.isVisible ();
    }

    /**
     * Callback that constructs the menu just before showing it.
     * Default implementation just creates an empty JPopupMenu.
     * A subclass should override this to populate the menu.
     * @return menu, or null to prevent popping up the menu
     */
    protected JPopupMenu makeMenu () {
        return new JPopupMenu ();
    }

    /**
     * Callback that shows the menu.
     * Default implementation positions the menu just below the trigger
     * that triggered it in the frame passed to the constructor.
     * A subclass can override this for different placement.
     * @param menu Menu about to be shown
     */
    protected void showMenu (JPopupMenu menu) {
        Point triggerLocation = 
            SwingUtilities.convertPoint (trigger.getParent (),
                                         trigger.getLocation (),
                                         frame);
        Dimension triggerSize = trigger.getSize ();
        
        int x = triggerLocation.x;
        int y = triggerLocation.y + triggerSize.height;

        menu.show (frame, x, y);
    }

    /**
     * Get the component that triggers this menu.
     * @return trigger component
     */
    public Component getTrigger () {
        return trigger;            
    }

    /**
     * Get the frame in which this menu will popup.
     * @return frame component
     */
    public Component getFrame () {
        return frame;
    }
}
