/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.swing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;


/**
 * A wrapper class for JTabbedPane that hides the tabs when there 
 * is only one tab.  Otherwise a similar interface to JTabbedPane.
 */
public class DisappearingTabbedPane extends JPanel {
    public static lapisx.util.Debug debug = lapisx.util.Debug.QUIET;

    /**
     * Map from tab component to tab title.
     */
    Map children;

    /**
     * State pattern.  One of EMPTY, ONE_TAB, or MULTIPLE_TABS.
     */
    InternalState state;

    /**
     * Panel shown for 0 or 1 tabs.
     */
    JPanel singleton;

    /**
     * Tabbed pane, shown for more than 1 tab.
     */
    JTabbedPane tabs;

    /**
     * Layout for this.  Flips between singleton (used in EMPTY and ONE_TAB
     * states) and tabs (in MULTIPLE_TABS state).
     */
    CardLayout cards;

    

    /**
     * Make a DisappearingTabbedPane with same defaults as new JTabbedPane().
     */
    public DisappearingTabbedPane () {
        this (new JTabbedPane ());
    }

    /**
     * Make a DisappearingTabbedPane with given tab placement.  Otherwise,
     * same defaults as new JTabbedPane(tabPlacement).
     *
     * @param tabPlacement  one of SwingConstants.TOP, BOTTOM, LEFT, RIGHT
     */
    public DisappearingTabbedPane (int tabPlacement) {
        this (new JTabbedPane (tabPlacement));
    }

    /**
     * Make a DisappearingTabbedPane that encapsulates the given JTabbedPane.
     *
     * @param tabs  JTabbedPane that will be added to the new 
     *              DisappearingTabbedPane
     * @requires    don't use the JTabbedPane object after putting it under
     *              DisappearingTabbedPane's control
     */
    public DisappearingTabbedPane (JTabbedPane tabs) {
        this.tabs = tabs;
       
        state = EMPTY;
        children = new HashMap ();

        setLayout (cards = new CardLayout ());

        super.add ("singleton", singleton = new JPanel ());
        singleton.setLayout (new CardLayout ());

        super.add ("tabs", tabs);
    }

    /**
     * Add a tab.
     *
     * @param title  Title shown on tab
     * @param comp   Component displayed in tab
     */
    public void addTab (String title, Component comp) {
	debug.println("attempting to add new component to Tabbed pane");
        if (children.containsKey (comp)){
	    debug.println("component exists, removing");
	    remove (comp);
	}
        state.addTab (title, comp);
	debug.println("added new component. tabs: " + getTabCount());
    }

    /** 
     * returns true if pane already exists in tabbed pane
     **/

    public boolean containsComp(Component pane) {
	return children.containsKey (pane);
    }


    /**
     * Remove a tab.
     *
     * @param comp  Component to remove
     */
    public void remove (Component comp) {
        if (!children.containsKey (comp)){
	    debug.println("component exists, removing");
            return;
	}
        state.removeTab (comp);
	debug.println("component removed. tabs: " + getTabCount());

    } 

    /**
     * Select a tab (bringing it to top).
     *
     * @param comp  Component to select
     */
    public void setSelectedComponent (Component comp) {
        state.selectTab (comp);
    }

    /**
     * Get the selected tab (returns null if there is no tab).
     * 
     * @return  component contained in the currently selected
     *          tab
     */
    public Component getSelectedComponent () {
        return state.getSelectedTab ();
    }

    /**
     * Get number of tabs.
     *
     * @return  number of tabs in this
     */
    public int getTabCount () {
        return children.size ();
    }


    /**
     * Flip between one-tab and multiple-tab display.
     *
     * @param card  "singleton" for singleton tab, "tabs" for tabbed pane
     */
    void flipTo (String card) {
        cards.show (this, card);
    }

    /**
     * State pattern.
     */
    static interface InternalState {
        void addTab (String title, Component comp);
        void removeTab (Component comp);
        void selectTab (Component comp);
	Component getSelectedTab ();
    }
    
    /**
     * State in which this contains no tabs.
     */
    final InternalState EMPTY = new InternalState () {
            public void addTab (String title, Component comp) {
                debug.assertion (children.size () == 0);
                debug.assertion (!children.containsKey (comp));

                singleton.add (title, comp);
                flipTo ("singleton");

                children.put (comp, title);

                state = ONE_TAB;
            }
            public void removeTab (Component comp) {
                debug.assertion (false); // shouldn't get here
            }
            public void selectTab (Component comp) {
                debug.assertion (false); // shouldn't get here
            }
	    public Component getSelectedTab () {
		return null;
	    }
        };

    /**
     * State in which this contains exactly one tab.
     */
    final InternalState ONE_TAB = new InternalState () {
            public void addTab (String title, Component comp) {
                debug.assertion (children.size () == 1);
                debug.assertion (!children.containsKey (comp));
                
                Component firstComp = 
                    (Component) children.keySet ().iterator ().next ();
                String firstTitle =
                    (String) children.get (firstComp);

                tabs.addTab (firstTitle, firstComp);
                tabs.addTab (title, comp);
                flipTo ("tabs");

                children.put (comp, title);
                debug.assertion (children.size () == 2);

                state = MULTIPLE_TABS;
            }
            public void removeTab (Component comp) {
                debug.assertion (children.containsKey (comp));
                children.remove (comp);
                singleton.remove (comp);
                state = EMPTY;
            }
            public void selectTab (Component comp) {
                debug.assertion (children.containsKey (comp));
                // don't have to do anything, because singleton tab is always
                // selected
            }
	    public Component getSelectedTab () {
		return tabs.getSelectedComponent();
	    }
        };
                
    /**
     * State in which this contains more than one tab.
     */
    final InternalState MULTIPLE_TABS = new InternalState () {
            public void addTab (String title, Component comp) {
                debug.assertion (children.size () > 1);
                debug.assertion (!children.containsKey (comp));

                tabs.addTab (title, comp);
                children.put (comp, title);
            }
            public void removeTab (Component comp) {
                debug.assertion (children.containsKey (comp));
                children.remove (comp);
                tabs.remove (comp);

                if (children.size () == 1) {
                    Component firstComp = 
                        (Component) children.keySet ().iterator ().next ();
                    String firstTitle =
                        (String) children.get (firstComp);

                    singleton.add (firstTitle, firstComp);
                    flipTo ("singleton");

                    state = ONE_TAB;
                }
            }
            public void selectTab (Component comp) {
                debug.assertion (children.containsKey (comp));
                tabs.setSelectedComponent (comp);
            } 
	    public Component getSelectedTab () {
		return tabs.getSelectedComponent();
	    }               
        };


    static int tabsCreated = 0;

    public static void main (String[] args) {
        final DisappearingTabbedPane tabs = new DisappearingTabbedPane ();
        final JToolBar toolbar = new JToolBar ();
        final JFrame f = new JFrame ("DisappearingTabbedPane"); 
        Container c = f.getContentPane ();
        c.add (tabs, BorderLayout.CENTER);
        c.add (toolbar, BorderLayout.SOUTH);

        toolbar.add (new AbstractAction ("Add Tab") {
                public void actionPerformed (ActionEvent event) {
                    String title = "Tab " + (tabsCreated++);
                    final JLabel comp = new JLabel (title);
                    tabs.addTab (title, comp);
                    tabs.setSelectedComponent (comp);
                    toolbar.add (new AbstractAction ("Remove " + title) {
                        public void actionPerformed (ActionEvent event) {
			    tabs.remove (comp);
			    toolbar.remove ((Component) event.getSource ());
			    f.pack ();
			}
                    });
                    f.pack ();
                }
            });
        
        f.addWindowListener (new WindowAdapter () {
            public void windowClosing (WindowEvent event) {
                System.exit(0);
            }
        });
        f.pack ();
        f.show ();
    }
            
}
