/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.swing;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.undo.*;


public class FileEditor extends JFrame {
    private String defaultTitle;
    private FileEditorPanel editorPanel;
    private UndoManager undo = new UndoManager () {
        };


    public FileEditor (String title, JTextComponent editor) {
        super (title);

        defaultTitle = title;

        Container contentPane = getContentPane ();
        contentPane.setLayout (new BorderLayout ());

	// Creat the file editor panel
	editorPanel = new FileEditorPanel(editor) {
		protected void handleInteractiveException(Exception exception) {
		    FileEditor.this.handleInteractiveException(exception);
		}

		protected FileChooser makeFileChooser() {
		    return FileEditor.this.makeFileChooser();
		}

		public void setFile (File file) {
		    editorPanel.setFile(file);
		    
		    if (file == null)
			setTitle (defaultTitle);
		    else
			setTitle (defaultTitle + " - " + file.getName ());
		}

	    };
	contentPane.add(editorPanel, BorderLayout.CENTER);

        // Create actions that depend on the editor.
        cutAction = new AcceleratedAction 
             ("Cu&t", 'X', KeyEvent.CTRL_MASK,
              getActionByName (DefaultEditorKit.cutAction));
        copyAction = new AcceleratedAction 
            ("&Copy", 'C', KeyEvent.CTRL_MASK,
             getActionByName (DefaultEditorKit.copyAction));
        pasteAction = new AcceleratedAction 
            ("&Paste", 'V', KeyEvent.CTRL_MASK,
             getActionByName (DefaultEditorKit.pasteAction));
        selectAllAction = new AcceleratedAction 
            ("Select &All", 'A', KeyEvent.CTRL_MASK,
             getActionByName (DefaultEditorKit.selectAllAction));


        // Create the menubar
        JMenuBar mb = new JMenuBar ();

        JMenu fileMenu = new AcceleratedMenu ("&File");
        fileMenu.add (newAction);
        fileMenu.add (openAction);
        fileMenu.add (saveAction);
        fileMenu.add (saveAsAction);
        fileMenu.add (closeAction);
        mb.add (fileMenu);

        JMenu editMenu = new AcceleratedMenu ("&Edit");
        editMenu.add (undoAction);
        editMenu.add (redoAction);
        editMenu.addSeparator ();
        editMenu.add (cutAction);
        editMenu.add (copyAction);
        editMenu.add (pasteAction);
        editMenu.add (selectAllAction);
        mb.add (editMenu);

        editMenu.addMenuListener (new MenuListener () {
                public void menuSelected(MenuEvent e) {
                    undoAction.setEnabled (undo.canUndo ());
                    redoAction.setEnabled (undo.canRedo ());

                    boolean selected = editorPanel.getEditorSelectionStart() < 
			editorPanel.getEditorSelectionEnd ();
                    cutAction.setEnabled (selected);
                    copyAction.setEnabled (selected);
                }
                public void menuDeselected(MenuEvent e) {
                    undoAction.setEnabled (true);
                    redoAction.setEnabled (true);
                    cutAction.setEnabled (true);
                    copyAction.setEnabled (true);
                }
                public void menuCanceled(MenuEvent e) {
                    undoAction.setEnabled (true);
                    redoAction.setEnabled (true);
                    cutAction.setEnabled (true);
                    copyAction.setEnabled (true);
                }
            });
                
        setJMenuBar (mb);


        CoalescingUndo.setUndoableEditListener(editorPanel.getEditorDocument(),
					       undo);

        pack ();
    }


    protected JTextComponent makeEditor () {
        return new JTextArea ();
    }


    protected void handleInteractiveException (Exception exception) {
        exception.printStackTrace ();
        JOptionPane.showMessageDialog (null, exception,
                                       "Error", JOptionPane.ERROR_MESSAGE);
    }


    /**
     * Action invoked by File/New command.  Clears the editor.
     */
    public Action newAction = new AcceleratedAction ("&New", 'N', 
						     KeyEvent.CTRL_MASK) {
        public void actionPerformed (ActionEvent evt) {
            if (okToClear ())
                clear ();
        }
    };
    

    /**
     * Action invoked by File/Open command.  Pops up a file-open dialog.
     */
    public Action openAction = new AcceleratedAction ("&Open...", 'O', 
						      KeyEvent.CTRL_MASK) {
        public void actionPerformed (ActionEvent evt) {
            if (okToClear ())
                open ();
        }
    };
    

    /**
     * Action invoked by File/Save command.  Saves the contents of the edito.
     */
    public Action saveAction = new AcceleratedAction ("&Save", 'S', 
						      KeyEvent.CTRL_MASK) {
        public void actionPerformed (ActionEvent evt) {
	    File file = editorPanel.getFile();

            if (file != null)
                save (file);
            else
                save ();
        }
    };
    

    /**
     * Action invoked by File/Save As command.  Pops up a save-as dialog.
     */
    public Action saveAsAction = new AcceleratedAction ("Save &As...") {
        public void actionPerformed (ActionEvent evt) {
            save ();
        }
    };
    

    /**
     * Action invoked by File/Close command.  Disposes of this editor.
     */
    public Action closeAction = new AcceleratedAction ("&Close") {
        public void actionPerformed (ActionEvent evt) {
            if (okToClear ()) {
                clear ();
                setVisible (false);
            }
        }
    };

    
    /**
     * Action invoked by Edit/Undo command.  Undoes one editing action.
     */
    public Action undoAction = new AcceleratedAction ("&Undo", 'Z', 
                                                      KeyEvent.CTRL_MASK) {
        public void actionPerformed (ActionEvent evt) {
            try {
                undo.undo ();
            } catch (CannotUndoException e) {
                getToolkit ().beep ();
            }
        }
    };


    /**
     * Action invoked by Edit/Redo command.  Redoes one editing action.
     */
    public Action redoAction = new AcceleratedAction ("&Redo", 'Y', 
                                                      KeyEvent.CTRL_MASK) {
        public void actionPerformed (ActionEvent evt) {
            try {
                undo.redo ();
            } catch (CannotRedoException e) {
                getToolkit ().beep ();
            }
        }
    };


    /**
     * Action invoked by Edit/Cut command.  Cuts current selection 
     * to clipboard.
     */
    public Action cutAction;


    /**
     * Action invoked by Edit/Copy command.  Copies current selection 
     * to clipboard.
     */
    public Action copyAction;


    /**
     * Action invoked by Edit/Paste action.  Pastes clipboard.
     */
    public Action pasteAction;


    /**
     * Action invoked by Edit/Select All command.  Selects entire editor.
     */
    public Action selectAllAction;


    /**
     * Gets original Edit/Cut, Copy, Paste, or Select-All actions 
     * from the editor.
     */
    Hashtable actions;
    Action getActionByName (String name) {
        if (actions == null) {
            actions = new Hashtable ();
            Action[] a = editorPanel.getEditor().getActions();
            for (int i = 0; i < a.length; ++i)
                actions.put (a[i].getValue (Action.NAME), a[i]);
        }
        return (Action) actions.get (name);
    }


    /**
     * Clear the editor.
     */
    public void clear () {
        editorPanel.clear();
        undo.discardAllEdits ();
    }


    /**
     * Checks whether or not the editor has unsaved content.
     */
    public boolean okToClear () {
	return editorPanel.okToClear();
    }
    

    /**
     * Pop up a file-open dialog to open a file into the editor.
     */
    public void open () {
	editorPanel.open();

    }
    

    protected FileChooser makeFileChooser () {
        return new FileChooser ();
    }


    /**
     * Open a file (or URL) in the editor.  Pops up a dialog box if
     * an error occurs.
     *
     * @param f File to open
     */
    public void open (File f) {
	editorPanel.open(f);
	undo.discardAllEdits ();
    }


    /**
     * Pop up a save-as dialog box, then saves the editor contents.
     */
    public boolean save () {
	return editorPanel.save();
    }


    /**
     * Save editor contents to a local file.
     *
     * @param f Local file to save
     */
    public boolean save (File f) {
	return editorPanel.save(f);
    }


    /**
     * Return editor contents
     *
     * @return editor contents as string
     */
    public String getText () {
	return editorPanel.getText();
    }


    /**
     * Set the editor contents.
     */
    public void setText (String text) {
	editorPanel.setText(text);
    }


    /**
     * Return the editor.
     *
     * @return editor as a JTextComponent
     */
    public JTextComponent getEditor () {
	return editorPanel.getEditor();
    }


    /**
     * Return the toolbar.
     *
     * @return toolbar as a JToolBar
     */
    public JToolBar getToolBar () {
	return editorPanel.getToolBar();
    }


    /**
     * Return the value of the dirty flag, which signals 
     * whether or not the editor's contents have been 
     * changed but not saved.
     *
     * @return the value of the dirty flag as a boolean
     */
    public boolean getDirty () {
	return editorPanel.getDirty();
    }


    /**
     * Set the dirty flag to the new boolean value.
     */
    public void setDirty (boolean dirty) {
	editorPanel.setDirty(dirty);
    }


    /**
     * Return the file from which the 
     * editor contents were loaded, or null if there 
     * was none.
     *
     * @return the name of the file from which the editor 
     *         was loaded (or null if none exists) as a 
     *         string
     */
    public File getFile () {
	return editorPanel.getFile();
    }


    public static void main (String[] args) {
        FileEditor fe = new FileEditor ("FileEditor", new JTextArea());
        fe.addWindowListener (new WindowAdapter () {
                public void windowClosing (WindowEvent event) {
                    System.exit (0);
                }
            });
        if (args.length > 0)
            fe.open (new File (args[0]));
        fe.show ();
    }

}
