/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.swing;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
//import java.util.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.undo.*;
//import javax.swing.filechooser.FileFilter;


public class FileEditorPanel extends JPanel {
    private JToolBar toolbar;
    private JTextComponent editor;
    private File file;      // file from which editor contents were
                            //  loaded (or null if no file yet)
    private boolean dirty = false;  // editor contents have been changed 
                                    // but not saved
    private UndoManager undo = new UndoManager ();


    public FileEditorPanel () {
	this(new JTextArea());
    }


    public FileEditorPanel (JTextComponent myEditor) {
        super ();

        setLayout (new BorderLayout ());

        // Set up the editor
        editor = myEditor;
        editor.setMargin (new Insets (2, 2, 2, 2));
        editor.setFont (new Font ("Monospaced", Font.PLAIN, 12));

        // Create the toolbar
        toolbar = new JToolBar (JToolBar.HORIZONTAL);
        toolbar.setFloatable (false);

        add (Layout.horizontal()
	     .add(toolbar)
	     .addGlue()
	     .done(), 
	     BorderLayout.NORTH);

        // Register dirty-flag listener
        Document doc = editor.getDocument ();
        doc.addDocumentListener (new DocumentListener () {
            public void changedUpdate (DocumentEvent event) {
            }
            public void insertUpdate (DocumentEvent event) {
                dirty = true;
            }
            public void removeUpdate (DocumentEvent event) {
                dirty = true;
            }
        });

	CoalescingUndo.setUndoableEditListener (editor.getDocument (), undo);

        // Wrap editor in a scrolling pane
        JScrollPane scrollPane = new JScrollPane (editor);
        scrollPane.setDoubleBuffered (true);
        scrollPane.setPreferredSize (new Dimension (500, 500));
        add (scrollPane, BorderLayout.CENTER);
    }


    protected void handleInteractiveException (Exception exception) {
        exception.printStackTrace ();
        JOptionPane.showMessageDialog (null, exception,
                                       "Error", JOptionPane.ERROR_MESSAGE);
    }


    /**
     * Clear the editor.
     */
    public void clear () {
        setText ("");
        dirty = false;
        setFile (null);
        undo.discardAllEdits ();
    }


    /**
     * Check whether or not the editor has unsaved content; 
     * if there is, pops up a dialog box to ask if editor 
     * contents should be saved.
     */
    public boolean okToClear () {
        if (!dirty)
            return true;

        switch (JOptionPane.showConfirmDialog (this, 
                                               "The current file has been " + 
					       "changed.  Do you want to " + 
					       "save it?",
                                               "Save First?", 
                                               JOptionPane.YES_NO_CANCEL_OPTION)) {
        case JOptionPane.YES_OPTION:
            return (file != null)
                ? save (file)
                : save ();
            
        case JOptionPane.NO_OPTION:
            return true;
           
        default:
            return false;
        }
    }

    
    /**
     * Pop up a file-open dialog to open a file into the editor.
     */
    public void open () {
        FileChooser chooser = makeFileChooser ();
            
        if (chooser.showOpenDialog (this) == JFileChooser.APPROVE_OPTION)
            open (chooser.getSelectedFile ());
    }
    

    /**
     * Make a file-chooser object.
     */
    protected FileChooser makeFileChooser () {
	return new FileChooser ();
    }


    /**
     * Open a file (or URL) in the editor.  Pops up a dialog box if
     * an error occurs.
     *
     * @param f File to open
     */
    public void open (File f) {
        try {
            Reader r = new FileReader (f);

            int n;
            StringBuffer contentBuf = new StringBuffer ();
            char[] buf = new char[256];
            while ((n = r.read (buf)) != -1)
                contentBuf.append (buf, 0, n);

            setText (contentBuf.toString ());
            dirty = false;
            editor.select (0, 0);
            setFile (f);
	    undo.discardAllEdits ();
        } catch (Exception e) {
            handleInteractiveException (e);
            dirty = false;
            clear ();
            return;
        }
    }


    /**
     * Pop up a save-as dialog box, then saves the editor contents.
     */
    public boolean save () {
        FileChooser chooser = makeFileChooser ();
        if (file != null)
            chooser.setSelectedFile (file);
        
        if (chooser.showSaveDialog (this) != JFileChooser.APPROVE_OPTION)
            return false;

        File f = chooser.getSelectedFile ();
            if (f.exists ()) 
                switch (JOptionPane.showConfirmDialog (FileEditorPanel.this, 
						       f + 
						       " already exists. " + 
						       "Overwrite?\n", 
						       "Overwrite?", 
                                                       JOptionPane.YES_NO_CANCEL_OPTION)) {
                    case JOptionPane.YES_OPTION:
			break;
                    case JOptionPane.NO_OPTION:
			// Pop up dialog box again
			return save ();
                    case JOptionPane.CANCEL_OPTION:
			return false;
                }

        return save (f);
    }


    /**
     * Save editor contents to a local file.
     *
     * @param f Local file to save
     */
    public boolean save (File f) {
        try {
            PrintWriter out = new PrintWriter (new FileWriter (f));
            out.print (editor.getText ());
            out.close ();
            dirty = false;
            setFile (f);
            return true;
        } catch (Exception e) {
            handleInteractiveException (e);
            return false;
        }
    }


    /**
     * Return editor contents.
     *
     * @return editor contents as string
     */
    public String getText () {
        return editor.getText ();
    }


    /**
     * Set the editor contents.
     */
    public void setText (String text) {
        editor.setText (text);
    }


    /**
     * Return the editor.
     *
     * @return editor as a JTextComponent
     */
    public JTextComponent getEditor () {
        return editor;    
    }


    /**
     * Return the starting index of the selection in 
     * the editor.
     */
    public int getEditorSelectionStart() {
	return editor.getSelectionStart();
    }


    /**
     * Return the ending index of the selection in 
     * the editor.
     */
    public int getEditorSelectionEnd() {
	return editor.getSelectionEnd();
    }


    /**
     * Return the document for the editor.
     */
    public Document getEditorDocument() {
	return editor.getDocument();
    }


    /**
     * Return the toolbar.
     *
     * @return toolbar as a JToolBar
     */
    public JToolBar getToolBar () {
        return toolbar;
    }


    /**
     * Return the value of the dirty flag, which signals 
     * whether or not the editor's contents have been 
     * changed but not saved.
     *
     * @return the value of the dirty flag as a boolean
     */
    public boolean getDirty () {
        return dirty;
    }


    /**
     * Set the dirty flag to the new boolean value.
     */
    public void setDirty (boolean dirty) {
        this.dirty = dirty;
    }


    /**
     * Return the file from which the 
     * editor contents were loaded, or null if there 
     * was none.
     *
     * @return the name of the file from which the editor 
     *         was loaded (or null if none exists) as a 
     *         string
     */
    public File getFile () {
        return file;
    }


    /**
     * Set the file from which the editor 
     * contents are loaded.
     */
    public void setFile (File file) {
        this.file = file;
    }


    /**
     * Add a file "menu" (actually a combo box) to the toolbar.
     */
    public void addFileMenu () {
	final JComboBox fileMenu = new JComboBox();

	fileMenu.addItem("File");
	fileMenu.addItem("  New");
	fileMenu.addItem("  Open...");
	fileMenu.addItem("  Save");
	fileMenu.addItem("  Save As...");

	Action fileMenuAction = new AbstractAction() {
	    public void actionPerformed (ActionEvent evt) {
		String selection = (String) fileMenu.getSelectedItem();

	        fileMenu.setSelectedItem("File");

		if (selection.equals("  New")) {
		    if (okToClear()) {
			clear();
		    }
		}
		else if (selection.equals("  Open...")) {
		    if (okToClear())
			open();
		}
		else if (selection.equals("  Save")) {
		    if (file != null)
			save(file);
		    else
			save();
		}
		else if (selection.equals("  Save As...")) {
		    save();
		}
	    }
	};
	fileMenu.addActionListener(fileMenuAction);
	
	toolbar.add(fileMenu);
    }

    
    /**
     * Add a Select Color button that allows the background color 
     * for the text editor to be changed.
     *
     * @effects adds a button that pops up a color chooser dialog
     *          box and sets the background color for the text 
     *          editor
     * @returns nothing
     */
    public void addColorChooser() {
	Action selectColorAction = new AcceleratedAction ("&Select Color") {
		public void actionPerformed (ActionEvent evt) {
		    Color c = JColorChooser.showDialog(FileEditorPanel.this, 
						       "Select Background Color",
						       editor.getBackground());
		    if (c != null) {
			editor.setBackground(c);
			editor.repaint();
		    }
		}
	    };

	toolbar.add(selectColorAction);
    }

}
