/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.swing;

import java.awt.*;
import java.awt.font.*;
import javax.swing.*;
//import javax.swing.text.*;
import javax.swing.text.html.*;
import javax.swing.plaf.*;
import javax.swing.plaf.metal.*;

public abstract class FontUtil {

    /**
     * Get height of a font.
     * @param f Font
     * @param c Component that will display the font.  Must be showing
     * on the screen, otherwise we can't get its Graphics context.
     * @return height of font (leading + ascent + descent)
     * @throws IllegalArgumentException if c is not showing on the screen.
     * 
     */
    public static int getHeight (Font f, Component c) {
        Graphics g = c.getGraphics ();
        if (g == null)
            throw new IllegalArgumentException ("component not visible");
        FontRenderContext frc = ((Graphics2D) g).getFontRenderContext ();
        LineMetrics lm = f.getLineMetrics ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", frc);
        return (int) Math.ceil (lm.getHeight ());
    }

    /**
     * Sets the size of a component's font.
     * @param c Component to change
     * @param size font size in points
     * @effects calls c.setFont() to change c's font to a font with the same font family and style, but
     * the given font size.
     */
    public static void resizeFont (Component c, int size) {
        Font f = c.getFont ();
        if (f == null)
            return;
        c.setFont (resizeFont (f, size));
    }

    /**
     * Enlarges or shrinks a font.
     * @param f Font to use
     * @param size new font size in points
     * @return font with same family and style as f, but the given size
     */
    public static Font resizeFont (Font f, int size) { 
        return new Font (f.getName (), f.getStyle (), size);
    }

    /**
     * Sets the size of a JEditorPane's HTML fonts.
     * @param c JEditorPane to change
     * @param size font size in points
     * @effects sets the size of body and paragraph fonts in c's HTML stylesheet
     */
    public static void resizeHTMLFont (JEditorPane c, int size) {
        HTMLEditorKit kit = (HTMLEditorKit) c.getEditorKitForContentType ("text/html");
        StyleSheet stylesheet = kit.getStyleSheet ();
        stylesheet.addRule ("body {font-size: " + size + "pt}");
        stylesheet.addRule ("p {font-size: " + size + "pt}");
    }

    /**
     * Makes a Metal theme with resized fonts.
     * @param size font size in points
     * @return MetalTheme that uses the given font size for all fonts.
     * Use {@link MetalLookAndFeel#setCurrentTheme} to install it.
     */
    public static MetalTheme makeResizedMetalTheme (final int size) {
        return new DefaultMetalTheme () {
                public String getName () { return "lapisx.swing.FontAdjustment"; }

                public FontUIResource getControlTextFont() { 
                    return new FontUIResource (resizeFont (super.getControlTextFont (), size));
                }
            
                public FontUIResource getSystemTextFont() { 
                    return new FontUIResource (resizeFont (super.getSystemTextFont (), size));
                }
            
                public FontUIResource getUserTextFont() { 
                    return new FontUIResource (resizeFont (super.getUserTextFont (), size));
                }
            
                public FontUIResource getMenuTextFont() { 
                    return new FontUIResource (resizeFont (super.getMenuTextFont (), size));
                }
            
                public FontUIResource getWindowTitleFont() { 
                    return new FontUIResource (resizeFont (super.getWindowTitleFont (), size));
                }
            };
    }

}

