/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */
package lapisx.swing;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.text.html.HTML;

// NIY: get/set selection of SELECT element (with multiple lines visible
// or multi select).  The relevant model is ListModel/ListSelectionModel.

/**
 * This class provides access to the form fields of an HTML document
 * displayed by a JEditorPane, so that an HTML form can be used for input
 * in a Java program.
 * <p>
 * Only certain types of form fields can be accessed by the get* and set*
 * methods in this class:
 * <ul>  <li> text fields (input type=text|password or textarea)
 *       <li> single-selection lists (select)
 *       <li> checkboxes (input type=checkbox)
 *       <li> radio buttons (input type=radio)
 * </ul>
 * Other form fields (such as buttons and multiple-selection lists)
 * can be accessed by calling getFieldModel() to obtain the field's
 * model object and accessing it directly.
 * <p>
 * <b>Warning:</b> form fields are not accessible until JEditorPane
 * has finished loading the HTML page.  To initialize form
 * fields, you should use a PropertyChangeListener that listens for
 * the "page" property to change:
 * <pre>
 * final JEditorPane editor = new JEditorPane (myURL);
 * editor.addPropertyChangeListener (new PropertyChangeListener () {
 *     public void propertyChange (PropertyChangeEvent event) {
 *         if ("page".equals (event.getPropertyName ()))
 *             HTMLFormAccess access = new HTMLFormAccess (editor);
 *             access.setStringValue (name, value); // etc.
 *         }
 *     });
 * </pre>
 */
public class HTMLFormAccess {
    JEditorPane editor;

    /**
     * Make an HTMLFormAccess object.
     * @param editor JEditorPane containing an HTML form to access.
     */
    public HTMLFormAccess (JEditorPane editor) {
        this.editor = editor;
    }

    /**
     * Get value of a text field or single-selection list.
     * @param name name attribute of form field to access (case-sensitive)
     * @return value of form field (empty string if it's a list with no selection).
     * @throws IllegalArgumentException if no field by that name is found
     * @throws ClassCastException if the field is found but is not a text field or single-selection list
     */
    public String getStringValue (String name) {
        Object model = getFieldModel (name);
        if (model == null)
            throw new IllegalArgumentException ("no field named '" + name + "'");

	if (model instanceof Document) {
            // for INPUT type=text and TEXTAREA
            Document field = (Document) model;
            try {
                return field.getText (0, field.getLength ());
            } catch (BadLocationException e) {
                throw new RuntimeException (e.getMessage ()); 
            }
        } else if (model instanceof ComboBoxModel) {
            // for SELECT (single selection, only 1 line visible)
            ComboBoxModel items = (ComboBoxModel)model;
            Object item = items.getSelectedItem ();
            if (item == null)
                return "";
            else
                return item.toString ();
        } else
            throw new ClassCastException ("field '" + name + "' is not a string model");
    }

    /**
     * Set value of a text field or single-selection list.
     * @param name name attribute of form field to access (case-sensitive)
     * @param value value to store in form field
     * @throws IllegalArgumentException if no field by that name is found,
     * or if the field is a single-selection list and the value is not one
     * of the choices
     * @throws ClassCastException if the field is found but is not a text field or single-selection list
     */
    public void setStringValue (String name, String value) {
        Object model = getFieldModel (name);
        if (model == null)
            throw new IllegalArgumentException ("no field named '" + name + "'");

        // Replace old contents with value
	if (model instanceof Document) {
            // for INPUT type=text and TEXTAREA
            Document field = (Document) model;
            try {
                field.remove (0, field.getLength ());
            } catch (BadLocationException e) {
            }
            
            try {
                field.insertString (0, value, null);
            } catch (BadLocationException e) {
            }
        }
        else if (model instanceof ComboBoxModel) {
            // for SELECT (single selection, only 1 line visible)
            ComboBoxModel items = (ComboBoxModel)model;
        found: {
                for (int i = 0, n = items.getSize (); i < n; ++i) {
                    Object item = items.getElementAt (i);
                    if (item != null && item.toString ().equals (value)) {
                        items.setSelectedItem (item);
                        break found;
                    }
                }
                throw new IllegalArgumentException
                    ("field '" + name + "' doesn't contain an option '" 
                     + value + "'");
            }
        } else
            throw new ClassCastException ("field '" + name 
                                          + "' is not a string");
    }

    /**
     * Gets value of a checkbox or radio button.
     * @param name name attribute of form field to access (case-sensitive)
     * @return true if form field is checked, false if not
     * @throws IllegalArgumentException if no field by that name is found
     * @throws ClassCastException if the field is found but is not a checkbox or radio button
     */
    public boolean getBooleanValue (String name) {
        Object model = getFieldModel (name);
        if (model == null)
            throw new IllegalArgumentException ("no field named '" + name + "'");

	if (model instanceof ButtonModel)
            // for INPUT=checkbox | radio
            return ((ButtonModel) model).isSelected ();
        else
            throw new ClassCastException ("field '" + name 
                                          + "' is not a boolean");
    }

    /**
     * Set value of a checkbox or radio button.
     * @param name name attribute of form field to access (case-sensitive)
     * @param value value to set in form field
     * @throws IllegalArgumentException if no field by that name is found
     * @throws ClassCastException if the field is found but is not a checkbox or radio button
     */
    public void setBooleanValue (String name, boolean value) {
        Object model = getFieldModel (name);
        if (model == null)
            throw new IllegalArgumentException ("no field named '" + name + "'");

        // Replace old contents with value
	if (model instanceof ButtonModel)
            ((ButtonModel) model).setSelected (value);
        else
            throw new ClassCastException ("field '" + name 
                                          + "' is not a boolean");
    }

    /**
     * Get the model for a form field.
     * @param name name attribute of form field to access (case-sensitive)
     * @return object representing the form field's model.  For text
     * fields, the model is a Document.  For buttons, checkboxes, and
     * radio buttons, it is a ButtonModel.  For lists, it is a
     * ListModel and a ListSelectionModel (implements both).
     * @throws IllegalArgumentException if no field by that name is found
     */
    public Object getFieldModel (String name) {
        Document doc = editor.getDocument ();

        Element e = findElement (doc.getDefaultRootElement (),
                                 HTML.Attribute.NAME, name);
        if (e == null)
            return null;

        Object model = e.getAttributes ()
                        .getAttribute(StyleConstants.ModelAttribute);
        return model;
    }

    /**
     * Search subtree rooted at element e for element with an attribute 
     * attr with the given value.
     */
    Element findElement (Element e, 
                         HTML.Attribute attr, 
                         String value) {
        if (e == null)
            return null;

        String myValue = (String)e.getAttributes ().getAttribute (attr);
        if (value.equals (myValue))
            return e;

        // Search through children
        for (int i = 0, n = e.getElementCount (); i < n; ++i) {
            Element found = findElement (e.getElement (i), attr, value);
            if (found != null)
                return found;
        }

        return null;
    }

}
