/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapisx.swing;

import javax.swing.*;
import javax.swing.event.*;
import lapisx.util.History;
import java.util.NoSuchElementException;

/**
 * An adapter for lapisx.util.History that allows a history to 
 * be used as the model for a JComboBox.
 * <P>
 * Example of use:
 * <PRE>
 * // make a history with room for 10 elements
 * HistoryComboBoxModel history = new HistoryComboBoxModel (10);
 * // make a combo box that displays the history in its drop-down list
 * JComboBox combobox = new JComboBox (history);
 * </PRE>
 */
public class HistoryComboBoxModel extends History implements ComboBoxModel {
    /**
     * Listeners on this object.
     */
    protected EventListenerList listenerList = new EventListenerList ();

    /**
     * Make a HistoryComboBoxModel.
     * @param max Maximum length of history
     */
    public HistoryComboBoxModel (int max) {
        super (max);
    }
    
    /**
     * Make a duplicate of another History.
     * @param h History to copy
     */
    public HistoryComboBoxModel (History h) {
        super (h);
    }

    ///////////
    /////////// ComboBoxModel interface
    ///////////

    public void addListDataListener(ListDataListener l) {
        listenerList.add(ListDataListener.class, l);
    }
    
    public void removeListDataListener(ListDataListener l) {
        listenerList.remove(ListDataListener.class, l);
    }

    public int getSize () {
        return size ();
    }
    
    public Object getElementAt (int index) {
        if (index < 0 || index > getSize ())
            throw new ArrayIndexOutOfBoundsException (index + " out of bounds");
        
        return history[linearToCircular (index)];
    }
    
    public Object getSelectedItem () {
        return get ();
    }
    
    public void setSelectedItem (Object obj) {
        try {
            jumpTo (obj);
        } catch (NoSuchElementException e) {
        }
    }
    
    ///////////
    /////////// Implementation
    ///////////

    protected void fireRemoved (int i, int j) {
        fire (ListDataEvent.INTERVAL_REMOVED, i, j);
    }
    
    protected void fireAdded (int i, int j) {
        fire (ListDataEvent.INTERVAL_ADDED, i, j);
    }
    
    protected void fireChanged (int i, int j) {
        fire (ListDataEvent.CONTENTS_CHANGED, i, j);
    }
    
    protected void fire (int eventType, int start, int end) {
        Object[] listeners = listenerList.getListenerList();
        ListDataEvent event = 
            new ListDataEvent(this, eventType,
                              circularToLinear (start), 
                              circularToLinear (end));
        
        for (int i = listeners.length - 2; i >= 0; i -= 2)
            if (listeners[i] == ListDataListener.class)
                switch (eventType) {
                case ListDataEvent.CONTENTS_CHANGED:
                    ((ListDataListener)listeners[i+1]).contentsChanged (event);
                    break;
                case ListDataEvent.INTERVAL_ADDED:
                    ((ListDataListener)listeners[i+1]).intervalAdded (event);
                    break;
                case ListDataEvent.INTERVAL_REMOVED:
                    ((ListDataListener)listeners[i+1]).intervalRemoved (event);
                    break;
                }
    }
    
    /**
     * Maps end-1 to 0, end-2 to 1, ..., start to L-1.
     */
    private int circularToLinear (int i) {
        return (end-1 - i + history.length) % history.length;
    }
    
    /**
     * Maps 0 to end-1, 1 to end-2, ..., L-1 to start.
     */
    private int linearToCircular (int i) {
        return (end-1 - i + history.length) % history.length;
    }
    
}

