/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.swing;

import java.awt.*;
import javax.swing.*;

public abstract class Layout {
    Container container;
    Size defaultSize = ANY_SIZE;
    Alignment defaultAlignment = CENTER;
    Insets defaultInsets = null;

    protected Layout (Container container) {
        this.container = container;
    }

    public static Layout horizontal () {
        return horizontal (new JPanel ());
    }

    public static Layout vertical () {
        return vertical (new JPanel ());
    }

    public static Layout border () {
        return border (new JPanel ());
    }

    public static Layout gridbag () {
        return gridbag (new JPanel ());
    }

    public static Layout horizontal (Container comp) {
        comp.setLayout (new BoxLayout (comp, BoxLayout.X_AXIS));
        return new BoxHandler (comp);
    }

    public static Layout vertical (Container comp) {
        comp.setLayout (new BoxLayout (comp, BoxLayout.Y_AXIS));
        return new BoxHandler (comp);
    }

    public static Layout border (Container comp) {
        comp.setLayout (new BorderLayout ());
        return new BorderHandler (comp);
    }

    public static Layout gridbag (Container comp) {
        comp.setLayout (new GridBagLayout ());
        return new GridBagHandler (comp);
    }

    public Layout add (Component comp) {
        return add (comp, defaultSize, defaultAlignment, defaultInsets);
    }

    public Layout add (Component comp, Size size) {
        return add (comp, size, defaultAlignment, defaultInsets);
    }

    public Layout add (Component comp, Alignment alignment) {
        return add (comp, defaultSize, alignment, defaultInsets);
    }

    public Layout add (Component comp, Size size, Alignment alignment) {
        return add (comp, size, alignment, defaultInsets);
    }

    public Layout add (Component comp, Insets insets) {
        return add (comp, defaultSize, defaultAlignment, insets);
    }

    public Layout add (Component comp, Size size, Insets insets) {
        return add (comp, size, defaultAlignment, insets);
    }

    public Layout add (Component comp, Alignment alignment, Insets insets) {
        return add (comp, defaultSize, alignment, insets);
    }

    public abstract Layout add (Component comp, Size size, Alignment alignment, Insets insets);

    public abstract Layout addGlue ();
    public abstract Layout nextRow ();
    public abstract Layout nextColumn ();

    public Container done () {
        return container;
    }

    public Layout setContainerInsets (int top, int left, 
                                      int bottom, int right) {
        ((JComponent)container).setBorder 
            (BorderFactory.createEmptyBorder (top, left, bottom, right));
        return this;
    }

    public Layout setInsets (int top, int left, int bottom, int right) {
        return setInsets (new Insets (top, left, bottom, right));
    }

    public Layout setInsets (Insets insets) {
        defaultInsets = insets;
        return this;
    }

    public Layout setAlignment (Alignment alignment) {
        defaultAlignment = alignment;
        return this;
    }

    public Layout setSize (Size size) {
        defaultSize = size;
        return this;
    }

    
    static class BoxHandler extends Layout {
        public BoxHandler (Container container) {
            super (container);
        }

        public Layout add (Component comp, Size size, 
                           Alignment alignment, Insets insets) {
            if (comp instanceof JComponent) {
                JComponent jc = (JComponent)comp;
                size.set (jc);
                alignment.set (jc);
                if (insets != null)
                    jc.setBorder (BorderFactory.createEmptyBorder 
                                  (insets.top, insets.left, 
                                   insets.bottom, insets.right));
            }
            container.add (comp);
            return this;
        }

        public Layout addGlue () {
            container.add (Box.createGlue ());
            return this;
        }

        public Layout nextRow () {
            throw new RuntimeException ("not supported by BoxLayout");
        }

        public Layout nextColumn () {
            throw new RuntimeException ("not supported by BoxLayout");
        }
    }

    static class GridBagHandler extends Layout {
        GridBagConstraints gbc = new GridBagConstraints ();
        static final Insets NO_INSETS = new Insets (0, 0, 0, 0);

        public GridBagHandler (Container container) {
            super (container);
            gbc.gridx = 0;
            gbc.gridy = 0;
        }

        public Layout add (Component comp, Size size, 
                           Alignment alignment, Insets insets) {
            size.set (gbc);
            alignment.set (gbc);
            gbc.insets = (insets != null) ? insets : NO_INSETS;
            container.add (comp, gbc);
            nextColumn ();
            return this;
        }

        public Layout addGlue () {
            throw new RuntimeException ("not supported by GridBagLayout");
        }

        public Layout nextRow () {
            ++gbc.gridy;
            gbc.gridx = 0;
            return this;
        }

        public Layout nextColumn () {
            ++gbc.gridx;
            return this;
        }
    }

    static class BorderHandler extends Layout {
        public BorderHandler (Container container) {
            super (container);
        }

        public Layout add (Component comp, Size size, 
                           Alignment alignment, Insets insets) {
            container.add (comp, alignment.getBorderName ());
            return this;
        }

        public Layout addGlue () {
            throw new RuntimeException ("not supported by BorderLayout");
        }

        public Layout nextRow () {
            throw new RuntimeException ("not supported by BorderLayout");
        }

        public Layout nextColumn () {
            throw new RuntimeException ("not supported by BorderLayout");
        }
    }

    public static interface Size {
        void set (JComponent c);
        void set (GridBagConstraints c);
    }
    public static final Size ANY_SIZE = new Size () {
        public void set (JComponent c) {
        }
        public void set (GridBagConstraints gbc) {
            gbc.fill = GridBagConstraints.BOTH;
            gbc.weightx = 1.0;
            gbc.weighty = 1.0;
        }
    };
    public static final Size FIXED_HEIGHT = new Size () {
        public void set (JComponent c) {
            c.setMaximumSize (new Dimension 
                              (Short.MAX_VALUE,
                               c.getPreferredSize ().height));
        }
        public void set (GridBagConstraints gbc) {
            gbc.fill = GridBagConstraints.HORIZONTAL;
            gbc.weightx = 1.0;
            gbc.weighty = 0.0;
        }
    };
    public static final Size FIXED_WIDTH = new Size () {
        public void set (JComponent c) {
            c.setMaximumSize (new Dimension 
                              (c.getPreferredSize ().width,
                               Short.MAX_VALUE));
        }
        public void set (GridBagConstraints gbc) {
            gbc.fill = GridBagConstraints.VERTICAL;
            gbc.weightx = 0.0;
            gbc.weighty = 1.0;
        }
    };
    public static final Size FIXED_SIZE = new Size () {
        public void set (JComponent c) {
            c.setMaximumSize (c.getPreferredSize ());
        }
        public void set (GridBagConstraints gbc) {
            gbc.fill = GridBagConstraints.NONE;
            gbc.weightx = 0.0;
            gbc.weighty = 0.0;
        }
    };


    public static interface Alignment {
        void set (JComponent c);
        void set (GridBagConstraints c);
        String getBorderName ();
    }
    public static final Alignment CENTER = new Alignment () {
        public void set (JComponent c) {
            c.setAlignmentX (0.5f);
            c.setAlignmentY (0.5f);
        }
        public void set (GridBagConstraints gbc) {
            gbc.anchor = GridBagConstraints.CENTER;
        }
        public String getBorderName () {
            return BorderLayout.CENTER;
        }
    };
    public static final Alignment NORTH = new Alignment () {
        public void set (JComponent c) {
            c.setAlignmentX (0.5f);
            c.setAlignmentY (0.0f);
        }
        public void set (GridBagConstraints gbc) {
            gbc.anchor = GridBagConstraints.NORTH;
        }
        public String getBorderName () {
            return BorderLayout.NORTH;
        }
    };
    public static final Alignment NORTHEAST = new Alignment () {
        public void set (JComponent c) {
            c.setAlignmentX (1.0f);
            c.setAlignmentY (0.0f);
        }
        public void set (GridBagConstraints gbc) {
            gbc.anchor = GridBagConstraints.NORTHEAST;
        }
        public String getBorderName () {
            throw new RuntimeException ("ambiguous with BorderLayout");
        }
    };
    public static final Alignment EAST = new Alignment () {
        public void set (JComponent c) {
            c.setAlignmentX (1.0f);
            c.setAlignmentY (0.5f);
        }
        public void set (GridBagConstraints gbc) {
            gbc.anchor = GridBagConstraints.EAST;
        }
        public String getBorderName () {
            return BorderLayout.EAST;
        }
    };
    public static final Alignment SOUTHEAST = new Alignment () {
        public void set (JComponent c) {
            c.setAlignmentX (1.0f);
            c.setAlignmentY (1.0f);
        }
        public void set (GridBagConstraints gbc) {
            gbc.anchor = GridBagConstraints.SOUTHEAST;
        }
        public String getBorderName () {
            throw new RuntimeException ("ambiguous with BorderLayout");
        }
    };
    public static final Alignment SOUTH = new Alignment () {
        public void set (JComponent c) {
            c.setAlignmentX (0.5f);
            c.setAlignmentY (1.0f);
        }
        public void set (GridBagConstraints gbc) {
            gbc.anchor = GridBagConstraints.SOUTH;
        }
        public String getBorderName () {
            return BorderLayout.SOUTH;
        }
    };
    public static final Alignment SOUTHWEST = new Alignment () {
        public void set (JComponent c) {
            c.setAlignmentX (0.0f);
            c.setAlignmentY (1.0f);
        }
        public void set (GridBagConstraints gbc) {
            gbc.anchor = GridBagConstraints.SOUTHWEST;
        }
        public String getBorderName () {
            throw new RuntimeException ("ambiguous with BorderLayout");
        }
    };
    public static final Alignment WEST = new Alignment () {
        public void set (JComponent c) {
            c.setAlignmentX (0.0f);
            c.setAlignmentY (0.5f);
        }
        public void set (GridBagConstraints gbc) {
            gbc.anchor = GridBagConstraints.WEST;
        }
        public String getBorderName () {
            return BorderLayout.WEST;
        }
    };
    public static final Alignment NORTHWEST = new Alignment () {
        public void set (JComponent c) {
            c.setAlignmentX (0.0f);
            c.setAlignmentY (0.0f);
        }
        public void set (GridBagConstraints gbc) {
            gbc.anchor = GridBagConstraints.NORTHWEST;
        }
        public String getBorderName () {
            throw new RuntimeException ("ambiguous with BorderLayout");
        }
    };
}
