/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

package lapisx.swing;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class OkCancelDialog extends JDialog {
    
    JPanel controlsPane;
    JButton okButton;
    JButton cancelButton;

    public OkCancelDialog (Frame owner, String title) {
        this (owner, title, true);
    }

    public OkCancelDialog (Frame owner, String title, boolean modal) {
        super (owner, title, modal);

        Container me = getContentPane ();
        me.setLayout (new BorderLayout ());
        
        me.add (controlsPane = new JPanel (), BorderLayout.CENTER);
        
        JPanel panel;
        me.add (panel = new JPanel (), BorderLayout.SOUTH);
        panel.setLayout (new FlowLayout ());
        
        JPanel subpanel;
        panel.add (subpanel = new JPanel ());
        subpanel.setLayout (new GridLayout (1, 0, 4, 0));

        subpanel.add (okButton = new JButton ("OK"));
        okButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                ok ();
            }
        });
        getRootPane ().setDefaultButton (okButton);
        
        subpanel.add (cancelButton = new JButton ("Cancel"));
        cancelButton.addActionListener (new ActionListener () {
            public void actionPerformed (ActionEvent event) {
                cancel ();
            }
        });
        
        addWindowListener (new WindowAdapter () {
            public void windowClosing (WindowEvent event) {
                cancel ();
            }
        });
    }        

    protected void setInitialPosition () {
        Component parent = getParent ();
        Dimension size = getSize();
        Dimension parentSize = (parent != null) ? parent.getSize() : Toolkit.getDefaultToolkit().getScreenSize();
        Point origin = (parent != null) ? parent.getLocationOnScreen () : new Point (0, 0);
        
        if (parentSize != null) {
            int x = Math.max (0, origin.x + (parentSize.width - size.width) / 2);
            int y = Math.max (0, origin.y + (parentSize.height - size.height) / 2);
            setLocation (x, y);
        }
    }
    
    public void show () {
        setInitialPosition ();
        registerKeystrokes ();
        super.show ();
    }
    
    public JPanel getControlsPane () {
        return controlsPane;
    }

    public JButton getOKButton () {
        return okButton;
    }

    public JButton getCancelButton () {
        return cancelButton;
    }

    void registerKeystrokes () {
        JRootPane root = getRootPane ();
        root.registerKeyboardAction(new ButtonPressAction (root, cancelButton, true),
                               KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0, false), 
                               JComponent.WHEN_IN_FOCUSED_WINDOW);
        root.registerKeyboardAction(new ButtonPressAction (root, cancelButton, false),
                               KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0, true), 
                               JComponent.WHEN_IN_FOCUSED_WINDOW);

        JButton defaultButton = root.getDefaultButton ();
        if (defaultButton != null)
            registerActionInAllTextFields (root, 
                                           new ButtonPressAction (root, defaultButton));
    }

    void registerActionInAllTextFields (Container container, Action action) {
        for (int i = 0, n = container.getComponentCount ();
             i < n;
             ++i) {
            Component comp = container.getComponent (i);
            
            if (comp instanceof TextField)
                ((TextField)comp).addActionListener (action);
            else if (comp instanceof JTextField)
                ((JTextField)comp).addActionListener (action);
            else if (comp instanceof Container)
                registerActionInAllTextFields ((Container)comp, action);
        }
    }
 
    static class ButtonPressAction extends AbstractAction {
        JButton owner;
        JRootPane root;
        boolean press;
        boolean pressAndRelease;
        ButtonPressAction(JRootPane root, JButton owner) {
            super("pressAndReleaseAction");
            this.root = root;
            this.owner = owner;
            this.press = true;
            this.pressAndRelease = true;
        }
            
        ButtonPressAction(JRootPane root, JButton owner, boolean press) {
            super(press? "pressedAction" : "releasedAction");
            this.root = root;
            this.owner = owner;
            this.press = press;
            this.pressAndRelease = false;
        }
        public void actionPerformed(ActionEvent e) {
            if (owner != null && SwingUtilities.getRootPane(owner) == root) {
                ButtonModel model = owner.getModel();
                if (pressAndRelease) {
                    model.setArmed(true);
                    model.setPressed(true);
                    model.setPressed(false);
                } else if (press) {
                    model.setArmed(true);
                    model.setPressed(true);
                } else {
                    model.setPressed(false);
                }
            }
        }
        public boolean isEnabled() {
            return owner.getModel().isEnabled();
        }
    }

    public void ok () {
        dispose ();
    }
    
    public void cancel () {
        dispose ();            
    }
}
