/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapisx.util;

import lapisx.io.NullOutputStream;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.io.PrintStream;
import java.io.OutputStream;

/**
 * Debugging output that can be enabled and disabled on a per-class
 * basis. Using Debug is a three-step process. First, declare a public
 * static (but not final) variable called debug in your class.  Its
 * default value should generally be Debug.QUIET:
 * 
 * <pre>
 *   import lapisx.util.Debug;
 * 
 *   public class MyClass {
 *      public static Debug debug = Debug.QUIET;
 * </pre>
 * 
 * Second, whenever you want to print debugging output, use the debug
 * variable instead of System.out or System.err:
 * 
 * <pre>
 *   debug.println ("this is my debugging message"); 
 *   // instead of System.err.println()
 *   ...
 * 
 *   try {
 *     ...
 *   } catch (Exception e) {
 *     debug.report (e); 
 *     // instead of e.printStackTrace()
 *     ...
 *   }
 * 
 *   debug.assertion (obj != null);
 *   // to check assertions
 * </pre>
 * 
 * Third, when you run your program, use {@link #setDebugLevel} to enable
 * and disable debugging on a class-by-class basis:
 *
 * <pre>
 *   Debug.setDebugLevel (MyClass.class, Debug.VERBOSE);
 * </pre>
 *
 * In LAPIS, you can use the -debug command-line parameter to enable
 * debugging output for the classes you want to debug:
 * 
 * <pre>
 *   lapis -debug MyClass,MyOtherClass,...
 * </pre>
 * 
 * If you have a set of related classes that are always debugged
 * simultaneously, then you may want them to share a single debug
 * variable for convenience.
 */
public abstract class Debug {
    /**
     * Enables all debugging output and assertion checks.  Assertion failures
     * throw a RuntimeException.  Output is sent to System.err.
     */
    public static final Debug VERBOSE = new Verbose ();

    /**
     * Enables only exception reports and assertion checks.
     * Disables other printing.  Assertion failures print an exception report,
     * but don't throw any exceptions, so computation can proceed.
     * Output is sent to System.err.
     */
    public static final Debug QUIET = new Quiet ();

    /**
     * Disables all debugging output and assertion checks.
     */
    public static final Debug NONE = new NoDebug ();


    /**
     * Get the debugging object currently in use by a class.
     * @param cls Class to query
     * @return Debug object found in the public static field "debug" of cls.
     * @throws NoSuchFieldException if cls has no public static field named "debug"
     */
    public static Debug getDebugLevel (Class cls)
        throws NoSuchFieldException {
        try {
            Field fld = cls.getField ("debug");
            if (fld.getType () != Debug.class
                || !Modifier.isStatic (fld.getModifiers ()))
                throw new NoSuchFieldException ();
            return (Debug) fld.get (null);
        } catch (IllegalArgumentException e) {
            throw new NoSuchFieldException ();
        } catch (IllegalAccessException e) {
            throw new NoSuchFieldException ();
        } catch (SecurityException e) {
            throw new NoSuchFieldException ();
        }
    }

    /**
     * Set the debugging object used by a class.  Call this method to enable or
     * disable the class's debugging output, e.g. 
     * <code>Debug.setDebugLevel (MyClass.class, Debug.VERBOSE)</code>.
     * @modifies cls
     * @param cls Class to modify
     * @throws NoSuchFieldException if cls has no public static field named "debug"
     */
    public static void setDebugLevel (Class cls, Debug level) 
        throws NoSuchFieldException {
        try {
            Field fld = cls.getField ("debug");
            if (fld.getType () != Debug.class
                || !Modifier.isStatic (fld.getModifiers ()))
                throw new NoSuchFieldException ();
            fld.set (null, level);
        } catch (IllegalArgumentException e) {
            throw new NoSuchFieldException ();
        } catch (IllegalAccessException e) {
            throw new NoSuchFieldException ();
        } catch (SecurityException e) {
            throw new NoSuchFieldException ();
        }
    }

    /**
     * Test whether calling print() or println() on this debug object will print anything.
     * @return true if print() prints, false if it's disabled.
     */
    public abstract boolean printEnabled ();

    /**
     * Test whether calling report() on this debug object will print anything.
     * @return true if report() prints, false if it's disabled.
     */
    public abstract boolean reportEnabled ();

    /**
     * Test whether calling assertion() on this debug object will do anything.
     * @return true if assertion() checks assertions, false if it's disabled.
     */
    public abstract boolean assertionEnabled ();

    /**
     * Print message (if this is enabled for printing)
     * @modifies this debug object's output stream (by default, System.err)
     * @param message message to print
     * @effects prints message to output stream, if this is enabled for printing
     */
    public abstract void print (String message);

    /**
     * Print message (if this is enabled for printing)
     * @modifies this debug object's output stream (by default, System.err)
     * @param message message to print
     * @effects prints message to output stream, followed by a newline, if 
     * this is enabled for printing
     */
    public abstract void println (String message);

    /**
     * Print an object (if this is enabled for printing)
     * @modifies this debug object's output stream (by default, System.err)
     * @param obj object to print
     * @effects prints obj.toString() to output stream, if this is enabled for printing
     */
    public abstract void print (Object obj);

    /**
     * Print an object (if this is enabled for printing)
     * @modifies this debug object's output stream (by default, System.err)
     * @param obj object to print
     * @effects prints obj.toString() to output stream, followed by a newline, if
     * this is enabled for printing
     */
    public abstract void println (Object obj);

    /**
     * Print the stack trace of an exception (if this is enabled for reporting exceptions)
     * @modifies this debug object's output stream (by default, System.err)
     * @param t exception
     * @effects prints t's stack trace to output stream, if this is enabled for reporting
     * exceptions.
     */
    public abstract void report (Throwable t);

    /**
     * Print a list of active threads (if this is enabled for printing)
     * @modifies this debug object's output stream (by default, System.err)
     * @effects prints list of active threads to output stream, if
     * this is enabled for printing
     */
    public abstract void printThreadInfo ();

    /**
     * Print stack trace of current execution point (if this is
     * enabled for debugging printing)
     * @modifies this debug object's output stream (by default, System.err)
     * @effects prints stack trace (starting with the call to this.printStackTrace), 
     * if this is enabled for printing
     */
    public abstract void printStackTrace ();

    /**
     * Test an assertion (if this is enabled for assertions)
     * @modifies this debug object's output stream (by default, System.err)
     * @param f result of assertion expression
     * @throws RuntimeException if assertion fails and this is Verbose.
     * @effects if assertion fails and this is Quiet, then merely reports the assertion
     * failure (in the form of a stack trace) to the output stream.
     */
    public abstract void assertion (boolean f);

    /**
     * Returns the OutputStream associated with this debug object.
     */
    public abstract OutputStream getOutputStream();
    

    /**
     * Debug object that prints all messages, exception stack traces, and checks all 
     * assertions.  Useful during test and debug.
     * <p>If you want to print to System.err, just use the singleton instance
     * Debug.VERBOSE rather than constructing a new Debug.Verbose instance.
     */
    public static class Verbose extends Debug { 
        protected PrintStream out;

        /**
         * Makes a Verbose debug object that prints to System.err.
         */
        public Verbose () {
            this (System.err);
        }

        /**
         * Makes a Verbose debug object that prints to the given stream.
         * @param out stream that will receive debugging output from this object.
         */
        public Verbose (PrintStream out) {
            this.out = out;
        }

        public boolean printEnabled () {
            return true;
        }

        public boolean reportEnabled () {
            return true;
        }

        public boolean assertionEnabled () {
            return true;
        }

        public void print (String message) {
            out.print (message);
            out.flush ();
        }

        public void println (String message) {
            out.println (message);
            out.flush ();
        }

        public void print (Object obj) {
            print (obj.toString ());
        }

        public void println (Object obj) {
            println (obj.toString ());
        }

        public void report (Throwable t) {
            t.printStackTrace (out);
            out.flush ();
        }

        public void printThreadInfo () {
            ThreadGroup g = Thread.currentThread().getThreadGroup ();
            Thread[] t = new Thread[g.activeCount ()];
            g.enumerate (t);
            out.println ("Active threads in " + g);
            for (int i=0; i<t.length; ++i)
                out.println (t[i]);
            out.flush ();
        }

        public void printStackTrace () {
            try {
                throw new Exception ();
            } catch (Exception e) {
                e.printStackTrace (out);
                out.flush ();
            }
        }

        public void assertion (boolean f) {
            if (!f)
                throw new RuntimeException ("assertion failure");
        }

	public OutputStream getOutputStream() {
	    return out;
	}

    }

    /**
     * Debug object that only prints exception stack traces and
     * assertion failures.  No other debug messages are printed.
     * Assertion failures only print a report, without throwing an
     * exception. This object is a good default debug level, even for
     * released code.
     * 
     * <p>If you want to print to System.err, just use the singleton
     * instance Debug.QUIET rather than constructing a new Debug.Quiet
     * instance.
     */
    public static class Quiet extends Verbose { 
        /**
         * Makes a Quiet debug object that sends exception reports to System.err.
         */
        public Quiet () {
        }

        /**
         * Makes a Quiet debug object that sends exception reports to the given stream.
         * @param out stream to receive exception reports.
         */
        public Quiet (PrintStream out) {
            super (out);
        }

        public boolean printEnabled () {
            return false;
        }

        public boolean reportEnabled () {
            return true;
        }

        public boolean assertionEnabled () {
            return true;
        }

        public void print (String message) {
        }
        public void println (String message) {
        }
        public void print (Object message) {
        }
        public void println (Object message) {
        }
        public void report (Throwable t) {
            t.printStackTrace (out);
            out.flush ();
        }
        public void printThreadInfo () {
        }
        public void printStackTrace () {
        }
        public void assertion (boolean f) {
            if (!f) {
                try {
                    throw new RuntimeException ("assertion failure");
                } catch (RuntimeException e) {
                    e.printStackTrace (out);
                    out.flush ();
                }
            }
        }
    }

    /**
     * Debug object that prints nothing and checks no assertions.
     * Rarely useful.
     *
     * <p>It is preferable to use the singleton instance Debug.NONE
     * rather than constructing new Debug.NoDebug instances.  The
     * Debug.NoDebug class is made public in case someone wants to
     * extend it.
     */
    public static class NoDebug extends Debug { 
	protected NullOutputStream out = NullOutputStream.getInstance();

        public boolean printEnabled () {
            return false;
        }

        public boolean reportEnabled () {
            return false;
        }

        public boolean assertionEnabled () {
            return false;
        }

        public void print (String message) {
        }
        public void println (String message) {
        }
        public void print (Object message) {
        }
        public void println (Object message) {
        }
        public void report (Throwable t) {
        }
        public void printThreadInfo () {
        }
        public void printStackTrace () {
        }
        public void assertion (boolean f) {
        }

	public OutputStream getOutputStream() {
	    return out;
	}
    }

}
