/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */
package lapisx.util;

import junit.framework.*;

import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * A collection of JUnit testcases to test the History class.
 */
public class HistoryTest extends TestCase {

    protected History h0;
    protected History h1;
    protected History h2;

    public HistoryTest (String name) {
        super(name);
    }

    protected void setUp () {
        h0 = new History (0);
        h1 = new History (1);
        h2 = new History (2);
    }        
    
    public static Test suite () {        
        return new TestSuite (HistoryTest.class);
    }

    /*
      Tests

      creating history of size 0, 1, n
      clearing an empty history, a full history

      events
    */
    public void testAdd () {
        h0.add ("1");
        assertEnum (new String[] {}, h0.elements());

        h1.clear ();
        h1.add ("1");
        assertEnum (new String[] {"1"}, h1.elements ());

        h1.add ("2");
        assertEnum (new String[] {"2"}, h1.elements ());

        h2.clear ();
        h2.add ("1");
        h2.add ("2");
        h2.add ("3");
        h2.add ("4");
        assertEnum (new String[] {"3","4"}, h2.elements ());
    }

    public void testClear () {
        h0.clear ();
        assertEnum (new String[] {}, h0.elements ());

        h2.add ("1");
        h2.add ("2");
        h2.clear ();
        assertEnum (new String[] {}, h2.elements ());

        h2.clear ();
        assertEnum (new String[] {}, h2.elements ());
    }

    public void testCopy () {
        History h0Dup = new History (h0);
        assertEnum (new String[] {}, h0Dup.elements ());

        h2.clear ();
        h2.add ("1");
        h2.add ("2");
        h2.add ("3");
        h2.add ("4");
        History h2Dup = new History (h2);
        assertEnum (new String[] {"3","4"}, h2Dup.elements ());
    }

    public void testPut () {
        h1.clear ();
        h1.add ("1");
        h1.put ("2");
        assertEnum (new String[] {"2"}, h1.elements ());

        h2.clear ();
        h2.add ("1");
        h2.put ("2");
        assertEnum (new String[] {"1","2"}, h2.elements ());
        h2.put ("3");
        assertEnum (new String[] {"2","3"}, h2.elements ());
        h2.put ("4");
        assertEnum (new String[] {"3","4"}, h2.elements ());
    }

    public void testForwardAndBack () {
        assertNull (h0.get ());
        assertNull (h0.peekForward ());
        assertFalse (h0.canForward ());
        assertNull (h0.peekBack ());
        assertFalse (h0.canBack ());
        assertEnum (new Object[] {}, h0.elements ());
        assertEnum (new Object[] {}, h0.forwardElements ());
        assertEnum (new Object[] {}, h0.backElements ());

        h2.clear ();
        assertNull (h2.get ());
        assertNull (h2.peekForward ());
        assertFalse (h2.canForward ());
        assertNull (h2.peekBack ());
        assertFalse (h2.canBack ());
        assertEnum (new Object[] {}, h2.elements ());
        assertEnum (new Object[] {}, h2.forwardElements ());
        assertEnum (new Object[] {}, h2.backElements ());

        h2.add ("1");
        assertEquals ("1", h2.get ());
        assertNull (h2.peekForward ());
        assertFalse (h2.canForward ());
        assertNull (h2.peekBack ());
        assertFalse (h2.canBack ());
        assertEnum (new Object[] {"1"}, h2.elements ());
        assertEnum (new Object[] {}, h2.forwardElements ());
        assertEnum (new Object[] {}, h2.backElements ());

        h2.add ("2");
        assertEquals ("2", h2.get ());
        assertEquals ("1", h2.peekBack ());
        assertTrue (h2.canBack ());
        assertNull (h2.peekForward ());
        assertFalse (h2.canForward ());
        assertEnum (new Object[] {"1", "2"}, h2.elements ());
        assertEnum (new Object[] {}, h2.forwardElements ());
        assertEnum (new Object[] {"1"}, h2.backElements ());
        
        assertEquals ("1", h2.back ());
        assertEquals ("1", h2.get ());
        assertNull (h2.peekBack ());
        assertFalse (h2.canBack ());
        assertEquals ("2", h2.peekForward ());
        assertTrue (h2.canForward ());
        assertEnum (new Object[] {"1", "2"}, h2.elements ());
        assertEnum (new Object[] {"2"}, h2.forwardElements ());
        assertEnum (new Object[] {}, h2.backElements ());

        h2.add ("3");
        assertEquals ("3", h2.get ());
        assertEquals ("2", h2.peekBack ());
        assertNull (h2.peekForward ());
        assertEnum (new Object[] {"2","3"}, h2.elements ());
        assertEnum (new Object[] {}, h2.forwardElements ());
        assertEnum (new Object[] {"2"}, h2.backElements ());

        assertEquals ("2", h2.back ());
        assertNull (h2.back ());
        assertEquals ("2", h2.get ());
        assertEnum (new Object[] {"2","3"}, h2.elements ());
        assertEnum (new Object[] {"3"}, h2.forwardElements ());
        assertEnum (new Object[] {}, h2.backElements ());

        assertEquals ("3", h2.forward ());
        assertNull (h2.forward ());
        assertEquals ("3", h2.get ());
        assertEnum (new Object[] {"2","3"}, h2.elements ());
        assertEnum (new Object[] {}, h2.forwardElements ());
        assertEnum (new Object[] {"2"}, h2.backElements ());
    }

    public void testReplace () {
        h2.clear ();
        h2.replace ("1");
        assertEnum (new String[] {"1"}, h2.elements ());

        h2.replace ("2");
        assertEnum (new String[] {"2"}, h2.elements ());

        h2.add ("3");
        h2.replace ("4");
        assertEnum (new String[] {"2","4"}, h2.elements ());

        h2.back ();
        h2.replace ("5");
        assertEnum (new String[] {"5","4"}, h2.elements ());
    }

    public void testToStartAndToEnd () {
        h2.clear ();
        assertNull (h2.toStart ());
        assertNull (h2.toEnd ());

        h2.add ("1");
        h2.add ("2");
        h2.add ("3");

        h2.toStart ();
        assertEquals ("2", h2.get ());

        h2.toEnd ();
        assertEquals ("3", h2.get ());
    }

    public void testSize () {
        assertTrue (h0.isEmpty ());
        assertTrue (h0.isFull ());
        assertEquals (0, h0.size ());

        h1.clear ();
        assertTrue (h1.isEmpty ());
        assertFalse (h1.isFull ());
        assertEquals (0, h1.size ());

        h1.add ("1");
        assertFalse (h1.isEmpty ());
        assertTrue (h1.isFull ());
        assertEquals (1, h1.size ());

        h2.clear ();
        h2.add ("1");
        assertFalse (h2.isEmpty ());
        assertFalse (h2.isFull ());
        assertEquals (1, h2.size ());

        h2.add ("2");
        assertFalse (h2.isEmpty ());
        assertTrue (h2.isFull ());
        assertEquals (2, h2.size ());

        h2.add ("3");
        assertFalse (h2.isEmpty ());
        assertTrue (h2.isFull ());
        assertEquals (2, h2.size ());
    }

    public void testContains () {
        assertFalse (h0.contains ("1"));

        h1.add ("1");
        assertTrue (h1.contains ("1"));

        h2.clear ();
        h2.add ("1");
        h2.add ("2");
        h2.add ("3");
        assertFalse (h2.contains ("1"));
        assertTrue (h2.contains ("2"));
        assertTrue (h2.contains ("3"));

        h2.clear ();
        assertFalse (h2.contains ("2"));
        assertFalse (h2.contains ("3"));
    }

    public void testExpand () {
        History h0 = new History (0);
        History h2 = new History (2);

        // expand shouldn't affect h0
        h0.expand ();
        h0.add ("1");
        assertEnum (new String[]{}, h0.elements ());

        // but it should double h2
        h2.add ("1");
        h2.add ("2");
        h2.add ("3");
        assertEnum (new String[]{"2","3"}, h2.elements ());
        h2.expand ();
        assertEnum (new String[]{"2","3"}, h2.elements ());
        h2.add ("4");
        h2.add ("5");
        h2.add ("6");
        assertEnum (new String[]{"3","4","5","6"}, h2.elements ());

        // double h2 again
        h2.clear ();
        h2.expand ();
        h2.add ("1");
        h2.add ("2");
        h2.add ("3");
        h2.add ("4");
        h2.add ("5");
        h2.add ("6");
        h2.add ("7");
        h2.add ("8");
        h2.add ("9");
        assertEnum (new String[]{"2","3","4","5","6","7","8","9"}, h2.elements ());
    }

    public void testJumpTo () {
        try {
            h0.jumpTo ("a");
            fail ("didn't throw exception");
        } catch (NoSuchElementException e) {}

        h1.clear ();
        try {
            h1.jumpTo ("a");
            fail ("didn't throw exception");
        } catch (NoSuchElementException e) {}

        h2.clear ();
        h2.add ("a");
        h2.add ("b");
        h2.jumpTo ("a");
        assertEquals ("a", h2.get ());

        h2.add ("c");
        try {
            h2.jumpTo ("a");
            fail ("didn't throw exception");
        } catch (NoSuchElementException e) {}

        h2.jumpTo ("c");
        assertEquals ("c", h2.get ());
    }

    /**
     * Compare an enumeration with an expected sequence of objects.
     */
    private void assertEnum (Object[] expected, Enumeration actual) {
        int i = 0;
        while (actual.hasMoreElements ()) {
            assertTrue (i < expected.length);
            assertEquals (expected[i], actual.nextElement ());
            ++i;
        }
        assertTrue (i == expected.length);
    }
}

