/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */


package lapisx.util;

import java.util.*;

/** 
 * This class implements the Set interface backed by a hash table.
 * Iterators returned by a SafeIteratorSet iterate over an
 * unmodifiable snapshot of the set, so it's safe to call add and
 * remove on the original set even while iterators are running.
 * Useful for storing a collection of event listeners.
 */
public class SafeIteratorSet implements Set {
    public static Debug debug = Debug.QUIET;

    HashSet set;
    int refcount = 0;  // number of iterators currently using set

    /**
     * Make an empty SafeIteratorSet.
     */
    public SafeIteratorSet () {
        set = new HashSet ();
    }

    /**
     * Make a SafeIteratorSet initialized with the elements of a collection.
     * @param c Initial members of the set
     */
    public SafeIteratorSet (Collection c) {
        set = new HashSet (c);
    }

    /**
     * Make a SafeIteratorSet with an initial capacity.
     * @param initialCapacity Initial capacity allocated to hash table
     */
    public SafeIteratorSet (int initialCapacity) {
        set = new HashSet (initialCapacity);
    }

    /**
     * Make a SafeIteratorSet with an initial capacity and load factor.
     * @param initialCapacity Initial capacity allocated to hash table
     * @param loadFactor Load factor for hash table.
     */
    public SafeIteratorSet (int initialCapacity, float loadFactor) {
        set = new HashSet (initialCapacity, loadFactor);
    }

    /**
     * Copy this set if any iterators are currently using it.
     * @effects If any iterators are in operation on this set, 
     * make a fresh copy of the set so that mutations do not affect
     * them.
     */
    private void copyOnWrite () {
        if (refcount > 0) {
            debug.println ("copying set, refcount=" + refcount);
            set = new HashSet (set);
            refcount = 0;
        }
    }

    public boolean add (Object o) {
        copyOnWrite ();
        return set.add (o);
    }

    public boolean addAll (Collection c) {
        copyOnWrite ();
        return set.addAll (c);
    }

    public void clear () {
        copyOnWrite ();
        set.clear ();
    }

    public boolean contains (Object o) {
        return set.contains (o);
    }

    public boolean containsAll (Collection c) {
        return set.containsAll (c);
    }

    public boolean equals (Object o) {
        return set.equals (o);
    }

    public int hashCode () {
        return set.hashCode ();
    }

    public boolean isEmpty () {
        return set.isEmpty ();
    }

    /**
     * Return an iterator over the elements in this set.  The iterator
     * operates over an unmodifiable snapshot of this set's elements,
     * so the set may be mutated without harm while the iterator is
     * operating.
     * @returns iterator over unmodifiable snapshot of this set
     */
    public Iterator iterator () {
        return new SafeIterator ();
    }

    private class SafeIterator implements Iterator {
        Set s;
        Iterator g;
        public SafeIterator () {
            g = set.iterator ();
            acquire ();
        }

        public boolean hasNext () { 
            if (released ()) 
                return false;

            boolean f = g.hasNext ();
            if (!f)
                release ();
            
            return f;
        }

        public Object next () {
            if (released ()) 
                throw new NoSuchElementException ();
            else
                try {
                    return g.next ();
                } catch (NoSuchElementException e) {
                    release ();
                    throw e;
                }
        }

        public void remove () {
            throw new UnsupportedOperationException ("Iterator.remove() not supported by SafeIteratorSet");
        }

        private void acquire () {
            s = set;
            ++refcount;
        }

        private void release () {
            if (s == set) {
                --refcount;
                debug.assertion (refcount >= 0);
                s = null;
            }
        }
        
        private boolean released () {
            return s == null;
        }
    }

    public boolean remove (Object o) {
        copyOnWrite ();
        return set.remove (o);
    }

    public boolean removeAll (Collection c) {
        copyOnWrite ();
        return set.removeAll (c);
    }

    public boolean retainAll (Collection c) {
        copyOnWrite ();
        return set.retainAll (c);
    }

    public int size () {
        return set.size ();
    }

    public Object[] toArray () {
        return set.toArray ();
    }

    public Object[] toArray (Object[] a) {
        return set.toArray (a);
    }

}
