/*
 * LAPIS lightweight structured text processing system
 *
 * Copyright (C) 1998-2002 Carnegie Mellon University,
 * Copyright (C) 2003 Massachusetts Institute of Technology.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation, version 2.
 *
 * LAPIS homepage: http://graphics.lcs.mit.edu/lapis/
 */

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <io.h>

#define JAR_FILES "../lib/*.jar"

static char *classpathEntries[] = {
    "../src", 
    "../lib",
    NULL
};

static char *pathEntries[] = {
    "../bin",
    "../lib",
    NULL
};

static char *javaExe = 
#ifdef CONSOLE
    "java";
#else
    "javag";
#endif

static char *fixedJavaArgs[] = {
    "-mx64M",
    "-Dlapis.dir=",
    "lapis.Main",
    NULL
};

static char *jviewExe = "jview";
static char *fixedJviewArgs[] = {
    "/d:lapis.dir=",
    "lapis.Main",
    NULL
};


// Get directory containing this executable.
char *GetAppDir () 
{
  static char appDir[2048];

  if (!*appDir) {
    int len = GetModuleFileName (NULL, appDir, sizeof(appDir));
    
    // Strip off everything after the last slash,
    // leaving only the directory 
    char *s = strrchr (appDir, '\\');
    if (s)
      s[1] = '\0';
    else
      appDir[0] = '\0';

#if DEBUG
    fprintf (stderr, "appdir = %s\n", appDir);
    fflush (stderr);
#endif
  }
  
  return appDir;
}

// convert forward slashes to backslashes
void ForwardToBackSlashes (char *s) {
    while ((s = strchr (s, '/')) != NULL)
	*s = '\\';
}

// convert backslashes to forward slashes
void BackToForwardSlashes (char *s) {
    while ((s = strchr (s, '\\')) != NULL)
	*s = '/';
}

char *AppendString (char *s, char *t) {
    char *result = malloc (strlen (s) + strlen (t) + 1);
    strcpy (result, s);
    strcat (result, t);
    return result;
}


int CalculatePathSpace (char *prefix, char **entries) {
  int size = 0;
  int prefixLen = strlen (prefix);
  while (*entries) {
      size += prefixLen // directory prefix
          + strlen (*entries) // entry itself
          + 1 // semicolon
          ;
    ++entries;
  }
  return size;
}

void AppendPathEntries (char *result, char *prefix, char **entries) {
  while (*entries) {
    strcat (result, prefix);
    strcat (result, *entries);
    strcat (result, ";");
    ++entries;
  }
}

void ExtendPath (char *name, char **entries)
{
  char *appdir = GetAppDir ();
  char *currpath = getenv (name);
  char *result;
  int i;

  if (currpath == NULL)
    currpath = "";

  for (i = 0; entries[i]; ++i)
      ForwardToBackSlashes (entries[i]);

  result = malloc (strlen (name) 
                      + 1 /* equal sign */
		      + CalculatePathSpace (appdir, entries)
		      + strlen (currpath));
  strcpy (result, name);
  strcat (result, "=");
  AppendPathEntries (result, appdir, entries);
  strcat (result, currpath);

  putenv (result);
#if DEBUG
  fprintf (stderr, "extended %s to %s\n", name, result);
  fflush (stderr);
#endif

  free (result);
}

char **FindFiles (char *filespec) {
  char *appdir = GetAppDir ();
  char *truefilespec = AppendString (appdir, filespec);
  char **files;
  int maxFiles;
  int nFiles;
  char *dir;
  char *s;
  long hf;
  struct _finddata_t f;
  int i;
  
  ForwardToBackSlashes (filespec);
  ForwardToBackSlashes (truefilespec);
  dir = strdup (filespec);
  s = strrchr (dir, '\\');
  if (s)
    s[1] = '\0';
  else
    dir[0] = '\0';

  maxFiles = 10;
  files = malloc (sizeof(*files) * maxFiles);
  nFiles = 0;

  hf = _findfirst (truefilespec, &f);
  if (hf != -1) {
    do {
      int i = nFiles;

      if (++nFiles >= maxFiles) {
	maxFiles *= 2;
	files = realloc (files, sizeof (*files) * maxFiles);
      }

      files[i] = malloc (strlen (dir) + strlen (f.name));
      strcpy (files[i], dir);
      strcat (files[i], f.name);
#if DEBUG
      fprintf (stderr, "found: %s\n", files[i]);
#endif
    } while (_findnext (hf, &f) == 0);
    _findclose (hf);
  }

  files[nFiles] = NULL;
  return files;
}

int CountArgs (char **args) {
  int n = 0;
  while (*args) {
    ++n;
    ++args;
  }
  return n;
}

char **MakeArgs (char *exename, char **fixedArgs, char **variableArgs) 
{
  int n = 2 + CountArgs (fixedArgs) + CountArgs (variableArgs);
  char **args = malloc (sizeof(char*) * n);
  int i = 0;
  int j;
  args[i++] = exename;

  for (j = 0; fixedArgs[j]; ++j) {
      char *arg = fixedArgs[j];
      if (strstr (arg, "lapis.dir") != NULL) {
          char *lapisdir = AppendString ("\"", AppendString (GetAppDir (), "\"\\.."));
	  arg = AppendString (arg, lapisdir);
	  BackToForwardSlashes (arg);
      }

    args[i++] = arg;
  }

  for (j = 0; variableArgs[j]; ++j)
    args[i++] = variableArgs[j];
  args[i++] = NULL;
  return args;
}

int RunJava (char **javaArgs, char **jviewArgs)
{
  int result;
  result = _spawnvp (_P_WAIT, javaArgs[0], javaArgs);
  if (result == -1 && errno != 0) {
    result = _spawnvp (_P_WAIT, jviewArgs[0], jviewArgs);

    if (result == -1 && errno != 0) {
      char *msg = strerror (errno);
      char *buf = malloc (20 + strlen (javaArgs[0]) + strlen (msg));
      sprintf (buf, "can't run %s or %s: %s", javaArgs[0], jviewArgs[0], msg);
#ifdef CONSOLE
      fprintf (stderr, "%s\n", buf);
#else
      MessageBox (NULL, buf, "Error", MB_OK | MB_ICONEXCLAMATION);
#endif
    }
  }
  return result;
}

int main (int argc, char **argv)
{
  char **javaArgs;
  char **jviewArgs;
  STARTUPINFO sinfo;
  PROCESS_INFORMATION pinfo;

  // Append CLASSPATH and PATH compile-time symbols
  // to the appropriate environment variables.
  ExtendPath ("CLASSPATH", classpathEntries);
  ExtendPath ("CLASSPATH", FindFiles (JAR_FILES));
  ExtendPath ("PATH", pathEntries);

#if DEBUG
  fprintf (stderr, "CLASSPATH=%s", getenv ("CLASSPATH"));
  fprintf (stderr, "PATH=%s", getenv ("PATH"));
#endif

  // Set up Java command line.
  javaArgs = MakeArgs (javaExe, fixedJavaArgs, argv+1);
  jviewArgs = MakeArgs (jviewExe, fixedJviewArgs, argv+1);

  // Invoke Java.
  return RunJava (javaArgs, jviewArgs);
}

#ifndef CONSOLE
//__declspec(dllimport) char **__initenv;

int WINAPI WinMain (HINSTANCE inst, HINSTANCE previnst, LPSTR cmdline, int cmdshow)
{
  //__initenv = _environ;
    return main (__argc, __argv);
}
#endif

